
#ifndef __HSCAL_BLACS_BIMPIBCAST_HSCALFUNC_HH
#define __HSCAL_BLACS_BIMPIBCAST_HSCALFUNC_HH

/*
*  The implementation here depends on the MPI package
*  Currently using only the simple method
*/

#include "../../../../include/scalapack.h"

void hscal_BI_MPI_Bcast_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, int root,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
   int src, dst;
   int relative_rank, mask;
   int rank = ctxt->scp->Iam;
   int size = ctxt->scp->Np;
   char scope = ctxt->scp->scope;

   if (N == 0)
   {
      return;
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "BCAST (%d, %d): N %d root %d comm size %d scope %c", 
        myrow, mycol, N, root, size, scope
   );

   relative_rank = (rank >= root) ? rank - root : rank - root + size;

   mask = 0x1;

   while (mask < size) {
      if (relative_rank & mask) {
         src = rank - mask;
         if (src < 0) src += size;
         /* Receive from src */
         break;
      }
      mask <<= 1;
   }

   mask >>= 1;
   while (mask > 0) {
      if (relative_rank + mask < size) {
         dst = rank + mask;
         if (dst >= size) dst -= size;
         /* send to dst */

         if (tcomm != NULL)
         {
            int idest, jdest;

            switch (scope)
            {
            case 'c':
            {
                HSCAL_DEBUG(
                     __FILE__, __FUNCTION__,__LINE__, 
                     "BCAST (%d, %d)->(%d, %d)\n", myrow, mycol, dst, mycol
                );
                TCOMM(myrow, mycol, dst, mycol, nprow, npcol) += N;
                break;
            }
            case 'r':
            {
                HSCAL_DEBUG(
                     __FILE__, __FUNCTION__,__LINE__, 
                     "BCAST (%d, %d)->(%d, %d)\n", myrow, mycol, myrow, dst
                );
                TCOMM(myrow, mycol, myrow, dst, nprow, npcol) += N;
                break;
            }
            case 'a':
            {
                idest = dst/npcol;
                jdest = dst%npcol;
                HSCAL_DEBUG(
                     __FILE__, __FUNCTION__,__LINE__, 
                     "BCAST (%d, %d)->(%d, %d)\n", myrow, mycol, idest, jdest
                );
                TCOMM(myrow, mycol, idest, jdest, nprow, npcol) += N;
                break;
            }
            default:
                break;
            }
         }
      }
      mask >>= 1;
   }

   return;
}

#endif

