
#ifndef __HSCAL_BLACS_BIMPATHBR_HSCALFUNC_HH
#define __HSCAL_BLACS_BIMPATHBR_HSCALFUNC_HH

#include "../../../../include/scalapack.h"

void hscal_BI_MpathBR_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, HSCALSDPTR send, int src, int npaths,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
   int pathlen;		/* the minimal length of each path */
   int mydist;		/* my distance from src */
   int faredge;		/* node at far end of path */
   int lastlong;	/* distance to node on end of last path with extra node */
   int Np, Iam, Np_1, dest;
   char scope = ctxt->scp->scope;

   if (N == 0)
   {
      return;
   }

   HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "N %d src %d npaths %d", N, src, npaths);

   Np = ctxt->scp->Np;
   Iam = ctxt->scp->Iam;
   Np_1 = Np - 1;
   if (npaths == FULLCON) npaths = Np_1;

   if (npaths > 0)
   {
      dest = (Iam+1) % Np;
      mydist = (Np + Iam - src) % Np;
   }
   else
   {
      dest = (Np_1+Iam) % Np;
      mydist = (Np + src - Iam) % Np;
      npaths = -npaths;
   }
/*
 * Make sure npaths is cool
 */
   if (npaths > Np_1) npaths = Np_1;

   pathlen = Np_1 / npaths;
   lastlong = (Np_1%npaths) * (pathlen+1);
   if (lastlong)
   {
      if (mydist <= lastlong) faredge = ((mydist-1)/(pathlen+1)+1)*(pathlen+1);
      else faredge = ((lastlong-1)/(pathlen+1)+1) * (pathlen+1)
		     + ((mydist-lastlong-1)/pathlen + 1) * pathlen;
   }
   else faredge = ((mydist-1)/pathlen + 1) * pathlen;

   if (mydist < faredge) {
      if (tcomm != NULL)
      {
         int idest, jdest;
         switch (scope)
         {
         case 'c':
         {      
             send(
               N, dest, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'r':
         {
             send(
               N, myrow, dest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'a':
         {
             idest = dest/npcol;
             jdest = dest%npcol;
             send(
               N, idest, jdest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         default:
             break;
         }
      }
   }
}

#endif
