
#ifndef __HSCAL_BLACS_BIMPATHBS_HSCALFUNC_HH
#define __HSCAL_BLACS_BIMPATHBS_HSCALFUNC_HH

#include "../../../../include/scalapack.h"

void hscal_BI_MpathBS_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, HSCALSDPTR send, int npaths,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
   int pathlen;		/* the length of each path */
   int dist;	        /* the distance to the node closest to src on each path */
   int pdest;           /* part of dest calculation -- saves unneeded ops */
   int lastlong;	/* number of paths with extra node */
   int Np, Iam, Np_1, dir;
   char scope = ctxt->scp->scope;

   if (N == 0)
   {
      return;
   }

   HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "N %d npaths %d", N, npaths);

   Np = ctxt->scp->Np;
   if (Np < 2) return;
   Iam = ctxt->scp->Iam;
   Np_1 = Np - 1;
   if (npaths == FULLCON) npaths = Np_1;

   if (npaths > 0)  /* paths are increasing rings */
   {
      pdest = Iam;
      dir = 1;
   }
   else             /* paths are decreasing rings */
   {
      pdest = Np + Iam;
      dir = -1;
      npaths = -npaths;
   }
/*
 * Ensure npaths is correct
 */
   if (npaths > Np_1) npaths = Np_1;
   pathlen = Np_1 / npaths;

/*
 * Loop over all long paths (paths with an extra node), if there are any
 */
   lastlong = (Np_1 % npaths) * (pathlen+1);  /* last node in long ring */
   for (dist=1; dist < lastlong; dist += pathlen+1) {
      int dest = (pdest+dir*dist)%Np;
      if (tcomm != NULL)
      {
         int idest, jdest;
         switch (scope)
         {
         case 'c':
         {      
             send(
               N, dest, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'r':
         {
             send(
               N, myrow, dest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'a':
         {
             idest = dest/npcol;
             jdest = dest%npcol;
             send(
               N, idest, jdest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         default:
             break;
         }
      }
   }

/*
 * Loop over all normal length paths
 */
   while (dist < Np)
   {
      int dest = (pdest+dir*dist)%Np;
      if (tcomm != NULL)
      {
         int idest, jdest;
         switch (scope)
         {
         case 'c':
         {      
             send(
               N, dest, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'r':
         {
             send(
               N, myrow, dest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'a':
         {
             idest = dest/npcol;
             jdest = dest%npcol;
             send(
               N, idest, jdest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         default:
             break;
         }
      }
      dist += pathlen;
   }
}

#endif
