
#ifndef __HSCAL_BLACS_BIMRINGCOMB_HSCALFUNC_HH
#define __HSCAL_BLACS_BIMRINGCOMB_HSCALFUNC_HH

#include "../../../../include/scalapack.h"

void hscal_BI_MringComb_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, HSCALVVFUNPTR Xvvop,
     int dest, int nrings,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
   int Np, Iam, i, inc, mysrc, mydest, Np_1;
   int mydist, ringlen, myring;
   int nearedge, faredge;  /* edge closest and farthest from dest */
   int REBS;               /* Is result leave-on-all? */
   char scope = ctxt->scp->scope;

   if (N == 0)
   {
      return;
   }

   HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "N %d dest %d nrings %d", N, dest, nrings);

   Np = ctxt->scp->Np;
   if (Np < 2) return;
   Iam = ctxt->scp->Iam;
   if (REBS = (dest == -1)) dest = 0;

   if (nrings > 0)
   {
      mydist = (Np + dest - Iam) % Np;
      inc = 1;
   }
   else
   {
      mydist = (Np + Iam - dest) % Np;
      inc = -1;
      nrings = -nrings;
   }
   Np_1 = Np - 1;
   if (nrings > Np_1) nrings = Np_1;

/*
 * If I'm not the destination
 */
   if (Iam != dest)
   {
      ringlen = Np_1 / nrings;
      myring = (mydist-1) / ringlen;
      if (myring >= nrings) myring = nrings - 1;
      nearedge = (myring*ringlen) + 1;
      faredge = nearedge + ringlen - 1;
      if (myring == nrings-1) faredge += Np_1 % nrings;
      if (mydist == nearedge) mydest = dest;
      else mydest = (Np + Iam + inc) % Np;
      if (mydist != faredge)
      {
         Xvvop(
           NC,
           nprow, npcol, myrow, mycol, tcomp
         );
      }
      if (tcomm != NULL)
      {
         int idest, jdest;
         switch (scope)
         {
         case 'c':
         {
             TCOMM(myrow, mycol, mydest, mycol, nprow, npcol) += N;
             break;
         }
         case 'r':
         {
             TCOMM(myrow, mycol, myrow, mydest, nprow, npcol) += N;
             break;
         }
         case 'a':
         {
             idest = mydest/npcol;
             jdest = mydest%npcol;
             TCOMM(myrow, mycol, idest, jdest, nprow, npcol) += N;
             break;
         }
         default:
             break;
         }
      }
      if (REBS) hscal_BI_MpathBR_nocomm(
                  ctxt, NC, N, hscal_BI_MPI_Send_nocomm, dest, nrings,
                  nprow, npcol, myrow, mycol, tcomp, tcomm
                );
   }
/*
 * If I'm the destination process
 */
   else
   {
      if (!ctxt->TopsRepeat)
      {
         for(i=nrings; i; i--)
         {
            Xvvop(
              NC,
              nprow, npcol, myrow, mycol, tcomp
            );
         }
      }
      else
      {
         ringlen = Np_1 / nrings;
         if (inc == 1) mysrc = (Np + Iam - 1) % Np;
         else mysrc = (Iam + 1) % Np;
         for(i=nrings; i; i--)
         {
            Xvvop(
              NC,
              nprow, npcol, myrow, mycol, tcomp
            );
            if (inc == 1) mysrc = (Np + mysrc - ringlen) % Np;
            else mysrc = (mysrc + ringlen) % Np;
         }
      }
      if (REBS) hscal_BI_MpathBS_nocomm(
                  ctxt, NC, N, hscal_BI_MPI_Send_nocomm, nrings,
                  nprow, npcol, myrow, mycol, tcomp, tcomm
                );
   }
}

#endif
