
#ifndef __HSCAL_BLACS_BIMPIREDUCE_HSCALFUNC_HH
#define __HSCAL_BLACS_BIMPIREDUCE_HSCALFUNC_HH

/*
*  The implementation here depends on the MPI package
*  Currently using only the simple method
*/
#include "../../../../include/scalapack.h"

void hscal_BI_MPI_Reduce_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, HSCALVVFUNPTR Xvvop, int root, 
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
     int mask, relrank, source, lroot;
     int rank = ctxt->scp->Iam;
     int size = ctxt->scp->Np;
     char scope = ctxt->scp->scope;

     if (N == 0)
     {
        return;
     }

     HSCAL_DEBUG(
          __FILE__, __FUNCTION__,__LINE__, 
          "REDUCE (%d, %d): N %d root %d comm size %d scope %c", 
          myrow, mycol, N, root, size, scope
     );

     lroot = root;
     mask = 0x1;
     relrank = (rank - lroot + size) % size;

     while (mask < size) {
        /* Receive */
        if ((mask & relrank) == 0) {
            source = (relrank | mask);
            if (source < size) {
                source = (source + lroot) % size;
                Xvvop(
                  NC, nprow, npcol, myrow, mycol, tcomp
                );
            }
        }
        else
        {
            /* I've received all that I'm going to.  Send my result to
               my parent */
            source = ((relrank & (~ mask)) + lroot) % size;

            if (tcomm != NULL)
            {
               int idest, jdest;

               switch (scope)
               {
               case 'c':
               {
                HSCAL_DEBUG(
                     __FILE__, __FUNCTION__,__LINE__, 
                     "REDUCE (%d, %d)->(%d, %d)\n", myrow, mycol, root, mycol
                );
                   TCOMM(myrow, mycol, root, mycol, nprow, npcol) += N;
                   break;
               }
               case 'r':
               {
                HSCAL_DEBUG(
                     __FILE__, __FUNCTION__,__LINE__, 
                     "REDUCE (%d, %d)->(%d, %d)\n", myrow, mycol, myrow, root
                );
                   TCOMM(myrow, mycol, myrow, root, nprow, npcol) += N;
                   break;
               }
               case 'a':
               {
                   idest = root/npcol;
                   jdest = root%npcol;
                HSCAL_DEBUG(
                     __FILE__, __FUNCTION__,__LINE__, 
                     "REDUCE (%d, %d)->(%d, %d)\n", myrow, mycol, idest, jdest
                );
                   TCOMM(myrow, mycol, idest, jdest, nprow, npcol) += N;
                   break;
               }
               default:
                   break;
               }
            }            

            break;
        }
        mask <<= 1;
     }

     return;
} 

#endif
