
#ifndef __HSCAL_BLACS_BISRINGBR_HSCALFUNC_HH
#define __HSCAL_BLACS_BISRINGBR_HSCALFUNC_HH

#include "../../../../include/scalapack.h"

void hscal_BI_SringBR_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, HSCALSDPTR send, int src,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
   int mydist;  	/* my distance from source */
   int Np, Iam, rightedge;
   char scope = ctxt->scp->scope;

   if (N == 0)
   {
      return;
   }

   HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "N %d", N);

   Np = ctxt->scp->Np;
   Iam = ctxt->scp->Iam;

   mydist = (Np + Iam - src) % Np;
   rightedge = Np/2;

/*
 * If I'm between source & right edge of split ring, send to right
 */
   if (mydist < rightedge)
   {
      int dest = (Iam + 1)%Np;
      if (tcomm != NULL)
      {
         int idest, jdest;
         switch (scope)
         {
         case 'c':
         {      
             send(
               N, dest, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'r':
         {
             send(
               N, myrow, dest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'a':
         {
             idest = dest/npcol;
             jdest = dest%npcol;
             send(
               N, idest, jdest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         default:
             break;
         }
      }
   }

/*
 * If I'm between source and left edge of split ring, send to left
 */
   else if (mydist > rightedge+1)
   {
      int dest = (Np + Iam - 1)%Np;
      if (tcomm != NULL)
      {
         int idest, jdest;
         switch (scope)
         {
         case 'c':
         {      
             send(
               N, dest, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'r':
         {
             send(
               N, myrow, dest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         case 'a':
         {
             idest = dest/npcol;
             jdest = dest%npcol;
             send(
               N, idest, jdest,
               nprow, npcol, myrow, mycol, tcomp, tcomm
             );
             break;
         }
         default:
             break;
         }
      }
   }
}

#endif
