
#ifndef __HSCAL_BLACS_BITREEBR_HSCALFUNC_HH
#define __HSCAL_BLACS_BITREEBR_HSCALFUNC_HH

#include "../../../../include/scalapack.h"

void hscal_BI_TreeBR_nocomm(
     HSCALBLACSCONTEXT *ctxt, int NC, int N, HSCALSDPTR send, int src, int nbranches,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
   int Np, Iam, i, j;
   int mydist;          /* my distance from src */
   int destdist;	/* the distance of the destination node */
   char scope = ctxt->scp->scope;

   if (N == 0)
   {
      return;
   }

   HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "N %d nbranches %d", N, nbranches);

   Np = ctxt->scp->Np;
   if (Np < 2) return;
   Iam = ctxt->scp->Iam;
   mydist = (Np + Iam - src) % Np;

/*
 * Go up to first step of tree where I send data to other nodes
 */
   for (i=nbranches; i < Np; i *= nbranches);
   for (i /= nbranches; (mydist%i); i /= nbranches);

/*
 * While I need to send data to others
 */
   while ( (i > 1) && !(mydist%i) )
   {
      i /= nbranches;
      j = 1;
      do
      {
	 destdist = mydist + j*i;
	 if (destdist < Np)
         {
            int dest = (src+destdist)%Np;

            if (tcomm != NULL)
            {
               int idest, jdest;

               switch (scope)
               {
               case 'c':
               {      
                   send(
                     N, dest, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                   );
                   break;
               }
               case 'r':
               {
                   send(
                     N, myrow, dest,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                   );
                   break;
               }
               case 'a':
               {
                   idest = dest/npcol;
                   jdest = dest%npcol;
                   send(
                     N, idest, jdest,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                   );
                   break;
               }
               default:
                   break;
               }
            }
         }
      }
      while(++j < nbranches);
   }

} /* end BI_TreeBR */

#endif
