
#ifndef __HSCAL_BLACS_CCTRSD2D_HSCALFUNC_HH
#define __HSCAL_BLACS_CCTRSD2D_HSCALFUNC_HH

#include "../../../include/scalapack.h"

void hscal_Cctrsd2d(
     HSCALBLACSCONTEXT *ctxt, 
     char *uplo, char *diag,
     int m, int n, int lda,
     int rdest, int cdest,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
/*
 *  -- V1.1 BLACS routine --
 *  University of Tennessee, May 1, 1996
 *  Written by Clint Whaley.
 *
 *  Purpose
 *  =======
 *  Locally-blocking point-to-point trapezoidal complex send.
 *
 *  Arguments
 *  =========
 *
 *  ConTxt  (input) Ptr to int
 *          Index into MyConTxts00 (my contexts array).
 *
 *  UPLO    (input) Ptr to char
 *          Specifies the part of the matrix to be sent.
 *          = 'U':      Upper trapezoidal part
 *          ELSE :      Lower trapezoidal part
 *
 *  DIAG    (input) Ptr to char
 *          Specifies whether the matrix is unit diagonal or not.
 *          = 'U':      Matrix is unit diagonal, diagonal not communicated.
 *          ELSE :      Matrix is not unit diagonal, diagonal is communicated.
 *
 *  M       (input) Ptr to int
 *          The number of rows of the matrix A.  M >= 0.
 *
 *  N       (input) Ptr to int
 *          The number of columns of the matrix A.  N >= 0.
 *
 *  LDA     (input) Ptr to int
 *          The leading dimension of the array A.  LDA >= M.
 *
 *  RDEST   (input) Ptr to int
 *          The process row of the destination process.
 *
 *  CDEST   (input) Ptr to int
 *          The process column of the destination process.
 *
 * ------------------------------------------------------------------------
 */
{
   char tuplo, tdiag;
   int tlda;
   int N, lth;

   tuplo = F2C_CharTrans(uplo);
   tdiag = F2C_CharTrans(diag);
   tuplo = Mlowcase(tuplo);
   tdiag = Mlowcase(tdiag);

   if (Mpval(lda) < Mpval(m)) tlda = Mpval(m);
   else tlda = Mpval(lda);

   hscal_BI_GetMpiTrType(tuplo, tdiag, Mpval(m), Mpval(n), tlda, &N);

   lth = N * sizeof(SCOMPLEX);

   hscal_BI_MPI_Send_nocomm(
     lth, Mpval(rdest), Mpval(cdest),
     nprow, npcol, myrow, mycol, tcomp, tcomm
   );
}  /* end of ctrsd2d */

#endif
