
#ifndef __HSCAL_BLACS_CSGESD2D_HSCALFUNC_HH
#define __HSCAL_BLACS_CSGESD2D_HSCALFUNC_HH

#include "../../../include/scalapack.h"

void hscal_Csgesd2d(
     HSCALBLACSCONTEXT *ctxt, 
     int m, int n, int lda,
     int rdest, int cdest,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
/*
 *  -- V1.1 BLACS routine --
 *  University of Tennessee, May 1, 1996
 *  Written by Clint Whaley.
 *
 *  Purpose
 *  =======
 *  Locally-blocking point-to-point general real send.
 *
 *  Arguments
 *  =========
 *
 *  ConTxt  (input) Ptr to int
 *          Index into MyConTxts00 (my contexts array).
 *
 *  M       (input) Ptr to int
 *          The number of rows of the matrix A.  M >= 0.
 *
 *  N       (input) Ptr to int
 *          The number of columns of the matrix A.  N >= 0.
 *
 *  A       (input) Ptr to real two dimensional array
 *          The m by n matrix A.  Fortran77 (column-major) storage
 *          assumed.
 *
 *  LDA     (input) Ptr to int
 *          The leading dimension of the array A.  LDA >= M.
 *
 *  RDEST   (input) Ptr to int
 *          The process row of the destination process.
 *
 *  CDEST   (input) Ptr to int
 *          The process column of the destination process.
 *
 * ------------------------------------------------------------------------
 */
{
   int N, lgth;

   ctxt->scp = &ctxt->pscp;
   N = Mpval(m) * Mpval(n);
   lgth = N * sizeof(float);
   hscal_BI_MPI_Send_nocomm(
     lgth, Mpval(rdest), Mpval(cdest),
     nprow, npcol, myrow, mycol, tcomp, tcomm
   );
}

#endif

