
#ifndef __HSCAL_BLACS_CSTRBR2D_HSCALFUNC_HH
#define __HSCAL_BLACS_CSTRBR2D_HSCALFUNC_HH

#include "../../../include/scalapack.h"

void hscal_Cstrbr2d(
     HSCALBLACSCONTEXT *ctxt, char *scope, char *top, char *uplo, char *diag,
     int m, int n, int lda, int rsrc, int csrc,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm 
)
{
   int N, lth, src, tlda, error, one=1;
   char ttop, tscope, tuplo, tdiag;

 /*
  * Fill up the uninitialized fields
  */
   ctxt->rscp.Iam = mycol;
   ctxt->cscp.Iam = myrow;
   ctxt->pscp.Iam = ctxt->ascp.Iam = myrow*npcol+mycol;

   ttop = F2C_CharTrans(top);
   ttop = Mlowcase(ttop);
   tscope = F2C_CharTrans(scope);
   tscope = Mlowcase(tscope);
   tdiag = F2C_CharTrans(diag);
   tdiag = Mlowcase(tdiag);
   tuplo = F2C_CharTrans(uplo);
   tuplo = Mlowcase(tuplo);

/*
 *  If the user has set the default broadcast topology, use it instead of
 *  BLACS default
 */
   if (ttop == ' ') ttop = DefBSTop;
   if (Mpval(m) <= Mpval(lda)) tlda = Mpval(lda);
   else tlda = Mpval(m);

   switch(tscope)
   {
   case 'r':
      ctxt->scp = &ctxt->rscp;
      src = Mpval(csrc);
      break;
   case 'c':
      ctxt->scp = &ctxt->cscp;
      src = Mpval(rsrc);
      break;
   case 'a':
      ctxt->scp = &ctxt->ascp;
      src = Mvkpnum(ctxt, Mpval(rsrc), Mpval(csrc));
      break;
   default:
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown scope '%c'",
                  tscope);
      return;
   }

   hscal_BI_GetMpiTrType(
        tuplo, tdiag, Mpval(m), Mpval(n), tlda,
        &N
   );

   lth = N * sizeof(float);

   switch(ttop)
   {
   case ' ':         /* If using default topology, use MPI native broadcast */
      hscal_BI_MPI_Bcast_nocomm(
        ctxt, N, lth, src,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'h':
   {
      hscal_BI_HypBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src,
        &error,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      if (error == NPOW2) 
      {
          hscal_BI_TreeBR_nocomm(
               ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, 2,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );
      }
      break;
   }
   case '1':
   case '2':
   case '3':
   case '4':
   case '5':
   case '6':
   case '7':
   case '8':
   case '9':
      hscal_BI_TreeBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, ttop-47,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 't':
      hscal_BI_TreeBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, ctxt->Nb_bs,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'i':
      hscal_BI_IdringBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, 1,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'd':
      hscal_BI_IdringBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, -1,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 's':
      hscal_BI_SringBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'm':
      hscal_BI_MpathBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, ctxt->Nr_bs,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   case 'f':
   {
      int npaths = FULLCON;
      hscal_BI_MpathBR_nocomm(
        ctxt, N, lth, hscal_BI_MPI_Send_nocomm, src, npaths,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      break;
   }
   default :
      BI_BlacsErr_nocomm(ctxt, __LINE__, __FILE__, "Unknown topology '%c'",
                  ttop);
      return;
   }
}

#endif
