/*
 * $Id: BEprint.c,v 1.7 2005/03/01 13:48:24 posypkin Exp $
 */


#include "be_print.h"

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

/**
 * Opens C files for code generation.
 *
 * @param name
 *
 * @return opened file handler
 *
 */
void report_error(char *fmt, ...);
FILE* open_c_file(char *name)
{
    int size;
    char* c_filename;
    char *p;
    FILE* rf;

    c_filename = (char*) malloc(MAX_IDENT_LEN);
    p = strrchr(name, '.');
    size = p ? (p - name) : strlen(name);
    strncpy(c_filename, name, size);
    strcpy(c_filename + size, ".c");

    if (!strcmp(output, "stdout")) 
	rf = stdout; 
    else
	rf = fopen(c_filename, "w");

    if(rf == NULL) 
	report_error("cannot open file [%s]\n", c_filename);
	
    file_names[pass_number - 1] = c_filename;
    return rf;
}

/* NAMES:    blank
   PURPOSE:  Type n space into file
   SYNOPSYS: file
             n
   RETURNS:  void
*/

void blank(int n)
{
    int i;
    for (i = 0; i < n; i++)
        print2file("", " ");
    cur_pos = n;
}

void delete_c_file()
{
    if (strcmp(output, "stdout")) {
        int i;
        for (i = 0; i < pass_number; i++) {
            remove(file_names[i]);
        }
    }
}

/* NAMES:    internal_error
   PURPOSE:  Writes into stderr message about internal error
   SYNOPSYS: void
   RETURNS:  void
*/

void internal_error()
{
    fprintf(stderr, "---------- mpC compiler internal error ----------\n");
    fprintf(stderr, "  source: %s line: %d\n",
            SourceFile(Position), RealPosition(Position).Line);
    fprintf(stderr, "  It is case to contact with mpC compiler group  \n");
    /*  fprintf(stderr,"-- phone              , fax             --\n"); */
    fprintf(stderr, "               e-mail mpc@ispras.ru              \n");
    fprintf(stderr, "-------------------------------------------------\n");
    delete_c_file();
    if(query_tree)
        *(int*) NULL = 1;
    exit(1);
}

void be_error(char *mes)
{
    fprintf(stderr, "error in %s: %d,%d -> %s\n",
            SourceFile(Position), RealPosition(Position).Line,
            RealPosition(Position).Column, mes);
    delete_c_file();
    exit(1);
}


int dummy_file()
{
    int ret = 0;
    if (pass_number > 1 && (host_stat || host_expr))
        ret = 1;
    return ret;
}

int print2file(char *debug_message, char *fmt, ...)
{
    int cnt;
    va_list ap;

    if (debug_message && strcmp(debug_message, "")) {
        fprintf(file, "%s", debug_message);
    }
    va_start(ap, fmt);
    cnt = vfprintf(file, fmt, ap);
    fflush(file);
    va_end(ap);
    return cnt;
}

void report_error(char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(stderr, fmt, ap);
    va_end(ap);
    internal_error();
}

void print_set_position(tPosition inPos)
{
    c_blank;
    print2file("", MPC_SET_POSITION "(%d,%d);\n",
                  RealPosition(inPos).Column, 
                  RealPosition(inPos).Line);    
}

















