#include "be_print.h"
#include "Tree.h"
#include "expr.h"

static int vector_rank(tTree inExpr);
static void print_lbound(tTree typ);

/**
 * Prints the scalar projection of a given expression.
 *
 * @param pExpr processing expression
 */
void print_sexpr(tTree pExpr)
{

    if (pExpr->MPC_Expr.Optm_Attribute.slot
        && pExpr->MPC_Expr.Optm_Attribute.generated) {
        print2file("", "%s%d", TMP_SLOT_PREFIX,
                   pExpr->MPC_Expr.Optm_Attribute.slot);
    } else {
        if (pExpr->MPC_Expr.Flag.InParentheses)
            cur_pos += print2file("", "(");

        switch (pExpr->Kind) {
        case kMPC_Const:
        case kMPC_IntConst:
        case kMPC_UIntConst:
        case kMPC_FloatConst:
            print2file("", "%s", pExpr->MPC_Const.SymbolicForm);
            break;
        case kMPC_StringLiteral:
            {
                tTree pNext = pExpr->MPC_StringLiteral.Next;

                print2file("", "%s", pExpr->MPC_StringLiteral.String);
                if (pNext != NoTree && pNext->Kind != kMPC_FreeNode) {
                    print_sexpr(pNext);
                }
            }
            break;
        case kMPC_Ident:
            {
                tTree pType = pExpr->MPC_Ident.Type;
                tTree pVar;
                PointerControl(pType);
                pVar = pExpr->MPC_Ident.Store;
                if (pVar != NoTree && pVar->Kind == kMPC_Var) {
                    tTree pRealExpr;

                    pRealExpr = real_net_type_param(pVar);
                    if (pRealExpr != NoTree) {
                        if (pRealExpr->Kind != kMPC_Var) {
                            print_sexpr(pRealExpr);
                        } else {
                            cur_pos +=
                                print2file("", "%s", VarName(pRealExpr));
                        }
                    } else {
                        print2file("", "%s", VarName(pVar));
                    }
                } else {
                    static char nameBuf[MAX_IDENT_LEN];
                    tIdent Ident = pExpr->MPC_Ident.Ident;

                    IdentControl(Ident);
                    GetString(Ident, nameBuf);
                    print2file("", "%s", nameBuf);
                }
            }
            break;
        case kMPC_CastExpr:
            {
                tTree pType = pExpr->MPC_CastExpr.Type;

                PointerControl(pType);
                print_Type_in_cast(pType);
                print_sexpr(pExpr->MPC_CastExpr.Operand);
            }
            break;
        case kMPC_NetCastExpr:
            print_sexpr(pExpr->MPC_NetCastExpr.Operand);
            /*
               {
               tTree pStoreNet = pExpr->MPC_Expr.StoreNet;

               PointerControl(pStoreNet);
               if (is_NetCast_asynch(pExpr)) {
               tTree pOperand = pExpr->MPC_NetCastExpr.Operand;

               PointerControl(pOperand);
               {
               tTree prev_act_net = cur_act_net;

               {
               if (flag.Asynchr) {
               pOperand->MPC_Expr.ActualEvalNet =
               pOperand->MPC_Expr.EvalNet;
               not_change_net = 1;
               print_sexpr(pOperand);
               } else {
               fill_ActualEvalNet(pOperand);
               cur_act_net = not_change_net ?
               prev_act_net : pOperand->MPC_Expr.
               ActualEvalNet;
               PointerControl(cur_act_net);
               print_sexpr(pOperand);
               cur_act_net = prev_act_net;
               }
               }
               }
               } else {
               fprintf(stderr,
               "Sorry, not bounded NetCast isn't supported\n");
               internal_error();
               }
               }
             */
            break;
        case kMPC_CoordExpr:
            {
                tTree pENet;

                if (net_to_change == NoTree)
                    pENet = SingleRegion(pExpr->MPC_CoordExpr.EvalNet);
                else
                    pENet = net_to_change;

                while (pENet->Kind == kMPC_Subnet &&
                       pENet->MPC_Subnet.Flexible) {
                    pENet = pENet->MPC_Subnet.Distribution;
                    PointerControl(pENet);
                }
                print2file("", "%s.coord[%d]", VarName(pENet),
                           pExpr->MPC_CoordExpr.CoordNumber);
            }
            break;
        case kMPC_Size_Of_Value:
        case kMPC_Size_Of_Type:
            {
                int CompileTime = pExpr->MPC_SizeofExpr.CompileTime;

                if (CompileTime) {
                    print2file("", "%d", pExpr->MPC_SizeofExpr.Value);
                } else if (pExpr->Kind == kMPC_Size_Of_Value) {
                    tTree pOperand;
                    tTree pType;

                    pOperand = pExpr->MPC_Size_Of_Value.Operand;
                    pType = pOperand->MPC_Expr.Type;
                    PointerControl(pOperand);
                    PointerControl(pType);
                    if (IS_DYN_ARRAY_OR_VECTOR(pType))
                        print2file("", "%s.Base.Size",
                                   DataVarName(pOperand->MPC_Expr.Type));
                    else if (pOperand->MPC_Expr.Flag.Peculiarity) {
                        if ((pOperand->MPC_Expr.Type->Kind ==
                             kMPC_ArrayType)) {
                            print2file("", "(");
                            while (pType->Kind == kMPC_ArrayType) {
                                if (STEP(pType) == 1)
                                    print2file("", "%d*",
                                               NUMBER_OF_COMPONENTS
                                               (pType));
                                else
                                    print2file("", "(%d*%d+1)*",
                                               NUMBER_OF_COMPONENTS(pType)
                                               - 1, STEP(pType));
                                pType = element_type(pType);
                                PointerControl(pType);
                            }
                            print2file("", "sizeof(");
                            print_Type_in_sizeof(pType);
                            print2file("", "))");
                        } else if (pType->Kind == kMPC_VectorType) {
                            print2file("", "(");
                            while (pType->Kind == kMPC_VectorType) {
                                print2file("", "%d*",
                                           NUMBER_OF_COMPONENTS(pType));
                                pType = element_type(pType);
                                PointerControl(pType);
                            }
                            print2file("", "sizeof(");
                            print_Type_in_sizeof(pType);
                            print2file("", "))");
                        } else if (pType->Kind == kMPC_PointerType) {
                            print2file("", "sizeof(void*)");
                        } else {
                            print2file("", "sizeof(");
                            print_Type_in_sizeof(pType);
                            print2file("", ")");
                        }
                    } else {
                        print2file("", "sizeof(");
                        print_sexpr(pOperand);
                        print2file("", ")");
                    }

                } else {
                    tTree pOperand;
                    pOperand = pExpr->MPC_Size_Of_Type.Operand;
                    PointerControl(pOperand);
                    if (IS_DYN_ARRAY_OR_VECTOR(pOperand))
                        print2file("", "%s.Base.Size",
                                   DataVarName(pOperand));
                    else if (is_dynamic_type(pOperand) == DYNAMIC_POINTER) {
                        print2file("", "sizeof(void*)");
                    } else {
                        print2file("", "sizeof(");
                        print_Type_in_sizeof(pOperand);
                        print2file("", ")");
                    }
                }
            }
            break;
        case kMPC_UnaryExpr:
            {
                tTree pOperand;
                tTree pType = pExpr->MPC_UnaryExpr.Type;
                int OpCode;
                char *ope[2] = { "+=", "-=" };
                int old_print_base_type;

                PointerControl(pType);
                pOperand = pExpr->MPC_UnaryExpr.Operand;
                PointerControl(pOperand);
                OpCode = pExpr->MPC_UnaryExpr.OpCode;
                if (OpCode == POST_INC || OpCode == POST_DEC) {
                    if (is_not_c_pointer(pType)) {
                        old_print_base_type = print_base_type;
                        print_base_type = 1;
                        print2file("", "(MPC_tmp_pp=(void**)&(");
                        print_sexpr(pOperand);
                        print2file("", "), MPC_tmp_pv=*MPC_tmp_pp,((");
                        print_Type(pType);
                        print2file("", ")*MPC_tmp_pp)%s%s,(",
                                   ope[OpCode - POST_INC],
                                   step_macro(pType));
                        print_Type(pType);
                        print2file("", ")MPC_tmp_pv)");
                        print_base_type = old_print_base_type;

                    } else {
                        print_sexpr(pOperand);
                        print2file("", "%s", op_represent[OpCode]);
                    }
                } else if (OpCode == PRE_INC || OpCode == PRE_DEC) {
                    if (is_not_c_pointer(pType)) {
                        print2file("", "(");
                        print_sexpr(pOperand);
                        print2file("", "%s%s)", ope[OpCode - PRE_INC],
                                   step_macro(pType));
                    } else {
                        print2file("", "%s", op_represent[OpCode]);
                        print_sexpr(pOperand);
                    }
                } else if (OpCode == STAR) {
                    if (!is_pointer_on_dynamic(pOperand->MPC_Expr.Type))
                        print2file("", "%s", op_represent[OpCode]);
                    print_sexpr(pOperand);
                } else if (OpCode == ADDRESS) {
                    if (is_dynamic_type(pOperand->MPC_Expr.Type) ==
                        DYNAMIC_ARRAY) {
                        if ((pOperand->MPC_Expr.Flag.Lvalue != 0)
                            &&
                            (!(is_dynamic_type
                               (element_type(pOperand->MPC_Expr.Type)) ==
                               DYNAMIC_ARRAY)))
                            print2file("", "%s", op_represent[OpCode]);
                    } else
                        print2file("", "%s", op_represent[OpCode]);
                    print_sexpr(pOperand);
                } else if ((OpCode >= ADDRESS && OpCode <= LOG_NOT) ||
                           OpCode == PLUS) {
                    print2file("", "%s", op_represent[OpCode]);
                    print_sexpr(pOperand);
                } else if (OpCode == BRACKETS) {
                    tTree pType = pExpr->MPC_Expr.Type;
                    PointerControl(pType);
                    if (global_index_level == MPC_NIL_INDEX)
                        print_sexpr(pOperand);
                    else {
                        if ((global_index_level - global_shift) == 1)
                            print_sexpr(pOperand);
                        else
                            print2file("", "%s%d", TMP_SLOT_PREFIX,
                                       pExpr->MPC_Expr.Optm_Attribute.
                                       first_slot + global_index_level -
                                       global_shift - 2);

                        print_index(pType, 0, global_shift);
                    }


                } else if (OpCode > LINEAR) {
                    print_sexpr(pOperand);
                } else {
                    fprintf(stderr,
                            "invalid op code of the unary expression = %d\n",
                            OpCode);
                    internal_error();
                }
            }
            break;
        case kMPC_BinaryExpr:
            {
                int depth;
                int OpCode;
                tTree pLop;
                tTree pRop;
                tTree pType = pExpr->MPC_Expr.Type;

                OpCode = pExpr->MPC_BinaryExpr.OpCode;
                not_change_net = set_change_net(pExpr);
                PointerControl(pType);
                pLop = pExpr->MPC_BinaryExpr.Loperand;
                PointerControl(pLop);
                pRop = pExpr->MPC_BinaryExpr.Roperand;
                PointerControl(pRop);
                OpCode = pExpr->MPC_BinaryExpr.OpCode;
                depth = var_array_depth(pExpr);
                if (OpCode == INDEX) {
                    tTree pLopType = Strip_Typedef(pLop->MPC_Expr.Type);
                    tTree pExprType = Strip_Typedef(pExpr->MPC_Expr.Type);
                    
                    pLopType = get_vector_terminal_type(pLopType);
                    pExprType = get_vector_terminal_type(pExprType);
                    
                    if (pLopType->Kind == kMPC_ArrayType) {
                        if (is_dynamic_type(pLopType)) {
                            if (is_dynamic_type(pExprType) ==
                                DYNAMIC_ARRAY) {
                                print2file("", "(");
                                print_sexpr(pLop);
                                cur_pos +=
                                    print2file("", "+%s(",
                                               starStepMacro(pLopType,
                                                             RIGHT));
				print_lbound(pLopType);
                                print_sexpr(pRop);
                                print2file("", "))");
                            } else {
                                print2file("", "(*(");
                                print_sexpr(pLop);
                                cur_pos +=
                                    print2file("", "+%s(",
                                               starStepMacro(pLopType,
                                                             RIGHT));
				print_lbound(pLopType);
                                print_sexpr(pRop);
                                print2file("", ")))");
                            }
                        } else {
                            print_sexpr(pLop);
                            print2file("", "[");
                            print2file("", "(");
			    print_lbound(pLopType);
                            print_sexpr(pRop);
                            print2file("", ")");
                            if (is_not_c_pointer(pLopType))
                                print2file("", "*%s",
                                           step_macro(pLopType));
                            print2file("", "]");
                        }
                    } else if (pLopType->Kind == kMPC_PointerType) {
                        if (is_pointer_on_dynamic(pLopType)) {
                            print2file("", "(");
                            print_sexpr(pLop);
                            cur_pos +=
                                print2file("", "+%s%s*(",
                                           DataVarName(pLopType),
                                           SHIFT_SIZE_POSTFIX);
                            print_sexpr(pRop);
                            print2file("", "))");
                        } else {
                            print_sexpr(pLop);
                            print2file("", "[");
                            print_sexpr(pRop);
                            if (is_not_c_pointer(pType))
                                print2file("", "*%s",
                                           step_macro(pLopType));
                            print2file("", "]");
                        }
                    } /*else {
                        print_sexpr(pLop);
                        print2file("", "[");
                        print_sexpr(pRop);
                        if (is_not_c_pointer(pLopType))
                            print2file("", "*%s", step_macro(pLopType));
                        print2file("", "]");
                    }*/

                } else if ((OpCode >= DOT && OpCode <= ARROW) ||
                           (OpCode >= MULT && OpCode <= COMMA &&
                            OpCode != CONDITION)) {
                    tTree pType = pExpr->MPC_BinaryExpr.Type;
                    tTree pLType = pLop->MPC_Expr.Type;
                    tTree pRType = pRop->MPC_Expr.Type;

                    PointerControl(pType);
                    PointerControl(pLType);
                    PointerControl(pRType);
                    if (OpCode == MINUS && is_not_c_pointer(pLType) &&
                        (pRType->Kind == kMPC_PointerType
                         || pRType->Kind == kMPC_ArrayType)) {
                        print2file("", "(");
                        print_sexpr(pLop);
                        print2file("", "%s", op_represent[OpCode]);
                        print_sexpr(pRop);
                        print2file("", ")/%s", step_macro(pLType));
                    } else if ((OpCode == PLUS || OpCode == MINUS) &&
                               is_not_c_pointer(pType)) {

                        if (pLType->Kind == kMPC_PointerType
                            || pLType->Kind == kMPC_ArrayType) {
                            print_sexpr(pLop);
                            cur_pos +=
                                print2file("", "%s", op_represent[OpCode]);
                            print2file("", "(");
                            print_sexpr(pRop);
                            print2file("", ")*%s", step_macro(pLType));
                        } else {
                            print2file("", "(");
                            print_sexpr(pLop);
                            print2file("", ")*%s", step_macro(pRType));
                            cur_pos +=
                                print2file("", "%s", op_represent[OpCode]);
                            print_sexpr(pRop);
                        }
                    } else if ((OpCode == MAX) || (OpCode == MIN)) {

                        tTree pType = pExpr->MPC_Expr.Type;
                        tTree pTermType = get_vector_terminal_type(pType);
                        print_Type_in_cast(pTermType);
                        print2file("", "%s(",
                                   max_min_func_name(OpCode, pTermType));
                        print_Type_in_cast(pTermType);
                        print_sexpr(pLop);
                        print2file("", ",");
                        print_Type_in_cast(pTermType);
                        print_sexpr(pRop);
                        print2file("", ")");

                    } else {
                        print_sexpr(pLop);
                        print2file("", "%s", op_represent[OpCode]);
                        print_sexpr(pRop);
		    }
		}
		else if (OpCode == DEFINED_OPERATION) {
		    print2file("", "%s(", pExpr->MPC_Expr.OperationName);
		    print_sexpr(pLop);
		    print2file("", ",");
		    print_sexpr(pRop);
                        print2file("", ")");
		} else {
                    fprintf(stderr,
                            "invalid op code of the binary expression = %d\n",
                            OpCode);
                    internal_error();
                }
            }
            break;
        case kMPC_TernaryExpr:
            {
                int OpCode;
                tTree pFop, pSop, pTop;

                pFop = pExpr->MPC_TernaryExpr.Foperand;
                PointerControl(pFop);
                pSop = pExpr->MPC_TernaryExpr.Soperand;
                PointerControl(pSop);
                pTop = pExpr->MPC_TernaryExpr.Toperand;
                PointerControl(pTop);
                OpCode = pExpr->MPC_TernaryExpr.OpCode;
                if (OpCode == CONDITION) {
                    print_sexpr(pFop);
                    fputs("?", file);
                    print_sexpr(pSop);
                    fputs(":", file);
                    print_sexpr(pTop);
                } else {
                    fprintf(stderr,
                            "invalid op code of the ternary expression = %d\n",
                            OpCode);
                    internal_error();
                }
                break;

        case kMPC_QuaternaryExpr:
                print_grid(pExpr, global_shift);
                break;

        case kMPC_CallExpr:
                if (is_basic_function_call(pExpr) && basic_fun_call == 1)
                    basic_fun_call++;
                print_call(pExpr);
                basic_fun_call = 0;
                break;
        default:
                break;
            }

        }
        if (pExpr->MPC_Expr.Flag.InParentheses)
            cur_pos += print2file("", ")");

    }


}

/**
 * Prints lower bound expression
 *
 * @param typ array type
 *
 */
static void print_lbound(tTree typ)
{
    if(typ->MPC_ArrayType.DynLbound == NoTree){
	if(typ->MPC_ArrayType.Lbound != 0)		
	    print2file("", "(-%d)+", typ->MPC_ArrayType.Lbound);
    } else {
	print2file("", "(-(");
	print_sexpr(typ->MPC_ArrayType.DynLbound);
	print2file("", "))+");
    }    
}

/**
 * Prints Grid exression.
 *
 * @param pExpr grid expression
 * @param shift index shift
 */
void print_grid(tTree pExpr, int shift)
{
    tTree pType;
    tTree pOpType;
    tTree pArrayType;
    tTree pArrayElemType;
    tTree pOp1;
    tTree pOp2;
    tTree pOp3;
    tTree pOp4;

    pType = pExpr->MPC_Expr.Type;
    PointerControl(pType);
    pOp1 = pExpr->MPC_QuaternaryExpr.Operand1;
    PointerControl(pOp1);
    pOpType = pOp1->MPC_Expr.Type;
    PointerControl(pOpType);
    pArrayType = get_type_pattern(pOpType, expr_depth(pExpr) - 1);
    PointerControl(pArrayType);
    pArrayElemType = pArrayType->MPC_ArrayType.ElementType;
    PointerControl(pArrayElemType);
    pOp2 = pExpr->MPC_QuaternaryExpr.Operand2;
    PointerControl(pOp2);
    pOp3 = pExpr->MPC_QuaternaryExpr.Operand3;
    PointerControl(pOp3);
    pOp4 = pExpr->MPC_QuaternaryExpr.Operand4;
    PointerControl(pOp4);

    if (is_dynamic_type(pArrayType)) {
        int dyn;

        dyn = is_dynamic_type(pArrayElemType);
        if (!dyn)
            print2file("", "(*");
        print2file("", "(");

        if ((global_index_level == MPC_NIL_INDEX)
            || (global_index_level - global_shift) == 1)
            print_sexpr(pOp1);
        else
            print2file("", "%s%d",
                       TMP_SLOT_PREFIX,
                       pOp1->MPC_Expr.Optm_Attribute.slot);


        print2file("", "+%s(", starStepMacro(pArrayType, RIGHT));
        if (pOp2->Kind != kMPC_FreeNode) {
            print_sexpr(pOp2);
            print2file("", "+");
        }
	print_lbound(pArrayType);
        print2file("", "%s", index_name(global_index_level - 1));
        if (pOp4->Kind != kMPC_FreeNode) {
            print2file("", "*(");
            print_sexpr(pOp4);
            print2file("", ")");
        }
        print2file("", "))");
        if (!dyn)
            print2file("", ")");
    } else {
        if ((global_index_level == MPC_NIL_INDEX)
            || (global_index_level - global_shift) == 1)
            print_sexpr(pOp1);
        else
            print2file("", "%s%d",
                       TMP_SLOT_PREFIX,
                       pOp1->MPC_Expr.Optm_Attribute.slot);

        print2file("", "[");
        print2file("", "(");
        if (pOp2->Kind != kMPC_FreeNode) {
            print_sexpr(pOp2);
            print2file("", "+");
        }
	print_lbound(pArrayType);
        print2file("", "%s", index_name(global_index_level - 1));
        if (pOp4->Kind != kMPC_FreeNode) {
            print2file("", "*(");
            print_sexpr(pOp4);
            print2file("", ")");
        }
        print2file("", ")");
        print2file("", "%s", starStepMacro(pArrayType, LEFT));
        print2file("", "]");
    }
}


/**
 * Returns the name of mpC RTS function for MAX/MIN operations.
 *
 * @param inOpCode operation code
 * @param inType operation type
 */
static char *mpc_max_min_function_name(int inOpCode, tTree inType)
{
    int op;
    tTree type;

    if (IS_PREFIX_REDUCTION(inOpCode))
        op = inOpCode - LINEAR;
    else if (IS_POSTFIX_REDUCTION(inOpCode))
        op = inOpCode - DISTRIBUTED;

    type = get_vector_terminal_type(inType);

    return max_min_func_name(op, type);
}

/**
 * Returns the expression kind
 *
 * @param inExpr processing expression
 */
int get_expr_kind(tTree inExpr)
{
    if (Tree_IsType(inExpr, kMPC_BinaryExpr) &&
        (inExpr->MPC_BinaryExpr.OpCode == ASSIGN)) {
        tTree lop;
        tTree rop;
        int l_distr;
        int r_distr;

        lop = inExpr->MPC_BinaryExpr.Loperand;
        rop = inExpr->MPC_BinaryExpr.Roperand;
        l_distr = lop->MPC_Expr.Flag.Distributed;
        r_distr = rop->MPC_Expr.Flag.Distributed;

        if (Tree_IsType(rop, kMPC_UnaryExpr) &&
            IS_POSTFIX_REDUCTION(rop->MPC_UnaryExpr.OpCode))
            return POSTFIX_REDUCTION;

        if (Tree_IsType(rop, kMPC_UnaryExpr) &&
            IS_PREFIX_REDUCTION(rop->MPC_UnaryExpr.OpCode))
            return PREFIX_REDUCTION;

        if ((inExpr->MPC_BinaryExpr.Loperand->MPC_Expr.StoreNet ==
             inExpr->MPC_BinaryExpr.Roperand->MPC_Expr.StoreNet) ||
            (inExpr->MPC_BinaryExpr.Roperand->MPC_Expr.StoreNet ==
             CONST_NET))
            return ASYNCH_ELEMENTWISE;
        else if (l_distr == r_distr)
            return PARALLEL_SEND_ASSIGN;
        else if ((l_distr == 0) && (r_distr == 1))
            return GATHER_ASSIGN;
        else if ((l_distr == 1) && (r_distr == 0)) {
            if (vector_rank(lop->MPC_Expr.Type) == vector_rank(rop->MPC_Expr.Type))
                return BCAST_ASSIGN;
            else
                return SCATTER_ASSIGN;
        }
    } else {
        return ASYNCH_ELEMENTWISE;
    }

}


/**
 * Returns the rank of the vector type. For non-vector types
 * it returns 0.
 *
 * @param inType processing type
 */
static int vector_rank(tTree inExpr)
{
    if (Tree_IsType(inExpr, kMPC_VectorType)) {
        return vector_rank(inExpr->MPC_VectorType.ElementType) + 1;
    } else
        return 0;
}


/**
 * Generates code for element-wise expression.
 *
 * @param inExpr processing expression
 */
static void print_asynch_elementwise(tTree inExpr)
{
    int depth;

    /** 
     * rreddy, 17/04/2009
     * For large mpC files, these comments could lead to large generated 
     * files resulting in massive compilation times.
     * The generation of these comments will now only take place in
     * Debug mode
     */
    if (MPCC_DEBUG)
    {
       print2file("", "/*begin asynch elementwise (line %d)*/\n", RealPosition(inExpr->MPC_Expr.Pos).Line);
    }

    depth = vector_rank(inExpr->MPC_Expr.Type);

    c_blank;

    if (depth == 0) {
        print_sexpr(inExpr);
        print2file("", ";\n");
    } else {
        int i;
        ExprPiece_traverse(NoTree, inExpr, set_tmp_slots);
        index_boundary = malloc(sizeof(struct INDEX_BOUNDARY) * depth);
        set_index_boundary(inExpr->MPC_Expr.Type);

        print2file("", "{\n");
        for (i = 0; i < depth; i++) {
            global_index_level = i;
            ExprPiece_traverse(NoTree, inExpr, declare_tmp_slots);
            c_blank;
            print2file("", "int %s = 0;\n", index_name(i));
            ExprPiece_traverse(NoTree, inExpr, calc_tmp_slots);
            c_blank;
            if (index_boundary[i].NumberOfComponents)
                print2file("", "for(;%s<%d;%s++) {\n",
                           index_name(i),
                           index_boundary[i].NumberOfComponents,
                           index_name(i));
            else {
                print2file("", "for(;%s<", index_name(i));
                print2file("", "%s.Number;",
                           DataVarName(index_boundary[i].Type));
                print2file("", "%s++) {\n", index_name(i));
            }
            cur_blank += SKIP;
        }

        global_index_level++;

        free(index_boundary);
        c_blank;

        print_sexpr(inExpr);

        print2file("", ";\n");

        for (i = 0; i < depth; i++)
            b_end;
        b_end;
    }

    /** 
     * rreddy, 17/04/2009
     * For large mpC files, these comments could lead to large generated 
     * files resulting in massive compilation times.
     * The generation of these comments will now only take place in
     * Debug mode
     */
    if (MPCC_DEBUG)
    {
       print2file("/*end asynch elementwise*/", "\n");
    }
}

/**
 * Generates code for reduction expression.
 * 
 * @param inExpr processing expression
 */
static void print_prefix_reduction(tTree inExpr)
{
    int depth, i;

    print2file("", "/*begin prefix reduction (%d)*/\n", RealPosition(inExpr->MPC_Expr.Pos).Line);
    depth =
        vector_rank(inExpr->MPC_BinaryExpr.Roperand->MPC_UnaryExpr.
                    Operand->MPC_Expr.Type);
    index_boundary = malloc(sizeof(struct INDEX_BOUNDARY) * depth);
    set_index_boundary(inExpr->MPC_BinaryExpr.Roperand->MPC_UnaryExpr.
                       Operand->MPC_Expr.Type);
    ExprPiece_traverse(NoTree, inExpr, set_tmp_slots);

    c_blank;
    print2file("", "{\n");

    cur_blank += SKIP;

    for (i = 0; i < depth; i++) {
        c_blank;
        print2file("", "int %s = 0;\n", index_name(i));
        global_index_level = i;
        global_shift = 1;
        ExprPiece_traverse(NoTree, inExpr->MPC_BinaryExpr.Loperand,
                           declare_tmp_slots);
        global_shift = 0;
        ExprPiece_traverse(NoTree, inExpr, declare_tmp_slots);

        global_shift = 1;
        ExprPiece_traverse(NoTree, inExpr->MPC_BinaryExpr.Loperand,
                           calc_tmp_slots);
        global_shift = 0;
        ExprPiece_traverse(NoTree, inExpr, calc_tmp_slots);
        c_blank;
        if (index_boundary[i].NumberOfComponents)
            print2file("", "for(;%s<%d;%s++) {\n",
                       index_name(i),
                       index_boundary[i].NumberOfComponents,
                       index_name(i));
        else {
            print2file("", "for(;%s<", index_name(i));
            print2file("", "%s.Number;",
                       DataVarName(index_boundary[i].Type));
            print2file("", "%s++) {\n", index_name(i));
        }
        cur_blank += SKIP;
    }

    global_index_level++;
    c_blank;
    print2file("", "if(%s == 0) {\n", index_name(0));
    cur_blank += SKIP;
    c_blank;
    global_shift = 1;
    print_sexpr(inExpr->MPC_BinaryExpr.Loperand);
    global_shift = 0;
    print2file("", " = ");
    print_sexpr(inExpr->MPC_BinaryExpr.Roperand->MPC_UnaryExpr.Operand);
    print2file("", ";\n");
    cur_blank -= SKIP;
    c_blank;
    print2file("", "} else {\n");

    cur_blank += SKIP;
    c_blank;
    global_shift = 1;
    print_sexpr(inExpr->MPC_BinaryExpr.Loperand);
    global_shift = 0;

    if ((inExpr->MPC_BinaryExpr.Roperand->MPC_UnaryExpr.OpCode == LIN_MAX)
        || (inExpr->MPC_BinaryExpr.Roperand->MPC_UnaryExpr.OpCode ==
            LIN_MIN)) {
        print2file("", " = %s(",
                   mpc_max_min_function_name(inExpr->MPC_BinaryExpr.
                                             Roperand->MPC_UnaryExpr.
                                             OpCode,
                                             inExpr->MPC_BinaryExpr.
                                             Roperand->MPC_UnaryExpr.
                                             Type));
        global_shift = 1;
        print_sexpr(inExpr->MPC_BinaryExpr.Loperand);
        global_shift = 0;
        print2file("", ",");
        print_sexpr(inExpr->MPC_BinaryExpr.Roperand->MPC_UnaryExpr.
                    Operand);
        print2file("", ")");

    } else {
        print2file("", " %s ",
                   lin_oper_code(inExpr->MPC_BinaryExpr.Roperand->
                                 MPC_UnaryExpr.OpCode));
        print_sexpr(inExpr->MPC_BinaryExpr.Roperand->MPC_UnaryExpr.
                    Operand);
    }
    print2file("", ";\n");
    b_end;

    for (i = 0; i < depth; i++)
        b_end;

    b_end;

    free(index_boundary);
    print2file("/*end prefix reduction*/", "\n");
}



/**
 * Prints address of Lvalue expression.
 *
 * @param inExpr processing expression
 */
static void print_address(tTree inExpr)
{
    global_index_level = MPC_NIL_INDEX;

    if (Tree_IsType(inExpr, kMPC_UnaryExpr)
        && (inExpr->MPC_UnaryExpr.OpCode == BRACKETS)) {
        print_sexpr(inExpr);
    } else {
        print2file("", "&(");
        print_sexpr(inExpr);
        print2file("", ")");
    }
}

/**
 *  Finds net enclosing given subnet.
 *
 *  @param tTree inSubnet given subnet
 */
 static find_enclosing_net(tTree inSubnet)
 {
   tTree net;

   net = inSubnet;

   while(net->Kind != kMPC_Net)
     net = net->MPC_Subnet.Distribution;

   return net;
 }
  
/**
 * Returns minal supernet for organizing send-recv operation.
 *
 * @param inLNet left operands distribution
 * @param inRNet right operands distribution
 * @param inLRel kind of relation left-subnet <-> supernet
 * @param inRRel kind of relation right-subnet <-> supernet
 */
static tTree get_minimal_supernet(tTree inLNet, tTree inRNet, tRelSpecifier* inLRel, tRelSpecifier* inRRel)
{
    tTree net;
    
    if(Tree_IsType(inLNet, kMPC_Subnet) && Tree_IsType(inRNet, kMPC_Subnet)){
            *inLRel = SUBNET_RELATION;
            *inRRel = SUBNET_RELATION;
            return inLNet->MPC_Subnet.Distribution;
    } else if(Tree_IsType(inLNet, kMPC_Net) && Tree_IsType(inRNet, kMPC_Subnet)){
        if(inRNet->MPC_Subnet.Distribution == inLNet){
            *inLRel = IDENTICAL_RELATION;
            *inRRel = SUBNET_RELATION;
            return inLNet;
        } else if(inLNet->MPC_Net.Distribution == inRNet){
            *inLRel = IDENTICAL_RELATION;
            *inRRel = PARENT_RELATION;
            return inLNet;
        } else if(Tree_IsType(inRNet->MPC_Subnet.Distribution, kMPC_Net) &&
            (inRNet->MPC_Subnet.Distribution->MPC_Net.Distribution == inLNet)) {
            *inLRel = PARENT_RELATION;
            *inRRel = SUBNET_RELATION;
            return inRNet->MPC_Subnet.Distribution;
        } else {
            report_error("get_minimal_supernet: failed <subnet|subnet>\n");
        }
    } else if(Tree_IsType(inLNet, kMPC_Subnet) && Tree_IsType(inRNet, kMPC_Net)){
        if(inLNet->MPC_Subnet.Distribution == inRNet){
            *inLRel = SUBNET_RELATION;
            *inRRel = IDENTICAL_RELATION;
            return inRNet;
        } else if(inRNet->MPC_Net.Distribution == inLNet){
            *inLRel = PARENT_RELATION;
            *inRRel = IDENTICAL_RELATION;
            return inRNet;
        } else if(Tree_IsType(inLNet->MPC_Subnet.Distribution, kMPC_Net) &&
            (inLNet->MPC_Subnet.Distribution->MPC_Net.Distribution == inRNet)) {
            *inLRel = SUBNET_RELATION;
            *inRRel = PARENT_RELATION;
            return inLNet->MPC_Subnet.Distribution;
        } else {
            report_error("get_minimal_supernet: failed <subnet|net>\n");
        }
    } else if(Tree_IsType(inLNet, kMPC_Net) && Tree_IsType(inRNet, kMPC_Net)){
        if(inLNet->MPC_Net.Distribution == inRNet){
            *inLRel = IDENTICAL_RELATION;
            *inRRel = PARENT_RELATION;
            return inLNet;
        } else if(inRNet->MPC_Net.Distribution == inLNet){
            *inLRel = PARENT_RELATION;
            *inRRel = IDENTICAL_RELATION;
            return inRNet;
        } else {
            report_error("get_minimal_supernet: failed <net|net>\n");
        }        
    } else {
            report_error("get_minimal_supernet: unrecognized case\n");
    }
    
}

/**
 * Refines the relation. Sets custom relation kind for cutting
 * operator.
 * 
 * @param inRel (input-output) relation and refined relation
 * @param inNet the subnet
 */
 static void refine_relation_kind(int* inRel, tTree inNet)
 {
    if((*inRel == SUBNET_RELATION) && Tree_IsType(inNet, kMPC_Subnet) 
        && (inNet->MPC_Subnet.Ident == NoIdent)) {
        
        if(!inNet->MPC_Subnet.SingleNode)
            report_error("non-single node in cutting operator\n");
            
        *inRel = CUSTOM_RELATION;            
    }
 }

/**
 * Checks whether network object inNet is a network cutting operator, i.e. [w: P].
 *
 * @param inNet netwrok object
 * @return 1 if inNet is a cutting operator, 0 otherwise
 */
static int is_network_cut(tTree inNet)
{
    if(Tree_IsType(inNet, kMPC_Subnet) && (inNet->MPC_Subnet.Ident == NoIdent))
    return 1;
    else
    return 0;
}

/**
 * Fills inString with corresponding macro name.
 *
 * @param inString (output) macro postfix
 * @param inRel the relation kind
 */
static void get_rel_kind_macro(char* inString, int inRel)
{
    static int rel_kind_number = 0;
    
    switch(inRel){
    case IDENTICAL_RELATION:
        sprintf(inString, "I");
        break;
    case PARENT_RELATION:
        sprintf(inString, "P");
        break;
    case SUBNET_RELATION:
        sprintf(inString, "S");
        break;
    case CUSTOM_RELATION:
        sprintf(inString, "C%d", rel_kind_number ++);
        break;
    }
}


/**
 * Prints node coordinates.
 *
 * @param inExpr preocessing expression
 */
static void  assign_node_coord(tTree inExpr)
{
    if(Tree_IsType(inExpr, kMPC_BinaryExpr) 
        && (inExpr->MPC_BinaryExpr.OpCode == EQUAL)){
        if(Tree_IsType(inExpr->MPC_BinaryExpr.Loperand, kMPC_CoordExpr)) {
            print2file("", "%s[%d] = ", 
                NODE_COORD, 
                inExpr->MPC_BinaryExpr.Loperand->MPC_CoordExpr.CoordNumber);            
            print_sexpr(inExpr->MPC_BinaryExpr.Roperand);
            print2file("", ";\\\n");
        } else if(Tree_IsType(inExpr->MPC_BinaryExpr.Roperand, kMPC_CoordExpr)) {
            print2file("", "%s[%d] = ", 
                NODE_COORD, 
                inExpr->MPC_BinaryExpr.Roperand->MPC_CoordExpr.CoordNumber);            
            print_sexpr(inExpr->MPC_BinaryExpr.Loperand);
            print2file("", ";\\\n");
        } else {
            report_error("print_node_coord: inconsistent predicate");
        }
    }
}  
 

/**
 * Prints node coordinates defintions.
 *
 * @param inSubnet single node subnet
 */
static void print_node_coords(tTree inSubnet)
{
    Expr_traverse(inSubnet->MPC_Subnet.Predicate, assign_node_coord);   
}


/**
 * Generates macro defintions for cutting operator predicates.
 *
 * @param inSuperNet supernet
 * @param inSubnet subnet
 * @param inName macro postfix
 */
void define_custom_rel_macros(tTree inSupernet, tTree inSubnet, char* inName)
{
    tTree net;
    
    net = net_parent(inSubnet);
    print2file("", "#define %s%s(SUBNET) ", NET_PREDICATE_PREFIX, inName);
    print2file("", "if(");
    print_sexpr(inSubnet->MPC_Subnet.Predicate);
    print2file("", ")\n");
    print2file("", "#define %s%s(NUM, SUBNET) ", MPC_GET_MY_NUM_PREFIX, inName);
    print2file("", "NUM = 0;\n");
    print2file("", "#define %s%s(NUM, SUBNET_NUM, NET, SUBNET) ", MPC_GET_ENCLOSED_NUM_PREFIX, inName);
    print2file("", "{\\\n");
    print2file("", "int %s[%d];\\\n", NODE_COORD, coord_number(net->MPC_Net.NetType));
    print_node_coords(inSubnet);
    print2file("", "NUM = MPC_Number(&%s, %s);\\\n", VarName(net), NODE_COORD);
    print2file("", "}\n");
}


/**
 * Constructs an ident for an auxilary left subnet for 
 * parallel assignment operator.
 *
 * @returns constructed ident
 */
static tIdent lazy_auxilarly_left_subnet_ident()
{
    static tIdent id = -1;
    
    if(id == -1)
    id = MakeIdent(LEFT_AUX_SUBNET_NAME, strlen(LEFT_AUX_SUBNET_NAME));

    return id;
}

/**
 * Constructs an ident for an auxilary right subnet for 
 * parallel assignment operator.
 *
 * @returns constructed ident
 */
static tIdent lazy_auxilarly_right_subnet_ident()
{
    static tIdent id = -1;
    
    if(id == -1)
    id = MakeIdent(RIGHT_AUX_SUBNET_NAME, strlen(RIGHT_AUX_SUBNET_NAME));

    return id;
}

/**
 * Generates code for any kind of parallel assignement using macros
 * from mpcassig.h.
 *
 * @param inExpr processing expression.
 */
static void print_parallel_assign(tTree inExpr, int inAssignKind)
{
    tTree LType;
    tTree RType;
    tTree Net;
    tTree LNet;
    tTree RNet;
    int lcut;
    int rcut;
    int LRank;
    int RRank;
    tRelSpecifier LRel;
    tRelSpecifier RRel;
    char *macro;
    char lkind[MAX_NAME_LEN];
    char rkind[MAX_NAME_LEN];
    tIdent oldlident;
    tIdent oldrident;
    int oldlflag;
    int oldrflag;



    print2file("", "/*begin parallel assign, line(%d)*/\n", RealPosition(inExpr->MPC_Expr.Pos).Line);
    
    c_blank;
    print2file("", "MPC_Source_position_set(%d,%d);\n",
               RealPosition(inExpr->MPC_Expr.Pos).Column,
               RealPosition(inExpr->MPC_Expr.Pos).Line);

    
    GET_ASSIGN_MACRO_NAME(macro, inAssignKind)

    LType = inExpr->MPC_BinaryExpr.Loperand->MPC_Expr.Type;
    RType = inExpr->MPC_BinaryExpr.Roperand->MPC_Expr.Type;
    LRank = vector_rank(LType);
    RRank = vector_rank(RType);

    LNet = inExpr->MPC_BinaryExpr.Loperand->MPC_Expr.StoreNet;
    RNet = inExpr->MPC_BinaryExpr.Roperand->MPC_Expr.StoreNet;
    Net  = get_minimal_supernet(LNet, RNet, &LRel, &RRel);

    lcut = is_network_cut(LNet);
    rcut = is_network_cut(RNet);    
    if(lcut || rcut) 
    {
        c_blank;
    print2file("", "if((");

    if(lcut)
        print_sexpr(LNet->MPC_Subnet.Predicate);
    else
            print2file("", "%s(&%s)", MEMBER, VarName(LNet));
    print2file("",")||(");
    if(rcut)
        print_sexpr(RNet->MPC_Subnet.Predicate);
    else
            print2file("", "%s(&%s)", MEMBER, VarName(RNet));
    
    print2file("", "))\n");
    c_blank;
    BRACE_OPEN;

    if(lcut) {
            oldlident = LNet->MPC_Subnet.Ident;
            oldlflag = LNet->MPC_Subnet.WithNumber;
        LNet->MPC_Subnet.Ident = lazy_auxilarly_left_subnet_ident();
            LNet->MPC_Subnet.WithNumber = 1;
        c_blank;            
        print_SubnetDesc(LNet, 1);
        }
    if(rcut) {
        oldrident = RNet->MPC_Subnet.Ident;
            oldrflag = RNet->MPC_Subnet.WithNumber;
        RNet->MPC_Subnet.Ident = lazy_auxilarly_right_subnet_ident();
            RNet->MPC_Subnet.WithNumber = 1;
        c_blank;
        print_SubnetDesc(RNet, 1);  
    }
    if(lcut) 
        print_create_subnet(LNet);        
    if(rcut) 
        print_create_subnet(RNet);  
    }

/*    
    refine_relation_kind(&LRel, LNet);
    refine_relation_kind(&RRel, RNet);
*/
    get_rel_kind_macro(lkind, LRel);
    get_rel_kind_macro(rkind, RRel);

/*    
    if(LRel == CUSTOM_RELATION) 
        define_custom_rel_macros(Net, LNet, lkind);

    if(RRel == CUSTOM_RELATION) 
        define_custom_rel_macros(Net, RNet, rkind);
*/

        
    c_blank;    
    print2file("", "%s(", macro);
    print2file("", "%s,", DataVarName(LType));
    print2file("", "%s,", DataVarName(RType));
    print2file("", "%s,", VarName(Net));
    print2file("", "%s,", VarName(LNet));
    print2file("", "%s,", lkind);
    print2file("", "%s,", VarName(RNet));
    print2file("", "%s,", rkind);
    print_address(inExpr->MPC_BinaryExpr.Loperand);
    print2file("", ",");
    print_address(inExpr->MPC_BinaryExpr.Roperand);
    print2file("", ")\n");

    if(lcut || rcut) 
    {
    if(lcut)
        print_free_subnet(LNet);
    if(rcut)
        print_free_subnet(RNet);        
    BRACE_CLOSE;
        if(lcut) {
        LNet->MPC_Subnet.Ident = oldlident;
            LNet->MPC_Subnet.WithNumber = oldlflag;
        }
        if(rcut) {
            RNet->MPC_Subnet.Ident = oldrident;
            RNet->MPC_Subnet.WithNumber = oldrflag;
        }        
    }

    print2file("/*end parallel assign*/", "\n");
}


void print_postfix_reduction(tTree inExpr)
{
    char *op;

    GET_POSTFIX_REDUCTION_OP_NAME(op,
                                  inExpr->MPC_BinaryExpr.Roperand->
                                  MPC_UnaryExpr.OpCode)

    print2file("", "/*begin postfix reduction line(%d)*/\n", RealPosition(inExpr->MPC_Expr.Pos).Line);
    c_blank;
    print2file("", "MPC_Source_position_set(%d,%d);\n",
               RealPosition(inExpr->MPC_Expr.Pos).Column,
               RealPosition(inExpr->MPC_Expr.Pos).Line);
    c_blank;
    print2file("", "%s(", MPC_REDUCTION_FUN_NAME);
    print2file("", "&%s,", VarName(inExpr->MPC_Expr.StoreNet));
    print_address(inExpr->MPC_BinaryExpr.Roperand->MPC_UnaryExpr.Operand);
    print2file("", ",");
    print_address(inExpr->MPC_BinaryExpr.Loperand);
    print2file("", ",");
    print2file("", "(MPC_Dt)(&%s),", DataVarName(inExpr->MPC_Expr.Type));
    print2file("", "(MPC_Dt)(&%s),",
               DataVarName(get_vector_terminal_type
                           (inExpr->MPC_Expr.Type)));
    print2file("", "%s);\n", op);
    print2file("/*end postfix reduction*/", "\n");
}


/**
 * The root function for expression generation
 *
 * @param inExpr processing expression
 */
void print_expr(tTree inExpr)
{
    int kind;
    
    kind = get_expr_kind(inExpr);
    

    switch (kind) {
    case ASYNCH_ELEMENTWISE:
        print_asynch_elementwise(inExpr);
        break;

    case PREFIX_REDUCTION:
        print_prefix_reduction(inExpr);
        break;

    case SCATTER_ASSIGN:
    case GATHER_ASSIGN:
    case BCAST_ASSIGN:
    case PARALLEL_SEND_ASSIGN:
        print_parallel_assign(inExpr, kind);
        break;

    case POSTFIX_REDUCTION:
        print_postfix_reduction(inExpr);
        break;
    }
}
