#ifndef _EXPR_
#define _EXPR_

#include "Tree.h"

/*
 * Basic expression kinds from the implementation point of view.
 */

/*
 * Expression Statement kinds
 */
#define ASYNCH_ELEMENTWISE 1

#define PREFIX_REDUCTION 2

#define SCATTER_ASSIGN 3

#define GATHER_ASSIGN 4

#define BCAST_ASSIGN 5

#define PARALLEL_SEND_ASSIGN 6

#define POSTFIX_REDUCTION 7

    

#define GET_ASSIGN_MACRO_NAME(STRING, KIND)\
    switch(KIND){\
    case SCATTER_ASSIGN:\
        STRING = "MPC_SCATTER_ASSIGN";\
        break;\
    case GATHER_ASSIGN:\
        STRING = "MPC_GATHER_ASSIGN";\
        break;\
    case BCAST_ASSIGN:\
        STRING = "MPC_BCAST_ASSIGN";\
        break;\
    case PARALLEL_SEND_ASSIGN:\
        STRING = "MPC_PARALLEL_SEND_ASSIGN";\
        break;\
    default:\
        STRING = "MPC_PARALLEL_ASSIGN";\
    }

/*
 * Relation specifications
 */
 
 typedef enum {
    IDENTICAL_RELATION,   /* superregion and subregion are identical */
    
    PARENT_RELATION,      /* subregion is a parent node of the superregion */
    
    SUBNET_RELATION,      /* subregion is a subnet of a superregion */
    
    CUSTOM_RELATION       /* support for cutting operator */
 } tRelSpecifier;


#define NET_PREDICATE_PREFIX "NET_PREDICATE_"

#define MPC_GET_MY_NUM_PREFIX "MPC_GET_MY_NUM_"

#define MPC_GET_ENCLOSED_NUM_PREFIX "MPC_GET_ENCLOSED_NUM_" 

#define MPC_REDUCTION_FUN_NAME "MPC_Reduce"

#define GET_POSTFIX_REDUCTION_OP_NAME(STRING, OP)\
    switch(OP){\
    case DISTR_LIN_MULT:\
        STRING = "MPC_PROD";\
        break;\
    case DISTR_LIN_MAX:\
        STRING = "MPC_MAX";\
        break;\
    case DISTR_LIN_MIN:\
        STRING = "MPC_MIN";\
        break;\
    case DISTR_LIN_PLUS:\
        STRING = "MPC_SUM";\
        break;\
    case DISTR_LIN_AND:\
        STRING = "MPC_BAND";\
        break;\
    case DISTR_LIN_EXCL_OR:\
        STRING = "MPC_BOR";\
        break;\
    case DISTR_LIN_INCL_OR:\
        STRING = "MPC_BXOR";\
        break;\
    case DISTR_LIN_LOG_AND:\
        STRING = "MPC_LAND";\
        break;\
    case DISTR_LIN_LOG_OR:\
        STRING = "MPC_LOR";\
        break;\
    }

int get_expr_kind(tTree inExpr);

void print_expr(tTree inExpr);



#endif
