#include "be_print.h"
#include "expr.h"

#define MAX_GRID_REGIONS 1000

#define Traverse(struct,field) {yyn=yyt->struct.field; \
  if(yyn != NoTree) refactor_traverse(&(yyt->struct.field),proc);}

static tDeclFlags commonDeclFlags = { 1, 0 };
static tStatFlags commonStatNoSourceFlags;
static tStatFlags commonStatSourceFlags;
static tTree tmpVarDecls;
static tTree statsSequence;
static tTree emptyStat;
static tTree tmpVarTermNode;
static tTree noTreeStat = NoTree;
static int tmp_var_index;
static tPosition exprPos;
static tPosition statPos;
static tTree globalRop;
static tIdent globalIdent;
static int isCrossing;

/*
 * Traverses type tree in order to add tag reference
 * declarations for previously declared structure and union types:
 * 
 * structs S
 *
 * @inExpr processing expression
 */
 static tTree enrich_with_equiv_SU(tTree inTree){
    tTree new, old;
    union Tree_Node* typ;
    
    if(Tree_IsType(inTree, kMPC_ArrayType)) {
        old = inTree->MPC_ArrayType.ElementType;
        new = enrich_with_equiv_SU(old);
        if(new != old) {
            typ = (union Tree_Node*) malloc (sizeof(union Tree_Node));
            * typ = * (union Tree_Node*) inTree;
            typ->MPC_ArrayType.ElementType = new;
            return typ;             
        } else 
            return inTree;
    } else if(Tree_IsType(inTree, kMPC_PointerType)) {
        old = inTree->MPC_PointerType.ElementType;
        new = enrich_with_equiv_SU(old);
        if(new != old) {
            typ = (union Tree_Node*) malloc (sizeof(union Tree_Node));
            * typ = * (union Tree_Node*) inTree;
            typ->MPC_PointerType.ElementType = new;
            return typ;             
        } else 
            return inTree;
    } else if(Tree_IsType(inTree, kMPC_SU_Type)) {
        if(inTree->MPC_SU_Type.MemberDecls == NoTree) 
            return inTree;
        else if(inTree->MPC_SU_Type.EquivType != NoTree){
            return inTree->MPC_SU_Type.EquivType;
        } else {
            assign_SU_tag(inTree);
            typ = (union Tree_Node*) malloc (sizeof(union Tree_Node));
            * typ = * (union Tree_Node*) inTree;
            typ->MPC_SU_Type.MemberDecls = NoTree;            
            typ->MPC_SU_Type.EquivType = inTree;
            inTree->MPC_SU_Type.EquivType = typ;
            return typ;
        } 
    } else 
        return inTree;
 }

/*
 * Constructs pointer type from array type (use in
 * expressions.
 *
 * @inExpr processing experession
 */
static tTree construct_scalar_type(tTree inTree)
{
    tTree tree;
    tTree rt;
    
    tree = enrich_with_equiv_SU(inTree);
    if (Tree_IsType(tree, kMPC_ArrayType)){
        rt = mMPC_PointerType(tree->MPC_ArrayType.Pos,
                                tree->MPC_ArrayType.Next,
                                tree->MPC_ArrayType.Prev,
                                tree->MPC_ArrayType.Flag,
                                tree->MPC_ArrayType.NumberOfComponents,
                                tree->MPC_ArrayType.Step,
                                tree->MPC_ArrayType.ElementType);
        rt->MPC_Type.Flag.Source = 0;           
    } else
        rt = tree;
    
    return rt;
}

/*
 * Constructs temporary variable's type.
 *
 * @inType processing expression
 */
static tTree construct_var_type(tTree inType)
{
    tTree rt;
    
    if (Tree_IsType(inType, kMPC_VectorType)) {
        tTree typ;

        typ = mMPC_ArrayType(exprPos,
                             NoTree,
                             NoTree,
                             inType->MPC_Type.Flag,
                             inType->MPC_VectorType.NumberOfComponents,
                             1,
                             construct_var_type(inType->MPC_VectorType.
                                                ElementType));

        typ->MPC_ArrayType.ArraySize = inType->MPC_VectorType.VectorSize;
        typ->MPC_Type.Flag.Source = 0;
        rt = typ;
    } else {
        rt = construct_scalar_type(get_vector_terminal_type(inType));
    }
    rt->MPC_Type.Flag.Const = 0;
    return rt;
}

/**
 * Constructs the corresponding vector of pointers
 *
 * @param inType processing type
 */
 static tTree  construct_vptype(tTree inType)
 {
    if (Tree_IsType(inType, kMPC_VectorType)) {
        tTree typ;

        typ = mMPC_VectorType(exprPos,
                             NoTree,
                             NoTree,
                             inType->MPC_Type.Flag,
                             inType->MPC_VectorType.NumberOfComponents,
                             1,                                     
                             construct_vptype(inType->MPC_VectorType.ElementType));

        typ->MPC_VectorType.VectorSize = inType->MPC_VectorType.VectorSize;
        typ->MPC_Type.Flag.Source = 0;       
        return typ;
    } else
        return mMPC_PointerType(exprPos, 
            NoTree, 
            NoTree, 
            inType->MPC_Type.Flag, 
            0,
            1, 
            inType);
    
 }
 
/** 
 * Transforms lvalue E to the form of *&E.
 *
 * @param inExpr processing expression
 */
 static tTree normalize_lvalue(tTree inExpr)
 {
    tTree typ, ptyp, expr;
    tExprFlags flags;
    
    typ = inExpr->MPC_Expr.Type;
    ptyp = construct_vptype(typ);
    
    flags = inExpr->MPC_Expr.Flag;
    flags.Lvalue = 0;
    flags.Const = 0;
    flags.Generated = 0;
  
    expr = mMPC_UnaryExpr(exprPos, 
        flags, 
        "", 
        ptyp, 
        inExpr->MPC_Expr.EvalNet, 
        inExpr->MPC_Expr.StoreNet, 
        ADDRESS, 
        inExpr);
    
    flags.Lvalue = 1;

    return mMPC_UnaryExpr(exprPos, 
        flags, 
        "", 
        typ, 
        inExpr->MPC_Expr.EvalNet, 
        inExpr->MPC_Expr.StoreNet, 
        STAR, 
        expr);
 }


/*
 * Returns temporary variable distribution.
 *
 * @param inTree processing expression
 *
 */
static tTree construct_var_distribution(tTree inTree)
{
    return inTree->MPC_Expr.StoreNet;
}

/*
 * Returns MPC_Var part of variable declaration.
 *
 * @param inTree processing expression    
 *
 */
static tTree construct_var(tTree inTree)
{
    char *name;
    tIdent ident;

    PointerControl(inTree);
    
    name = (char *) malloc(MAX_IDENT_LEN);

    sprintf(name, "%s%d", MPC_TMP_NAME_PREFIX, tmp_var_index++);

    ident = MakeIdent(name, strlen(name));
    
    return mMPC_Var(inTree->MPC_Expr.Pos,
                    ident,
                    commonDeclFlags,
                    NoTree,
                    NoTree,
                    construct_var_type(inTree->MPC_Expr.Type),
                    construct_var_distribution(inTree), NoTree);
}

/* 
 * Produces temporary variable declaration for a
 * given expression.
 *
 * @param inTree processing expression
 */
static tTree construct_var_decl(tTree inTree)
{
    tTree notree;
    tTree var;
    
    PointerControl(inTree);
    
    var = construct_var(inTree);

    tmpVarDecls = mMPC_VarDecl(inTree->MPC_Expr.Pos,
                               tmpVarDecls,
                               tmpVarTermNode,
                               AUTO,
                               commonDeclFlags,
                               get_base_type(var->MPC_Var.Type),
                               var);
    if(Tree_IsType(tmpVarDecls->MPC_VarDecl.Next, kMPC_VarDecl)) 
        tmpVarDecls->MPC_VarDecl.Next->MPC_VarDecl.Prev = tmpVarDecls;
    
    notree = mMPC_FreeNode(tmpVarDecls);
    tmpVarDecls->MPC_VarDecl.Var->MPC_Var.Next = notree;
    tmpVarDecls->MPC_VarDecl.Var->MPC_Var.Prev = notree;

    return tmpVarDecls;
}


/*
 * Returns temporary variable expression.
 * 
 * @param inVarDecl temporary variable declaration
 * @param inExpr expression for replacing by temporary expression
 */
static tTree construct_tmp_expr(tTree inVarDecl, tTree inExpr)
{
    tTree typ;
    tTree var;
    tTree ident;
    tTree tmp;
    tExprFlags flags;

    
    flags = inExpr->MPC_Expr.Flag;
    flags.Lvalue = 1;
    flags.Const = 0;
    flags.Generated = 0;
    flags.Source = 0;

    var = inVarDecl->MPC_VarDecl.Var;
    typ = var->MPC_Var.Type;
    ident = mMPC_Ident(exprPos,
                       flags,
                       "",
                       typ,
                       inExpr->MPC_Expr.EvalNet,
                       inExpr->MPC_Expr.StoreNet,
                       var->MPC_Var.Ident, NoTree);


    if (Tree_IsType(inExpr->MPC_Expr.Type, kMPC_VectorType))
        return mMPC_UnaryExpr(exprPos,
                              flags,
                              "",
                              inExpr->MPC_Expr.Type,
                              inExpr->MPC_Expr.EvalNet,
                              inExpr->MPC_Expr.StoreNet, BRACKETS, ident);
    else
        return ident;
}

/**
 * Checks whether processing assignement causes
 * send-receive operations.
 *
 * @param inExpr processing expression
 */
static int is_exchange_assign(tTree inExpr)
{
    int kind;

    kind = get_expr_kind(inExpr);

    if ((kind == SCATTER_ASSIGN)
        || (kind == BCAST_ASSIGN)
        || (kind == GATHER_ASSIGN)
        || (kind == PARALLEL_SEND_ASSIGN))
        return 1;
    else
        return 0;
}

/** 
 * Looks for a specific identifier in a tree.
 * 
 * @param inExpr traversal node
 */
 static void find_crossing_rop(tTree inExpr)
 {
    if(Tree_IsType(inExpr, kMPC_UnaryExpr) && 
       inExpr->MPC_UnaryExpr.OpCode == STAR)
       isCrossing = 1;
    else if(Tree_IsType(inExpr, kMPC_Ident)) {
        if(globalIdent == inExpr->MPC_Ident.Ident) 
            isCrossing = 1;
    }        
 }
 
 /**
  * Looks for identifiers or indirections in a tree.
  *
  * @param inExpr traversal node
  */ 
 static void find_crossing(tTree inExpr) 
 {
    if(Tree_IsType(inExpr, kMPC_UnaryExpr) && 
       inExpr->MPC_UnaryExpr.OpCode == STAR)
       isCrossing = 1;
    else if(Tree_IsType(inExpr, kMPC_Ident)) {
        globalIdent = inExpr->MPC_Ident.Ident;
        Expr_traverse(globalRop, find_crossing_rop);
    }        
 }
 
/**
 * Checks whether left and right operands of the given 
 * assignement are disjoint in memory.
 *
 * @param inExpr processing exression
 */
static int is_crossing_assign(tTree inExpr)
{
    if(Tree_IsType(inExpr, kMPC_BinaryExpr) && 
       IS_ASSIGN_OPER(inExpr->MPC_BinaryExpr.OpCode) &&
       !is_exchange_assign(inExpr)) {
        
        isCrossing = 0;
        globalRop = inExpr->MPC_BinaryExpr.Roperand;        
        Expr_traverse(inExpr->MPC_BinaryExpr.Loperand, find_crossing);
        return isCrossing;
    } else 
        return 0;
}

/**
 * Attaches expression to the sequence of assignements.
 *
 * @param inExpr expression to attach
 */
static void attach_expr(tTree inExpr)
{
    if (Tree_IsType(inExpr, kMPC_Exprs)) 
        statsSequence = mMPC_ExprStat(exprPos,
                                      statsSequence,
                                      emptyStat,
                                      commonStatNoSourceFlags,
                                      inExpr->MPC_Exprs.Expr->
                                      MPC_Expr.EvalNet, inExpr);
    else 
        statsSequence = mMPC_ExprStat(exprPos,
                                      statsSequence,
                                      emptyStat,
                                      commonStatNoSourceFlags,
                                      inExpr->MPC_Expr.EvalNet,
                                      inExpr);
}
 


/**
 * Attaches temporary variable assignement to the sequence 
 * of statements. Returns the produced expression.
 *
 * @param inTmpExpr temporary storage
 * @param inExpr expression to store
 */
static tTree attach_tmp_stat(tTree inLExpr, tTree inRExpr)
{
    tTree newExpr;
    tTree newEvalNet;
    tExprFlags flags;

    flags = inLExpr->MPC_Expr.Flag;
    flags.Lvalue = 0;
    flags.Const = 0;
    flags.Generated = 0;
    flags.Source = 0;

    newEvalNet =
        mMPC_NetList(emptyStat, emptyStat, inLExpr->MPC_Expr.StoreNet);

    if (inLExpr->MPC_Expr.StoreNet != inRExpr->MPC_Expr.StoreNet) {
        newEvalNet =
            mMPC_NetList(newEvalNet, emptyStat,
                         inRExpr->MPC_Expr.StoreNet);
        flags.Asynchr = 0;
    }

    newExpr = mMPC_BinaryExpr(exprPos,
                              flags,
                              "",
                              inLExpr->MPC_Expr.Type,
                              newEvalNet,
                              inLExpr->MPC_Expr.StoreNet,
                              ASSIGN, inLExpr, inRExpr);

    statsSequence = mMPC_ExprStat(exprPos,
                                  statsSequence,
                                  emptyStat,
                                  commonStatNoSourceFlags, newEvalNet, newExpr);


    return newExpr;
}

/**
 * Checks whether designated vector (or scalar) 
 * has layout determined by the type of expression.
 * 
 * For instance, the layout of a vector *(a[]) could not be determined from the vector type.
 *
 * @param inExpr vector designator
 */
 static int is_regular_layout(tTree inExpr)
 {
    if(Tree_IsType(inExpr->MPC_Expr.Type,kMPC_VectorType)) {
        if(Tree_IsType(inExpr, kMPC_UnaryExpr) && (inExpr->MPC_UnaryExpr.OpCode == BRACKETS))
            return is_regular_layout(inExpr->MPC_UnaryExpr.Operand);
        else if(Tree_IsType(inExpr, kMPC_NetCastExpr))
            return is_regular_layout(inExpr->MPC_NetCastExpr.Operand);
        else if(Tree_IsType(inExpr, kMPC_QuaternaryExpr))
            return is_regular_layout(inExpr->MPC_QuaternaryExpr.Operand1);
        else 
            return 0;
    } else 
        return 1;
 }


/**
 * Splits given expression into several ones.
 *
 * The following notation is used below:
 *
 * T(E) -- temporal l-value expression suitable for holding values
 * of E
 * 
 * '*=*' -- simple assignment operator in synchronious assignement
 *  'x=' -- simple or compound assignmentoperator
 *
 * Split use the following rewrite rules:
 *
 * 1) Handles the most general case of assignement:
 *
 *    E(BE(=,L,R)));
 *    ----------------------------------------------------
 *    BE(=,T(R),R); BE(=,T(L),T(R)); BE(=,L,T(L)); E(T(L));
 *
 * 2) Applicable to synchron assignement if R has regular layout
 *    form
 *
 *    E(BE(*=*,L,R))
 *    ----------------------------------------------------
 *    BE(*=*,T(L),R); BE(=,L,T(L)); E(T(L));
 *
 * 3) Applicable to synchron assignments with no parent
 *    
 *    BE(*=*,L,R)
 *    ----------------------------------------------------
 *    BE(=,T(R),R); BE(*=*,L,T(R));
 *
 * 4) Applicable to asynchron assignements 
 *    
 *    E(BE(x=,L,R));
 *    ----------------------------------------------------
 *    BE(=,T(R),R); E(BE(x=,L,T(R)));
 *
 * 5) Applicable to Postfix and Prefix reduction 
 *
 *    E(UE(x:O));
 *    ----------------------------------------------------
 *    BE(=,T(x:O),UE(x:O)); E(T(x:O));
 *
 * 6) Appicable to NetCast expressions
 *
 *    E(NC(O))
 *    ----------------------------------------------------
 *    BE(=,T(O),O); E(NC(T(O));
 *
 * 7) E(NC(LO))
 *    ----------------------------------------------------
 *    BE(=,T(&LO),&LO); E(NC(*T(&LO)))
 *
 * @param inParent expression one level higher then processing one
 * @param inEdge the graph edge to which point to attach
 * @param inExpr processing expression
 */
static void split_expr(tTree inParentExpr, tTree* inEdge)
{
    tTree tmpExpr;
    tTree newExpr1;
    tTree newExpr2;
    tTree inExpr;
    
    inExpr = *inEdge;

    if (inExpr == NoTree) {
    } else if (is_exchange_assign(inExpr)) {
        tTree tmpExpr1;
        tTree tmpExpr2;

        tmpExpr1 = ((inParentExpr == NoTree) && is_regular_layout(inExpr->MPC_BinaryExpr.Loperand)) ?
                   NoTree :
                   construct_tmp_expr(construct_var_decl
                               (inExpr->MPC_BinaryExpr.Loperand),
                               inExpr->MPC_BinaryExpr.Loperand);
            
        
        
        tmpExpr2 = (inExpr->MPC_BinaryExpr.Roperand->MPC_Expr.Flag.Lvalue && 
                    is_regular_layout(inExpr->MPC_BinaryExpr.Roperand)) ?
                   NoTree :
                   construct_tmp_expr(construct_var_decl
                               (inExpr->MPC_BinaryExpr.Roperand),
                               inExpr->MPC_BinaryExpr.Roperand);
    
        if((tmpExpr1 != NoTree) && (tmpExpr2 != NoTree)) {
            newExpr1 = attach_tmp_stat(inExpr->MPC_BinaryExpr.Loperand, tmpExpr1);
            attach_tmp_stat(tmpExpr1, tmpExpr2);
            newExpr2 = attach_tmp_stat(tmpExpr2, inExpr->MPC_BinaryExpr.Roperand);
        } else if((tmpExpr1 != NoTree) && (tmpExpr2 == NoTree)) {
            newExpr1 = attach_tmp_stat(inExpr->MPC_BinaryExpr.Loperand, tmpExpr1);
            newExpr2 = attach_tmp_stat(tmpExpr1, inExpr->MPC_BinaryExpr.Roperand);
        } else if((tmpExpr1 == NoTree) && (tmpExpr2 != NoTree)) {
            newExpr1 = inExpr;
            attach_expr(inExpr);
            newExpr2 = attach_tmp_stat(tmpExpr2, inExpr->MPC_BinaryExpr.Roperand);
            inExpr->MPC_BinaryExpr.Roperand = tmpExpr2;
        } else {
            attach_expr(inExpr);
            newExpr1 = inExpr;
            newExpr2 = inExpr;
        }

        if(tmpExpr1)
            *inEdge = tmpExpr1;

        split_expr(newExpr1, &(newExpr1->MPC_BinaryExpr.Loperand));
        split_expr(newExpr2, &(newExpr2->MPC_BinaryExpr.Roperand));
    } else if (is_crossing_assign(inExpr)) {
        tmpExpr =
            construct_tmp_expr(construct_var_decl
                               (inExpr->MPC_BinaryExpr.Roperand),
                               inExpr->MPC_BinaryExpr.Roperand);

        if(inParentExpr == NoTree) 
            attach_expr(inExpr);
            
        newExpr2 =
            attach_tmp_stat(tmpExpr, inExpr->MPC_BinaryExpr.Roperand);
            
        inExpr->MPC_BinaryExpr.Roperand = tmpExpr;

        split_expr(inExpr, &(inExpr->MPC_BinaryExpr.Loperand));
        split_expr(newExpr2, &(newExpr2->MPC_BinaryExpr.Roperand));
    } else if (Tree_IsType(inExpr, kMPC_UnaryExpr) &&
               IS_LINEAR_OPER(inExpr->MPC_UnaryExpr.OpCode)) {
        tmpExpr = construct_tmp_expr(construct_var_decl(inExpr), inExpr);

        newExpr1 = attach_tmp_stat(tmpExpr, inExpr);
        *inEdge = tmpExpr;
        split_expr(inExpr, &(inExpr->MPC_UnaryExpr.Operand));
    } else if (Tree_IsType(inExpr, kMPC_UnaryExpr) &&
                IS_POSTFIX_REDUCTION(inExpr->MPC_UnaryExpr.OpCode)) {
        tTree tmpExpr1, tmpExpr2;
        
        tmpExpr1 = construct_tmp_expr(construct_var_decl(inExpr), inExpr);
        newExpr1 = attach_tmp_stat(tmpExpr1, inExpr);
        tmpExpr2 = construct_tmp_expr(
            construct_var_decl(inExpr->MPC_UnaryExpr.Operand), 
            inExpr->MPC_UnaryExpr.Operand);
        newExpr2 = attach_tmp_stat(tmpExpr2, 
            inExpr->MPC_UnaryExpr.Operand);
        inExpr->MPC_UnaryExpr.Operand = tmpExpr2;
        *inEdge = tmpExpr1;
        split_expr(newExpr2, &(newExpr2->MPC_BinaryExpr.Roperand));
    } else if(Tree_IsType(inExpr, kMPC_NetCastExpr)){ 
        if(inExpr->MPC_NetCastExpr.Flag.SideEffects) {
            if(inExpr->MPC_NetCastExpr.Flag.Lvalue) {
                tTree addrExpr;
                
                inExpr->MPC_NetCastExpr.Operand = 
                    normalize_lvalue(inExpr->MPC_NetCastExpr.Operand);
                    
                addrExpr = inExpr->MPC_NetCastExpr.Operand->MPC_UnaryExpr.Operand;

                tmpExpr = construct_tmp_expr(
                    construct_var_decl(addrExpr),
                    addrExpr);
                    
                newExpr1 = attach_tmp_stat(tmpExpr, addrExpr);
                inExpr->MPC_NetCastExpr.Operand->MPC_UnaryExpr.Operand = tmpExpr;
                split_expr(newExpr1, &(newExpr1->MPC_BinaryExpr.Roperand));                
            }
            else{                     
                tmpExpr = construct_tmp_expr(
                    construct_var_decl(inExpr->MPC_NetCastExpr.Operand),
                    inExpr->MPC_NetCastExpr.Operand);

                newExpr1 = attach_tmp_stat(tmpExpr, inExpr->MPC_NetCastExpr.Operand);
                inExpr->MPC_NetCastExpr.Operand = tmpExpr;
                split_expr(newExpr1, &(newExpr1->MPC_BinaryExpr.Loperand));                
            }
        } else {
            split_expr(inExpr, &(inExpr->MPC_NetCastExpr.Operand));
        }
    } 
    else {
        if (inParentExpr == NoTree) 
            attach_expr(inExpr);



        switch (inExpr->Kind) {
        case kMPC_Exprs:
            {
                tTree p;
                for (p = inExpr;
                     !Tree_IsType(p, kMPC_FreeNode); p = p->MPC_Exprs.Next)
                    split_expr(p, &(p->MPC_Exprs.Expr));
            }
            break;

        case kMPC_CastExpr:
            split_expr(inExpr, &(inExpr->MPC_CastExpr.Operand));
            break;
        case kMPC_NetCastExpr:
            split_expr(inExpr, &(inExpr->MPC_NetCastExpr.Operand));
            break;
        case kMPC_CoordExpr:
            split_expr(inExpr, &(inExpr->MPC_CoordExpr.Operand));
            break;
        case kMPC_Size_Of_Value:
            split_expr(inExpr, &(inExpr->MPC_Size_Of_Value.Operand));
            break;
        case kMPC_UnaryExpr:
            split_expr(inExpr, &(inExpr->MPC_UnaryExpr.Operand));
            break;
        case kMPC_BinaryExpr:
            split_expr(inExpr, &(inExpr->MPC_BinaryExpr.Loperand));
            split_expr(inExpr, &(inExpr->MPC_BinaryExpr.Roperand));
            break;
        case kMPC_TernaryExpr:
            split_expr(inExpr, &(inExpr->MPC_TernaryExpr.Foperand));
            split_expr(inExpr, &(inExpr->MPC_TernaryExpr.Soperand));
            split_expr(inExpr, &(inExpr->MPC_TernaryExpr.Toperand));
            break;
        case kMPC_QuaternaryExpr:
            split_expr(inExpr, &(inExpr->MPC_QuaternaryExpr.Operand1));
            split_expr(inExpr, &(inExpr->MPC_QuaternaryExpr.Operand2));
            split_expr(inExpr, &(inExpr->MPC_QuaternaryExpr.Operand3));
            split_expr(inExpr, &(inExpr->MPC_QuaternaryExpr.Operand4));
            break;
        case kMPC_CallExpr:
            split_expr(inExpr, &(inExpr->MPC_CallExpr.ArgList));
            split_expr(inExpr, &(inExpr->MPC_CallExpr.NetworkArgList));
            break;
        default:
            break;
        }
    }
}


/*
 * Reevaluates distributions for expression statement.
 *
 * @param tTree inExprStat
 */
static void reevaluate_distrib(tTree inExprStat)
{
    tTree newEvalNet;
    tTree expr;
    
    expr = inExprStat->MPC_ExprStat.Expr;
    
    if(Tree_IsType(expr, kMPC_BinaryExpr)) {    
        newEvalNet = mMPC_NetList(emptyStat, emptyStat, 
                        expr->MPC_BinaryExpr.Loperand->MPC_Expr.StoreNet);

        if (expr->MPC_BinaryExpr.Loperand->MPC_Expr.StoreNet != 
            expr->MPC_BinaryExpr.Roperand->MPC_Expr.StoreNet) 
            newEvalNet = mMPC_NetList(newEvalNet, emptyStat,
                         expr->MPC_BinaryExpr.Roperand->MPC_Expr.StoreNet);
    } else {
        newEvalNet = mMPC_NetList(emptyStat, emptyStat, 
                        expr->MPC_Expr.StoreNet);
    }
    
    inExprStat->MPC_ExprStat.ExecNet = newEvalNet;
}


/*
 * Refactors expression statements.
 *
 * @param inTree processing expression statement.
 */
static tTree refactor_expr_stat(tTree inTree)
{
    tTree compound;

    exprPos = inTree->MPC_ExprStat.Pos;
    compound = mMPC_Compound(exprPos,
                             inTree->MPC_ExprStat.Next,
                             inTree->MPC_ExprStat.Prev,
                             commonStatNoSourceFlags,
                             inTree->MPC_ExprStat.ExecNet,
                             NO, NoTree, NoTree);
    PointerControl(compound);

    tmpVarDecls = mMPC_FreeNode(compound);
    tmpVarTermNode = mMPC_FreeNode(compound);
    statsSequence = mMPC_FreeNode(compound);
    emptyStat = mMPC_FreeNode(compound);
   
    /*Expr_traverse(inTree->MPC_ExprStat.Expr, bring_out_netgrids);*/
    split_expr(NoTree, &(inTree->MPC_ExprStat.Expr));

    compound->MPC_Compound.Decls = tmpVarDecls;
    compound->MPC_Compound.Stats = statsSequence;

#ifdef MPC_DEBUGGER
    if(DBG_CODE)
        compound->MPC_Compound.Origin = inTree;
#endif
    
    {
        tTree stat;
    
        for(stat = statsSequence; 
            !Tree_IsType(stat, kMPC_FreeNode); 
            stat = stat->MPC_ExprStat.Next)
        reevaluate_distrib(stat);
    }

    return compound;
}

/* 
 * Refactors  statements.
 *
 * @param inTree processing expression
 */
static void refactor_proc(tTree* inEdge)
{
    tTree inTree;
    
    inTree = *inEdge;
    
    if (Tree_IsType(inTree, kMPC_ExprStat) 
        && (inTree->MPC_ExprStat.Expr != NoTree)
        && ((!inTree->MPC_ExprStat.Expr->MPC_Expr.Flag.Asynchr)
            || inTree->MPC_ExprStat.Expr->MPC_Expr.Flag.Peculiarity)) {
        tTree prev;
        tTree next;
        tTree me;
        

        prev = inTree->MPC_Stat.Prev;
        next = inTree->MPC_Stat.Next;

        me = refactor_expr_stat(inTree);
        
        if (Tree_IsType(prev, kMPC_Stat))
            prev->MPC_Stat.Next = me;
        
        /*
        else if (Tree_IsType(prev, kMPC_FreeNode))
            prev->MPC_FreeNode.Parent->MPC_Compound.Stats = me;
        */
        if (Tree_IsType(next, kMPC_Stat))
            next->MPC_Stat.Prev = me;
            
        me->MPC_Stat.Prev = prev;
        me->MPC_Stat.Next = next;            
        
        if(inEdge)
            *inEdge = me;
    }
}



/**
 * Rewrites the slection statement according to the
 * following rules.
 *
 * if(E) S;
 * --------------------------------
 * {tE [d$] T; T = E; if(T) S;}
 *
 * if(E) S;
 * --------------------------------
 * {tE [d$] T; T = E; if(T) S;}
 *
 * switch(E) S;
 * --------------------------------
 * {tE [d$] T; T = E; switch(T) S;}
 *
 * @param inTree processing statement
 */
 static tTree rewrite_selection(tTree inTree)
 {
    tTree rtree;
    tTree texpr;
    tTree *expr;
    tTree tvar;
    tTree net;
        
    net = inTree->MPC_Stat.ExecNet->MPC_NetList.Net;
    
    switch(inTree->Kind) {
    case kMPC_If:
        expr = &(inTree->MPC_If.Expr);
    break;
    case kMPC_IfElse:
        expr = &(inTree->MPC_IfElse.Expr);
    break;
    case kMPC_Switch:
        expr = &(inTree->MPC_Switch.Expr);
    break;
    }
           
    exprPos = (*expr)->MPC_Expr.BegPos;
    if(
      ((!(*expr)->MPC_Expr.Flag.Distributed) && 
       (!net->MPC_NetOrSubnet.SingleNode))||
      (*expr)->MPC_Expr.Flag.Peculiarity) {    
        rtree = mMPC_Compound(statPos,
                          NoTree,
                          NoTree,
                          commonStatSourceFlags,
                          inTree->MPC_Stat.ExecNet,
                          NO, NoTree, NoTree);                 
        tmpVarDecls = mMPC_FreeNode(rtree);
        tmpVarTermNode = mMPC_FreeNode(rtree);
        statsSequence = mMPC_FreeNode(rtree);
        emptyStat = mMPC_FreeNode(rtree);
        tvar = construct_var_decl(*expr);
        tvar->MPC_VarDecl.Var->MPC_Var.Distribution = net;
        texpr = construct_tmp_expr(tvar, *expr);
        texpr->MPC_Expr.StoreNet = net;
        texpr->MPC_Expr.EvalNet = inTree->MPC_Stat.ExecNet;
        if(!net->MPC_NetOrSubnet.SingleNode) 
            texpr->MPC_Expr.Flag.Distributed = 1;
        attach_tmp_stat(texpr, *expr);
            *expr = texpr;
        statsSequence->MPC_Stat.Next = inTree;
        inTree->MPC_Stat.Prev = statsSequence;
        inTree->MPC_Stat.Next = emptyStat;
        rtree->MPC_Compound.Decls = tmpVarDecls;
        rtree->MPC_Compound.Stats = statsSequence;         
    } else 
        rtree = NoTree;
    
    return rtree;
}

/*
 * Returns a compound statement for temporary
 * expression evaluation:
 *
 * E
 * -------------
 * {tE [d$] T; T = E; if(!T) break;}
 *
 * @param inDistr d$
 * @param inExpr expression  E
 */
static tTree construct_control_compound(tTree inDistr, tTree inExpr) 
{
    tTree rtree;
    tTree net;
    tTree ifstat;
    tTree ifexpr;
    tTree brstat;
    tTree emptstat;
    tTree tvar;
    tTree texpr;
    tTree term;
        
    exprPos = inExpr->MPC_Expr.BegPos;      
    net = inDistr->MPC_NetList.Net;
    rtree = mMPC_Compound(exprPos,
                     NoTree,
                     NoTree,
                     commonStatSourceFlags,
                     inDistr,
                     NO, NoTree, NoTree);                 
    tmpVarDecls = mMPC_FreeNode(rtree);
    tmpVarTermNode = mMPC_FreeNode(rtree);
    statsSequence = mMPC_FreeNode(rtree);
    emptyStat = mMPC_FreeNode(rtree);
    
    tvar = construct_var_decl(inExpr);
    tvar->MPC_VarDecl.Var->MPC_Var.Distribution = net;
    texpr = construct_tmp_expr(tvar, inExpr);
    texpr->MPC_Expr.StoreNet = net;
    texpr->MPC_Expr.EvalNet = inDistr;
    if(!net->MPC_NetOrSubnet.SingleNode) 
        texpr->MPC_Expr.Flag.Distributed = 1;
    attach_tmp_stat(texpr, inExpr);
    
    ifstat = mMPC_IfElse(exprPos, 
                emptyStat, 
                statsSequence, 
                commonStatNoSourceFlags,
                inDistr, 
                texpr, 
                NoTree,
                NoTree);
    term = mMPC_FreeNode(ifstat);
    emptstat =  mMPC_ExprStat(exprPos, 
                term, 
                term,                 
                commonStatNoSourceFlags,
                inDistr,
                NoTree);
    brstat = mMPC_Break(exprPos, 
                term, 
                term, 
                commonStatNoSourceFlags,
                inDistr);
   
    ifstat->MPC_IfElse.Then = emptstat;
    ifstat->MPC_IfElse.Else = brstat;   
    
    statsSequence->MPC_Stat.Next = ifstat;
    ifstat->MPC_Stat.Prev = statsSequence;
    ifstat->MPC_Stat.Next = emptyStat;
    rtree->MPC_Compound.Decls = tmpVarDecls;
    rtree->MPC_Compound.Stats = statsSequence;         
        
    return rtree;
} 

/*
 * Generates expression 1
 *
 * @param inExpr previous expression
 */
static tTree construct_true_expr(tTree inExpr)
{
    tExprFlags flags;
    
    flags = inExpr->MPC_Expr.Flag;
    flags.Lvalue = 0;
    flags.Const = 1;
    flags.Generated = 0;
    flags.Source = 0;


    return  mMPC_IntConst (exprPos, 
                    flags,
                    "", 
                    inExpr->MPC_Expr.Type, 
                    CONST_NET_LIST, 
                    CONST_NET,
                    "1", 
                    1);

}

/*
 * Rewrites while loop according to the following rule:
 *
 * while(E) S
 * --------------------------------
 * while(1) {{tE [d$] T; T = E; if(!T) break;} S}
 *
 * @param inTree source statement
 */
 static tTree rewrite_while(tTree inTree){
    tTree rtree;
    tTree control;
    tTree execNet;
    tTree comp;
    tTree stat;
    
    execNet = inTree->MPC_Stat.ExecNet;
    
    if(
      ((!inTree->MPC_While.Expr->MPC_Expr.Flag.Distributed) && 
       (!execNet->MPC_NetList.Net->MPC_NetOrSubnet.SingleNode))||
      inTree->MPC_While.Expr->MPC_Expr.Flag.Peculiarity) {    
        control = construct_control_compound(execNet, inTree->MPC_While.Expr);
        comp = mMPC_Compound(statPos,
                     NoTree,
                     NoTree,
                     commonStatSourceFlags,
                     execNet,
                     NO, NoTree, NoTree);                 
        comp->MPC_Compound.Decls = mMPC_FreeNode(comp);  
        comp->MPC_Compound.Stats = control;
        control->MPC_Stat.Prev = mMPC_FreeNode(comp);
        control->MPC_Stat.Next = inTree->MPC_While.Stats;
        inTree->MPC_While.Stats->MPC_Stat.Prev = control;
        inTree->MPC_While.Stats->MPC_Stat.Next = mMPC_FreeNode(comp);
        inTree->MPC_While.Stats = comp;
        inTree->MPC_While.Expr = construct_true_expr(inTree->MPC_While.Expr);                
        comp->MPC_Stat.Prev = mMPC_FreeNode(inTree);
        comp->MPC_Stat.Next = mMPC_FreeNode(inTree);
        
        rtree = inTree;
        rtree->MPC_Stat.Next = NoTree;
        rtree->MPC_Stat.Prev = NoTree;
    } else
        rtree = NoTree;
        
    return rtree;        

 }

/*
 * Rewrites do-while loop according to the following rule:
 *
 * do S while(E)
 * --------------------------------
 * do{{tE [d$] T; T = E; if(!T) break;} S} while(1)
 *
 * @param inTree source statement
 */
 static tTree rewrite_dowhile(tTree inTree){
    tTree rtree;
    tTree control;
    tTree execNet;
    tTree comp;
    tTree stat;
    
    execNet = inTree->MPC_Stat.ExecNet;
    
    if(
      ((!inTree->MPC_DoWhile.Expr->MPC_Expr.Flag.Distributed) && 
       (!execNet->MPC_NetList.Net->MPC_NetOrSubnet.SingleNode))||
      inTree->MPC_DoWhile.Expr->MPC_Expr.Flag.Peculiarity) {    
        control = construct_control_compound(execNet, inTree->MPC_DoWhile.Expr);
        comp = mMPC_Compound(exprPos,
                     NoTree,
                     NoTree,
                     commonStatSourceFlags,
                     execNet,
                     NO, NoTree, NoTree);                 
        comp->MPC_Compound.Decls = mMPC_FreeNode(comp);  
        comp->MPC_Compound.Stats = inTree->MPC_DoWhile.Stats;
        control->MPC_Stat.Next = mMPC_FreeNode(comp);
        control->MPC_Stat.Prev = inTree->MPC_DoWhile.Stats;
        inTree->MPC_DoWhile.Stats->MPC_Stat.Prev = mMPC_FreeNode(comp);
        inTree->MPC_DoWhile.Stats->MPC_Stat.Next = control;
        inTree->MPC_DoWhile.Stats = comp;
        inTree->MPC_DoWhile.Expr = construct_true_expr(inTree->MPC_DoWhile.Expr);                
        comp->MPC_Stat.Prev = mMPC_FreeNode(inTree);
        comp->MPC_Stat.Next = mMPC_FreeNode(inTree);
        
        rtree = inTree;
        rtree->MPC_Stat.Next = NoTree;
        rtree->MPC_Stat.Prev = NoTree;
    } else
        rtree = NoTree;
        
    return rtree;        

 }


/*
 * Rewrites for loop according to the following rule:
 *
 * for(E1; E2; E3) S
 * -----------------------------------------------------
 * E1; for(;;) {{tE2 [d$] T; T = E2; if(!T) break;} S E3;}
 *
 * Expressions are extracted only if necessary.
 *
 * @param inTree source statement
 */
 static tTree rewrite_for(tTree inTree){
    tTree rtree;
    tTree control;
    tTree execNet;
    tTree comp;
    tTree stat;
    tTree* initExpr;
    tTree* condExpr;
    tTree* reinitExpr;
    int initExtract;
    int condExtract;
    int reinitExtract;
    
    execNet = inTree->MPC_Stat.ExecNet;
    
    if(inTree->MPC_For.Init) {
        initExpr = &(inTree->MPC_For.Init);
        initExtract = (*initExpr)->MPC_Expr.Flag.Peculiarity;
    } else
        initExtract = 0;
    
    if(inTree->MPC_For.Cond) {        
        condExpr = &(inTree->MPC_For.Cond);
        condExtract = ((!(*condExpr)->MPC_Expr.Flag.Distributed) && 
           (!execNet->MPC_NetList.Net->MPC_NetOrSubnet.SingleNode))||
        (*condExpr)->MPC_Expr.Flag.Peculiarity ;
    } else 
        condExtract = 0;
    
    if(inTree->MPC_For.Reinit) {
        reinitExpr = &(inTree->MPC_For.Reinit);
        reinitExtract = (*reinitExpr)->MPC_Expr.Flag.Peculiarity;
    } else
        reinitExtract = 0;
    
    
      
    if(condExtract || initExtract || reinitExtract) {    
      
        if(initExtract) {
            rtree = mMPC_ExprStat((*initExpr)->MPC_Expr.BegPos,
                                  inTree,
                                  NoTree,                                  
                                  commonStatNoSourceFlags,
                                  (*initExpr)->MPC_Expr.EvalNet, 
                                  *initExpr);
            inTree->MPC_Stat.Prev = rtree;                                  
            inTree->MPC_Stat.Next = NoTree;
            *initExpr = NoTree;
        } else  {
            rtree = inTree;
            rtree->MPC_Stat.Next = NoTree;
            rtree->MPC_Stat.Prev = NoTree;
        }
        
        if(condExtract || reinitExtract) {
                    
            comp = mMPC_Compound(statPos,
                     NoTree,
                     NoTree,
                     commonStatSourceFlags,
                     execNet,
                     NO, NoTree, NoTree);                 
            if(condExtract) 
                control = construct_control_compound(execNet, inTree->MPC_For.Cond);
                
            comp->MPC_Compound.Decls = mMPC_FreeNode(comp);  
            if(condExtract) {
                comp->MPC_Compound.Stats = control;
                control->MPC_Stat.Prev = mMPC_FreeNode(comp);
                control->MPC_Stat.Next = inTree->MPC_For.Stats;
                inTree->MPC_For.Stats->MPC_Stat.Prev = control;
                *condExpr = NoTree;
            } else {
                comp->MPC_Compound.Stats = inTree;
                inTree->MPC_For.Stats->MPC_Stat.Prev = mMPC_FreeNode(comp);
            }
            
            if(reinitExtract){ 
                inTree->MPC_For.Stats->MPC_Stat.Next = mMPC_ExprStat(
                                  (*reinitExpr)->MPC_Expr.BegPos,
                                  mMPC_FreeNode(comp),
                                  inTree->MPC_For.Stats,                                  
                                  commonStatNoSourceFlags,
                                  (*reinitExpr)->MPC_Expr.EvalNet, 
                                  *reinitExpr);
                *reinitExpr = NoTree;
            }                                  
            else
                inTree->MPC_For.Stats->MPC_Stat.Next = mMPC_FreeNode(comp);
                
                
            inTree->MPC_For.Stats = comp;
        
            comp->MPC_Stat.Prev = mMPC_FreeNode(inTree);
            comp->MPC_Stat.Next = mMPC_FreeNode(inTree);
        }
        
    } else
        rtree = NoTree;
        
    return rtree;        

 }

/*
 * Rewrites the return statement.
 *
 * return E
 * -------------
 * {tE [d$] T; T = E; return T;}
 *
 * @param inStats processing statement
 */
static tTree rewrite_return(tTree inStat) 
{
    tTree rtree;
    tTree expr;
    tTree net;
    tTree ifstat;
    tTree ifexpr;
    tTree brstat;
    tTree emptstat;
    tTree tvar;
    tTree texpr;
    tTree term;
 
    exprPos = statPos;
    expr = inStat->MPC_Return.Expr;       
    net = inStat->MPC_Return.ExecNet->MPC_NetList.Net;
    if((expr == NoTree) || !expr->MPC_Expr.Flag.Peculiarity)
        rtree = NoTree;
    else {
        rtree = mMPC_Compound(statPos,
                     NoTree,
                     NoTree,
                     commonStatSourceFlags,
                     inStat->MPC_Return.ExecNet,
                     NO, NoTree, NoTree);                 
        tmpVarDecls = mMPC_FreeNode(rtree);
        tmpVarTermNode = mMPC_FreeNode(rtree);
        statsSequence = mMPC_FreeNode(rtree);
        emptyStat = mMPC_FreeNode(rtree);
    
        if(expr != NoTree) {
            tvar = construct_var_decl(expr);
            tvar->MPC_VarDecl.Var->MPC_Var.Distribution = net;
            texpr = construct_tmp_expr(tvar, expr);
            texpr->MPC_Expr.StoreNet = net;
            texpr->MPC_Expr.EvalNet = inStat->MPC_Return.ExecNet;
            if(!net->MPC_NetOrSubnet.SingleNode) 
                texpr->MPC_Expr.Flag.Distributed = 1;
        } else
            texpr = NoTree;
        
        statsSequence = mMPC_Return (statPos, 
                statsSequence, 
                emptyStat, 
                commonStatNoSourceFlags,
                inStat->MPC_Return.ExecNet, 
                texpr);
    
        if(expr!= NoTree)
            attach_tmp_stat(texpr, expr);
    
        statsSequence->MPC_Stat.Flag.Source = 1;
        
        rtree->MPC_Compound.Decls = tmpVarDecls;
        rtree->MPC_Compound.Stats = statsSequence;         
    }
        
    return rtree;
} 


/* 
 * Rewrites iteration and selection statements.
 *
 * @param inTree processing expression
 */
static void rewrite_stats(tTree* inEdge)
{
    tTree inTree;
    
    inTree = *inEdge;
    
    if (Tree_IsType(inTree, kMPC_Stat)){
        
        tTree prev;
        tTree next;
        tTree me;
        
        me = NoTree;
        prev = inTree->MPC_Stat.Prev;
        next = inTree->MPC_Stat.Next;
        statPos = inTree->MPC_Stat.Pos;
    
        switch(inTree->Kind) {
        case kMPC_If:
        case kMPC_IfElse:
        case kMPC_Switch:
            me = rewrite_selection(inTree);
        break;
        case kMPC_While:
            me = rewrite_while(inTree);
        break;        
        case kMPC_DoWhile:
            me = rewrite_dowhile(inTree);
        break;        
        case kMPC_For:
            me = rewrite_for(inTree);
        break;        
        case kMPC_Return:
            me = rewrite_return(inTree);
        break;
        }
        
        if (me != NoTree) {           
            tTree tail;
            
            tail =  me;        
            while(tail->MPC_Stat.Next != NoTree)
                tail = tail->MPC_Stat.Next;
             
            
            if (Tree_IsType(prev, kMPC_Stat))
                prev->MPC_Stat.Next = me;
        
            if (Tree_IsType(next, kMPC_Stat))
                next->MPC_Stat.Prev = tail;
            
            me->MPC_Stat.Prev = prev;
            tail->MPC_Stat.Next = next;            
        
            if(inEdge)
                *inEdge = me;
                        
        }
    }
}


/* Traverses expression tree and calls proc function for
 * all traversed nodes.
 *
 * @param inTree processing tree
 * @param proc function to call
 */
static void refactor_traverse(tTree* inEdge, void (*proc) (tTree*))
{
    tTree yyt;

    yyt = *inEdge;

    if (yyt != NoTree) {
        tTree yyn;

        if (Tree_IsType(yyt, kMPC_Stats)) {
            switch (yyt->Kind) {
            case kMPC_ExprStat:
                break;
            case kMPC_If:
                Traverse(MPC_If, Stats);
                break;
            case kMPC_IfElse:
                Traverse(MPC_IfElse, Then);
                Traverse(MPC_IfElse, Else);
                break;
            case kMPC_Switch:
                Traverse(MPC_Switch, Stats);
                break;
            case kMPC_While:
                Traverse(MPC_While, Stats);
                break;
            case kMPC_DoWhile:
                Traverse(MPC_DoWhile, Stats);
                break;
            case kMPC_For:
                Traverse(MPC_For, Stats);
                break;
#ifdef _3XX
            case kMPC_ParScheme:
                Traverse(MPC_ParScheme, Stats);
                break;
#endif /*_3XX*/
            case kMPC_Compound:
                {
                    tTree old_cur_block = cur_block;
                    cur_block = yyt;
                    Traverse(MPC_Compound, Decls);
                    Traverse(MPC_Compound, Stats);
                    cur_block = old_cur_block;
                }
                break;
            default:
                break;
            }


            Traverse(MPC_Stats, Next);
        } else if (Tree_IsType(yyt, kMPC_Decls)) {
            if (yyt->Kind == kMPC_Function) {
                Traverse(MPC_Function, Stats);
            }
            Traverse(MPC_Decls, Next);
        }
        proc(inEdge);

    }
}

/**
 * Splits expression statement into a seqence of expression statements
 * in order to evealute expressions which could not be implemneted by
 * one system of loops.
 *
 * @param inTree expression statement tree
 */
void refactor(tTree inTree)
{
    commonDeclFlags.Source = 0;
    commonStatSourceFlags.Source = 1;    
    commonStatNoSourceFlags.Source = 0;
    refactor_traverse(&(inTree), rewrite_stats);
    refactor_traverse(&(inTree), refactor_proc);
}


/**
 * Constructs corresponding vector type from an array type of [].
 * Subject to remove in the nearest future.
 *
 * @param inTree processing type
 */
static tTree TMP_vectorize_array_type(tTree inTree)
{
    if(Tree_IsType(inTree, kMPC_ArrayType)) {
        tTree tree;
        
        tree =  mMPC_VectorType(inTree->MPC_ArrayType.Pos,
                                inTree->MPC_ArrayType.Next,
                                inTree->MPC_ArrayType.Prev,
                                inTree->MPC_ArrayType.Flag,
                                inTree->MPC_ArrayType.NumberOfComponents,
                                inTree->MPC_ArrayType.Step,
                                TMP_vectorize_array_type(inTree->MPC_ArrayType.ElementType));
        tree->MPC_VectorType.VectorSize = inTree->MPC_ArrayType.ArraySize;
        tree->MPC_VectorType.DynStep = inTree->MPC_ArrayType.DynStep;
        return tree;
    } else 
        return inTree;
    
} 


/**
 * Fixes front-end bug in [] processing. The [] result's
 * type shoul be vector not array.
 * Subject to remove in the nearest future.
 *
 * @param inTree processing node
 */
void TMP_rebracket(tTree inTree)
{
    if(Tree_IsType(inTree, kMPC_UnaryExpr) && (inTree->MPC_UnaryExpr.OpCode == BRACKETS)) 
        inTree->MPC_UnaryExpr.Type = TMP_vectorize_array_type(inTree->MPC_UnaryExpr.Type);    
}

 
 
