/*****************************************************************************/
/* mpCstd - contains MPC functions which stands for malloc, free and others  */
/*          from stdlib called in generated C programs                       */ 
/* Coded by Dm. Arapov for A.Lastovtsky 1995-1996                            */
/* Release 1.10.96                                                           */
/*****************************************************************************/
#include "stdlib.h"
#include "mpCinternal.h"
#include <mpi.h>

#define MPC_MAX_FUN_BODY(prefix, type)\
 type MPC_##prefix##_max(type inA, type inB)\
 {\
    return (inA > inB) ? inA : inB;\
 }
  
#define MPC_MIN_FUN_BODY(prefix, type)\
 type MPC_##prefix##_min(type inA, type inB)\
 {\
    return (inA < inB) ? inA : inB;\
 }


MPC_MAX_FUN_BODY(int, int)
MPC_MAX_FUN_BODY(unsigned_int, unsigned int)
MPC_MAX_FUN_BODY(char, char)
MPC_MAX_FUN_BODY(unsigned_char, unsigned char)
MPC_MAX_FUN_BODY(short, short)
MPC_MAX_FUN_BODY(unsigned_short, unsigned short)
MPC_MAX_FUN_BODY(long, long)
MPC_MAX_FUN_BODY(unsigned_long, unsigned long)
MPC_MAX_FUN_BODY(double, double)
MPC_MAX_FUN_BODY(float, float)
MPC_MAX_FUN_BODY(pointer, void*)

MPC_MIN_FUN_BODY(int, int)
MPC_MIN_FUN_BODY(unsigned_int, unsigned int)
MPC_MIN_FUN_BODY(char, char)
MPC_MIN_FUN_BODY(unsigned_char, unsigned char)
MPC_MIN_FUN_BODY(short, short)
MPC_MIN_FUN_BODY(unsigned_short, unsigned short)
MPC_MIN_FUN_BODY(long, long)
MPC_MIN_FUN_BODY(unsigned_long, unsigned long)
MPC_MIN_FUN_BODY(double, double)
MPC_MIN_FUN_BODY(float, float)
MPC_MIN_FUN_BODY(pointer, void*)



/*********************** MPC_Wtime *******************************************/
double MPC_Wtime( void )
  {
    return MPI_Wtime();
  } /* MPC_Wtime */

/*********************** MPC_Malloc ******************************************/
void* MPC_Malloc(Int size)
  {
    void* ptr;
    ptr = malloc(size);
    if (ptr == NULL) 
      {
        MPC_Error_raise("Can not allocate memory MPC_Malloc returned NULL");
        MPC_Abort(MPC_ERR_NOMEM);
      } /* if */
    return ptr;
  } /* MPC_Malloc */

/*********************** MPC_Free ********************************************/
void MPC_Free(void* ptr)
  {
     free(ptr);
  } /* MPC_Free */

/*********************** MPC_Realloc *****************************************/
void* MPC_Realloc(void* ptr, Int size)
  {
    void* temp = realloc(ptr,size);
    if (temp == NULL) {
      MPC_Error_raise("Can not allocate memory MPC_Realloc returned NULL");
      MPC_Abort(MPC_ERR_NOMEM);
    } /* if */	
    return (temp);
  } /* MPC_Realloc */
  
/*********************** MPC_Processor_name ***** Nov 2001 ***AL*************/
int MPC_Processor_name(char **name, int *resultlen)
  {
    int temp;
    *name = calloc(MPI_MAX_PROCESSOR_NAME, sizeof(char));
    return MPI_Get_processor_name(*name, resultlen?resultlen:&temp);
  } /* MPC_Processor_name */
 

