
/*****************************************************************************/
/* mpCtopo.h - contains all declarations of topology functions, we recommend */
/*             to use.                                                       */
/* Coded by Dm. Arapov for A.Lastovetsky 1995-1996                           */
/* Release 1.10.96                                                           */
/*****************************************************************************/
/*                            Revision history                               */
/*                                                                           */
/* DD-MM-YYYY          Description (Developer Name)                          */
/* ----------          ------------------------------------------------------*/
/* 01-11-2007  --      Consolidation of the MPC CVS rts and MPC3XX rts       */
/*                     into a single rts (Ravi Reddy)                        */
/* 20-11-2007  --      All the definitions moved to this header file         */
/*                     (Ravi Reddy)                                          */
/*****************************************************************************/

#ifndef _MPC_TOPO_INCLUDE_HH_
#define _MPC_TOPO_INCLUDE_HH_

#include "mpCinternal.h"

#define MPC_MAX_ENVTOPOSTRINGLENGTH 32768

int MPC_Processes_num;    /* number of processes in computing space */

/***new***MPC_Hierarchy_descr*****@@@*******************************/

typedef struct MPC_Link_hierarchy_descr {
  Int level;                            /* Level of link hierarchy */
  double linkspeed[MPC_TRANSFER_DATA_RANGE];
                                        /* 1.0/(link speed) on different sizes of */
                                        /* transferred data messages, sec/byte */
  struct MPC_Link_hierarchy_descr *top; /* points to higher link descriptor */
} MPC_Hi_descr;

/**corrected**@@@**/
typedef struct MPC_GROUP_DESCR {
  Int maxproc;                     /* total number of processess */
  Int numproc;                     /* number of hired processes*/ 
  Int scalability;                 /* number of processors */
  Int processor_power;             /* power of one processor */
  double max_nodepower;            /* current power of a free node */
  double* nodeweights;             /* weights of allocated virtual processors */
  Int*    nodenumbers;             /* and their linear numbers in mpC network */
  MPC_Hi_descr *link_descr;        /* points to link descriptor for the group @@@*/
} MPC_Group_descr; 

Int _iMPC_NGroups;
MPC_Group_descr* _iMPC_Groups;

extern Int* _iMPC_recon_net_MASK;
extern Int  _iMPC_recon_net_V;

#define ngroups  _iMPC_NGroups
#define groups   _iMPC_Groups

#define nprocs       _iMPC_NProcs
#define rperformance _iMPC_rperformance

#define recon_net_V     _iMPC_recon_net_V
#define recon_net_MASK  _iMPC_recon_net_MASK

/********* new new *********scheme***********@@@***/

MPC_Topo_graph **MPC_Ptgs;
double (*MPC_Current_mapping_estimation)(MPC_Topo_graph **, const Int *,
                                         Int, Int **, Int **);
Int *MPC_Environment_buffer;
Int MPC_Size_of_environment_buffer;
Int MPC_Local_env_built;
Int MPC_Local_env_updated_by_net_operation;
Int MPC_Local_env_updated_by_recon;

#define MPC_LINK_NODE_TAG (-2)
#define MPC_NO_PARENT_TAG (-1)
#define MPC_TEMP_GROUP_TAG (-10)

MPC_Hi_descr *link_nodes;

Int MPC_The_topology_data(       /* topology info encoding, used by MPC_Net_create */
  MPC_Net* net,                  /* the net which topology is encoded */
  Int* plength,                  /* size of the message, containing  topo info in ints */
  MPC_Topology *atopology,       /* message containing encoded topology info */
  MPC_Topology_free *afree_topology); /* function to free topology info pointed by atopology */

Int MPC_The_strategy( 
  MPC_Node parent,               /* absolute number of the node, which will be parent of the net */
  Int parent_rank,               /* linear number of the parent in the net */
  Int power,                     /* number of nodes to allocate+1 */
  Int topology_length,           /* size of topology info for requested net */
  MPC_Topology topology,         /* the requested net topology */
  Int total_nodes,               /* number of nodes in the computational space */
  Int* map );

Int MPC_The_init_strategy(void); /* used to initialize topology handling module */

Int MPC_The_processors_info(     /* function, which returns phisical reformance info */
    Int* num_of_processors,
    double** relative_performance );

Int MPC_Get_number_of_processors(void);    /* function, which returns the number of phisical processors */

Int MPC_Get_processors_info(     /* function, which returns phisical processors dispatcher' map */
    Int* imap,
    double* map );

Int MPC_Set_processors_info(      /* replaces the dispatcher' map by the given one on the host */
    Int  *map );

#endif /* _MPC_TOPO_INCLUDE_HH_ */

