#ifndef _MPC_ASSIGN_
#define _MPC_ASSIGN_


#define MPC_ERR_ASSIGN_WRONG_POWERS "Operands types are the same but assign isn't broadcast or parallel send"
#define MPC_ERR_ASSIGN_WRONG_RANKS "Illegal operands ranks in assignement"
#define MPC_ERR_ASSIGN_UNDEF_KIND "Undefined assign kind"

enum ASSIGN_KINDS
{
  MPC_PARALLEL_SEND_ASSIGN_KIND,
  MPC_BCAST_ASSIGN_KIND,
  MPC_GATHER_ASSIGN_KIND,
  MPC_SCATTER_ASSIGN_KIND
};

#define MPC_ASSIG_DEBUG

#define STD_MPC_GET_NET_POWER(POWER, NET) POWER = NET.power;
#define MPC_GET_NET_POWER(POWER, NET) STD_MPC_GET_NET_POWER(POWER, NET)

#define MPC_GET_ASSIGN_KIND(KIND, LRANK, RRANK, LNET, RNET)\
{\
    int lpower;\
    int rpower;\
    \
    MPC_GET_NET_POWER(lpower, LNET)\
    MPC_GET_NET_POWER(rpower, RNET)\
    switch(LRANK - RRANK){\
        case 0:\
            if(lpower == rpower)\
                KIND = MPC_PARALLEL_SEND_ASSIGN_KIND;\
            else if(rpower == 1)\
                KIND = MPC_BCAST_ASSIGN_KIND;\
            else\
                MPC_Error_raise(MPC_ERR_ASSIGN_WRONG_POWERS);\
            break;\
        case 1:\
            KIND = MPC_GATHER_ASSIGN_KIND;\
            break;\
        case -1:\
            KIND = MPC_SCATTER_ASSIGN_KIND;\
            break;\
        default:\
            MPC_Error_raise(MPC_ERR_ASSIGN_WRONG_RANKS);\
            break;\
    }\
}


#define NET_PREDICATE_I(NET) if(MPC_Is_member(&NET))

#define NET_PREDICATE_S(NET) if(MPC_Is_member(&NET))

#define NET_PREDICATE_P(NET) if(MPC_Is_member(&NET))

#define MPC_GET_MY_NUM_I(NUM, SUBNET) NUM = SUBNET.rank;

#define MPC_GET_MY_NUM_S(NUM, SUBNET) NUM = SUBNET.rank;

#define MPC_GET_MY_NUM_P(NUM, SUBNET) NUM = SUBNET.rank;

#define MPC_GET_ENCLOSED_NUM_I(NUM, SUBNET_NUM, NET, SUBNET) NUM = SUBNET_NUM;

#define MPC_GET_ENCLOSED_NUM_S(NUM, SUBNET_NUM, NET, SUBNET) NUM = SUBNET.mapping[SUBNET_NUM];

#define MPC_GET_ENCLOSED_NUM_P(NUM, SUBNET_NUM, NET, SUBNET) NUM = MPC_Parent(&NET);

#define MPC_PARALLEL_SEND_ASSIGN_SEND(LTYPE, RTYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
{\
    int to;\
    int from;\
    \
    MPC_GET_MY_NUM_##RNET_KIND(from, RNET)\
    MPC_GET_ENCLOSED_NUM_##LNET_KIND(to, from, NET, LNET)\
    MPC_Send_copy_conv(&NET,\
        RADDR,\
       (NET.rank ==  to) ? LADDR : NULL,\
       (MPC_Dt)(&RTYPE),\
       (MPC_Dt)(&LTYPE),\
       to);\
 }

#define MPC_PARALLEL_SEND_ASSIGN_RECV(TYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
{\
    int to;\
    int from;\
    MPC_GET_MY_NUM_##LNET_KIND(to, LNET)\
    MPC_GET_ENCLOSED_NUM_##RNET_KIND(from, to, NET, RNET)\
    MPC_Recv_copy(&NET,\
        LADDR,\
       (MPC_Dt)(&TYPE),\
       from);\
}

#define MPC_PARALLEL_SEND_ASSIGN(LTYPE, RTYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
    NET_PREDICATE_##RNET_KIND(RNET)\
        MPC_PARALLEL_SEND_ASSIGN_SEND(LTYPE, RTYPE, NET,  LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
    NET_PREDICATE_##LNET_KIND(LNET)\
        MPC_PARALLEL_SEND_ASSIGN_RECV(LTYPE, NET,  LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)


#define MPC_BCAST_ASSIGN_SEND(LTYPE, RTYPE, NET, LNET, LNET_KIND, LADDR, RADDR)\
{\
    int to;\
    int from;\
    \
    for(from = 0; from < LNET.power; from ++){\
        MPC_GET_ENCLOSED_NUM_##LNET_KIND(to, from, NET,LNET)\
        MPC_Send_copy_conv(&NET,\
            RADDR,\
            (NET.rank ==  to) ? LADDR : NULL,\
            (MPC_Dt)(&RTYPE),\
            (MPC_Dt)(&LTYPE),\
            to);\
    }\
 }

#define MPC_BCAST_ASSIGN_RECV(TYPE, NET, RNET, RNET_KIND, LADDR)\
{\
    int from;\
    MPC_GET_ENCLOSED_NUM_##RNET_KIND(from, 0, NET, RNET)\
    MPC_Recv_copy(&NET,\
        LADDR,\
       (MPC_Dt)(&TYPE),\
       from);\
}

#define MPC_BCAST_ASSIGN(LTYPE, RTYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
    NET_PREDICATE_##RNET_KIND(RNET)\
        MPC_BCAST_ASSIGN_SEND(LTYPE, RTYPE, NET, LNET, LNET_KIND, LADDR, RADDR)\
    NET_PREDICATE_##LNET_KIND(LNET)\
        MPC_BCAST_ASSIGN_RECV(LTYPE, NET, RNET, RNET_KIND, LADDR)

#define MPC_SCATTER_ASSIGN_SEND(LTYPE, RTYPE, NET, LNET, LNET_KIND, LADDR, RADDR)\
{\
    int to;\
    int from;\
    \
    for(from = 0; from < LNET.power; from ++){\
        MPC_GET_ENCLOSED_NUM_##LNET_KIND(to, from, NET, LNET)\
        MPC_Send_copy_conv(&NET,\
            ((char*)RADDR) + from * RTYPE.Step * RTYPE.Type->Base.Size,\
            (NET.rank ==  to) ? LADDR : NULL,\
            (MPC_Dt)(RTYPE.Type),\
            (MPC_Dt)(&LTYPE),\
            to);\
    }\
 }

#define MPC_SCATTER_ASSIGN_RECV(TYPE, NET, RNET, RNET_KIND, LADDR)\
{\
    int from;\
    MPC_GET_ENCLOSED_NUM_##RNET_KIND(from, 0, NET, RNET)\
    MPC_Recv_copy(&NET,\
        LADDR,\
       (MPC_Dt)(&TYPE),\
       from);\
 }

#define MPC_SCATTER_ASSIGN(LTYPE, RTYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
    NET_PREDICATE_##RNET_KIND(RNET)\
        MPC_SCATTER_ASSIGN_SEND(LTYPE, RTYPE, NET, LNET, LNET_KIND, LADDR, RADDR)\
    NET_PREDICATE_##LNET_KIND(LNET)\
        MPC_SCATTER_ASSIGN_RECV(LTYPE, NET, RNET, RNET_KIND, LADDR)

#define MPC_GATHER_ASSIGN_SEND(LTYPE, RTYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
{\
    int to;\
    int from;\
    \
    MPC_GET_MY_NUM_##RNET_KIND(from, RNET)\
    MPC_GET_ENCLOSED_NUM_##LNET_KIND(to, 0, NET, LNET)\
    MPC_Send_copy_conv(&NET,\
        RADDR,\
       (NET.rank ==  to) ? ((char*)LADDR) + from * LTYPE.Step * LTYPE.Type->Base.Size: NULL,\
       (MPC_Dt)(&RTYPE),\
       (MPC_Dt)(LTYPE.Type),\
       to);\
 }

#define MPC_GATHER_ASSIGN_RECV(TYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR)\
{\
    int to;\
    int from;\
    for(to = 0; to < RNET.power; to ++){\
        MPC_GET_ENCLOSED_NUM_##RNET_KIND(from, to, NET, RNET)\
        MPC_Recv_copy(&NET,\
         ((char*)LADDR) + to * TYPE.Step * TYPE.Type->Base.Size,\
         (MPC_Dt)(TYPE.Type),\
         from);\
    }\
}

#define MPC_GATHER_ASSIGN(LTYPE, RTYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
    NET_PREDICATE_##RNET_KIND(RNET)\
        MPC_GATHER_ASSIGN_SEND(LTYPE, RTYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR, RADDR)\
    NET_PREDICATE_##LNET_KIND(LNET)\
        MPC_GATHER_ASSIGN_RECV(LTYPE, NET, LNET, LNET_KIND, RNET, RNET_KIND, LADDR)



#define MPC_PARALLEL_ASSIGN(LTYPE, RTYPE, LRANK, RRANK, NET, LNET, RNET, LADDR, RADDR)\
{\
    int kind;\
    \
    MPC_GET_ASSIGN_KIND(kind, LRANK, RRANK, LNET, RNET)\
    switch(kind){\
    case MPC_PARALLEL_SEND_ASSIGN_KIND:\
        MPC_PARALLEL_SEND_ASSIGN(LTYPE, RTYPE, NET, LNET, RNET, LADDR, RADDR)\
    break;\
    case MPC_BCAST_ASSIGN_KIND:\
        MPC_BCAST_ASSIGN(LTYPE, RTYPE, NET, LNET, RNET, LADDR, RADDR)\
    break;\
    case MPC_SCATTER_ASSIGN_KIND:\
        MPC_SCATTER_ASSIGN(LTYPE, RTYPE, NET, LNET, RNET, LADDR, RADDR)\
    break;\
    default:\
        MPC_Error_raise(MPC_ERR_ASSIGN_UNDEF_KIND);\
    break;\
    }\
}


#define MPC_SUBNET_DECLARE(SUBNET, NUM, PARENT, SCOPE, LINE, COLUMN)\
          MPC_Net SUBNET = {MPC_NULL_WEB,\
                    NULL,0,NULL,\
                    SCOPE,\
                    NULL,\
                    MPC_UNINITED_RANK,\
                    0,&PARENT,\
                    NULL,0,MPC_MULTI_ROOT,\
                    NULL,NULL,NULL,NULL,NULL,NULL,\
                    MPC_NOT_BENCH,0,NULL,NULL,\
                    0,#SUBNET,LINE,COLUMN,0};


#define MPC_SUBNET_CREATE(SUPERNET,SUBNET,PREDICATE,IS_FLEXIBLE)\
if (MPC_Is_member (&SUPERNET))\
  {\
    int MPC_ind = 0, MPC_sub_pow = 0, MPC_co[MPC_MAX_COORDS];\
    int *MPC_old_coord = SUPERNET.coord;\
    int *MPC_map;\
    MPC_map = (int *) MPC_Malloc (SUPERNET.power * sizeof (int));\
    SUPERNET.coord = MPC_co;\
    for (; MPC_ind < SUPERNET.power; MPC_ind++)\
      {\
	MPC_Coord_set (&SUPERNET, MPC_ind);\
	if (PREDICATE)\
	  {\
	    MPC_map[MPC_sub_pow] = MPC_ind;\
	    if (MPC_ind == SUPERNET.rank)\
	      {\
		SUBNET.rank = MPC_sub_pow;\
	      }\
	    MPC_sub_pow++;\
	  }\
      }\
    SUPERNET.coord = MPC_old_coord;\
    if (MPC_sub_pow)\
      {\
	MPC_map = (int *) MPC_Realloc (MPC_map, MPC_sub_pow * sizeof (int));\
	SUBNET.mapping = MPC_map;\
	SUBNET.power = MPC_sub_pow;\
      }\
    else\
      {\
	MPC_Error_raise ("Empty subnet");\
      }\
    MPC_Subnet_create (&SUBNET, 1);\
  }
#endif












