
    #include <math.h>
    #include <stdio.h>
    #include <string.h>
    #include <stdlib.h>
    #include <sys/time.h>

    #include <hmpi.h>

    #include "mxm_i.c"
    #include "mxm_benchmark_dgemm.c"
    #include "Execute_scalapack_outer_product.c"

    #include "pdgemm.c"
    #include "pdgemm_impl.c"

    int main(int argc, char **argv)
    {
        int    COORDN = 2;
        int    i, j, k, rc;
        int    *w, *h, *trow, *tcol, colw,
               gme, gsize, myi, myj;

        int    hostnamelen;
        char   hostname[MPI_MAX_PROCESSOR_NAME];

        int    MSIZE = n/kb;

        double gstart, gend;
        double astart, aend;

        HMPI_Group gid;
        MPI_Comm mxmcomm;

        rc = HMPI_Init(&argc, &argv);

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems initializing MPI...Exiting...\n");
        }

        gstart = MPI_Wtime();

        MPI_Get_processor_name(
           hostname,
           &hostnamelen
        );

        /*
        * My global rank
        */
        rc  = MPI_Comm_rank(
                 HMPI_COMM_WORLD,
                 &gme
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems getting rank...Exiting...\n");
        }

        /*
        * Size of the communication universe
        */
        rc  = MPI_Comm_size(
                 HMPI_COMM_WORLD,
                 &gsize
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems getting size...Exiting...\n");
        }

        /*
        * The initial benchmark
        */
        int num_of_computers = HMPI_Get_number_of_computers();
        int pp               = sqrt(num_of_computers);
        int qq               = num_of_computers/pp;

        int benchmark_m      = n/pp;
        int benchmark_n      = n/qq;

        /*
        * Get my coordinates within the (p,q) grid
        */
        if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
        {
            myi = gme;
            myj = 0;
        }
        else
        {
           myi = -1;
           myj = -1;
        }

        /*
        * Refresh the speeds
        */
        { 
           int output_p;
           int input_p[5] = 
               {
                 myi,
                 myj,
                 benchmark_m, 
                 benchmark_n, 
                 kb
           };

           rc = HMPI_Recon_t(
                    &Benchmark_function,
                    input_p,
                    5,
                    &output_p
           ); 
   
           if  (rc != HMPI_OK)
           {
              printf("Panic: HMPI_Recon failed \n");
           }
       }

       HMPI_Debug(1);

       {
           int model_count = 1+1+1+1;
           int *model_params = (int*)malloc(
                                     sizeof(int)
                                     *
                                     model_count
           );
    
           if (model_params == NULL)
           {
              printf("No memory to allocate performanc model parameters\n");
           }          

           model_params[0]                          = pp;
           model_params[1]                          = qq;
           model_params[1+1]                        = n; 
           model_params[1+1+1]                      = kb; 

           rc = HMPI_Group_heuristic_pauto_create(
                    &gid,
                    &MPC_NetType_hetpdgemm,
                    user_heuristic_fn_nonlinear,
                    model_params,
                    model_count
           );

           if (rc != HMPI_OK)
           {
              return rc;
           }

           free(model_params);
       }

       HMPI_Debug(0);

       gend = MPI_Wtime();

       if (HMPI_Is_free())
       {
          goto LEXIT;
       }

       astart = MPI_Wtime();

       mxmcomm = *(MPI_Comm*)HMPI_Get_comm(&gid);

       int p, q;

/*
* Broadcast the optimal grid arrangement
*/
       if (HMPI_Is_parent(&gid))
       {
          int nd;
          int *d;

          HMPI_Group_topology(
              &gid,
              &nd,
              &d
          );

          p = d[0];

          free(d);

          rc = MPI_Bcast(
                  &p,
                  1,
                  MPI_INT,
                  0,
                  mxmcomm
          );
       }
       else
       {
          rc = MPI_Bcast(
                  &p,
                  1,
                  MPI_INT,
                  0,
                  mxmcomm
          );
       }

       if (rc != MPI_SUCCESS)
       {
          printf("Problems broadcasting p\n");
       }

       if (HMPI_Is_parent(&gid))
       {
          int nd;
          int *d;

          HMPI_Group_topology(
              &gid,
              &nd,
              &d
          );

          q = d[1];

          free(d);

          rc = MPI_Bcast(
                  &q,
                  1,
                  MPI_INT,
                  0,
                  mxmcomm
          );
       }
       else
       {
          rc = MPI_Bcast(
                  &q,
                  1,
                  MPI_INT,
                  0,
                  mxmcomm
          );
       }

       if (rc != MPI_SUCCESS)
       {
          printf("Problems broadcasting q\n");
       }

       gme = HMPI_Group_rank(&gid);

       {
          int nc, *c;

          HMPI_Group_coordof(
              &gid,
              &nc,
              &c
          );

          myi = c[0];
          myj = c[1];

          printf("<----MAPPING---->: coordinates=(%d,%d), hostname=%s\n", myi, myj, hostname);

          free(c);
       }

       sprintf(dofpname, "%s_%d_%d.txt", hostname, myi, myj);

       if ((print_computation_times_to_file == 0)
            && (print_communication_times_to_file == 0)
            && (CHECK_ARGUMENTS == 0)
            && (CHECK_RESULTS == 0)
            && (VERBOSE == 0
           )
       )
       {
          ;
       }
       else
       {
          dofp = fopen(dofpname, "a");
       }

       if (HMPI_Is_parent(&gid))
       {
          printf("Optimal arrangement is (%d,%d)\n", p, q);
       }

       w = (int*)malloc(
                 sizeof(int)
                 *
                 (p*q*p*q)
       );

       if (w == NULL)
       {
          printf("(%d): Cannot allocate w\n", gme);
       }

       h = (int*)malloc(
                 sizeof(int)
                 *
                (p*q*p*q)
       );

       if (h == NULL)
       {
          printf("(%d): Cannot allocate h\n", gme);
       }

       trow = (int*)malloc(
                    sizeof(int)
                    *
                   (p*q)
       );

       if (trow == NULL)
       {
          printf("(%d): Cannot allocate trow\n", gme);
       }

       tcol = (int*)malloc(
                    sizeof(int)
                    *
                    (p*q)
       );

       if (tcol == NULL)
       {
          printf("(%d): Cannot allocate tcol\n", gme);
       }

       if (HMPI_Is_parent(&gid))
       {
           rc = HMPI_Partition_matrix_2d(
                    p,
                    q,
                    1,
                    NULL,
                    NULL,
                    NULL,
                    MSIZE,
                    MSIZE,
                    HMPI_COLUMN_BASED,
                    w,
                    h,
                    trow,
                    tcol,
                    NULL,
                    NULL
           );

           if (rc != HMPI_OK)
           {
              printf("(%d): Problems partitioning\n", gme);
           }
       }

       /*
       * Broadcast the new matrix distribution
       * w, h
       */
       rc = MPI_Bcast(
               w,
               p*q*p*q,
               MPI_INT,
               HMPI_Group_parent(&gid),
               mxmcomm
       );

       if (rc != MPI_SUCCESS)
       {
          printf("(%d):Problems broadcasting w\n", gme);
       }

       rc = MPI_Bcast(
               h,
               p*q*p*q,
               MPI_INT,
               HMPI_Group_parent(&gid),
               mxmcomm
       );

       if (rc != MPI_SUCCESS)
       {
          printf("(%d):Problems broadcasting h\n", gme);
       }

       /*
       * Execute the algorithm
       */
       rc    = Execute_scalapack_outer_product_algo(
                      gme, myi, myj,
                      MSIZE,
                      p, q,
                      w, h,
                      mxmcomm
       );

       if (rc != 0)
       {
          printf("Problems executing the algorithm\n");
       }

       {
          free(w);
          free(h);
          free(trow);
          free(tcol);
       }

       aend = MPI_Wtime();

       {
          printf(
             "%s: n=%d, Total computation time=%3f, Total communication time=%3f, MXM execution time(sec)=%3f\n",
             dofpname,
             n,
             total_comp_time,
             total_comm_time,
             (aend - astart)
          );
       }

       /*
        * Close the file for diagnostic information
        */
       if ((print_computation_times_to_file == 0)
           && (print_communication_times_to_file == 0)
            && (CHECK_ARGUMENTS == 0)
            && (CHECK_RESULTS == 0)
            && (VERBOSE == 0
          )
       )
       {
          ;
       }
       else
       {
          fclose(dofp);
       }

LEXIT:
       {
          HMPI_Group_free(
             &gid
          );
       }

       HMPI_Finalize(0);

       return 0;
    }
