
    #include <math.h>
    #include <stdio.h>
    #include <stdlib.h>
    #include <sys/time.h>

    #include <hmpi.h>

    #include "mxm_i.c"
    #include "mxm_benchmark_dgemm.c"
    #include "Execute_scalapack_outer_product.c"

    #include "pdgemm.c"

    int main(int argc, char **argv)
    {
        int    i, j, k, rc;
        int    *w, *h, *trow, *tcol, colw,
               gme, gsize, myi, myj;

        double start, end, *gspeeds = NULL;
        double myspeed, myetime;

        int    hostnamelen;
        char   hostname[MPI_MAX_PROCESSOR_NAME];

        int    MSIZE = n/kb;

        rc = HMPI_Init(&argc, &argv);

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems initializing MPI...Exiting...\n");
        }

        /*
        * My global rank
        */
        rc  = MPI_Comm_rank(
                 HMPI_COMM_WORLD,
                 &gme
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems getting rank...Exiting...\n");
        }

        /*
        * Size of the communication universe
        */
        rc  = MPI_Comm_size(
                 HMPI_COMM_WORLD,
                 &gsize
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems getting size...Exiting...\n");
        }

        /*
        * My global coordinates
        */
        {
           int COORDN = 2;
           const int ppar[] = {p, q};
           int* mycoords = (int*)malloc(
                                 sizeof(int)
                                 *
                                 COORDN
           );

           if (mycoords == NULL)
           {
              printf("%d:Cannot allocate mycoords\n", gme);
           }

           rank2coord(gme, ppar, mycoords);

           myi = mycoords[0];
           myj = mycoords[1];

           free(mycoords);
        }

        /*
        * Open the file for writing diagnostic output
        */
        MPI_Get_processor_name(
           hostname,
           &hostnamelen
        );
        sprintf(dofpname, "%s_%d_%d.txt", hostname, myi, myj);

        if ((print_computation_times_to_file == 0)
            && (print_communication_times_to_file == 0)
            && (CHECK_ARGUMENTS == 0)
            && (CHECK_RESULTS == 0)
            && (VERBOSE == 0
           )
        )
        {
           ;
        }
        else
        {
           dofp = fopen(dofpname, "a");
        }

        w = (int*)malloc(
                  sizeof(int)
                  *
                  (p*q*p*q)
        );

        if (w == NULL)
        {
           printf("(%d): Cannot allocate w\n", gme);
        }

        h = (int*)malloc(
                  sizeof(int)
                  *
                 (p*q*p*q)
        );

        if (h == NULL)
        {
           printf("(%d): Cannot allocate h\n", gme);
        }

        trow = (int*)malloc(
                     sizeof(int)
                     *
                    (p*q)
        );

        if (trow == NULL)
        {
           printf("(%d): Cannot allocate trow\n", gme);
        }

        tcol = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q)
        );

        if (tcol == NULL)
        {
           printf("(%d): Cannot allocate tcol\n", gme);
        }

       /*
       * Execute the benchmark with correct
       * values of m & n now
       */
       int benchmark_m = n/p;
       int benchmark_n = n/q;

       {
          int output_p;
          int input_p[5] =
              {
                 myi,
                 myj,
                 benchmark_m,
                 benchmark_n,
                 kb
          };

          rc = HMPI_Recon_t(
                   &Benchmark_function,
                   input_p,
                   5,
                   &output_p
          );

          if  (rc != HMPI_OK)
          {
            printf("Panic: HMPI_Recon failed \n");
          }
       }

       /*
       * Get the speeds
       */
       gspeeds = (double*)malloc(
                          sizeof(double)
                          *
                          gsize
       );

       if (gspeeds == NULL)
       {
          printf("(%d): Cannot allocate gspeeds\n", gme);
       }

       rc = HMPI_Group_performances(
                HMPI_COMM_WORLD_GROUP,
                gspeeds
       );

       if (rc != HMPI_OK)
       {
          printf(
             "MAIN: Problems getting the performances"
             "...Exiting...\n"
          );
       }

       start = MPI_Wtime();

       HMPI_Group gid;

       if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
       {
          rc = HMPI_Partition_matrix_2d(
                    p,
                    q,
                    1,
                    gspeeds,
                    NULL,
                    NULL,
                    MSIZE,
                    MSIZE,
                    HMPI_COLUMN_BASED,
                    w,
                    h,
                    trow,
                    tcol,
                    NULL,
                    NULL
          );

          if (rc != HMPI_OK)
          {
             printf("(%d): Problems partitioning\n", gme);
          }

          int *model_params;
          int model_count;

          model_count = 1+1+1+1+(p*q*p*q)+(p*q*p*q);
          model_params = (int*)malloc(
                               sizeof(int)
                               *
                               model_count
          );

          if (model_params == NULL)
          {
             printf("No memory to allocate performanc model parameters\n");
          }          

          model_params[0]                      = p; 
          model_params[1]                      = q; 
          model_params[1+1]                    = n; 
          model_params[1+1+1]                  = kb; 

          for (i = 0; i < (p*q*p*q); i++) 
          {         
             model_params[1+1+1+1+i]           = w[i];
             model_params[1+1+1+1+(p*q*p*q)+i] = h[i];
          } 

          rc = HMPI_Group_create(
                   &gid,
                   &MPC_NetType_hetpdgemm,
                   model_params,
                   model_count
          );

          if (rc != HMPI_OK)
          {
             return rc;
          }

          free(model_params);
       }

       if (HMPI_Is_free())
       {
          rc = HMPI_Group_create(
                   &gid,
                   &MPC_NetType_hetpdgemm,
                   NULL,
                   0
          );

          if (rc != HMPI_OK)
          {
             return rc;
          }
       }

       MPI_Comm mxmcomm = *(MPI_Comm*)HMPI_Get_comm(&gid);

       rc = HMPI_Group_performances(
                &gid,
                gspeeds
       );

       if (rc != HMPI_OK)
       {
          printf(
             "MAIN: Problems getting the performances"
             "...Exiting...\n"
          );
       }

       if (HMPI_Is_parent(&gid))
       {
           rc = HMPI_Partition_matrix_2d(
                    p,
                    q,
                    1,
                    gspeeds,
                    NULL,
                    NULL,
                    MSIZE,
                    MSIZE,
                    HMPI_COLUMN_BASED,
                    w,
                    h,
                    trow,
                    tcol,
                    NULL,
                    NULL
           );

           if (rc != HMPI_OK)
           {
              printf("(%d): Problems partitioning\n", gme);
           }
       }

       free(gspeeds);

       /*
       * Broadcast the new matrix distribution
       * w, h
       */
       rc = MPI_Bcast(
               w,
               p*q*p*q,
               MPI_INT,
               HMPI_Group_parent(&gid),
               mxmcomm
       );

       if (rc != MPI_SUCCESS)
       {
          printf("(%d):Problems broadcasting w\n", gme);
       }

       rc = MPI_Bcast(
               h,
               p*q*p*q,
               MPI_INT,
               HMPI_Group_parent(&gid),
               mxmcomm
       );

       if (rc != MPI_SUCCESS)
       {
          printf("(%d):Problems broadcasting h\n", gme);
       }

       gme   = HMPI_Group_rank(&gid);

       {
          int nc, *c;

          HMPI_Group_coordof(
              &gid,
              &nc,
              &c
          );

          myi = c[0];
          myj = c[1];

          free(c);
       }

       /*
       * Print the mapping
       */
       printf("%s(new rank %d): New coordinates (%d,%d)\n", dofpname, gme, myi, myj);

       /*
       * Execute the algorithm
       */
       rc    = Execute_scalapack_outer_product_algo(
                      gme, myi, myj,
                      MSIZE,
                      p, q,
                      w, h,
                      mxmcomm
       );

       if (rc != 0)
       {
          printf("Problems executing the algorithm\n");
       }

       {
          HMPI_Group_free(
             &gid
          );
       }

       end = MPI_Wtime();

       {
          printf(
             "%s: n=%d, Total computation time=%3f, Total communication time=%3f, MXM execution time(sec)=%3f\n",
             dofpname,
             n,
             total_comp_time,
             total_comm_time,
             (end - start)
          );
       }

       {
          free(w);
          free(h);
          free(trow);
          free(tcol);
       }

       /*
        * Close the file for diagnostic information
        */
       if ((print_computation_times_to_file == 0)
           && (print_communication_times_to_file == 0)
            && (CHECK_ARGUMENTS == 0)
            && (CHECK_RESULTS == 0)
            && (VERBOSE == 0
          )
       )
       {
          ;
       }
       else
       {
          fclose(dofp);
       }

       HMPI_Finalize(0);

       return 0;
    }
