
    #ifndef _MXM_HH
    #define _MXM_HH

    #include "parameters.h"

    /*
    * The communication and computation times for each step are 
    * printed in the file for each process
    *
    * The file has the name with the format
    * <hostname>_I_J.txt
    *
    * where (I,J) are the process coordinates
    */
    int    print_computation_times_to_file = 0;  /* 1/0 */
    int    print_communication_times_to_file = 0;  /* 1/0 */

    /*
    * The total communication and computation times are printed 
    * at the end for every process
    */
    double total_comm_time = 0;
    double total_comp_time = 0;

    /*
    * Set the flags below only when the matrices are very small
    * Otherwise the output will be outrageous
    */
    int    CHECK_ARGUMENTS      = 0; /* 1/0 */
    int    CHECK_RESULTS        = 0; /* 1/0 */
    int    VERBOSE              = 0; /* 1/0 */

    /*
    * Tag for MPI messages
    */
    int    MPI_MSG_TAG_FOR_A    = 0xf1;
    int    MPI_MSG_TAG_FOR_B    = 0xf2;

    /*
    * Matrices are initialized with these numbers
    */
    double MXM_CONSTANT_NUMBER1 = 1.5;
    double MXM_CONSTANT_NUMBER2 = 2.5;

    /*
    * Diagnostic output file
    */
    #define  DIAG_FILE_LEN        128
    char   dofpname[DIAG_FILE_LEN];
    FILE   *dofp;

    /*
    * Local modules
    */
    int InitializeMatricesLocal
    (
        int m,
        int n,
        double *a,
        double *b,
        double *c
    );

    int dgemm_local_optimized(
        const int *myh,
        const int *myw,
        const int *kb,
        double    *A,
        double    *B,
        double    *C,
        double *etime
    );

    int dgemm_local_noptimized(
        const int *myh,
        const int *myw,
        const int *kb,
        double    *A,
        double    *B,
        double    *C,
        double *etime
    );

    int Execute_scalapack_outer_product_algo(
        const int me,
        const int myi,
        const int myj,
        const int n,
        const int p,
        const int q,
        const int *w,
        const int *h,
        MPI_Comm  mxmcomm
    );

    int mxm_benchmark_function(
        const int myi,
        const int myj,
        const int myh,
        const int myw,
        const int kb,
        double *etime
    );

    void rank2coord
    (
       int pnum,
       const int *ppar,
       int *pcoord
    );

    int Print_rectangle_2d_local(
        int p,
        int q,
        int m,
        int n,
        const int *w,
        const int *h
    );

   #endif

