
    #include <math.h>
    #include <stdio.h>
    #include <stdlib.h>
    #include <string.h>
    #include <sys/time.h>

    #include <hmpi.h>

    #include "mxm_i.c"
    #include "mxm_benchmark_dgemm.c"
    #include "Execute_scalapack_outer_product.c"

    int main(int argc, char **argv)
    {
        int    i, j, k, rc;
        int    *w, *h, *trow, *tcol, colw,
               gme, gsize, myi, myj;

        double start, end, *gspeeds = NULL;
        double myspeed, myetime;

        int    hostnamelen;
        char   hostname[MPI_MAX_PROCESSOR_NAME];

        int    MSIZE = n/kb;

        rc = MPI_Init(&argc, &argv);

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems initializing MPI...Exiting...\n");
        }

        /*
        * My global rank
        */
        rc  = MPI_Comm_rank(
                 MPI_COMM_WORLD,
                 &gme
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems getting rank...Exiting...\n");
        }

        if (gme >= (p*q))
        {
           MPI_Finalize();
 
           return 0;
        }

        /*
        * Size of the communication universe
        */
        rc  = MPI_Comm_size(
                 MPI_COMM_WORLD,
                 &gsize
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems getting size...Exiting...\n");
        }

        /*
        * My global coordinates
        */
        {
           int COORDN = 2;
           const int ppar[] = {p, q};
           int* mycoords = (int*)malloc(
                                 sizeof(int)
                                 *
                                 COORDN
           );

           if (mycoords == NULL)
           {
              printf("%d:Cannot allocate mycoords\n", gme);
           }

           rank2coord(gme, ppar, mycoords);

           myi = mycoords[0];
           myj = mycoords[1];

           free(mycoords);
        }

        /*
        * Open the file for writing diagnostic output
        */
        MPI_Get_processor_name(
           hostname,
           &hostnamelen
        );
        sprintf(dofpname, "%s_%d_%d.txt", hostname, myi, myj);

        if ((print_computation_times_to_file == 0)
            && (print_communication_times_to_file == 0)
            && (CHECK_ARGUMENTS == 0)
            && (CHECK_RESULTS == 0)
            && (VERBOSE == 0
           )
        )
        {
           ;
        }
        else
        {
           dofp = fopen(dofpname, "a");
        }

        /*
        * On homogeneous platforms, we do not worry about speed
        * We assume it is the same for all the processors
        */
        if (homogeneous == 'n')
        {
           /*
           * The initial benchmark
           */
           int benchmark_m = n/p;
           int benchmark_n = n/q;

           /*
           * Ignore the first execution of the benchmark
           */
           rc = mxm_benchmark_function(
                myi, myj,
                benchmark_m, benchmark_n, kb, 
                &myetime
           );

           if (rc != 0)
           {
              printf("(%d): Problems executing benchmark function\n", gme);
           }
   
           rc = mxm_benchmark_function(
                myi, myj,
                benchmark_m, benchmark_n, kb, 
                &myetime
           );

           if (rc != 0)
           {
              printf("(%d): Problems executing benchmark function\n", gme);
           }
   
           if (myetime == 0.)
           {
              myspeed         = 0.;
           }
           else
           {   
              myspeed         = (2.0*benchmark_m*0.001*benchmark_n*0.001*kb)/myetime;
	   }

           /*
           * Gather the speeds at process 0
           */
           if (gme == 0)
           {
              gspeeds = (double*)malloc(
                                 sizeof(double)
                                 *
                                 gsize
              );

              if (gspeeds == NULL)
              {
                 printf("(%d): Cannot allocate gspeeds\n", gme);
              }
           }

           rc = MPI_Gather(
                &myspeed,
                1,
                MPI_DOUBLE,
                gspeeds,
                1,
                MPI_DOUBLE,
                0,
                MPI_COMM_WORLD
           );
   
           if (rc != MPI_SUCCESS)
           {
              printf("(%d):Problems gathering all speeds\n", gme);
           }

           /*
           * Print the speeds
           */
           if (gme == 0)
           {
              printf("The speeds are \n");

              for (i = 0; i < p; i++)
              {
                 for (j = 0; j < q; j++)
                 {
                     printf("%f, ", gspeeds[i*q+j]);
                 }
                 printf("\n");
              }

              printf("\n");
           }
        }

        {
           w = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q*p*q)
           );

           if (w == NULL)
           {
              printf("(%d): Cannot allocate w\n", gme);
           }

           h = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q*p*q)
           );

           if (h == NULL)
           {
              printf("(%d): Cannot allocate h\n", gme);
           }

           trow = (int*)malloc(
                        sizeof(int)
                        *
                        (p*q)
           );

           if (trow == NULL)
           {
              printf("(%d): Cannot allocate trow\n", gme);
           }

           tcol = (int*)malloc(
                        sizeof(int)
                        *
                        (p*q)
           );

           if (tcol == NULL)
           {
              printf("(%d): Cannot allocate tcol\n", gme);
           }
        }

        if (gme == 0)
        {
           rc = HMPI_Partition_matrix_2d(
                    p,
                    q,
                    1,
                    gspeeds,
                    NULL,
                    NULL,
                    MSIZE,
                    MSIZE,
                    HMPI_COLUMN_BASED,
                    w,
                    h,
                    trow,
                    tcol,
                    NULL,
                    NULL
           );
   
           if (rc != HMPI_OK)
           {
              printf("(%d): Problems partitioning\n", gme);
           }
       }

       if (homogeneous == 'n')
       {
          if (gme == 0)
          {
             free(gspeeds);
          }
       }

       /*
       * Broadcast the matrix distribution
       * w, h, trow, tcol
       */
       rc = MPI_Bcast(
               w,
               p*q*p*q,
               MPI_INT,
               0,
               MPI_COMM_WORLD
       );

       if (rc != MPI_SUCCESS)
       {
          printf("(%d):Problems broadcasting w\n", gme);
       }

       rc = MPI_Bcast(
               h,
               p*q*p*q,
               MPI_INT,
               0,
               MPI_COMM_WORLD
       );

       if (rc != MPI_SUCCESS)
       {
          printf("(%d):Problems broadcasting h\n", gme);
       }

       rc = MPI_Bcast(
               trow,
               p*q,
               MPI_INT,
               0,
               MPI_COMM_WORLD
       );

       if (rc != MPI_SUCCESS)
       {
          printf("(%d):Problems broadcasting trow\n", gme);
       }

       rc = MPI_Bcast(
               tcol,
               p*q,
               MPI_INT,
               0,
               MPI_COMM_WORLD
       );

       if (rc != MPI_SUCCESS)
       {
          printf("(%d):Problems broadcasting tcol\n", gme);
       }

       if ((VERBOSE > 0) && (gme != 0))
       {
          fprintf(dofp, "Partitioning of matrix is: \n");

          Print_rectangle_2d_local(
               p, q, MSIZE, MSIZE,
               w, h
          );

          fflush(dofp);
       }

       if ((VERBOSE > 0) && (gme == 0))
       {
          printf("Partitioning of matrix is: \n");

          Print_rectangle_2d_local_stdin(
               p, q, MSIZE, MSIZE,
               w, h
          );
       }

       start = MPI_Wtime();

       rc    = Execute_scalapack_outer_product_algo(
                      gme, myi, myj,
                      MSIZE,
                      p, q,
                      w, h, trow, tcol
       );

       if (rc != 0)
       {
          printf("Problems executing the algorithm\n");
       }

       end = MPI_Wtime();

       {
          printf(
             "%s: n=%d, "
             "Total computation time=%3f, "
             "Total communication time(A)=%3f, "
             "Total communication time(B)=%3f, "
             "Total communication time=%3f, "
             "MXM execution time(sec)=%3f\n",
             dofpname,
             n,
             total_comp_time,
             total_A_comm_time,
             total_B_comm_time,
             total_comm_time,
             (end - start)
          );
       }

       {
          free(w);
          free(h);
          free(trow);
          free(tcol);
       }

       /*
        * Close the file for diagnostic information
        */
       if ((print_computation_times_to_file == 0)
           && (print_communication_times_to_file == 0)
            && (CHECK_ARGUMENTS == 0)
            && (CHECK_RESULTS == 0)
            && (VERBOSE == 0
          )
       )
       {
          ;
       }
       else
       {
          fclose(dofp);
       }

       MPI_Finalize();

       return 0;
    }
