
    #include <math.h>
    #include <stdio.h>
    #include <stdlib.h>
    #include <sys/time.h>

    #include "mxm_i.h"

    /*-----------------------------------------------------------------------------------------*/
    /*
    * Timing of what code fragments could be crucial in this benchmark code
    * In the PDGEMM algorithm, for the simplest case, the local matrix c
    * is only allocated once (not for every step k).
    * Local matrices a and b are allocated for all steps k (excepting for the
    * step for which the pivot is the current process)
    */

    int mxm_benchmark_function(
        const int myi,
        const int myj,
        const int myh,
        const int myw,
        const int kb,
        double *etime
    )
    {
        int i, lda, ldb, ldc;
        int m, n, k, rc;

        char TRANSAC   = 'N';
        char *TRANSA   = &TRANSAC;
        char TRANSBC   = 'N';
        char *TRANSB   = &TRANSBC;

        double tstart, tend;
        double *A, *B, *C;
        double  ALPHA  = 1.;
        double  BETA   = 1.;

        if ((myw == 0)
            || (myh == 0
           )
        )
        {
           *etime = 0.;
           return 0;
        }

        m = myh;
        n = myw;
        k = kb;

        C = (double*)malloc(
                     sizeof(double)
                     *
                     (m*n)
        );

        if (C == NULL)
        {
           printf("(%d, %d):Cannot allocate C, m=%d, n=%d \n", myi, myj, m, n);
        }

        for (i = 0; i < (m*n); i++)
        {
            C[i] = 0.0;
        }

        tstart = MPI_Wtime();

        /*
        * Allocate the full matrices
        */
        A = (double*)malloc(
                     sizeof(double)
                     *
                     (m*k)
        );

        if (A == NULL)
        {
           printf("(%d, %d):Cannot allocate A, m=%d, k=%d\n", myi, myj, m, k);
        }

        B = (double*)malloc(
                     sizeof(double)
                     *
                     (k*n)
        );

        if (B == NULL)
        {
           printf("(%d, %d):Cannot allocate B, k=%d, n=%d\n", myi, myj, k, n);
        }

        for (i = 0; i < (m*k); i++)
        {
            A[i] = MXM_CONSTANT_NUMBER1;
        }

        for (i = 0; i < (k*n); i++)
        {
            B[i] = MXM_CONSTANT_NUMBER2;
        }

        /*
        * C := alpha*op( A )*op( B ) + beta*C
        */
        lda = ((m > 1) ? m : 1);
        ldb = ((k > 1) ? k : 1);
        ldc = ((m > 1) ? m : 1);

        dgemm_(
              TRANSA,
              TRANSB,
              &m,
              &n,
              &k,
              &ALPHA,
              A,
              &lda,
              B,
              &ldb,
              &BETA,
              C,
              &ldc
        );

        free(A);
        free(B);

        tend = MPI_Wtime();

        *etime = tend - tstart;
  
        {
           double t1 = 2*myw*0.01;
           double t2 = t1*myh*0.01;
           double t3 = t2*kb*0.01;
           double mflops = t3/(*etime);
           printf("(%d,%d): Time of execution of benchmark code %f, speed=%f\n",
                  myi, myj, *etime, mflops);
        }

        free(C);

        return 0;
    }

    /*-----------------------------------------------------------------------------------------*/
