
    #include <math.h>
    #include <stdio.h>
    #include <string.h>
    #include <stdlib.h>
    #include <sys/time.h>

    #include <hmpi.h>

    #include "mxm_i.c"
    #include "mxm_benchmark_dgemm.c"
    #include "Execute_scalapack_outer_product.c"

    #include "pdgemm.c"

    int main(int argc, char **argv)
    {
        int    COORDN = 2;
        int    i, j, k, rc;
        int    *w, *h, *trow, *tcol, colw,
               gme, gsize, myi, myj;

        double start, end, *gspeeds = NULL;
        double myspeed, myetime;

        int    hostnamelen;
        char   hostname[MPI_MAX_PROCESSOR_NAME];

        int    MSIZE = n/kb;

        rc = HMPI_Init(&argc, &argv);

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems initializing MPI...Exiting...\n");
        }

        /*
        * My global rank
        */
        rc  = MPI_Comm_rank(
                 HMPI_COMM_WORLD,
                 &gme
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems getting rank...Exiting...\n");
        }

        /*
        * Size of the communication universe
        */
        rc  = MPI_Comm_size(
                 HMPI_COMM_WORLD,
                 &gsize
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN:Problems getting size...Exiting...\n");
        }

        MPI_Get_processor_name(
           hostname,
           &hostnamelen
        );

        /*
        * The initial benchmark
        */
        int num_of_computers = HMPI_Get_number_of_computers();
        int pp               = sqrt(num_of_computers);
        int qq               = num_of_computers/pp;

        int benchmark_m      = n/pp;
        int benchmark_n      = n/qq;

        /*
        * Get my coordinates within the (p,q) grid
        */
        if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
        {
            const int ppar[] = {pp, qq};
            int* mycoords = (int*)malloc(
                                  sizeof(int)
                                  *
                                  COORDN
            );

            if (mycoords == NULL)
            {
               printf("%d:Cannot allocate mycoords\n", gme);
            }

            rank2coord(gme, ppar, mycoords);

            myi = mycoords[0];
            myj = mycoords[1];

            free(mycoords);
         }
         else
         {
            myi = -1;
            myj = -1;
         }

         {
           int output_p;
           int input_p[5] =
               {
                 myi,
                 myj,
                 benchmark_m,
                 benchmark_n,
                 kb
           };

           /*
           * Ignore the first run
           */
           rc = HMPI_Recon_t(
                    &Benchmark_function,
                    input_p,
                    5,
                    &output_p
           );

           if  (rc != HMPI_OK)
           {
              printf("Panic: HMPI_Recon failed \n");
           }

           rc = HMPI_Recon_t(
                    &Benchmark_function,
                    input_p,
                    5,
                    &output_p
           );

           if  (rc != HMPI_OK)
           {
              printf("Panic: HMPI_Recon failed \n");
           }
        }

        /*
        * Process arrangements and Number of process arrangements
        * at the host process
        */
        int *parr, nparr;

        MPI_Comm *hmpiproccomm = (MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

        if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
        {
           /*
           * Evaluate the process arrangements
           */
           rc = Get_process_arrangements(
                    gsize,
                    &parr,
                    &nparr             
           );

           if (rc != HMPI_OK)
           {
              return rc;
           }

           printf("Total number of processes %d, Number of process arrangements %d\n", gsize, nparr);
        }

        rc = MPI_Bcast(
                &nparr,
                1,
                MPI_INT,
                HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
                *hmpiproccomm
        );
       
        if (rc != MPI_SUCCESS)
        {
           return rc;
        }

        if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
        {
           ;
        }
        else
        {
           parr = (int*)malloc(
                        sizeof(int)
                        *
                        (nparr*COORDN)
           );

           if (parr == NULL)
           {
              printf("Cannot allocate parr\n");
           }
        }

        rc = MPI_Bcast(
                parr,
                (nparr*COORDN),
                MPI_INT,
                HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
                *hmpiproccomm
        );
       
        if (rc != MPI_SUCCESS)
        {
           return rc;
        }

        if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
        {
           start = MPI_Wtime();
        }

        for (k = 0; k < nparr; k++)
        {
           int p = parr[2*k];
           int q = parr[2*k+1];

           if (((MSIZE%p) != 0) || ((MSIZE%q) != 0))
           {
              continue;
           }

           if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
           {
              printf("Evaluating process arrangement:(%d,%d) ", p, q);
           }

           gspeeds = (double*)malloc(
                              sizeof(double)
                              *
                              gsize
           );

           if (gspeeds == NULL)
           {
              printf("(%d): Cannot allocate gspeeds\n", gme);
           }

           rc = HMPI_Group_performances(
                    HMPI_COMM_WORLD_GROUP,
                    gspeeds
           );
    
           if (rc != HMPI_OK)
           {
              printf(
                 "MAIN: Problems getting the performances"
                 "...Exiting...\n"
              );
           }

           if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
           {
              w = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q*p*q)
              );
   
              if (w == NULL)
              {
                 printf("(%d): Cannot allocate w\n", gme);
              }
   
              h = (int*)malloc(
                        sizeof(int)
                        *
                        (p*q*p*q)
              );

              if (h == NULL)
              {
                 printf("(%d): Cannot allocate h\n", gme);
              }

              trow = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (trow == NULL)
              {
                 printf("(%d): Cannot allocate trow\n", gme);
              }

              tcol = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (tcol == NULL)
              {
                 printf("(%d): Cannot allocate tcol\n", gme);
              }

              rc = HMPI_Partition_matrix_2d(
                    p,
                    q,
                    1,
                    gspeeds,
                    NULL,
                    NULL,
                    MSIZE,
                    MSIZE,
                    HMPI_COLUMN_BASED,
                    w,
                    h,
                    trow,
                    tcol,
                    NULL,
                    NULL
              );
   
              if (rc != HMPI_OK)
              {
                 printf("(%d): Problems partitioning\n", gme);
              }

              int *model_params;
              int model_count;

              model_count = 1+1+1+1+1+1+(p*q*p*q)+(p*q*p*q);
              model_params = (int*)malloc(
                                   sizeof(int)
                                   *
                                   model_count
              );
    
              if (model_params == NULL)
              {
                 printf("No memory to allocate performanc model parameters\n");
              }          

              model_params[0]                          = p; 
              model_params[1]                          = q; 
              model_params[1+1]                        = pp;
              model_params[1+1+1]                      = qq;
              model_params[1+1+1+1]                    = n; 
              model_params[1+1+1+1+1]                  = kb; 

              for (i = 0; i < (p*q*p*q); i++) 
              {         
                 model_params[1+1+1+1+1+1+i]           = w[i];
                 model_params[1+1+1+1+1+1+(p*q*p*q)+i] = h[i];
              } 

              double timeof = HMPI_Timeof(
                                  &MPC_NetType_hetpdgemm,
                                  model_params,
                                  model_count
              );

              free(model_params);
              free(w);
              free(h);
              free(trow);
              free(tcol);

              printf("TIMEOF: %f\n", timeof);
           }

           free(gspeeds);
       }

       free(parr);

       if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
       {
          end = MPI_Wtime();

          printf("Time to evaluate the arrangements %f seconds\n", (end-start));
       }

       HMPI_Finalize(0);

       return 0;
    }
