
    #include <math.h>
    #include <stdio.h>
    #include <stdlib.h>

    #include "mxm_i.h"

   /*-----------------------------------------------------*/

   int 
   Test_size_i
   (
       int gsize,
       int numcoord,
       const int *dim
   )
   {
       int i;
       int product = 1;

       for (i = 0; i < numcoord; i++)
       {
           product = product*dim[i];
       }

       if (product > gsize)
       {
          return 1;
       }

       return 0;
   }

    /*----------------------------------------------------------------------------*/

   int
   Get_process_arrangements_recursively_i
   (
       int d,
       int gsize,
       int numcoord,
       int *coord,
       int **parr,
       int *nparr,
       int *max_nparr
   )
   {
       int i, j, y, rc;
       int totalp;
       int *modelp, paramc;

       if (d > numcoord)
       {
          return 0;
       }

       for (y = 1; y <= gsize; y++)
       {
           coord[d] = y;

           if ((d+1) == numcoord)
           {
              int global_size;
              int only_one_proc;

              /*
               * We have a combination here
               * Test if the combination against
               * the number of processes available.
               */
              int greater = Test_size_i(
                                gsize,
                                numcoord,
                                coord
              );

              if (greater)
              {
                 continue;
              }

              for (i = 0; i < numcoord; i++)
              {
                 (*parr)[(*nparr)*numcoord + i] = coord[i];
              }

              (*nparr)++;

              if ((*nparr) >= (*max_nparr))
              {
                 int tmp_max_nparr = (*max_nparr)*numcoord;
                 int *tmp_parr = (*parr);

                 (*max_nparr) *= 2;

                 (*parr) = (int*)malloc(
                                 sizeof(int)
                                 *
                                 numcoord
                                 *
                                 (*max_nparr)
                 );

                 if ((*parr) == NULL)
                 {
                    return MPC_ERR_NOMEM;
                 }
 
                 memcpy((*parr), tmp_parr, tmp_max_nparr*sizeof(int));

                 free(tmp_parr);
              }

              continue;
           }

           rc = Get_process_arrangements_recursively_i(
                    d+1,
                    gsize,
                    numcoord,
                    coord,
                    parr,
                    nparr,
                    max_nparr
            );

           if (rc != HMPI_OK)
           {
              return rc;
           }
       }

       return HMPI_OK;
   }

    /*----------------------------------------------------------------------------*/

   int
   Get_process_arrangements
   (
       const int gsize,
       int **parr,
       int *nparr
   )
   {
       int i, j, rc;
       int numcoord = 2, *coord;
       int max_nparr = MAX_INITIAL_PROC_ARRMTS;

       *nparr = 0;

       coord = (int*)malloc(
                     sizeof(int)
                     *
                     numcoord
       );

       if (coord == NULL)
       {
          return MPC_ERR_NOMEM;
       }

       (*parr) = (int*)malloc(
                       sizeof(int)
                       *
                       numcoord
                       *
                       max_nparr
       );

       if ((*parr) == NULL)
       {
          return MPC_ERR_NOMEM;
       }

       rc = Get_process_arrangements_recursively_i(
                0,
                gsize,
                numcoord,
                coord,
                parr,
                nparr,
                &max_nparr
       );

       free(coord);

       return rc;
   }

    /*----------------------------------------------------------------------------*/

    void rank2coord
    (
       int pnum,
       const int *ppar,
       int *pcoord
    )
    {
       int tmp;
       tmp =  * (ppar + 1);
        * pcoord = pnum / tmp;
       pnum = pnum % tmp;
        * (pcoord + 1) = pnum;
    }

    /*----------------------------------------------------------------------------*/

    int InitializeMatrices
    (
        int m,
        int n,
        int k,
        double *a,
        double *b,
        double *c
    )
    {
        int i;

        for (i = 0; i < (m*k); i++)
        {
            a[i] = MXM_CONSTANT_NUMBER1;
        }

        for (i = 0; i < (k*n); i++)
        {
            b[i] = MXM_CONSTANT_NUMBER2;
        }

        for (i = 0; i < (m*n); i++)
        {
            c[i] = 0.0;
        }

        return HMPI_OK;
    }

    /*----------------------------------------------------------------------------*/

    int Print_rectangle_2d_local(
        int p,
        int q,
        int m,
        int n,
        const int *w,
        const int *h
    )
    {
       int i, j, k, l;

       fprintf(dofp, "The widths of rectangles are:\n");

       /*
       * COLUMN BASED; HENCE ONLY widths of first row are enough
       */
       {
           for (j = 0; j < q; j++)
           {
               fprintf(dofp, "%d ", w[HMPI_RECT_INDEX(0, j, 0, j, p, q)]);
           }
           fprintf(dofp, "\n");
       }

       fprintf(dofp, "\n");

       fprintf(dofp, "The heights of rectangles are:\n");

       for (i = 0; i < p; i++)
       {
           for (j = 0; j < q; j++)
           {
               for (k = 0; k < p; k++)
               {
                   for (l = 0; l < q; l++)
                   {
                      fprintf(dofp, "%d ", h[HMPI_RECT_INDEX(i, j, k, l, p, q)]);
                   }
               }
               fprintf(dofp, "\n");
           }
       }

       fprintf(dofp, "\n");

       return 0;
    }

    /*----------------------------------------------------------------------------*/

    int Print_rectangle_2d_local_stdin(
        int p,
        int q,
        int m,
        int n,
        const int *w,
        const int *h
    )
    {
       int i, j, k, l;

       printf("The widths of rectangles are:\n");

       /*
       * COLUMN BASED; HENCE ONLY widths of first row are enough
       */
       {
           for (j = 0; j < q; j++)
           {
               printf("%d ", w[HMPI_RECT_INDEX(0, j, 0, j, p, q)]);
           }
           printf("\n");
       }

       printf("\n");

       printf("The heights of rectangles are:\n");

       for (i = 0; i < p; i++)
       {
           for (j = 0; j < q; j++)
           {
               for (k = 0; k < p; k++)
               {
                   for (l = 0; l < q; l++)
                   {
                      printf("%d ", h[HMPI_RECT_INDEX(i, j, k, l, p, q)]);
                   }
               }
               printf("\n");
           }
       }

       printf("\n");

       return 0;
    }

    /*----------------------------------------------------------------------------*/

