
#ifndef __HSCAL_CB_CGEMV_HH
#define __HSCAL_CB_CGEMV_HH



#include "../../../include/hscalapack.h"

int hscal_cgemv(
    char *trans, 
    int *m, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
)
{
    int info;
    int lenx, leny;
    hscallogical noconj;
    hscalcomplex *alpha, *beta;

    hscaldoublereal tm, tn;

    *tcomp = 0.;
    info = 0;

    if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")
	    ) {
	info = 1;
    } else if (*m < 0) {
	info = 2;
    } else if (*n < 0) {
	info = 3;
    } else if (*lda < max(1,*m)) {
	info = 6;
    } else if (*incx == 0) {
	info = 8;
    } else if (*incy == 0) {
	info = 11;
    }

    if (info != 0) {
	return HSCAL_SUCCESS;
    }

    alpha = (hscalcomplex*)calpha;
    beta = (hscalcomplex*)cbeta;

    if (*m == 0 || *n == 0 || alpha->r == 0.f && alpha->i == 0.f && (beta->r 
	    == 1.f && beta->i == 0.f)) {
	return HSCAL_SUCCESS;
    }

    noconj = hscal_lsame(trans, "T");

    if (hscal_lsame(trans, "N")) {
	lenx = *n;
	leny = *m;
    } else {
	lenx = *m;
	leny = *n;
    }

    if (beta->r != 1.f || beta->i != 0.f) {
	if (*incy == 1) {
	    if (beta->r == 0.f && beta->i == 0.f) {
                *tcomp += 3*leny;
	    } else {
                *tcomp += 10*leny;
	    }
	} else {
	    if (beta->r == 0.f && beta->i == 0.f) {
                *tcomp += 4*leny;
	    } else {
                *tcomp += 11*leny;
	    }
	}
    }

    if (alpha->r == 0.f && alpha->i == 0.f) {
	return HSCAL_SUCCESS;
    }

    tm = (double)(*m);
    tn = (double)(*n);

    if (hscal_lsame(trans, "N")) {
	if (*incy == 1) {
           *tcomp += (tm * 13 + 11) * tn;
	} else {
           *tcomp += (tm * 14 + 13) * tn;
	}
    } else {
	if (*incx == 1) {
           if (noconj) {
              *tcomp += (tm * 12 + 15) * tn;
           } else {
              *tcomp += (tm * 14 + 15) * tn;
           }
	} else {
           if (noconj) {
              *tcomp += (tm * 13 + 16) * tn;
           } else {
              *tcomp += (tm * 15 + 16) * tn;
           }
	}
    }

    return HSCAL_SUCCESS;
}

#endif

