
#ifndef __HSCAL_CB_CHERK_HH
#define __HSCAL_CB_CHERK_HH



#include "../../../include/hscalapack.h"

int hscal_cherk(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    char   *cbeta,
    int    *ldc, 
    double *tcomp
)
{
    int a_dim1, a_offset, c_dim1, c_offset, i__1, i__2, i__3, i__4, i__5, 
	    i__6;
    int info;
    int i__, j, l;
    int nrowa;
    hscallogical upper;

    hscalreal *alpha = (hscalreal*)calpha;
    hscalreal *beta  = (hscalreal*)cbeta;

    hscaldoublereal tn, tk;

    *tcomp = 0.;

    if (hscal_lsame(trans, "N")) {
	nrowa = *n;
    } else {
	nrowa = *k;
    }
    upper = hscal_lsame(uplo, "U");
    info = 0;
    if (! upper && ! hscal_lsame(uplo, "L")) {
	info = 1;
    } else if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, 
	    "C")) {
	info = 2;
    } else if (*n < 0) {
	info = 3;
    } else if (*k < 0) {
	info = 4;
    } else if (*lda < max(1,nrowa)) {
	info = 7;
    } else if (*ldc < max(1,*n)) {
	info = 10;
    }
    if (info != 0) {
	return HSCAL_SUCCESS;
    }

    if (*n == 0 || (*alpha == 0.f || *k == 0) && *beta == 1.f) {
	return HSCAL_SUCCESS;
    }

    tn = (double)(*n);
    tk = (double)(*k);

    if (*alpha == 0.f) {
	if (upper) {
	    if (*beta == 0.f) {
                *tcomp += (((tn+1)*tn)/2)*4;
	    } else {
                *tcomp += (((tn-1)*tn)/2)*8 + 7*tn;
	    }
	} else {
	    if (*beta == 0.f) {
                *tcomp += 4*(((tn+1)*tn)/2);
	    } else {
                *tcomp += (((tn-1)*tn)/2)*8 + 7*tn;
	    }
	}
	return HSCAL_SUCCESS;
    }

    if (hscal_lsame(trans, "N")) {
	if (upper) {
            if (*beta == 0.f) {
               *tcomp += (((tn+1)*tn)/2)*4;
            } else if (*beta != 1.f) {
               *tcomp += (((tn-1)*tn)/2)*8 + 7*tn;
            } else {
               *tcomp += 7*tn;
            }
            *tcomp += tk*(24*tn + (((tn-1)*tn)/2) * 16);
	} else {
            if (*beta == 0.f) {
               *tcomp += (((tn+1)*tn)/2)*4;
            } else if (*beta != 1.f) {
               *tcomp += (((tn-1)*tn)/2)*8 + 7*tn;
            } else {
               *tcomp += 7*tn;
            }
            *tcomp += tk*(24*tn + (((tn-1)*tn)/2) * 16);
	}
    } else {
	if (upper) {
            *tcomp += tk*17*(((tn-1)*tn)/2);
            if (*beta == 0.f) {
               *tcomp += 6*(((tn-1)*tn)/2);
            } else {
               *tcomp += 12*(((tn-1)*tn)/2);
            }
            *tcomp += tk*16*tn;
            if (*beta == 0.f) {
               *tcomp += 5*tn;
            } else {
               *tcomp += 9*tn;
            }
	} else {
            *tcomp += tk*16*tn;
            if (*beta == 0.f) {
               *tcomp += 5*tn;
            } else {
               *tcomp += 9*tn;
            }
            *tcomp += tk*17*(((tn-1)*tn)/2);
            if (*beta == 0.f) {
               *tcomp += 6*(((tn-1)*tn)/2);
            } else {
               *tcomp += 12*(((tn-1)*tn)/2);
            }
	}
    }

    return HSCAL_SUCCESS;
}

#endif
