
#ifndef __HSCAL_CB_CTRSM_HH
#define __HSCAL_CB_CTRSM_HH



#include "../../../include/hscalapack.h"

int hscal_ctrsm(
    char *side, 
    char *uplo, 
    char *transa, 
    char *diag, 
    int *m, 
    int *n, 
    char *alphac,
    int *lda, 
    int *ldb,
    double *tcomp
)
{
    hscalcomplex c_b1 = {1.f,0.f};
    hscalcomplex *alpha = (hscalcomplex*)alphac;

    int a_dim1, a_offset, b_dim1, b_offset, i__1, i__2, i__3, i__4, i__5, 
	    i__6, i__7;
    int info;
    int i__, j, k;
    hscallogical lside;
    int nrowa;
    hscallogical upper;
    hscallogical noconj, nounit;
    hscaldoublereal tm, tn;    

    *tcomp = 0.;

    lside = hscal_lsame(side, "L");
    if (lside) {
	nrowa = *m;
    } else {
	nrowa = *n;
    }
    noconj = hscal_lsame(transa, "T");
    nounit = hscal_lsame(diag, "N");
    upper = hscal_lsame(uplo, "U");
    info = 0;
    if (! lside && ! hscal_lsame(side, "R")) {
	info = 1;
    } else if (! upper && ! hscal_lsame(uplo, "L")) {
	info = 2;
    } else if (! hscal_lsame(transa, "N") && ! hscal_lsame(transa,
	     "T") && ! hscal_lsame(transa, "C")) {
	info = 3;
    } else if (! hscal_lsame(diag, "U") && ! hscal_lsame(diag, 
	    "N")) {
	info = 4;
    } else if (*m < 0) {
	info = 5;
    } else if (*n < 0) {
	info = 6;
    } else if (*lda < max(1,nrowa)) {
	info = 9;
    } else if (*ldb < max(1,*m)) {
	info = 11;
    }
    if (info != 0) {
	return HSCAL_SUCCESS;
    }

    if (*n == 0) {
	return HSCAL_SUCCESS;
    }

    tm = (double)(*m);
    tn = (double)(*n);

    if (alpha->r == 0.f && alpha->i == 0.f) {
       *tcomp += tn*tm*4;
	return HSCAL_SUCCESS;
    }

    if (lside) {
	if (hscal_lsame(transa, "N")) {
	    if (upper) {
               if (alpha->r != 1.f || alpha->i != 0.f) {
                  *tcomp += tm*10*tn;
               }
               if (nounit) {
                  *tcomp += tn*18*tm;
               }
               *tcomp += 16*tn*(((tm-1)*tm)/2);
	    } else {
               if (alpha->r != 1.f || alpha->i != 0.f) {
                  *tcomp += tn*tm*10;
               }
               if (nounit) {
                  *tcomp += tn*tm*18;
               }
               *tcomp += tn*16*(((tm-1)*tm)/2);
	    }
	} else {
	    if (upper) {
               *tcomp += tm*10*tn;
               if (noconj) {
                  *tcomp += tn*14*(((tm-1)*tm)/2);
                  if (nounit) {
                     *tcomp += tm*14*tn;
                  }
               } else {
                  *tcomp += 17*tn*(((tm-1)*tm)/2);
                  if (nounit) {
                     *tcomp += tm*17*tn;
                  }
               }
               *tcomp += tm*4*tn;
	    } else {
               *tcomp += tm*10*tn;
               if (noconj) {
                  *tcomp += tn*14*(((tm-1)*tm)/2);
                  if (nounit) {
                     *tcomp += tm*14*tn;
                  }
               } else {
                  *tcomp += 17*tn*(((tm-1)*tm)/2);
                  if (nounit) {
                     *tcomp += tm*17*tn;
                  }
               }
               *tcomp += tm*4*tn;
	    }
	}
    } else {
	if (hscal_lsame(transa, "N")) {
	    if (upper) {
               if (alpha->r != 1.f || alpha->i != 0.f) {
                  *tcomp += tm*10*tn;
               }
               *tcomp += tm*16*(((tn-1)*tn)/2);
               if (nounit) {
                  *tcomp += (14 + tm*10)*tn;
               }
	    } else {
               if (alpha->r != 1.f || alpha->i != 0.f) {
                  *tcomp += tm*10*tn;
               }
               *tcomp += tm*16*(((tn-1)*tn)/2);
               if (nounit) {
                  *tcomp += (14 + tm*10)*tn;
               }
	    }
	} else {
	    if (upper) {
               if (nounit) {
                  if (noconj) {
                     *tcomp += 14*tn;
                  } else {
                     *tcomp += 17*tn;
                  }
                  *tcomp += tm*10*tn;
               }
               if (noconj) {
                  *tcomp += 4*(((tn-1)*tn)/2);
               } else {
                  *tcomp += 7*(((tn-1)*tn)/2);
               }
               *tcomp += tm*14*(((tn-1)*tn)/2);
               if (alpha->r != 1.f || alpha->i != 0.f) {
                  *tcomp += tm*10*tn;
               }
	    } else {
               if (nounit) {
                  if (noconj) {
                     *tcomp += 14*tn;
                  } else {
                     *tcomp += 17*tn;
                  }
                  *tcomp += tm*10*tn;
               }
               if (noconj) {
                  *tcomp += 4*(((tn-1)*tn)/2);
               } else {
                  *tcomp += 7*(((tn-1)*tn)/2);
               }
               *tcomp += tm*14*(((tn-1)*tn)/2);
               if (alpha->r != 1.f || alpha->i != 0.f) {
                  *tcomp += tm*10*tn;
               }
	    }
	}
    }
    return HSCAL_SUCCESS;
}

#endif
