
#ifndef __HSCAL_CB_SGEMM_HH
#define __HSCAL_CB_SGEMM_HH



#include "../../../include/hscalapack.h"

int hscal_sgemm(
    char   *transa, 
    char   *transb, 
    int    *m, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb, 
    char   *cbeta, 
    int    *ldc,
    double *tcomp
)
{
    int a_dim1, a_offset, b_dim1, b_offset, c_dim1, c_offset, i__1, i__2, 
	    i__3;

    hscalreal *alpha = (hscalreal*)calpha;
    hscalreal *beta  = (hscalreal*)cbeta;

    hscaldoublereal tm, tn, tk;

    int info;
    hscallogical nota, notb;
    int i__, j, l, ncola;
    int nrowa, nrowb;

    nota = hscal_lsame(transa, "N");
    notb = hscal_lsame(transb, "N");
    if (nota) {
	nrowa = *m;
	ncola = *k;
    } else {
	nrowa = *k;
	ncola = *m;
    }
    if (notb) {
	nrowb = *k;
    } else {
	nrowb = *n;
    }

    info = 0;
    *tcomp = 0.;

    if (! nota && ! hscal_lsame(transa, "C") && ! hscal_lsame(
	    transa, "T")) {
	info = 1;
    } else if (! notb && ! hscal_lsame(transb, "C") && ! 
	    hscal_lsame(transb, "T")) {
	info = 2;
    } else if (*m < 0) {
	info = 3;
    } else if (*n < 0) {
	info = 4;
    } else if (*k < 0) {
	info = 5;
    } else if (*lda < max(1,nrowa)) {
	info = 8;
    } else if (*ldb < max(1,nrowb)) {
	info = 10;
    } else if (*ldc < max(1,*m)) {
	info = 13;
    }
    if (info != 0) {
	return HSCAL_SUCCESS;
    }

    if (*m == 0 || *n == 0 || (*alpha == 0.f || *k == 0) && *beta == 1.f) {
	return HSCAL_SUCCESS;
    }

    tm = (double)(*m);
    tn = (double)(*n);
    tk = (double)(*k);

    if (*alpha == 0.f) {
	if (*beta == 0.f) {
            *tcomp += tn*tm*2;
	} else {
            *tcomp += tn*tm*3;
	}
	return HSCAL_SUCCESS;
    }

    if (notb) {
	if (nota) {
            if (*beta == 0.f) {
               *tcomp += tm*2*tn;
            } else if (*beta != 1.f) {
               *tcomp += tm*3*tn;
            }
            *tcomp += tk*(3 + tm * 6)*tn;
	} else {
            *tcomp += tk*6*tn*tm;
            if (*beta == 0.f) {
               *tcomp += 3*tn*tm;
            } else {
               *tcomp += 5*tn*tm;
            }
	}
    } else {
	if (nota) {
            if (*beta == 0.f) {
               *tcomp += tm*2*tn;
            } else if (*beta != 1.f) {
               *tcomp += tm*3*tn;
            }
            *tcomp += tk*(3 + tm * 6)*tn;
	} else {
            *tcomp += tk*6*tn*tm;
            if (*beta == 0.f) {
               *tcomp += 3*tn*tm;
            } else {
               *tcomp += 5*tn*tm;
            }
	}
    }

    return HSCAL_SUCCESS;
}

#endif

