
#ifndef __HSCAL_CB_SGEMV_HH
#define __HSCAL_CB_SGEMV_HH



#include "../../../include/hscalapack.h"

int hscal_sgemv(
    char *trans, 
    int *m, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
)
{
    int info;
    int lenx, leny;
    float *alpha, *beta;
    hscaldoublereal tm, tn;

    *tcomp = 0.;

    info = 0;

    if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")
	    ) {
	info = 1;
    } else if (*m < 0) {
	info = 2;
    } else if (*n < 0) {
	info = 3;
    } else if (*lda < max(1,*m)) {
	info = 6;
    } else if (*incx == 0) {
	info = 8;
    } else if (*incy == 0) {
	info = 11;
    }

    if (info != 0) {
       return HSCAL_SUCCESS;
    }

    alpha = (float*)calpha;
    beta = (float*)cbeta;

    if (*m == 0 || *n == 0 || *alpha == 0.f && *beta == 1.f) {
       return HSCAL_SUCCESS;
    }

    if (hscal_lsame(trans, "N")) {
	lenx = *n;
	leny = *m;
    } else {
	lenx = *m;
	leny = *n;
    }

    if (*beta != 1.f) {
	if (*incy == 1) {
	    if (*beta == 0.f) {
               *tcomp += leny;
	    } else {
               *tcomp += leny;
	    }
	} else {
	    if (*beta == 0.f) {
               *tcomp += 2*leny;
	    } else {
               *tcomp += 2*leny;
	    }
	}
    }

    if (*alpha == 0.f) {
       return HSCAL_SUCCESS;
    }

    tm = (double)(*m);
    tn = (double)(*n);

    if (hscal_lsame(trans, "N")) {
	if (*incy == 1) {
           *tcomp += (tm*2+3) * tn;
	} else {
           *tcomp += (tm*3+4) * tn;
	}
    } else {
	if (*incx == 1) {
           *tcomp += (tm*2+5) * tn;
	} else {
           *tcomp += (tm*3+6) * tn;
	}
    }

    return HSCAL_SUCCESS;
}

#endif

