
#ifndef __HSCALAPACK_PBCMATADD_HSCALFUNC_HH
#define __HSCALAPACK_PBCMATADD_HSCALFUNC_HH

#include "../../../include/hscalapack.h"

int hscal_pbcmatadd(
    hscalinteger *icontxt, char *mode, hscalinteger *m, 
    hscalinteger *n, hscalcomplex *alpha, hscalinteger *lda, hscalcomplex *beta, 
    hscalinteger *ldb,
    double *tcomp
)
{
    hscalinteger c__1 = 1;

    /* System generated locals */
    hscalinteger a_dim1, a_offset, b_dim1, b_offset, i__1, i__2, i__3, i__4, i__5;
    hscalcomplex q__1, q__2, q__3, q__4;

    /* Local variables */
    hscalinteger i__, j;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PBCMATADD performs the matrix add operation B := alpha*A + beta*B, */
/*  where alpha and beta are scalars, and A and B are m-by-n */
/*  upper/lower trapezoidal matrices, or rectangular matrices. */

/*  Arguments */
/*  ========= */

/*  ICONTXT (input) INTEGER */
/*          ICONTXT is the BLACS mechanism for partitioning communication */
/*          space.  A defining property of a context is that a message in */
/*          a context cannot be sent or received in another context.  The */
/*          BLACS context includes the definition of a grid, and each */
/*          process' coordinates in it. */

/*  MODE    (input) CHARACTER*1 */
/*          Specifies the part of the matrix A, or (conjugate) transposed */
/*          matrix A to be added to the matrix B, */
/*          = 'U':  Upper triangular part */
/*                  up(B) = alpha*up(A) + beta*up(B) */
/*          = 'L':  Lower triangular part */
/*                  lo(B) = alpha*lo(A) + beta*lo(B) */
/*          = 'T':  Transposed matrix A */
/*                  B = alpha*A**T + beta*B */
/*          = 'C':  Conjugate transposed matrix A */
/*                  B = alpha*A**H + beta*B */
/*          Otherwise: B = alpha*A + beta*B */
/*              if M = LDA = LDB: use one BLAS loop */
/*              if MODE = 'V' :   columnwise copy using BLAS if possible */
/*              else :            use double loops */

/*  M       (input) INTEGER */
/*          M specifies the number of columns of the matrix A if */
/*          MODE != 'T'/'C', and it specifies the number of rows of the */
/*          matrix A otherwise.  It also specifies the number of rows of */
/*          the matrix B. M >= 0. */

/*  N       (input) INTEGER */
/*          N specifies the number of rows of the matrix A if */
/*          MODE != 'T'/'C', and it specifies the number of columns of */
/*          the matrix A otherwise. It also specifies the number of */
/*          columns of the matrix B. N >= 0. */

/*  ALPHA   (input) COMPLEX */
/*          ALPHA specifies the scalar alpha. */

/*  LDA     (input) INTEGER */
/*          The leading dimension of the array A.  LDA >= max(1,M) if */
/*          MODE != 'T'/'C'.  And LDA >= max(1,N) if MODE = 'T'/'C'. */

/*  BETA    (input) COMPLEX */
/*          BETA specifies the scalar beta. */

/*  LDB     (input) INTEGER */
/*          The leading dimension of the array B.  LDB >= max(1,M). */

/*  ===================================================================== */

    *tcomp = 0.;

    /* Function Body */
    if (*m <= 0 || *n <= 0 || alpha->r == 0.f && alpha->i == 0.f && (beta->r 
	    == 1.f && beta->i == 0.f)) {
	return HSCAL_SUCCESS;
    }

/*     A is upper triangular or upper trapezoidal, */

    if (hscal_lsame(mode, "U")) {
	if (alpha->r == 0.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 4*i__2;
		}
	    } else {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 12*i__2;
		}
	    }

	} else if (alpha->r == 1.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 6*i__2;
		}
	    } else if (beta->r == 1.f && beta->i == 0.f) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 10*i__2;
		}
	    } else {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 16*i__2;
		}
	    }

	} else {
	    if (beta->r == 0.f && beta->i == 0.f) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 12*i__2;
		}
	    } else if (beta->r == 1.f && beta->i == 0.f) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 16*i__2;
		}
	    } else {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 22*i__2;
		}
	    }
	}

/*     A is lower triangular or upper trapezoidal, */

    } else if (hscal_lsame(mode, "L")) {
	if (alpha->r == 0.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
            if ((*n) >= (*m))
            {
               *tcomp += 4*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 4*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else {
            if ((*n) >= (*m))
            {
               *tcomp += 12*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 12*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    }

	} else if (alpha->r == 1.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
            if ((*n) >= (*m))
            {
               *tcomp += 6*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 6*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else if (beta->r == 1.f && beta->i == 0.f) {
            if ((*n) >= (*m))
            {
               *tcomp += 10*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 10*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else {
            if ((*n) >= (*m))
            {
               *tcomp += 16*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 16*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    }

	} else {
	    if (beta->r == 0.f && beta->i == 0.f) {
            if ((*n) >= (*m))
            {
               *tcomp += 12*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 12*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else if (beta->r == 1.f && beta->i == 0.f) {
            if ((*n) >= (*m))
            {
               *tcomp += 16*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 16*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else {
            if ((*n) >= (*m))
            {
               *tcomp += 22*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 22*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    }
	}

/*     If MODE = 'Transpose' */

    } else if (hscal_lsame(mode, "T")) {
	if (alpha->r == 0.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
            *tcomp += 4*((*n) * (*m));
	    } else {
            *tcomp += 12*((*n) * (*m));
	    }

	} else if (alpha->r == 1.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
            *tcomp += 6*((*n) * (*m));
	    } else if (beta->r == 1.f && beta->i == 0.f) {
            *tcomp += 10*((*n) * (*m));
	    } else {
            *tcomp += 16*((*n) * (*m));
	    }

	} else {
	    if (beta->r == 0.f && beta->i == 0.f) {
            *tcomp += 12*((*n) * (*m));
	    } else if (beta->r == 1.f && beta->i == 0.f) {
            *tcomp += 16*((*n) * (*m));
	    } else {
            *tcomp += 22*((*n) * (*m));
	    }
	}

/*     If MODE = 'Conjugate', */

    } else if (hscal_lsame(mode, "C")) {
	if (alpha->r == 0.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
            *tcomp += 4*((*n) * (*m));
	    } else {
            *tcomp += 12*((*n) * (*m));
	    }

	} else if (alpha->r == 1.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
            *tcomp += 9*((*n) * (*m));
	    } else if (beta->r == 1.f && beta->i == 0.f) {
            *tcomp += 13*((*n) * (*m));
	    } else {
            *tcomp += 17*((*n) * (*m));
	    }

	} else {
	    if (beta->r == 0.f && beta->i == 0.f) {
            *tcomp += 15*((*n) * (*m));
	    } else if (beta->r == 1.f && beta->i == 0.f) {
            *tcomp += 19*((*n) * (*m));
	    } else {
            *tcomp += 25*((*n) * (*m));
	    }
	}

/*     Other cases (for genral matrix additions) */

    } else {
	if (alpha->r == 0.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
            *tcomp += 4*((*n) * (*m));
	    } else {
		if (*m == *ldb) {
		    i__1 = *m * *n;
                {
                double t;
		    hscal_cscal(
                     &i__1, (char*)beta, &c__1,
                     &t
                );
                *tcomp += t;
                }
		} else if (hscal_lsame(mode, "V")) {
		    i__1 = *n;
		    {
                  double t;
			hscal_cscal(
                       m, (char*)beta, &c__1,
                       &t
                  );
                  *tcomp += (t * i__1);
		    }
		} else {
                *tcomp += 12*((*n) * (*m));
		}
	    }

	} else if (alpha->r == 1.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
		if (*m == *lda && *m == *ldb) {
		    i__1 = *m * *n;
                {
                double t;
		    hscal_ccopy(
                     &i__1, &c__1, &c__1,
                     &t
                );
                *tcomp += t;
                }
		} else if (hscal_lsame(mode, "V")) {
		    i__1 = *n;
		    {
                  double t;
			hscal_ccopy(
                       m, &c__1, &c__1,
                       &t
                  );
                  *tcomp += (t * i__1);
		    }
		} else {
                *tcomp += 6*((*n) * (*m));
		}

	    } else if (beta->r == 1.f && beta->i == 0.f) {
            *tcomp += 10*((*n) * (*m));
	    } else {
            *tcomp += 16*((*n) * (*m));
	    }

	} else {
	    if (beta->r == 0.f && beta->i == 0.f) {
            *tcomp += 12*((*n) * (*m));
	    } else if (beta->r == 1.f && beta->i == 0.f) {
		if (*m == *lda && *m == *ldb) {
		    i__1 = *m * *n;
                {
                double t;
		    hscal_caxpy(
                     &i__1, (char*)alpha, &c__1, &c__1, &t);
                *tcomp += t;
                }
		} else if (hscal_lsame(mode, "V")) {
		    i__1 = *n;
		    {
                  double t;
			hscal_caxpy(
                       m, (char*)alpha, &c__1, 
                       &c__1, &t
                  );
                  *tcomp += (t * i__1);
		    }
		} else {
                *tcomp += 16*((*n) * (*m));
		}

	    } else {
            *tcomp += 22*((*n) * (*m));
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of PBCMATADD */

}

#endif
