
#ifndef __HSCALAPACK_PBCVECADD_HSCALFUNC_HH
#define __HSCALAPACK_PBCVECADD_HSCALFUNC_HH

#include "../../../include/hscalapack.h"

int hscal_pbcvecadd(
    hscalinteger *icontxt, char *mode, hscalinteger *n, 
    hscalcomplex *alpha, hscalinteger *incx, hscalcomplex *beta,
    hscalinteger *incy,
    double *tcomp
)
{
    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4;
    hscalcomplex q__1, q__2, q__3, q__4;

    /* Local variables */
    hscalinteger i__, ix, iy;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */

/*     .. */

/*  Purpose */
/*  ======= */

/*  PBCVECADD performs a vector X to be added to Y */
/*    Y := alpha*op(X) + beta*Y, */
/*  where alpha and beta are scalars, and X and Y are n vectors, */
/*  and op(X) = X**H if MODE = 'C', */

/*  Arguments */
/*  ========= */

/*  ICONTXT (input) INTEGER */
/*          ICONTXT is the BLACS mechanism for partitioning communication */
/*          space.  A defining property of a context is that a message in */
/*          a context cannot be sent or received in another context.  The */
/*          BLACS context includes the definition of a grid, and each */
/*          process' coordinates in it. */

/*  MODE   (input) CHARACTER*1 */
/*          Specifies the transposed, or conjugate transposed vector X */
/*          to be added to the vector Y */
/*          = 'C': Conjugate vector X is added for hscalcomplex data set. */
/*                 Y = alpha * X**H + beta * Y */
/*          ELSE : Vector X is added.  Y = alpha*X + beta*Y */
/*                 if MODE = 'V', BLAS routine may be used. */

/*  N       (input) INTEGER */
/*          The number of elements of the vectors X and Y to be added. */
/*          N >= 0. */

/*  ALPHA   (input) COMPLEX */
/*          ALPHA specifies the scalar alpha. */

/*  INCX    (input) INTEGER */
/*          INCX specifies the increment for the elements of X. */
/*          INCX <> 0. */

/*  BETA    (input) COMPLEX */
/*          BETA specifies the scalar beta. */

/*  INCY  - (input) INTEGER */
/*          INCY specifies the increment for the elements of Y. */
/*          INCY <> 0. */

/*  ===================================================================== */

    *tcomp = 0.;

    /* Function Body */
    if (*n <= 0 || alpha->r == 0.f && alpha->i == 0.f && (beta->r == 1.f && 
	    beta->i == 0.f)) {
	return HSCAL_SUCCESS;
    }

    if (alpha->r == 0.f && alpha->i == 0.f) {
	if (beta->r == 0.f && beta->i == 0.f) {
	    if (*incy == 1) {
            *tcomp += 2*(*n);
	    } else {
            *tcomp += 3*(*n);
	    }

	} else {
	    if (hscal_lsame(mode, "V")) {
            {
            double t;
		hscal_cscal(
                 n, (char*)beta, incy,
                 &t
            );
            *tcomp += t;
            }
	    } else if (*incy == 1) {
            *tcomp += 8*(*n);
	    } else {
            *tcomp += 9*(*n);
	    }
	}

    } else if (! hscal_lsame(mode, "C")) {
	if (alpha->r == 1.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
		if (hscal_lsame(mode, "V")) {
                {
                double t;
		    hscal_ccopy(
                     n, incx, incy,
                     &t
                );
                *tcomp += t;
                }
		} else if (*incx == 1 && *incy == 1) {
                *tcomp += 2*(*n);
		} else {
                *tcomp += 4*(*n);
		}

	    } else if (beta->r == 1.f && beta->i == 0.f) {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 4*(*n);
		} else {
                *tcomp += 6*(*n);
		}

	    } else {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 10*(*n);
		} else {
                *tcomp += 12*(*n);
		}
	    }

	} else {
	    if (beta->r == 0.f && beta->i == 0.f) {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 8*(*n);
		} else {
                *tcomp += 4*(*n);
		}

	    } else if (beta->r == 1.f && beta->i == 0.f) {
		if (hscal_lsame(mode, "V")) {
                {
                double t;
		    hscal_caxpy(
                     n, (char*)alpha, incx, incy,
                     &t
                );
                *tcomp += t;
                }
		} else if (*incx == 1 && *incy == 1) {
                *tcomp += 10*(*n);
		} else {
                *tcomp += 12*(*n);
		}

	    } else {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 16*(*n);
		} else {
                *tcomp += 18*(*n);
		}
	    }
	}

    } else {
	if (alpha->r == 1.f && alpha->i == 0.f) {
	    if (beta->r == 0.f && beta->i == 0.f) {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 5*(*n);
		} else {
                *tcomp += 7*(*n);
		}

	    } else if (beta->r == 1.f && beta->i == 0.f) {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 7*(*n);
		} else {
                *tcomp += 9*(*n);
		}

	    } else {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 13*(*n);
		} else {
                *tcomp += 15*(*n);
		}
	    }

	} else {
	    if (beta->r == 0.f && beta->i == 0.f) {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 11*(*n);
		} else {
                *tcomp += 13*(*n);
		}

	    } else if (beta->r == 1.f && beta->i == 0.f) {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 13*(*n);
		} else {
                *tcomp += 15*(*n);
		}

	    } else {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 19*(*n);
		} else {
                *tcomp += 21*(*n);
		}
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of PBCVECADD */

}

#endif
