
#ifndef __HSCALAPACK_PBDMATADD_HSCALFUNC_HH
#define __HSCALAPACK_PBDMATADD_HSCALFUNC_HH

#include "../../../include/hscalapack.h"

int hscal_pbdmatadd(
    hscalinteger *icontxt, char *mode, hscalinteger *m, 
    hscalinteger *n, hscaldoublereal *alpha, hscalinteger *lda, 
    hscaldoublereal *beta, hscalinteger *ldb, 
    double *tcomp
)
{
    hscalinteger c__1 = 1;

    /* System generated locals */
    hscalinteger a_dim1, a_offset, b_dim1, b_offset, i__1, i__2;

    /* Local variables */
    hscalinteger i__, j;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PBDMATADD performs the matrix add operation B := alpha*A + beta*B, */
/*  where alpha and beta are scalars, and A and B are m-by-n */
/*  upper/lower trapezoidal matrices, or rectangular matrices. */

/*  Arguments */
/*  ========= */

/*  ICONTXT (input) hscalinteger */
/*          ICONTXT is the BLACS mechanism for partitioning communication */
/*          space.  A defining property of a context is that a message in */
/*          a context cannot be sent or received in another context.  The */
/*          BLACS context includes the definition of a grid, and each */
/*          process' coordinates in it. */

/*  MODE    (input) CHARACTER*1 */
/*          Specifies the part of the matrix A, or (conjugate) transposed */
/*          matrix A to be added to the matrix B, */
/*          = 'U':  Upper triangular part */
/*                  up(B) = alpha*up(A) + beta*up(B) */
/*          = 'L':  Lower triangular part */
/*                  lo(B) = alpha*lo(A) + beta*lo(B) */
/*          = 'T':  Transposed matrix A */
/*                  B = alpha*A**T + beta*B */
/*          = 'C':  Conjugate transposed matrix A */
/*                  B = alpha*A**H + beta*B */
/*          Otherwise: B = alpha*A + beta*B */
/*              if M = LDA = LDB: use one BLAS loop */
/*              if MODE = 'V' :   columnwise copy using BLAS if possible */
/*              else :            use double loops */

/*  M       (input) hscalinteger */
/*          M specifies the number of columns of the matrix A if */
/*          MODE != 'T'/'C', and it specifies the number of rows of the */
/*          matrix A otherwise.  It also specifies the number of rows of */
/*          the matrix B. M >= 0. */

/*  N       (input) hscalinteger */
/*          N specifies the number of rows of the matrix A if */
/*          MODE != 'T'/'C', and it specifies the number of columns of */
/*          the matrix A otherwise. It also specifies the number of */
/*          columns of the matrix B. N >= 0. */

/*  ALPHA   (input) DOUBLE PRECISION */
/*          ALPHA specifies the scalar alpha. */

/*  LDA     (input) hscalinteger */
/*          The leading dimension of the array A.  LDA >= max(1,M) if */
/*          MODE != 'T'/'C'.  And LDA >= max(1,N) if MODE = 'T'/'C'. */

/*  BETA    (input) DOUBLE PRECISION */
/*          BETA specifies the scalar beta. */

/*  LDB     (input) hscalinteger */
/*          The leading dimension of the array B.  LDB >= max(1,M). */

/*  ===================================================================== */

    *tcomp = 0.;

    /* Function Body */
    if (*m <= 0 || *n <= 0 || *alpha == 0. && *beta == 1.) {
	return HSCAL_SUCCESS;
    }

/*     A is upper triangular or upper trapezoidal, */

    if (hscal_lsame(mode, "U")) {
	if (*alpha == 0.) {
	    if (*beta == 0.) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 2*i__2;
		}
	    } else {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                    *tcomp += 5*i__2;
                }
	    }

	} else if (*alpha == 1.) {
	    if (*beta == 0.) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 4*i__2;
		}
	    } else if (*beta == 1.) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 7*i__2;
		}
	    } else {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 7*i__2;
		}
	    }

	} else {
	    if (*beta == 0.) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 5*i__2;
		}
	    } else if (*beta == 1.) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 8*i__2;
		}
	    } else {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = min(j,*m);
                *tcomp += 9*i__2;
		}
	    }
	}

/*     A is lower triangular or upper trapezoidal, */

    } else if (hscal_lsame(mode, "L")) {
	if (*alpha == 0.) {
	    if (*beta == 0.) {
            if ((*n) >= (*m))
            {
               *tcomp += 2*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 2*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else {
            if ((*n) >= (*m))
            {
               *tcomp += 5*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 5*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    }

	} else if (*alpha == 1.) {
	    if (*beta == 0.) {
            if ((*n) >= (*m))
            {
               *tcomp += 4*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 4*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else if (*beta == 1.) {
            if ((*n) >= (*m))
            {
               *tcomp += 7*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 7*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else {
            if ((*n) >= (*m))
            {
               *tcomp += 8*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 8*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    }

	} else {
	    if (*beta == 0.) {
            if ((*n) >= (*m))
            {
               *tcomp += 5*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 5*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else if (*beta == 1.) {
            if ((*n) >= (*m))
            {
               *tcomp += 8*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 8*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    } else {
            if ((*n) >= (*m))
            {
               *tcomp += 9*(((*m)*(*m+1))/2);
            }
            else
            {
               *tcomp += 9*(((*n) * (*m)) - (((*n)*(*n-1))/2));
            }
	    }
	}

/*     If MODE = 'Transpose'/'Conjugate' */

    } else if (hscal_lsame(mode, "T") || hscal_lsame(mode, "C")) {
	if (*alpha == 0.) {
	    if (*beta == 0.) {
            *tcomp += 2*((*n) * (*m));
	    } else {
            *tcomp += 5*((*n) * (*m));
	    }

	} else if (*alpha == 1.) {
	    if (*beta == 0.) {
            *tcomp += 4*((*n) * (*m));
	    } else if (*beta == 1.) {
            *tcomp += 7*((*n) * (*m));
	    } else {
            *tcomp += 8*((*n) * (*m));
	    }

	} else {
	    if (*beta == 0.) {
            *tcomp += 5*((*n) * (*m));
	    } else if (*beta == 1.) {
            *tcomp += 8*((*n) * (*m));
	    } else {
            *tcomp += 9*((*n) * (*m));
	    }
	}

/*     Other cases (for genral matrix additions) */

    } else {
	if (*alpha == 0.) {
	    if (*beta == 0.) {
            *tcomp += 2*((*n) * (*m));
	    } else {
		if (*m == *ldb) {
		    i__1 = *m * *n;
                {
                double t;
		    hscal_dscal(
                     &i__1, (char*)beta, &c__1,
                     &t
                );
                *tcomp += t;
                }
		} else if (hscal_lsame(mode, "V")) {
		    i__1 = *n;
		    {
                  double t;
			hscal_dscal(
                       m, (char*)beta, &c__1,
                       &t
                  );
                  *tcomp += (t * i__1);
		    }
		} else {
                *tcomp += 5*((*n) * (*m));
		}
	    }

	} else if (*alpha == 1.) {
	    if (*beta == 0.) {
		if (*m == *lda && *m == *ldb) {
		    i__1 = *m * *n;
                {
                double t;
		    hscal_dcopy(
                     &i__1, &c__1, &c__1,
                     &t
                );
                *tcomp += t;
                }
		} else if (hscal_lsame(mode, "V")) {
		    i__1 = *n;
		    {
                  double t;
			hscal_dcopy(
                       m, &c__1, &c__1,
                       &t
                  );
                  *tcomp += (t * i__1);
		    }
		} else {
                *tcomp += 4*((*n) * (*m));
		}

	    } else if (*beta == 1.) {
            *tcomp += 7*((*n) * (*m));
	    } else {
            *tcomp += 8*((*n) * (*m));
	    }

	} else {
	    if (*beta == 0.) {
            *tcomp += 5*((*n) * (*m));
	    } else if (*beta == 1.) {
		if (*m == *lda && *m == *ldb) {
		    i__1 = *m * *n;
                {
                double t;
		    hscal_daxpy(
                     &i__1, (char*)alpha, &c__1, &c__1,
                     &t
                );
                *tcomp += t;
                }
		} else if (hscal_lsame(mode, "V")) {
		    i__1 = *n;
		    {
                  double t;
			hscal_daxpy(
                       m, (char*)alpha, &c__1, &c__1,
                       &t
                  );
                  *tcomp += (t * i__1);
		    }
		} else {
                *tcomp += 8*((*n) * (*m));
		}

	    } else {
            *tcomp += 9*((*n) * (*m));
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of PBDMATADD */

}

#endif
