
#ifndef __HSCALAPACK_PBDTRGET_HSCALFUNC_HH
#define __HSCALAPACK_PBDTRGET_HSCALFUNC_HH

/*
*  Include files
*/
#include "../../../include/hscalapack.h"

int hscal_pbdtrget(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *mnb, hscalinteger *lda, hscalinteger *mcrow,
    hscalinteger *mccol, hscalinteger *igd,
    hscalinteger *myrow, hscalinteger *mycol, hscalinteger *nprow, hscalinteger *npcol,
    double *tcomp, int *tcomm
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2;

    /* Local variables */
    hscalinteger klen, kmod, nlen, kint;
    hscalreal temp;
    hscalinteger nnum, kint2;
    hscalinteger ntlen, kppos;

    HSCALBLACSCONTEXT   * blacsctxt;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PBDTRGET forms a row block of A from scattered row subblocks if */
/*  ADIST = 'R', or forms a column block of A from scattered column */
/*  subblocks,  if ADIST = 'C'. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Variables .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */

/*     if A is a row block, it needs to communicate columnwise. */

    /* Function Body */

    Cblacs_gridinit_nocomm(&blacsctxt, "r", *nprow, *npcol);

    if (hscal_lsame(adist, "R")) {
	kppos = (*nprow + *myrow - *mcrow) % *nprow;
	if (kppos % *igd == 0) {
	    kint = *igd;
	    nlen = *n;
/* Computing MIN */
	    i__1 = *nprow / *igd, i__2 = *mnb - *mccol;
	    nnum = min(i__1,i__2);
	    temp = (hscalreal) nnum;
	    ntlen = *n * nnum;
	    nnum = *igd * nnum;
	    if (kppos >= nnum) {
		goto L30;
	    }
	    kppos %= *nprow;

L10:
	    if (temp > 1.f) {
		kint2 = kint << 1;
		kmod = kppos % kint2;

		if (kmod == 0) {
		    if (kppos + kint < nnum) {
			klen = ntlen - kppos / kint2 * (kint2 / *igd) * *n;
/* Computing MIN */
			i__1 = klen - nlen;
			klen = min(i__1,nlen);
			i__1 = (*myrow + kint) % *nprow;
			nlen += klen;
		    }
		} else {
		    i__1 = (*nprow + *myrow - kint) % *nprow;
		    hscal_Cdgesd2d(blacsctxt, *m, nlen, *lda, i__1, 
		         *mycol,
                         *nprow, *npcol, *myrow, *mycol, tcomp, tcomm
                    );
		    goto L30;
		}

		kint = kint2;
		temp /= 2.f;
		goto L10;
	    }
	}

/*     if A is a column block, it needs to communicate rowwise. */

    } else if (hscal_lsame(adist, "C")) {

	kppos = (*npcol + *mycol - *mccol) % *npcol;
	if (kppos % *igd == 0) {
	    kint = *igd;
	    nlen = *n;
/* Computing MIN */
	    i__1 = *npcol / *igd, i__2 = *mnb - *mcrow;
	    nnum = min(i__1,i__2);
	    temp = (hscalreal) nnum;
	    ntlen = *n * nnum;
	    nnum = *igd * nnum;
	    if (kppos >= nnum) {
		goto L30;
	    }
	    kppos %= *npcol;

L20:
	    if (temp > 1.f) {
		kint2 = kint << 1;
		kmod = kppos % kint2;

		if (kmod == 0) {
		    if (kppos + kint < nnum) {
			klen = ntlen - kppos / kint2 * (kint2 / *igd) * *n;
/* Computing MIN */
			i__1 = klen - nlen;
			klen = min(i__1,nlen);
			i__1 = (*mycol + kint) % *npcol;
			nlen += klen;
		    }
		} else {
		    i__1 = (*npcol + *mycol - kint) % *npcol;
		    hscal_Cdgesd2d(blacsctxt, *m, nlen, *lda, *myrow, i__1,
                         *nprow, *npcol, *myrow, *mycol, tcomp, tcomm
                    );
		    goto L30;
		}

		kint = kint2;
		temp /= 2.f;
		goto L20;
	    }
	}
    }

L30:

    Cblacs_gridexit_nocomm(blacsctxt);

    return HSCAL_SUCCESS;

/*     End of PBDTRGET */

} /* pbdtrget_ */

#endif

