
#ifndef __HSCALAPACK_PBDVECADD_HSCALFUNC_HH
#define __HSCALAPACK_PBDVECADD_HSCALFUNC_HH

#include "../../../include/hscalapack.h"

int hscal_pbdvecadd(
    hscalinteger *icontxt, char *mode, hscalinteger *n, 
    hscaldoublereal *alpha, hscalinteger *incx, hscaldoublereal *beta, 
    hscalinteger *incy,
    double *tcomp
)
{
    /* System generated locals */
    hscalinteger i__1;

    /* Local variables */
    hscalinteger i__, ix, iy;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */

/*     .. */

/*  Purpose */
/*  ======= */

/*  PBDVECADD performs a vector X to be added to Y */
/*    Y := alpha*op(X) + beta*Y, */
/*  where alpha and beta are scalars, and X and Y are n vectors, */
/*  and op(X) = X**H if MODE = 'C', */

/*  Arguments */
/*  ========= */

/*  ICONTXT (input) INTEGER */
/*          ICONTXT is the BLACS mechanism for partitioning communication */
/*          space.  A defining property of a context is that a message in */
/*          a context cannot be sent or received in another context.  The */
/*          BLACS context includes the definition of a grid, and each */
/*          process' coordinates in it. */

/*  MODE   (input) CHARACTER*1 */
/*          Specifies the transposed, or conjugate transposed vector X */
/*          to be added to the vector Y */
/*          = 'C': Conjugate vector X is added for complex data set. */
/*                 Y = alpha * X**H + beta * Y */
/*          ELSE : Vector X is added.  Y = alpha*X + beta*Y */
/*                 if MODE = 'V', BLAS routine may be used. */

/*  N       (input) INTEGER */
/*          The number of elements of the vectors X and Y to be added. */
/*          N >= 0. */

/*  ALPHA   (input) DOUBLE PRECISION */
/*          ALPHA specifies the scalar alpha. */

/*  INCX    (input) INTEGER */
/*          INCX specifies the increment for the elements of X. */
/*          INCX <> 0. */

/*  BETA    (input) DOUBLE PRECISION */
/*          BETA specifies the scalar beta. */

/*  INCY  - (input) INTEGER */
/*          INCY specifies the increment for the elements of Y. */
/*          INCY <> 0. */

/*  ===================================================================== */

    *tcomp = 0.;

    /* Function Body */
    if (*n <= 0 || *alpha == 0. && *beta == 1.) {
	return HSCAL_SUCCESS;
    }

    if (*alpha == 0.) {
	if (*beta == 0.) {
	    if (*incy == 1) {
            *tcomp += (*n);
	    } else {
            *tcomp += 2*(*n);
	    }

	} else {
	    if (hscal_lsame(mode, "V")) {
            {
            double t;
		hscal_dscal(
                 n, (char*)beta, incy,
                 &t
            );
            *tcomp += t;
            }
	    } else if (*incy == 1) {
            *tcomp += 2*(*n);
	    } else {
            *tcomp += 3*(*n);
	    }
	}

    } else {
	if (*alpha == 1.) {
	    if (*beta == 0.) {
		if (hscal_lsame(mode, "V")) {
                {
                double t;
		    hscal_dcopy(
                     n, incx, incy,
                     &t
                );
                *tcomp += t;
                }
		} else if (*incx == 1 && *incy == 1) {
                *tcomp += (*n);
		} else {
                *tcomp += 3*(*n);
		}

	    } else if (*beta == 1.) {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 2*(*n);
		} else {
                *tcomp += 4*(*n);
		}

	    } else {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 3*(*n);
		} else {
                *tcomp += 5*(*n);
		}
	    }

	} else {
	    if (*beta == 0.) {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 2*(*n);
		} else {
                *tcomp += 3*(*n);
		}

	    } else if (*beta == 1.) {
		if (hscal_lsame(mode, "V")) {
                {
                double t;
		    hscal_daxpy(
                     n, (char*)alpha, incx, incy,
                     &t
                );
                *tcomp += t;
                }
		} else if (*incx == 1 && *incy == 1) {
                *tcomp += 3*(*n);
		} else {
                *tcomp += 5*(*n);
		}

	    } else {
		if (*incx == 1 && *incy == 1) {
                *tcomp += 4*(*n);
		} else {
                *tcomp += 6*(*n);
		}
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of PBDVECADD */

}

#endif
