
#ifndef __HSCALAPACK_PBSTRAN_HSCALFUNC_HH
#define __HSCALAPACK_PBSTRAN_HSCALFUNC_HH

/*
*  Include files
*/
#include "../../../include/hscalapack.h"

/* ======================================================================= */
/*     SUBROUTINE PBSTRAN                                                  */
/* ======================================================================= */

int hscal_pbstran(
    hscalinteger *icontxt, char *adist, char *trans, 
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda, 
    hscalreal *beta, hscalinteger *ldc, hscalinteger *iarow, 
    hscalinteger *iacol, hscalinteger *icrow, hscalinteger *iccol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
)
{
    HSCALBLACSCONTEXT   * blacsctxt;

    hscalinteger c__0 = 0;
    hscalreal c_b14 = 0.f;
    hscalreal c_b31 = 1.f;

    /* System generated locals */
    hscalinteger a_dim1, a_offset, c_dim1, c_offset, i__1, i__2, i__3, i__4;
    hscalinteger i__;
    hscalinteger ml, mp, mq, np, nq, mq0, np0, igd, lcm;
    hscalinteger idex, jdex, info, lcmp, lcmq;
    hscalinteger jccol, mccol;
    hscalreal tbeta;
    hscalinteger mrcol, jcrow, mcrow, mrrow;
    hscallogical colform, rowform;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     Jaeyoung Choi, Oak Ridge National Laboratory */
/*     Jack Dongarra, University of Tennessee and Oak Ridge National Lab. */
/*     David Walker,  Oak Ridge National Laboratory */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PBSTRAN  transposes  a column block to row block, or a row block to */
/*  column block by reallocating data distribution. */

/*     C := A^T + beta*C, or C := A^C + beta*C */

/*  where A is an M-by-N matrix  and C is an N-by-M matrix, and the size */
/*  of M or N is limited to its block size NB. */

/*  The first elements  of the matrices A, and C  should  be  located  at */
/*  the beginnings of their first blocks. (not the middle of the blocks.) */

/*  Parameters */
/*  ========== */

/*  ICONTXT (input) INTEGER */
/*          ICONTXT is the BLACS mechanism for partitioning communication */
/*          space.  A defining property of a context is that a message in */
/*          a context cannot be sent or received in another context.  The */
/*          BLACS context includes the definition of a grid, and each */
/*          process' coordinates in it. */

/*  ADIST  - (input) CHARACTER*1 */
/*           ADIST specifies whether A is a column block or a row block. */

/*              ADIST = 'C',  A is a column block */
/*              ADIST = 'R',  A is a row block */

/*  TRANS  - (input) CHARACTER*1 */
/*           TRANS specifies whether the transposed format is transpose */
/*           or conjugate transpose.  If the matrices A and C are real, */
/*           the argument is ignored. */

/*              TRANS = 'T',  transpose */
/*              TRANS = 'C',  conjugate transpose */

/*  M      - (input) INTEGER */
/*           M specifies the (global) number of rows of the matrix (block */
/*           column or block row) A and of columns of the matrix C. */
/*           M >= 0. */

/*  N      - (input) INTEGER */
/*           N specifies the (global) number of columns of the matrix */
/*           (block column or block row) A  and of columns of the matrix */
/*           C.  N >= 0. */

/*  NB     - (input) INTEGER */
/*           NB specifies  the column block size of the matrix A and the */
/*           row block size of the matrix C when ADIST = 'C'.  Otherwise, */
/*           it specifies  the row block size of the matrix A and the */
/*           column block size of the matrix C. NB >= 1. */

/*  LDA     (input) INTEGER */
/*          LDA specifies the leading dimension of (local) A as declared */
/*          in the calling (sub) program.  LDA >= MAX(1,Mp) when */
/*          ADIST = 'C', or LDA >= MAX(1,M) otherwise. */

/*  BETA    (input) REAL */
/*          BETA specifies scaler beta. */

/*  LDC     (input) INTEGER */
/*          LDC specifies the leading dimension of (local) C as declared */
/*          in the calling (sub) program. LDC >= MAX(1,N) when ADIST='C', */
/*          or LDC >= MAX(1,Np) otherwise. */

/*  IAROW   (input) INTEGER */
/*          IAROW specifies  a row  of the process  template, */
/*          which holds the first block  of the matrix  A. If A is a row */
/*          of blocks (ADIST = 'R') and all rows of processes have a copy */
/*          of A, then set IAROW = -1. */

/*  IACOL   (input) INTEGER */
/*          IACOL specifies  a column of the process template, */
/*          which holds  the first block  of the matrix A.  If  A is  a */
/*          column of blocks (ADIST = 'C') and all columns of processes */
/*          have a copy of A, then set IACOL = -1. */

/*  ICROW   (input) INTEGER */
/*          ICROW specifies the current row process which holds */
/*          the first block  of the matrix C,  which is transposed of A. */
/*          If C is a row of blocks (ADIST = 'C') and the transposed */
/*          row block C is distributed all rows of processes, set */
/*          ICROW = -1. */

/*  ICCOL   (input) INTEGER */
/*          ICCOL specifies  the current column process which holds */
/*          the first block of the matrix C,  which is transposed of A. */
/*          If C is a column of blocks (ADIST = 'R') and the transposed */
/*          column block C is distributed all columns of processes, */
/*          set ICCOL = -1. */

/*  Parameters Details */
/*  ================== */

/*  Lx      It is  a local portion  of L  owned  by  a process,  (L is */
/*          replaced by M, or N,  and x is replaced by either p (=NPROW) */
/*          or q (=NPCOL)).  The value is  determined by  L, LB, x,  and */
/*          MI, where  LB is  a block size  and  MI is a  row  or column */
/*          position  in a process template.  Lx is  equal to  or less */
/*          than Lx0 = CEIL( L, LB*x ) * LB. */

/*  Communication Scheme */
/*  ==================== */

/*  The communication scheme of the routine is set to '1-tree', which is */
/*  fan-out.  (For details, see BLACS user's guide.) */

/*  Memory Requirement of WORK */
/*  ========================== */

/*  Mqb  = CEIL( M, NB*NPCOL ) */
/*  Npb  = CEIL( N, NB*NPROW ) */
/*  LCMQ = LCM / NPCOL */
/*  LCMP = LCM / NPROW */

/*  (1) ADIST = 'C' */
/*   (a) IACOL != -1 */
/*       Size(WORK) = N * CEIL(Mqb,LCMQ)*NB */
/*   (b) IACOL = -1 */
/*       Size(WORK) = N * CEIL(Mqb,LCMQ)*NB * MIN(LCMQ,CEIL(M,NB)) */

/*  (2) ADIST = 'R' */
/*   (a) IAROW != -1 */
/*       Size(WORK) = M * CEIL(Npb,LCMP)*NB */
/*   (b) IAROW = -1 */
/*       Size(WORK) = M * CEIL(Npb,LCMP)*NB * MIN(LCMP,CEIL(N,NB)) */

/*  Notes */
/*  ----- */
/*  More precise space can be computed as */

/*  CEIL(Mqb,LCMQ)*NB => NUMROC( NUMROC(M,NB,0,0,NPCOL), NB, 0, 0, LCMQ ) */
/*  CEIL(Npb,LCMP)*NB => NUMROC( NUMROC(N,NB,0,0,NPROW), NB, 0, 0, LCMP ) */

/*  ===================================================================== */

    /* Function Body */
    if (*m == 0 || *n == 0) {
	return HSCAL_SUCCESS;
    }

    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

    colform = hscal_lsame(adist, "C");
    rowform = hscal_lsame(adist, "R");

/*     Test the input parameters. */

    info = 0;
    if (! colform && ! rowform) {
	info = 2;
    } else if (*m < 0) {
	info = 4;
    } else if (*n < 0) {
	info = 5;
    } else if (*nb < 1) {
	info = 6;
    } else if (*iarow < -1 || *iarow >= nprow || *iarow == -1 && colform) {
	info = 12;
    } else if (*iacol < -1 || *iacol >= npcol || *iacol == -1 && rowform) {
	info = 13;
    } else if (*icrow < -1 || *icrow >= nprow || *icrow == -1 && rowform) {
	info = 14;
    } else if (*iccol < -1 || *iccol >= npcol || *iccol == -1 && colform) {
	info = 15;
    }

L10:
    if (info != 0) {
	i__1 = -(info);
      hscal_PB_Cabort(
           "PBSTRAN",
           i__1,
           nprow, npcol, myrow, mycol
      );
      Cblacs_gridexit_nocomm(blacsctxt);
	return HSCAL_SUCCESS;
    }

/*     Start the operations. */

/*     LCM : the least common multiple of NPROW and NPCOL */

    lcm = ilcm_(&nprow, &npcol);
    lcmp = lcm / nprow;
    lcmq = lcm / npcol;
    igd = npcol / lcmp;

/*     When A is a column block */

    if (colform) {

/*       Form  C <== A'  ( A is a column block ) */
/*                                         _ */
/*                                        | | */
/*                                        | | */
/*            _____________               | | */
/*           |______C______|     <==      |A| */
/*                                        | | */
/*                                        | | */
/*                                        |_| */

/*       MRROW : row relative position in template from IAROW */
/*       MRCOL : column relative position in template from ICCOL */

	mrrow = (nprow + myrow - *iarow) % nprow;
	mrcol = (npcol + mycol - *iccol) % npcol;
	jcrow = *icrow;
	if (*icrow == -1) {
	    jcrow = *iarow;
	}

	mp = numroc_(m, nb, &myrow, iarow, &nprow);
	mq = numroc_(m, nb, &mycol, iccol, &npcol);
	i__1 = numroc_(m, nb, &c__0, &c__0, &npcol);
	mq0 = numroc_(&i__1, nb, &c__0, &c__0, &lcmq);

	if (*lda < mp && (*iacol == mycol || *iacol == -1)) {
	    info = 8;
	} else if (*ldc < *n && (*icrow == myrow || *icrow == -1)) {
	    info = 11;
	}
	if (info != 0) {
	    goto L10;
	}

/*       When a column process of IACOL has a column block A, */

	if (*iacol >= 0) {
	    tbeta = 0.f;
	    if (myrow == jcrow) {
		tbeta = *beta;
	    }

/* Computing MIN */
	    i__2 = lcm, i__3 = hscal_iceil(m, nb);
	    i__1 = min(i__2,i__3) - 1;
	    for (i__ = 0; i__ <= i__1; ++i__) {
		mcrow = (i__ % nprow + *iarow) % nprow;
		mccol = (i__ % npcol + *iccol) % npcol;
		if (lcmq == 1) {
		    mq0 = numroc_(m, nb, &i__, &c__0, &npcol);
		}
		jdex = i__ / npcol * *nb;

/*           A source node copies the blocks to WORK, and send it */

		if (myrow == mcrow && mycol == *iacol) {

/*             The source node is a destination node */

		    idex = i__ / nprow * *nb;
		    if (myrow == jcrow && mycol == mccol) {
			i__2 = mp - idex;
			hscal_pbstr2at(
                       icontxt, "Col", trans, &i__2, n, nb,
			     lda, &tbeta,
			     ldc, &lcmp, &lcmq,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

/*             The source node sends blocks to a destination node */

		    } else {
			i__2 = mp - idex;
			i__3 = lcmp * *nb;
			hscal_pbstr2bt(
                       icontxt, "Col", trans, &i__2, n, nb,
			     lda, &c_b14, n, 
                       &i__3,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
			hscal_Csgesd2d(
                       blacsctxt, *n, mq0, *n, jcrow, mccol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }

/*           A destination node receives the copied blocks */

		} else if (myrow == jcrow && mycol == mccol) {
		    if (lcmq == 1 && tbeta == 0.f) {
                   ;
		    } else {
			i__2 = mq - jdex;
			hscal_pbstr2af(
                       icontxt, "Row", n, &i__2, nb, n, &tbeta, ldc, 
                       &lcmp, &lcmq, &mq0,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }
		}
/* L20: */
	    }

/*         Broadcast a row block of C in each column of template */

	    if (*icrow == -1) {
		if (myrow == jcrow) {
		    hscal_Csgebs2d(
                     blacsctxt, "Col", "1-tree", *n, mq,
                     *ldc, 
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Csgebr2d(
                     blacsctxt, "Col", "1-tree", *n, mq,
			   *ldc, jcrow, mycol, 
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
	    }

/*       When all column procesors have a copy of the column block A, */

	} else {
	    if (lcmq == 1) {
		mq0 = mq;
	    }

/*         Processors, which have diagonal blocks of A, copy them to */
/*         WORK array in transposed form */

	    i__1 = lcmp - 1;
	    for (i__ = 0; i__ <= i__1; ++i__) {
		if (mrcol == (nprow * i__ + mrrow) % npcol) {
		    if (lcmq == 1 && (*icrow == -1 || *icrow == myrow)) {
			i__2 = mp - i__ * *nb;
			i__3 = lcmp * *nb;
			hscal_pbstr2bt(
                       icontxt, "Col", trans, &i__2, n, nb,
			     lda, beta, ldc, &i__3,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    } else {
			i__2 = mp - i__ * *nb;
			i__3 = lcmp * *nb;
			hscal_pbstr2bt(
                       icontxt, "Col", trans, &i__2, n, nb,
			     lda, &c_b14, n,
                       &i__3,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }
		}
/* L30: */
	    }

/*         Get diagonal blocks of A for each column of the template */

	    mcrow = (mrcol % nprow + *iarow) % nprow;
	    if (lcmq > 1) {
		mccol = (npcol + mycol - *iccol) % npcol;
		i__1 = hscal_iceil(m, nb);
		hscal_pbstrget(
                 icontxt, "Row", n, &mq0, &i__1, n, &mcrow,
                 &mccol, &igd, &myrow, &mycol, &nprow, &npcol,
                 tcomp, tcomm
            );
	    }

/*         Broadcast a row block of WORK in every row of template */

	    if (*icrow == -1) {
		if (myrow == mcrow) {
		    if (lcmq > 1) {
			hscal_pbstrsrt(
                       icontxt, "Row", n, &mq, nb, n, 
                       beta, ldc, &lcmp, &lcmq, &mq0,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    hscal_Csgebs2d(
                     blacsctxt, "Col", "1-tree", *n, mq,
			   *ldc,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Csgebr2d(
                     blacsctxt, "Col", "1-tree", *n, mq,
			   *ldc, mcrow, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}

/*         Send a row block of WORK to the destination row */

	    } else {
		if (lcmq == 1) {
		    if (myrow == mcrow) {
			if (myrow != *icrow) {
			    hscal_Csgesd2d(
                           blacsctxt, *n, mq, *n, *icrow, mycol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );
			}
		    } else if (myrow == *icrow) {
			if (*beta == 0.) {
                     ;
			} else {
			   pbsmatadd_tcomp_macro(
                          icontxt, "G", n, &mq, &c_b31,
                          n, beta, ldc, 
                          nprow, npcol, myrow, mycol, tcomp
                     );
			}
		    }

		} else {
/* Computing MIN */
/* Computing MAX */
		    i__3 = 0, i__4 = hscal_iceil(m, nb) - mccol;
		    i__1 = lcmq, i__2 = max(i__3,i__4);
		    ml = mq0 * min(i__1,i__2);
		    if (myrow == mcrow) {
			if (myrow != *icrow) {
			    hscal_Csgesd2d(
                           blacsctxt, *n, ml, *n, *icrow,
                           mycol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );
			}
		    } else if (myrow == *icrow) {
				;
		    }

		    if (myrow == *icrow) {
			hscal_pbstrsrt(
                       icontxt, "Row", n, &mq, nb, n, 
                       beta, ldc, &lcmp, &lcmq, &mq0,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		}
	    }

	}

/*     When A is a row block */

    } else {

/*        Form  C <== A'  ( A is a row block ) */
/*            _ */
/*           | | */
/*           | | */
/*           | |                _____________ */
/*           |C|      <==      |______A______| */
/*           | | */
/*           | | */
/*           |_| */

/*        MRROW : row relative position in template from ICROW */
/*        MRCOL : column relative position in template from IACOL */

	mrrow = (nprow + myrow - *icrow) % nprow;
	mrcol = (npcol + mycol - *iacol) % npcol;
	jccol = *iccol;
	if (*iccol == -1) {
	    jccol = *iacol;
	}

	np = numroc_(n, nb, &myrow, icrow, &nprow);
	nq = numroc_(n, nb, &mycol, iacol, &npcol);
	i__1 = numroc_(n, nb, &c__0, &c__0, &nprow);
	np0 = numroc_(&i__1, nb, &c__0, &c__0, &lcmp);

	if (*lda < *m && (*iarow == myrow || *iarow == -1)) {
	    info = 8;
	} else if (*ldc < np && (*iccol == mycol || *iccol == -1)) {
	    info = 11;
	}
	if (info != 0) {
	    goto L10;
	}

/*        When a row process of IAROW has a row block A, */

	if (*iarow >= 0) {
	    tbeta = 0.f;
	    if (mycol == jccol) {
		tbeta = *beta;
	    }

/* Computing MIN */
	    i__2 = lcm, i__3 = hscal_iceil(n, nb);
	    i__1 = min(i__2,i__3) - 1;
	    for (i__ = 0; i__ <= i__1; ++i__) {
		mcrow = (i__ % nprow + *icrow) % nprow;
		mccol = (i__ % npcol + *iacol) % npcol;
		if (lcmp == 1) {
		    np0 = numroc_(n, nb, &i__, &c__0, &nprow);
		}
		idex = i__ / nprow * *nb;

/*            A source node copies the blocks to WORK, and send it */

		if (myrow == *iarow && mycol == mccol) {

/*              The source node is a destination node */

		    jdex = i__ / npcol * *nb;
		    if (myrow == mcrow && mycol == jccol) {
			i__2 = nq - jdex;
			hscal_pbstr2at(
                       icontxt, "Row", trans, m, &i__2, nb,
			     lda, &tbeta,
			     ldc, &lcmp, &lcmq,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

/*              The source node sends blocks to a destination node */

		    } else {
			i__2 = nq - jdex;
			i__3 = lcmq * *nb;
			hscal_pbstr2bt(
                       icontxt, "Row", trans, m, &i__2, nb,
			     lda, &c_b14, &np0, &i__3,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
			hscal_Csgesd2d(
                       blacsctxt, np0, *m, np0, mcrow, jccol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }

/*           A destination node receives the copied blocks */

		} else if (myrow == mcrow && mycol == jccol) {
		    if (lcmp == 1 && tbeta == 0.f) {
                   ;
		    } else {
			i__2 = np - idex;
			hscal_pbstr2af(
                       icontxt, "Col", &i__2, m, nb, &np0, &tbeta, 
                       ldc, &lcmp, &lcmq, &np0, 
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }
		}
/* L40: */
	    }

/*         Broadcast a column block of WORK in each row of template */

	    if (*iccol == -1) {
		if (mycol == jccol) {
		    hscal_Csgebs2d(
                     blacsctxt, "Row", "1-tree", np, *m,
			   *ldc,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Csgebr2d(
                     blacsctxt, "Row", "1-tree", np, *m,
			   *ldc, myrow, jccol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
	    }

/*       When all row procesors have a copy of the row block A, */

	} else {
	    if (lcmp == 1) {
		np0 = np;
	    }

/*         Processors, which have diagonal blocks of A, copy them to */
/*         WORK array in transposed form */

	    i__1 = lcmq - 1;
	    for (i__ = 0; i__ <= i__1; ++i__) {
		if (mrrow == (npcol * i__ + mrcol) % nprow) {
		    if (lcmp == 1 && (*iccol == -1 || *iccol == mycol)) {
			i__2 = nq - i__ * *nb;
			i__3 = lcmq * *nb;
			hscal_pbstr2bt(
                       icontxt, "Row", trans, m, &i__2, nb,
			     lda, beta, ldc, &i__3,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    } else {
			i__2 = nq - i__ * *nb;
			i__3 = lcmq * *nb;
			hscal_pbstr2bt(
                       icontxt, "Row", trans, m, &i__2, nb,
			     lda, &c_b14, &np0, &i__3,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }
		}
/* L50: */
	    }

/*         Get diagonal blocks of A for each row of the template */

	    mccol = (mrrow % npcol + *iacol) % npcol;
	    if (lcmp > 1) {
		mcrow = (nprow + myrow - *icrow) % nprow;
		i__1 = hscal_iceil(n, nb);
		hscal_pbstrget(
                 icontxt, "Col", &np0, m, &i__1, &np0, &mcrow, &mccol, 
                 &igd, &myrow, &mycol, &nprow, &npcol,
                 tcomp, tcomm
            );
	    }

/*         Broadcast a column block of WORK in every column of template */

	    if (*iccol == -1) {
		if (mycol == mccol) {
		    if (lcmp > 1) {
			hscal_pbstrsrt(
                       icontxt, "Col", &np, m, nb, &np0, 
			     beta, ldc, &lcmp, &lcmq, &np0,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    hscal_Csgebs2d(
                     blacsctxt, "Row", "1-tree", np, *m,
			   *ldc,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Csgebr2d(
                     blacsctxt, "Row", "1-tree", np, *m,
			   *ldc, myrow, mccol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}

/*         Send a column block of WORK to the destination column */

	    } else {
		if (lcmp == 1) {
		    if (mycol == mccol) {
			if (mycol != *iccol) {
			    hscal_Csgesd2d(
                           blacsctxt, np, *m, np, myrow, *iccol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );
			}
		    } else if (mycol == *iccol) {
			if (*beta == 0.) {
                     ;
			} else {
			    pbsmatadd_tcomp_macro(
                           icontxt, "G", &np, m, &c_b31,
				   &np, beta, ldc, 
                           nprow, npcol, myrow, mycol, tcomp
                      );
			}
		    }

		} else {
/* Computing MIN */
/* Computing MAX */
		    i__3 = 0, i__4 = hscal_iceil(n, nb) - mcrow;
		    i__1 = lcmp, i__2 = max(i__3,i__4);
		    ml = *m * min(i__1,i__2);
		    if (mycol == mccol) {
			if (mycol != *iccol) {
			    hscal_Csgesd2d(
                           blacsctxt, np0, ml, np0, myrow, *iccol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );
			}
		    } else if (mycol == *iccol) {
                  ;
		    }

		    if (mycol == *iccol) {
			hscal_pbstrsrt(
                       icontxt, "Col", &np, m, nb, &np0, 
			     beta, ldc, &lcmp, &lcmq, &np0,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		}
	    }

	}
    }

    Cblacs_gridexit_nocomm(blacsctxt);

    return HSCAL_SUCCESS;

/*     End of PBSTRAN */

}

/* ======================================================================= */
/*     SUBROUTINE PBSTR2AT */
/* ======================================================================= */

int hscal_pbstr2at(
    hscalinteger *icontxt, char *adist, char *trans, 
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda, 
    hscalreal *beta, hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
)
{
    hscalinteger c__0 = 0;
    hscalreal c_b14 = 0.;
    hscalreal c_b31 = 1.;

    /* System generated locals */
    hscalinteger a_dim1, a_offset, b_dim1, b_offset, i__1, i__2, i__3;

    /* Local variables */
    hscalinteger k;
    hscalinteger ia, ib, intv, jntv;

    /* Function Body */
    if (*lcmp == *lcmq) {
	pbsmatadd_tcomp_macro(
           icontxt, trans, n, m, &c_b31, lda, beta, ldb,
           nprow, npcol, myrow, mycol, tcomp
      );

    } else {

/*        If A is a column block ( ADIST = 'C' ), */

	if (hscal_lsame(adist, "C")) {
	    intv = *lcmp * *nb;
	    jntv = *lcmq * *nb;
	    ia = 1;
	    ib = 1;
	    i__1 = hscal_iceil(m, &intv);
	    for (k = 1; k <= i__1; ++k) {
/* Computing MIN */
		i__3 = *m - ia + 1;
		i__2 = min(i__3,*nb);
		pbsmatadd_tcomp_macro(
                 icontxt, trans, n, &i__2, &c_b31, 
                 lda, beta, ldb,
                 nprow, npcol, myrow, mycol, tcomp
            );
		ia += intv;
		ib += jntv;
/* L10: */
	    }

/*        If A is a row block ( ADIST = 'R' ), */

	} else {
	    intv = *lcmp * *nb;
	    jntv = *lcmq * *nb;
	    ia = 1;
	    ib = 1;
	    i__1 = hscal_iceil(n, &jntv);
	    for (k = 1; k <= i__1; ++k) {
/* Computing MIN */
		i__3 = *n - ia + 1;
		i__2 = min(i__3,*nb);
		pbsmatadd_tcomp_macro(
                 icontxt, trans, &i__2, m, &c_b31,
		     lda, beta, ldb,
                 nprow, npcol, myrow, mycol, tcomp
            );
		ia += jntv;
		ib += intv;
/* L20: */
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of PBSTR2AT */

}

/* ======================================================================= */
/*     SUBROUTINE PBSTR2BT */
/* ======================================================================= */

int hscal_pbstr2bt(
    hscalinteger *icontxt, char *adist, char *trans, 
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda, 
    hscalreal *beta, hscalinteger *ldb, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
)
{
    hscalinteger c__0 = 0;
    hscalreal c_b14 = 0.;
    hscalreal c_b31 = 1.;

    /* System generated locals */
    hscalinteger a_dim1, a_offset, b_dim1, b_offset, i__1, i__2, i__3;

    /* Local variables */
    hscalinteger k;
    hscalinteger ia, ib;

    /* Function Body */
    if (*intv == *nb) {
	pbsmatadd_tcomp_macro(
           icontxt, trans, n, m, &c_b31, lda, beta, ldb,
           nprow, npcol, myrow, mycol, tcomp
      );

    } else {

/*        If A is a column block ( ADIST = 'C' ), */

	if (hscal_lsame(adist, "C")) {
	    ia = 1;
	    ib = 1;
	    i__1 = hscal_iceil(m, intv);
	    for (k = 1; k <= i__1; ++k) {
/* Computing MIN */
		i__3 = *m - ia + 1;
		i__2 = min(i__3,*nb);
		pbsmatadd_tcomp_macro(
                 icontxt, trans, n, &i__2, &c_b31,
		     lda, beta, ldb,
                 nprow, npcol, myrow, mycol, tcomp
            );
		ia += *intv;
		ib += *nb;
/* L10: */
	    }

/*        If A is a row block (ADIST = 'R'), */

	} else {
	    ia = 1;
	    ib = 1;
	    i__1 = hscal_iceil(n, intv);
	    for (k = 1; k <= i__1; ++k) {
/* Computing MIN */
		i__3 = *n - ia + 1;
		i__2 = min(i__3,*nb);
		pbsmatadd_tcomp_macro(
                 icontxt, trans, &i__2, m, &c_b31,
		     lda, beta, ldb,
                 nprow, npcol, myrow, mycol, tcomp
            );
		ia += *intv;
		ib += *nb;
/* L20: */
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of PBSTR2BT */

}

/* ======================================================================= */
/*     SUBROUTINE PBSTR2AF */
/* ======================================================================= */

int hscal_pbstr2af(
    hscalinteger *icontxt, char *adist, hscalinteger *m, 
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda, 
    hscalreal *beta, hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq, 
    hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
)
{
    hscalinteger c__0 = 0;
    hscalreal c_b14 = 0.;
    hscalreal c_b31 = 1.;

    /* System generated locals */
    hscalinteger a_dim1, a_offset, b_dim1, b_offset, i__1, i__2, i__3;

    /* Local variables */
    hscalinteger k;
    hscalinteger ja, jb, intv;

    /* Function Body */
    if (hscal_lsame(adist, "R")) {
	intv = *nb * *lcmq;
	ja = 1;
	jb = 1;
	i__1 = hscal_iceil(nint, nb);
	for (k = 1; k <= i__1; ++k) {
/* Computing MIN */
	    i__3 = *n - jb + 1;
	    i__2 = min(i__3,*nb);
	    pbsmatadd_tcomp_macro(
               icontxt, "G", m, &i__2, &c_b31,
	         lda, beta, ldb,
               nprow, npcol, myrow, mycol, tcomp
          );
	    ja += *nb;
	    jb += intv;
/* L10: */
	}

/*     if( LSAME( ADIST, 'C' ) ) then */

    } else {
	intv = *nb * *lcmp;
	ja = 1;
	jb = 1;
	i__1 = hscal_iceil(nint, nb);
	for (k = 1; k <= i__1; ++k) {
/* Computing MIN */
	    i__3 = *m - jb + 1;
	    i__2 = min(i__3,*nb);
	    pbsmatadd_tcomp_macro(
               icontxt, "G", &i__2, n, &c_b31, lda, 
	         beta, ldb, 
               nprow, npcol, myrow, mycol, tcomp
          );
	    ja += *nb;
	    jb += intv;
/* L20: */
	}
    }

    return HSCAL_SUCCESS;

/*     End of PBSTR2AF */

}

/* ======================================================================= */

#endif
