
#ifndef __HSCALAPACK_PBSTRNV_HSCALFUNC_HH
#define __HSCALAPACK_PBSTRNV_HSCALFUNC_HH

/*
*  Include files
*/
#include "../../../include/hscalapack.h"

/* ======================================================================= */
/*     SUBROUTINE PBSTRNV                                                  */
/* ======================================================================= */

int hscal_pbstrnv(
    hscalinteger *icontxt, char *xdist, char *trans, 
    hscalinteger *n, hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, 
    hscalreal *beta, hscalinteger *incy, hscalinteger *ixrow, 
    hscalinteger *ixcol, hscalinteger *iyrow, hscalinteger *iycol, 
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
)
{
    HSCALBLACSCONTEXT   * blacsctxt;

    hscalinteger c__0 = 0;
    hscalreal c_b12 = 0.f;
    hscalinteger c__1 = 1;
    hscalreal c_b49 = 1.f;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4;

    /* Local variables */
    hscalinteger i__;
    hscalinteger nn, np, nq, jz, kz, np0, np1, nq0, nq1, igd, lcm;
    hscalinteger idex, jdex, info, lcmp, lcmq;
    hscalinteger mccol;
    hscalreal tbeta;
    hscalinteger mrcol, jycol, mcrow, mrrow, jyrow;
    hscallogical colform, rowform;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     Jaeyoung Choi, Oak Ridge National Laboratory */
/*     Jack Dongarra, University of Tennessee and Oak Ridge National Lab. */
/*     David Walker,  Oak Ridge National Laboratory */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PBSTRNV transposes a column vector to row vector, or a row vector to */
/*  column vector by reallocating data distribution. */

/*     Y := X' */

/*  where X and Y are N vectors. */

/*  Parameters */
/*  ========== */

/*  ICONTXT (input) INTEGER */
/*          ICONTXT is the BLACS mechanism for partitioning communication */
/*          space.  A defining property of a context is that a message in */
/*          a context cannot be sent or received in another context.  The */
/*          BLACS context includes the definition of a grid, and each */
/*          process' coordinates in it. */

/*  XDIST   (input) CHARACTER*1 */
/*          XDIST specifies whether X is a column vector or a row vector, */

/*            XDIST = 'C',  X is a column vector (distributed columnwise) */
/*            XDIST = 'R',  X is a row vector    (distributed rowwise) */

/*  TRANS   (input) CHARACTER*1 */
/*          TRANS specifies whether the transposed format is transpose */
/*          or conjugate transpose.  If the vectors X and Y are real, */
/*          the argument is ignored. */

/*             TRANS = 'T',  transpose */
/*             TRANS = 'C',  conjugate transpose */

/*  N       (input) INTEGER */
/*          N specifies the (global) number of the vector X and the */
/*          vector Y.  N >= 0. */

/*  NB      (input) INTEGER */
/*          NB specifies the block size of vectors X and Y.  NB >= 0. */

/*  NZ      (input) INTEGER */
/*          NZ is the column offset to specify the column distance from */
/*          the beginning of the block to the first element of the */
/*          vector X, and the row offset to the first element of the */
/*          vector Y if XDIST = 'C'. */
/*          Otherwise, it is row offset to specify the row distance */
/*          from the beginning of the block to the first element of the */
/*          vector X, and the column offset to the first element of the */
/*          vector Y.  0 < NZ <= NB. */

/*  INCX    (input) INTEGER */
/*          INCX specifies the increment for the elements of X. */
/*          INCX <> 0. */

/*  BETA    (input) DOUBLE PRECISION */
/*          BETA specifies scaler beta. */

/*  INCY    (input) INTEGER */
/*          INCY specifies the increment for the elements of Y. */
/*          INCY <> 0. */

/*  IXROW   (input) INTEGER */
/*          IXROW specifies a row of the process template, which holds */
/*          the first element of the vector X. If X is a row vector and */
/*          all rows of processes have a copy of X, then set IXROW = -1. */

/*  IXCOL   (input) INTEGER */
/*          IXCOL specifies  a column of the process template, */
/*          which holds the first element of the vector X.  If  X is  a */
/*          column block and all columns of processes have a copy of X, */
/*          then set IXCOL = -1. */

/*  IYROW   (input) INTEGER */
/*          IYROW specifies the current row process which holds the */
/*          first element of the vector Y, which is transposed of X. */
/*          If X  is a column vector and the transposed  row vector Y is */
/*          distributed all rows of processes, set IYROW = -1. */

/*  IYCOL   (input) INTEGER */
/*          IYCOL specifies  the current column process  which holds */
/*          the first element of the vector Y, which is transposed of Y. */
/*          If X is a row block and the transposed column vector Y is */
/*          distributed all columns of processes, set IYCOL = -1. */

/*  Parameters Details */
/*  ================== */

/*  Nx      It is a local portion  of N owned by a process, where x is */
/*          replaced by  either p (=NPROW) or q (=NPCOL)).  The value is */
/*          determined by N, NB, NZ, x, and MI, where NB is a block size, */
/*          NZ is a offset from the beginning of the block,  and MI is a */
/*          row or column position  in a process template. Nx is equal */
/*          to  or less than Nx0 = CEIL( N+NZ, NB*x ) * NB. */

/*  Communication Scheme */
/*  ==================== */

/*  The communication scheme of the routine is set to '1-tree', which is */
/*  fan-out.  (For details, see BLACS user's guide.) */

/*  Memory Requirement of WORK */
/*  ========================== */

/*  NN   = N + NZ */
/*  Npb  = CEIL( NN, NB*NPROW ) */
/*  Nqb  = CEIL( NN, NB*NPCOL ) */
/*  LCMP = LCM / NPROW */
/*  LCMQ = LCM / NPCOL */

/*   (1) XDIST = 'C' */
/*     (a) IXCOL != -1 */
/*         Size(WORK) = CEIL(Nqb,LCMQ)*NB */
/*     (b) IXCOL = -1 */
/*         Size(WORK) = CEIL(Nqb,LCMQ)*NB * MIN(LCMQ,CEIL(NN,NB)) */

/*   (2) XDIST = 'R' */
/*     (a) IXROW != -1 */
/*         Size(WORK) = CEIL(Npb,LCMP)*NB */
/*     (b) IXROW = -1 */
/*         Size(WORK) = CEIL(Npb,LCMP)*NB * MIN(LCMP,CEIL(NN,NB)) */

/*  Notes */
/*  ----- */
/*  More precise space can be computed as */

/*  CEIL(Npb,LCMP)*NB => NUMROC( NUMROC(NN,NB,0,0,NPROW), NB, 0, 0, LCMP) */
/*  CEIL(Nqb,LCMQ)*NB => NUMROC( NUMROC(NN,NB,0,0,NPCOL), NB, 0, 0, LCMQ) */

/*  ===================================================================== */

    /* Function Body */
    if (*n == 0) {
	return HSCAL_SUCCESS;
    }

    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

    colform = hscal_lsame(xdist, "C");
    rowform = hscal_lsame(xdist, "R");

/*     Test the input parameters. */

    info = 0;
    if (! colform && ! rowform) {
	info = 2;
    } else if (*n < 0) {
	info = 4;
    } else if (*nb < 1) {
	info = 5;
    } else if (*nz < 0 || *nz >= *nb) {
	info = 6;
    } else if (*incx == 0) {
	info = 8;
    } else if (*incy == 0) {
	info = 11;
    } else if (*ixrow < -1 || *ixrow >= nprow || *ixrow == -1 && colform) {
	info = 12;
    } else if (*ixcol < -1 || *ixcol >= npcol || *ixcol == -1 && rowform) {
	info = 13;
    } else if (*iyrow < -1 || *iyrow >= nprow || *iyrow == -1 && rowform) {
	info = 14;
    } else if (*iycol < -1 || *iycol >= npcol || *iycol == -1 && colform) {
	info = 15;
    }

L10:
    if (info != 0) {
	i__1 = -(info);
      hscal_PB_Cabort(
           "PBSTRNV",
           i__1,
           nprow, npcol, myrow, mycol
      );
      Cblacs_gridexit_nocomm(blacsctxt);
	return HSCAL_SUCCESS;
    }

/*     Start the operations. */

/*     LCM : the least common multiple of NPROW and NPCOL */

    lcm = ilcm_(&nprow, &npcol);
    lcmp = lcm / nprow;
    lcmq = lcm / npcol;
    igd = npcol / lcmp;
    nn = *n + *nz;

/*     When x is a column vector */

    if (colform) {

/*       Form  y <== x'  ( x is a column vector ) */

/*                                        || */
/*                                        || */
/*            _____________               || */
/*            -----(y)-----      <==     (x) */
/*                                        || */
/*                                        || */
/*                                        || */

	if (*ixrow < 0 || *ixrow >= nprow) {
	    info = 12;
	} else if (*ixcol < -1 || *ixcol >= npcol) {
	    info = 13;
	} else if (*iyrow < -1 || *iyrow >= nprow) {
	    info = 14;
	} else if (*iycol < 0 || *iycol >= npcol) {
	    info = 15;
	}
	if (info != 0) {
	    goto L10;
	}

/*       MRROW : row relative position in template from IXROW */
/*       MRCOL : column relative position in template from IYCOL */

	mrrow = (nprow + myrow - *ixrow) % nprow;
	mrcol = (npcol + mycol - *iycol) % npcol;
	jyrow = *iyrow;
	if (*iyrow == -1) {
	    jyrow = *ixrow;
	}

	np = numroc_(&nn, nb, &myrow, ixrow, &nprow);
	if (mrrow == 0) {
	    np -= *nz;
	}
	nq = numroc_(&nn, nb, &mycol, iycol, &npcol);
	if (mrcol == 0) {
	    nq -= *nz;
	}
	i__1 = numroc_(&nn, nb, &c__0, &c__0, &npcol);
	nq0 = numroc_(&i__1, nb, &c__0, &c__0, &lcmq);

/*       When a column process of IXCOL has a column block A, */

	if (*ixcol >= 0) {
	    tbeta = 0.f;
	    if (myrow == jyrow) {
		tbeta = *beta;
	    }
	    kz = *nz;

/* Computing MIN */
	    i__2 = lcm, i__3 = hscal_iceil(&nn, nb);
	    i__1 = min(i__2,i__3) - 1;
	    for (i__ = 0; i__ <= i__1; ++i__) {
		mcrow = (i__ % nprow + *ixrow) % nprow;
		mccol = (i__ % npcol + *iycol) % npcol;
		if (lcmq == 1) {
		    nq0 = numroc_(&nn, nb, &i__, &c__0, &npcol);
		}
		jdex = i__ / npcol * *nb;
		if (mrcol == 0) {
/* Computing MAX */
		    i__2 = 0, i__3 = jdex - *nz;
		    jdex = max(i__2,i__3);
		}

/*           A source node copies the blocks to WORK, and send it */

		if (myrow == mcrow && mycol == *ixcol) {

/*             The source node is a destination node */

		    idex = i__ / nprow * *nb;
		    if (mrrow == 0) {
/* Computing MAX */
			i__2 = 0, i__3 = idex - *nz;
			idex = max(i__2,i__3);
		    }
		    if (myrow == jyrow && mycol == mccol) {
			i__2 = np - idex;
			hscal_pbstr2b1(
                       icontxt, trans, &i__2, nb, &kz,
			     incx, &tbeta,
                       incy, &lcmp, &lcmq, 
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

/*             The source node sends blocks to a destination node */

		    } else {
			i__2 = np - idex;
			hscal_pbstr2b1(
                       icontxt, trans, &i__2, nb, &kz,
			     incx, &c_b12, &c__1, &lcmp, &c__1,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
			i__2 = nq0 - kz;
			hscal_Csgesd2d(
                       blacsctxt, c__1, i__2, c__1,
                       jyrow, mccol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }

/*           A destination node receives the copied vector */

		} else if (myrow == jyrow && mycol == mccol) {
		    if (lcmq == 1 && tbeta == 0.f) {
			i__2 = nq0 - kz;
		    } else {
			i__2 = nq0 - kz;
			i__2 = nq - jdex;
			i__3 = lcmq * *nb;
			hscal_pbstr2a1(
                       icontxt, &i__2, nb, &kz, &c__1, &tbeta, incy, &i__3,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }
		}
		kz = 0;
/* L20: */
	    }

/*         Broadcast a row block of WORK in each column of template */

	    if (*iyrow == -1) {
		if (myrow == jyrow) {
		    hscal_Csgebs2d(
                     blacsctxt, "Col", "1-tree", c__1, nq,
                     *incy,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Csgebr2d(
                     blacsctxt, "Col", "1-tree", c__1, nq,
                     *incy, jyrow, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
	    }

/*       When all column procesors have a copy of the column block A, */

	} else {
	    if (lcmq == 1) {
		nq0 = nq;
	    }

/*         Processors, which have diagonal blocks of X, copy them to */
/*         WORK array in transposed form */

	    kz = 0;
	    if (mrrow == 0) {
		kz = *nz;
	    }
	    jz = 0;
	    if (mrrow == 0 && mycol == *iycol) {
		jz = *nz;
	    }

	    i__1 = lcmp - 1;
	    for (i__ = 0; i__ <= i__1; ++i__) {
		if (mrcol == (nprow * i__ + mrrow) % npcol) {
/* Computing MAX */
		    i__2 = 0, i__3 = i__ * *nb - kz;
		    idex = max(i__2,i__3);
		    if (lcmq == 1 && (*iyrow == -1 || *iyrow == myrow)) {
			i__2 = np - idex;
			hscal_pbstr2b1(
                       icontxt, trans, &i__2, nb, &jz,
			     incx, beta, incy, &lcmp, &c__1,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    } else {
			i__2 = np - idex;
			hscal_pbstr2b1(
                       icontxt, trans, &i__2, nb, &jz,
                       incx, &c_b12, &c__1, &lcmp, &c__1,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }
		}
/* L30: */
	    }

/*         Get diagonal blocks of A for each column of the template */

	    mcrow = (mrcol % nprow + *ixrow) % nprow;
	    if (lcmq > 1) {
		mccol = (npcol + mycol - *iycol) % npcol;
		i__1 = hscal_iceil(&nn, nb);
		hscal_pbstrget(
                 icontxt, "Row", &c__1, &nq0, &i__1, &c__1,
                 &mcrow, &mccol, &igd, &myrow, &mycol, &nprow, &npcol,
                 tcomp, tcomm
            );
	    }

/*         Broadcast a row block of WORK in every row of template */

	    if (*iyrow == -1) {
		if (myrow == mcrow) {
		    if (lcmq > 1) {
			kz = 0;
			if (mycol == *iycol) {
			    kz = *nz;
			}
			hscal_pbstrst1(
                       icontxt, "Row", &nq, nb, &kz, &c__1, beta, 
                       incy, &lcmp, &lcmq, &nq0,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    hscal_Csgebs2d(
                     blacsctxt, "Col", "1-tree", c__1, nq,
		         *incy,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Csgebr2d(
                     blacsctxt, "Col", "1-tree", c__1, nq,
			   *incy, mcrow, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}

/*         Send a row block of WORK to the destination row */

	    } else {
		if (lcmq == 1) {
		    if (myrow == mcrow) {
			if (myrow != *iyrow) {
			    hscal_Csgesd2d(
                           blacsctxt, c__1, nq0, c__1, 
                           *iyrow, mycol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );
			}
		    } else if (myrow == *iyrow) {
			if (*beta == 0.) {
                     ;
			} else {
			    pbsvecadd_tcomp_macro(
                           icontxt, "G", &nq0, &c_b49, &c__1, beta, 
                           incy,
                           nprow, npcol, myrow, mycol, tcomp
                      );
			}
		    }

		} else {
/* Computing MIN */
/* Computing MAX */
		    i__3 = 0, i__4 = hscal_iceil(&nn, nb) - mccol;
		    i__1 = lcmq, i__2 = max(i__3,i__4);
		    nq1 = nq0 * min(i__1,i__2);
		    if (myrow == mcrow) {
			if (myrow != *iyrow) {
			    hscal_Csgesd2d(
                           blacsctxt, c__1, nq1, c__1, 
				   *iyrow, mycol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );
			}
		    } else if (myrow == *iyrow) {
                  ;
		    }

		    if (myrow == *iyrow) {
			kz = 0;
			if (mycol == *iycol) {
			    kz = *nz;
			}
			hscal_pbstrst1(
                       icontxt, "Row", &nq, nb, &kz, &c__1, beta, 
                       incy, &lcmp, &lcmq, &nq0,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		}
	    }
	}

/*     When x is a row vector */

    } else {

/*       Form  y <== x'  ( x is a row block ) */

/*           || */
/*           || */
/*           ||               _____________ */
/*          (y)      <==      -----(x)----- */
/*           || */
/*           || */
/*           || */

	if (*ixrow < -1 || *ixrow >= nprow) {
	    info = 12;
	} else if (*ixcol < 0 || *ixcol >= npcol) {
	    info = 13;
	} else if (*iyrow < 0 || *iyrow >= nprow) {
	    info = 14;
	} else if (*iycol < -1 || *iycol >= npcol) {
	    info = 15;
	}
	if (info != 0) {
	    goto L10;
	}

/*       MRROW : row relative position in template from IYROW */
/*       MRCOL : column relative position in template from IXCOL */

	mrrow = (nprow + myrow - *iyrow) % nprow;
	mrcol = (npcol + mycol - *ixcol) % npcol;
	jycol = *iycol;
	if (*iycol == -1) {
	    jycol = *ixcol;
	}

	np = numroc_(&nn, nb, &myrow, iyrow, &nprow);
	if (mrrow == 0) {
	    np -= *nz;
	}
	nq = numroc_(&nn, nb, &mycol, ixcol, &npcol);
	if (mrcol == 0) {
	    nq -= *nz;
	}
	i__1 = numroc_(&nn, nb, &c__0, &c__0, &nprow);
	np0 = numroc_(&i__1, nb, &c__0, &c__0, &lcmp);

/*       When a row process of IXROW has a row block A, */

	if (*ixrow >= 0) {
	    tbeta = 0.f;
	    if (mycol == jycol) {
		tbeta = *beta;
	    }
	    kz = *nz;

/* Computing MIN */
	    i__2 = lcm, i__3 = hscal_iceil(&nn, nb);
	    i__1 = min(i__2,i__3) - 1;
	    for (i__ = 0; i__ <= i__1; ++i__) {
		mcrow = (i__ % nprow + *iyrow) % nprow;
		mccol = (i__ % npcol + *ixcol) % npcol;
		if (lcmp == 1) {
		    np0 = numroc_(&nn, nb, &i__, &c__0, &nprow);
		}
		jdex = i__ / nprow * *nb;
		if (mrrow == 0) {
/* Computing MAX */
		    i__2 = 0, i__3 = jdex - *nz;
		    jdex = max(i__2,i__3);
		}

/*           A source node copies the blocks to WORK, and send it */

		if (myrow == *ixrow && mycol == mccol) {

/*             The source node is a destination node */

		    idex = i__ / npcol * *nb;
		    if (mrcol == 0) {
/* Computing MAX */
			i__2 = 0, i__3 = idex - *nz;
			idex = max(i__2,i__3);
		    }
		    if (myrow == mcrow && mycol == jycol) {
			i__2 = nq - idex;
			hscal_pbstr2b1(
                       icontxt, trans, &i__2, nb, &kz,
			     incx, &tbeta,
                       incy, &lcmq, &lcmp,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

/*             The source node sends blocks to a destination node */

		    } else {
			i__2 = nq - idex;
			hscal_pbstr2b1(
                       icontxt, trans, &i__2, nb, &kz,
			     incx, &c_b12, &c__1, &lcmq, &c__1,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
			i__2 = np0 - kz;
			hscal_Csgesd2d(
                       blacsctxt, c__1, i__2, c__1,
			     mcrow, jycol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }

/*           A destination node receives the copied blocks */

		} else if (myrow == mcrow && mycol == jycol) {
		    if (lcmp == 1 && tbeta == 0.f) {
			i__2 = np0 - kz;
		    } else {
			i__2 = np0 - kz;
			i__2 = np - jdex;
			i__3 = lcmp * *nb;
			hscal_pbstr2a1(
                       icontxt, &i__2, nb, &kz, &c__1, &tbeta,
                       incy, &i__3,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }
		}
		kz = 0;
/* L40: */
	    }

/*         Broadcast a column vector Y in each row of template */

	    if (*iycol == -1) {
		if (mycol == jycol) {
		    hscal_Csgebs2d(
                     blacsctxt, "Row", "1-tree", c__1, np,
			   *incy,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Csgebr2d(
                     blacsctxt, "Row", "1-tree", c__1, np,
                     *incy, myrow, jycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
	    }

/*       When all row procesors have a copy of the row block A, */

	} else {
	    if (lcmp == 1) {
		np0 = np;
	    }

/*         Processors, which have diagonal blocks of A, copy them to */
/*         WORK array in transposed form */

	    kz = 0;
	    if (mrcol == 0) {
		kz = *nz;
	    }
	    jz = 0;
	    if (mrcol == 0 && myrow == *iyrow) {
		jz = *nz;
	    }

	    i__1 = lcmq - 1;
	    for (i__ = 0; i__ <= i__1; ++i__) {
		if (mrrow == (npcol * i__ + mrcol) % nprow) {
/* Computing MAX */
		    i__2 = 0, i__3 = i__ * *nb - kz;
		    idex = max(i__2,i__3);
		    if (lcmp == 1 && (*iycol == -1 || *iycol == mycol)) {
			i__2 = nq - idex;
			hscal_pbstr2b1(
                       icontxt, trans, &i__2, nb, &jz,
			     incx, beta, incy, &lcmq, &c__1,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    } else {
			i__2 = nq - idex;
			hscal_pbstr2b1(
                       icontxt, trans, &i__2, nb, &jz,
			     incx, &c_b12, &c__1, &lcmq, &c__1,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
		    }
		}
/* L50: */
	    }

/*         Get diagonal blocks of A for each row of the template */

	    mccol = (mrrow % npcol + *ixcol) % npcol;
	    if (lcmp > 1) {
		mcrow = (nprow + myrow - *iyrow) % nprow;
		i__1 = hscal_iceil(&nn, nb);
		hscal_pbstrget(
                 icontxt, "Col", &c__1, &np0, &i__1, &c__1,
                 &mcrow, &mccol, &igd, &myrow, &mycol, &nprow, &npcol,
                 tcomp, tcomm
            );
	    }

/*         Broadcast a column block of WORK in every column of template */

	    if (*iycol == -1) {
		if (mycol == mccol) {
		    if (lcmp > 1) {
			kz = 0;
			if (myrow == *iyrow) {
			    kz = *nz;
			}
			hscal_pbstrst1(
                       icontxt, "Col", &np, nb, &kz, &c__1, beta, 
                       incy, &lcmp, &lcmq, &np0,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    hscal_Csgebs2d(
                     blacsctxt, "Row", "1-tree", c__1, np, 
			   *incy,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Csgebr2d(
                     blacsctxt, "Row", "1-tree", c__1, np,
			   *incy, myrow, mccol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}

/*         Send a column block of WORK to the destination column */

	    } else {
		if (lcmp == 1) {
		    if (mycol == mccol) {
			if (mycol != *iycol) {
			    hscal_Csgesd2d(
                           blacsctxt, c__1, np, c__1,
                           myrow, *iycol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );
			}
		    } else if (mycol == *iycol) {
			if (*beta == 0.) {
			} else {
			    pbsvecadd_tcomp_macro(
                           icontxt, "G", &np, &c_b49, &c__1, 
                           beta, incy,
                           nprow, npcol, myrow, mycol, tcomp
                      );
			}
		    }

		} else {
/* Computing MIN */
/* Computing MAX */
		    i__3 = 0, i__4 = hscal_iceil(&nn, nb) - mcrow;
		    i__1 = lcmp, i__2 = max(i__3,i__4);
		    np1 = np0 * min(i__1,i__2);
		    if (mycol == mccol) {
			if (mycol != *iycol) {
			    hscal_Csgesd2d(
                           blacsctxt, c__1, np1, c__1, 
                           myrow, *iycol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );
			}
		    } else if (mycol == *iycol) {
                  ;
		    }

		    if (mycol == *iycol) {
			kz = 0;
			if (myrow == *iyrow) {
			    kz = *nz;
			}
			hscal_pbstrst1(
                       icontxt, "Col", &np, nb, &kz, 
                       &c__1, beta, incy, &lcmp, &lcmq, &np0,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		}
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of PBSTRNV */

}

/* ======================================================================= */
/*     SUBROUTINE PBSTR2A1                                                 */
/* ======================================================================= */

int hscal_pbstr2a1(
    hscalinteger *icontxt, hscalinteger *n, hscalinteger *nb, 
    hscalinteger *nz, hscalinteger *incx, hscalreal *beta, 
    hscalinteger *incy, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
)
{
    hscalinteger c__0 = 0;
    hscalreal c_b12 = 0.f;
    hscalinteger c__1 = 1;
    hscalreal c_b49 = 1.f;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3;

    /* Local variables */
    hscalinteger k;
    hscalinteger ix, iy, jz, iter;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*     y <== x */
/*     y is a scattered vector, copied from a condensed vector x. */

    /* Function Body */
    ix = 0;
    iy = 0;
    jz = *nz;
    i__1 = *n + *nz;
    iter = hscal_iceil(&i__1, intv);

    if (iter > 1) {
	i__1 = *nb - jz;
	pbsvecadd_tcomp_macro(
           icontxt, "G", &i__1, &c_b49, incx,  
           beta, incy, 
           nprow, npcol, myrow, mycol, tcomp
      );
	ix = ix + *nb - jz;
	iy = iy + *intv - jz;
	jz = 0;

	i__1 = iter - 1;
	for (k = 2; k <= i__1; ++k) {
	    pbsvecadd_tcomp_macro(
               icontxt, "G", nb, &c_b49, incx, 
               beta, incy, 
               nprow, npcol, myrow, mycol, tcomp
          );
	    ix += *nb;
	    iy += *intv;
/* L10: */
	}
    }

/* Computing MIN */
    i__2 = *n - iy, i__3 = *nb - jz;
    i__1 = min(i__2,i__3);
    pbsvecadd_tcomp_macro(
         icontxt, "G", &i__1, &c_b49, incx, beta,
	   incy, 
         nprow, npcol, myrow, mycol, tcomp
    );

    return HSCAL_SUCCESS;

/*     End of PBSTR2A1 */

}


/* ======================================================================= */
/*     SUBROUTINE PBSTR2B1                                                 */
/* ======================================================================= */

int hscal_pbstr2b1(
    hscalinteger *icontxt, char *trans, hscalinteger *n, 
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, 
    hscalreal *beta, hscalinteger *incy, hscalinteger *jinx, hscalinteger *jiny, 
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
)
{
    hscalinteger c__0 = 0;
    hscalreal c_b12 = 0.f;
    hscalinteger c__1 = 1;
    hscalreal c_b49 = 1.f;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3;

    /* Local variables */
    hscalinteger k;
    hscalinteger ix, iy, jz, iter, lenx, leny;

/*  -- PB-BLAS routine (version 2.1) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory. */
/*     April 28, 1996 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*     y <== x + beta * y */
/*     y is a condensed vector, copied from a scattered vector x */

/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Parameters .. */
/*     .. */
/*     .. Local Variables .. */

    /* Function Body */
    if (*jinx == 1 && *jiny == 1) {
	pbsvecadd_tcomp_macro(
           icontxt, trans, n, &c_b49, incx, beta, incy, 
           nprow, npcol, myrow, mycol, tcomp
	);

    } else {
	ix = 0;
	iy = 0;
	jz = *nz;
	lenx = *nb * *jinx;
	leny = *nb * *jiny;
	i__1 = *n + *nz;
	iter = hscal_iceil(&i__1, &lenx);

	if (iter > 1) {
	    i__1 = *nb - jz;
	    pbsvecadd_tcomp_macro(
               icontxt, trans, &i__1, &c_b49,
               incx, beta, incy, 
               nprow, npcol, myrow, mycol, tcomp
          );
	    ix = ix + lenx - jz;
	    iy = iy + leny - jz;
	    jz = 0;

	    i__1 = iter - 1;
	    for (k = 2; k <= i__1; ++k) {
		pbsvecadd_tcomp_macro(
                 icontxt, trans, nb, &c_b49,
		     incx, beta, incy,
                 nprow, npcol, myrow, mycol, tcomp
            );
		ix += lenx;
		iy += leny;
/* L10: */
	    }
	}

/* Computing MIN */
	i__2 = *n - ix, i__3 = *nb - jz;
	i__1 = min(i__2,i__3);
	pbsvecadd_tcomp_macro(
           icontxt, trans, &i__1, &c_b49, incx, 
	     beta, incy,
           nprow, npcol, myrow, mycol, tcomp
      );
    }

    return HSCAL_SUCCESS;

/*     End of PBSTR2B1 */

}

#endif
