/* ---------------------------------------------------------------------
*
*  -- PBLAS auxiliary routine (version 2.0) --
*     University of Tennessee, Knoxville, Oak Ridge National Laboratory,
*     and University of California, Berkeley.
*     April 1, 1998
*
*  ---------------------------------------------------------------------
*/
#ifndef __HSCAL_PB_CABORT_HSCALFUNC_HH
#define __HSCAL_PB_CABORT_HSCALFUNC_HH

/*
*  Include files
*/
#include "../../../include/hscalapack.h"

void hscal_PB_Cabort(
     char * ROUT, int INFO,
     int nprow, int npcol, int myrow, int mycol
)
{
/*
*  Purpose
*  =======
*
*  PB_Cabort is an error handler for the PBLAS  routines.  This  routine
*  displays an error message on  stderr  by calling  PB_Cwarn, and halts
*  execution by calling Cblacs_abort().
*
*  Arguments
*  =========
*
*  ROUT    (global input) pointer to CHAR
*          On entry, ROUT specifies the name of the routine calling this
*          error handler.
*
*  INFO    (local input) INTEGER
*          The error code computed by the calling PBLAS routine.
*          = 0:  no error found
*          < 0:  If the  i-th  argument is an array and the j-entry  had
*                an illegal value, then  INFO = -(i*100+j),  if the i-th
*                argument  is  a  scalar  and had an illegal value, then
*                INFO = -i.
*
*  -- Written on April 1, 1998 by
*     R. Clint Whaley, University of Tennessee, Knoxville 37996, USA.
*
*  ---------------------------------------------------------------------
*/
/*
*  .. Local Scalars ..
*/
/* ..
*  .. External Functions ..
*/
/* ..
*  .. Executable Statements ..
*
*/
   if( INFO < 0 )
   {
/*
*  Display an error message
*/
      if( INFO < DESCMULT )
      {
         hscal_PB_Cwarn( 
              nprow, npcol, myrow, mycol, 
              -1, ROUT,
              "Parameter number %d had an illegal value", -INFO
         );
      }
      else
      {
         hscal_PB_Cwarn( 
              nprow, npcol, myrow, mycol,
              -1, ROUT,
              "Parameter number %d, entry number %d had an illegal value",
              (-INFO) / DESCMULT, (-INFO) % DESCMULT 
         );
      }
   }
   else
   {
/*
*  Error code is incorrect, it should be negative
*/
      hscal_PB_Cwarn( 
           nprow, npcol, myrow, mycol, 
           -1, ROUT,
           "Positive error code %d returned!!!", INFO 
      );
   }

/*
*  End of PB_Cabort
*/
}

#endif
