/* ---------------------------------------------------------------------
*
*  -- PBLAS auxiliary routine (version 2.0) --
*     University of Tennessee, Knoxville, Oak Ridge National Laboratory,
*     and University of California, Berkeley.
*     April 1, 1998
*
*  ---------------------------------------------------------------------
*/
#ifndef __HSCAL_PB_CPTRSMAB0_C_HSCALFUNC_HH
#define __HSCAL_PB_CPTRSMAB0_C_HSCALFUNC_HH

/*
*  Include files
*/
#include "../../../include/hscalapack.h"

void hscal_PB_CptrsmAB0_C(
     PBTYP_T * TYPE, char * SIDE, char * UPLO, char * DIAG,
     int M, int N, char * ALPHA, 
     int IA, int JA, int * DESCA, 
     int IB, int JB, int * DESCB,
     int * DESCC,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           btop, * negone, * one, * talpha, * zero;
   int            Acol, Acurcol, Acurrow, Aii, Aimb, Aimb1, Ainb, Ainb1, Ajj,
                  Ald, Almb1, Alnb1, Amb, Amp0, Anq0, An, Anb, Arow, Bcol, Bii,
                  Bimb, Bimb1, Binb, Binb1, Bjj, Bld, Bmb, Bmp0, Bnb, Bnq0,
                  Brow, Cld, ctxt, k=1, kb, kblks, kbprev, ktmp, lside, size, upper;
   HSCALBLACSCONTEXT   * blacsctxt;

/* ..
*  .. Executable Statements ..
*
*/
   size  = TYPE->size;
   lside = ( Mupcase( SIDE[0] ) == CLEFT );
/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);
/*
*  Retrieve sub( A )'s local information: Aii, Ajj, Arow, Acol, Ald
*/
   Ald = DESCA[LLD_];
   PB_Cinfog2l( IA, JA, DESCA, nprow, npcol, myrow, mycol, &Aii, &Ajj,
                &Arow, &Acol );
/*
*  Retrieve sub( B )'s local information: Bii, Bjj, Brow, Bcol, Bld ...
*/
   Bimb = DESCB[IMB_]; Binb = DESCB[INB_];
   Bmb  = DESCB[MB_ ]; Bnb  = DESCB[NB_ ]; Bld = DESCB[LLD_];
   PB_Cinfog2l( IB, JB, DESCB, nprow, npcol, myrow, mycol, &Bii, &Bjj,
                &Brow, &Bcol );
/*
*  Shorcuts when sub( B ) spans only one process row or column
*/
   if( lside )
   {
      if( !( PB_Cspan( M, IB, Bimb, Bmb, DESCB[RSRC_], nprow ) ) )
      {
         Binb1 = PB_Cfirstnb( N, JB, Binb, Bnb );
         PB_Cdescset( DESCC, M, N, M, Binb1, Bmb, Bnb, Brow, Bcol, ctxt, Bld );
         Bnq0  = PB_Cnumroc( N, 0, Binb1, Bnb, mycol, Bcol, npcol );

         if( ( Bnq0 > 0 ) &&
             ( ( ( Brow >= 0 ) && ( myrow == Brow ) ) || ( Brow < 0 ) ) )
         {
            ctrsm_tcomp_macro( 
                 C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                 C2F_CHAR( DIAG ), &M, &Bnq0, ALPHA, &Ald, &Bld,
                 nprow, npcol, myrow, mycol, tcomp
            );
         }
         Cblacs_gridexit_nocomm(blacsctxt);
         return;
      }
   }
   else
   {
      if( !( PB_Cspan( N, JB, Binb, Bnb, DESCB[CSRC_], npcol ) ) )
      {
         Bimb1 = PB_Cfirstnb( M, IB, Bimb, Bmb );
         PB_Cdescset( DESCC, M, N, Bimb1, N, Bmb, Bnb, Brow, Bcol, ctxt, Bld );
         Bmp0  = PB_Cnumroc( M, 0, Bimb1, Bmb, myrow, Brow, nprow );

         if( ( Bmp0 > 0 ) &&
             ( ( ( Bcol >= 0 ) && ( mycol == Bcol ) ) || ( Bcol < 0 ) ) )
         {
            ctrsm_tcomp_macro( 
                 C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                 C2F_CHAR( DIAG ), &Bmp0, &N, ALPHA, &Ald, &Bld,
                 nprow, npcol, myrow, mycol, tcomp
            );
         }
         Cblacs_gridexit_nocomm(blacsctxt);
         return;
      }
   }
/*
*  Handle the general case now
*/
   An     = ( lside ? M : N );
   upper  = ( Mupcase( UPLO[0] ) == CUPPER  );
   talpha = ALPHA;
   negone = TYPE->negone;  one   = TYPE->one;     zero  = TYPE->zero;

/*
*  Compute more local information for sub( A ) and sub( B )
*/
   Aimb  = DESCA[IMB_]; Ainb = DESCA[INB_];
   Amb   = DESCA[MB_ ]; Anb  = DESCA[NB_ ];
   Aimb1 = PB_Cfirstnb( An, IA, Aimb, Amb );
   Almb1 = PB_Clastnb ( An, IA, Aimb, Amb );
   Amp0  = PB_Cnumroc( An, 0, Aimb1, Amb, myrow, Arow, nprow );
   Ainb1 = PB_Cfirstnb( An, JA, Ainb, Anb );
   Alnb1 = PB_Clastnb ( An, JA, Ainb, Anb );
   Anq0  = PB_Cnumroc( An, 0, Ainb1, Anb, mycol, Acol, npcol );
   if( ( Amp0 > 0 ) && ( Anq0 > 0 ) ) ;

   Bimb1 = PB_Cfirstnb( M, IB, Bimb, Bmb );
   Bmp0  = PB_Cnumroc( M, 0, Bimb1, Bmb, myrow, Brow, nprow );
   Binb1 = PB_Cfirstnb( N, JB, Binb, Bnb );
   Bnq0  = PB_Cnumroc( N, 0, Binb1, Bnb, mycol, Bcol, npcol );
   if( ( Bmp0 > 0 ) && ( Bnq0 > 0 ) ) ;

   if( lside )
   {
      Cld = M;
      PB_Cdescset( DESCC, M, N, M, Binb1, Bmb, Bnb, -1, Bcol, ctxt, Cld );
      if( Bnq0 > 0 ) { ; }
      else           { Cblacs_gridexit_nocomm(blacsctxt); return; }

      kblks = ( An > Aimb1 ? ( An - Aimb1 - 1 ) / Amb + 2 : 1 );
      btop  = *PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );

      if( upper )
      {
         Acurrow = PB_Cindxg2p( An-1, Aimb1, Amb, Arow, Arow, nprow );
         kb      = Almb1;
/*
*  Solve last block of rows of sub( B ) and broadcast it vertically to update
*  the rest of sub( B )
*/
         if( myrow == Acurrow )
         {
            ctrsm_tcomp_macro( 
                 C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                 C2F_CHAR( DIAG ), &kb, &Bnq0, ALPHA, &Ald, &Bld,
                 nprow, npcol, myrow, mycol, tcomp

            );
            hscal_Ccgebs2d( 
                 blacsctxt, COLUMN, &btop, kb, Bnq0, Bld,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
            cmmadd_tcomp_macro( 
                  &kb, &Bnq0, one, &Bld, zero, &Cld,
                  nprow, npcol, myrow, mycol, tcomp
            );
            Amp0 -= kb;
            Bmp0 -= kb;
         }
         else
         {
            hscal_Ccgebr2d( 
                 blacsctxt, COLUMN, &btop, kb, Bnq0, Cld, Acurrow, mycol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
         Acurrow = MModSub1( Acurrow, nprow );
         An     -= ( kbprev = kb );
         Anq0   -= kb;
         kblks  -= 1;
/*
*  Lookahead
*/
         while( kblks > 0 )
         {
            kb = ( kblks == 1 ? Aimb1 : Amb );

            if( myrow == Acurrow )
            {
/*
*  Update the current block of rows of sub( B ) with block of rows of sub( B )
*  of previous step
*/
              cgemm_tcomp_macro( 
                   C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &kb, &Bnq0,
                   &kbprev, negone, 
                   &Ald, &Cld, talpha, &Bld,
                   nprow, npcol, myrow, mycol, tcomp
              );
/*
*  Solve the current block of rows of sub( B )
*/
               ctrsm_tcomp_macro( 
                    C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                    C2F_CHAR( DIAG ), &kb, &Bnq0, one, &Ald, &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Broadcast the current block of rows of sub( B ) for next update
*/
               hscal_Ccgebs2d( 
                    blacsctxt, COLUMN, &btop, kb, Bnq0, Bld,
                    nprow, npcol, myrow, mycol, tcomp, tcomm
               );
               cmmadd_tcomp_macro( 
                     &kb, &Bnq0, one, &Bld, zero, &Cld,
                     nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Finish update of the remaining blocks of rows of sub( B ) with block of rows
*  of sub( B ) of previous step
*/
               if( ( ktmp = Amp0 - kb ) > 0 )
               {
                  cgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &ktmp, &Bnq0,
                       &kbprev, negone, &Ald, &Cld, talpha,
                       &Bld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
               Amp0 -= kb;
               Bmp0 -= kb;
            }
            else
            {
/*
*  Update the remaining rows of sub( B ) with block of rows of sub( B ) of
*  previous step
*/
               if( Amp0 > 0 )
               {
                  cgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &Amp0, &Bnq0,
                       &kbprev, negone, &Ald, &Cld, talpha,
                       &Bld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
/*
*  Receive the current block of rows of sub( B ) for next update
*/
               hscal_Ccgebr2d( 
                    blacsctxt, COLUMN, &btop, kb, Bnq0, Cld, Acurrow, mycol,
                    nprow, npcol, myrow, mycol, tcomp, tcomm
               );
            }

            Acurrow = MModSub1( Acurrow, nprow );
            An     -= ( kbprev = kb );
            Anq0   -= kb;
            talpha  = one;
            kblks  -= 1;
         }
      }
      else
      {
         Acurrow = Arow;
         kb      = Aimb1;
/*
*  Solve first block of rows of sub( B ) and broadcast it vertically to update
*  the rest of sub( B )
*/
         if( myrow == Acurrow )
         {
            ctrsm_tcomp_macro( 
                 C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                 C2F_CHAR( DIAG ), &kb, &Bnq0, ALPHA, &Ald,
                 &Bld,
                 nprow, npcol, myrow, mycol, tcomp
            );
            hscal_Ccgebs2d( 
                 blacsctxt, COLUMN, &btop, kb, Bnq0, Bld,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
            cmmadd_tcomp_macro( 
                  &kb, &Bnq0, one, &Bld, zero, &Cld,
                  nprow, npcol, myrow, mycol, tcomp
            );
            Amp0 -= kb;
         }
         else
         {
            hscal_Ccgebr2d( 
                 blacsctxt, COLUMN, &btop, kb, Bnq0, Cld, Acurrow, mycol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
         Acurrow = MModAdd1( Acurrow, nprow );
         kbprev  = kb;
         k      += 1;
/*
*  Lookahead
*/
         while( k <= kblks )
         {
            kb = ( k == kblks ? Almb1 : Amb );

            if( myrow == Acurrow )
            {
/*
*  Update the current block of rows of sub( B ) with block of rows of sub( B )
*  of previous step
*/
               cgemm_tcomp_macro( 
                    C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &kb, &Bnq0,
                    &kbprev, negone, 
                    &Ald, &Cld, talpha,
                    &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Solve the current block of rows of sub( B )
*/
               ctrsm_tcomp_macro( 
                    C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                    C2F_CHAR( DIAG ), &kb, &Bnq0, one, &Ald,
                    &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Broadcast the current block of rows of sub( B ) for next update
*/
               hscal_Ccgebs2d( 
                    blacsctxt, COLUMN, &btop, kb, Bnq0, Bld,
                    nprow, npcol, myrow, mycol, tcomp, tcomm
               );
               cmmadd_tcomp_macro( 
                     &kb, &Bnq0, one, &Bld, zero, &Cld,
                     nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Finish update of the remaining blocks of rows of sub( B ) with block of rows
*  of sub( B ) of previous step
*/
               if( ( ktmp = Amp0 - kb ) > 0 )
               {
                  cgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &ktmp, &Bnq0,
                       &kbprev, negone,
                       &Ald, &Cld, talpha,
                       &Bld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
               Amp0 -= kb;
            }
            else
            {
/*
*  Update the remaining rows of sub( B ) with block of rows of sub( B ) of
*  previous step
*/
               if( Amp0 > 0 )
               {
                  cgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &Amp0, &Bnq0,
                       &kbprev, negone, 
                       &Ald, &Cld, talpha,
                       &Bld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
/*
*  Receive the current block of rows of sub( B ) for next update
*/
               hscal_Ccgebr2d( 
                    blacsctxt, COLUMN, &btop, kb, Bnq0, Cld, Acurrow,
                      mycol,
                    nprow, npcol, myrow, mycol, tcomp, tcomm
               );
            }

            Acurrow = MModAdd1( Acurrow, nprow );
            kbprev  = kb;
            talpha  = one;
            k      += 1;
         }
      }
   }
   else
   {
      Cld = MAX( 1, Bmp0 );
      PB_Cdescset( DESCC, M, N, Bimb1, N, Bmb, Bnb, Brow, -1, ctxt, Cld );
      if( Bmp0 > 0 ) { ; }
      else           { Cblacs_gridexit_nocomm(blacsctxt); return; }

      kblks = ( An > Ainb1 ? ( An - Ainb1 - 1 ) / Anb + 2 : 1 );
      btop  = *PB_Ctop( &ctxt, BCAST, ROW, TOP_GET );

      if( upper )
      {
         Acurcol = Acol;
         kb      = Ainb1;
/*
*  Solve first block of columns of sub( B ) and broadcast it horizontally to
*  update the rest of sub( B )
*/
         if( mycol == Acurcol )
         {
            ctrsm_tcomp_macro( 
                 C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                 C2F_CHAR( DIAG ), &Bmp0, &kb, ALPHA, &Ald,
                 &Bld,
                 nprow, npcol, myrow, mycol, tcomp
            );
            hscal_Ccgebs2d( 
                 blacsctxt, ROW, &btop, Bmp0, kb, Bld,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
            cmmadd_tcomp_macro( 
                  &Bmp0, &kb, one, &Bld, zero, &Cld,
                  nprow, npcol, myrow, mycol, tcomp
            );
            Anq0 -= kb;
         }
         else
         {
            hscal_Ccgebr2d( 
                 blacsctxt, ROW, &btop, Bmp0, kb, Cld, myrow, Acurcol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
         Acurcol = MModAdd1( Acurcol, npcol );
         kbprev  = kb;
         k      += 1;
/*
*  Lookahead
*/
         while( k <= kblks )
         {
            kb = ( k == kblks ? Alnb1 : Anb );

            if( mycol == Acurcol )
            {
/*
*  Update the current block of columns of sub( B ) with block of columns of
*  sub( B ) of previous step
*/
               cgemm_tcomp_macro( 
                    C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &Bmp0, &kb,
                    &kbprev, negone, &Cld,
                    &Ald, talpha,
                    &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Solve the current block of columns of sub( B )
*/
               ctrsm_tcomp_macro( 
                    C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                    C2F_CHAR( DIAG ), &Bmp0, &kb, one, &Ald,
                    &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Broadcast the current block of columns of sub( B ) for next update
*/
               hscal_Ccgebs2d( 
                    blacsctxt, ROW, &btop, Bmp0, kb, Bld,
                    nprow, npcol, myrow, mycol, tcomp, tcomm
               );
               cmmadd_tcomp_macro( 
                     &Bmp0, &kb, one, &Bld, zero, &Cld,
                     nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Finish update of the remaining blocks of columns of sub( B ) with block of
*  columns of sub( B ) of previous step
*/
               if( ( ktmp = Anq0 - kb ) > 0 )
               {
                  cgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &Bmp0, &ktmp,
                       &kbprev, negone, 
                       &Cld, &Ald,
                       talpha, &Bld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
               Anq0 -= kb;
            }
            else
            {
/*
*  Update the remaining columns of sub( B ) with block of columns of sub( B )
*  of previous step
*/
               if( Anq0 > 0 )
               {
                  cgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &Bmp0, &Anq0,
                       &kbprev, negone, 
                       &Cld, &Ald,
                       talpha, &Bld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
/*
*  Receive the current block of columns of sub( B ) for next update
*/
               hscal_Ccgebr2d( 
                    blacsctxt, ROW, &btop, Bmp0, kb, Cld, myrow, Acurcol,
                    nprow, npcol, myrow, mycol, tcomp, tcomm
               );
            }

            Acurcol = MModAdd1( Acurcol, npcol );
            kbprev  = kb;
            talpha  = one;
            k      += 1;
         }
      }
      else
      {
         Acurcol = PB_Cindxg2p( An-1, Ainb1, Anb, Acol, Acol, npcol );
         kb      = Alnb1;
/*
*  Solve last block of columns of sub( B ) and broadcast it horizontally to
*  update the rest of sub( B )
*/
         if( mycol == Acurcol )
         {
            ctrsm_tcomp_macro( 
                 C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                 C2F_CHAR( DIAG ), &Bmp0, &kb, ALPHA, &Ald, &Bld,
                 nprow, npcol, myrow, mycol, tcomp
            );
            hscal_Ccgebs2d( 
                 blacsctxt, ROW, &btop, Bmp0, kb, Bld,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
            cmmadd_tcomp_macro( 
                  &Bmp0, &kb, one, &Bld, zero, &Cld,
                  nprow, npcol, myrow, mycol, tcomp
            );
            Anq0 -= kb;
            Bnq0 -= kb;
         }
         else
         {
            hscal_Ccgebr2d( 
                 blacsctxt, ROW, &btop, Bmp0, kb, Cld, myrow, Acurcol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
         Acurcol = MModSub1( Acurcol, npcol );
         An     -= ( kbprev = kb );
         Amp0   -= kb;
         kblks  -= 1;
/*
*  Lookahead
*/
         while( kblks > 0 )
         {
            kb = ( kblks == 1 ? Ainb1 : Anb );

            if( mycol == Acurcol )
            {
/*
*  Update the current block of columns of sub( B ) with block of columns of
*  sub( B ) of previous step
*/
               cgemm_tcomp_macro( 
                    C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &Bmp0, &kb,
                    &kbprev, negone, &Cld, &Ald, talpha, &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Solve the current block of columns of sub( B )
*/
               ctrsm_tcomp_macro( 
                    C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( NOTRAN ),
                    C2F_CHAR( DIAG ), &Bmp0, &kb, one, &Ald, &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Broadcast the current block of columns of sub( B ) for next update
*/
               hscal_Ccgebs2d( 
                    blacsctxt, ROW, &btop, Bmp0, kb, Bld,
                    nprow, npcol, myrow, mycol, tcomp, tcomm
               );
               cmmadd_tcomp_macro( 
                     &Bmp0, &kb, one, &Bld, zero, &Cld,
                      nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Finish update of the remaining blocks of columns of sub( B ) with block of
*  columns of sub( B ) of previous step
*/
               if( ( ktmp = Anq0 - kb ) > 0 )
               {
                  cgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &Bmp0, &ktmp,
                       &kbprev, negone, &Cld, &Ald, talpha,
                       &Bld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
               Anq0 -= kb;
               Bnq0 -= kb;
            }
            else
            {
/*
*  Update the remaining columns of sub( B ) with block of columns of sub( B )
*  of previous step
*/
               if( Anq0 > 0 )
               {
                  cgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( NOTRAN ), &Bmp0, &Anq0,
                       &kbprev, negone, &Cld, &Ald, talpha,
                       &Bld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
/*
*  Receive the current block of columns of sub( B ) for next update
*/
               hscal_Ccgebr2d( 
                    blacsctxt, ROW, &btop, Bmp0, kb, Cld, myrow, Acurcol,
                    nprow, npcol, myrow, mycol, tcomp, tcomm
               );
            }

            Acurcol = MModSub1( Acurcol, npcol );
            An     -= ( kbprev = kb );
            Amp0   -= kb;
            talpha  = one;
            kblks  -= 1;
         }
      }
   }

   Cblacs_gridexit_nocomm(blacsctxt);

/*
*  End of PB_CptrsmAB0
*/
}

#endif
