/* ---------------------------------------------------------------------
*
*  -- PBLAS auxiliary routine (version 2.0) --
*     University of Tennessee, Knoxville, Oak Ridge National Laboratory,
*     and University of California, Berkeley.
*     April 1, 1998
*
*  ---------------------------------------------------------------------
*/
#ifndef __HSCAL_PB_CPTRSMAB1_Z_HSCALFUNC_HH
#define __HSCAL_PB_CPTRSMAB1_Z_HSCALFUNC_HH

/*
*  Include files
*/
#include "../../../include/hscalapack.h"

void hscal_PB_CptrsmAB1_Z(
     PBTYP_T * TYPE, char * SIDE, char * UPLO, char * TRANSA,
     char * DIAG, int M, int N, char * ALPHA, 
     int IA, int JA, int * DESCA,  
     int IB, int JB, int * DESCB,
     int * DESCC,
     int nprow, int npcol, int myrow, int mycol, 
     double *tcomp, int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           * negone, * one;
   int            Acol, Acurcol, Acurrow, Aii, Aimb, Aimb1, Ainb, Ainb1, Ajj,
                  Ald, Almb1, Alnb1, Amb, Amp0, An, Anb, Anq0, Anxtrow, Anxtcol,
                  Arow, Bcol, Bii, Bimb, Binb, Bjj, Bld, Bmb, Bmp0, Bnb, Bnq0,
                  Brow, Cld, ctxt, k=1, kb, kblks, lside, size, upper;
   HSCALBLACSCONTEXT   * blacsctxt;

/*
*  .. Local Arrays ..
*/

/* ..
*  .. Executable Statements ..
*
*/
   size  = TYPE->size;
   lside = ( Mupcase( SIDE[0] ) == CLEFT );
/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);
/*
*  Retrieve sub( A )'s local information: Aii, Ajj, Arow, Acol, Ald
*/
   Ald  = DESCA[LLD_];
   PB_Cinfog2l( IA, JA, DESCA, nprow, npcol, myrow, mycol, &Aii, &Ajj,
                &Arow, &Acol );
/*
*  Retrieve sub( B )'s local information: Bii, Bjj, Brow, Bcol, Bld ...
*/
   Bimb = DESCB[IMB_]; Binb = DESCB[INB_];
   Bmb  = DESCB[MB_ ]; Bnb  = DESCB[NB_ ]; Bld = DESCB[LLD_];
   PB_Cinfog2l( IB, JB, DESCB, nprow, npcol, myrow, mycol, &Bii, &Bjj,
                &Brow, &Bcol );
/*
*  Shorcuts when sub( B ) spans only one process row or column
*/
   if( lside )
   {
      Bnq0 = PB_Cnumroc( N, JB, Binb, Bnb, mycol, DESCB[CSRC_], npcol );
      if( Bnq0 <= 0 ) 
      {
         Cblacs_gridexit_nocomm(blacsctxt);
         return;
      }

      Bmp0 = PB_Cnumroc( M, IB, Bimb, Bmb, myrow, DESCB[RSRC_], nprow );

      if( !( PB_Cspan( M, IB, Bimb, Bmb, DESCB[RSRC_], nprow ) ) )
      {
         if( Bmp0 > 0 )
         {
            zmmadd_tcomp_macro( 
                  &M, &Bnq0, TYPE->negone, &DESCC[LLD_], ALPHA,
                  &Bld,
                  nprow, npcol, myrow, mycol, tcomp
            );
            ztrsm_tcomp_macro( 
                 C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( TRANSA ),
                 C2F_CHAR( DIAG ), &M, &Bnq0, TYPE->one, &Ald, &Bld,
                 nprow, npcol, myrow, mycol, tcomp
            );
         }
         Cblacs_gridexit_nocomm(blacsctxt);
         return;
      }
      if( Bmp0 > 0 ) ;
   }
   else
   {
      Bmp0 = PB_Cnumroc( M, IB, Bimb, Bmb, myrow, DESCB[RSRC_], nprow );
      if( Bmp0 <= 0 ) 
      {
         Cblacs_gridexit_nocomm(blacsctxt);
         return;
      }

      Bnq0 = PB_Cnumroc( N, JB, Binb, Bnb, mycol, DESCB[CSRC_], npcol );

      if( !( PB_Cspan( N, JB, Binb, Bnb, DESCB[CSRC_], npcol ) ) )
      {
         if( Bnq0 > 0 )
         {
            zmmadd_tcomp_macro( 
                  &Bmp0, &N, TYPE->negone, &DESCC[LLD_], ALPHA,
                  &Bld,
                  nprow, npcol, myrow, mycol, tcomp
            );
            ztrsm_tcomp_macro( 
                 C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( TRANSA ),
                 C2F_CHAR( DIAG ), &Bmp0, &N, TYPE->one, &Ald, &Bld,
                 nprow, npcol, myrow, mycol, tcomp
            );
         }
         Cblacs_gridexit_nocomm(blacsctxt);
         return;
      }
      if( Bnq0 > 0 ) ;
   }
/*
*  Handle the general case now
*/
   An     = ( lside ? M : N );
   upper  = ( Mupcase( UPLO[0] ) == CUPPER  );
   negone = TYPE->negone;  one  = TYPE->one;

/*
*  Compute more local information for sub( A )
*/
   Aimb  = DESCA[IMB_]; Ainb = DESCA[INB_];
   Amb   = DESCA[MB_ ]; Anb  = DESCA[NB_ ];
   Aimb1 = PB_Cfirstnb( An, IA, Aimb, Amb );
   Almb1 = PB_Clastnb ( An, IA, Aimb, Amb );
   Amp0  = PB_Cnumroc( An, 0, Aimb1, Amb, myrow, Arow, nprow );
   Ainb1 = PB_Cfirstnb( An, JA, Ainb, Anb );
   Alnb1 = PB_Clastnb ( An, JA, Ainb, Anb );
   Anq0  = PB_Cnumroc( An, 0, Ainb1, Anb, mycol, Acol, npcol );
   if( ( Amp0 > 0 ) && ( Anq0 > 0 ) ) ;

   Cld = DESCC[LLD_];

   if( lside )
   {
      kblks = ( An > Aimb1 ? ( An - Aimb1 - 1 ) / Amb + 2 : 1 );

      if( upper )
      {
         Acurrow = Arow;
         Anxtrow = MModAdd1( Acurrow, nprow );

         while( k <= kblks )
         {
            kb  = ( k == 1 ? Aimb1 : ( k == kblks ? Almb1 : Amb ) );
            An -= kb;

            if( myrow == Acurrow )
            {
/*
*  Add contribution of previous blocks of rows of sub( B ) to part of the
*  current block of rows of sub( B )
*/
               zmmadd_tcomp_macro( 
                     &kb, &Bnq0, negone, &Cld, ALPHA, &Bld,
                     nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Solve updated and current part of block of rows of sub( B )
*/
               ztrsm_tcomp_macro( 
                    C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( TRANSA ),
                    C2F_CHAR( DIAG ), &kb, &Bnq0, one, &Ald,
                    &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Add contribution of part of the current block of rows of sub( B ) to the
*  remaining of the contribution of previous blocks of rows of sub( B ). Send
*  this remaining part to next process row.
*/
               if( An > 0 )
               {
                  zgemm_tcomp_macro( 
                       C2F_CHAR( TRANSA ), C2F_CHAR( NOTRAN ), &An, &Bnq0, &kb,
                       one, &Ald, &Bld,
                       one, &Cld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
                  hscal_Czgesd2d( 
                       blacsctxt, An, Bnq0, Cld,
                       Anxtrow, mycol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
               }
            }
            else if( myrow == Anxtrow )
            {
/*
*  Receive contribution of previous blocks of rows of sub( B ) to be added to
*  next block of rows of sub( B )
*/
               if( An > 0 ) ;
            }

            Acurrow  = Anxtrow;
            Anxtrow  = MModAdd1( Acurrow, nprow );
            k       += 1;
         }
      }
      else
      {
         k       = kblks;
         Acurrow = PB_Cindxg2p( An-1, Aimb1, Amb, Arow, Arow, nprow );
         Anxtrow = MModSub1( Acurrow, nprow );

         while( k > 0 )
         {
            kb  = ( k == 1 ? Aimb1 : ( k == kblks ? Almb1 : Amb ) );
            An -= kb;

            if( myrow == Acurrow )
            {
/*
*  Add contribution of previous blocks of rows of sub( B ) to part of the
*  current block of rows of sub( B )
*/
               zmmadd_tcomp_macro( 
                     &kb, &Bnq0, negone, &Cld, ALPHA, &Bld,
                     nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Solve updated and current part of block of rows of sub( B )
*/
               ztrsm_tcomp_macro( 
                    C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( TRANSA ),
                    C2F_CHAR( DIAG ), &kb, &Bnq0, one, &Ald, &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Add contribution of part of the current block of rows of sub( B ) to the
*  remaining of the contribution of previous blocks of rows of sub( B ). Send
*  this remaining part to next process row.
*/
               if( An > 0 )
               {
                  zgemm_tcomp_macro( 
                       C2F_CHAR( TRANSA ), C2F_CHAR( NOTRAN ), &An, &Bnq0, &kb,
                       one, &Ald, &Bld, one, &Cld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
                  hscal_Czgesd2d(
                       blacsctxt, An, Bnq0, Cld, Anxtrow, mycol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
               }
               Amp0 -= kb;
               Bmp0 -= kb;
            }
            else if( myrow == Anxtrow )
            {
/*
*  Receive contribution of previous blocks of rows of sub( B ) to be added to
*  next block of rows of sub( B )
*/
               if( An > 0 ) ;
            }

            Anq0    -= kb;
            Acurrow  = Anxtrow;
            Anxtrow  = MModSub1( Acurrow, nprow );
            k       -= 1;
         }
      }
   }
   else
   {
      kblks = ( An > Ainb1 ? ( An - Ainb1 - 1 ) / Anb + 2 : 1 );

      if( upper )
      {
         k       = kblks;
         Acurcol = PB_Cindxg2p( An-1, Ainb1, Anb, Acol, Acol, npcol );
         Anxtcol = MModSub1( Acurcol, npcol );

         while( k > 0 )
         {
            kb  = ( k == 1 ? Ainb1 : ( k == kblks ? Alnb1 : Anb ) );
            An -= kb;

            if( mycol == Acurcol )
            {
/*
*  Add contribution of previous blocks of columns of sub( B ) to part of the
*  current block of columns of sub( B )
*/
               zmmadd_tcomp_macro( 
                     &Bmp0, &kb, negone, &Cld, ALPHA, &Bld,
                     nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Solve updated and current part of block of columns of sub( B )
*/
               ztrsm_tcomp_macro( 
                    C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( TRANSA ),
                    C2F_CHAR( DIAG ), &Bmp0, &kb, one, &Ald, &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Add contribution of part of the current block of columns of sub( B ) to the
*  remaining of the contribution of previous blocks of columns of sub( B ).
*  Send this remaining part to next process column.
*/
               if( An > 0 )
               {
                  zgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( TRANSA ), &Bmp0, &An, &kb,
                       one, &Bld, &Ald, one, &Cld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
                  hscal_Czgesd2d(
                       blacsctxt, Bmp0, An, Cld, myrow, Anxtcol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
               }
               Anq0 -= kb;
               Bnq0 -= kb;
            }
            else if( mycol == Anxtcol )
            {
/*
*  Receive contribution of previous blocks of columns of sub( B ) to be added
*  to next block of columns of sub( B )
*/
               if( An > 0 ) ;
            }

            Amp0   -= kb;
            Acurcol = Anxtcol;
            Anxtcol = MModSub1( Acurcol, npcol );
            k      -= 1;
         }
      }
      else
      {
         Acurcol = Acol;
         Anxtcol = MModAdd1( Acurcol, npcol );

         while( k <= kblks )
         {
            kb  = ( k == 1 ? Ainb1 : ( k == kblks ? Alnb1 : Anb ) );
            An -= kb;

            if( mycol == Acurcol )
            {
/*
*  Add contribution of previous blocks of columns of sub( B ) to part of the
*  current block of columns of sub( B )
*/
               zmmadd_tcomp_macro( 
                     &Bmp0, &kb, negone, &Cld, ALPHA, &Bld,
                     nprow, npcol, myrow, mycol, tcomp
               );
/*
*              Solve updated and current part of block of columns of sub( B )
*/
               ztrsm_tcomp_macro( 
                    C2F_CHAR( SIDE ), C2F_CHAR( UPLO ), C2F_CHAR( TRANSA ),
                    C2F_CHAR( DIAG ), &Bmp0, &kb, one, &Ald,
                    &Bld,
                    nprow, npcol, myrow, mycol, tcomp
               );
/*
*  Add contribution of part of the current block of columns of sub( B ) to the
*  remaining of the contribution of previous blocks of columns of sub( B ).
*  Send this remaining part to next process column.
*/
               if( An > 0 )
               {
                  zgemm_tcomp_macro( 
                       C2F_CHAR( NOTRAN ), C2F_CHAR( TRANSA ), &Bmp0, &An, &kb,
                       one, &Bld, &Ald,
                       one, &Cld,
                       nprow, npcol, myrow, mycol, tcomp
                  );
                  hscal_Czgesd2d( 
                       blacsctxt, Bmp0, An, Cld,
                       myrow, Anxtcol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );
               }
            }
            else if( mycol == Anxtcol )
            {
/*
*  Receive contribution of previous blocks of columns of sub( B ) to be added
*  to next block of columns of sub( B ).
*/
               if( An > 0 ) ;
            }

            Acurcol  = Anxtcol;
            Anxtcol  = MModAdd1( Acurcol, npcol );
            k       += 1;
         }
      }
   }

   Cblacs_gridexit_nocomm(blacsctxt);

/*
*  End of PB_CptrsmAB1
*/
}

#endif
