/* ---------------------------------------------------------------------
*
*  -- PBLAS auxiliary routine (version 2.0) --
*     University of Tennessee, Knoxville, Oak Ridge National Laboratory,
*     and University of California, Berkeley.
*     April 1, 1998
*
*  ---------------------------------------------------------------------
*/

#ifndef __HSCAL_PB_CWARN_HSCALFUNC_HH
#define __HSCAL_PB_CWARN_HSCALFUNC_HH


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
/*
*  Include files
*/
#include "../../../include/hscalapack.h"

void hscal_PB_Cwarn( 
     int nprow, int npcol, int myrow, int mycol,
     int LINE, char * ROUT, char * FORM, ... 
)
{
/*
*  Purpose
*  =======
*
*  PB_Cwarn  is  an error handler for the PBLAS routines.  This  routine
*  displays an error message on stderr.
*
*  Arguments
*  =========
*
*  ICTXT   (local input) INTEGER
*          On entry,  ICTXT  specifies the BLACS context handle, indica-
*          ting the global  context of the operation. The context itself
*          is global, but the value of ICTXT is local.
*
*  LINE    (local input) INTEGER
*          On entry,  LINE  specifies the line  number in the file where
*          the error has occured. When  LINE is not a valid line number,
*
*  ROUT    (global input) pointer to CHAR
*          On entry, ROUT specifies the name of the routine calling this
*          error handler.
*
*  FORM    (local input) pointer to CHAR
*          On entry,  FORM  is a  control  string  specifying the format
*          conversion of its following arguments.
*
*  ...     (local input)
*          On entry,  FORM  is a  control  string  specifying the format
*          On entry,  the expressions that are to be  evaluated and con-
*          verted  according  to the formats in the control string  FORM
*          and then placed in the output stream.
*
*  -- Written on April 1, 1998 by
*     R. Clint Whaley, University of Tennessee, Knoxville 37996, USA.
*
*  ---------------------------------------------------------------------
*/
   va_list        argptr;
   int            iam;
   char           cline[100];
/* ..
*  .. External Functions ..
*/
   va_start( argptr, FORM );
   vsprintf( cline, FORM, argptr );
   va_end( argptr );

   if( nprow != -1 ) iam = myrow*npcol+mycol;
   else              iam = -1;
/*
*  Display an error message
*/
   if( LINE <= 0 )
   {
      (void) fprintf( stderr, "%s'%s'\n%s{%d,%d}, %s%d, %s'%s' total prows=%d, total pcols=%d.\n\n",
                      "PBLAS ERROR ", cline, "from ", myrow, mycol, "pnum=",
                      iam, ", in routine ", ROUT , nprow, npcol);
   }
   else
   {
      (void) fprintf( stderr, "%s'%s'\n%s{%d,%d}, %s%d, %s%d%s'%s' total prows=%d, total pcols=%d.\n\n",
                      "PBLAS ERROR ", cline, "from ", myrow, mycol, "pnum=",
                      iam, ", on line ", LINE,
                      " of routine ", ROUT , nprow, npcol);
   }
/*
*  End of PB_Cwarn
*/
}

#endif
