
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_cagemv_benchmark(
    char    *trans, 
    int     *m, 
    int     *n, 
    hscalreal    *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *x, 
    int     *incx, 
    hscalreal    *beta, 
    hscalreal    *y, 
    int     *incy
)
{
    return hscal_cagemv_benchmark_i(
           trans,
           (hscalinteger*)m,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           x,
           (hscalinteger*)incx,
           beta,
           y,
           (hscalinteger*)incy
    );
}

/*------------------------------------------------------------------*/

int hscal_cagemv_benchmark_i(
    char *trans, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscalreal *alpha, 
    hscalcomplex *a, 
    hscalinteger *lda, 
    hscalcomplex *x, 
    hscalinteger *incx, 
    hscalreal *beta, 
    hscalreal *y, 
    hscalinteger *incy
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2, i__3, i__4;
    hscalreal r__1, r__2, r__3, r__4;

    /* Local variables */
    hscalinteger i__, j, ix, iy, jx, jy, kx, ky, info;
    hscalreal absx, temp;
    hscalinteger lenx, leny;
    hscalreal talpha;

/*  -- PBLAS auxiliary routine (version 2.0) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     April 1, 1998 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  CAGEMV performs one of the matrix-vector operations */

/*     y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( conjg( A' ) )*abs( x ) + abs( beta*y ), */

/*  where  alpha  and  beta  are hscalreal scalars, y is a real vector, x is a */
/*  vector and A is an m by n matrix. */

/*  Arguments */
/*  ========= */

/*  TRANS   (input) CHARACTER*1 */
/*          On entry,  TRANS  specifies the  operation to be performed as */
/*          follows: */

/*             TRANS = 'N' or 'n': */
/*                y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ) */

/*             TRANS = 'T' or 't': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*             TRANS = 'C' or 'c': */
/*                y := abs( alpha )*abs( conjg( A' ) )*abs( x ) + */
/*                     abs( beta*y ) */

/*  M       (input) INTEGER */
/*          On entry, M  specifies the number of rows of the matrix  A. M */
/*          must be at least zero. */

/*  N       (input) INTEGER */
/*          On entry, N  specifies the number of columns of the matrix A. */
/*          N must be at least zero. */

/*  ALPHA   (input) REAL */
/*          On entry, ALPHA specifies the hscalreal scalar alpha. */

/*  A       (input) COMPLEX array of dimension ( LDA, n ). */
/*          On entry, A  is an array of dimension ( LDA, N ). The leading */
/*          m by n part of the array  A  must contain the matrix of coef- */
/*          ficients. */

/*  LDA     (input) INTEGER */
/*          On entry, LDA specifies the leading dimension of the array A. */
/*          LDA must be at least max( 1, M ). */

/*  X       (input) COMPLEX array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCX ) ) when TRANS = 'N' or 'n' and  at */
/*          least ( 1 + ( m - 1 )*abs( INCX ) ) otherwise.  Before entry, */
/*          the incremented array X must contain the vector x. */

/*  INCX    (input) INTEGER */
/*          On entry, INCX specifies the increment for the elements of X. */
/*          INCX must not be zero. */

/*  BETA    (input) REAL */
/*          On entry,  BETA  specifies the hscalreal scalar beta. When BETA is */
/*          supplied as zero then Y need not be set on input. */

/*  Y       (input/output) REAL array of dimension at least */
/*          ( 1 + ( m - 1 )*abs( INCY ) ) when TRANS = 'N' or 'n' and  at */
/*          least ( 1 + ( n - 1 )*abs( INCY ) ) otherwise.  Before  entry */
/*          with BETA non-zero, the incremented array  Y must contain the */
/*          vector y. On exit, the incremented array  Y is overwritten by */
/*          the updated vector y. */

/*  INCY    (input) INTEGER */
/*          On entry, INCY specifies the increment for the elements of Y. */
/*          INCY must not be zero. */

/*  -- Written on April 1, 1998 by */
/*     Antoine Petitet, University  of  Tennessee, Knoxville 37996, USA. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Statement Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Test the input parameters. */

    /* Parameter adjustments */
    a_dim1 = *lda;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --x;
    --y;

    /* Function Body */
    info = 0;
    if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")) {
	info = 1;
    } else if (*m < 0) {
	info = 2;
    } else if (*n < 0) {
	info = 3;
    } else if (*lda < max(1,*m)) {
	info = 6;
    } else if (*incx == 0) {
	info = 8;
    } else if (*incy == 0) {
	info = 11;
    }
    if (info != 0) {
	hscal_xerbla("CAGEMV", &info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible. */

    if (*m == 0 || *n == 0 || *alpha == 0.f && *beta == 1.f) {
	return HSCAL_SUCCESS;
    }

/*     Set  LENX  and  LENY, the lengths of the vectors x and y, and set */
/*     up the start points in  X  and  Y. */

    if (hscal_lsame(trans, "N")) {
	lenx = *n;
	leny = *m;
    } else {
	lenx = *m;
	leny = *n;
    }
    if (*incx > 0) {
	kx = 1;
    } else {
	kx = 1 - (lenx - 1) * *incx;
    }
    if (*incy > 0) {
	ky = 1;
    } else {
	ky = 1 - (leny - 1) * *incy;
    }

/*     Start the operations. In this version the elements of A are */
/*     accessed sequentially with one pass through A. */

/*     First form  y := abs( beta*y ). */

    if (*incy == 1) {
	if (*beta == 0.f) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = 0.f;
/* L10: */
	    }
	} else if (*beta == 1.f) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (r__1 = y[i__], dabs(r__1));
/* L20: */
	    }
	} else {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (r__1 = *beta * y[i__], dabs(r__1));
/* L30: */
	    }
	}
    } else {
	iy = ky;
	if (*beta == 0.f) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = 0.f;
		iy += *incy;
/* L40: */
	    }
	} else if (*beta == 1.f) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (r__1 = y[iy], dabs(r__1));
		iy += *incy;
/* L50: */
	    }
	} else {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (r__1 = *beta * y[iy], dabs(r__1));
		iy += *incy;
/* L60: */
	    }
	}
    }

    if (*alpha == 0.f) {
	return HSCAL_SUCCESS;
    }

    talpha = dabs(*alpha);

    if (hscal_lsame(trans, "N")) {

/*        Form  y := abs( alpha ) * abs( A ) * abs( x ) + y. */

	jx = kx;
	if (*incy == 1) {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		i__2 = jx;
		absx = (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = hscal_r_imag(&x[jx])
			, dabs(r__2));
		if (absx != 0.f) {
		    temp = talpha * absx;
		    i__2 = *m;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			y[i__] += temp * ((r__1 = a[i__3].r, dabs(r__1)) + (
				r__2 = hscal_r_imag(&a[i__ + j * a_dim1]), dabs(
				r__2)));
                        
/* L70: */
		    }
		}
		jx += *incx;
                
/* L80: */
	    }
	} else {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		i__2 = jx;
		absx = (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = hscal_r_imag(&x[jx])
			, dabs(r__2));
		if (absx != 0.f) {
		    temp = talpha * absx;
		    iy = ky;
		    i__2 = *m;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			y[iy] += temp * ((r__1 = a[i__3].r, dabs(r__1)) + (
				r__2 = hscal_r_imag(&a[i__ + j * a_dim1]), dabs(
				r__2)));
			iy += *incy;
                        
/* L90: */
		    }
		}
		jx += *incx;
                
/* L100: */
	    }
	}

    } else {

/*        Form  y := abs( alpha ) * abs( A' ) * abs( x ) + y. */

	jy = ky;
	if (*incx == 1) {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		temp = 0.f;
		i__2 = *m;
		for (i__ = 1; i__ <= i__2; ++i__) {
		    i__3 = i__ + j * a_dim1;
		    i__4 = i__;
		    temp += ((r__1 = a[i__3].r, dabs(r__1)) + (r__2 = hscal_r_imag(&
			    a[i__ + j * a_dim1]), dabs(r__2))) * ((r__3 = x[
			    i__4].r, dabs(r__3)) + (r__4 = hscal_r_imag(&x[i__]), 
			    dabs(r__4)));
                    
/* L110: */
		}
		y[jy] += talpha * temp;
		jy += *incy;
                
/* L120: */
	    }
	} else {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		temp = 0.f;
		ix = kx;
		i__2 = *m;
		for (i__ = 1; i__ <= i__2; ++i__) {
		    i__3 = i__ + j * a_dim1;
		    i__4 = ix;
		    temp += ((r__1 = a[i__3].r, dabs(r__1)) + (r__2 = hscal_r_imag(&
			    a[i__ + j * a_dim1]), dabs(r__2))) * ((r__3 = x[
			    i__4].r, dabs(r__3)) + (r__4 = hscal_r_imag(&x[ix]), 
			    dabs(r__4)));
		    ix += *incx;
                    
/* L130: */
		}
		y[jy] += talpha * temp;
		jy += *incy;
                
/* L140: */
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of CAGEMV */

} /* cagemv_ */

/*------------------------------------------------------------------*/

int hscal_sagemv_benchmark(
    char *trans, 
    int  *m, 
    int  *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int  *lda, 
    hscalreal *x, 
    int  *incx, 
    hscalreal *beta, 
    hscalreal *y,
    int  *incy
)
{
    return hscal_sagemv_benchmark_i(
           trans,
           (hscalinteger*)m,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           x,
           (hscalinteger*)incx,
           beta,
           y,
           (hscalinteger*)incy
    );
}

/*------------------------------------------------------------------*/

int hscal_sagemv_benchmark_i(
    char *trans, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    hscalinteger *lda, 
    hscalreal *x, 
    hscalinteger *incx, 
    hscalreal *beta, 
    hscalreal *y,
    hscalinteger *incy
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2;
    hscalreal r__1;

    /* Local variables */
    hscalinteger i__, j, ix, iy, jx, jy, kx, ky, info;
    hscalreal absx, temp;
    hscalinteger lenx, leny;
    hscalreal talpha;

/*  -- PBLAS auxiliary routine (version 2.0) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     April 1, 1998 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  SAGEMV performs one of the matrix-vector operations */

/*     y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ), */

/*  where  alpha  and  beta  are hscalreal scalars, y is a real vector, x is a */
/*  vector and A is an m by n matrix. */

/*  Arguments */
/*  ========= */

/*  TRANS   (input) CHARACTER*1 */
/*          On entry,  TRANS  specifies the  operation to be performed as */
/*          follows: */

/*             TRANS = 'N' or 'n': */
/*                y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ) */

/*             TRANS = 'T' or 't': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*             TRANS = 'C' or 'c': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*  M       (input) INTEGER */
/*          On entry, M  specifies the number of rows of the matrix  A. M */
/*          must be at least zero. */

/*  N       (input) INTEGER */
/*          On entry, N  specifies the number of columns of the matrix A. */
/*          N must be at least zero. */

/*  ALPHA   (input) REAL */
/*          On entry, ALPHA specifies the hscalreal scalar alpha. */

/*  A       (input) REAL array of dimension ( LDA, n ). */
/*          On entry, A  is an array of dimension ( LDA, N ). The leading */
/*          m by n part of the array  A  must contain the matrix of coef- */
/*          ficients. */

/*  LDA     (input) INTEGER */
/*          On entry, LDA specifies the leading dimension of the array A. */
/*          LDA must be at least max( 1, M ). */

/*  X       (input) REAL array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCX ) ) when TRANS = 'N' or 'n' and  at */
/*          least ( 1 + ( m - 1 )*abs( INCX ) ) otherwise.  Before entry, */
/*          the incremented array X must contain the vector x. */

/*  INCX    (input) INTEGER */
/*          On entry, INCX specifies the increment for the elements of X. */
/*          INCX must not be zero. */

/*  BETA    (input) REAL */
/*          On entry,  BETA  specifies the hscalreal scalar beta. When BETA is */
/*          supplied as zero then Y need not be set on input. */

/*  Y       (input/output) REAL array of dimension at least */
/*          ( 1 + ( m - 1 )*abs( INCY ) ) when TRANS = 'N' or 'n' and  at */
/*          least ( 1 + ( n - 1 )*abs( INCY ) ) otherwise.  Before  entry */
/*          with BETA non-zero, the incremented array  Y must contain the */
/*          vector y. On exit, the incremented array  Y is overwritten by */
/*          the updated vector y. */

/*  INCY    (input) INTEGER */
/*          On entry, INCY specifies the increment for the elements of Y. */
/*          INCY must not be zero. */

/*  -- Written on April 1, 1998 by */
/*     Antoine Petitet, University  of  Tennessee, Knoxville 37996, USA. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Test the input parameters. */

    /* Parameter adjustments */
    a_dim1 = *lda;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --x;
    --y;

    /* Function Body */
    info = 0;
    if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")) {
	info = 1;
    } else if (*m < 0) {
	info = 2;
    } else if (*n < 0) {
	info = 3;
    } else if (*lda < max(1,*m)) {
	info = 6;
    } else if (*incx == 0) {
	info = 8;
    } else if (*incy == 0) {
	info = 11;
    }
    if (info != 0) {
	hscal_xerbla("SAGEMV", &info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible. */

    if (*m == 0 || *n == 0 || *alpha == 0.f && *beta == 1.f) {
	return HSCAL_SUCCESS;
    }

/*     Set  LENX  and  LENY, the lengths of the vectors x and y, and set */
/*     up the start points in  X  and  Y. */

    if (hscal_lsame(trans, "N")) {
	lenx = *n;
	leny = *m;
    } else {
	lenx = *m;
	leny = *n;
    }
    if (*incx > 0) {
	kx = 1;
    } else {
	kx = 1 - (lenx - 1) * *incx;
    }
    if (*incy > 0) {
	ky = 1;
    } else {
	ky = 1 - (leny - 1) * *incy;
    }

/*     Start the operations. In this version the elements of A are */
/*     accessed sequentially with one pass through A. */

/*     First form  y := abs( beta*y ). */

    if (*incy == 1) {
	if (*beta == 0.f) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = 0.f;
/* L10: */
	    }
	} else if (*beta == 1.f) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (r__1 = y[i__], dabs(r__1));
/* L20: */
	    }
	} else {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (r__1 = *beta * y[i__], dabs(r__1));
/* L30: */
	    }
	}
    } else {
	iy = ky;
	if (*beta == 0.f) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = 0.f;
		iy += *incy;
/* L40: */
	    }
	} else if (*beta == 1.f) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (r__1 = y[iy], dabs(r__1));
		iy += *incy;
/* L50: */
	    }
	} else {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (r__1 = *beta * y[iy], dabs(r__1));
		iy += *incy;
/* L60: */
	    }
	}
    }

    if (*alpha == 0.f) {
	return HSCAL_SUCCESS;
    }

    talpha = dabs(*alpha);

    if (hscal_lsame(trans, "N")) {

/*        Form  y := abs( alpha ) * abs( A ) * abs( x ) + y. */

	jx = kx;
	if (*incy == 1) {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		absx = (r__1 = x[jx], dabs(r__1));
		if (absx != 0.f) {
		    temp = talpha * absx;
		    i__2 = *m;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			y[i__] += temp * (r__1 = a[i__ + j * a_dim1], dabs(
				r__1));
                        
/* L70: */
		    }
		}
		jx += *incx;
                
/* L80: */
	    }
	} else {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		absx = (r__1 = x[jx], dabs(r__1));
		if (absx != 0.f) {
		    temp = talpha * absx;
		    iy = ky;
		    i__2 = *m;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			y[iy] += temp * (r__1 = a[i__ + j * a_dim1], dabs(
				r__1));
			iy += *incy;
                        
/* L90: */
		    }
		}
		jx += *incx;
                
/* L100: */
	    }
	}

    } else {

/*        Form  y := abs( alpha ) * abs( A' ) * abs( x ) + y. */

	jy = ky;
	if (*incx == 1) {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		temp = 0.f;
		i__2 = *m;
		for (i__ = 1; i__ <= i__2; ++i__) {
		    temp += (r__1 = a[i__ + j * a_dim1] * x[i__], dabs(r__1));
                    
/* L110: */
		}
		y[jy] += talpha * temp;
		jy += *incy;
                
/* L120: */
	    }
	} else {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		temp = 0.f;
		ix = kx;
		i__2 = *m;
		for (i__ = 1; i__ <= i__2; ++i__) {
		    temp += (r__1 = a[i__ + j * a_dim1] * x[ix], dabs(r__1));
		    ix += *incx;
                    
/* L130: */
		}
		y[jy] += talpha * temp;
		jy += *incy;
                
/* L140: */
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of SAGEMV */

} /* sagemv_ */

/*------------------------------------------------------------------*/

int hscal_dagemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublereal * alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *x, 
    int     *incx,
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    int     *incy
)
{
    return hscal_dagemv_benchmark_i(
           trans,
           (hscalinteger*)m,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           x,
           (hscalinteger*)incx,
           beta,
           y,
           (hscalinteger*)incy
    );
}

/*------------------------------------------------------------------*/

int hscal_dagemv_benchmark_i(
    char *trans, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscaldoublereal * alpha, 
    hscaldoublereal *a, 
    hscalinteger *lda, 
    hscaldoublereal *x, 
    hscalinteger *incx,
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    hscalinteger *incy
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2;
    hscaldoublereal d__1;

    /* Local variables */
    hscalinteger i__, j, ix, iy, jx, jy, kx, ky, info;
    hscaldoublereal absx, temp;
    hscalinteger lenx, leny;
    hscaldoublereal talpha;


/*  -- PBLAS auxiliary routine (version 2.0) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     April 1, 1998 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  DAGEMV performs one of the matrix-vector operations */

/*     y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ), */

/*  where  alpha  and  beta  are hscalreal scalars, y is a real vector, x is a */
/*  vector and A is an m by n matrix. */

/*  Arguments */
/*  ========= */

/*  TRANS   (input) CHARACTER*1 */
/*          On entry,  TRANS  specifies the  operation to be performed as */
/*          follows: */

/*             TRANS = 'N' or 'n': */
/*                y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ) */

/*             TRANS = 'T' or 't': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*             TRANS = 'C' or 'c': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*  M       (input) INTEGER */
/*          On entry, M  specifies the number of rows of the matrix  A. M */
/*          must be at least zero. */

/*  N       (input) INTEGER */
/*          On entry, N  specifies the number of columns of the matrix A. */
/*          N must be at least zero. */

/*  ALPHA   (input) DOUBLE PRECISION */
/*          On entry, ALPHA specifies the hscalreal scalar alpha. */

/*  A       (input) DOUBLE PRECISION array of dimension ( LDA, n ). */
/*          On entry, A  is an array of dimension ( LDA, N ). The leading */
/*          m by n part of the array  A  must contain the matrix of coef- */
/*          ficients. */

/*  LDA     (input) INTEGER */
/*          On entry, LDA specifies the leading dimension of the array A. */
/*          LDA must be at least max( 1, M ). */

/*  X       (input) DOUBLE PRECISION array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCX ) ) when TRANS = 'N' or 'n' and  at */
/*          least ( 1 + ( m - 1 )*abs( INCX ) ) otherwise.  Before entry, */
/*          the incremented array X must contain the vector x. */

/*  INCX    (input) INTEGER */
/*          On entry, INCX specifies the increment for the elements of X. */
/*          INCX must not be zero. */

/*  BETA    (input) DOUBLE PRECISION */
/*          On entry,  BETA  specifies the hscalreal scalar beta. When BETA is */
/*          supplied as zero then Y need not be set on input. */

/*  Y       (input/output) DOUBLE PRECISION array of dimension at least */
/*          ( 1 + ( m - 1 )*abs( INCY ) ) when TRANS = 'N' or 'n' and  at */
/*          least ( 1 + ( n - 1 )*abs( INCY ) ) otherwise.  Before  entry */
/*          with BETA non-zero, the incremented array  Y must contain the */
/*          vector y. On exit, the incremented array  Y is overwritten by */
/*          the updated vector y. */

/*  INCY    (input) INTEGER */
/*          On entry, INCY specifies the increment for the elements of Y. */
/*          INCY must not be zero. */

/*  -- Written on April 1, 1998 by */
/*     Antoine Petitet, University  of  Tennessee, Knoxville 37996, USA. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Test the input parameters. */

    /* Parameter adjustments */
    a_dim1 = *lda;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --x;
    --y;

    /* Function Body */
    info = 0;
    if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")) {
	info = 1;
    } else if (*m < 0) {
	info = 2;
    } else if (*n < 0) {
	info = 3;
    } else if (*lda < max(1,*m)) {
	info = 6;
    } else if (*incx == 0) {
	info = 8;
    } else if (*incy == 0) {
	info = 11;
    }
    if (info != 0) {
	hscal_xerbla("DAGEMV", &info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible. */

    if (*m == 0 || *n == 0 || *alpha == 0. && *beta == 1.) {
	return HSCAL_SUCCESS;
    }

/*     Set  LENX  and  LENY, the lengths of the vectors x and y, and set */
/*     up the start points in  X  and  Y. */

    if (hscal_lsame(trans, "N")) {
	lenx = *n;
	leny = *m;
    } else {
	lenx = *m;
	leny = *n;
    }
    if (*incx > 0) {
	kx = 1;
    } else {
	kx = 1 - (lenx - 1) * *incx;
    }
    if (*incy > 0) {
	ky = 1;
    } else {
	ky = 1 - (leny - 1) * *incy;
    }

/*     Start the operations. In this version the elements of A are */
/*     accessed sequentially with one pass through A. */

/*     First form  y := abs( beta*y ). */

    if (*incy == 1) {
	if (*beta == 0.) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = 0.;
/* L10: */
	    }
	} else if (*beta == 1.) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (d__1 = y[i__], abs(d__1));
/* L20: */
	    }
	} else {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (d__1 = *beta * y[i__], abs(d__1));
/* L30: */
	    }
	}
    } else {
	iy = ky;
	if (*beta == 0.) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = 0.;
		iy += *incy;
/* L40: */
	    }
	} else if (*beta == 1.) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (d__1 = y[iy], abs(d__1));
		iy += *incy;
/* L50: */
	    }
	} else {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (d__1 = *beta * y[iy], abs(d__1));
		iy += *incy;
/* L60: */
	    }
	}
    }

    if (*alpha == 0.) {
	return HSCAL_SUCCESS;
    }

    talpha = abs(*alpha);

    if (hscal_lsame(trans, "N")) {

/*        Form  y := abs( alpha ) * abs( A ) * abs( x ) + y. */

	jx = kx;
	if (*incy == 1) {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		absx = (d__1 = x[jx], abs(d__1));
		if (absx != 0.) {
		    temp = talpha * absx;
		    i__2 = *m;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			y[i__] += temp * (d__1 = a[i__ + j * a_dim1], abs(
				d__1));
                        
/* L70: */
		    }
		}
		jx += *incx;
                
/* L80: */
	    }
	} else {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		absx = (d__1 = x[jx], abs(d__1));
		if (absx != 0.) {
		    temp = talpha * absx;
		    iy = ky;
		    i__2 = *m;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			y[iy] += temp * (d__1 = a[i__ + j * a_dim1], abs(d__1)
				);
			iy += *incy;
                        
/* L90: */
		    }
		}
		jx += *incx;
                
/* L100: */
	    }
	}

    } else {

/*        Form  y := abs( alpha ) * abs( A' ) * abs( x ) + y. */

	jy = ky;
	if (*incx == 1) {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		temp = 0.;
		i__2 = *m;
		for (i__ = 1; i__ <= i__2; ++i__) {
		    temp += (d__1 = a[i__ + j * a_dim1] * x[i__], abs(d__1));
                    
/* L110: */
		}
		y[jy] += talpha * temp;
		jy += *incy;
                
/* L120: */
	    }
	} else {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		temp = 0.;
		ix = kx;
		i__2 = *m;
		for (i__ = 1; i__ <= i__2; ++i__) {
		    temp += (d__1 = a[i__ + j * a_dim1] * x[ix], abs(d__1));
		    ix += *incx;
                    
/* L130: */
		}
		y[jy] += talpha * temp;
		jy += *incy;
                
/* L140: */
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of DAGEMV */

} /* dagemv_ */

/*------------------------------------------------------------------*/

int hscal_zagemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *x, 
    int     *incx, 
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    int     *incy
)
{
    return hscal_zagemv_benchmark_i(
           trans,
           (hscalinteger*)m,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           x,
           (hscalinteger*)incx,
           beta,
           y,
           (hscalinteger*)incy
    );
}

/*------------------------------------------------------------------*/

int hscal_zagemv_benchmark_i(
    char *trans, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscaldoublereal *alpha, 
    hscaldoublecomplex *a, 
    hscalinteger *lda, 
    hscaldoublecomplex *x, 
    hscalinteger *incx, 
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    hscalinteger *incy
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2, i__3, i__4;
    hscaldoublereal d__1, d__2, d__3, d__4;

    /* Local variables */
    hscalinteger i__, j, ix, iy, jx, jy, kx, ky, info;
    hscaldoublereal absx, temp;
    hscalinteger lenx, leny;
    hscaldoublereal talpha;

/*  -- PBLAS auxiliary routine (version 2.0) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     April 1, 1998 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  ZAGEMV performs one of the matrix-vector operations */

/*     y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( conjg( A' ) )*abs( x ) + abs( beta*y ), */

/*  where  alpha  and  beta  are hscalreal scalars, y is a real vector, x is a */
/*  vector and A is an m by n matrix. */

/*  Arguments */
/*  ========= */

/*  TRANS   (input) CHARACTER*1 */
/*          On entry,  TRANS  specifies the  operation to be performed as */
/*          follows: */

/*             TRANS = 'N' or 'n': */
/*                y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ) */

/*             TRANS = 'T' or 't': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*             TRANS = 'C' or 'c': */
/*                y := abs( alpha )*abs( conjg( A' ) )*abs( x ) + */
/*                     abs( beta*y ) */

/*  M       (input) INTEGER */
/*          On entry, M  specifies the number of rows of the matrix  A. M */
/*          must be at least zero. */

/*  N       (input) INTEGER */
/*          On entry, N  specifies the number of columns of the matrix A. */
/*          N must be at least zero. */

/*  ALPHA   (input) DOUBLE PRECISION */
/*          On entry, ALPHA specifies the hscalreal scalar alpha. */

/*  A       (input) COMPLEX*16 array of dimension ( LDA, n ). */
/*          On entry, A  is an array of dimension ( LDA, N ). The leading */
/*          m by n part of the array  A  must contain the matrix of coef- */
/*          ficients. */

/*  LDA     (input) INTEGER */
/*          On entry, LDA specifies the leading dimension of the array A. */
/*          LDA must be at least max( 1, M ). */

/*  X       (input) COMPLEX*16 array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCX ) ) when TRANS = 'N' or 'n' and  at */
/*          least ( 1 + ( m - 1 )*abs( INCX ) ) otherwise.  Before entry, */
/*          the incremented array X must contain the vector x. */

/*  INCX    (input) INTEGER */
/*          On entry, INCX specifies the increment for the elements of X. */
/*          INCX must not be zero. */

/*  BETA    (input) DOUBLE PRECISION */
/*          On entry,  BETA  specifies the hscalreal scalar beta. When BETA is */
/*          supplied as zero then Y need not be set on input. */

/*  Y       (input/output) DOUBLE PRECISION array of dimension at least */
/*          ( 1 + ( m - 1 )*abs( INCY ) ) when TRANS = 'N' or 'n' and  at */
/*          least ( 1 + ( n - 1 )*abs( INCY ) ) otherwise.  Before  entry */
/*          with BETA non-zero, the incremented array  Y must contain the */
/*          vector y. On exit, the incremented array  Y is overwritten by */
/*          the updated vector y. */

/*  INCY    (input) INTEGER */
/*          On entry, INCY specifies the increment for the elements of Y. */
/*          INCY must not be zero. */

/*  -- Written on April 1, 1998 by */
/*     Antoine Petitet, University  of  Tennessee, Knoxville 37996, USA. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Statement Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Test the input parameters. */

    /* Parameter adjustments */
    a_dim1 = *lda;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --x;
    --y;

    /* Function Body */
    info = 0;
    if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")) {
	info = 1;
    } else if (*m < 0) {
	info = 2;
    } else if (*n < 0) {
	info = 3;
    } else if (*lda < max(1,*m)) {
	info = 6;
    } else if (*incx == 0) {
	info = 8;
    } else if (*incy == 0) {
	info = 11;
    }
    if (info != 0) {
	hscal_xerbla("ZAGEMV", &info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible. */

    if (*m == 0 || *n == 0 || *alpha == 0. && *beta == 1.) {
	return HSCAL_SUCCESS;
    }

/*     Set  LENX  and  LENY, the lengths of the vectors x and y, and set */
/*     up the start points in  X  and  Y. */

    if (hscal_lsame(trans, "N")) {
	lenx = *n;
	leny = *m;
    } else {
	lenx = *m;
	leny = *n;
    }
    if (*incx > 0) {
	kx = 1;
    } else {
	kx = 1 - (lenx - 1) * *incx;
    }
    if (*incy > 0) {
	ky = 1;
    } else {
	ky = 1 - (leny - 1) * *incy;
    }

/*     Start the operations. In this version the elements of A are */
/*     accessed sequentially with one pass through A. */

/*     First form  y := abs( beta*y ). */

    if (*incy == 1) {
	if (*beta == 0.) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = 0.;
/* L10: */
	    }
	} else if (*beta == 1.) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (d__1 = y[i__], abs(d__1));
/* L20: */
	    }
	} else {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (d__1 = *beta * y[i__], abs(d__1));
/* L30: */
	    }
	}
    } else {
	iy = ky;
	if (*beta == 0.) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = 0.;
		iy += *incy;
/* L40: */
	    }
	} else if (*beta == 1.) {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (d__1 = y[iy], abs(d__1));
		iy += *incy;
/* L50: */
	    }
	} else {
	    i__1 = leny;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (d__1 = *beta * y[iy], abs(d__1));
		iy += *incy;
/* L60: */
	    }
	}
    }

    if (*alpha == 0.) {
	return HSCAL_SUCCESS;
    }

    talpha = abs(*alpha);

    if (hscal_lsame(trans, "N")) {

/*        Form  y := abs( alpha ) * abs( A ) * abs( x ) + y. */

	jx = kx;
	if (*incy == 1) {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		i__2 = jx;
		absx = (d__1 = x[i__2].r, abs(d__1)) + (d__2 = hscal_d_imag(&x[jx]),
			 abs(d__2));
		if (absx != 0.) {
		    temp = talpha * absx;
		    i__2 = *m;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			y[i__] += temp * ((d__1 = a[i__3].r, abs(d__1)) + (
				d__2 = hscal_d_imag(&a[i__ + j * a_dim1]), abs(d__2)
				));
                        
/* L70: */
		    }
		}
		jx += *incx;
                
/* L80: */
	    }
	} else {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		i__2 = jx;
		absx = (d__1 = x[i__2].r, abs(d__1)) + (d__2 = hscal_d_imag(&x[jx]),
			 abs(d__2));
		if (absx != 0.) {
		    temp = talpha * absx;
		    iy = ky;
		    i__2 = *m;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			y[iy] += temp * ((d__1 = a[i__3].r, abs(d__1)) + (
				d__2 = hscal_d_imag(&a[i__ + j * a_dim1]), abs(d__2)
				));
			iy += *incy;
                        
/* L90: */
		    }
		}
		jx += *incx;
                
/* L100: */
	    }
	}

    } else {

/*        Form  y := abs( alpha ) * abs( A' ) * abs( x ) + y. */

	jy = ky;
	if (*incx == 1) {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		temp = 0.;
		i__2 = *m;
		for (i__ = 1; i__ <= i__2; ++i__) {
		    i__3 = i__ + j * a_dim1;
		    i__4 = i__;
		    temp += ((d__1 = a[i__3].r, abs(d__1)) + (d__2 = hscal_d_imag(&
			    a[i__ + j * a_dim1]), abs(d__2))) * ((d__3 = x[
			    i__4].r, abs(d__3)) + (d__4 = hscal_d_imag(&x[i__]), 
			    abs(d__4)));
                    
/* L110: */
		}
		y[jy] += talpha * temp;
		jy += *incy;
                
/* L120: */
	    }
	} else {
	    i__1 = *n;
	    for (j = 1; j <= i__1; ++j) {
		temp = 0.;
		ix = kx;
		i__2 = *m;
		for (i__ = 1; i__ <= i__2; ++i__) {
		    i__3 = i__ + j * a_dim1;
		    i__4 = ix;
		    temp += ((d__1 = a[i__3].r, abs(d__1)) + (d__2 = hscal_d_imag(&
			    a[i__ + j * a_dim1]), abs(d__2))) * ((d__3 = x[
			    i__4].r, abs(d__3)) + (d__4 = hscal_d_imag(&x[ix]), abs(
			    d__4)));
		    ix += *incx;
                    
/* L130: */
		}
		y[jy] += talpha * temp;
		jy += *incy;
                
/* L140: */
	    }
	}
    }

    return HSCAL_SUCCESS;

/*     End of ZAGEMV */

} /* zagemv_ */

/*------------------------------------------------------------------*/
