
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_catrmv_benchmark(
    char *uplo,
    char *trans,
    char *diag,
    int     *n,
    hscalreal * alpha,
    hscalcomplex *a,
    int     *lda,
    hscalcomplex *x,
    int     *incx,
    hscalreal *beta,
    hscalreal *y,
    int     *incy
)
{
    return hscal_catrmv_benchmark_i(
           uplo,
           trans,
           diag,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           x,
           (hscalinteger*)incx,
           beta,
           y,
           (hscalinteger*)incy
    );
}

/*------------------------------------------------------------------*/

int hscal_catrmv_benchmark_i(
    char *uplo,
    char *trans,
    char *diag,
    hscalinteger *n,
    hscalreal * alpha,
    hscalcomplex *a,
    hscalinteger *lda,
    hscalcomplex *x,
    hscalinteger *incx,
    hscalreal *beta,
    hscalreal *y,
    hscalinteger *incy
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2, i__3, i__4;
    hscalreal r__1, r__2, r__3, r__4;

    /* Local variables */
    static hscalinteger i__, j, ix, iy, jx, jy, kx, ky, info;
    static hscalreal absx, temp;
    static hscalreal talpha;
    static hscallogical nounit;

/*  -- PBLAS auxiliary routine (version 2.0) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     April 1, 1998 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  CATRMV performs one of the matrix-vector operations */

/*     y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( conjg( A' ) )*abs( x ) + abs( beta*y ), */

/*  where  alpha  and  beta  are hscalreal scalars, y is a real vector, x is a */
/*  vector and A is an n by n unit or non-unit, upper or lower triangular */
/*  matrix. */

/*  Arguments */
/*  ========= */

/*  UPLO    (input) CHARACTER*1 */
/*          On entry,  UPLO  specifies  whether the matrix is an upper or */
/*          lower triangular matrix as follows: */

/*             UPLO = 'U' or 'u'   A is an upper triangular matrix. */

/*             UPLO = 'L' or 'l'   A is a lower triangular matrix. */

/*  TRANS   (input) CHARACTER*1 */
/*          On entry,  TRANS  specifies  the operation to be performed as */
/*          follows: */

/*             TRANS = 'N' or 'n': */
/*                y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ) */

/*             TRANS = 'T' or 't': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*             TRANS = 'C' or 'c': */
/*                y := abs( alpha )*abs( conjg( A' ) )*abs( x ) + */
/*                     abs( beta*y ) */

/*  DIAG    (input) CHARACTER*1 */
/*          On entry, DIAG  specifies whether or not A is unit triangular */
/*          as follows: */

/*             DIAG = 'U' or 'u'  A is assumed to be unit triangular. */

/*             DIAG = 'N' or 'n'  A is not assumed to be unit triangular. */

/*  N       (input) INTEGER */
/*          On entry, N specifies the order of the matrix A. N must be at */
/*          least zero. */

/*  ALPHA   (input) REAL */
/*          On entry, ALPHA specifies the hscalreal scalar alpha. */

/*  A       (input) COMPLEX array */
/*          On entry,  A  is an array of dimension (LDA,N).  Before entry */
/*          with UPLO = 'U' or 'u', the leading n by n part of the  array */
/*          A must contain the upper triangular part of the matrix A  and */
/*          the  strictly  lower  triangular part of A is not referenced. */
/*          When  UPLO = 'L' or 'l', the leading n by n part of the array */
/*          A  must contain the lower triangular part of the matrix A and */
/*          the  strictly  upper trapezoidal part of A is not referenced. */
/*          Note that when  DIAG = 'U' or 'u', the diagonal elements of A */
/*          are not referenced either, but are assumed to be unity. */

/*  LDA     (input) INTEGER */
/*          On entry, LDA specifies the leading dimension of the array A. */
/*          LDA must be at least max( 1, N ). */

/*  X       (input) COMPLEX array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCX ) ).  Before entry, the incremented */
/*          array X must contain the vector x. */

/*  INCX    (input) INTEGER */
/*          On entry, INCX specifies the increment for the elements of X. */
/*          INCX must not be zero. */

/*  BETA    (input) REAL */
/*          On entry,  BETA  specifies the hscalreal scalar beta. When BETA is */
/*          supplied as zero then Y need not be set on input. */

/*  Y       (input/output) REAL array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCY ) ).  Before  entry with  BETA non- */
/*          zero, the  incremented array  Y must contain the vector y. On */
/*          exit, the  incremented array  Y is overwritten by the updated */
/*          vector y. */

/*  INCY    (input) INTEGER */
/*          On entry, INCY specifies the increment for the elements of Y. */
/*          INCY must not be zero. */

/*  -- Written on April 1, 1998 by */
/*     Antoine Petitet, University  of  Tennessee, Knoxville 37996, USA. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Statement Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Test the input parameters. */

    /* Parameter adjustments */
    a_dim1 = *lda;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --x;
    --y;

    /* Function Body */
    info = 0;
    if (! hscal_lsame(uplo, "U") && ! hscal_lsame(uplo, "L")) {
	info = 1;
    } else if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")) {
	info = 2;
    } else if (! hscal_lsame(diag, "U") && ! hscal_lsame(diag, "N")) {
	info = 3;
    } else if (*n < 0) {
	info = 4;
    } else if (*lda < max(1,*n)) {
	info = 7;
    } else if (*incx == 0) {
	info = 9;
    } else if (*incy == 0) {
	info = 12;
    }
    if (info != 0) {
	hscal_xerbla("CATRMV", &info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible. */

    if (*n == 0 || *alpha == 0.f && *beta == 1.f) {
	return HSCAL_SUCCESS;
    }

    nounit = hscal_lsame(diag, "N");

/*     Set up the start points in  X  and  Y. */

    if (*incx > 0) {
	kx = 1;
    } else {
	kx = 1 - (*n - 1) * *incx;
    }
    if (*incy > 0) {
	ky = 1;
    } else {
	ky = 1 - (*n - 1) * *incy;
    }

/*     Start the operations. In this version the elements of A are */
/*     accessed sequentially with one pass through A. */

/*     First form  y := abs( beta*y ). */

    if (*incy == 1) {
	if (*beta == 0.f) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = 0.f;
/* L10: */
	    }
	} else if (*beta == 1.f) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (r__1 = y[i__], dabs(r__1));
/* L20: */
	    }
	} else {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (r__1 = *beta * y[i__], dabs(r__1));
/* L30: */
	    }
	}
    } else {
	iy = ky;
	if (*beta == 0.f) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = 0.f;
		iy += *incy;
/* L40: */
	    }
	} else if (*beta == 1.f) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (r__1 = y[iy], dabs(r__1));
		iy += *incy;
/* L50: */
	    }
	} else {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (r__1 = *beta * y[iy], dabs(r__1));
		iy += *incy;
/* L60: */
	    }
	}
    }

    if (*alpha == 0.f) {
	return HSCAL_SUCCESS;
    }

    talpha = dabs(*alpha);

    if (hscal_lsame(trans, "N")) {

/*        Form  y := abs( alpha ) * abs( A ) * abs( x ) + y. */

	if (hscal_lsame(uplo, "U")) {
	    jx = kx;
	    if (*incy == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = jx;
		    absx = (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = hscal_r_imag(&x[
			    jx]), dabs(r__2));
		    if (absx != 0.f) {
			temp = talpha * absx;
			i__2 = j - 1;
			for (i__ = 1; i__ <= i__2; ++i__) {
			    i__3 = i__ + j * a_dim1;
			    y[i__] += temp * ((r__1 = a[i__3].r, dabs(r__1)) 
				    + (r__2 = hscal_r_imag(&a[i__ + j * a_dim1]), 
				    dabs(r__2)));
                            
/* L70: */
			}

			if (nounit) {
			    i__2 = j + j * a_dim1;
			    y[j] += temp * ((r__1 = a[i__2].r, dabs(r__1)) + (
				    r__2 = hscal_r_imag(&a[j + j * a_dim1]), dabs(
				    r__2)));
			} else {
			    y[j] += temp;
			}
		    }
		    jx += *incx;
                    
/* L80: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = jx;
		    absx = (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = hscal_r_imag(&x[
			    jx]), dabs(r__2));
		    if (absx != 0.f) {
			temp = talpha * absx;
			iy = ky;
			i__2 = j - 1;
			for (i__ = 1; i__ <= i__2; ++i__) {
			    i__3 = i__ + j * a_dim1;
			    y[iy] += temp * ((r__1 = a[i__3].r, dabs(r__1)) + 
				    (r__2 = hscal_r_imag(&a[i__ + j * a_dim1]), 
				    dabs(r__2)));
			    iy += *incy;
                            
/* L90: */
			}

			if (nounit) {
			    i__2 = j + j * a_dim1;
			    y[iy] += temp * ((r__1 = a[i__2].r, dabs(r__1)) + 
				    (r__2 = hscal_r_imag(&a[j + j * a_dim1]), dabs(
				    r__2)));
			} else {
			    y[iy] += temp;
			}
		    }
		    jx += *incx;
                    
/* L100: */
		}

	    }

	} else {

	    jx = kx;
	    if (*incy == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = jx;
		    absx = (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = hscal_r_imag(&x[
			    jx]), dabs(r__2));
		    if (absx != 0.f) {

			temp = talpha * absx;

			if (nounit) {
			    i__2 = j + j * a_dim1;
			    y[j] += temp * ((r__1 = a[i__2].r, dabs(r__1)) + (
				    r__2 = hscal_r_imag(&a[j + j * a_dim1]), dabs(
				    r__2)));
			} else {
			    y[j] += temp;
			}

			i__2 = *n;
			for (i__ = j + 1; i__ <= i__2; ++i__) {
			    i__3 = i__ + j * a_dim1;
			    y[i__] += temp * ((r__1 = a[i__3].r, dabs(r__1)) 
				    + (r__2 = hscal_r_imag(&a[i__ + j * a_dim1]), 
				    dabs(r__2)));
                            
/* L110: */
			}
		    }
		    jx += *incx;
                    
/* L120: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = jx;
		    absx = (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = hscal_r_imag(&x[
			    jx]), dabs(r__2));
		    if (absx != 0.f) {
			temp = talpha * absx;
			iy = ky + (j - 1) * *incy;

			if (nounit) {
			    i__2 = j + j * a_dim1;
			    y[iy] += temp * ((r__1 = a[i__2].r, dabs(r__1)) + 
				    (r__2 = hscal_r_imag(&a[j + j * a_dim1]), dabs(
				    r__2)));
			} else {
			    y[iy] += temp;
			}

			i__2 = *n;
			for (i__ = j + 1; i__ <= i__2; ++i__) {
			    iy += *incy;
			    i__3 = i__ + j * a_dim1;
			    y[iy] += temp * ((r__1 = a[i__3].r, dabs(r__1)) + 
				    (r__2 = hscal_r_imag(&a[i__ + j * a_dim1]), 
				    dabs(r__2)));
                            
/* L130: */
			}
		    }
		    jx += *incx;
                    
/* L140: */
		}

	    }

	}

    } else {

/*        Form  y := abs( alpha ) * abs( A' ) * abs( x ) + y. */

	if (hscal_lsame(uplo, "U")) {
	    jy = ky;
	    if (*incx == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    temp = 0.f;

		    i__2 = j - 1;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			i__4 = i__;
			temp += ((r__1 = a[i__3].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&a[i__ + j * a_dim1]), dabs(r__2))) * (
				(r__3 = x[i__4].r, dabs(r__3)) + (r__4 = 
				hscal_r_imag(&x[i__]), dabs(r__4)));
                        
/* L150: */
		    }

		    if (nounit) {
			i__2 = j + j * a_dim1;
			i__3 = j;
			temp += ((r__1 = a[i__2].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&a[j + j * a_dim1]), dabs(r__2))) * ((
				r__3 = x[i__3].r, dabs(r__3)) + (r__4 = 
				hscal_r_imag(&x[j]), dabs(r__4)));
		    } else {
			i__2 = j;
			temp += (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&x[j]), dabs(r__2));
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L160: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    temp = 0.f;
		    ix = kx;
		    i__2 = j - 1;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			i__4 = ix;
			temp += ((r__1 = a[i__3].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&a[i__ + j * a_dim1]), dabs(r__2))) * (
				(r__3 = x[i__4].r, dabs(r__3)) + (r__4 = 
				hscal_r_imag(&x[ix]), dabs(r__4)));
			ix += *incx;
                        
/* L170: */
		    }

		    if (nounit) {
			i__2 = j + j * a_dim1;
			i__3 = ix;
			temp += ((r__1 = a[i__2].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&a[j + j * a_dim1]), dabs(r__2))) * ((
				r__3 = x[i__3].r, dabs(r__3)) + (r__4 = 
				hscal_r_imag(&x[ix]), dabs(r__4)));
		    } else {
			i__2 = ix;
			temp += (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&x[ix]), dabs(r__2));
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L180: */
		}

	    }

	} else {

	    jy = ky;
	    if (*incx == 1) {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    if (nounit) {
			i__2 = j + j * a_dim1;
			i__3 = j;
			temp = ((r__1 = a[i__2].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&a[j + j * a_dim1]), dabs(r__2))) * ((
				r__3 = x[i__3].r, dabs(r__3)) + (r__4 = 
				hscal_r_imag(&x[j]), dabs(r__4)));
		    } else {
			i__2 = j;
			temp = (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&x[j]), dabs(r__2));
		    }

		    i__2 = *n;
		    for (i__ = j + 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			i__4 = i__;
			temp += ((r__1 = a[i__3].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&a[i__ + j * a_dim1]), dabs(r__2))) * (
				(r__3 = x[i__4].r, dabs(r__3)) + (r__4 = 
				hscal_r_imag(&x[i__]), dabs(r__4)));
                        
/* L190: */
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L200: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    ix = kx + (j - 1) * *incx;

		    if (nounit) {
			i__2 = j + j * a_dim1;
			i__3 = ix;
			temp = ((r__1 = a[i__2].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&a[j + j * a_dim1]), dabs(r__2))) * ((
				r__3 = x[i__3].r, dabs(r__3)) + (r__4 = 
				hscal_r_imag(&x[ix]), dabs(r__4)));
		    } else {
			i__2 = ix;
			temp = (r__1 = x[i__2].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&x[ix]), dabs(r__2));
		    }

		    i__2 = *n;
		    for (i__ = j + 1; i__ <= i__2; ++i__) {
			ix += *incx;
			i__3 = i__ + j * a_dim1;
			i__4 = ix;
			temp += ((r__1 = a[i__3].r, dabs(r__1)) + (r__2 = 
				hscal_r_imag(&a[i__ + j * a_dim1]), dabs(r__2))) * (
				(r__3 = x[i__4].r, dabs(r__3)) + (r__4 = 
				hscal_r_imag(&x[ix]), dabs(r__4)));
                        
/* L210: */
		    }
		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L220: */
		}
	    }
	}

    }

    return HSCAL_SUCCESS;

/*     End of CATRMV */

} /* catrmv_ */

/*------------------------------------------------------------------*/

int hscal_datrmv_benchmark(
    char *uplo,
    char *trans,
    char *diag,
    int     *n,
    double * alpha,
    double *a,
    int     *lda,
    double *x,
    int     *incx,
    double *beta,
    double *y,
    int     *incy
)
{
    return hscal_datrmv_benchmark_i(
           uplo,
           trans,
           diag,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           x,
           (hscalinteger*)incx,
           beta,
           y,
           (hscalinteger*)incy
    );
}

/*------------------------------------------------------------------*/

int hscal_datrmv_benchmark_i(
    char *uplo,
    char *trans,
    char *diag,
    hscalinteger *n,
    double * alpha,
    double *a,
    hscalinteger *lda,
    double *x,
    hscalinteger *incx,
    double *beta,
    double *y,
    hscalinteger *incy
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2;
    hscaldoublereal d__1;

    /* Local variables */
    static hscalinteger i__, j, ix, iy, jx, jy, kx, ky, info;
    static hscaldoublereal absx, temp;
    static hscaldoublereal talpha;
    static hscallogical nounit;

/*  -- PBLAS auxiliary routine (version 2.0) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     April 1, 1998 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  DATRMV performs one of the matrix-vector operations */

/*     y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ), */

/*  where  alpha  and  beta  are hscalreal scalars, y is a real vector, x is a */
/*  vector and A is an n by n unit or non-unit, upper or lower triangular */
/*  matrix. */

/*  Arguments */
/*  ========= */

/*  UPLO    (input) CHARACTER*1 */
/*          On entry,  UPLO  specifies  whether the matrix is an upper or */
/*          lower triangular matrix as follows: */

/*             UPLO = 'U' or 'u'   A is an upper triangular matrix. */

/*             UPLO = 'L' or 'l'   A is a lower triangular matrix. */

/*  TRANS   (input) CHARACTER*1 */
/*          On entry,  TRANS  specifies  the operation to be performed as */
/*          follows: */

/*             TRANS = 'N' or 'n': */
/*                y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ) */

/*             TRANS = 'T' or 't': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*             TRANS = 'C' or 'c': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*  DIAG    (input) CHARACTER*1 */
/*          On entry, DIAG  specifies whether or not A is unit triangular */
/*          as follows: */

/*             DIAG = 'U' or 'u'  A is assumed to be unit triangular. */

/*             DIAG = 'N' or 'n'  A is not assumed to be unit triangular. */

/*  N       (input) INTEGER */
/*          On entry, N specifies the order of the matrix A. N must be at */
/*          least zero. */

/*  ALPHA   (input) DOUBLE PRECISION */
/*          On entry, ALPHA specifies the hscalreal scalar alpha. */

/*  A       (input) DOUBLE PRECISION array */
/*          On entry,  A  is an array of dimension (LDA,N).  Before entry */
/*          with UPLO = 'U' or 'u', the leading n by n part of the  array */
/*          A must contain the upper triangular part of the matrix A  and */
/*          the  strictly  lower  triangular part of A is not referenced. */
/*          When  UPLO = 'L' or 'l', the leading n by n part of the array */
/*          A  must contain the lower triangular part of the matrix A and */
/*          the  strictly  upper trapezoidal part of A is not referenced. */
/*          Note that when  DIAG = 'U' or 'u', the diagonal elements of A */
/*          are not referenced either, but are assumed to be unity. */

/*  LDA     (input) INTEGER */
/*          On entry, LDA specifies the leading dimension of the array A. */
/*          LDA must be at least max( 1, N ). */

/*  X       (input) DOUBLE PRECISION array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCX ) ).  Before entry, the incremented */
/*          array X must contain the vector x. */

/*  INCX    (input) INTEGER */
/*          On entry, INCX specifies the increment for the elements of X. */
/*          INCX must not be zero. */

/*  BETA    (input) DOUBLE PRECISION */
/*          On entry,  BETA  specifies the hscalreal scalar beta. When BETA is */
/*          supplied as zero then Y need not be set on input. */

/*  Y       (input/output) DOUBLE PRECISION array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCY ) ).  Before  entry with  BETA non- */
/*          zero, the  incremented array  Y must contain the vector y. On */
/*          exit, the  incremented array  Y is overwritten by the updated */
/*          vector y. */

/*  INCY    (input) INTEGER */
/*          On entry, INCY specifies the increment for the elements of Y. */
/*          INCY must not be zero. */

/*  -- Written on April 1, 1998 by */
/*     Antoine Petitet, University  of  Tennessee, Knoxville 37996, USA. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Test the input parameters. */

    /* Parameter adjustments */
    a_dim1 = *lda;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --x;
    --y;

    /* Function Body */
    info = 0;
    if (! hscal_lsame(uplo, "U") && ! hscal_lsame(uplo, "L")) {
	info = 1;
    } else if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")) {
	info = 2;
    } else if (! hscal_lsame(diag, "U") && ! hscal_lsame(diag, "N")) {
	info = 3;
    } else if (*n < 0) {
	info = 4;
    } else if (*lda < max(1,*n)) {
	info = 7;
    } else if (*incx == 0) {
	info = 9;
    } else if (*incy == 0) {
	info = 12;
    }
    if (info != 0) {
	hscal_xerbla("DATRMV", &info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible. */

    if (*n == 0 || *alpha == 0. && *beta == 1.) {
	return HSCAL_SUCCESS;
    }

    nounit = hscal_lsame(diag, "N");

/*     Set up the start points in  X  and  Y. */

    if (*incx > 0) {
	kx = 1;
    } else {
	kx = 1 - (*n - 1) * *incx;
    }
    if (*incy > 0) {
	ky = 1;
    } else {
	ky = 1 - (*n - 1) * *incy;
    }

/*     Start the operations. In this version the elements of A are */
/*     accessed sequentially with one pass through A. */

/*     First form  y := abs( beta*y ). */

    if (*incy == 1) {
	if (*beta == 0.) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = 0.;
/* L10: */
	    }
	} else if (*beta == 1.) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (d__1 = y[i__], abs(d__1));
/* L20: */
	    }
	} else {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (d__1 = *beta * y[i__], abs(d__1));
/* L30: */
	    }
	}
    } else {
	iy = ky;
	if (*beta == 0.) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = 0.;
		iy += *incy;
/* L40: */
	    }
	} else if (*beta == 1.) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (d__1 = y[iy], abs(d__1));
		iy += *incy;
/* L50: */
	    }
	} else {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (d__1 = *beta * y[iy], abs(d__1));
		iy += *incy;
/* L60: */
	    }
	}
    }

    if (*alpha == 0.) {
	return HSCAL_SUCCESS;
    }

    talpha = abs(*alpha);

    if (hscal_lsame(trans, "N")) {

/*        Form  y := abs( alpha ) * abs( A ) * abs( x ) + y. */

	if (hscal_lsame(uplo, "U")) {
	    jx = kx;
	    if (*incy == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    absx = (d__1 = x[jx], abs(d__1));
		    if (absx != 0.) {
			temp = talpha * absx;
			i__2 = j - 1;
			for (i__ = 1; i__ <= i__2; ++i__) {
			    y[i__] += temp * (d__1 = a[i__ + j * a_dim1], abs(
				    d__1));
                            
/* L70: */
			}

			if (nounit) {
			    y[j] += temp * (d__1 = a[j + j * a_dim1], abs(
				    d__1));
			} else {
			    y[j] += temp;
			}
		    }
		    jx += *incx;
                    
/* L80: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    absx = (d__1 = x[jx], abs(d__1));
		    if (absx != 0.) {
			temp = talpha * absx;
			iy = ky;
			i__2 = j - 1;
			for (i__ = 1; i__ <= i__2; ++i__) {
			    y[iy] += temp * (d__1 = a[i__ + j * a_dim1], abs(
				    d__1));
			    iy += *incy;
                            
/* L90: */
			}

			if (nounit) {
			    y[iy] += temp * (d__1 = a[j + j * a_dim1], abs(
				    d__1));
			} else {
			    y[iy] += temp;
			}
		    }
		    jx += *incx;
                    
/* L100: */
		}

	    }

	} else {

	    jx = kx;
	    if (*incy == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    absx = (d__1 = x[jx], abs(d__1));
		    if (absx != 0.) {

			temp = talpha * absx;

			if (nounit) {
			    y[j] += temp * (d__1 = a[j + j * a_dim1], abs(
				    d__1));
			} else {
			    y[j] += temp;
			}

			i__2 = *n;
			for (i__ = j + 1; i__ <= i__2; ++i__) {
			    y[i__] += temp * (d__1 = a[i__ + j * a_dim1], abs(
				    d__1));
                            
/* L110: */
			}
		    }
		    jx += *incx;
                    
/* L120: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    absx = (d__1 = x[jx], abs(d__1));
		    if (absx != 0.) {
			temp = talpha * absx;
			iy = ky + (j - 1) * *incy;

			if (nounit) {
			    y[iy] += temp * (d__1 = a[j + j * a_dim1], abs(
				    d__1));
			} else {
			    y[iy] += temp;
			}

			i__2 = *n;
			for (i__ = j + 1; i__ <= i__2; ++i__) {
			    iy += *incy;
			    y[iy] += temp * (d__1 = a[i__ + j * a_dim1], abs(
				    d__1));
                            
/* L130: */
			}
		    }
		    jx += *incx;
                    
/* L140: */
		}

	    }

	}

    } else {

/*        Form  y := abs( alpha ) * abs( A' ) * abs( x ) + y. */

	if (hscal_lsame(uplo, "U")) {
	    jy = ky;
	    if (*incx == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    temp = 0.;

		    i__2 = j - 1;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			temp += (d__1 = a[i__ + j * a_dim1] * x[i__], abs(
				d__1));
                        
/* L150: */
		    }

		    if (nounit) {
			temp += (d__1 = a[j + j * a_dim1] * x[j], abs(d__1));
		    } else {
			temp += (d__1 = x[j], abs(d__1));
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L160: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    temp = 0.;
		    ix = kx;
		    i__2 = j - 1;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			temp += (d__1 = a[i__ + j * a_dim1] * x[ix], abs(d__1)
				);
			ix += *incx;
                        
/* L170: */
		    }

		    if (nounit) {
			temp += (d__1 = a[j + j * a_dim1] * x[ix], abs(d__1));
		    } else {
			temp += (d__1 = x[ix], abs(d__1));
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L180: */
		}

	    }

	} else {

	    jy = ky;
	    if (*incx == 1) {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    if (nounit) {
			temp = (d__1 = a[j + j * a_dim1] * x[j], abs(d__1));
		    } else {
			temp = (d__1 = x[j], abs(d__1));
		    }

		    i__2 = *n;
		    for (i__ = j + 1; i__ <= i__2; ++i__) {
			temp += (d__1 = a[i__ + j * a_dim1] * x[i__], abs(
				d__1));
                        
/* L190: */
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L200: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    ix = kx + (j - 1) * *incx;

		    if (nounit) {
			temp = (d__1 = a[j + j * a_dim1] * x[ix], abs(d__1));
		    } else {
			temp = (d__1 = x[ix], abs(d__1));
		    }

		    i__2 = *n;
		    for (i__ = j + 1; i__ <= i__2; ++i__) {
			ix += *incx;
			temp += (d__1 = a[i__ + j * a_dim1] * x[ix], abs(d__1)
				);
                        
/* L210: */
		    }
		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L220: */
		}
	    }
	}

    }

    return HSCAL_SUCCESS;

/*     End of DATRMV */

} /* datrmv_ */

/*------------------------------------------------------------------*/

int hscal_satrmv_benchmark(
    char *uplo,
    char *trans,
    char *diag,
    int  *n,
    hscalreal * alpha,
    hscalreal *a,
    int  *lda,
    hscalreal *x,
    int  *incx,
    hscalreal *beta,
    hscalreal *y,
    int  *incy
)
{
    return hscal_satrmv_benchmark_i(
           uplo,
           trans,
           diag,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           x,
           (hscalinteger*)incx,
           beta,
           y,
           (hscalinteger*)incy
    );
}

/*------------------------------------------------------------------*/

int hscal_satrmv_benchmark_i(
    char *uplo,
    char *trans,
    char *diag,
    hscalinteger *n,
    hscalreal * alpha,
    hscalreal *a,
    hscalinteger *lda,
    hscalreal *x,
    hscalinteger *incx,
    hscalreal *beta,
    hscalreal *y,
    hscalinteger *incy
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2;
    hscalreal r__1;

    /* Local variables */
    static hscalinteger i__, j, ix, iy, jx, jy, kx, ky, info;
    static hscalreal absx, temp;
    static hscalreal talpha;
    static hscallogical nounit;

/*  -- PBLAS auxiliary routine (version 2.0) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     April 1, 1998 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  SATRMV performs one of the matrix-vector operations */

/*     y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ), */

/*  where  alpha  and  beta  are hscalreal scalars, y is a real vector, x is a */
/*  vector and A is an n by n unit or non-unit, upper or lower triangular */
/*  matrix. */

/*  Arguments */
/*  ========= */

/*  UPLO    (input) CHARACTER*1 */
/*          On entry,  UPLO  specifies  whether the matrix is an upper or */
/*          lower triangular matrix as follows: */

/*             UPLO = 'U' or 'u'   A is an upper triangular matrix. */

/*             UPLO = 'L' or 'l'   A is a lower triangular matrix. */

/*  TRANS   (input) CHARACTER*1 */
/*          On entry,  TRANS  specifies  the operation to be performed as */
/*          follows: */

/*             TRANS = 'N' or 'n': */
/*                y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ) */

/*             TRANS = 'T' or 't': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*             TRANS = 'C' or 'c': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*  DIAG    (input) CHARACTER*1 */
/*          On entry, DIAG  specifies whether or not A is unit triangular */
/*          as follows: */

/*             DIAG = 'U' or 'u'  A is assumed to be unit triangular. */

/*             DIAG = 'N' or 'n'  A is not assumed to be unit triangular. */

/*  N       (input) INTEGER */
/*          On entry, N specifies the order of the matrix A. N must be at */
/*          least zero. */

/*  ALPHA   (input) REAL */
/*          On entry, ALPHA specifies the hscalreal scalar alpha. */

/*  A       (input) REAL array */
/*          On entry,  A  is an array of dimension (LDA,N).  Before entry */
/*          with UPLO = 'U' or 'u', the leading n by n part of the  array */
/*          A must contain the upper triangular part of the matrix A  and */
/*          the  strictly  lower  triangular part of A is not referenced. */
/*          When  UPLO = 'L' or 'l', the leading n by n part of the array */
/*          A  must contain the lower triangular part of the matrix A and */
/*          the  strictly  upper trapezoidal part of A is not referenced. */
/*          Note that when  DIAG = 'U' or 'u', the diagonal elements of A */
/*          are not referenced either, but are assumed to be unity. */

/*  LDA     (input) INTEGER */
/*          On entry, LDA specifies the leading dimension of the array A. */
/*          LDA must be at least max( 1, N ). */

/*  X       (input) REAL array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCX ) ).  Before entry, the incremented */
/*          array X must contain the vector x. */

/*  INCX    (input) INTEGER */
/*          On entry, INCX specifies the increment for the elements of X. */
/*          INCX must not be zero. */

/*  BETA    (input) REAL */
/*          On entry,  BETA  specifies the hscalreal scalar beta. When BETA is */
/*          supplied as zero then Y need not be set on input. */

/*  Y       (input/output) REAL array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCY ) ).  Before  entry with  BETA non- */
/*          zero, the  incremented array  Y must contain the vector y. On */
/*          exit, the  incremented array  Y is overwritten by the updated */
/*          vector y. */

/*  INCY    (input) INTEGER */
/*          On entry, INCY specifies the increment for the elements of Y. */
/*          INCY must not be zero. */

/*  -- Written on April 1, 1998 by */
/*     Antoine Petitet, University  of  Tennessee, Knoxville 37996, USA. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Test the input parameters. */

    /* Parameter adjustments */
    a_dim1 = *lda;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --x;
    --y;

    /* Function Body */
    info = 0;
    if (! hscal_lsame(uplo, "U") && ! hscal_lsame(uplo, "L")) {
	info = 1;
    } else if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")) {
	info = 2;
    } else if (! hscal_lsame(diag, "U") && ! hscal_lsame(diag, "N")) {
	info = 3;
    } else if (*n < 0) {
	info = 4;
    } else if (*lda < max(1,*n)) {
	info = 7;
    } else if (*incx == 0) {
	info = 9;
    } else if (*incy == 0) {
	info = 12;
    }
    if (info != 0) {
	hscal_xerbla("SATRMV", &info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible. */

    if (*n == 0 || *alpha == 0.f && *beta == 1.f) {
	return HSCAL_SUCCESS;
    }

    nounit = hscal_lsame(diag, "N");

/*     Set up the start points in  X  and  Y. */

    if (*incx > 0) {
	kx = 1;
    } else {
	kx = 1 - (*n - 1) * *incx;
    }
    if (*incy > 0) {
	ky = 1;
    } else {
	ky = 1 - (*n - 1) * *incy;
    }

/*     Start the operations. In this version the elements of A are */
/*     accessed sequentially with one pass through A. */

/*     First form  y := abs( beta*y ). */

    if (*incy == 1) {
	if (*beta == 0.f) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = 0.f;
/* L10: */
	    }
	} else if (*beta == 1.f) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (r__1 = y[i__], dabs(r__1));
/* L20: */
	    }
	} else {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (r__1 = *beta * y[i__], dabs(r__1));
/* L30: */
	    }
	}
    } else {
	iy = ky;
	if (*beta == 0.f) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = 0.f;
		iy += *incy;
/* L40: */
	    }
	} else if (*beta == 1.f) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (r__1 = y[iy], dabs(r__1));
		iy += *incy;
/* L50: */
	    }
	} else {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (r__1 = *beta * y[iy], dabs(r__1));
		iy += *incy;
/* L60: */
	    }
	}
    }

    if (*alpha == 0.f) {
	return HSCAL_SUCCESS;
    }

    talpha = dabs(*alpha);

    if (hscal_lsame(trans, "N")) {

/*        Form  y := abs( alpha ) * abs( A ) * abs( x ) + y. */

	if (hscal_lsame(uplo, "U")) {
	    jx = kx;
	    if (*incy == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    absx = (r__1 = x[jx], dabs(r__1));
		    if (absx != 0.f) {
			temp = talpha * absx;
			i__2 = j - 1;
			for (i__ = 1; i__ <= i__2; ++i__) {
			    y[i__] += temp * (r__1 = a[i__ + j * a_dim1], 
				    dabs(r__1));
                            
/* L70: */
			}

			if (nounit) {
			    y[j] += temp * (r__1 = a[j + j * a_dim1], dabs(
				    r__1));
			} else {
			    y[j] += temp;
			}
		    }
		    jx += *incx;
                    
/* L80: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    absx = (r__1 = x[jx], dabs(r__1));
		    if (absx != 0.f) {
			temp = talpha * absx;
			iy = ky;
			i__2 = j - 1;
			for (i__ = 1; i__ <= i__2; ++i__) {
			    y[iy] += temp * (r__1 = a[i__ + j * a_dim1], dabs(
				    r__1));
			    iy += *incy;
                            
/* L90: */
			}

			if (nounit) {
			    y[iy] += temp * (r__1 = a[j + j * a_dim1], dabs(
				    r__1));
			} else {
			    y[iy] += temp;
			}
		    }
		    jx += *incx;
                    
/* L100: */
		}

	    }

	} else {

	    jx = kx;
	    if (*incy == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    absx = (r__1 = x[jx], dabs(r__1));
		    if (absx != 0.f) {

			temp = talpha * absx;

			if (nounit) {
			    y[j] += temp * (r__1 = a[j + j * a_dim1], dabs(
				    r__1));
			} else {
			    y[j] += temp;
			}

			i__2 = *n;
			for (i__ = j + 1; i__ <= i__2; ++i__) {
			    y[i__] += temp * (r__1 = a[i__ + j * a_dim1], 
				    dabs(r__1));
                            
/* L110: */
			}
		    }
		    jx += *incx;
                    
/* L120: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    absx = (r__1 = x[jx], dabs(r__1));
		    if (absx != 0.f) {
			temp = talpha * absx;
			iy = ky + (j - 1) * *incy;

			if (nounit) {
			    y[iy] += temp * (r__1 = a[j + j * a_dim1], dabs(
				    r__1));
			} else {
			    y[iy] += temp;
			}

			i__2 = *n;
			for (i__ = j + 1; i__ <= i__2; ++i__) {
			    iy += *incy;
			    y[iy] += temp * (r__1 = a[i__ + j * a_dim1], dabs(
				    r__1));
                            
/* L130: */
			}
		    }
		    jx += *incx;
                    
/* L140: */
		}

	    }

	}

    } else {

/*        Form  y := abs( alpha ) * abs( A' ) * abs( x ) + y. */

	if (hscal_lsame(uplo, "U")) {
	    jy = ky;
	    if (*incx == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    temp = 0.f;

		    i__2 = j - 1;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			temp += (r__1 = a[i__ + j * a_dim1] * x[i__], dabs(
				r__1));
                        
/* L150: */
		    }

		    if (nounit) {
			temp += (r__1 = a[j + j * a_dim1] * x[j], dabs(r__1));
		    } else {
			temp += (r__1 = x[j], dabs(r__1));
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L160: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    temp = 0.f;
		    ix = kx;
		    i__2 = j - 1;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			temp += (r__1 = a[i__ + j * a_dim1] * x[ix], dabs(
				r__1));
			ix += *incx;
                        
/* L170: */
		    }

		    if (nounit) {
			temp += (r__1 = a[j + j * a_dim1] * x[ix], dabs(r__1))
				;
		    } else {
			temp += (r__1 = x[ix], dabs(r__1));
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L180: */
		}

	    }

	} else {

	    jy = ky;
	    if (*incx == 1) {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    if (nounit) {
			temp = (r__1 = a[j + j * a_dim1] * x[j], dabs(r__1));
		    } else {
			temp = (r__1 = x[j], dabs(r__1));
		    }

		    i__2 = *n;
		    for (i__ = j + 1; i__ <= i__2; ++i__) {
			temp += (r__1 = a[i__ + j * a_dim1] * x[i__], dabs(
				r__1));
                        
/* L190: */
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L200: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    ix = kx + (j - 1) * *incx;

		    if (nounit) {
			temp = (r__1 = a[j + j * a_dim1] * x[ix], dabs(r__1));
		    } else {
			temp = (r__1 = x[ix], dabs(r__1));
		    }

		    i__2 = *n;
		    for (i__ = j + 1; i__ <= i__2; ++i__) {
			ix += *incx;
			temp += (r__1 = a[i__ + j * a_dim1] * x[ix], dabs(
				r__1));
                        
/* L210: */
		    }
		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L220: */
		}
	    }
	}

    }

    return HSCAL_SUCCESS;

/*     End of SATRMV */

} /* satrmv_ */

/*------------------------------------------------------------------*/

int hscal_zatrmv_benchmark(
    char *uplo,
    char *trans,
    char *diag,
    int     *n,
    double * alpha,
    hscaldoublecomplex *a,
    int     *lda,
    hscaldoublecomplex *x,
    int     *incx,
    double *beta,
    double *y,
    int     *incy
)
{
    return hscal_zatrmv_benchmark_i(
           uplo,
           trans,
           diag,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           x,
           (hscalinteger*)incx,
           beta,
           y,
           (hscalinteger*)incy
    );
}

/*------------------------------------------------------------------*/

int hscal_zatrmv_benchmark_i(
    char *uplo,
    char *trans,
    char *diag,
    hscalinteger *n,
    double * alpha,
    hscaldoublecomplex *a,
    hscalinteger *lda,
    hscaldoublecomplex *x,
    hscalinteger *incx,
    double *beta,
    double *y,
    hscalinteger *incy
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2, i__3, i__4;
    hscaldoublereal d__1, d__2, d__3, d__4;

    /* Local variables */
    static hscalinteger i__, j, ix, iy, jx, jy, kx, ky, info;
    static hscaldoublereal absx, temp;
    static hscaldoublereal talpha;
    static hscallogical nounit;

/*  -- PBLAS auxiliary routine (version 2.0) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     April 1, 1998 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  ZATRMV performs one of the matrix-vector operations */

/*     y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ), */

/*     or */

/*     y := abs( alpha )*abs( conjg( A' ) )*abs( x ) + abs( beta*y ), */

/*  where  alpha  and  beta  are hscalreal scalars, y is a real vector, x is a */
/*  vector and A is an n by n unit or non-unit, upper or lower triangular */
/*  matrix. */

/*  Arguments */
/*  ========= */

/*  UPLO    (input) CHARACTER*1 */
/*          On entry,  UPLO  specifies  whether the matrix is an upper or */
/*          lower triangular matrix as follows: */

/*             UPLO = 'U' or 'u'   A is an upper triangular matrix. */

/*             UPLO = 'L' or 'l'   A is a lower triangular matrix. */

/*  TRANS   (input) CHARACTER*1 */
/*          On entry,  TRANS  specifies  the operation to be performed as */
/*          follows: */

/*             TRANS = 'N' or 'n': */
/*                y := abs( alpha )*abs( A )*abs( x )+ abs( beta*y ) */

/*             TRANS = 'T' or 't': */
/*                y := abs( alpha )*abs( A' )*abs( x ) + abs( beta*y ) */

/*             TRANS = 'C' or 'c': */
/*                y := abs( alpha )*abs( conjg( A' ) )*abs( x ) + */
/*                     abs( beta*y ) */

/*  DIAG    (input) CHARACTER*1 */
/*          On entry, DIAG  specifies whether or not A is unit triangular */
/*          as follows: */

/*             DIAG = 'U' or 'u'  A is assumed to be unit triangular. */

/*             DIAG = 'N' or 'n'  A is not assumed to be unit triangular. */

/*  N       (input) INTEGER */
/*          On entry, N specifies the order of the matrix A. N must be at */
/*          least zero. */

/*  ALPHA   (input) DOUBLE PRECISION */
/*          On entry, ALPHA specifies the hscalreal scalar alpha. */

/*  A       (input) COMPLEX*16 array */
/*          On entry,  A  is an array of dimension (LDA,N).  Before entry */
/*          with UPLO = 'U' or 'u', the leading n by n part of the  array */
/*          A must contain the upper triangular part of the matrix A  and */
/*          the  strictly  lower  triangular part of A is not referenced. */
/*          When  UPLO = 'L' or 'l', the leading n by n part of the array */
/*          A  must contain the lower triangular part of the matrix A and */
/*          the  strictly  upper trapezoidal part of A is not referenced. */
/*          Note that when  DIAG = 'U' or 'u', the diagonal elements of A */
/*          are not referenced either, but are assumed to be unity. */

/*  LDA     (input) INTEGER */
/*          On entry, LDA specifies the leading dimension of the array A. */
/*          LDA must be at least max( 1, N ). */

/*  X       (input) COMPLEX*16 array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCX ) ).  Before entry, the incremented */
/*          array X must contain the vector x. */

/*  INCX    (input) INTEGER */
/*          On entry, INCX specifies the increment for the elements of X. */
/*          INCX must not be zero. */

/*  BETA    (input) DOUBLE PRECISION */
/*          On entry,  BETA  specifies the hscalreal scalar beta. When BETA is */
/*          supplied as zero then Y need not be set on input. */

/*  Y       (input/output) DOUBLE PRECISION array of dimension at least */
/*          ( 1 + ( n - 1 )*abs( INCY ) ).  Before  entry with  BETA non- */
/*          zero, the  incremented array  Y must contain the vector y. On */
/*          exit, the  incremented array  Y is overwritten by the updated */
/*          vector y. */

/*  INCY    (input) INTEGER */
/*          On entry, INCY specifies the increment for the elements of Y. */
/*          INCY must not be zero. */

/*  -- Written on April 1, 1998 by */
/*     Antoine Petitet, University  of  Tennessee, Knoxville 37996, USA. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Statement Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Test the input parameters. */

    /* Parameter adjustments */
    a_dim1 = *lda;
    a_offset = 1 + a_dim1;
    a -= a_offset;
    --x;
    --y;

    /* Function Body */
    info = 0;
    if (! hscal_lsame(uplo, "U") && ! hscal_lsame(uplo, "L")) {
	info = 1;
    } else if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, "T") && ! hscal_lsame(trans, "C")) {
	info = 2;
    } else if (! hscal_lsame(diag, "U") && ! hscal_lsame(diag, "N")) {
	info = 3;
    } else if (*n < 0) {
	info = 4;
    } else if (*lda < max(1,*n)) {
	info = 7;
    } else if (*incx == 0) {
	info = 9;
    } else if (*incy == 0) {
	info = 12;
    }
    if (info != 0) {
	hscal_xerbla("ZATRMV", &info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible. */

    if (*n == 0 || *alpha == 0. && *beta == 1.) {
	return HSCAL_SUCCESS;
    }

    nounit = hscal_lsame(diag, "N");

/*     Set up the start points in  X  and  Y. */

    if (*incx > 0) {
	kx = 1;
    } else {
	kx = 1 - (*n - 1) * *incx;
    }
    if (*incy > 0) {
	ky = 1;
    } else {
	ky = 1 - (*n - 1) * *incy;
    }

/*     Start the operations. In this version the elements of A are */
/*     accessed sequentially with one pass through A. */

/*     First form  y := abs( beta*y ). */

    if (*incy == 1) {
	if (*beta == 0.) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = 0.;
/* L10: */
	    }
	} else if (*beta == 1.) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (d__1 = y[i__], abs(d__1));
/* L20: */
	    }
	} else {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[i__] = (d__1 = *beta * y[i__], abs(d__1));
/* L30: */
	    }
	}
    } else {
	iy = ky;
	if (*beta == 0.) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = 0.;
		iy += *incy;
/* L40: */
	    }
	} else if (*beta == 1.) {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (d__1 = y[iy], abs(d__1));
		iy += *incy;
/* L50: */
	    }
	} else {
	    i__1 = *n;
	    for (i__ = 1; i__ <= i__1; ++i__) {
		y[iy] = (d__1 = *beta * y[iy], abs(d__1));
		iy += *incy;
/* L60: */
	    }
	}
    }

    if (*alpha == 0.) {
	return HSCAL_SUCCESS;
    }

    talpha = abs(*alpha);

    if (hscal_lsame(trans, "N")) {

/*        Form  y := abs( alpha ) * abs( A ) * abs( x ) + y. */

	if (hscal_lsame(uplo, "U")) {
	    jx = kx;
	    if (*incy == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = jx;
		    absx = (d__1 = x[i__2].r, abs(d__1)) + (d__2 = hscal_d_imag(&x[
			    jx]), abs(d__2));
		    if (absx != 0.) {
			temp = talpha * absx;
			i__2 = j - 1;
			for (i__ = 1; i__ <= i__2; ++i__) {
			    i__3 = i__ + j * a_dim1;
			    y[i__] += temp * ((d__1 = a[i__3].r, abs(d__1)) + 
				    (d__2 = hscal_d_imag(&a[i__ + j * a_dim1]), abs(
				    d__2)));
                            
/* L70: */
			}

			if (nounit) {
			    i__2 = j + j * a_dim1;
			    y[j] += temp * ((d__1 = a[i__2].r, abs(d__1)) + (
				    d__2 = hscal_d_imag(&a[j + j * a_dim1]), abs(
				    d__2)));
			} else {
			    y[j] += temp;
			}
		    }
		    jx += *incx;
                    
/* L80: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = jx;
		    absx = (d__1 = x[i__2].r, abs(d__1)) + (d__2 = hscal_d_imag(&x[
			    jx]), abs(d__2));
		    if (absx != 0.) {
			temp = talpha * absx;
			iy = ky;
			i__2 = j - 1;
			for (i__ = 1; i__ <= i__2; ++i__) {
			    i__3 = i__ + j * a_dim1;
			    y[iy] += temp * ((d__1 = a[i__3].r, abs(d__1)) + (
				    d__2 = hscal_d_imag(&a[i__ + j * a_dim1]), abs(
				    d__2)));
			    iy += *incy;
                            
/* L90: */
			}

			if (nounit) {
			    i__2 = j + j * a_dim1;
			    y[iy] += temp * ((d__1 = a[i__2].r, abs(d__1)) + (
				    d__2 = hscal_d_imag(&a[j + j * a_dim1]), abs(
				    d__2)));
			} else {
			    y[iy] += temp;
			}
		    }
		    jx += *incx;
                    
/* L100: */
		}

	    }

	} else {

	    jx = kx;
	    if (*incy == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = jx;
		    absx = (d__1 = x[i__2].r, abs(d__1)) + (d__2 = hscal_d_imag(&x[
			    jx]), abs(d__2));
		    if (absx != 0.) {

			temp = talpha * absx;

			if (nounit) {
			    i__2 = j + j * a_dim1;
			    y[j] += temp * ((d__1 = a[i__2].r, abs(d__1)) + (
				    d__2 = hscal_d_imag(&a[j + j * a_dim1]), abs(
				    d__2)));
			} else {
			    y[j] += temp;
			}

			i__2 = *n;
			for (i__ = j + 1; i__ <= i__2; ++i__) {
			    i__3 = i__ + j * a_dim1;
			    y[i__] += temp * ((d__1 = a[i__3].r, abs(d__1)) + 
				    (d__2 = hscal_d_imag(&a[i__ + j * a_dim1]), abs(
				    d__2)));
                            
/* L110: */
			}
		    }
		    jx += *incx;
                    
/* L120: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    i__2 = jx;
		    absx = (d__1 = x[i__2].r, abs(d__1)) + (d__2 = hscal_d_imag(&x[
			    jx]), abs(d__2));
		    if (absx != 0.) {
			temp = talpha * absx;
			iy = ky + (j - 1) * *incy;

			if (nounit) {
			    i__2 = j + j * a_dim1;
			    y[iy] += temp * ((d__1 = a[i__2].r, abs(d__1)) + (
				    d__2 = hscal_d_imag(&a[j + j * a_dim1]), abs(
				    d__2)));
			} else {
			    y[iy] += temp;
			}

			i__2 = *n;
			for (i__ = j + 1; i__ <= i__2; ++i__) {
			    iy += *incy;
			    i__3 = i__ + j * a_dim1;
			    y[iy] += temp * ((d__1 = a[i__3].r, abs(d__1)) + (
				    d__2 = hscal_d_imag(&a[i__ + j * a_dim1]), abs(
				    d__2)));
                            
/* L130: */
			}
		    }
		    jx += *incx;
                    
/* L140: */
		}

	    }

	}

    } else {

/*        Form  y := abs( alpha ) * abs( A' ) * abs( x ) + y. */

	if (hscal_lsame(uplo, "U")) {
	    jy = ky;
	    if (*incx == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    temp = 0.;

		    i__2 = j - 1;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			i__4 = i__;
			temp += ((d__1 = a[i__3].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&a[i__ + j * a_dim1]), abs(d__2))) * ((
				d__3 = x[i__4].r, abs(d__3)) + (d__4 = hscal_d_imag(
				&x[i__]), abs(d__4)));
                        
/* L150: */
		    }

		    if (nounit) {
			i__2 = j + j * a_dim1;
			i__3 = j;
			temp += ((d__1 = a[i__2].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&a[j + j * a_dim1]), abs(d__2))) * ((
				d__3 = x[i__3].r, abs(d__3)) + (d__4 = hscal_d_imag(
				&x[j]), abs(d__4)));
		    } else {
			i__2 = j;
			temp += (d__1 = x[i__2].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&x[j]), abs(d__2));
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L160: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    temp = 0.;
		    ix = kx;
		    i__2 = j - 1;
		    for (i__ = 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			i__4 = ix;
			temp += ((d__1 = a[i__3].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&a[i__ + j * a_dim1]), abs(d__2))) * ((
				d__3 = x[i__4].r, abs(d__3)) + (d__4 = hscal_d_imag(
				&x[ix]), abs(d__4)));
			ix += *incx;
                        
/* L170: */
		    }

		    if (nounit) {
			i__2 = j + j * a_dim1;
			i__3 = ix;
			temp += ((d__1 = a[i__2].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&a[j + j * a_dim1]), abs(d__2))) * ((
				d__3 = x[i__3].r, abs(d__3)) + (d__4 = hscal_d_imag(
				&x[ix]), abs(d__4)));
		    } else {
			i__2 = ix;
			temp += (d__1 = x[i__2].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&x[ix]), abs(d__2));
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L180: */
		}

	    }

	} else {

	    jy = ky;
	    if (*incx == 1) {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    if (nounit) {
			i__2 = j + j * a_dim1;
			i__3 = j;
			temp = ((d__1 = a[i__2].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&a[j + j * a_dim1]), abs(d__2))) * ((
				d__3 = x[i__3].r, abs(d__3)) + (d__4 = hscal_d_imag(
				&x[j]), abs(d__4)));
		    } else {
			i__2 = j;
			temp = (d__1 = x[i__2].r, abs(d__1)) + (d__2 = hscal_d_imag(
				&x[j]), abs(d__2));
		    }

		    i__2 = *n;
		    for (i__ = j + 1; i__ <= i__2; ++i__) {
			i__3 = i__ + j * a_dim1;
			i__4 = i__;
			temp += ((d__1 = a[i__3].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&a[i__ + j * a_dim1]), abs(d__2))) * ((
				d__3 = x[i__4].r, abs(d__3)) + (d__4 = hscal_d_imag(
				&x[i__]), abs(d__4)));
                        
/* L190: */
		    }

		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L200: */
		}

	    } else {

		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {

		    ix = kx + (j - 1) * *incx;

		    if (nounit) {
			i__2 = j + j * a_dim1;
			i__3 = ix;
			temp = ((d__1 = a[i__2].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&a[j + j * a_dim1]), abs(d__2))) * ((
				d__3 = x[i__3].r, abs(d__3)) + (d__4 = hscal_d_imag(
				&x[ix]), abs(d__4)));
		    } else {
			i__2 = ix;
			temp = (d__1 = x[i__2].r, abs(d__1)) + (d__2 = hscal_d_imag(
				&x[ix]), abs(d__2));
		    }

		    i__2 = *n;
		    for (i__ = j + 1; i__ <= i__2; ++i__) {
			ix += *incx;
			i__3 = i__ + j * a_dim1;
			i__4 = ix;
			temp += ((d__1 = a[i__3].r, abs(d__1)) + (d__2 = 
				hscal_d_imag(&a[i__ + j * a_dim1]), abs(d__2))) * ((
				d__3 = x[i__4].r, abs(d__3)) + (d__4 = hscal_d_imag(
				&x[ix]), abs(d__4)));
                        
/* L210: */
		    }
		    y[jy] += talpha * temp;
		    jy += *incy;
                    
/* L220: */
		}
	    }
	}

    }

    return HSCAL_SUCCESS;

/*     End of ZATRMV */

} /* zatrmv_ */

/*------------------------------------------------------------------*/
