
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_ctradd_benchmark(
    char *uplo, 
    char *trans, 
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *beta, 
    hscalcomplex *c__,
    int     *ldc
)
{
    return hscal_ctradd_benchmark_i(
           uplo,
           trans,
           (hscalinteger*)m,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           beta,
           c__,
           (hscalinteger*)ldc
    );
}

/*------------------------------------------------------------------*/

int hscal_ctradd_benchmark_i(
    char *uplo, 
    char *trans, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    hscalinteger *lda, 
    hscalcomplex *beta, 
    hscalcomplex *c__,
    hscalinteger *ldc
)
{
    /*
    * This is a SCALAPACK TOOLS routine
    * 'trans' or 'uplo' is ignored for the present
    */
    cmatadd_(
       m,
       n,
       (float*)alpha,
       (float*)a,
       lda,
       (float*)beta,
       (float*)c__,
       ldc
    );

    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_dtradd_benchmark(
    char *uplo, 
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *beta, 
    hscaldoublereal *c__, 
    int     *ldc
)
{
    return hscal_dtradd_benchmark_i(
           uplo,
           trans,
           (hscalinteger*)m,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           beta,
           c__,
           (hscalinteger*)ldc
    );
}

/*------------------------------------------------------------------*/

int hscal_dtradd_benchmark_i(
    char *uplo, 
    char *trans, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    hscalinteger *lda, 
    hscaldoublereal *beta, 
    hscaldoublereal *c__, 
    hscalinteger *ldc
)
{
    /*
    * This is a SCALAPACK TOOLS routine
    * 'trans' or 'uplo' is ignored for the present
    */
    dmatadd_(
       m,
       n,
       (double*)alpha,
       (double*)a,
       lda,
       (double*)beta,
       (double*)c__,
       ldc
    );

    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_stradd_benchmark(
    char *uplo, 
    char *trans, 
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *beta, 
    hscalreal *c__, 
    int     * ldc
)
{
    return hscal_stradd_benchmark_i(
           uplo,
           trans,
           (hscalinteger*)m,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           beta,
           c__,
           (hscalinteger*)ldc
    );
}

/*------------------------------------------------------------------*/

int hscal_stradd_benchmark_i(
    char *uplo, 
    char *trans, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    hscalinteger *lda, 
    hscalreal *beta, 
    hscalreal *c__, 
    hscalinteger * ldc
)
{
    /*
    * This is a SCALAPACK TOOLS routine
    * 'trans' or 'uplo' is ignored for the present
    */
    smatadd_(
       m,
       n,
       (float*)alpha,
       (float*)a,
       lda,
       (float*)beta,
       (float*)c__,
       ldc
    );

    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_ztradd_benchmark(
    char *uplo, 
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
)
{
    return hscal_ztradd_benchmark_i(
           uplo,
           trans,
           (hscalinteger*)m,
           (hscalinteger*)n,
           alpha,
           a,
           (hscalinteger*)lda,
           beta,
           c__,
           (hscalinteger*)ldc
    );
}

/*------------------------------------------------------------------*/

int hscal_ztradd_benchmark_i(
    char *uplo, 
    char *trans, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    hscalinteger *lda, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    hscalinteger *ldc
)
{
    /*
    * This is a SCALAPACK TOOLS routine
    * 'trans' or 'uplo' is ignored for the present
    */
    zmatadd_(
       m,
       n,
       (double*)alpha,
       (double*)a,
       lda,
       (double*)beta,
       (double*)c__,
       ldc
    );

    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
