
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_ccopy_benchmark(
    int     *n,
    hscalcomplex *cx,
    int     *incx,
    hscalcomplex *cy,
    int     *incy
)
{
    ccopy_(
           n,
           (char*)cx,
           incx,
           (char*)cy,
           incy
    );
    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_dcopy_benchmark(
    int        *n,
    hscaldoublereal *dx,
    int        *incx,
    hscaldoublereal *dy,
    int        *incy
)
{
    dcopy_(
           n,
           (char*)dx,
           incx,
           (char*)dy,
           incy
    );
    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_icopy_benchmark(
    int *n,
    int *dx,
    int *incx,
    int *dy,
    int *incy
)
{
    /* System generated locals */
    hscalinteger i__1;
    /* Local variables */
    static hscalinteger i__, m, ix, iy, mp1;
/*     copies a vector, x, to a vector, y.
       uses unrolled loops for increments equal to one.
       jack dongarra, linpack, 3/11/78.
       modified 12/3/93, array(1) declarations changed to array(*)
       Parameter adjustments */
    --dy;
    --dx;
    /* Function Body */
    if (*n <= 0) {
        return HSCAL_SUCCESS;
    }
    if (*incx == 1 && *incy == 1) {
        goto L20;
    }
/*        code for unequal increments or equal increments
            not equal to 1 */
    ix = 1;
    iy = 1;
    if (*incx < 0) {
        ix = (-(*n) + 1) * *incx + 1;
    }
    if (*incy < 0) {
        iy = (-(*n) + 1) * *incy + 1;
    }
    i__1 = *n;
    for (i__ = 1; i__ <= i__1; ++i__) {
        dy[iy] = dx[ix];
        ix += *incx;
        iy += *incy;
/* L10: */
    }
    return HSCAL_SUCCESS;
/*        code for both increments equal to 1
          clean-up loop */
L20:
    m = *n % 7;
    if (m == 0) {
        goto L40;
    }
    i__1 = m;
    for (i__ = 1; i__ <= i__1; ++i__) {
        dy[i__] = dx[i__];
/* L30: */
    }
    if (*n < 7) {
        return HSCAL_SUCCESS;
    }
L40:
    mp1 = m + 1;
    i__1 = *n;
    for (i__ = mp1; i__ <= i__1; i__ += 7) {
        dy[i__] = dx[i__];
        dy[i__ + 1] = dx[i__ + 1];
        dy[i__ + 2] = dx[i__ + 2];
        dy[i__ + 3] = dx[i__ + 3];
        dy[i__ + 4] = dx[i__ + 4];
        dy[i__ + 5] = dx[i__ + 5];
        dy[i__ + 6] = dx[i__ + 6];
/* L50: */
    }
    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_scopy_benchmark(
    int     *n,
    hscalreal    *sx,
    int     *incx,
    hscalreal    *sy,
    int     *incy
)
{
    scopy_(
           n,
           (char*)sx,
           incx,
           (char*)sy,
           incy
    );
    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_zcopy_benchmark(
    int     *n,
    hscaldoublecomplex *zx,
    int     *incx,
    hscaldoublecomplex *zy,
    int     *incy
)
{
    zcopy_(
           n,
           (char*)zx,
           incx,
           (char*)zy,
           incy
    );
    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

