
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcatrmv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int UPLOI      =  params[0];
   int TRANSI     =  params[1];
   int DIAGI      =  params[1+1];
   int N          =  params[1+1+1];
   float ALPHA    =  params[1+1+1+1];
   int IA         =  params[1+1+1+1+1];
   int JA         =  params[1+1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   float BETA     =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int IY         =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int JY         =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];
   int *DESCY     = &params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1];
   int INCY       =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_];

   char UPLOC     = UPLOI;
   char UPLO[2]   = { UPLOC, '\0' };
   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };
   char DIAGC     = DIAGI;
   char DIAG[2]   = { DIAGC, '\0' };

   hscalcomplex *a, *x;
   hscalreal    *y;

   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);

   int lda               = reconn;
   int RECONX            = ( 1 + ( reconn - 1 )*abs( INCX ) );
   int RECONY            = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "UPLO %s TRANS %s DIAG %s N %d reconn %d ALPHA %f IA %d JA %d IX %d JX %d "
           "INCX %d BETA %f IY %d JY %d INCY %d",
           UPLO, TRANS, DIAG, N, reconn, ALPHA, IA, JA, IX, JX,
           INCX, BETA, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconn*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (hscalreal*)malloc(
              sizeof(hscalreal)
              *
              RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconn*reconn; i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 0.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS catrmv routine
   hscal_catrmv_benchmark(
     UPLO,
     TRANS,
     DIAG,
     &reconn,
     &ALPHA,
     a,
     &lda,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcatrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications initialization flags (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *TRANS;
   (*model_params)[1+1]                                             = *DIAG;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1]   = *nprow;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcatrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *TRANS;
   (*model_params)[1+1]                                             = *DIAG;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcatrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pcatrmv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcatrmv_timeof(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcatrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcatrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcatrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcatrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcatrmv_group_heuristic_pauto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcatrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psatrmv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int UPLOI      =  params[0];
   int TRANSI     =  params[1];
   int DIAGI      =  params[1+1];
   int N          =  params[1+1+1];
   float ALPHA    =  params[1+1+1+1];
   int IA         =  params[1+1+1+1+1];
   int JA         =  params[1+1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   float BETA     =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int IY         =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int JY         =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];
   int *DESCY     = &params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1];
   int INCY       =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_];

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };
   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };
   char DIAGC    = DIAGI;
   char DIAG[2]  = { DIAGC, '\0' };

   float *a, *x, *y;

   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);
   int lda               = reconn;
   int RECONX            = ( 1 + ( reconn - 1 )*abs( INCX ) );
   int RECONY            = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "UPLO %s TRANS %s DIAG %s N %d reconn %d ALPHA %f IA %d JA %d IX %d JX %d "
                  "INCX %d BETA %f IY %d JY %d INCY %d",
                  UPLO, TRANS, DIAG, N, reconn, ALPHA, IA, JA, IX, JX,
                  INCX, BETA, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (float*)malloc(
               sizeof(float)
               *
               (reconn*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (float*)malloc(
               sizeof(float)
               *
               RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (float*)malloc(
               sizeof(float)
               *
               RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < (reconn*reconn); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 0.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS satrmv routine
   hscal_satrmv_benchmark(
     UPLO,
     TRANS,
     DIAG,
     &reconn,
     &ALPHA,
     a,
     &lda,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psatrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications initialization flags (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *TRANS;
   (*model_params)[1+1]                                             = *DIAG;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1]   = *nprow;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psatrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *TRANS;
   (*model_params)[1+1]                                             = *DIAG;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_psatrmv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psatrmv_timeof(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psatrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psatrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_group_heuristic_pauto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdatrmv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int UPLOI      =  params[0];
   int TRANSI     =  params[1];
   int DIAGI      =  params[1+1];
   int N          =  params[1+1+1];
   double ALPHA   =  params[1+1+1+1];
   int IA         =  params[1+1+1+1+1];
   int JA         =  params[1+1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   double BETA    =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int IY         =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int JY         =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];
   int *DESCY     = &params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1];
   int INCY       =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_];

   char UPLOC     = UPLOI;
   char UPLO[2]   = { UPLOC, '\0' };
   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };
   char DIAGC     = DIAGI;
   char DIAG[2]   = { DIAGC, '\0' };

   double *a, *x, *y;

   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);

   int lda               = reconn;
   int RECONX            = ( 1 + ( reconn - 1 )*abs( INCX ) );
   int RECONY            = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "UPLO %s TRANS %s DIAG %s N %d reconn %d ALPHA %f IA %d JA %d IX %d JX %d "
                  "INCX %d BETA %f IY %d JY %d INCY %d",
                  UPLO, TRANS, DIAG, N, reconn, ALPHA, IA, JA, IX, JX,
                  INCX, BETA, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (double*)malloc(
                sizeof(double)
                *
                (reconn*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (double*)malloc(
                sizeof(double)
                *
                RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (double*)malloc(
                sizeof(double)
                *
                RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconn*reconn; i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 0.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS datrmv routine
   hscal_datrmv_benchmark(
     UPLO,
     TRANS,
     DIAG,
     &reconn,
     &ALPHA,
     a,
     &lda,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, pdgemv recon time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pdatrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pdatrmv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdatrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *TRANS;
   (*model_params)[1+1]                                             = *DIAG;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdatrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications initialization flags (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *TRANS;
   (*model_params)[1+1]                                             = *DIAG;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1]   = *nprow;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdatrmv_timeof(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow, 
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdatrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdatrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow, 
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdatrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdatrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdatrmv_group_heuristic_pauto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdatrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzatrmv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int UPLOI      =  params[0];
   int TRANSI     =  params[1];
   int DIAGI      =  params[1+1];
   int N          =  params[1+1+1];
   double ALPHA   =  params[1+1+1+1];
   int IA         =  params[1+1+1+1+1];
   int JA         =  params[1+1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   double BETA    =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int IY         =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int JY         =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];
   int *DESCY     = &params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1];
   int INCY       =  params[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_];

   char UPLOC     = UPLOI;
   char UPLO[2]   = { UPLOC, '\0' };
   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };
   char DIAGC     = DIAGI;
   char DIAG[2]   = { DIAGC, '\0' };

   hscaldoublecomplex *a, *x;
   hscaldoublereal    *y;

   int num_of_computers = HMPI_Get_number_of_computers();
   int reconn = (((N/num_of_computers) >= 1) ? (N/num_of_computers) : 1);

   int lda               = reconn;
   int RECONX            = ( 1 + ( reconn - 1 )*abs( INCX ) );
   int RECONY            = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "UPLO %s TRANS %s DIAG %s N %d reconn %d ALPHA %f IA %d JA %d IX %d JX %d "
                  "INCX %d BETA %f IY %d JY %d INCY %d",
                  UPLO, TRANS, DIAG, N, reconn, ALPHA, IA, JA, IX, JX,
                  INCX, BETA, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (hscaldoublecomplex*)malloc(
                       sizeof(hscaldoublecomplex)
                       *
                       (reconn*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (hscaldoublecomplex*)malloc(
                       sizeof(hscaldoublecomplex)
                       *
                       RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (hscaldoublereal*)malloc(
                    sizeof(hscaldoublereal)
                    *
                    RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconn*reconn; i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zatrmv routine
   hscal_zatrmv_benchmark(
     UPLO,
     TRANS,
     DIAG,
     &reconn,
     &ALPHA,
     a,
     &lda,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzatrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *TRANS;
   (*model_params)[1+1]                                             = *DIAG;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzatrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications initialization flags (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *TRANS;
   (*model_params)[1+1]                                             = *DIAG;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1]   = *nprow;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzatrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pzatrmv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzatrmv_timeof(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzatrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzatrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzatrmv_modelp_noauto(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzatrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzatrmv_group_heuristic_pauto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzatrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzatrmv_modelp(
                UPLO,
                TRANS,
                DIAG,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
