
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcgeadd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   int IA         =  params[1+1+CNARG+1];
   int JA         =  params[1+1+CNARG+1+1];
   int *DESCA     = &params[1+1+CNARG+1+1+1];
   int IC         =  params[1+1+CNARG+1+1+1+DLEN_+CNARG];
   int JC         =  params[1+1+CNARG+1+1+1+DLEN_+CNARG+1];
   int *DESCC     = &params[1+1+CNARG+1+1+1+DLEN_+CNARG+1+1];

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   hscalcomplex ALPHA, BETA;
   hscalcomplex *a, *c;

   int reconm            = hscal_pilaenv_recon("C");
   int reconn            = hscal_pilaenv_recon("C");
    
   int lda, ldc;

   ALPHA.r        = params[1+1+1];
   ALPHA.i        = params[1+1+1+1];

   BETA.r         = params[1+1+CNARG+1+1+1+DLEN_];
   BETA.i         = params[1+1+CNARG+1+1+1+DLEN_+1];

   if (TRANSC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANS %s M %d N %d reconm %d reconn %d "
           "ALPHA (%f, %f) IA %d JA %d BETA (%f, %f) IC %d JC %d",
           TRANS, M, N, reconm, reconn, 
           ALPHA.r, ALPHA.i, IA, JA, BETA.r, BETA.i, IC, JC
      );
   }

   //
   // Allocate the matrix A and C
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconn*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A and matrix C
   for (i = 0; i < (reconm*reconn); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;

      c[i].r = 2.;
      c[i].i = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cgeadd routine
   hscal_cgeadd_benchmark(
     TRANS,
     &reconm,
     &reconn,
     &ALPHA,
     a,
     &lda,
     &BETA,
     c,
     &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A and matrix C
   free(a);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcgeadd_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.f)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.f)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.f)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.f)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                       = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                     = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1]                   = *IC;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+1]                 = *JC;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+1+1+DLEN_]         = *nprow;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+1+1+DLEN_+1]       = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+CNARG+1+1+1+i]                     = DESCA[i];
       (*model_params)[1+1+CNARG+1+1+1+DLEN_+CNARG+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcgeadd_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.f)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.f)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.f)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.f)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                       = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                     = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1]                   = *IC;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+1]                 = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+CNARG+1+1+1+i]                     = DESCA[i];
       (*model_params)[1+1+CNARG+1+1+1+DLEN_+CNARG+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgeadd_recon(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pcgeadd_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcgeadd_timeof(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcgeadd_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcgeadd_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgeadd_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgeadd_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgeadd_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgeadd_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzgeadd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   int IA         =  params[1+1+CNARG+1];
   int JA         =  params[1+1+CNARG+1+1];
   int *DESCA     = &params[1+1+CNARG+1+1+1];
   int IC         =  params[1+1+CNARG+1+1+1+DLEN_+CNARG];
   int JC         =  params[1+1+CNARG+1+1+1+DLEN_+CNARG+1];
   int *DESCC     = &params[1+1+CNARG+1+1+1+DLEN_+CNARG+1+1];

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   hscaldoublecomplex ALPHA, BETA;
   hscaldoublecomplex *a, *c;

   int reconm            = hscal_pilaenv_recon("Z");
   int reconn            = hscal_pilaenv_recon("Z");
    
   int lda, ldc;

   ALPHA.r        = params[1+1+1];
   ALPHA.i        = params[1+1+1+1];

   BETA.r         = params[1+1+CNARG+1+1+1+DLEN_];
   BETA.i         = params[1+1+CNARG+1+1+1+DLEN_+1];

   if (TRANSC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANS %s M %d N %d reconm %d reconn %d "
           "ALPHA (%f, %f) IA %d JA %d BETA (%f, %f) IC %d JC %d",
           TRANS, M, N, reconm, reconn, 
           ALPHA.r, ALPHA.i, IA, JA, BETA.r, BETA.i, IC, JC
      );
   }

   //
   // Allocate the matrix A and C
   a = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A and matrix C
   for (i = 0; i < (reconm*reconn); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;

      c[i].r = 2.;
      c[i].i = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zgeadd routine
   hscal_zgeadd_benchmark(
     TRANS,
     &reconm,
     &reconn,
     &ALPHA,
     a,
     &lda,
     &BETA,
     c,
     &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A and matrix C
   free(a);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzgeadd_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                       = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                     = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1]                   = *IC;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+1]                 = *JC;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+1+1+DLEN_]         = *nprow;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+1+1+DLEN_+1]       = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+CNARG+1+1+1+i]                     = DESCA[i];
       (*model_params)[1+1+CNARG+1+1+1+DLEN_+CNARG+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzgeadd_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+1+DLEN_]                       = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1]                     = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1]                   = *IC;
   (*model_params)[1+1+1+1+1+1+1+DLEN_+1+1+1]                 = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+CNARG+1+1+1+i]                     = DESCA[i];
       (*model_params)[1+1+CNARG+1+1+1+DLEN_+CNARG+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgeadd_recon(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pzgeadd_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzgeadd_timeof(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzgeadd_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzgeadd_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgeadd_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgeadd_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgeadd_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgeadd_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/


double hscal_pdgeadd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   double ALPHA   =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   double BETA    =  params[1+1+1+1+1+1+DLEN_];
   int IC         =  params[1+1+1+1+1+1+DLEN_+1];
   int JC         =  params[1+1+1+1+1+1+DLEN_+1+1];
   int *DESCC     = &params[1+1+1+1+1+1+DLEN_+1+1+1];

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   double *a, *c;

   int reconm            = hscal_pilaenv_recon("D");
   int reconn            = hscal_pilaenv_recon("D");
    
   int lda, ldc;

   if (TRANSC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANS %s M %d N %d reconm %d reconn %d "
           "ALPHA %f IA %d JA %d BETA %f IC %d JC %d",
           TRANS, M, N, reconm, reconn, 
           ALPHA, IA, JA, BETA, IC, JC
      );
   }

   //
   // Allocate the matrix A and C
   a = (double*)malloc(
                 sizeof(double)
                 *
                 (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (double*)malloc(
                 sizeof(double)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A and matrix C
   for (i = 0; i < (reconm*reconn); i++)
   {
      a[i] = 2.;
      c[i] = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dgeadd routine
   hscal_dgeadd_benchmark(
     TRANS,
     &reconm,
     &reconn,
     &ALPHA,
     a,
     &lda,
     &BETA,
     c,
     &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A and matrix C
   free(a);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgeadd_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                         = betai;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                       = *IC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1]                     = *JC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1+DLEN_]             = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1+DLEN_+1]           = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+i]                 = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgeadd_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                         = betai;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                       = *IC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1]                     = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+i]                 = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_recon(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pdgeadd_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdgeadd_timeof(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdgeadd_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgeadd_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/


double hscal_psgeadd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   float ALPHA    =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   float BETA     =  params[1+1+1+1+1+1+DLEN_];
   int IC         =  params[1+1+1+1+1+1+DLEN_+1];
   int JC         =  params[1+1+1+1+1+1+DLEN_+1+1];
   int *DESCC     = &params[1+1+1+1+1+DLEN_+1+1+1];

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   float *a, *c;

   int reconm            = hscal_pilaenv_recon("S");
   int reconn            = hscal_pilaenv_recon("S");
    
   int lda, ldc;

   if (TRANSC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANS %s M %d N %d reconm %d reconn %d "
           "ALPHA %f IA %d JA %d BETA %f IC %d JC %d",
           TRANS, M, N, reconm, reconn, 
           ALPHA, IA, JA, BETA, IC, JC
      );
   }

   //
   // Allocate the matrix A and C
   a = (float*)malloc(
                 sizeof(float)
                 *
                 (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (float*)malloc(
                 sizeof(float)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A and matrix C
   for (i = 0; i < (reconm*reconn); i++)
   {
      a[i] = 2.;
      c[i] = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sgeadd routine
   hscal_sgeadd_benchmark(
     TRANS,
     &reconm,
     &reconn,
     &ALPHA,
     a,
     &lda,
     &BETA,
     c,
     &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A and matrix C
   free(a);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psgeadd_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                         = betai;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                       = *IC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1]                     = *JC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1+DLEN_]             = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1+DLEN_+1]           = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+i]                 = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psgeadd_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                         = betai;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                       = *IC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1]                     = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+i]                 = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgeadd_recon(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_psgeadd_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psgeadd_timeof(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psgeadd_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psgeadd_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgeadd_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgeadd_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgeadd_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgeadd_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgeadd_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
