
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcgemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSAI    =  params[0];
   int TRANSBI    =  params[1];
   int M          =  params[1+1];
   int N          =  params[1+1+1];
   int K          =  params[1+1+1+1];
   int IA         =  params[1+1+1+1+CNARG+1];
   int JA         =  params[1+1+1+1+CNARG+1+1];
   int *DESCA     = &params[1+1+1+1+CNARG+1+1+1];
   int IB         =  params[1+1+1+1+CNARG+1+1+1+DLEN_];
   int JB         =  params[1+1+1+1+CNARG+1+1+1+DLEN_+1];
   int *DESCB     = &params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1];
   int IC         =  params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_+CNARG];
   int JC         =  params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_+CNARG+1];
   int *DESCC     = &params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_+CNARG+1+1];

   char TRANSAC   = TRANSAI;
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = TRANSBI;
   char TRANSB[2] = { TRANSBC, '\0' };

   hscalcomplex ALPHA, BETA;
   hscalcomplex *a, *b, *c;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int reconk           = hscal_pilaenv_recon("C");

   int lda, ldb, ldc;

   ALPHA.r        = params[1+1+1+1+1];
   ALPHA.i        = params[1+1+1+1+1+1];

   BETA.r         = params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_];
   BETA.i         = params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_+1];

   if (TRANSAC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSBC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANSA %s TRANSB %s "
           "M %d N %d K %d "
           "reconm %d reconn %d reconk %d"
           "ALPHA (%f, %f) "
           "IA %d JA %d IB %d JB %d "
           "BETA (%f, %f) "
           "IC %d JC %d",
           TRANSA, TRANSB, 
           M, N, K, 
           reconm, reconn, reconk, 
           ALPHA.r, ALPHA.i,
           IA, JA, IB, JB, 
           BETA.r, BETA.i, 
           IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cgemm routine
   hscal_cgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcgemm_modelp_noauto(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.f)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.f)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.f)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.f)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANSA;
   (*model_params)[1]                                                 = *TRANSB;
   (*model_params)[1+1]                                               = *M;
   (*model_params)[1+1+1]                                             = *N;
   (*model_params)[1+1+1+1]                                           = *K;
   (*model_params)[1+1+1+1+1]                                         = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1]                                       = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]           = *JC;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]   = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]   = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcgemm_modelp(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.f)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.f)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.f)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.f)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +
                  1        /* M */
                  +
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANSA;
   (*model_params)[1]                                                 = *TRANSB;
   (*model_params)[1+1]                                               = *M;
   (*model_params)[1+1+1]                                             = *N;
   (*model_params)[1+1+1+1]                                           = *K;
   (*model_params)[1+1+1+1+1]                                         = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1]                                       = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]           = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]   = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgemm_recon(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pcgemm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcgemm_timeof(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemm_modelp_noauto(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcgemm_group_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemm_modelp_noauto(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgemm_group_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgemm_group_heuristic_pauto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgemm_group_heuristic_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdgemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSAI       =  params[0];
   int TRANSBI       =  params[1];
   int M             =  params[1+1];
   int N             =  params[1+1+1];
   int K             =  params[1+1+1+1];
   hscaldoublereal ALPHA  =  params[1+1+1+1+1];
   int IA            =  params[1+1+1+1+1+1];
   int JA            =  params[1+1+1+1+1+1+1];
   int *DESCA        = &params[1+1+1+1+1+1+1+1];
   int IB            =  params[1+1+1+1+1+1+1+1+DLEN_];
   int JB            =  params[1+1+1+1+1+1+1+1+DLEN_+1];
   int *DESCB        = &params[1+1+1+1+1+1+1+1+DLEN_+1+1];
   hscaldoublereal BETA   =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   int IC            =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int JC            =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int *DESCC        = &params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];

   char TRANSAC   = TRANSAI;
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = TRANSBI;
   char TRANSB[2] = { TRANSBC, '\0' };

   hscaldoublereal *a, *b, *c;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int reconk            = hscal_pilaenv_recon("D");

   int lda, ldb, ldc;

   if (TRANSAC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSBC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANSA %s TRANSB %s "
           "M %d N %d K %d "
           "reconm %d reconn %d reconk %d"
           "ALPHA %f "
           "IA %d JA %d IB %d JB %d "
           "BETA %f "
           "IC %d JC %d",
           TRANSA, TRANSB, 
           M, N, K, 
           reconm, reconn, reconk, 
           ALPHA,
           IA, JA, IB, JB, 
           BETA, 
           IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dgemm routine
   hscal_dgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgemm_modelp_noauto(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANSA;
   (*model_params)[1]                                               = *TRANSB;
   (*model_params)[1+1]                                             = *M;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = *K;
   (*model_params)[1+1+1+1+1]                                       = alphai;
   (*model_params)[1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *JC;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+DLEN_]     = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+DLEN_+1]   = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgemm_modelp(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +
                  1        /* M */
                  +
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANSA;
   (*model_params)[1]                                               = *TRANSB;
   (*model_params)[1+1]                                             = *M;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = *K;
   (*model_params)[1+1+1+1+1]                                       = alphai;
   (*model_params)[1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemm_recon(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pdgemm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdgemm_timeof(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemm_modelp_noauto(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdgemm_group_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemm_modelp_noauto(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemm_group_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemm_group_heuristic_pauto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemm_group_heuristic_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psgemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSAI       =  params[0];
   int TRANSBI       =  params[1];
   int M             =  params[1+1];
   int N             =  params[1+1+1];
   int K             =  params[1+1+1+1];
   hscalreal ALPHA        =  params[1+1+1+1+1];
   int IA            =  params[1+1+1+1+1+1];
   int JA            =  params[1+1+1+1+1+1+1];
   int *DESCA        = &params[1+1+1+1+1+1+1+1];
   int IB            =  params[1+1+1+1+1+1+1+1+DLEN_];
   int JB            =  params[1+1+1+1+1+1+1+1+DLEN_+1];
   int *DESCB        = &params[1+1+1+1+1+1+1+1+DLEN_+1+1];
   hscalreal BETA         =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   int IC            =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int JC            =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int *DESCC        = &params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];

   char TRANSAC   = TRANSAI;
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = TRANSBI;
   char TRANSB[2] = { TRANSBC, '\0' };

   hscalreal *a, *b, *c;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int reconk            = hscal_pilaenv_recon("S");

   int lda, ldb, ldc;

   if (TRANSAC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSBC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANSA %s TRANSB %s "
           "M %d N %d K %d "
           "reconm %d reconn %d reconk %d"
           "ALPHA %f "
           "IA %d JA %d IB %d JB %d "
           "BETA %f "
           "IC %d JC %d",
           TRANSA, TRANSB, 
           M, N, K, 
           reconm, reconn, reconk, 
           ALPHA,
           IA, JA, IB, JB, 
           BETA, 
           IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.f;
      a[i] = 2.f;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.f;
      b[i] = 2.f;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.f;
      c[i] = 1.f;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sgemm routine
   hscal_sgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psgemm_modelp_noauto(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANSA;
   (*model_params)[1]                                               = *TRANSB;
   (*model_params)[1+1]                                             = *M;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = *K;
   (*model_params)[1+1+1+1+1]                                       = alphai;
   (*model_params)[1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *JC;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+DLEN_]     = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+DLEN_+1]   = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psgemm_modelp(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +
                  1        /* M */
                  +
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANSA;
   (*model_params)[1]                                               = *TRANSB;
   (*model_params)[1+1]                                             = *M;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = *K;
   (*model_params)[1+1+1+1+1]                                       = alphai;
   (*model_params)[1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgemm_recon(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_psgemm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psgemm_timeof(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psgemm_modelp_noauto(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psgemm_group_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgemm_modelp_noauto(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgemm_group_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgemm_group_heuristic_pauto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgemm_group_heuristic_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzgemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSAI    =  params[0];
   int TRANSBI    =  params[1];
   int M          =  params[1+1];
   int N          =  params[1+1+1];
   int K          =  params[1+1+1+1];
   int IA         =  params[1+1+1+1+CNARG+1];
   int JA         =  params[1+1+1+1+CNARG+1+1];
   int *DESCA     = &params[1+1+1+1+CNARG+1+1+1];
   int IB         =  params[1+1+1+1+CNARG+1+1+1+DLEN_];
   int JB         =  params[1+1+1+1+CNARG+1+1+1+DLEN_+1];
   int *DESCB     = &params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1];
   int IC         =  params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_+CNARG];
   int JC         =  params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_+CNARG+1];
   int *DESCC     = &params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_+CNARG+1+1];

   char TRANSAC   = TRANSAI;
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = TRANSBI;
   char TRANSB[2] = { TRANSBC, '\0' };

   hscaldoublecomplex ALPHA, BETA;
   hscaldoublecomplex *a, *b, *c;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int reconk            = hscal_pilaenv_recon("Z");
    
   int lda, ldb, ldc;

   ALPHA.r        = params[1+1+1+1+1];
   ALPHA.i        = params[1+1+1+1+1+1];

   BETA.r         = params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_];
   BETA.i         = params[1+1+1+1+CNARG+1+1+1+DLEN_+1+1+DLEN_+1];

   if (TRANSAC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSBC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANSA %s TRANSB %s "
           "M %d N %d K %d "
           "reconm %d reconn %d reconk %d"
           "ALPHA (%f, %f) "
           "IA %d JA %d IB %d JB %d "
           "BETA (%f, %f) "
           "IC %d JC %d",
           TRANSA, TRANSB, 
           M, N, K, 
           reconm, reconn, reconk, 
           ALPHA.r, ALPHA.i,
           IA, JA, IB, JB, 
           BETA.r, BETA.i, 
           IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zgemm routine
   hscal_zgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzgemm_modelp_noauto(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANSA;
   (*model_params)[1]                                                 = *TRANSB;
   (*model_params)[1+1]                                               = *M;
   (*model_params)[1+1+1]                                             = *N;
   (*model_params)[1+1+1+1]                                           = *K;
   (*model_params)[1+1+1+1+1]                                         = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1]                                       = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]           = *JC;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]   = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]   = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzgemm_modelp(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +
                  1        /* M */
                  +
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANSA;
   (*model_params)[1]                                                 = *TRANSB;
   (*model_params)[1+1]                                               = *M;
   (*model_params)[1+1+1]                                             = *N;
   (*model_params)[1+1+1+1]                                           = *K;
   (*model_params)[1+1+1+1+1]                                         = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1]                                       = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]           = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]   = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_recon(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pzgemm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzgemm_timeof(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemm_modelp_noauto(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_group_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemm_modelp_noauto(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_group_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_group_heuristic_pauto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_group_heuristic_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemm_modelp(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

