
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_pack_pzgemm_modelp_noauto_optimal_b(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * reconm,
   int * reconn,
   int * reconk,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* reconm */
                  +
                  1        /* reconn */
                  +
                  1        /* Optimal B */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                   = *TRANSA;
   (*model_params)[1]                                                   = *TRANSB;
   (*model_params)[1+1]                                                 = *M;
   (*model_params)[1+1+1]                                               = *N;
   (*model_params)[1+1+1+1]                                             = *K;
   (*model_params)[1+1+1+1+1]                                           = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1]                                         = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                             = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                           = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JC;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *reconm;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1]   = *reconn;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1] = *reconk;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1+1] = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]   = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
