
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcgemv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   hscalcomplex ALPHA, BETA;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   int IA         =  params[1+1+1+2];
   int JA         =  params[1+1+1+2+1];
   int *DESCA     = &params[1+1+1+2+1+1];
   int IX         =  params[1+1+1+2+1+1+DLEN_];
   int JX         =  params[1+1+1+2+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+2+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_];
   int IY         =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2];
   int JY         =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1];
   int *DESCY     = &params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1];
   int INCY       =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_];

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   hscalcomplex *a, *x, *y;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int lda               = reconm;
   int RECONX, RECONY;

   ALPHA.r        = params[1+1+1];
   ALPHA.i        = params[1+1+1+1];

   BETA.r         = params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1];
   BETA.i         = params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+1];

   if (TRANSC == 'N')
   {
      RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );
      RECONY = ( 1 + ( reconm - 1 )*abs( INCY ) );
   }
   else
   {
      RECONX = ( 1 + ( reconm - 1 )*abs( INCX ) );
      RECONY = ( 1 + ( reconn - 1 )*abs( INCY ) );
   }

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANS %s M %d N %d reconM %d recon N %d ALPHA (%f, %f) IA %d JA %d IX %d JX %d "
           "INCX %d BETA (%f, %f) IY %d JY %d INCY %d",
           TRANS, M, N, reconm, reconn, ALPHA.r, ALPHA.i, IA, JA, IX, JX,
           INCX, BETA.r, BETA.i, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconm*reconn; i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i].r = 0.;
      y[i].i = 0.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cgemv routine
   hscal_cgemv_benchmark(
     TRANS,
     &reconm,
     &reconn,
     &ALPHA,
     a,
     &lda,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcgemv_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.f)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.f)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.f)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.f)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  CNARG    /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count), 
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                             = *TRANS;
   (*model_params)[1]                                             = *M;
   (*model_params)[1+1]                                           = *N;
   (*model_params)[1+1+1]                                         = alphai[REAL_PART];
   (*model_params)[1+1+1+1]                                       = alphai[IMAG_PART];
   (*model_params)[1+1+1+2]                                       = *IA;
   (*model_params)[1+1+1+2+1]                                     = *JA;
   (*model_params)[1+1+1+2+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1]                 = betai[REAL_PART];
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+1]               = betai[IMAG_PART];
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2]               = *IY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1]             = *JY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_]     = *INCY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_+1]   = *nprow;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+2+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+2+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcgemv_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.f)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.f)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.f)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.f)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  CNARG    /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+2]                                   = *IA;
   (*model_params)[1+1+1+2+1]                                 = *JA;
   (*model_params)[1+1+1+2+1+1+DLEN_]                         = *IX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1]                       = *JX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_]               = *INCX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1]             = betai[REAL_PART];
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+1]           = betai[IMAG_PART];
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2]           = *IY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1]         = *JY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_] = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+2+1+1+i]                         = DESCA[i];
       (*model_params)[1+1+1+2+1+1+DLEN_+1+1+i]               = DESCX[i];
       (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+i] = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pcgemv_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcgemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemv_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcgemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemv_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgemv_group_heuristic_pauto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psgemv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   float ALPHA    =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   float BETA     =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int IY         =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int JY         =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];
   int *DESCY     = &params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1];
   int INCY       =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_];

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   float *a, *x, *y;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int lda               = reconm;
   int RECONX, RECONY;

   if (TRANSC == 'N')
   {
      RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );
      RECONY = ( 1 + ( reconm - 1 )*abs( INCY ) );
   }
   else
   {
      RECONX = ( 1 + ( reconm - 1 )*abs( INCX ) );
      RECONY = ( 1 + ( reconn - 1 )*abs( INCY ) );
   }

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "TRANS %s M %d N reconM %d reconN %d %d ALPHA %f IA %d JA %d IX %d JX %d "
                  "INCX %d BETA %f IY %d JY %d INCY %d",
                  TRANS, M, N, reconm, reconn, ALPHA, IA, JA, IX, JX,
                  INCX, BETA, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (float*)malloc(
               sizeof(float)
               *
               (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (float*)malloc(
               sizeof(float)
               *
               RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (float*)malloc(
               sizeof(float)
               *
               RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconm*reconn; i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 0.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sgemv routine
   hscal_sgemv_benchmark(
     TRANS,
     &reconm,
     &reconn,
     &ALPHA,
     a,
     &lda,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psgemv_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                             = *TRANS;
   (*model_params)[1]                                             = *M;
   (*model_params)[1+1]                                           = *N;
   (*model_params)[1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1]   = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psgemv_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   float * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                         = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                       = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]               = *INCX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]             = betai;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]           = *IY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]         = *JY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_] = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                         = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]               = DESCX[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i] = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_psgemv_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psgemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psgemv_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psgemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgemv_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgemv_group_heuristic_pauto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdgemv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   double ALPHA   =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IX         =  params[1+1+1+1+1+1+DLEN_];
   int JX         =  params[1+1+1+1+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+1+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   double BETA    =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int IY         =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int JY         =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];
   int *DESCY     = &params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1];
   int INCY       =  params[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_];

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   double *a, *x, *y;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int lda               = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "TRANS %s M %d N %d reconM %d reconN %d ALPHA %f IA %d JA %d IX %d JX %d "
                  "INCX %d BETA %f IY %d JY %d INCY %d",
                  TRANS, M, N, reconm, reconn, ALPHA, IA, JA, IX, JX,
                  INCX, BETA, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (double*)malloc(
                sizeof(double)
                *
                (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (double*)malloc(
                sizeof(double)
                *
                (reconn)
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (double*)malloc(
                sizeof(double)
                *
                reconm
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconm*reconn; i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < reconn; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < reconm; i++)
   {
      y[i] = 0.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dgemv routine
   hscal_dgemv_benchmark(
     TRANS,
     &reconm,
     &reconn,
     &ALPHA,
     a,
     &lda,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, pdgemv recon time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pdgemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pdgemv_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgemv_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                         = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                       = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]               = *INCX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]             = betai;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]           = *IY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]         = *JY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_] = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                         = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]               = DESCX[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i] = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgemv_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  1        /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                             = *TRANS;
   (*model_params)[1]                                             = *M;
   (*model_params)[1+1]                                           = *N;
   (*model_params)[1+1+1]                                         = alphai;
   (*model_params)[1+1+1+1]                                       = *IA;
   (*model_params)[1+1+1+1+1]                                     = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]                 = betai;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]               = *IY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1]             = *JY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_]     = *INCY;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1]   = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+DLEN_+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdgemv_timeof(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow, 
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemv_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdgemv_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow, 
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemv_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemv_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemv_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemv_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzgemv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   hscaldoublecomplex ALPHA, BETA;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   int IA         =  params[1+1+1+2];
   int JA         =  params[1+1+1+2+1];
   int *DESCA     = &params[1+1+1+2+1+1];
   int IX         =  params[1+1+1+2+1+1+DLEN_];
   int JX         =  params[1+1+1+2+1+1+DLEN_+1];
   int *DESCX     = &params[1+1+1+2+1+1+DLEN_+1+1];
   int INCX       =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_];
   int IY         =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2];
   int JY         =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1];
   int *DESCY     = &params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1];
   int INCY       =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_];

   char TRANSC    = TRANSI;
   char TRANS[2]  = { TRANSC, '\0' };

   hscaldoublecomplex *a, *x, *y;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int lda               = reconm;
   int RECONX, RECONY;

   ALPHA.r        =  params[1+1+1];
   ALPHA.i        =  params[1+1+1+1];

   BETA.r         =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1];
   BETA.i         =  params[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+1];

   if (TRANSC == 'N')
   {
      RECONX = ( 1 + ( reconn - 1 )*abs( INCX ) );
      RECONY = ( 1 + ( reconm - 1 )*abs( INCY ) );
   }
   else
   {
      RECONX = ( 1 + ( reconm - 1 )*abs( INCX ) );
      RECONY = ( 1 + ( reconn - 1 )*abs( INCY ) );
   }

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "TRANS %s M %d N %d reconM %d reconN %d ALPHA (%f, %f) IA %d JA %d IX %d JX %d "
                  "INCX %d BETA (%f, %f) IY %d JY %d INCY %d",
                  TRANS, M, N, reconm, reconn, ALPHA.r, ALPHA.i, IA, JA, IX, JX,
                  INCX, BETA.r, BETA.i, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (hscaldoublecomplex*)malloc(
                       sizeof(hscaldoublecomplex)
                       *
                       (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (hscaldoublecomplex*)malloc(
                       sizeof(hscaldoublecomplex)
                       *
                       RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (hscaldoublecomplex*)malloc(
                       sizeof(hscaldoublecomplex)
                       *
                       RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconm*reconn; i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i].r = 2.;
      x[i].i = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i].r = 2.;
      y[i].i = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zgemv routine
   hscal_zgemv_benchmark(
     TRANS,
     &reconm,
     &reconn,
     &ALPHA,
     a,
     &lda,
     x,
     &INCX,
     &BETA,
     y,
     &INCY
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzgemv_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  CNARG    /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                         = *TRANS;
   (*model_params)[1]                                         = *M;
   (*model_params)[1+1]                                       = *N;
   (*model_params)[1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+2]                                   = *IA;
   (*model_params)[1+1+1+2+1]                                 = *JA;
   (*model_params)[1+1+1+2+1+1+DLEN_]                         = *IX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1]                       = *JX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_]               = *INCX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1]             = betai[REAL_PART];
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+1]           = betai[IMAG_PART];
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2]           = *IY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1]         = *JY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_] = *INCY;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+2+1+1+i]                         = DESCA[i];
       (*model_params)[1+1+1+2+1+1+DLEN_+1+1+i]               = DESCX[i];
       (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+i] = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzgemv_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */        
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */        
                  +
                  CNARG    /* BETA */        
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),  
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                             = *TRANS;
   (*model_params)[1]                                             = *M;
   (*model_params)[1+1]                                           = *N;
   (*model_params)[1+1+1]                                         = alphai[REAL_PART];
   (*model_params)[1+1+1+1]                                       = alphai[IMAG_PART];
   (*model_params)[1+1+1+2]                                       = *IA;
   (*model_params)[1+1+1+2+1]                                     = *JA;
   (*model_params)[1+1+1+2+1+1+DLEN_]                             = *IX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1]                           = *JX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_]                   = *INCX;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1]                 = betai[REAL_PART];
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+1]               = betai[IMAG_PART];
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2]               = *IY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1]             = *JY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_]     = *INCY;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_+1]   = *nprow;
   (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+DLEN_+1+1] = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+2+1+1+i]                             = DESCA[i];
       (*model_params)[1+1+1+2+1+1+DLEN_+1+1+i]                   = DESCX[i];
       (*model_params)[1+1+1+2+1+1+DLEN_+1+1+DLEN_+1+2+1+1+i]     = DESCY[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      {
         rc = HMPI_Recon_t(
                  &hscal_pzgemv_benchmark_function,
                  model_params,
                  model_count,
                  &output_p
         );

         if (rc != HMPI_SUCCESS)
         {
            free(model_params);
            return rc;
         }
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzgemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemv_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemv_modelp_noauto(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_group_heuristic_pauto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgemv_modelp(
                TRANS,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IX,
                JX,
                DESCX,
                INCX,
                BETA,
                IY,
                JY,
                DESCY,
                INCY,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
