
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_psger_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int M          =  params[0];
   int N          =  params[1];
   float ALPHA    =  params[1+1];
   int IX         =  params[1+1+1];
   int JX         =  params[1+1+1+1];
   int *DESCX     = &params[1+1+1+1+1];
   int INCX       =  params[1+1+1+1+1+DLEN_];
   int IY         =  params[1+1+1+1+1+DLEN_+1];
   int JY         =  params[1+1+1+1+1+DLEN_+1+1];
   int *DESCY     = &params[1+1+1+1+1+DLEN_+1+1+1];
   int INCY       =  params[1+1+1+1+1+DLEN_+1+1+1+DLEN_];
   int IA         =  params[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1];
   int JA         =  params[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1];
   int *DESCA     = &params[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1];

   float *a, *x, *y;

   int reconm            = hscal_pilaenv_recon("S");
   int reconn            = hscal_pilaenv_recon("S");
   int lda               = reconm;
   int RECONX            = ( 1 + ( reconm - 1 )*abs( INCX ) );
   int RECONY            = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "M %d N %d reconm %d reconn %d ALPHA %f IA %d JA %d IX %d JX %d "
                  "INCX %d IY %d JY %d INCY %d",
                  M, N, reconm , reconn, ALPHA, IA, JA, IX, JX,
                  INCX, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (float*)malloc(
               sizeof(float)
               *
               (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (float*)malloc(
               sizeof(float)
               *
               RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (float*)malloc(
               sizeof(float)
               *
               RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconm*reconn; i++)
   {
      a[i] = 0.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sger routine
   hscal_sger_benchmark(
     &reconm,
     &reconn,
     &ALPHA,
     x,
     &INCX,
     y,
     &INCY,
     a,
     &lda
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psger_modelp_noauto(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */ 
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* IA */
                  +
                  1        /* JA */
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *M;
   (*model_params)[1]                                           = *N;
   (*model_params)[1+1]                                         = alphai;
   (*model_params)[1+1+1]                                       = *IX;
   (*model_params)[1+1+1+1]                                     = *JX;
   (*model_params)[1+1+1+1+1+DLEN_]                             = *INCX;
   (*model_params)[1+1+1+1+1+DLEN_+1]                           = *IY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]                         = *JY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_]                 = *INCY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1]               = *IA;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1]             = *JA;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1+DLEN_]     = *nprow;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1+DLEN_+1]   = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                             = DESCX[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+i]                 = DESCY[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1+i]     = DESCA[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psger_modelp(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
)
{
   int i;
   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */ 
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* IA */
                  +
                  1        /* JA */
                  +
                  DLEN_    /* DESCA */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *M;
   (*model_params)[1]                                           = *N;
   (*model_params)[1+1]                                         = alphai;
   (*model_params)[1+1+1]                                       = *IX;
   (*model_params)[1+1+1+1]                                     = *JX;
   (*model_params)[1+1+1+1+1+DLEN_]                             = *INCX;
   (*model_params)[1+1+1+1+1+DLEN_+1]                           = *IY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]                         = *JY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_]                 = *INCY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1]               = *IA;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1]             = *JA;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                             = DESCX[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+i]                 = DESCY[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1+i]     = DESCA[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psger_recon(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psger_modelp(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_psger_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psger_timeof(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psger_modelp_noauto(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psger_group_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psger_modelp_noauto(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psger_group_auto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psger_modelp(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psger_group_heuristic_pauto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psger_modelp(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psger_group_heuristic_auto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psger_modelp(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdger_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int M          =  params[0];
   int N          =  params[1];
   double ALPHA   =  params[1+1];
   int IX         =  params[1+1+1];
   int JX         =  params[1+1+1+1];
   int *DESCX     = &params[1+1+1+1+1];
   int INCX       =  params[1+1+1+1+1+DLEN_];
   int IY         =  params[1+1+1+1+1+DLEN_+1];
   int JY         =  params[1+1+1+1+1+DLEN_+1+1];
   int *DESCY     = &params[1+1+1+1+1+DLEN_+1+1+1];
   int INCY       =  params[1+1+1+1+1+DLEN_+1+1+1+DLEN_];
   int IA         =  params[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1];
   int JA         =  params[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1];
   int *DESCA     = &params[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1];

   double *a, *x, *y;

   int reconm            = hscal_pilaenv_recon("D");
   int reconn            = hscal_pilaenv_recon("D");
   int lda               = reconm;
   int RECONX            = ( 1 + ( reconm - 1 )*abs( INCX ) );
   int RECONY            = ( 1 + ( reconn - 1 )*abs( INCY ) );

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__,
                  "M %d N %d reconm %d reconn %d ALPHA %f IA %d JA %d IX %d JX %d "
                  "INCX %d IY %d JY %d INCY %d",
                  M, N, reconm , reconn, ALPHA, IA, JA, IX, JX,
                  INCX, IY, JY, INCY
      );
   }

   //
   // Allocate the matrix A, vectors x and y
   a = (double*)malloc(
               sizeof(double)
               *
               (reconm*reconn)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   x = (double*)malloc(
               sizeof(double)
               *
               RECONX
   );

   if (x == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   y = (double*)malloc(
               sizeof(double)
               *
               RECONY
   );


   if (y == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrix A, vectors x and y
   for (i = 0; i < reconm*reconn; i++)
   {
      a[i] = 0.;
   }

   for (i = 0; i < RECONX; i++)
   {
      x[i] = 2.;
   }

   for (i = 0; i < RECONY; i++)
   {
      y[i] = 2.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dger routine
   hscal_dger_benchmark(
     &reconm,
     &reconn,
     &ALPHA,
     x,
     &INCX,
     y,
     &INCY,
     a,
     &lda
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrix A, vectors x and y
   free(a);
   free(x);
   free(y);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdger_modelp_noauto(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */ 
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* IA */
                  +
                  1        /* JA */
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *M;
   (*model_params)[1]                                           = *N;
   (*model_params)[1+1]                                         = alphai;
   (*model_params)[1+1+1]                                       = *IX;
   (*model_params)[1+1+1+1]                                     = *JX;
   (*model_params)[1+1+1+1+1+DLEN_]                             = *INCX;
   (*model_params)[1+1+1+1+1+DLEN_+1]                           = *IY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]                         = *JY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_]                 = *INCY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1]               = *IA;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1]             = *JA;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1+DLEN_]     = *nprow;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1+DLEN_+1]   = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                             = DESCX[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+i]                 = DESCY[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1+i]     = DESCA[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdger_modelp(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IX */
                  +
                  1        /* JX */ 
                  +
                  DLEN_    /* DESCX */
                  +
                  1        /* INCX */
                  +
                  1        /* IY */
                  +
                  1        /* JY */
                  +
                  DLEN_    /* DESCY */
                  +
                  1        /* INCY */
                  +
                  1        /* IA */
                  +
                  1        /* JA */
                  +
                  DLEN_    /* DESCA */
   ;

   *model_params = (int*)malloc(
                         sizeof(int)
                         *
                         (*model_count)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                           = *M;
   (*model_params)[1]                                           = *N;
   (*model_params)[1+1]                                         = alphai;
   (*model_params)[1+1+1]                                       = *IX;
   (*model_params)[1+1+1+1]                                     = *JX;
   (*model_params)[1+1+1+1+1+DLEN_]                             = *INCX;
   (*model_params)[1+1+1+1+1+DLEN_+1]                           = *IY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]                         = *JY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_]                 = *INCY;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1]               = *IA;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1]             = *JA;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                             = DESCX[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+i]                 = DESCY[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1+1+1+i]     = DESCA[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdger_recon(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdger_modelp(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pdger_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdger_timeof(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdger_modelp_noauto(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdger_group_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdger_modelp_noauto(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdger_group_auto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdger_modelp(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdger_group_heuristic_pauto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdger_modelp(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdger_group_heuristic_auto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdger_modelp(
                M,
                N,
                ALPHA,
                IX,
                JX,
                DESCX,
                INCX,
                IY,
                JY,
                DESCY,
                INCY,
                IA,
                JA,
                DESCA,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
