
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pstran_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int M          =  params[0];
   int N          =  params[1];
   hscalreal ALPHA     =  params[1+1];
   int IA         =  params[1+1+1];
   int JA         =  params[1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1];
   hscalreal BETA      =  params[1+1+1+1+1+DLEN_];
   int IC         =  params[1+1+1+1+1+DLEN_+1];
   int JC         =  params[1+1+1+1+1+DLEN_+1+1];
   int *DESCC     = &params[1+1+1+1+1+DLEN_+1+1+1];

   char TRANSAC   = 'N';
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = 'N';
   char TRANSB[2] = { TRANSBC, '\0' };

   hscalreal *a, *b, *c;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int reconk            = hscal_pilaenv_recon("S");

   int lda, ldb, ldc;

   lda = reconm;
   ldb = reconk;
   ldc = reconm;
    
   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "M %d N %d reconm %d reconn %d "
           "ALPHA %f IA %d JA %d BETA %f IC %d JC %d",
           M, N, reconm, reconn, 
           ALPHA, IA, JA, BETA, IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.f;
      a[i] = 2.f;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.f;
      b[i] = 2.f;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.f;
      c[i] = 1.f;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sgemm routine
   hscal_sgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;
   
   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pstran_modelp_noauto(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                     = *M;
   (*model_params)[1]                                     = *N;
   (*model_params)[1+1]                                   = alphai;
   (*model_params)[1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+DLEN_]                       = betai;
   (*model_params)[1+1+1+1+1+DLEN_+1]                     = *IC;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]                   = *JC;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_]           = *nprow;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1]         = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                       = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+i]           = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pstran_modelp(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                     = *M;
   (*model_params)[1]                                     = *N;
   (*model_params)[1+1]                                   = alphai;
   (*model_params)[1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+DLEN_]                       = betai;
   (*model_params)[1+1+1+1+1+DLEN_+1]                     = *IC;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]                   = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                       = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+i]           = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstran_recon(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pstran_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pstran_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pstran_timeof(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pstran_modelp_noauto(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pstran_group_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstran_modelp_noauto(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstran_group_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstran_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstran_group_heuristic_pauto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pstran_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstran_group_heuristic_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstran_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdtran_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int M                =  params[0];
   int N                =  params[1];
   hscaldoublereal ALPHA     =  params[1+1];
   int IA               =  params[1+1+1];
   int JA               =  params[1+1+1+1];
   int *DESCA           = &params[1+1+1+1+1];
   hscaldoublereal BETA      =  params[1+1+1+1+1+DLEN_];
   int IC               =  params[1+1+1+1+1+DLEN_+1];
   int JC               =  params[1+1+1+1+1+DLEN_+1+1];
   int *DESCC           = &params[1+1+1+1+1+DLEN_+1+1+1];

   char TRANSAC   = 'N';
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = 'N';
   char TRANSB[2] = { TRANSBC, '\0' };

   hscaldoublereal *a, *b, *c;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int reconk           = hscal_pilaenv_recon("D");

   int lda, ldb, ldc;

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "M %d N %d reconm %d reconn %d "
           "ALPHA %f IA %d JA %d BETA %f IC %d JC %d",
           M, N, reconm, reconn, 
           ALPHA, IA, JA, BETA, IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dgemm routine
   hscal_dgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdtran_modelp_noauto(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                     = *M;
   (*model_params)[1]                                     = *N;
   (*model_params)[1+1]                                   = alphai;
   (*model_params)[1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+DLEN_]                       = betai;
   (*model_params)[1+1+1+1+1+DLEN_+1]                     = *IC;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]                   = *JC;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_]           = *nprow;
   (*model_params)[1+1+1+1+1+DLEN_+1+1+1+DLEN_+1]         = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                       = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+i]           = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdtran_modelp(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                     = *M;
   (*model_params)[1]                                     = *N;
   (*model_params)[1+1]                                   = alphai;
   (*model_params)[1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+DLEN_]                       = betai;
   (*model_params)[1+1+1+1+1+DLEN_+1]                     = *IC;
   (*model_params)[1+1+1+1+1+DLEN_+1+1]                   = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                       = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN_+1+1+1+i]           = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtran_recon(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdtran_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pdtran_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdtran_timeof(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdtran_modelp_noauto(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdtran_group_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtran_modelp_noauto(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtran_group_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtran_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtran_group_heuristic_pauto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdtran_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtran_group_heuristic_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtran_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
