
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pctranu_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int M          =  params[0];
   int N          =  params[1];
   int IA         =  params[1+CNARG+1];
   int JA         =  params[1+CNARG+1+1];
   int *DESCA     = &params[1+CNARG+1+1+1];
   int IC         =  params[1+CNARG+1+1+1+DLEN_+CNARG];
   int JC         =  params[1+CNARG+1+1+1+DLEN_+CNARG+1];
   int *DESCC     = &params[1+CNARG+1+1+1+DLEN_+CNARG+1+1];

   hscalcomplex ALPHA, BETA;

   char TRANSAC   = 'N';
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = 'N';
   char TRANSB[2] = { TRANSBC, '\0' };

   hscalcomplex *a, *b, *c;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int reconk           = hscal_pilaenv_recon("C");

   int lda, ldb, ldc;

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   ALPHA.r        = params[1+1];
   ALPHA.i        = params[1+1+1];

   BETA.r         = params[1+CNARG+1+1+1+DLEN_];
   BETA.i         = params[1+CNARG+1+1+1+DLEN_+1];

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "M %d N %d reconm %d reconn %d "
           "ALPHA (%f, %f) IA %d JA %d BETA (%f, %f) IC %d JC %d",
           M, N, reconm, reconn, 
           ALPHA.r, ALPHA.i, IA, JA, BETA.r, BETA.i, IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cgemm routine
   hscal_cgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pctranu_modelp_noauto(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.f)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.f)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.f)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.f)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                       = *M;
   (*model_params)[1]                                       = *N;
   (*model_params)[1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                       = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                     = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1]                   = *IC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1]                 = *JC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1+1+DLEN_]         = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1+1+DLEN_+1]       = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+CNARG+1+1+1+i]                     = DESCA[i];
       (*model_params)[1+CNARG+1+1+1+DLEN_+CNARG+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pctranu_modelp(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.f)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.f)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.f)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.f)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                       = *M;
   (*model_params)[1]                                       = *N;
   (*model_params)[1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                       = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                     = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1]                   = *IC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1]                 = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+CNARG+1+1+1+i]                     = DESCA[i];
       (*model_params)[1+CNARG+1+1+1+DLEN_+CNARG+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctranu_recon(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pctranu_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pctranu_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pctranu_timeof(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pctranu_modelp_noauto(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pctranu_group_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctranu_modelp_noauto(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctranu_group_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctranu_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctranu_group_heuristic_pauto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pctranu_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctranu_group_heuristic_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctranu_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pztranu_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int M          =  params[0];
   int N          =  params[1];
   int IA         =  params[1+CNARG+1];
   int JA         =  params[1+CNARG+1+1];
   int *DESCA     = &params[1+CNARG+1+1+1];
   int IC         =  params[1+CNARG+1+1+1+DLEN_+CNARG];
   int JC         =  params[1+CNARG+1+1+1+DLEN_+CNARG+1];
   int *DESCC     = &params[1+CNARG+1+1+1+DLEN_+CNARG+1+1];

   hscaldoublecomplex ALPHA, BETA;

   char TRANSAC   = 'N';
   char TRANSA[2] = { TRANSAC, '\0' };
   
   char TRANSBC   = 'N';
   char TRANSB[2] = { TRANSBC, '\0' };

   hscaldoublecomplex *a, *b, *c;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int reconm           = M/p;
   int reconn           = N/q;
   int reconk            = hscal_pilaenv_recon("Z");
    
   int lda, ldb, ldc;

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   ALPHA.r        = params[1+1];
   ALPHA.i        = params[1+1+1];

   BETA.r         = params[1+CNARG+1+1+1+DLEN_];
   BETA.i         = params[1+CNARG+1+1+1+DLEN_+1];

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "M %d N %d reconm %d reconn %d "
           "ALPHA (%f, %f) IA %d JA %d BETA (%f, %f) IC %d JC %d",
           M, N, reconm, reconn, 
           ALPHA.r, ALPHA.i, IA, JA, BETA.r, BETA.i, IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zgemm routine
   hscal_zgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pztranu_modelp_noauto(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                       = *M;
   (*model_params)[1]                                       = *N;
   (*model_params)[1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                       = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                     = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1]                   = *IC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1]                 = *JC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1+1+DLEN_]         = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1+1+DLEN_+1]       = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+CNARG+1+1+1+i]                     = DESCA[i];
       (*model_params)[1+CNARG+1+1+1+DLEN_+CNARG+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pztranu_modelp(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   *model_count = 1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  CNARG    /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                       = *M;
   (*model_params)[1]                                       = *N;
   (*model_params)[1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN_]                       = betai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+DLEN_+1]                     = betai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1]                   = *IC;
   (*model_params)[1+1+1+1+1+1+DLEN_+1+1+1]                 = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+CNARG+1+1+1+i]                     = DESCA[i];
       (*model_params)[1+CNARG+1+1+1+DLEN_+CNARG+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_recon(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pztranu_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pztranu_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pztranu_timeof(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pztranu_modelp_noauto(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_group_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztranu_modelp_noauto(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_group_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztranu_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_group_heuristic_pauto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pztranu_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_group_heuristic_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztranu_modelp(
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
