
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pctrmm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int SIDEI      =  params[0];
   int UPLOI      =  params[1];
   int TRANSI     =  params[1+1];
   int DIAGI      =  params[1+1+1];
   int M          =  params[1+1+1+1];
   int N          =  params[1+1+1+1+1];
   int IA         =  params[1+1+1+1+1+CNARG+1];
   int JA         =  params[1+1+1+1+1+CNARG+1+1];
   int *DESCA     = &params[1+1+1+1+1+CNARG+1+1+1];
   int IB         =  params[1+1+1+1+1+CNARG+1+1+1+DLEN_];
   int JB         =  params[1+1+1+1+1+CNARG+1+1+1+DLEN_+1];
   int *DESCB     = &params[1+1+1+1+1+CNARG+1+1+1+DLEN_+1+1];

   char SIDEC     = SIDEI;
   char SIDE[2]   = { SIDEC, '\0' };

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };

   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };

   char DIAGC    = DIAGI;
   char DIAG[2]  = { DIAGC, '\0' };

   hscalcomplex ALPHA;
   hscalcomplex *a, *b, *c;
   int reconm, reconn, reconk;
   int lda, ldb, ldc;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;

   ALPHA.r        = params[1+1+1+1+1+1];
   ALPHA.i        = params[1+1+1+1+1+1+1];

   if (TRANSC == 'N')
   {
      reconm  = M/p;
      reconn  = N/q;
      reconk  = hscal_pilaenv_recon("C");
   }
   else
   {
      reconm  = hscal_pilaenv_recon("C");
      reconn  = M/p;
      reconk  = N/q;
   }

   if (TRANSC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "SIDE %s UPLO %s TRANS %s DIAG %s M %d N %d reconm %d reconn %d reconk %d ALPHA (%f, %f) "
           "IA %d JA %d IB %d JB %d",
           SIDE, UPLO, TRANS, DIAG, M, N, reconm, reconn, reconk, ALPHA.r, ALPHA.i,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cgemm routine
   hscal_cgemm_benchmark(
        TRANS,
        TRANS,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &ALPHA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pctrmm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   *model_count = 1        /* SIDE */
                  +        
                  1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *SIDE;
   (*model_params)[1]                                               = *UPLO;
   (*model_params)[1+1]                                             = *TRANS;
   (*model_params)[1+1+1]                                           = *DIAG;
   (*model_params)[1+1+1+1]                                         = *M;
   (*model_params)[1+1+1+1+1]                                       = *N;
   (*model_params)[1+1+1+1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_]                       = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1]                     = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]             = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]           = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+1+i]                       = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]             = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pctrmm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   *model_count = 1        /* SIDE */
                  +        
                  1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *SIDE;
   (*model_params)[1]                                               = *UPLO;
   (*model_params)[1+1]                                             = *TRANS;
   (*model_params)[1+1+1]                                           = *DIAG;
   (*model_params)[1+1+1+1]                                         = *M;
   (*model_params)[1+1+1+1+1]                                       = *N;
   (*model_params)[1+1+1+1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_]                       = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1]                     = *JB;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+1+i]                       = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]             = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctrmm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pctrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pctrmm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pctrmm_timeof(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmm_modelp_noauto(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pctrmm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmm_modelp_noauto(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctrmm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctrmm_group_heuristic_pauto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pctrmm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pctrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pztrmm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int SIDEI      =  params[0];
   int UPLOI      =  params[1];
   int TRANSI     =  params[1+1];
   int DIAGI      =  params[1+1+1];
   int M          =  params[1+1+1+1];
   int N          =  params[1+1+1+1+1];
   int IA         =  params[1+1+1+1+1+CNARG+1];
   int JA         =  params[1+1+1+1+1+CNARG+1+1];
   int *DESCA     = &params[1+1+1+1+1+CNARG+1+1+1];
   int IB         =  params[1+1+1+1+1+CNARG+1+1+1+DLEN_];
   int JB         =  params[1+1+1+1+1+CNARG+1+1+1+DLEN_+1];
   int *DESCB     = &params[1+1+1+1+1+CNARG+1+1+1+DLEN_+1+1];

   char SIDEC     = SIDEI;
   char SIDE[2]   = { SIDEC, '\0' };

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };

   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };

   char DIAGC    = DIAGI;
   char DIAG[2]  = { DIAGC, '\0' };

   hscaldoublecomplex ALPHA;

   hscaldoublecomplex *a, *b, *c;
   int reconm, reconn, reconk;
   int lda, ldb, ldc;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;

   ALPHA.r        = params[1+1+1+1+1+1];
   ALPHA.i        = params[1+1+1+1+1+1+1];

   if (TRANSC == 'N')
   {
      reconm  = M/p;
      reconn  = N/q;
      reconk  = hscal_pilaenv_recon("Z");
   }
   else
   {
      reconm  = hscal_pilaenv_recon("Z");
      reconn  = M/p;
      reconk  = N/q;
   }

   if (TRANSC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "SIDE %s UPLO %s TRANS %s DIAG %s M %d N %d reconm %d reconn %d reconk %d ALPHA (%f, %f) "
           "IA %d JA %d IB %d JB %d",
           SIDE, UPLO, TRANS, DIAG, M, N, reconm, reconn, reconk, ALPHA.r, ALPHA.i,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zgemm routine
   hscal_zgemm_benchmark(
        TRANS,
        TRANS,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &ALPHA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pztrmm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   *model_count = 1        /* SIDE */
                  +        
                  1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *SIDE;
   (*model_params)[1]                                               = *UPLO;
   (*model_params)[1+1]                                             = *TRANS;
   (*model_params)[1+1+1]                                           = *DIAG;
   (*model_params)[1+1+1+1]                                         = *M;
   (*model_params)[1+1+1+1+1]                                       = *N;
   (*model_params)[1+1+1+1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_]                       = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1]                     = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]             = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]           = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+1+i]                       = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]             = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pztrmm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   *model_count = 1        /* SIDE */
                  +        
                  1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  CNARG    /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *SIDE;
   (*model_params)[1]                                               = *UPLO;
   (*model_params)[1+1]                                             = *TRANS;
   (*model_params)[1+1+1]                                           = *DIAG;
   (*model_params)[1+1+1+1]                                         = *M;
   (*model_params)[1+1+1+1+1]                                       = *N;
   (*model_params)[1+1+1+1+1+1]                                     = alphai[REAL_PART];
   (*model_params)[1+1+1+1+1+1+1]                                   = alphai[IMAG_PART];
   (*model_params)[1+1+1+1+1+1+1+1]                                 = *IA;
   (*model_params)[1+1+1+1+1+1+1+1+1]                               = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_]                       = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1]                     = *JB;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+1+i]                       = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]             = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pztrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pztrmm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pztrmm_timeof(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmm_modelp_noauto(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmm_modelp_noauto(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_group_heuristic_pauto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pztrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdtrmm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int SIDEI        =  params[0];
   int UPLOI        =  params[1];
   int TRANSI       =  params[1+1];
   int DIAGI        =  params[1+1+1];
   int M            =  params[1+1+1+1];
   int N            =  params[1+1+1+1+1];
   hscaldoublereal ALPHA =  params[1+1+1+1+1+1];
   int IA           =  params[1+1+1+1+1+1+1];
   int JA           =  params[1+1+1+1+1+1+1+1];
   int *DESCA       = &params[1+1+1+1+1+1+1+1+1];
   int IB           =  params[1+1+1+1+1+1+1+1+1+DLEN_];
   int JB           =  params[1+1+1+1+1+1+1+1+1+DLEN_+1];
   int *DESCB       = &params[1+1+1+1+1+1+1+1+1+DLEN_+1+1];

   char SIDEC     = SIDEI;
   char SIDE[2]   = { SIDEC, '\0' };

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };

   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };

   char DIAGC    = DIAGI;
   char DIAG[2]  = { DIAGC, '\0' };

   double *a, *b, *c;
   int reconm, reconn, reconk;
   int lda, ldb, ldc;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;

   if (TRANSC == 'N')
   {
      reconm  = M/p;
      reconn  = N/q;
      reconk  = hscal_pilaenv_recon("D");
   }
   else
   {
      reconm  = hscal_pilaenv_recon("D");
      reconn  = M/p;
      reconk  = N/q;
   }

   if (TRANSC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "SIDE %s UPLO %s TRANS %s DIAG %s M %d N %d reconm %d reconn %d reconk %d ALPHA %f "
           "IA %d JA %d IB %d JB %d",
           SIDE, UPLO, TRANS, DIAG, M, N, reconm, reconn, reconk, ALPHA,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (double*)malloc(
                 sizeof(double)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (double*)malloc(
                 sizeof(double)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (double*)malloc(
                 sizeof(double)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dgemm routine
   hscal_dgemm_benchmark(
        TRANS,
        TRANS,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &ALPHA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdtrmm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   *model_count = 1        /* SIDE */
                  +        
                  1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *SIDE;
   (*model_params)[1]                                               = *UPLO;
   (*model_params)[1+1]                                             = *TRANS;
   (*model_params)[1+1+1]                                           = *DIAG;
   (*model_params)[1+1+1+1]                                         = *M;
   (*model_params)[1+1+1+1+1]                                       = *N;
   (*model_params)[1+1+1+1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                         = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                       = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]               = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]             = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                         = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]               = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdtrmm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   *model_count = 1        /* SIDE */
                  +        
                  1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *SIDE;
   (*model_params)[1]                                               = *UPLO;
   (*model_params)[1+1]                                             = *TRANS;
   (*model_params)[1+1+1]                                           = *DIAG;
   (*model_params)[1+1+1+1]                                         = *M;
   (*model_params)[1+1+1+1+1]                                       = *N;
   (*model_params)[1+1+1+1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                         = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                       = *JB;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                         = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]               = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdtrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pdtrmm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdtrmm_timeof(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmm_modelp_noauto(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmm_modelp_noauto(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmm_group_heuristic_pauto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdtrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pstrmm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int SIDEI        =  params[0];
   int UPLOI        =  params[1];
   int TRANSI       =  params[1+1];
   int DIAGI        =  params[1+1+1];
   int M            =  params[1+1+1+1];
   int N            =  params[1+1+1+1+1];
   hscalreal ALPHA       =  params[1+1+1+1+1+1];
   int IA           =  params[1+1+1+1+1+1+1];
   int JA           =  params[1+1+1+1+1+1+1+1];
   int *DESCA       = &params[1+1+1+1+1+1+1+1+1];
   int IB           =  params[1+1+1+1+1+1+1+1+1+DLEN_];
   int JB           =  params[1+1+1+1+1+1+1+1+1+DLEN_+1];
   int *DESCB       = &params[1+1+1+1+1+1+1+1+1+DLEN_+1+1];

   char SIDEC     = SIDEI;
   char SIDE[2]   = { SIDEC, '\0' };

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };

   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };

   char DIAGC    = DIAGI;
   char DIAG[2]  = { DIAGC, '\0' };

   float *a, *b, *c;
   int reconm, reconn, reconk;
   int lda, ldb, ldc;

   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;

   if (TRANSC == 'N')
   {
      reconm  = M/p;
      reconn  = N/q;
      reconk  = hscal_pilaenv_recon("S");
   }
   else
   {
      reconm  = hscal_pilaenv_recon("S");
      reconn  = M/p;
      reconk  = N/q;
   }

   if (TRANSC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "SIDE %s UPLO %s TRANS %s DIAG %s M %d N %d reconm %d reconn %d reconk %d ALPHA %f "
           "IA %d JA %d IB %d JB %d",
           SIDE, UPLO, TRANS, DIAG, M, N, reconm, reconn, reconk, ALPHA,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (float*)malloc(
                 sizeof(float)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (float*)malloc(
                 sizeof(float)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (float*)malloc(
                 sizeof(float)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sgemm routine
   hscal_sgemm_benchmark(
        TRANS,
        TRANS,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &ALPHA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pstrmm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   *model_count = 1        /* SIDE */
                  +        
                  1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *SIDE;
   (*model_params)[1]                                               = *UPLO;
   (*model_params)[1+1]                                             = *TRANS;
   (*model_params)[1+1+1]                                           = *DIAG;
   (*model_params)[1+1+1+1]                                         = *M;
   (*model_params)[1+1+1+1+1]                                       = *N;
   (*model_params)[1+1+1+1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                         = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                       = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]               = *nprow;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]             = *npcol;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                         = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]               = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pstrmm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   *model_count = 1        /* SIDE */
                  +        
                  1        /* UPLO */
                  +        
                  1        /* TRANS */
                  +        
                  1        /* DIAG */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *SIDE;
   (*model_params)[1]                                               = *UPLO;
   (*model_params)[1+1]                                             = *TRANS;
   (*model_params)[1+1+1]                                           = *DIAG;
   (*model_params)[1+1+1+1]                                         = *M;
   (*model_params)[1+1+1+1+1]                                       = *N;
   (*model_params)[1+1+1+1+1+1]                                     = alphai;
   (*model_params)[1+1+1+1+1+1+1]                                   = *IA;
   (*model_params)[1+1+1+1+1+1+1+1]                                 = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_]                         = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1]                       = *JB;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+1+i]                         = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+1+DLEN_+1+1+i]               = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstrmm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pstrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pstrmm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pstrmm_timeof(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmm_modelp_noauto(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pstrmm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmm_modelp_noauto(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstrmm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstrmm_group_heuristic_pauto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pstrmm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pstrmm_modelp(
                SIDE,
                UPLO,
                TRANS,
                DIAG,
                M,
                N,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
