/*
* PBLAS level-2 PZTRMV routine context
* 
* Ravi Reddy
* Version 1.0
* School of Computer Science and Informatics, 
* UCD, 
* Dublin,
* August 21, 2007
*/

/*
*  Include files
*/
#include "../../include/hscalapack.h"

int hscal_pztrmv_ctxt(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int      * N,
   int      * IA,
   int      * JA,
   int      * DESCA,
   int      * IX,
   int      * JX,
   int      * DESCX,
   int      * INCX,
   int      * ICTXT
)
{
/*
*  Purpose
*  =======
*
*  PZTRMV  performs one of the matrix-vector operations
*
*  sub( X ) := sub( A )*sub( X )   or   sub( X ) := sub( A )'*sub( X )
*
*  or
*
*  sub( X ) := conjg( sub( A )' )*sub( X ),
*
*  where
*
*     sub( A ) denotes A(IA:IA+N-1,JA:JA+N-1), and,
*
*     sub( X ) denotes X(IX,JX:JX+N-1) if INCX = M_X,
*                      X(IX:IX+N-1,JX) if INCX = 1 and INCX <> M_X.
*
*  sub( X )  is an n element subvector and  sub( A ) is an n by n  unit,
*  or non-unit, upper or lower triangular submatrix.
*
*  Notes
*  =====
*
*  A description  vector  is associated with each 2D block-cyclicly dis-
*  tributed matrix.  This  vector  stores  the  information  required to
*  establish the  mapping  between a  matrix entry and its corresponding
*  process and memory location.
*
*  In  the  following  comments,   the character _  should  be  read  as
*  "of  the  distributed  matrix".  Let  A  be a generic term for any 2D
*  block cyclicly distributed matrix.  Its description vector is DESC_A:
*
*  NOTATION         STORED IN       EXPLANATION
*  ---------------- --------------- ------------------------------------
*  DTYPE_A (global) DESCA[ DTYPE_ ] The descriptor type.
*  CTXT_A  (global) DESCA[ CTXT_  ] The BLACS context handle, indicating
*                                   the NPROW x NPCOL BLACS process grid
*                                   A  is  distributed over. The context
*                                   itself  is  global,  but  the handle
*                                   (the integer value) may vary.
*  M_A     (global) DESCA[ M_     ] The  number of rows in the distribu-
*                                   ted matrix A, M_A >= 0.
*  N_A     (global) DESCA[ N_     ] The number of columns in the distri-
*                                   buted matrix A, N_A >= 0.
*  IMB_A   (global) DESCA[ IMB_   ] The number of rows of the upper left
*                                   block of the matrix A, IMB_A > 0.
*  INB_A   (global) DESCA[ INB_   ] The  number  of columns of the upper
*                                   left   block   of   the  matrix   A,
*                                   INB_A > 0.
*  MB_A    (global) DESCA[ MB_    ] The blocking factor used to  distri-
*                                   bute the last  M_A-IMB_A  rows of A,
*                                   MB_A > 0.
*  NB_A    (global) DESCA[ NB_    ] The blocking factor used to  distri-
*                                   bute the last  N_A-INB_A  columns of
*                                   A, NB_A > 0.
*  RSRC_A  (global) DESCA[ RSRC_  ] The process row over which the first
*                                   row of the matrix  A is distributed,
*                                   NPROW > RSRC_A >= 0.
*  CSRC_A  (global) DESCA[ CSRC_  ] The  process column  over  which the
*                                   first column of  A  is  distributed.
*                                   NPCOL > CSRC_A >= 0.
*  LLD_A   (local)  DESCA[ LLD_   ] The  leading dimension  of the local
*                                   array  storing  the  local blocks of
*                                   the distributed matrix A,
*                                   IF( Lc( 1, N_A ) > 0 )
*                                      LLD_A >= MAX( 1, Lr( 1, M_A ) )
*                                   ELSE
*                                      LLD_A >= 1.
*
*  Let K be the number of  rows of a matrix A starting at the global in-
*  dex IA,i.e, A( IA:IA+K-1, : ). Lr( IA, K ) denotes the number of rows
*  that the process of row coordinate MYROW ( 0 <= MYROW < NPROW ) would
*  receive if these K rows were distributed over NPROW processes.  If  K
*  is the number of columns of a matrix  A  starting at the global index
*  JA, i.e, A( :, JA:JA+K-1, : ), Lc( JA, K ) denotes the number  of co-
*  lumns that the process MYCOL ( 0 <= MYCOL < NPCOL ) would  receive if
*  these K columns were distributed over NPCOL processes.
*
*  The values of Lr() and Lc() may be determined via a call to the func-
*  tion PB_Cnumroc:
*  Lr( IA, K ) = PB_Cnumroc( K, IA, IMB_A, MB_A, MYROW, RSRC_A, NPROW )
*  Lc( JA, K ) = PB_Cnumroc( K, JA, INB_A, NB_A, MYCOL, CSRC_A, NPCOL )
*
*  Arguments
*  =========
*
*  UPLO    (global input) CHARACTER*1
*          On entry,  UPLO  specifies whether the submatrix  sub( A ) is
*          an upper or lower triangular submatrix as follows:
*
*             UPLO = 'U' or 'u'   sub( A ) is an upper triangular
*                                 submatrix,
*
*             UPLO = 'L' or 'l'   sub( A ) is a  lower triangular
*                                 submatrix.
*
*  TRANS   (global input) CHARACTER*1
*          On entry,  TRANS  specifies the  operation to be performed as
*          follows:
*
*             TRANS = 'N' or 'n'   sub( X ) := sub( A )  * sub( X ).
*
*             TRANS = 'T' or 't'   sub( X ) := sub( A )' * sub( X ).
*
*             TRANS = 'C' or 'c'
*                         sub( X ) := conjg( sub( A )' ) * sub( X ).
*
*  DIAG    (global input) CHARACTER*1
*          On entry,  DIAG  specifies  whether or not  sub( A )  is unit
*          triangular as follows:
*
*             DIAG = 'U' or 'u'  sub( A )  is  assumed to be unit trian-
*                                gular,
*
*             DIAG = 'N' or 'n'  sub( A ) is not assumed to be unit tri-
*                                angular.
*
*  N       (global input) INTEGER
*          On entry,  N specifies the order of the  submatrix  sub( A ).
*          N must be at least zero.
*
*  IA      (global input) INTEGER
*          On entry, IA  specifies A's global row index, which points to
*          the beginning of the submatrix sub( A ).
*
*  JA      (global input) INTEGER
*          On entry, JA  specifies A's global column index, which points
*          to the beginning of the submatrix sub( A ).
*
*  DESCA   (global and local input) INTEGER array
*          On entry, DESCA  is an integer array of dimension DLEN_. This
*          is the array descriptor for the matrix A.
*
*  IX      (global input) INTEGER
*          On entry, IX  specifies X's global row index, which points to
*          the beginning of the submatrix sub( X ).
*
*  JX      (global input) INTEGER
*          On entry, JX  specifies X's global column index, which points
*          to the beginning of the submatrix sub( X ).
*
*  DESCX   (global and local input) INTEGER array
*          On entry, DESCX  is an integer array of dimension DLEN_. This
*          is the array descriptor for the matrix X.
*
*  INCX    (global input) INTEGER
*          On entry,  INCX   specifies  the  global  increment  for  the
*          elements of  X.  Only two values of  INCX   are  supported in
*          this version, namely 1 and M_X. INCX  must not be zero.
*
*
*  ---------------------------------------------------------------------
*/
/*
*  .. Local Scalars ..
*/
    int            rc, p, q;
    HMPI_Group     *gid;
    MPI_Comm       *pztrmvcomm;
    HSCALCONTEXT   *hctxt; 
    double         algotime;
   
/*
* Initialize the context ID
*/
    *ICTXT = HSCAL_CTXT_UNDEFINED;

/*
* Dynamically refresh the speeds of the processors
*/
    rc = hscal_pztrmv_recon(
              UPLO,
              TRANS,
              DIAG,
              N,
              IA,
              JA,
              DESCA,
              IX,
              JX,
              DESCX,
              INCX
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_RECON;
    }

/*
* Create a HeteroMPI group with the optimal grid arrangement 
* of processes
*/
    gid  = (HMPI_Group*)malloc(sizeof(HMPI_Group));

    if (gid == NULL)
    {
       return HSCAL_ERR_LOW_MEMORY;
    }

    rc = hscal_pztrmv_group_heuristic_pauto_create(
              UPLO,
              TRANS,
              DIAG,
              N,
              IA,
              JA,
              DESCA,
              IX,
              JX,
              DESCX,
              INCX,
              gid,
              hscal_model_pztrmv,
              hscal_heuristic_fn_lean
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_GROUP_CREATION;
    }

    if (HMPI_Is_free())
    {
       *ICTXT = HSCAL_CTXT_FOR_FREE_PROCS;

       hctxt  = (HSCALCONTEXT*)malloc(sizeof(HSCALCONTEXT));

       if (hctxt == NULL)
       {
          return HSCAL_ERR_LOW_MEMORY;
       }

       hctxt->time = -1.;
       hctxt->optp = -1;
       hctxt->optq = -1;
       hctxt->gid  = gid;

       rc = hscal_store_gid_to_ctxt_mapping(ICTXT, hctxt);

       return rc;
    }

    pztrmvcomm = (MPI_Comm*)HMPI_Get_comm(gid);

    if (pztrmvcomm == NULL)
    {
       return HSCAL_ERR_NULL_COMM; 
    } 

/* 
* Broadcast the estimated execution time 
*/ 
    algotime = HMPI_Group_timeof(gid);
/*
* Broadcast the optimal grid arrangement
*/
    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       p = d[0];

       free(d);

       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *pztrmvcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *pztrmvcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       q = d[1];

       free(d);

       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *pztrmvcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *pztrmvcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       HSCAL_DEBUG(
            __FILE__, __FUNCTION__, __LINE__,
            "Optimal number of process rows =%d, Optimal number of process columns =%d, " 
            "Estimated execution time=%f ",
            p, q, algotime
       );       
    }

/*
* Translate communicator to a BLACS handle
*/
    *ICTXT = Csys2blacs_handle(*pztrmvcomm);

/*
* Form BLACS context
*/
    Cblacs_gridinit(ICTXT, "r", p, q);

/*
* Store the context to gid mapping
*/
    hctxt  = (HSCALCONTEXT*)malloc(sizeof(HSCALCONTEXT)); 
 
    if (hctxt == NULL) 
    { 
       return HSCAL_ERR_LOW_MEMORY; 
    } 
 
    hctxt->time = algotime; 
    hctxt->optp = p; 
    hctxt->optq = q; 
    hctxt->gid  = gid; 
 
    rc = hscal_store_gid_to_ctxt_mapping(ICTXT, hctxt); 

    return rc;
}
