
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcamax_recon_tcomp(
   int n,
   int incx
)
{
    double tcomp = 0;

    int num_of_computers = HMPI_Get_number_of_computers();
    int reconn = (((n/num_of_computers) >= 1) ? (n/num_of_computers) : 1);
    
    hscal_icamax(
         &reconn,
         &incx,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pcamax_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pcamax_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pcamax_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pcamax_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcamax_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pcamax_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pcamax_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           cbtop, cctop, rbtop, rctop;
   int            Xcol, Xgindx, Xi, Xii, Ximb, Xinb, Xj, Xjj, Xlindx, Xld, Xmb,
                  Xnb, Xnp, Xnq, Xrow, Xsrc, ctxt, dist, dst, idumm, info, k,
                  maxpos, mydist, src, size;
   HSCALBLACSCONTEXT   * blacsctxt;
   PBTYP_T        * type;
/*
*  .. Local Arrays ..
*/
   int            Xd[DLEN_];

   double         * tcomp_l  =  tcomp;
   int            * tcomm_l  =  tcomm;

/* ..
*  .. Executable Statements ..
*
*/
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );

   if (init_llds_f == 1)
   {
      HSCAL_PXAMAX_INIT_LLDS(
           N,
           Xd, IX, INCX, 
           myrow, nprow
      );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 701 + CTXT_ ) : 0 ) ) )
   {
      hscal_PB_Cchkvec(
           "PCAMAX", "X", N, 1, Xi, Xj, Xd, INCX, 7,
           nprow, npcol, myrow, mycol,
           &info
      );
   }

   if( info )
   {
      hscal_PB_Cabort(
           "PCAMAX", 
           info,
           nprow, npcol, myrow, mycol
      );

      return HSCAL_ERR_INPUT; 
   }

/*
*  Quick return if possible
*/
   if( N == 0 ) return HSCAL_SUCCESS;
/*
*  Retrieve process grid information
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*
*  Retrieve sub( X )'s local information: Xii, Xjj, Xrow, Xcol
*/
   PB_Cinfog2l( Xi, Xj, Xd, nprow, npcol, myrow, mycol, &Xii, &Xjj,
                &Xrow, &Xcol );
/*
*  Handle degenerate case separately, sub( X )'s scope is just one process
*/
   if( ( INCX == 1 ) && ( Xd[M_] == 1 ) && ( N == 1 ) )
   {
/*
*  Make sure I own some data and compute INDX and AMAX
*/
      if( ( ( myrow == Xrow ) || ( Xrow < 0 ) ) &&
          ( ( mycol == Xcol ) || ( Xcol < 0 ) ) )
      {
         if (tcomp_l != NULL)
         {
            TCOMPL((myrow), (mycol), (nprow), (npcol)) += 2;
         }
      }
   }
   else if( INCX == Xd[M_] )
   {
/*
*  sub( X ) resides in (a) process row(s)
*/
      if( ( myrow == Xrow ) || ( Xrow < 0 ) )
      {
         rctop = *PB_Ctop( &ctxt, COMBINE, ROW, TOP_GET );

         if( ( rctop == CTOP_DEFAULT ) || ( rctop == CTOP_TREE1 ) )
         {
/*
*  Inline the 1-tree combine for communication savings
*/
            Xinb = Xd[INB_ ]; Xnb = Xd[NB_ ]; Xsrc = Xd[CSRC_];
            Xnq = PB_Cnumroc( N, Xj, Xinb, Xnb, mycol, Xsrc, npcol );
/*
*  Make sure I own some data and compute local INDX and AMAX
*/
            if( Xnq > 0 )
            {
               Xld = Xd[LLD_];

               icamax_tcomp_macro( 
                     &Xnq, &Xld,
                     nprow, npcol, myrow, mycol, tcomp_l
               );
            }
            else
            {
               ;
            }
/*
*  Combine the local results using a 1-tree topology within process column 0
*  if npcol > 1 or Xcol >= 0, i.e sub( X ) is distributed.
*/
            if( ( npcol >= 2 ) && ( Xcol >= 0 ) )
            {
               mydist = mycol;
               k      = 1;
l_10:
               if( mydist & 1 )
               {
                  dist = k * ( mydist - 1 );
                  dst  = MPosMod( dist, npcol );
                  hscal_Ccgesd2d( 
                       blacsctxt, 2, 1, 2, myrow, dst,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
                  goto l_20;
               }
               else
               {
                  dist = mycol + k;
                  src  = MPosMod( dist, npcol );

                  if( mycol < src )
                  {
                     ;
                  }
                  mydist >>= 1;
               }
               k <<= 1;

               if( k < npcol ) goto l_10;
l_20:
/*
*  Process column 0 broadcasts the combined values of INDX and AMAX within
*  their process row.
*/
               rbtop = *PB_Ctop( &ctxt, BCAST, ROW, TOP_GET );
               if( mycol == 0 )
               {
                  hscal_Ccgebs2d( 
                       blacsctxt, ROW, &rbtop, 2, 1, 2,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
               }
               else
               {
                  hscal_Ccgebr2d( 
                       blacsctxt, ROW, &rbtop, 2, 1, 2,
                       myrow, 0,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
               }
            }
/*
*  Set INDX and AMAX to the replicated answers contained in work. If AMAX is
*  zero, then select a coherent INDX.
*/

         }
         else
         {
/*
*  Otherwise use the current topology settings to combine the results
*/
            Xinb = Xd[INB_ ]; Xnb = Xd[NB_ ]; Xsrc = Xd[CSRC_];
            Xnq = PB_Cnumroc( N, Xj, Xinb, Xnb, mycol, Xsrc, npcol );
/*
*  Make sure I own some data and compute local INDX and AMAX
*/
            if( Xnq > 0 )
            {
/*
*  Compute the local maximum and its corresponding local index
*/
               Xld = Xd[LLD_];

               icamax_tcomp_macro( 
                     &Xnq, &Xld,
                     nprow, npcol, myrow, mycol, tcomp_l
               );
            }
            else
            {
               ;
            }

            if( Xcol >= 0 )
            {
/*
*  Combine leave on all the local maximum if Xcol >= 0, i.e sub( X ) is
*  distributed
*/
               hscal_Ccgamx2d( 
                    blacsctxt, ROW, &rctop, 1, 1, 1,
                    &idumm, &maxpos, 1, -1, mycol,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );

/*
*  Broadcast the corresponding global index
* 
*  We assume the worse case of communications happening at any cost
*/
               {
                  rbtop = *PB_Ctop( &ctxt, BCAST, ROW, TOP_GET );
                  if( mycol == maxpos )
                  {
                     hscal_Cigebs2d( 
                          blacsctxt, ROW, &rbtop, 1, 1, 1,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  }
                  else
                  {
                     hscal_Cigebr2d( 
                          blacsctxt, ROW, &rbtop, 1, 1, 1,
                          myrow, maxpos,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  }
               }
            }
            else
            {
/*
*  sub( X ) is not distributed. If AMAX is zero, then select a coherent INDX.
*/
               ;
            }
         }
      }
   }
   else
   {
/*
*  sub( X ) resides in (a) process column(s)
*/
      if( ( mycol == Xcol ) || ( Xcol < 0 ) )
      {
         cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );

         if( ( cctop == CTOP_DEFAULT ) || ( cctop == CTOP_TREE1 ) )
         {
/*
*  Inline the 1-tree combine for communication savings
*/
            Ximb = Xd[IMB_ ]; Xmb = Xd[MB_ ]; Xsrc = Xd[RSRC_];
            Xnp = PB_Cnumroc( N, Xi, Ximb, Xmb, myrow, Xsrc, nprow );
/*
*  Make sure I own some data and compute local INDX and AMAX
*/
            if( Xnp > 0 )
            {
               Xld     = Xd[LLD_];

               icamax_tcomp_macro( 
                     &Xnp, &INCX,
                     nprow, npcol, myrow, mycol, tcomp_l
               );
            }
            else
            {
               ;
            }
/*
*  Combine the local results using a 1-tree topology within process row 0
*  if nprow > 1 or Xrow >= 0, i.e sub( X ) is distributed.
*/
            if( ( nprow >= 2 ) && ( Xrow >= 0 ) )
            {
               mydist = myrow;
               k      = 1;
l_30:
               if( mydist & 1 )
               {
                  dist = k * ( mydist - 1 );
                  dst  = MPosMod( dist, nprow );
                  hscal_Ccgesd2d( 
                       blacsctxt, 2, 1, 2, dst, mycol,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
                  goto l_40;
               }
               else
               {
                  dist = myrow + k;
                  src  = MPosMod( dist, nprow );

                  if( myrow < src )
                  {
                     ;
                  }
                  mydist >>= 1;
               }
               k <<= 1;

               if( k < nprow ) goto l_30;
l_40:
/*
*  Process row 0 broadcasts the combined values of INDX and AMAX within their
*  process column.
*/
               cbtop = *PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );
               if( myrow == 0 )
               {
                  hscal_Ccgebs2d( 
                       blacsctxt, COLUMN, &cbtop, 2, 1, 2,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
               }
               else
               {
                  hscal_Ccgebr2d( 
                       blacsctxt, COLUMN, &cbtop, 2, 1, 2,
                       0, mycol,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
               }
            }
/*
*  Set INDX and AMAX to the replicated answers contained in work. If AMAX is
*  zero, then select a coherent INDX.
*/

         }
         else
         {
/*
*  Otherwise use the current topology settings to combine the results
*/
            Ximb = Xd[IMB_ ]; Xmb = Xd[MB_ ]; Xsrc = Xd[RSRC_];
            Xnp = PB_Cnumroc( N, Xi, Ximb, Xmb, myrow, Xsrc, nprow );
/*
*  Make sure I own some data and compute local INDX and AMAX
*/

            if( Xnp > 0 )
            {
/*
*  Compute the local maximum and its corresponding local index
*/
               Xld = Xd[LLD_];

               icamax_tcomp_macro( 
                     &Xnp, &INCX,
                     nprow, npcol, myrow, mycol, tcomp_l
               );
            }
            else
            {
               ;
            }

            if( Xrow >= 0 )
            {
/*
*  Combine leave on all the local maximum if Xrow >= 0, i.e sub( X ) is
*  distributed.
*/
               hscal_Ccgamx2d( 
                    blacsctxt, COLUMN, &cctop, 1, 1, 1,
                    &maxpos, &idumm, 1, -1, mycol,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
/*
*  Broadcast the corresponding global index
* 
*  We assume the worse case of communications happening at any cost
*/

               {
                  cbtop = *PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );
                  if( myrow == maxpos )
                  {
                     hscal_Cigebs2d( 
                          blacsctxt, COLUMN, &cbtop, 1, 1, 1,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  }
                  else
                  {
                     hscal_Cigebr2d( 
                          blacsctxt, COLUMN, &cbtop, 1, 1, 1,
                          maxpos, mycol,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  }
               }
            }
            else
            {
/*
*  sub( X ) is not distributed. If AMAX is zero, then select a coherent INDX.
*/
               ;
            }
         }
      }
   }

   Cblacs_gridexit_nocomm(blacsctxt);

/*
*  End of PCAMAX
*/
   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcamax_tcomp_tcomm(
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcamax_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N, 
                IX, JX, DESCX, INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcamax_info(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcamax_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcamax_info_i(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcamax_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
