
#include <stdlib.h>
#include <stdio.h>

#include "../../include/scalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcgeru_recon_tcomp(
   int m,
   int n,
   int alphai[CNARG],
   int incx,
   int incy
)
{
    double tcomp = 0;

    int reconm   = hscal_pilaenv_recon("C");
    int reconn   = hscal_pilaenv_recon("C");
    int lda      = reconm;

    float alpha[CNARG];

    alpha[REAL_PART]  = alphai[REAL_PART];
    alpha[IMAG_PART]  = alphai[IMAG_PART];
    
    hscal_cgeru(
         &reconm,
         &reconn,
         ((char*)alpha),
         &incx,
         &incy,
         &lda,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pcgeru_my_tcomp(
   int myrow,
   int mycol,
   int M,
   int N,
   int ALPHA[CNARG],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pcgeru_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N, 
          ALPHA, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          IA, JA, DESCA, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pcgeru_my_tcomm(
   int myrow,
   int mycol,
   int M,
   int N,
   int ALPHA[CNARG],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pcgeru_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N, 
          ALPHA, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          IA, JA, DESCA, 
          nprow, npcol,
          NULL, tcomm
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgeru_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int M,
   int N,
   int ALPHA[CNARG],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pcgeru_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N, 
          ALPHA, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          IA, JA, DESCA, 
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pcgeru_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int M,
   int N,
   int ALPHAI[CNARG],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   int            Acol, Ai, Aii, Aimb1, Ainb1, Aj, Ajj, Ald, Amb, Amp, Anb,
                  Anq, Arow, XAfr, Xi, Xj, YAfr, Yi, Yj, ctxt, info, ione=1;
   PBTYP_T        * type;
/*
*  .. Local Arrays ..
*/
   int            Ad[DLEN_], Ad0[DLEN_], XAd[DLEN_], Xd[DLEN_], YAd[DLEN_],
                  Yd[DLEN_];
   float ALPHA[CNARG];

   ALPHA[REAL_PART]       = ALPHAI[REAL_PART];
   ALPHA[IMAG_PART]       = ALPHAI[IMAG_PART];

/* ..
*  .. Executable Statements ..
*
*/
   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
   PB_CargFtoC( IY, JY, DESCY, &Yi, &Yj, Yd );

   if (init_llds_f == 1)
   {
      HSCAL_PXGERU_INIT_LLDS(
           N, 
           Ad, 
           Xd, IX, INCX, 
           Yd, IY, INCY, 
           myrow, nprow
      );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d",
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Yd[M_]=%d Yd[MB_]=%d myrow=%d Yd[RSRC_]=%d nprow=%d Yd[LLD_]=%d", 
        Yd[M_], Yd[MB_], myrow, Yd[RSRC_], nprow, Yd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 701 + CTXT_ ) : 0 ) ) )
   {
      hscal_PB_Cchkvec( 
           "PCGERU", "X", M, 1, Xi, Xj, Xd, INCX,  7, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec( 
           "PCGERU", "Y", N, 2, Yi, Yj, Yd, INCY, 12, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkmat( 
           "PCGERU", "A", M, 1, N, 2, Ai, Aj, Ad, 17, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PCGERU",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( ( M == 0 ) || ( N == 0 ) ||
       ( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) ) )
   {
      return HSCAL_SUCCESS;
   }

/*
*  Get type structure
*/
   type = PB_Cctypeset();

/*
*  Compute descriptor Ad0 for sub( A )
*/
   PB_Cdescribe( M, N, Ai, Aj, Ad, nprow, npcol, myrow, mycol, &Aii, &Ajj,
                 &Ald, &Aimb1, &Ainb1, &Amb, &Anb, &Arow, &Acol, Ad0 );

/*
*  Replicate sub( X ) in process columns spanned by sub( A ) -> XA
*/
   {
   char *CONJUG_l = NOCONJG;
   char *ROWCOL_l = COLUMN; 
   int K_l = 1;
   char *XROC_l = ( INCX == Xd[M_] ? ROW : COLUMN );

   hscal_PB_CInV_C(
          type, CONJUG_l, ROWCOL_l, M, N, Ad0, K_l, Xi, Xj, Xd,
          XROC_l, XAd,
          nprow, npcol, myrow, mycol, tcomp, tcomm
   );
   }
/*
*  Replicate sub( Y ) in process rows spanned by sub( A ) -> YA
*/
   {
   char *CONJUG_l = NOCONJG;
   char *ROWCOL_l = ROW; 
   int K_l = 1;
   char *YROC_l = ( INCY == Yd[M_] ? ROW : COLUMN );

   hscal_PB_CInV_C( 
          type, CONJUG_l, ROWCOL_l,    M, N, Ad0, K_l, Yi, Yj, Yd,
          YROC_l, YAd,
          nprow, npcol, myrow, mycol, tcomp, tcomm
   );
   }
/*
*  Local rank-1 update iff I own some data
*/
   Amp = PB_Cnumroc( M, 0, Aimb1, Amb, myrow, Arow, nprow );
   Anq = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );

   if( ( Amp > 0 ) && ( Anq > 0 ) )
   {
      cgeru_tcomp_macro( 
           &Amp, &Anq, ((char *) ALPHA), &ione, &YAd[LLD_],
           &Ald,
           nprow, npcol, myrow, mycol, tcomp
      );
   }
/*
*  End of PCGERU
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgeru_tcomp_tcomm(
   int M,
   int N,
   int ALPHA[CNARG],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcgeru_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                M, N, 
                ALPHA,
                IX, JX, DESCX, INCX,
                IY, JY, DESCY, INCY,
                IA, JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgeru_info(
   int * M,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcgeru_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N, 
                alphai,
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgeru_info_i(
   int * M,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcgeru_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N, 
                alphai,
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
