
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcher2k_recon_tcomp(
   int uploi,
   int transi,
   int n,
   int k, 
   int alphai[CNARG],
   int betai
)
{
    double tcomp = 0;

    char   uploc   = uploi;
    char   *uplo   = &uploc;

    char   transc  = transi;
    char   *trans  = &transc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    float alpha[CNARG], beta = betai;

    alpha[REAL_PART] = alphai[REAL_PART];
    alpha[IMAG_PART] = alphai[IMAG_PART];

    if (transc == 'N')
    {
       reconm  = n/p;
       reconn  = hscal_pilaenv_recon("C");
       reconk  = k/q;
    }
    else
    {
       reconm  = hscal_pilaenv_recon("C");
       reconn  = n/p;
       reconk  = k/q;
    }

    if (transc == 'N')
    {
       lda = reconm;
    }
    else
    {
       lda = reconk;
    }

    if (transc == 'N')
    {
       ldb = reconk;
    }
    else
    {
       ldb = reconn;
    }

    ldc = reconm;

    hscal_cgemm(
         trans,
         trans,
         &reconm,
         &reconn,
         &reconk,
         ((char*)alpha),
         &lda,
         &ldb,
         ((char*)alpha),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pcher2k_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int N,
   int K,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pcher2k_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI,
          N, K,
          ALPHA, 
          IA, JA, DESCA, 
          IB, JB, DESCB, 
          BETA,
          IC, JC, DESCC, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pcher2k_my_tcomm(
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int N,
   int K,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pcher2k_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI,
          N, K,
          ALPHA, 
          IA, JA, DESCA,
          IB, JB, DESCB,
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcher2k_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLOI,
   int TRANSI,
   int N,
   int K,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pcher2k_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI,
          N, K,
          ALPHA, 
          IA, JA, DESCA,
          IB, JB, DESCB,
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pcher2k_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int N,
   int K,
   int ALPHAI[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETAI,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           DirA, OpC, OpR, TopC, TopR, TranOp, UploC, ctop, ctopsave,
                  rtop, rtopsave;
   int            Ai, Aj, Bi, Bj, ChooseABC, Ci, Cj, ForceTop, ctxt, info,
                  nb, notran, upper;
   double         ABCest, ABest, tmp1, tmp2, tmp3, tmp4;
   PBTYP_T        * type;
/*
*  .. Local Arrays ..
*/
   int            Ad[DLEN_], Bd[DLEN_], Cd[DLEN_];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

   char UPLOC     = UPLOI, *UPLO = &UPLOC;
   char TRANSC    = TRANSI, *TRANS = &TRANSC;

   float ALPHA[CNARG], BETA = BETAI;

   ALPHA[REAL_PART]       = ALPHAI[REAL_PART];
   ALPHA[IMAG_PART]       = ALPHAI[IMAG_PART];

/* ..
*  .. Executable Statements ..
*
*/
   upper  = ( ( UploC  = Mupcase( F2C_CHAR( UPLO  )[0] ) ) ==  CUPPER );
   notran = ( ( TranOp = Mupcase( F2C_CHAR( TRANS )[0] ) ) == CNOTRAN );

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IB, JB, DESCB, &Bi, &Bj, Bd );
   PB_CargFtoC( IC, JC, DESCC, &Ci, &Cj, Cd );

   if (init_llds_f == 1)
   {
      HSCAL_PXHER2K_INIT_LLDS(
           Ad,
           Bd,
           Cd,
           myrow, nprow
      );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Bd[M_]=%d Bd[MB_]=%d myrow=%d Bd[RSRC_]=%d nprow=%d Bd[LLD_]=%d",
        Bd[M_], Bd[MB_], myrow, Bd[RSRC_], nprow, Bd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Cd[M_]=%d Cd[MB_]=%d myrow=%d Cd[RSRC_]=%d nprow=%d Cd[LLD_]=%d", 
        Cd[M_], Cd[MB_], myrow, Cd[RSRC_], nprow, Cd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 901 + CTXT_ ) : 0 ) ) )
   {
      if( ( !upper ) && ( UploC != CLOWER ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PCHER2K", "Illegal UPLO = %c\n", UploC 
         );
         info = -1;
      }
      else if( ( !notran ) && ( TranOp != CCOTRAN ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PCHER2K", "Illegal TRANS = %c\n", TranOp 
         );
         info = -2;
      }
      if( notran )
      {
         hscal_PB_Cchkmat( 
              "PCHER2K", "A", N, 3, K, 4, Ai, Aj, Ad,  9,
              nprow, npcol, myrow, mycol,
              &info
         );
         hscal_PB_Cchkmat( 
              "PCHER2K", "B", N, 3, K, 4, Bi, Bj, Bd, 13,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      else
      {
         hscal_PB_Cchkmat( 
              "PCHER2K", "A", K, 4, N, 3, Ai, Aj, Ad,  9,
              nprow, npcol, myrow, mycol,
              &info 
         );
         hscal_PB_Cchkmat( 
              "PCHER2K", "B", K, 4, N, 3, Bi, Bj, Bd, 13,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      hscal_PB_Cchkmat(    
           "PCHER2K", "C", N, 3, N, 3, Ci, Cj, Cd, 18,
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PCHER2K",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( ( N == 0 ) ||
       ( ( ( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) ) ||
           ( K == 0                      ) ) &&
         ( ( BETA == ONE )       ) ) )
   {
      return HSCAL_SUCCESS;
   }

/*
*  Get type structure
*/
   type = PB_Cctypeset();

/*
*  And when alpha or K is zero
*/
   if( ( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) ) ||
       ( K == 0 ) )
   {
      if( BETA == ZERO )
      {
         hscal_PB_Cplapad_C( 
           type, &UploC, NOCONJG, N, N, type->zero, type->zero,
           Ci, Cj, Cd,
           nprow, npcol, myrow, mycol, tcomp_l
         );
      }
      else
      {
         hscal_PB_Cplascal_C( 
           type, &UploC, CONJG,   N, N, ((char *) &BETA),
           Ci, Cj, Cd,
           nprow, npcol, myrow, mycol, tcomp_l
         );
      }

      return HSCAL_SUCCESS;
   }

/*
*  Algorithm selection is based on approximation of the communication volume
*  for distributed and aligned operands.
*
*  ABCest: operands sub( A ), sub( B ) and sub( C ) are communicated (K >> N)
*  ABest : only sub( A ) and sub( B ) are communicated               (N >> K)
*/
   if( notran )
   {
      tmp1   = DNROC( N, Cd[MB_], nprow );
      tmp3   = DNROC( K, Ad[NB_], npcol ); tmp4 = DNROC( K, Bd[NB_], npcol );
      ABCest = (double)(N) *
        ( ( ( ( Ad[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp3 ) +
          ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp4 ) +
          ( ( ( ( Ad[CSRC_] == -1 ) && ( Bd[CSRC_] == -1 ) ) ||
              ( npcol == 1 ) ) ? ZERO : CBRATIO * tmp1 / TWO ) );
      tmp1   = DNROC( N, Cd[MB_], nprow ); tmp2 = DNROC( N, Cd[NB_], npcol );
      tmp3   = DNROC( N, Ad[MB_], nprow ); tmp4 = DNROC( N, Bd[MB_], nprow );
      ABest  = (double)(K) *
               ( ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 ) +
                 ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 ) +
                 TWO * ( nprow == 1 ? ZERO : tmp2 ) +
                 MAX( tmp2, tmp3 ) + MAX( tmp2, tmp4 ) );
   }
   else
   {
      tmp2   = DNROC( N, Cd[NB_], npcol );
      tmp3   = DNROC( K, Ad[MB_], nprow ); tmp4 = DNROC( K, Bd[MB_], nprow );
      ABCest = (double)(N) *
        ( ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp3 ) +
          ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp4 ) +
          ( ( ( ( Ad[RSRC_] == -1 ) && ( Bd[RSRC_] == -1 ) ) ||
              ( nprow == 1 ) ) ? ZERO : CBRATIO * tmp2 / TWO ) );
      tmp1   = DNROC( N, Cd[MB_], nprow ); tmp2 = DNROC( N, Cd[NB_], npcol );
      tmp3   = DNROC( N, Ad[NB_], npcol ); tmp4 = DNROC( N, Bd[NB_], npcol );
      ABest  = (double)(K) *
               ( ( ( ( Ad[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 ) +
                 ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 ) +
                 TWO * ( npcol == 1 ? ZERO : tmp1 ) +
                 MAX( tmp1, tmp3 ) + MAX( tmp1, tmp4 ) );
   }
/*
*  Shift a little the cross-over point between both algorithms.
*/
   ChooseABC = ( ( 1.4 * ABCest ) <= ABest );
/*
*  BLACS topologies are enforced iff N and K are strictly greater than the
*  logical block size returned by pilaenv_. Otherwise, it is assumed that the
*  routine calling this routine has already selected an adequate topology.
*/
   nb       = hscal_pilaenv( C2F_CHAR( &type->type ) );
   ForceTop = ( ( N > nb ) && ( K > nb ) );

   if( ChooseABC )
   {
      if( notran )
      {
         OpC  = CBCAST;
         ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_GET );

         if( ForceTop )
         {
            OpR  = CCOMBINE;
            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_GET );

            rtopsave = rtop;
            ctopsave = ctop;

            if( upper ) { TopR = CTOP_IRING; TopC = CTOP_DRING; }
            else        { TopR = CTOP_DRING; TopC = CTOP_IRING; }

            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, &TopC );
            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    &TopR );
/*
*  Remove the next line when the BLACS combine operations support ring
*  topologies
*
*            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_DEFAULT );
*/
         }

         DirA = ( ctop == CTOP_DRING ? CBACKWARD : CFORWARD );
      }
      else
      {
         OpR  = CBCAST;
         rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_GET );

         if( ForceTop )
         {
            OpC  = CCOMBINE;
            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_GET );

            rtopsave = rtop;
            ctopsave = ctop;

            if( upper ) { TopR = CTOP_IRING; TopC = CTOP_DRING; }
            else        { TopR = CTOP_DRING; TopC = CTOP_IRING; }

            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    &TopR );
            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, &TopC );
/*
*  Remove the next line when the BLACS combine operations support ring
*  topologies
*
*            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_DEFAULT );
*/
         }

         DirA = ( rtop == CTOP_DRING ? CBACKWARD : CFORWARD );
      }

      {
      char *CONJUG_l = CONJG;

      hscal_PB_Cpsyr2kAC_C( 
        type, &DirA, CONJUG_l,   &UploC, ( notran ? NOTRAN : COTRAN ),
        N, K, ((char *)ALPHA), Ai, Aj, Ad,
        Bi, Bj, Bd, ((char *)&BETA), Ci,
        Cj, Cd,
        nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
      }
   }
   else
   {
      if( notran )
      {
         OpR  = CBCAST;
         rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_GET );

         if( ForceTop )
         {
            OpC  = CBCAST;
            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_GET );

            rtopsave = rtop;
            ctopsave = ctop;
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
            if( ( rtop != CTOP_DRING ) && ( rtop != CTOP_IRING ) &&
                ( rtop != CTOP_SRING ) )
               rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_SRING );
            if( ( ctop != CTOP_DRING ) && ( ctop != CTOP_IRING ) &&
                ( ctop != CTOP_SRING ) )
               ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_SRING );
         }

         DirA = ( rtop == CTOP_DRING ? CBACKWARD : CFORWARD );
      }
      else
      {
         OpC  = CBCAST;
         ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_GET );

         if( ForceTop )
         {
            OpR  = CBCAST;
            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_GET );

            rtopsave = rtop;
            ctopsave = ctop;
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
            if( ( rtop != CTOP_DRING ) && ( rtop != CTOP_IRING ) &&
                ( rtop != CTOP_SRING ) )
               rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_SRING );
            if( ( ctop != CTOP_DRING ) && ( ctop != CTOP_IRING ) &&
                ( ctop != CTOP_SRING ) )
               ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_SRING );
         }

         DirA = ( ctop == CTOP_DRING ? CBACKWARD : CFORWARD );
      }

      {
      char *CONJUG_l = CONJG;

      hscal_PB_Cpsyr2kA_C( 
        type, &DirA, CONJUG_l,   &UploC, ( notran ? NOTRAN : COTRAN ),
        N, K, ((char *)ALPHA), Ai, Aj, Ad,
        Bi, Bj, Bd, ((char *)&BETA), Ci, Cj,
        Cd,
        nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
      }
   }
/*
*  Restore the BLACS topologies when necessary.
*/
   if( ForceTop )
   {
      rtopsave = *PB_Ctop( &ctxt, &OpR, ROW,    &rtopsave );
      ctopsave = *PB_Ctop( &ctxt, &OpC, COLUMN, &ctopsave );
   }
/*
*  End of PCHER2K
*/
	
   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcher2k_tcomp_tcomm(
   int UPLOI,
   int TRANSI,
   int N,
   int K,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcher2k_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI,
                N, K,
                ALPHA, 
                IA, JA, DESCA, 
                IB, JB, DESCB, 
                BETA,
                IC, JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcher2k_info(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai;

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcher2k_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI,
                *N, *K,
                alphai,
                *IA, *JA, DESCA, 
                *IB, *JB, DESCB, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcher2k_info_i(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai;

   if (ALPHA[REAL_PART] == 0.f)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.f)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.f)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.f)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcher2k_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI,
                *N, *K,
                alphai,
                *IA, *JA, DESCA, 
                *IB, *JB, DESCB, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
