
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdcopy_recon_tcomp(
   int n,
   int incx,
   int incy
)
{
    double tcomp = 0;

    int num_of_computers = HMPI_Get_number_of_computers();
    int reconn = (((n/num_of_computers) >= 1) ? (n/num_of_computers) : 1);
    
    hscal_dcopy(
         &reconn,
         &incx,
         &incy,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdcopy_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdcopy_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdcopy_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdcopy_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdcopy_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdcopy_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdcopy_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   int            Xi, Xj, Yi, Yj, ctxt, info;
   PBTYP_T        * type;

/*
*  .. Local Arrays ..
*/
   int            Xd[DLEN_], Yd[DLEN_];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

/* ..
*  .. Executable Statements ..
*
*/
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
   PB_CargFtoC( IY, JY, DESCY, &Yi, &Yj, Yd );

   if (init_llds_f == 1)
   {
   HSCAL_PXCOPY_INIT_LLDS(
        N, 
        Xd, IX, INCX, 
        Yd, IY, INCY, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Yd[M_]=%d Yd[MB_]=%d myrow=%d Yd[RSRC_]=%d nprow=%d Yd[LLD_]=%d", 
        Yd[M_], Yd[MB_], myrow, Yd[RSRC_], nprow, Yd[LLD_]
   );

/*
*  Test the input parameters
*/
   if( !( info = ( ( nprow == -1 ) ? -( 501 + CTXT_ ) : 0 ) ) )
   {
      hscal_PB_Cchkvec( 
           "PDCOPY", "X", N, 1, Xi, Xj, Xd, INCX,  5, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec( 
           "PDCOPY", "Y", N, 1, Yi, Yj, Yd, INCY, 10, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info ) 
   { 
      hscal_PB_Cabort( 
           "PDCOPY", info,
           nprow, npcol, myrow, mycol
      ); 

      return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( N == 0 ) return HSCAL_SUCCESS;

/*
*  Get type structure
*/
   type = PB_Cdtypeset();
/*
*  Start the operations
*/
   if( INCX == Xd[M_] )
   {
      hscal_PB_Cpaxpby_D( 
           type, NOCONJG, 1, N, type->one, Xi,
           Xj, Xd, ROW,    type->zero, Yi, Yj, Yd,
           ( INCY == Yd[M_] ? ROW : COLUMN ),
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
   }
   else
   {
      hscal_PB_Cpaxpby_D( 
           type, NOCONJG, N, 1, type->one, Xi,
           Xj, Xd, COLUMN, type->zero, Yi, Yj, Yd,
           ( INCY == Yd[M_] ? ROW : COLUMN ),
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
   }

/*
*  End of pdcopy
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdcopy_tcomp_tcomm(
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdcopy_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N, 
                IX, JX, DESCX, INCX,
                IY, JY, DESCY, INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdcopy_info(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdcopy_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdcopy_info_i(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdcopy_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
