
#include <stdlib.h>
#include <stdio.h>

#include "../../include/scalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdgeadd_recon_tcomp(
   int transi,
   int m,
   int n, 
   int alphai,
   int betai
)
{
    double tcomp = 0.;

    char   transc  = transi;
    char   *trans  = &transc;

    int reconm     = hscal_pilaenv_recon("D");
    int reconn     = hscal_pilaenv_recon("D");

    int lda        = reconm;
    int ldc        = reconm;

    double alpha, beta;

    alpha          = alphai;
    beta           = betai;

    hscal_dmatadd(
         &m,
         &n,
         (char*)(&alpha),
         &lda,
         (char*)(&beta),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdgeadd_my_tcomp(
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETA,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdgeadd_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI, M, N,
          ALPHA, 
          IA, JA, DESCA, 
          BETA,
          IC, JC, DESCC, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_my_tcomm(
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETA,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdgeadd_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI, M, N,
          ALPHA, 
          IA, JA, DESCA,
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int TRANSI,
   int M,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETA,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdgeadd_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI, M, N,
          ALPHA, 
          IA, JA, DESCA, 
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int ALPHAI,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETAI,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           DirA, DirC, ctop, rtop;
   int            Ai, Aj, Ci, Cj, TrA, ctxt, info,
                  notran;
/*
*  .. Local Arrays ..
*/
   int            Ad[DLEN_], Cd[DLEN_];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

   char   TRANSC  = TRANSI, *TRANS = &TRANSC;
   double ALPHA   = ALPHAI;
   double BETA    = BETAI;

/* ..
*  .. Executable Statements ..
*
*/
   notran = ( ( TrA = Mupcase( F2C_CHAR( TRANS )[0] ) ) == CNOTRAN );

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IC, JC, DESCC, &Ci, &Cj, Cd );

   if (init_llds_f == 1)
   {
   HSCAL_PXGEADD_INIT_LLDS(
        Ad, 
        Cd, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Cd[M_]=%d Cd[MB_]=%d myrow=%d Cd[RSRC_]=%d nprow=%d Cd[LLD_]=%d", 
        Cd[M_], Cd[MB_], myrow, Cd[RSRC_], nprow, Cd[LLD_]
   );

/*
*  Test the input parameters
*/
   if( !( info = ( ( nprow == -1 ) ? -( 801 + CTXT_ ) : 0 ) ) )
   {
      if( ( !notran ) && ( TrA != CTRAN ) && ( TrA != CCOTRAN ) )
      {
         hscal_PB_Cwarn( 
              nprow, npcol, myrow, mycol,
              __LINE__, "PDGEADD", "Illegal TRANS = %c\n", TrA 
         );
         info = -1;
      }
      if( notran )
      {
         hscal_PB_Cchkmat( 
              "PDGEADD", "A", M, 2, N, 3, Ai, Aj, Ad,  8,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      else
      {
         hscal_PB_Cchkmat( 
              "PDGEADD", "A", N, 3, M, 2, Ai, Aj, Ad,  8,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      hscal_PB_Cchkmat( 
           "PDGEADD", "C", M, 2, N, 3, Ci, Cj, Cd, 13, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info ) 
   { 
      hscal_PB_Cabort( 
           "PDGEADD", info,
           nprow, npcol, myrow, mycol
      );

      return HSCAL_ERR_INPUT;
   }

/*
*  Quick return if possible
*/
   if( ( M == 0 ) || ( N == 0 ) ||
       ( ( ALPHA == ZERO ) && ( BETA == ONE ) ) )
   {
      return HSCAL_SUCCESS;
   }

/*
*  And when alpha is zero
*/
   if( ALPHA == ZERO )
   {
      if( BETA == ZERO )
      {
         hscal_PB_Cplapad_D( 
              PB_Cdtypeset(), ALL, NOCONJG, M, N, ((char *)&BETA),
              ((char *)&BETA), Ci, Cj, Cd,
              nprow, npcol, myrow, mycol, tcomp_l
         );
      }
      else
      {
         hscal_PB_Cplascal_D( 
              PB_Cdtypeset(), ALL, NOCONJG, M, N, ((char *)&BETA),
              Ci, Cj, Cd,
              nprow, npcol, myrow, mycol, tcomp_l
         );
      }

      return HSCAL_SUCCESS;
   }
/*
*  Start the operations
*/
/*
*  This operation mainly involves point-to-point send and receive communication.
*  There is therefore no particular BLACS topology to recommend. Still, one can
*  choose the main loop direction in which the operands will be added. This
*  selection is based on the current setting for the BLACS broadcast operations.
*/
   if( notran )
   {
      rtop = *PB_Ctop( &ctxt, BCAST, ROW,    TOP_GET );
      ctop = *PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );

      if( M <= N )
      {
         DirA = ( rtop == CTOP_DRING ? CBACKWARD : CFORWARD );
         DirC = ( ctop == CTOP_DRING ? CBACKWARD : CFORWARD );
      }
      else
      {
         DirA = ( ctop == CTOP_DRING ? CBACKWARD : CFORWARD );
         DirC = ( rtop == CTOP_DRING ? CBACKWARD : CFORWARD );
      }

      {
      char *CONJUG_l = NOCONJG;
      hscal_PB_Cpgeadd_D( 
        PB_Cdtypeset(), (&DirA), (&DirC), CONJUG_l, M, N,
        ((char *) &ALPHA), Ai, Aj, Ad,
        ((char *) &BETA), Ci, Cj, Cd,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      }
   }
   else
   {
      char *CONJUG_l = NOCONJG;
      hscal_PB_Cptran_D(
        PB_Cdtypeset(), CONJUG_l, M, N, ((char *) &ALPHA),
        Ai, Aj, Ad, ((char *) &BETA),
        Ci, Cj, Cd,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }

/*
*  End of PDGEADD
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_tcomp_tcomm(
   int TRANSI,
   int M,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETA,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdgeadd_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSI, M, N,
                ALPHA, 
                IA, JA, DESCA, 
                BETA,
                IC, JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_info(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double* BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int TRANSI = *TRANS;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdgeadd_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSI, *M, *N,
                alphai,
                *IA, *JA, DESCA, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgeadd_info_i(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double* BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int TRANSI = *TRANS;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdgeadd_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSI, *M, *N,
                alphai,
                *IA, *JA, DESCA, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
