
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdnrm2_recon_tcomp(
   int n,
   int incx
)
{
    double tcomp = 0;

    int num_of_computers = HMPI_Get_number_of_computers();
    int reconn = (((n/num_of_computers) >= 1) ? (n/num_of_computers) : 1);
    
    hscal_dnrm2(
         &reconn,
         &incx,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdnrm2_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdnrm2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdnrm2_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdnrm2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdnrm2_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdnrm2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdnrm2_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           top;
   int            Xcol, Xi, Xii, Xj, Xjj, Xld, Xnp, Xnq, Xrow, ctxt, dst, dist,
                  info, k, mydist, src;
   HSCALBLACSCONTEXT   * blacsctxt;

/*
*  .. Local Arrays ..
*/
   int            Xd[DLEN_];

   double         * tcomp_l  =  tcomp;
   int            * tcomm_l  =  tcomm;
/* ..
*  .. Executable Statements ..
*
*/
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );

   if (init_llds_f == 1)
   {
   HSCAL_PXNRM2_INIT_LLDS(
        N,
        Xd, IX, INCX, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

/*
*  Test the input parameters
*/
   if( !( info = ( ( nprow == -1 ) ? -( 601 + CTXT_ ) : 0 ) ) )
   {
      hscal_PB_Cchkvec(
           "PDNRM2", "X", N, 1, Xi, Xj, Xd, INCX, 6, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PDNRM2",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }

/*
*  Quick return if possible
*/
   if( N == 0 ) return HSCAL_SUCCESS;
/*
*  Retrieve process grid information
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*
*  Retrieve sub( X )'s local information: Xii, Xjj, Xrow, Xcol
*/
   PB_Cinfog2l( Xi, Xj, Xd, nprow, npcol, myrow, mycol, &Xii, &Xjj,
                &Xrow, &Xcol );
/*
*  Handle degenerate case separately, sub( X )'s scope is just one process
*/
   if( ( N == 1 ) && ( INCX == 1 ) && ( Xd[M_] == 1 ) )
   {
/*
*  Make sure I own some data and compute NORM2
*/
      if( ( ( myrow == Xrow ) || ( Xrow < 0 ) ) &&
          ( ( mycol == Xcol ) || ( Xcol < 0 ) ) )
      {
         if (tcomp_l != NULL)
         {
            TCOMPL((myrow), (mycol), (nprow), (npcol)) += 1;
         }
      }
   }
   else if( INCX == Xd[M_] )
   {
/*
*  sub( X ) resides in (a) process row(s)
*/
      if( ( myrow == Xrow ) || ( Xrow < 0 ) )
      {
/*
*        Initialize SCALE and SSQ
*/

/*
*  Make sure I own some data and compute local sum of squares
*/
         Xnq = PB_Cnumroc( N, Xj, Xd[INB_], Xd[NB_], mycol, Xd[CSRC_], npcol );
         if( Xnq > 0 )
         {
            if (tcomp_l != NULL)
            {
               TCOMPL((myrow), (mycol), (nprow), (npcol)) += Xnq*5;
            }
         }
/*
*  If Xnq <= 0, SCALE is zero and SSQ is one (see initialization above)
*/
         if( ( npcol >= 2 ) && ( Xcol >= 0 ) )
         {
/*
*  Combine the local sum of squares using a 1-tree topology within process row
*  0 if npcol > 1 and Xcol >= 0, i.e sub( X ) is distributed.
*/

            mydist  = mycol;
            k       = 1;
l_10:
            if( mydist & 1 )
            {
               dist = k * ( mydist - 1 );
               dst  = MPosMod( dist, npcol );
               hscal_Cdgesd2d( 
                    blacsctxt, 2, 1, 2, myrow, dst,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
               goto l_20;
            }
            else
            {
               dist = mycol + k;
               src  = MPosMod( dist, npcol );

               if( mycol < src )
               {
                 if (tcomp_l != NULL)
                 {
                   TCOMPL((myrow), (mycol), (nprow), (npcol)) += 5;
                 }
               }
               mydist >>= 1;
            }
            k <<= 1;

            if( k < npcol ) goto l_10;
l_20:
/*
*  Process column 0 broadcasts the combined values of SCALE and SSQ within their
*  process row.
*/
            top = *PB_Ctop( &ctxt, BCAST, ROW, TOP_GET );
            if( mycol == 0 )
            {
               hscal_Cdgebs2d( 
                    blacsctxt, ROW, &top, 2, 1, 2,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
            }
            else
            {
               hscal_Cdgebr2d( 
                    blacsctxt, ROW, &top, 2, 1, 2,
                    myrow, 0,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
            }
/*
*  Compute NORM2 redundantly NORM2  = WORK( 1 ) * SQRT( WORK( 2 ) )
*/
            ;
         }
         else
         {
/*
*  Compute NORM2 redundantly ( sub( X ) is not distributed )
*/
            ;
         }
      }
   }
   else
   {
/*
*  sub( X ) resides in (a) process column(s)
*/
      if( ( mycol == Xcol ) || ( Xcol < 0 ) )
      {
/*
*  Initialize SCALE and SSQ
*/

/*
*  Make sure I own some data and compute local sum of squares
*/
         Xnp = PB_Cnumroc( N, Xi, Xd[IMB_], Xd[MB_], myrow, Xd[RSRC_], nprow );
         if( Xnp > 0 )
         {
            if (tcomp_l != NULL)
            {
               TCOMPL((myrow), (mycol), (nprow), (npcol)) += Xnp*5;
            }
         }
/*
*  If Xnp <= 0, SCALE is zero and SSQ is one (see initialization above)
*/
         if( ( nprow >= 2 ) && ( Xrow >= 0 ) )
         {
/*
*  Combine the local sum of squares using a 1-tree topology within process
*  column 0 if nprow > 1 and Xrow >= 0, i.e sub( X ) is distributed.
*/

            mydist  = myrow;
            k       = 1;
l_30:
            if( mydist & 1 )
            {
               dist = k * ( mydist - 1 );
               dst  = MPosMod( dist, nprow );
               hscal_Cdgesd2d( 
                    blacsctxt, 2, 1, 2, dst, mycol,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
               goto l_40;
            }
            else
            {
               dist = myrow + k;
               src  = MPosMod( dist, nprow );

               if( myrow < src )
               {
                  if (tcomp_l != NULL)
                  {
                    TCOMPL((myrow), (mycol), (nprow), (npcol)) += 5;
                  }
               }
               mydist >>= 1;
            }
            k <<= 1;

            if( k < nprow ) goto l_30;
l_40:
/*
*  Process column 0 broadcasts the combined values of SCALE and SSQ within their
*  process column
*/
            top = *PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );
            if( myrow == 0 )
            {
               hscal_Cdgebs2d( 
                    blacsctxt, COLUMN, &top, 2, 1, 2,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
            }
            else
            {
               hscal_Cdgebr2d( 
                    blacsctxt, COLUMN, &top, 2, 1, 2,
                    0, mycol,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
            }
/*
*  Compute NORM2 redundantly NORM2 = WORK[0] * SQRT( WORK[1] )
*/
            ;
         }
         else
         {
/*
*  Compute NORM2 redundantly ( sub( X ) is not distributed )
*/
            ;
         }
      }
   }

   Cblacs_gridexit_nocomm(blacsctxt);

/*
*  End of pdnrm2
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdnrm2_tcomp_tcomm(
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdnrm2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N, 
                IX, JX, DESCX, INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdnrm2_info(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdnrm2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdnrm2_info_i(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdnrm2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
