
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdsymv_recon_tcomp(
   int uploi,
   int n,
   int alphai,
   int incx,
   int betai,
   int incy
)
{
    double tcomp = 0;

    char uploc   = uploi;
    char *uplo   = &uploc;

    char transc  = 'N';
    char *trans  = &transc;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int reconm           = n/p;
    int reconn           = n/p;
    int lda      = reconn;

    double alpha = alphai;
    double beta  = betai;
    
    hscal_dgemv(
         trans,
         &reconm,
         &reconn,
         ((char*)&alpha),
         &lda,
         &incx,
         ((char*)&beta),
         &incy,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdsymv_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdsymv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI,
          N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdsymv_my_tcomm(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdsymv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI,
          N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdsymv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLOI,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdsymv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI,
          N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdsymv_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int ALPHAI,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETAI,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           UploA, * one, top;
   int            Acol, Ai, Aii, Aimb1, Ainb1, Aj, Ajj, Akp, Akq, Ald, Amb,
                  Amp, Amp0, Anb, Anq, Anq0, Arow, XCfr, XCld, XRfr, XRld,
                  Xi, Xj, YCfr, YCpbY, YCsum, YCld, YRfr, YRpbY, YRsum,
                  YRld, Ycol, Yi, Yii, YisRow, Yj, Yjj, Yld, Ynp, Ynq, Yrow,
                  ctxt, info, ione=1, k, kb, ktmp, nb, size, upper;
   double         * tbeta;     
   PBTYP_T        * type;
   HSCALBLACSCONTEXT   * blacsctxt;

   char           UPLOC = UPLOI, *UPLO = &UPLOC;
   double         ALPHA = ALPHAI;
   double         BETA = BETAI;

/*
*  .. Local Arrays ..
*/
   int            Ad [DLEN_], Ad0[DLEN_], XCd[DLEN_], XRd[DLEN_], Xd[DLEN_],
                  YCd[DLEN_], YRd[DLEN_], Yd [DLEN_];

/* ..
*  .. Executable Statements ..
*
*/
   upper = ( ( UploA  = Mupcase( F2C_CHAR( UPLO )[0] ) ) == CUPPER );

   HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "UPPER %s", ((upper) ? "YES" : "NO"));

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
   PB_CargFtoC( IY, JY, DESCY, &Yi, &Yj, Yd );

   if (init_llds_f == 1)
   {
   HSCAL_PXSYMV_INIT_LLDS(
        N, 
        Ad, 
        Xd, IX, INCX, 
        Yd, IY, INCY, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Yd[M_]=%d Yd[MB_]=%d myrow=%d Yd[RSRC_]=%d nprow=%d Yd[LLD_]=%d", 
        Yd[M_], Yd[MB_], myrow, Yd[RSRC_], nprow, Yd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 701 + CTXT_ ) : 0 ) ) )
   {
      if( ( !upper ) && ( UploA != CLOWER ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, __FILE__, "Illegal UPLO = %c\n", UploA 
         );
         info = -1;
      }
      hscal_PB_Cchkmat(
           "PDSYMV", "A", N, 2, N, 2, Ai, Aj, Ad,  7, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec(
           "PDSYMV", "X", N, 2, Xi, Xj, Xd, INCX, 11, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec(
           "PDSYMV", "Y", N, 2, Yi, Yj, Yd, INCY, 17, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PDSYMV",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( ( N == 0 ) ||
       ( ( ALPHA == ZERO ) && ( BETA == ONE ) ) )
   {
      return HSCAL_SUCCESS;
   }

/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*
*  Get type structure
*/
   type = PB_Cdtypeset();
/*
*  When alpha is zero
*/
   if( ALPHA == ZERO )
   {
      if( INCY == Yd[M_] )
      {
/*
*  sub( Y ) resides in (a) process row(s)
*/
           {
              {
/*
*  Retrieve sub( Y )'s local information: Yii, Yjj, Yrow, Ycol
*/
                 PB_Cinfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, &Yii, &Yjj,
                              &Yrow, &Ycol );

                 if( ( myrow == Yrow ) || ( Yrow < 0 ) )
                 {
/*
*  Make sure I own some data and scale sub( Y )
*/
                     Ynq = PB_Cnumroc( N, Yj, Yd[INB_], Yd[NB_], mycol, Yd[CSRC_],
                                       npcol );
                     if( Ynq > 0 )
                     {
                        Yld = Yd[LLD_];
                        if( BETA == ZERO )
                        {
                           dset_tcomp_macro( 
                               &Ynq, ((char *) &BETA), &Yld,
                               nprow, npcol, myrow, mycol, tcomp
                           );
                        }
                        else
                        {
                           dscal_tcomp_macro( 
                                &Ynq, ((char *) &BETA), &Yld,
                                nprow, npcol, myrow, mycol, tcomp
                           );
                        }
                     }
                 }
              }
           }
      }
      else
      {
           /*
           *  sub( Y ) resides in (a) process column(s)
           */
           {
              {
                 /*
                 *  Retrieve sub( Y )'s local information: Yii, Yjj, Yrow, Ycol
                 */
                 PB_Cinfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, &Yii, &Yjj,
                              &Yrow, &Ycol );

/*
*  sub( Y ) resides in (a) process column(s)
*/
                 if( ( mycol == Ycol ) || ( Ycol < 0 ) )
                 {
/*
*  Make sure I own some data and scale sub( Y )
*/
                     Ynp = PB_Cnumroc( N, Yi, Yd[IMB_], Yd[MB_], myrow, Yd[RSRC_],
                                       nprow );
                     if( Ynp > 0 )
                     {
                        if( BETA == ZERO )
                        {
                           dset_tcomp_macro(
                               &Ynp, ((char *) &BETA), &INCY,
                               nprow, npcol, myrow, mycol, tcomp
                           );
                        }
                        else
                        {
                           dscal_tcomp_macro( 
                                &Ynp, ((char *) &BETA), &INCY,
                                nprow, npcol, myrow, mycol, tcomp
                           );
                        }
                     }
                 }
              }
           }
      }

      Cblacs_gridexit_nocomm(blacsctxt);

      return HSCAL_SUCCESS;
   }

   {
      {
/*
*  Compute descriptor Ad0 for sub( A )
*/
   PB_Cdescribe( N, N, Ai, Aj, Ad, nprow, npcol, myrow, mycol, &Aii, &Ajj,
                 &Ald, &Aimb1, &Ainb1, &Amb, &Anb, &Arow, &Acol, Ad0 );
/*
*  Reuse sub( Y ) and/or create vectors YR in process rows and YC in process
*  columns spanned by sub( A )
*/
   if( ( YisRow = ( INCY == Yd[M_] ) ) != 0 )
   {
      char * ZEROIT_l = INIT;
      int K_l = 1;
      char *ROWCOL_l = ROW;

      hscal_PB_CInOutV_D( 
                type, ROWCOL_l, N, N, Ad0, K_l, ((char *)&BETA), 
                Yi, Yj, Yd, ROWCOL_l, ((char**)(&tbeta)), YRd, 
                (&YRsum), (&YRpbY), 
                nprow, npcol, myrow, mycol, tcomp 
      );

      ROWCOL_l = COLUMN;
      hscal_PB_COutV_D(
              type, ROWCOL_l, ZEROIT_l, N, N, Ad0, K_l, YCd, &YCsum,
              nprow, npcol, myrow, mycol, tcomp
      );
   }
   else
   {
      char * ZEROIT_l = INIT;
      int K_l = 1;
      char *ROWCOL_l = COLUMN;

      hscal_PB_CInOutV_D( 
                type, ROWCOL_l, N, N, Ad0, K_l, ((char *)&BETA), 
                Yi, Yj, Yd, ROWCOL_l, ((char**)(&tbeta)), YCd, 
                (&YCsum), (&YCpbY),
                nprow, npcol, myrow, mycol, tcomp 
      );

      ROWCOL_l = ROW;
      hscal_PB_COutV_D( 
              type, ROWCOL_l, ZEROIT_l, N, N, Ad0, K_l, YRd, &YRsum,
              nprow, npcol, myrow, mycol, tcomp
      );
   }
/*
*  Replicate sub( X ) in process rows (XR) and process columns (XC) spanned by
*  sub( A )
*/
   if( INCX == Xd[M_] )
   {
      char *CONJUG_l = NOCONJG;
      char *ROWCOL_l = ROW;
      char *ROWCOL_l_2;
      int K_l = 1;

      hscal_PB_CInV_D(
             type, CONJUG_l, ROWCOL_l, N, N, Ad0, K_l, Xi, Xj, Xd,
             ROWCOL_l, XRd, 
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );

      ROWCOL_l = COLUMN;
      ROWCOL_l_2 = ROW;
      hscal_PB_CInV_D(
             type, CONJUG_l, ROWCOL_l, N, N, Ad0, K_l, 0, 0, XRd,
             ROWCOL_l_2, XCd,
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
   else
   {
      char *CONJUG_l = NOCONJG;
      char *ROWCOL_l = COLUMN;
      char *ROWCOL_l_2;
      int K_l = 1;

      hscal_PB_CInV_D( 
             type, CONJUG_l, ROWCOL_l, N, N, Ad0, K_l, Xi, Xj,  Xd,
             ROWCOL_l, XCd,
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );

      ROWCOL_l = ROW;
      ROWCOL_l_2 = COLUMN;
      hscal_PB_CInV_D( 
             type, NOCONJG, ROWCOL_l, N, N, Ad0, K_l, 0,  0, XCd,
             ROWCOL_l_2, XRd,
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }

   one = type->one;
/*
*  Local matrix-vector multiply iff I own some data
*/
   Aimb1 = Ad0[IMB_ ]; Ainb1 = Ad0[INB_ ]; Amb = Ad0[MB_]; Anb = Ad0[NB_];
   Acol  = Ad0[CSRC_]; Arow  = Ad0[RSRC_];
   Amp   = PB_Cnumroc( N, 0, Aimb1, Amb, myrow, Arow, nprow );
   Anq   = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );

   if( ( Amp > 0 ) && ( Anq > 0 ) )
   {
      size = type->size;
      XCld = XCd[LLD_]; XRld = XRd[LLD_]; YCld = YCd[LLD_]; YRld = YRd[LLD_];
/*
*  Scale YR or YC in the case sub( Y ) has been reused
*/
      if( YisRow )
      {
/*
*  YR resides in (a) process row(s)
*/
         if( !YRpbY )
         {
            if( ( myrow == YRd[RSRC_] ) || ( YRd[RSRC_] < 0 ) )
            {
/*
*  Make sure I own some data and scale YR
*/
               if( Anq > 0 )
               {
                  if( tbeta[REAL_PART] == ZERO )
                  {
                     dset_tcomp_macro( 
                         &Anq, ((char *) tbeta), &YRld,
                         nprow, npcol, myrow, mycol, tcomp
                     );
                  }
                  else
                  {
                     dscal_tcomp_macro( 
                          &Anq, ((char *) tbeta), &YRld,
                          nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
            }
         }
      }
      else
      {
/*
*  YC resides in (a) process column(s)
*/
         if( !YCpbY )
         {
            if( ( mycol == YCd[CSRC_] ) || ( YCd[CSRC_] < 0 ) )
            {
/*
*  Make sure I own some data and scale YC
*/
               if( Amp > 0 )
               {
                  if( tbeta[REAL_PART] == ZERO )
                  {
                     dset_tcomp_macro( 
                         &Amp, ((char *) tbeta), &ione,
                         nprow, npcol, myrow, mycol, tcomp
                     );
                  }
                  else
                  {
                     dscal_tcomp_macro( 
                          &Amp, ((char *) tbeta), &ione, 
                          nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
            }
         }
      }
/*
*  Computational partitioning size is computed as the product of the logical
*  value returned by pilaenv_ and 2 * lcm( nprow, npcol ).
*/
      nb = 2 * hscal_pilaenv( C2F_CHAR( &type->type ) ) *
           PB_Clcm( ( Arow >= 0 ? nprow : 1 ), ( Acol >= 0 ? npcol : 1 ) );

      if( upper )
      {
         for( k = 0; k < N; k += nb )
         {
            kb   = N - k; kb = MIN( kb, nb );
            Akp  = PB_Cnumroc( k,  0, Aimb1, Amb, myrow, Arow, nprow );
            Akq  = PB_Cnumroc( k,  0, Ainb1, Anb, mycol, Acol, npcol );
            Anq0 = PB_Cnumroc( kb, k, Ainb1, Anb, mycol, Acol, npcol );
            if( Akp > 0 && Anq0 > 0 )
            {
               dgemv_tcomp_macro( 
                    C2F_CHAR( NOTRAN ), &Akp, &Anq0, ((char *)&ALPHA),
                    &Ald, &XRld, one, &ione,
                    nprow, npcol, myrow, mycol, tcomp
               );
               dgemv_tcomp_macro( 
                    C2F_CHAR( TRAN   ), &Akp, &Anq0, ((char *)&ALPHA),
                    &Ald, &ione, one,
                    &YRld,
                    nprow, npcol, myrow, mycol, tcomp
               );
            }
            
            {
               char *SIDE_l = LEFT;
               char *UPLO_l = UPPER;
               int K_l = 1;

               hscal_PB_Cpsym( 
                      type, type, SIDE_l, UPLO_l, kb, K_l, ((char *) &ALPHA),
                      k, k, Ad0, XCld,
                      XRld, YCld, YRld,
                      hscal_PB_Ctzsymv_D,
                      nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }
      else
      {
         for( k = 0; k < N; k += nb )
         {
            kb  = N - k; ktmp = k + ( kb = MIN( kb, nb ) );
            Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
            Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
            
            {
                char *SIDE_l = LEFT;
                char *UPLO_l = LOWER;
                int K_l = 1;

                hscal_PB_Cpsym( 
                      type, type, SIDE_l, UPLO_l, kb, K_l, ((char *) &ALPHA),
                      k, k, Ad0, XCld,
                      XRld, YCld, YRld,
                      hscal_PB_Ctzsymv_D,
                      nprow, npcol, myrow, mycol, tcomp
                );
            }

            Akp  = PB_Cnumroc( ktmp, 0, Aimb1, Amb, myrow, Arow, nprow );
            Amp0 = Amp - Akp;
            Anq0 = PB_Cnumroc( kb,   k, Ainb1, Anb, mycol, Acol, npcol );
            if( Amp0 > 0 && Anq0 > 0 )
            {
               dgemv_tcomp_macro( 
                    C2F_CHAR( NOTRAN ), &Amp0, &Anq0, ((char *) &ALPHA),
                    &Ald, &XRld, one, &ione,
                    nprow, npcol, myrow, mycol, tcomp
               );
               dgemv_tcomp_macro( 
                    C2F_CHAR( TRAN   ), &Amp0, &Anq0, ((char *) &ALPHA),
                    &Ald, &ione, one, &YRld,
                    nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }
   }

   if( YisRow )
   {
/*
*  Combine the partial column results into YC
*/
      if( YCsum )
      {
         YCd[CSRC_] = 0;
         if( Amp > 0 )
         {
            top        = *PB_Ctop( &ctxt, COMBINE, ROW, TOP_GET );
            hscal_Cdgsum2d( 
              blacsctxt, ROW, &top, Amp, 1, YCd[LLD_], myrow, 0,
              nprow, npcol, myrow, mycol, tcomp, tcomm 
            );
         }
      }
/*
*  Combine the partial row results into YR
*/
      if( YRsum && ( Anq > 0 ) )
      {
         top = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );
         hscal_Cdgsum2d( 
             blacsctxt, COLUMN, &top, 1, Anq, YRd[LLD_], YRd[RSRC_],
             mycol,
             nprow, npcol, myrow, mycol, tcomp, tcomm 
         );
      }

/*
*  YR := YR + YC
*/
      {
               int N_l = 1;
               int IA_l = 0;
               int JA_l = 0;
               char *CONJUG_l = NOCONJG;

               hscal_PB_Cpaxpby_D( 
                 type, CONJUG_l, N, N_l, one, IA_l, JA_l, YCd, COLUMN, one,
                 IA_l, JA_l, YRd, ROW,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
               );
      }
/*
*  sub( Y ) := beta * sub( Y ) + YR (if necessary)
*/
      if( YRpbY )
      {
               int N_l = 1;
               int IA_l = 0;
               int JA_l = 0;
               char *CONJUG_l = NOCONJG;

               hscal_PB_Cpaxpby_D( 
                 type, CONJUG_l, N_l, N, one, IA_l, JA_l, YRd, ROW,
                 ((char *)&BETA), Yi, Yj, Yd, ROW,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
               );
      }
   }
   else
   {
/*
*  Combine the partial row results into YR
*/
      if( YRsum )
      {
         YRd[RSRC_] = 0;
         if( Anq > 0 )
         {
            top = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );
            hscal_Cdgsum2d(
               blacsctxt, COLUMN, &top, 1, Anq, YRd[LLD_], 0,
               mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
      }
/*
*  Combine the partial column results into YC
*/
      if( YCsum && ( Amp > 0 ) )
      {
         top = *PB_Ctop( &ctxt, COMBINE, ROW, TOP_GET );
         hscal_Cdgsum2d( 
            blacsctxt, ROW, &top, Amp, 1, YCd[LLD_], myrow,
            YCd[CSRC_],
            nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
/*
*  YC := YR + YC
*/
      {
               int N_l = 1;
               int IA_l = 0;
               int JA_l = 0;
               char *CONJUG_l = NOCONJG;

               hscal_PB_Cpaxpby_D( 
                  type, CONJUG_l, N_l, N, one, IA_l, JA_l, YRd, ROW, one,
                  0, 0, YCd, COLUMN,
                  nprow, npcol, myrow, mycol, tcomp, tcomm
               );
      }
/*
*  sub( Y ) := beta * sub( Y ) + YC (if necessary)
*/
      if( YCpbY )
      {
               int N_l = 1;
               int IA_l = 0;
               int JA_l = 0;
               char *CONJUG_l = NOCONJG;

               hscal_PB_Cpaxpby_D( 
                  type, CONJUG_l, N, N_l, one, IA_l, JA_l, YCd,
                  COLUMN, ((char *)&BETA), Yi, Yj, Yd,
                  COLUMN,
                  nprow, npcol, myrow, mycol, tcomp, tcomm
               );
      }
   }
/*
*  End of PDSYMV
*/
      }
   }

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdsymv_tcomp_tcomm(
   int UPLOI,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdsymv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI,
                N, 
                ALPHA, IA, JA, DESCA, 
                IX, JX, DESCX, INCX, BETA,
                IY, JY, DESCY, INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdsymv_info(
   F_CHAR_T UPLO,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI = *UPLO;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdsymv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI,
                *N, 
                alphai, *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX, betai,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdsymv_info_i(
   F_CHAR_T UPLO,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI = *UPLO;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdsymv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI,
                *N, 
                alphai, *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX, betai,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
