
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdtrmv_recon_tcomp(
   int uploi,
   int transi,
   int diagi,
   int n,
   int incx
)
{
    double tcomp = 0;

    char uploc   = uploi;
    char *uplo   = &uploc;
    char transc  = transi;
    char *trans  = &transc;
    char diagc   = diagi;
    char *diag   = &diagc;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int reconn           = n/p;
    int lda      = reconn;
    
    hscal_dtrmv(
         uplo,
         trans,
         diag,
         &reconn,
         &lda,
         &incx,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f", tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdtrmv_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdtrmv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          IA, JA, DESCA, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_my_tcomm(
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdtrmv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          IA, JA, DESCA, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          NULL, tcomm
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdtrmv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          IA, JA, DESCA, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           DiagA, TranOp, UploA, Xroc, * one, * tbeta, top, *zero;
   int            Acol, Ai, Aii, Aimb1, Ainb1, Aj, Ajj, Akp, Akq, Ald, Amb,
                  Amp, Amp0, Anb, Anq, Anq0, Arow, XAfr, XAld, Xcol, Xi, Xii,
                  Xj, Xjj, Xld, Xnp, Xnq, Xrow, YAfr, YAld, YApbY, YAsum,
                  ctxt, info, ione=1, k, kb, ktmp, nb, notran,
                  nounit, size, upper;
   PBTYP_T        * type;
   HSCALBLACSCONTEXT   * blacsctxt;

   char UPLOC    = UPLOI, *UPLO = &UPLOC;
   char TRANSC   = TRANSI, *TRANS = &TRANSC;
   char DIAGC    = DIAGI, *DIAG = &DIAGC;

/*
*  .. Local Arrays ..
*/
   int            Ad [DLEN_], Ad0[DLEN_], XAd[DLEN_], Xd[DLEN_], YAd[DLEN_];
/* ..
*  .. Executable Statements ..
*
*/
   upper  = ( ( UploA  = Mupcase( F2C_CHAR( UPLO  )[0] ) ) ==  CUPPER );
   notran = ( ( TranOp = Mupcase( F2C_CHAR( TRANS )[0] ) ) == CNOTRAN );
   nounit = ( ( DiagA  = Mupcase( F2C_CHAR( DIAG  )[0] ) ) == CNOUNIT );

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );

   if (init_llds_f == 1)
   {
   HSCAL_PXTRMV_INIT_LLDS(
        N, 
        Ad, 
        Xd, IX, INCX, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 801 + CTXT_ ) : 0 ) ) )
   {
      if( ( !upper ) && ( UploA != CLOWER ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PDTRMV", "Illegal UPLO = %c\n", UploA 
         );
         info = -1;
      }
      else if( ( !notran ) && ( TranOp != CTRAN ) && ( TranOp != CCOTRAN ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PDTRMV", "Illegal TRANS = %c\n", TranOp 
         );
         info = -2;
      }
      else if( ( !nounit ) && ( DiagA != CUNIT ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PDTRMV", "Illegal DIAG = %c\n", DiagA 
         );
         info = -3;
      }
      hscal_PB_Cchkmat(
           "PDTRMV", "A", N, 4, N, 4, Ai, Aj, Ad,  8, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec(
           "PDTRMV", "X", N, 4, Xi, Xj, Xd, INCX, 12, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PDTRMV",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( N == 0 ) return HSCAL_SUCCESS;
/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*
*  Get type structure
*/
   type = PB_Cdtypeset();
   size = type->size; one  = type->one; zero = type->zero;
/*
*  Compute descriptor Ad0 for sub( A )
*/
   PB_Cdescribe( N, N, Ai, Aj, Ad, nprow, npcol, myrow, mycol, &Aii, &Ajj,
                 &Ald, &Aimb1, &Ainb1, &Amb, &Anb, &Arow, &Acol, Ad0 );

   Xroc = ( INCX == Xd[M_] ? CROW : CCOLUMN );

   if( notran )
   {
/*
*  Replicate sub( X ) in process rows spanned by sub( A ) -> XA
*/
      {
         char *CONJUG_l = NOCONJG;
         char *ROWCOL_l = ROW;
         int K_l = 1;
         hscal_PB_CInV_D( 
                type, CONJUG_l, ROWCOL_l, N, N, Ad0, K_l, Xi, Xj, Xd,
                &Xroc, XAd,
                nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
/*
*  Reuse sub( X ) and/or create vector YA in process columns spanned by sub( A )
*/
      {
         int K_l = 1;
         char *ROWCOL_l = COLUMN;
         hscal_PB_CInOutV_D( 
                type, ROWCOL_l, N, N, Ad0, K_l, one, Xi, Xj, Xd,
                &Xroc, &tbeta, YAd, (&YAsum), (&YApbY),
                nprow, npcol, myrow, mycol, tcomp 
         );
      }
/*
*  If sub( X ) is distributed in (a) process column(s), then zero it.
*/
      if( Xroc == CCOLUMN )
      {
/*
*  Retrieve sub( X )'s local information: Xii, Xjj, Xrow, Xcol
*/
         PB_Cinfog2l( Xi, Xj, Xd, nprow, npcol, myrow, mycol, &Xii, &Xjj, &Xrow,
                      &Xcol );
/*
*  sub( X ) resides in (a) process columns(s)
*/
         if( ( mycol == Xcol ) || ( Xcol < 0 ) )
         {
/*
*  Make sure I own some data and scale sub( X )
*/
            Xnp = PB_Cnumroc( N, Xi, Xd[IMB_], Xd[MB_], myrow, Xd[RSRC_],
                              nprow );
            if( Xnp > 0 )
            {
               dset_tcomp_macro( 
                   &Xnp, zero, &ione,
                   nprow, npcol, myrow, mycol, tcomp 
               );
            }
         }
      }
   }
   else
   {
/*
*  Replicate sub( X ) in process columns spanned by sub( A ) -> XA
*/
      {
         char *CONJUG_l = NOCONJG;
         char *ROWCOL_l = COLUMN;
         int K_l = 1;
         hscal_PB_CInV_D( 
               type, CONJUG_l, ROWCOL_l, N, N, Ad0, K_l, Xi, Xj, Xd,
               &Xroc, XAd,
               nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
/*
*  Reuse sub( X ) and/or create vector YA in process rows spanned by sub( A )
*/
      {
         int K_l = 1;
         char *ROWCOL_l = ROW;
         hscal_PB_CInOutV_D( 
                type, ROWCOL_l,    N, N, Ad0, K_l, one, Xi, Xj, Xd,
                &Xroc, &tbeta, YAd, &YAsum, &YApbY,
                nprow, npcol, myrow, mycol, tcomp 
         );
      }
/*
*  If sub( X ) is distributed in (a) process row(s), then zero it.
*/
      if( Xroc == CROW )
      {
/*
*  Retrieve sub( X )'s local information: Xii, Xjj, Xrow, Xcol
*/
         PB_Cinfog2l( Xi, Xj, Xd, nprow, npcol, myrow, mycol, &Xii, &Xjj, &Xrow,
                      &Xcol );
/*
*  sub( X ) resides in (a) process row(s)
*/
         if( ( myrow == Xrow ) || ( Xrow < 0 ) )
         {
/*
*  Make sure I own some data and scale sub( X )
*/
            Xnq = PB_Cnumroc( N, Xj, Xd[INB_], Xd[NB_], mycol, Xd[CSRC_],
                              npcol );
            if( Xnq > 0 )
            {
               Xld = Xd[LLD_];
               dset_tcomp_macro( 
                   &Xnq, zero, &Xld,
                   nprow, npcol, myrow, mycol, tcomp 
               );
            }
         }
      }
   }
/*
*  Local matrix-vector multiply iff I own some data
*/
   Aimb1 = Ad0[IMB_ ]; Ainb1 = Ad0[INB_ ]; Amb = Ad0[MB_]; Anb = Ad0[NB_];
   Acol  = Ad0[CSRC_]; Arow  = Ad0[RSRC_];
   Amp = PB_Cnumroc( N, 0, Aimb1, Amb, myrow, Arow, nprow );
   Anq = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );

   if( ( Amp > 0 ) && ( Anq > 0 ) )
   {
      XAld = XAd[LLD_]; YAld = YAd[LLD_];
/*
*  Computational partitioning size is computed as the product of the logical
*  value returned by pilaenv_ and 2 * lcm( nprow, npcol ).
*/
      nb = 2 * hscal_pilaenv( C2F_CHAR( &type->type ) ) *
           PB_Clcm( ( Arow >= 0 ? nprow : 1 ), ( Acol >= 0 ? npcol : 1 ) );

      if( upper )
      {
         if( notran )
         {
            for( k = 0; k < N; k += nb )
            {
               kb   = N - k; kb = MIN( kb, nb );
               Akp  = PB_Cnumroc( k,  0, Aimb1, Amb, myrow, Arow, nprow );
               Akq  = PB_Cnumroc( k,  0, Ainb1, Anb, mycol, Acol, npcol );
               Anq0 = PB_Cnumroc( kb, k, Ainb1, Anb, mycol, Acol, npcol );
               if( Akp > 0 && Anq0 > 0 )
               {
                  dgemv_tcomp_macro( 
                       TRANS, &Akp, &Anq0, one, &Ald,
                       &XAld, one, &ione,
                       nprow, npcol, myrow, mycol, tcomp 
                  );
               }
               hscal_PB_Cptrm( 
                 type, type, LEFT, UPPER, &TranOp, &DiagA, kb, 1, one,
                 k, k, Ad0, XAld,
                 YAld, hscal_PB_Ctztrmv_D,
                 nprow, npcol, myrow, mycol, tcomp 
               );
            }
         }
         else
         {
            for( k = 0; k < N; k += nb )
            {
               kb   = N - k; kb = MIN( kb, nb );
               Akp  = PB_Cnumroc( k,  0, Aimb1, Amb, myrow, Arow, nprow );
               Akq  = PB_Cnumroc( k,  0, Ainb1, Anb, mycol, Acol, npcol );
               Anq0 = PB_Cnumroc( kb, k, Ainb1, Anb, mycol, Acol, npcol );
               if( Akp > 0 && Anq0 > 0 )
               {
                  dgemv_tcomp_macro( 
                       TRANS, &Akp, &Anq0, one, &Ald, &ione, one, &YAld,
                       nprow, npcol, myrow, mycol, tcomp 
                  );
               }
               hscal_PB_Cptrm( 
                     type, type, LEFT, UPPER, &TranOp, &DiagA, kb, 1, one,
                     k, k, Ad0, XAld,
                     YAld, hscal_PB_Ctztrmv_D,
                     nprow, npcol, myrow, mycol, tcomp 
               );
            }
         }
      }
      else
      {
         if( notran )
         {
            for( k = 0; k < N; k += nb )
            {
               kb  = N - k; ktmp = k + ( kb = MIN( kb, nb ) );
               Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
               Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
               hscal_PB_Cptrm( 
                    type, type, LEFT, LOWER, &TranOp, &DiagA, kb, 1, one,
                    k, k, Ad0, XAld,
                    YAld, hscal_PB_Ctztrmv_D,
                    nprow, npcol, myrow, mycol, tcomp 
               );
               Akp  = PB_Cnumroc( ktmp, 0, Aimb1, Amb, myrow, Arow, nprow );
               Amp0 = Amp - Akp;
               Anq0 = PB_Cnumroc( kb,   k, Ainb1, Anb, mycol, Acol, npcol );
               if( Amp0 > 0 && Anq0 > 0 )
               {
                  dgemv_tcomp_macro( 
                       TRANS, &Amp0, &Anq0, one,
                       &Ald,
                       &XAld, one,
                       &ione,
                       nprow, npcol, myrow, mycol, tcomp 
                  );
               }
            }
         }
         else
         {
            for( k = 0; k < N; k += nb )
            {
               kb = N - k; ktmp = k + ( kb = MIN( kb, nb ) );
               Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
               Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
               hscal_PB_Cptrm( 
                       type, type, LEFT, LOWER, &TranOp, &DiagA, kb, 1, one,
                       k, k, Ad0, XAld,
                       YAld, hscal_PB_Ctztrmv_D,
                       nprow, npcol, myrow, mycol, tcomp 
               );
               Akp  = PB_Cnumroc( ktmp, 0, Aimb1, Amb, myrow, Arow, nprow );
               Amp0 = Amp - Akp;
               Anq0 = PB_Cnumroc( kb,   k, Ainb1, Anb, mycol, Acol, npcol );
               if( Amp0 > 0 && Anq0 > 0 )
               {
                  dgemv_tcomp_macro( 
                       TRANS, &Amp0, &Anq0, one,
                       &Ald,
                       &ione, one,
                       &YAld,
                       nprow, npcol, myrow, mycol, tcomp 
                  );
               }
            }
         }
      }
   }

   if( notran )
   {
/*
*  Combine the partial column results into YA
*/
      if( YAsum && ( Amp > 0 ) )
      {
         top = *PB_Ctop( &ctxt, COMBINE, ROW, TOP_GET );
         hscal_Cdgsum2d( 
                 blacsctxt, ROW, &top, Amp, 1, YAd[LLD_], 
                 myrow, YAd[CSRC_],
                 nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
/*
*  sub( X ) := YA (if necessary)
*/
      if( YApbY )
      {
         int N_l = 1;
         int IA_l = 0;
         int JA_l = 0;
         char *CONJUG_l = NOCONJG;

         hscal_PB_Cpaxpby_D( 
           type, CONJUG_l, N, N_l, one, IA_l, JA_l, YAd, COLUMN, zero,
           Xi, Xj, Xd, &Xroc,
           nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
   }
   else
   {
/*
*  Combine the partial row results into YA
*/
      if( YAsum && ( Anq > 0 ) )
      {
         top = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );
         hscal_Cdgsum2d( 
            blacsctxt, COLUMN, &top, 1, Anq, YAd[LLD_], YAd[RSRC_],
            mycol,
            nprow, npcol, myrow, mycol, tcomp, tcomm
          );
      }
/*
*  sub( X ) := YA (if necessary)
*/
      if( YApbY )
      {
         int N_l = 1;
         int IA_l = 0;
         int JA_l = 0;
         char *CONJUG_l = NOCONJG;

         hscal_PB_Cpaxpby_D( 
            type, CONJUG_l, N_l, N, one, IA_l, JA_l, YAd, ROW, zero,
            Xi, Xj, Xd, &Xroc,
            nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
   }
/*
*  End of PDTRMV
*/
   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_tcomp_tcomm(
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdtrmv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                N, 
                IA, JA, DESCA, 
                IX, JX, DESCX, INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_info(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;
   int DIAGI  = *DIAG;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdtrmv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                *N, 
                *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdtrmv_info_i(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;
   int DIAGI  = *DIAG;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdtrmv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                *N, 
                *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
