
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_psatrmv_recon_tcomp(
   int uploi,
   int transi,
   int diagi,
   int n,
   int alphai,
   int incx,
   int betai,
   int incy
)
{
    double tcomp = 0;

    char uploc   = uploi;
    char *uplo   = &uploc;
    char transc  = transi;
    char *trans  = &transc;
    char diagc   = diagi;
    char *diag   = &diagc;

    int num_of_computers = HMPI_Get_number_of_computers();
    int reconn = (((n/num_of_computers) >= 1) ? (n/num_of_computers) : 1);
    int lda      = reconn;

    float alpha  = alphai;
    float beta   = betai;

    hscal_satrmv(
         uplo,
         trans,
         diag,
         &reconn,
         ((char*)&alpha),
         &lda,
         &incx,
         ((char*)&beta),
         &incy,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f", tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_psatrmv_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }
  
   rc = hscal_psatrmv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_my_tcomm(
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psatrmv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_psatrmv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int ALPHAI,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETAI,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           DiagA, TranOp, UploA, Yroc, * one, top;
   int            Acol, Ai, Aii, Aimb1, Ainb1, Aj, Ajj, Akp, Akq, Ald, Amb,
                  Amp, Amp0, Anb, Anq, Anq0, Arow, XAfr, XAld, Xi, Xj, YAfr,
                  YAld, YApbY, YAsum, Ycol, Yi, Yii, Yj, Yjj, Yld, Ynp, Ynq,
                  Yrow, ctxt, info, ione=1, k, kb, ktmp, nb,
                  notran, nounit, size, upper, usiz;
   float          * tbeta;
   PBTYP_T        * type, * utyp;
   HSCALBLACSCONTEXT   * blacsctxt;

   char UPLOC     = UPLOI, *UPLO = &UPLOC;
   char TRANSC    = TRANSI, *TRANS = &TRANSC;
   char DIAGC     = DIAGI, *DIAG = &DIAGC;

   float ALPHA    = ALPHAI;
   float BETA     = BETAI;

/*
*  .. Local Arrays ..
*/
   int            Ad [DLEN_], Ad0[DLEN_], XAd[DLEN_], Xd[DLEN_], YAd[DLEN_],
                  Yd [DLEN_];
/* ..
*  .. Executable Statements ..
*
*/
   upper  = ( ( UploA  = Mupcase( F2C_CHAR( UPLO  )[0] ) ) ==  CUPPER );
   notran = ( ( TranOp = Mupcase( F2C_CHAR( TRANS )[0] ) ) == CNOTRAN );
   nounit = ( ( DiagA  = Mupcase( F2C_CHAR( DIAG  )[0] ) ) == CNOUNIT );

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
   PB_CargFtoC( IY, JY, DESCY, &Yi, &Yj, Yd );

   if (init_llds_f == 1)
   {
   HSCAL_PXATRMV_INIT_LLDS(
        N, 
        Ad, 
        Xd, IX, INCX, 
        Yd, IY, INCY, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Yd[M_]=%d Yd[MB_]=%d myrow=%d Yd[RSRC_]=%d nprow=%d Yd[LLD_]=%d", 
        Yd[M_], Yd[MB_], myrow, Yd[RSRC_], nprow, Yd[LLD_]
   );

/*
*  Test the input parameters
*/
   if( !( info = ( ( nprow == -1 ) ? -( 801 + CTXT_ ) : 0 ) ) )
   {
      if( ( !upper ) && ( UploA != CLOWER ) )
      {
         hscal_PB_Cwarn( 
              nprow, npcol, myrow, mycol, 
              __LINE__, "PSATRMV", "Illegal UPLO = %c\n", UploA 
         );
         info = -1;
      }
      else if( ( !notran ) && ( TranOp != CTRAN ) && ( TranOp != CCOTRAN ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol, 
              __LINE__, "PSATRMV", "Illegal TRANS = %c\n", TranOp 
         );
         info = -2;
      }
      else if( ( !nounit ) && ( DiagA != CUNIT ) )
      {
         hscal_PB_Cwarn( 
              nprow, npcol, myrow, mycol, 
              __LINE__, "PSATRMV", "Illegal DIAG = %c\n", DiagA 
         );
         info = -3;
      }
      hscal_PB_Cchkmat( 
           "PSATRMV", "A", N, 4, N, 4, Ai, Aj, Ad,  9, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec( 
           "PSATRMV", "X", N, 4, Xi, Xj, Xd, INCX, 13, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec( 
           "PSATRMV", "Y", N, 4, Yi, Yj, Yd, INCY, 19, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info ) 
   { 
      hscal_PB_Cabort( 
           "PSATRMV", info,
           nprow, npcol, myrow, mycol
      ); 

      return HSCAL_ERR_INPUT;
   }

/*
*  Quick return if possible
*/
   if( ( N == 0 ) || ( ( ALPHA == ZERO ) &&
                       ( BETA  == ONE  ) ) ) return HSCAL_SUCCESS;
/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);
/*
*  Get type structure
*/
   type = utyp = PB_Cstypeset();
   size = usiz = type->size;
/*
*  and when alpha is zero
*/
   if( ALPHA == ZERO )
   {
/*
*  Retrieve sub( Y )'s local information: Yii, Yjj, Yrow, Ycol
*/
      PB_Cinfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, &Yii, &Yjj,
                   &Yrow, &Ycol );

      if( INCY == Yd[M_] )
      {
/*
*  sub( Y ) resides in (a) process row(s)
*/
         if( ( myrow == Yrow ) || ( Yrow < 0 ) )
         {
/*
*  Make sure I own some data and scale sub( Y )
*/
            Ynq = PB_Cnumroc( N, Yj, Yd[INB_], Yd[NB_], mycol, Yd[CSRC_],
                              npcol );
            if( Ynq > 0 )
            {
               Yld = Yd[LLD_];
               sascal_tcomp_macro( 
                     &Ynq, ((char *) &BETA), &Yld,
                     nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }
      else
      {
/*
*  sub( Y ) resides in (a) process column(s)
*/
         if( ( mycol == Ycol ) || ( Ycol < 0 ) )
         {
/*
*  Make sure I own some data and scale sub( Y )
*/
            Ynp = PB_Cnumroc( N, Yi, Yd[IMB_], Yd[MB_], myrow, Yd[RSRC_],
                              nprow );
            if( Ynp > 0 )
            {
               sascal_tcomp_macro( 
                     &Ynp, ((char *) &BETA), &INCY,
                     nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }

      Cblacs_gridexit_nocomm(blacsctxt);

      return HSCAL_SUCCESS;
   }

/*
*  Compute descriptor Ad0 for sub( A )
*/
   PB_Cdescribe( N, N, Ai, Aj, Ad, nprow, npcol, myrow, mycol, &Aii, &Ajj,
                 &Ald, &Aimb1, &Ainb1, &Amb, &Anb, &Arow, &Acol, Ad0 );

   Yroc = ( INCY == Yd[M_] ? CROW : CCOLUMN );

   if( notran )
   {
/*
*  Reuse sub( Y ) and/or create vector YA in process columns spanned by sub( A )
*/
      int K_l = 1;
      char *ROWCOL_l = COLUMN;
      char *CONJUG_l = NOCONJG;

      hscal_PB_CInOutV_S( 
                  utyp, ROWCOL_l, N, N, Ad0, K_l, ((char *) &BETA), 
                  Yi, Yj, Yd, &Yroc, ((char**)(&tbeta)), YAd,
                  &YAsum, &YApbY,
                  nprow, npcol, myrow, mycol, tcomp 
      );
/*
*  Replicate sub( X ) in process rows spanned by sub( A ) -> XA
*/
      ROWCOL_l = ROW;
      hscal_PB_CInV_S( 
             type, CONJUG_l, ROWCOL_l,    N, N, Ad0, K_l, Xi, Xj, Xd,
             ( INCX == Xd[M_] ? ROW : COLUMN ), XAd, 
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
   else
   {
/*
*  Reuse sub( Y ) and/or create vector YA in process rows spanned by sub( A )
*/
      int K_l = 1;
      char *ROWCOL_l = ROW;
      char *CONJUG_l = NOCONJG;

      hscal_PB_CInOutV_S( 
            utyp, ROWCOL_l,    N, N, Ad0, K_l, ((char *) &BETA), 
            Yi, Yj, Yd, &Yroc, ((char**)(&tbeta)), YAd,
            &YAsum, &YApbY,
            nprow, npcol, myrow, mycol, tcomp 
      );
/*
*  Replicate sub( X ) in process columns spanned by sub( A ) -> XA
*/
      ROWCOL_l = COLUMN;
      hscal_PB_CInV_S( 
           type, CONJUG_l, ROWCOL_l, N, N, Ad0, K_l, Xi, Xj, Xd,
           ( INCX == Xd[M_] ? ROW : COLUMN ), XAd, 
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }

   one   = type->one;
/*
*  Local matrix-vector multiply iff I own some data
*/
   Aimb1 = Ad0[IMB_ ]; Ainb1 = Ad0[INB_ ]; Amb = Ad0[MB_]; Anb = Ad0[NB_];
   Acol  = Ad0[CSRC_]; Arow  = Ad0[RSRC_];
   Amp = PB_Cnumroc( N, 0, Aimb1, Amb, myrow, Arow, nprow );
   Anq = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );

   if( ( Amp > 0 ) && ( Anq > 0 ) )
   {
      XAld = XAd[LLD_]; YAld = YAd[LLD_];
/*
*  Scale YA in the case sub( Y ) has been reused
*/
      if( notran && !( YApbY ) )
      {
/*
*  YA resides in (a) process column(s)
*/
         if( ( mycol == YAd[CSRC_] ) || ( YAd[CSRC_] < 0 ) )
         {
/*
*  Make sure I own some data and scale YA
*/
            if( Amp > 0 )
            {
               sascal_tcomp_macro( 
                     &Amp, ((char *) tbeta), &ione,
                     nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }
      else if( !( notran ) && !( YApbY ) )
      {
/*
*  YA resides in (a) process row(s)
*/
         if( ( myrow == YAd[RSRC_] ) || ( YAd[RSRC_] < 0 ) )
         {
/*
*  Make sure I own some data and scale YA
*/
            if( Anq > 0 )
            {
               sascal_tcomp_macro( 
                     &Anq, ((char *) tbeta), &YAld,
                     nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }
/*
*  Computational partitioning size is computed as the product of the logical
*  value returned by pilaenv_ and 2 * lcm( nprow, npcol )
*/
      nb = 2 * hscal_pilaenv( C2F_CHAR( &utyp->type ) ) *
           PB_Clcm( ( Arow >= 0 ? nprow : 1 ), ( Acol >= 0 ? npcol : 1 ) );

      if( upper )
      {
         if( notran )
         {
            for( k = 0; k < N; k += nb )
            {
               kb   = N - k; kb = MIN( kb, nb );
               Akp  = PB_Cnumroc( k,  0, Aimb1, Amb, myrow, Arow, nprow );
               Akq  = PB_Cnumroc( k,  0, Ainb1, Anb, mycol, Acol, npcol );
               Anq0 = PB_Cnumroc( kb, k, Ainb1, Anb, mycol, Acol, npcol );
               if( Akp > 0 && Anq0 > 0 )
               {
                  sagemv_tcomp_macro( 
                        TRANS, &Akp, &Anq0, ((char *) &ALPHA),
                        &Ald,
                        &XAld, one,
                        &ione,
                        nprow, npcol, myrow, mycol, tcomp
                  );
               }
               {
                  char *SIDE_l = LEFT;
                  char *UPLO_l = UPPER;
                  int K_l = 1;
                  hscal_PB_Cptrm( 
                   type, utyp, SIDE_l, UPLO_l, &TranOp, &DiagA, kb, K_l,
                   ((char *) &ALPHA), k, k, Ad0,
                   XAld,
                   YAld, hscal_PB_Ctzatrmv_S,
                   nprow, npcol, myrow, mycol, tcomp
                  );
               }
            }
         }
         else
         {
            for( k = 0; k < N; k += nb )
            {
               kb   = N - k; kb = MIN( kb, nb );
               Akp  = PB_Cnumroc( k,  0, Aimb1, Amb, myrow, Arow, nprow );
               Akq  = PB_Cnumroc( k,  0, Ainb1, Anb, mycol, Acol, npcol );
               Anq0 = PB_Cnumroc( kb, k, Ainb1, Anb, mycol, Acol, npcol );
               if( Akp > 0 && Anq0 > 0 )
               {
                  sagemv_tcomp_macro( 
                        TRANS, &Akp, &Anq0, ((char *) &ALPHA),
                        &Ald, &ione,
                        one, &YAld,
                        nprow, npcol, myrow, mycol, tcomp
                  );
               }
               {
                  char *SIDE_l = LEFT;
                  char *UPLO_l = UPPER;
                  int K_l = 1;
                  hscal_PB_Cptrm( 
                    type, utyp, SIDE_l, UPLO_l, &TranOp, &DiagA, kb, K_l,
                    ((char *) &ALPHA), k, k, Ad0,
                    XAld,
                    YAld, hscal_PB_Ctzatrmv_S,
                    nprow, npcol, myrow, mycol, tcomp
                  );
               }
            }
         }
      }
      else
      {
         if( notran )
         {
            for( k = 0; k < N; k += nb )
            {
               kb  = N - k; ktmp = k + ( kb = MIN( kb, nb ) );
               Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
               Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
               {
                  char *SIDE_l = LEFT;
                  char *UPLO_l = LOWER;
                  int K_l = 1;
                  hscal_PB_Cptrm( 
                       type, utyp, SIDE_l, UPLO_l, &TranOp, &DiagA, kb, K_l,
                       ((char *) &ALPHA), k, k, Ad0,
                       XAld,
                       YAld, hscal_PB_Ctzatrmv_S,
                       nprow, npcol, myrow, mycol, tcomp
                   );
               }
               Akp  = PB_Cnumroc( ktmp, 0, Aimb1, Amb, myrow, Arow, nprow );
               Amp0 = Amp - Akp;
               Anq0 = PB_Cnumroc( kb,   k, Ainb1, Anb, mycol, Acol, npcol );
               if( Amp0 > 0 && Anq0 > 0 )
               {
                  sagemv_tcomp_macro( 
                        TRANS, &Amp0, &Anq0, ((char *) &ALPHA),
                        &Ald,
                        &XAld, one,
                        &ione,
                        nprow, npcol, myrow, mycol, tcomp
                  );
               }
            }
         }
         else
         {
            for( k = 0; k < N; k += nb )
            {
               kb  = N - k; ktmp = k + ( kb = MIN( kb, nb ) );
               Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
               Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
               {
                  char *SIDE_l = LEFT;
                  char *UPLO_l = LOWER;
                  int K_l = 1;
                  hscal_PB_Cptrm( 
                       type, utyp, SIDE_l, UPLO_l, &TranOp, &DiagA, kb, K_l,
                       ((char *) &ALPHA), k, k, Ad0,
                       XAld,
                       YAld, hscal_PB_Ctzatrmv_S,
                       nprow, npcol, myrow, mycol, tcomp
                  );
               }
               Akp  = PB_Cnumroc( ktmp, 0, Aimb1, Amb, myrow, Arow, nprow );
               Amp0 = Amp - Akp;
               Anq0 = PB_Cnumroc( kb,   k, Ainb1, Anb, mycol, Acol, npcol );
               if( Amp0 > 0 && Anq0 > 0 )
               {
                  sagemv_tcomp_macro(
                        TRANS, &Amp0, &Anq0, one,
                        &Ald,
                        &ione, one,
                        &YAld,
                        nprow, npcol, myrow, mycol, tcomp
                  );
               }
            }
         }
      }
   }

   if( notran )
   {
/*
*  Combine the partial column results into YA
*/
      if( YAsum && ( Amp > 0 ) )
      {
         top = *PB_Ctop( &ctxt, COMBINE, ROW, TOP_GET );
         hscal_Csgsum2d( 
             blacsctxt, ROW, &top, Amp, 1, YAd[LLD_], myrow,
                   YAd[CSRC_],
             nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
   }
   else
   {
/*
*  Combine the partial row results into YA
*/
      if( YAsum && ( Anq > 0 ) )
      {
         top = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );
         hscal_Csgsum2d( 
             blacsctxt, COLUMN, &top, 1, Anq, YAd[LLD_], YAd[RSRC_],
                   mycol,
             nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
   }
/*
*  sub( Y ) := beta * sub( Y ) + YA (if necessary)
*/
   if( YApbY )
   {
/*
*  Retrieve sub( Y )'s local information: Yii, Yjj, Yrow, Ycol
*/
      PB_Cinfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, &Yii, &Yjj, &Yrow,
                   &Ycol );

      if( INCY == Yd[M_] )
      {
/*
*  sub( Y ) resides in (a) process row(s)
*/
         if( ( myrow == Yrow ) || ( Yrow < 0 ) )
         {
/*
*  Make sure I own some data and scale sub( Y )
*/
            Ynq = PB_Cnumroc( N, Yj, Yd[INB_], Yd[NB_], mycol, Yd[CSRC_],
                              npcol );
            if( Ynq > 0 )
            {
               Yld = Yd[LLD_];
               sascal_tcomp_macro( 
                     &Ynq, ((char *) &BETA), &Yld,
                     nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }
      else
      {
/*
*  sub( Y ) resides in (a) process column(s)
*/
         if( ( mycol == Ycol ) || ( Ycol < 0 ) )
         {
/*
*  Make sure I own some data and scale sub( Y )
*/
            Ynp = PB_Cnumroc( N, Yi, Yd[IMB_], Yd[MB_], myrow, Yd[RSRC_],
                              nprow );
            if( Ynp > 0 )
            {
               sascal_tcomp_macro( 
                     &Ynp, ((char *) &BETA), &INCY,
                     nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }

      if( notran )
      {
         int N_l = 1;
         int IA_l = 0;
         int JA_l = 0;
         char *CONJUG_l = NOCONJG;

         hscal_PB_Cpaxpby_S( 
           utyp, CONJUG_l, N, N_l, one, IA_l, JA_l, YAd, COLUMN, one,
           Yi, Yj, Yd, &Yroc,
           nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
      else
      {
         int N_l = 1;
         int IA_l = 0;
         int JA_l = 0;
         char *CONJUG_l = NOCONJG;

         hscal_PB_Cpaxpby_S( 
           utyp, CONJUG_l, N_l, N, one, IA_l, JA_l, YAd, ROW,    one,
           Yi, Yj, Yd, &Yroc,
           nprow, npcol, myrow, mycol, tcomp, tcomm
         );
      }
   }
/*
*  End of PSATRMV
*/

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_tcomp_tcomm(
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int ALPHA,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psatrmv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                N, 
                ALPHA, IA, JA, DESCA, 
                IX, JX, DESCX, INCX, BETA,
                IY, JY, DESCY, INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_info(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;
   int DIAGI  = *DIAG;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   for (i = 0; i < (nprow*npcol); i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < (nprow*npcol*nprow*npcol); i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psatrmv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                *N, 
                alphai, *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX, betai,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psatrmv_info_i(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;
   int DIAGI  = *DIAG;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.f)
   {
      betai = 0;
   }
   else if (*BETA == 1.f)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psatrmv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                *N, 
                alphai, *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX, betai,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
