
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_psswap_recon_tcomp(
   int n,
   int incx,
   int incy
)
{
    double tcomp = 0;

    int num_of_computers = HMPI_Get_number_of_computers();
    int reconn = (((n/num_of_computers) >= 1) ? (n/num_of_computers) : 1);
    
    hscal_sswap(
         &reconn,
         &incx,
         &incy,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_psswap_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psswap_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_psswap_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psswap_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psswap_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_psswap_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_psswap_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           Xscope, Yscope, * one, * top, tran, * zero;
   int            OneBlock, OneDgrid, RRorCC, Square, Xcol, Xi, XisD, XisR,
                  Xinb1D, XinbD, XisRow, Xii, Xj, Xjj, Xld, Xlinc, Xm, XmyprocD,
                  XmyprocR, Xn, XnbD, XnpD, XnprocsD, XnprocsR, XprocD, XprocR,
                  Xroc, Xrow, Ycol, Yi, Yii, Yinb1D, YinbD, YisD, YisR, YisRow,
                  Yj, Yjj, Yld, Ylinc, Ym, YmyprocD, YmyprocR, Yn, YnbD, YnpD,
                  YnprocsD, YnprocsR, YprocD, YprocR, Yroc, Yrow, cdst, csrc,
                  ctxt, dst, gcdPQ, info, ione=1, k, l, lcmPQ, lcmb,
                  npq, p, q, rdst, rsrc, src, size;
   PBTYP_T        * type;
   PB_VM_T        VM;
   HSCALBLACSCONTEXT   * blacsctxt;

/*
*  .. Local Arrays ..
*/
   int            Xd[DLEN_], Yd[DLEN_];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

/* ..
*  .. Executable Statements ..
*
*/
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
   PB_CargFtoC( IY, JY, DESCY, &Yi, &Yj, Yd );

   if (init_llds_f == 1)
   {
   HSCAL_PXSWAP_INIT_LLDS(
        N, 
        Xd, IX, INCX, 
        Yd, IY, INCY, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Yd[M_]=%d Yd[MB_]=%d myrow=%d Yd[RSRC_]=%d nprow=%d Yd[LLD_]=%d", 
        Yd[M_], Yd[MB_], myrow, Yd[RSRC_], nprow, Yd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 501 + CTXT_ ) : 0 ) ) )
   {
      hscal_PB_Cchkvec( 
           "PSSWAP", "X", N, 1, Xi, Xj, Xd, INCX,  5, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec( 
           "PSSWAP", "Y", N, 1, Yi, Yj, Yd, INCY, 10, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PSSWAP",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( N == 0 ) return HSCAL_SUCCESS;
/*
*  Retrieve process grid information
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*
*  Determine if sub( X ) is distributed or not
*/
   if( ( XisRow = ( INCX == Xd[M_] ) ) != 0 )
      XisD = ( ( Xd[CSRC_] >= 0 ) && ( ( XnprocsD = npcol ) > 1 ) );
   else
      XisD = ( ( Xd[RSRC_] >= 0 ) && ( ( XnprocsD = nprow ) > 1 ) );
/*
*  Determine if sub( Y ) is distributed or not
*/
   if( ( YisRow = ( INCY == Yd[M_] ) ) != 0 )
      YisD = ( ( Yd[CSRC_] >= 0 ) && ( ( YnprocsD = npcol ) > 1 ) );
   else
      YisD = ( ( Yd[RSRC_] >= 0 ) && ( ( YnprocsD = nprow ) > 1 ) );
/*
*  Are sub( X ) and sub( Y ) both row or column vectors ?
*/
   RRorCC = ( ( XisRow && YisRow ) || ( !( XisRow ) && !( YisRow ) ) );
/*
*  XisD && YisD <=> both vector operands are indeed distributed
*/
   if( XisD && YisD )
   {
/*
*  Retrieve sub( X )'s local information: Xii, Xjj, Xrow, Xcol ...
*/
      PB_Cinfog2l( Xi, Xj, Xd, nprow, npcol, myrow, mycol, &Xii, &Xjj,
                   &Xrow, &Xcol );
      if( XisRow )
      {
         XinbD  = Xd[INB_ ]; XnbD     = Xd[NB_ ]; Xld      = Xd[LLD_];
         Xlinc  = Xld;
         XprocD = Xcol;      XmyprocD = mycol;
         XprocR = Xrow;      XmyprocR = myrow;    XnprocsR = nprow;
         XisR   = ( ( Xrow == -1 ) || ( XnprocsR == 1 ) );
         Mfirstnb( Xinb1D, N, Xj, XinbD, XnbD );
      }
      else
      {
         XinbD  = Xd[IMB_ ]; XnbD     = Xd[MB_ ]; Xld      = Xd[LLD_];
         Xlinc  = 1;
         XprocD = Xrow;      XmyprocD = myrow;
         XprocR = Xcol;      XmyprocR = mycol;    XnprocsR = npcol;
         XisR   = ( ( Xcol == -1 ) || ( XnprocsR == 1 ) );
         Mfirstnb( Xinb1D, N, Xi, XinbD, XnbD );
      }
/*
*  Retrieve sub( Y )'s local information: Yii, Yjj, Yrow, Ycol ...
*/
      PB_Cinfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, &Yii, &Yjj,
                   &Yrow, &Ycol );
      if( YisRow )
      {
         YinbD  = Yd[INB_ ]; YnbD     = Yd[NB_ ]; Yld      = Yd[LLD_];
         Ylinc  = Yld;
         YprocD = Ycol;      YmyprocD = mycol;
         YprocR = Yrow;      YmyprocR = myrow;    YnprocsR = nprow;
         YisR   = ( ( Yrow == -1 ) || ( YnprocsR == 1 ) );
         Mfirstnb( Yinb1D, N, Yj, YinbD, YnbD );
      }
      else
      {
         YinbD  = Yd[IMB_ ]; YnbD     = Yd[MB_ ]; Yld      = Yd[LLD_];
         Ylinc  = 1;
         YprocD = Yrow;      YmyprocD = myrow;
         YprocR = Ycol;      YmyprocR = mycol;    YnprocsR = npcol;
         YisR   = ( ( Ycol == -1 ) || ( YnprocsR == 1 ) );
         Mfirstnb( Yinb1D, N, Yi, YinbD, YnbD );
      }
/*
*  Do sub( X ) and sub( Y ) span more than one process ?
*/
      OneDgrid = ( ( XnprocsD ==  1 ) && ( YnprocsD ==  1 ) );
      OneBlock = ( ( Xinb1D   >= N ) && ( Yinb1D   >= N ) );
/*
*  Are sub( X ) and sub( Y ) distributed in the same manner ?
*/
      Square   = ( ( Xinb1D   ==   Yinb1D ) && ( XnbD == YnbD ) &&
                   ( XnprocsD == YnprocsD ) );

      if( !( XisR ) )
      {
/*
*  sub( X ) is distributed but not replicated
*/
         if( YisR )
         {
/*
*  If sub( X ) is not replicated, but sub( Y ) is, a process row or column
*  YprocR need to be selected. It will contain the non-replicated vector to
*  swap sub( X ) with.
*/
            if( RRorCC )
            {
/*
*  sub( X ) and sub( Y ) are both row or column vectors
*/
               if( ( OneDgrid || OneBlock || Square ) && ( XprocD == YprocD ) )
               {
/*
*  sub( X ) and sub( Y ) start in the same process row or column XprocD=YprocD.
*  Enforce a purely local operation by choosing YprocR to be equal to XprocR.
*/
                  YprocR = XprocR;
               }
               else
               {
/*
*  Otherwise, communication has to occur, so choose the next process row or
*  column for YprocR to maximize the number of links, i.e reduce contention.
*/
                  YprocR = MModAdd1( XprocR, XnprocsR );
               }
            }
            else
            {
/*
*  sub( X ) and sub( Y ) are distributed in orthogonal directions, what is
*  chosen for YprocR does not really matter. Select the process origin.
*/
               YprocR = XprocD;
            }
         }
         else
         {
/*
*  Neither sub( X ) nor sub( Y ) are replicated. If I am not in process row or
*  column XprocR and not in process row or column YprocR, then quick return.
*/
            if( ( XmyprocR != XprocR ) && ( YmyprocR != YprocR ) )
            {
               Cblacs_gridexit_nocomm(blacsctxt);
               return HSCAL_SUCCESS;
            }
         }
      }
      else
      {
/*
*  sub( X ) is distributed and replicated (so no quick return possible)
*/
         if( YisR )
         {
/*
*  sub( Y ) is distributed and replicated as well
*/
            if( RRorCC )
            {
/*
*  sub( X ) and sub( Y ) are both row or column vectors
*/
               if( ( OneDgrid || OneBlock || Square ) && ( XprocD == YprocD ) )
               {
/*
*  sub( X ) and sub( Y ) start in the same process row or column XprocD=YprocD.
*  Enforce a purely local operation by choosing XprocR and YprocR to be equal
*  to zero.
*/
                  XprocR = YprocR = 0;
               }
               else
               {
/*
*  Otherwise, communication has to occur, so select YprocR to be zero and the
*  next process row or column for XprocR in order to maximize the number of
*  used links, i.e reduce contention.
*/
                  YprocR = 0;
                  XprocR = MModAdd1( YprocR, YnprocsR );
               }
            }
            else
            {
/*
*  sub( X ) and sub( Y ) are distributed in orthogonal directions, select the
*  origin processes.
*/
               XprocR = YprocD;
               YprocR = XprocD;
            }
         }
         else
         {
/*
*  sub( Y ) is distributed, but not replicated
*/
            if( RRorCC )
            {
/*
*  sub( X ) and sub( Y ) are both row or column vectors
*/
               if( ( OneDgrid || OneBlock || Square ) && ( XprocD == YprocD ) )
               {
/*
*  sub( X ) and sub( Y ) start in the same process row or column XprocD=YprocD.
*  Enforce a purely local operation by choosing XprocR to be equal to YprocR.
*/
                  XprocR = YprocR;
               }
               else
               {
/*
*  Otherwise, communication has to occur, so choose the next process row or
*  column for XprocR to maximize the number of links, i.e reduce contention.
*/
                  XprocR = MModAdd1( YprocR, YnprocsR );
               }
            }
            else
            {
/*
*  sub( X ) and sub( Y ) are distributed in orthogonal directions, what is
*  chosen for XprocR does not really matter. Select the origin process.
*/
               XprocR = YprocD;
            }
         }
      }
/*
*  Even if sub( X ) and/or sub( Y ) are replicated, only two process row or
*  column are active, namely XprocR and YprocR. If any of those operands is
*  replicated, broadcast will occur (unless there is an easy way out).
*/
      type = PB_Cstypeset(); size = type->size;
/*
*  A purely local operation occurs iff the operands start in the same process
*  and, if either the grid is mono-dimensional or there is a single local block
*  to be swapped or if both operands are aligned.
*/
      if( ( (    RRorCC   && ( XprocD == YprocD ) && ( XprocR == YprocR ) ) ||
            ( !( RRorCC ) && ( XprocD == YprocR ) && ( XprocR == YprocD ) ) ) &&
          ( OneDgrid || OneBlock || ( RRorCC && Square ) ) )
      {
         if( ( !XisR &&         ( XmyprocR == XprocR ) &&
               !YisR &&         ( YmyprocR == YprocR ) ) ||
             ( !XisR && YisR && ( YmyprocR == YprocR ) ) ||
             ( !YisR && XisR && ( XmyprocR == XprocR ) ) ||
             (  XisR && YisR                           ) )
         {
            XnpD = PB_Cnumroc( N, 0, Xinb1D, XnbD, XmyprocD, XprocD,
                               XnprocsD );
            YnpD = PB_Cnumroc( N, 0, Yinb1D, YnbD, YmyprocD, YprocD,
                               YnprocsD );
            if( ( XnpD > 0 ) && ( YnpD > 0 ) )
            {
               sswap_tcomp_macro(
                    &XnpD,
                    &Xlinc,
                    &Ylinc,
                    nprow, npcol, myrow, mycol, tcomp_l
               );
            }

            if( RRorCC && XisR && YisR ) 
            {
               Cblacs_gridexit_nocomm(blacsctxt);
               return HSCAL_SUCCESS;
            }
         }
      }
      else if( ( RRorCC && OneDgrid ) || OneBlock || Square )
      {
/*
*  Otherwise, it may be possible to swap the distributed vectors in a single
*  message exchange iff the grid is mono-dimensional and the operands are
*  distributed in the same direction, or there is just one block to be exchanged
*  or if both operands are similarly distributed in their respective direction.
*/
         if( RRorCC && ( XprocR != YprocR ) )
         {
/*
*  Both operands are distributed in the same direction, but reside in different
*  process rows or columns.
*/
            if( XmyprocR == XprocR )
            {
               XnpD = PB_Cnumroc( N, 0, Xinb1D, XnbD, XmyprocD, XprocD,
                                  XnprocsD );
               if( XnpD > 0 )
               {
                  dst = YprocD + MModSub( XmyprocD, XprocD, XnprocsD );
                  dst = MPosMod( dst, YnprocsD );
                  if( XisRow )
                  {
                     hscal_Csgesd2d( 
                          blacsctxt, 1, XnpD, Xld, YprocR, dst,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  }
                  else
                  {
                     hscal_Csgesd2d( 
                          blacsctxt, XnpD, 1, Xld, dst, YprocR,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  }
               }
            }
            if( YmyprocR == YprocR )
            {
               YnpD = PB_Cnumroc( N, 0, Yinb1D, YnbD, YmyprocD, YprocD,
                                  YnprocsD );
               if( YnpD > 0 )
               {
                  dst = XprocD + MModSub( YmyprocD, YprocD, YnprocsD );
                  dst = MPosMod( dst, XnprocsD );
                  if( YisRow )
                  {
                     hscal_Csgesd2d( 
                          blacsctxt, 1, YnpD, Yld, XprocR, dst,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  }
                  else
                  {
                     hscal_Csgesd2d( 
                          blacsctxt, YnpD, 1, Yld, dst, XprocR,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  }
               }
            }
         }
         else
         {
/*
*  General case when just one message needs to be exchanged
*/
            if( XmyprocR == XprocR )
            {
/*
*  The processes owning a piece of sub( X ) send it to the corresponding
*  process owning s piece of sub ( Y ).
*/
               XnpD = PB_Cnumroc( N, 0, Xinb1D, XnbD, XmyprocD, XprocD,
                                  XnprocsD );
               if( XnpD > 0 )
               {
                  dst = YprocD + MModSub( XmyprocD, XprocD, XnprocsD );
                  dst = MPosMod( dst, YnprocsD );
                  if( YisRow ) { rdst = YprocR; cdst = dst; }
                  else         { rdst = dst; cdst = YprocR; }

                  if( ( myrow == rdst ) && ( mycol == cdst ) )
                  {
                     sswap_tcomp_macro( 
                          &XnpD,  &Xlinc, &Ylinc,
                          nprow, npcol, myrow, mycol, tcomp_l
                     );
                  }
                  else
                  {
                     if( XisRow )
                     {
                        hscal_Csgesd2d( 
                             blacsctxt, 1, XnpD, Xld, rdst, cdst,
                             nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                        );
                     }
                     else
                     {
                        hscal_Csgesd2d( 
                             blacsctxt, XnpD, 1, Xld, rdst, cdst,
                             nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                        );
                     }
                  }
               }
            }
            if( YmyprocR == YprocR )
            {
/*
*  The processes owning a piece of sub( Y ) receive the corresponding piece
*  of sub( X ) and send the piece of sub( Y ) they own to the same process.
*/
               YnpD = PB_Cnumroc( N, 0, Yinb1D, YnbD, YmyprocD, YprocD,
                                  YnprocsD );
               if( YnpD > 0 )
               {
                  src = XprocD + MModSub( YmyprocD, YprocD, YnprocsD );
                  src = MPosMod( src, XnprocsD );
                  if( XisRow ) { rsrc = XprocR; csrc = src; }
                  else         { rsrc = src; csrc = XprocR; }

                  if( ( myrow != rsrc ) || ( mycol != csrc ) )
                  {
                     if( XisRow )
                        ;
                     else
                        ;
                     if( YisRow )
                     {
                        hscal_Csgesd2d( 
                             blacsctxt, 1, YnpD, Yld, rsrc, csrc,
                             nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                        );
                     }
                     else
                     {
                        hscal_Csgesd2d( 
                             blacsctxt, YnpD, 1, Yld, rsrc, csrc,
                             nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                        );
                     }
                     scopy_tcomp_macro( 
                          &YnpD, &ione, &Ylinc,
                          nprow, npcol, myrow, mycol, tcomp_l
                     );
                  }
               }
            }
            if( XmyprocR == XprocR )
            {
/*
*  The processes owning a piece of sub( X ) receive the corresponding piece
*  of sub( Y ).
*/
               if( XnpD > 0 )
               {
                  if( ( myrow != rdst ) || ( mycol != cdst ) )
                  {
                     if( YisRow )
                        ;
                     else
                        ;
                     scopy_tcomp_macro( 
                          &XnpD, &ione, &Xlinc,
                          nprow, npcol, myrow, mycol, tcomp_l
                     );
                  }
               }
            }
         }
      }
      else if( ( XmyprocR == XprocR ) || ( YmyprocR == YprocR ) )
      {
/*
*  General case
*/
         tran   = ( RRorCC ? CNOTRAN : CTRAN );
         if( XisRow ) { Xscope = CCOLUMN; Xm = 1; rsrc = XprocR; }
         else         { Xscope = CROW;    Xn = 1; csrc = XprocR; }
         if( YisRow ) { Yscope = CCOLUMN; Ym = 1; rdst = YprocR; }
         else         { Yscope = CROW;    Yn = 1; cdst = YprocR; }
         lcmb   = PB_Clcm( XnprocsD * XnbD, YnprocsD * YnbD );
         one    = type->one; zero = type->zero;
         gcdPQ  = PB_Cgcd( XnprocsD, YnprocsD );
         lcmPQ  = ( XnprocsD / gcdPQ ) * YnprocsD;

         for( k = 0; k < gcdPQ; k++ )
         {
            p = 0; q = k;

            for( l = 0; l < lcmPQ; l++ )
            {
               Xroc = MModAdd( XprocD, p, XnprocsD );
               Yroc = MModAdd( YprocD, q, YnprocsD );

               if( ( XmyprocD == Xroc ) || ( YmyprocD == Yroc ) )
               {
                  XnpD = PB_Cnumroc( N, 0, Xinb1D, XnbD, Xroc, XprocD,
                                     XnprocsD );
                  YnpD = PB_Cnumroc( N, 0, Yinb1D, YnbD, Yroc, YprocD,
                                     YnprocsD );
                  PB_CVMinit( &VM, 0, XnpD, YnpD, Xinb1D, Yinb1D, XnbD, YnbD,
                              p, q, XnprocsD, YnprocsD, lcmb );
                  if( npq = PB_CVMnpq( &VM ) )
                  {
                     if( (    RRorCC   && ( Xroc ==   Yroc ) &&
                           ( XprocR == YprocR ) ) ||
                         ( !( RRorCC ) && ( Xroc == YprocR ) &&
                           ( XprocR == Yroc   ) ) )
                     {
/*
*  If I am at the intersection of the process cross, or simply common to the
*  process rows or columns owning sub( X ) and sub( Y )
*/
                        if( ( YmyprocD == Yroc ) && ( YmyprocR == YprocR ) )
                        {
                           hscal_PB_CVMswp_S( 
                                type, &VM, ROW, &Xscope, &tran, npq,
                                Xlinc, Ylinc,
                                nprow, npcol, myrow, mycol, tcomp_l
                           );
                        }
                     }
                     else
                     {
/*
*  Perform the message exchange: pack the data I own, send it, receive the
*  remote data, and unpack it.
*/
                        if( ( XmyprocR == XprocR ) && ( XmyprocD == Xroc ) )
                        {
                           if( XisRow ) { Xn = npq; }
                           else         { Xm = npq; }
                           if( YisRow ) { Yn = npq; cdst = Yroc; }
                           else         { Ym = npq; rdst = Yroc; }

                           hscal_PB_CVMpack_S( 
                                type, &VM, ROW, &Xscope, PACKING, NOTRAN,
                                npq, 1, one, Xld, zero, Xm,
                                nprow, npcol, myrow, mycol, tcomp_l
                           );
                           
                           hscal_Csgesd2d(
                                blacsctxt, Xm, Xn, Xm, rdst, cdst,
                                nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                           );

                           hscal_PB_CVMpack_S( 
                                type, &VM, ROW, &Xscope, UNPACKING,
                                &tran, npq, 1, zero, Xld, one, Ym,
                                nprow, npcol, myrow, mycol, tcomp_l
                           );
                        }
                        if( ( YmyprocR == YprocR ) && ( YmyprocD == Yroc  ) )
                        {
                           if( XisRow ) { Xn = npq; csrc = Xroc; }
                           else         { Xm = npq; rsrc = Xroc; }
                           if( YisRow ) { Yn = npq; }
                           else         { Ym = npq; }

                           hscal_PB_CVMpack_S( 
                                type, &VM, COLUMN, &Yscope, PACKING,
                                NOTRAN, npq, 1, one, Yld, zero,
                                Ym,
                                nprow, npcol, myrow, mycol, tcomp_l
                           );

                           hscal_Csgesd2d( 
                                blacsctxt, Ym, Yn, Ym, rsrc, csrc,
                                nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                           );

                           hscal_PB_CVMpack_S( 
                                type, &VM, COLUMN, &Yscope, UNPACKING,
                                &tran, npq, 1, zero, Yld, one, Xm,
                                nprow, npcol, myrow, mycol, tcomp_l
                           );
                        }
                     }
                  }
               }
               p = MModAdd1( p, XnprocsD );
               q = MModAdd1( q, YnprocsD );
            }
         }
      }

      if( XisR )
      {
/*
*  Replicate sub( X ) when necessary
*/
         XnpD = PB_Cnumroc( N, 0, Xinb1D, XnbD, XmyprocD, XprocD, XnprocsD );
         if( XnpD > 0 )
         {
            if( XisRow )
            {
               top = PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );
               if( XmyprocR == XprocR )
                  hscal_Csgebs2d( 
                       blacsctxt, COLUMN, top, 1, XnpD, Xld,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
               else
                  hscal_Csgebr2d( 
                       blacsctxt, COLUMN, top, 1, XnpD, Xld, XprocR, XmyprocD,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
            }
            else
            {
               top = PB_Ctop( &ctxt, BCAST, ROW,    TOP_GET );
               if( XmyprocR == XprocR )
                  hscal_Csgebs2d( 
                       blacsctxt, ROW,    top, XnpD, 1, Xld,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
               else
                  hscal_Csgebr2d( 
                       blacsctxt, ROW,    top, XnpD, 1, Xld, XmyprocD, XprocR,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
            }
         }
      }

      if( YisR )
      {
/*
*  Replicate sub( Y ) when necessary
*/
         YnpD = PB_Cnumroc( N, 0, Yinb1D, YnbD, YmyprocD, YprocD, YnprocsD );
         if( YnpD > 0 )
         {
            if( YisRow )
            {
               top = PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );
               if( YmyprocR == YprocR )
                  hscal_Csgebs2d( 
                       blacsctxt, COLUMN, top, 1, YnpD, Yld,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
               else
                  hscal_Csgebr2d( 
                       blacsctxt, COLUMN, top, 1, YnpD, Yld, YprocR, YmyprocD,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
            }
            else
            {
               top = PB_Ctop( &ctxt, BCAST, ROW,    TOP_GET );
               if( YmyprocR == YprocR )
                  hscal_Csgebs2d( 
                       blacsctxt, ROW,    top, YnpD, 1, Yld,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
               else
                  hscal_Csgebr2d( 
                       blacsctxt, ROW,    top, YnpD, 1, Yld, YmyprocD, YprocR,
                       nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                  );
            }
         }
      }
   }
   else if( !( XisD ) && YisD )
   {
/*
*  sub( X ) is not distributed and sub( Y ) is distributed.
*/
      hscal_PB_CpswapND_S(
           PB_Cstypeset(), N, Xi, Xj, Xd, INCX,
           Yi, Yj, Yd, INCY,
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
   }
   else if( XisD && !( YisD ) )
   {
/*
*  sub( X ) is distributed and sub( Y ) is not distributed.
*/
      hscal_PB_CpswapND_S( 
           PB_Cstypeset(), N, Yi, Yj, Yd, INCY,
           Xi, Xj, Xd, INCX,
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
   }
   else
   {
/*
*  Neither sub( X ) nor sub( Y ) are distributed.
*/
      hscal_PB_CpswapNN_S( 
           PB_Cstypeset(), N, Xi, Xj, Xd, INCX,
           Yi, Yj, Yd, INCY,
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
   }

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psswap_tcomp_tcomm(
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psswap_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N, 
                IX, JX, DESCX, INCX,
                IY, JY, DESCY, INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psswap_info(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psswap_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psswap_info_i(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psswap_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
