
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "../../include/scalapack.h"

/*------------------------------------------------------------------*/

double hscal_pssyr2_recon_tcomp(
   int uploi,
   int n,
   int alphai,
   int incx,
   int incy
)
{
    double tcomp = 0;

    char uploc   = uploi;
    char *uplo   = &uploc;

    int reconn   = hscal_pilaenv_recon("S");
    int lda      = reconn;

    float alpha  = alphai;

    hscal_ssyr2(
         uplo,
         &reconn,
         ((char*)&alpha),
         &incx,
         &incy,
         &lda,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pssyr2_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int ALPHA,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pssyr2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, N, 
          ALPHA, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          IA, JA, DESCA, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pssyr2_my_tcomm(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int ALPHA,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pssyr2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, N, 
          ALPHA, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pssyr2_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLOI,
   int N,
   int ALPHA,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pssyr2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, N, 
          ALPHA, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pssyr2_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int ALPHAI,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           UploA;
   int            Acol, Ai, Aii, Aimb1, Ainb1, Aj, Ajj, Akp, Akq, Ald, Amb,
                  Amp, Amp0, Anb, Anq, Anq0, Arow, XCfr, XCld, XRfr, XRld,
                  Xi, Xj, YCfr, YCld, YRfr, YRld, Yi, Yj, ctxt, info, ione=1,
                  k, kb, ktmp, nb, size, upper;
   PBTYP_T        * type;
/*
*  .. Local Arrays ..
*/
   int            Ad  [DLEN_], Ad0 [DLEN_], XCd0[DLEN_], XRd0[DLEN_], Xd[DLEN_],
                  YCd0[DLEN_], YRd0[DLEN_], Yd  [DLEN_];

   char UPLOC     = UPLOI, *UPLO = &UPLOC;
   float ALPHA    = ALPHAI;

/* ..
*  .. Executable Statements ..
*
*/
   upper = ( ( UploA  = Mupcase( F2C_CHAR( UPLO )[0] ) ) == CUPPER );
   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
   PB_CargFtoC( IY, JY, DESCY, &Yi, &Yj, Yd );

   if (init_llds_f == 1)
   {
   HSCAL_PXSYR2_INIT_LLDS(
        N, 
        Ad, 
        Xd, IX, INCX, 
        Yd, IY, INCY, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d",
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Yd[M_]=%d Yd[MB_]=%d myrow=%d Yd[RSRC_]=%d nprow=%d Yd[LLD_]=%d", 
        Yd[M_], Yd[MB_], myrow, Yd[RSRC_], nprow, Yd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 701 + CTXT_ ) : 0 ) ) )
   {
      if( ( !upper ) && ( UploA != CLOWER ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PSSYR2", "Illegal UPLO = %c\n", UploA 
         );
         info = -1;
      }
      hscal_PB_Cchkvec(
           "PSSYR2", "X", N, 2, Xi, Xj, Xd, INCX,  7, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec(
           "PSSYR2", "Y", N, 2, Yi, Yj, Yd, INCY, 12, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkmat(
           "PSSYR2", "A", N, 2, N, 2, Ai, Aj, Ad, 17, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PSSYR2",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( (N == 0) || ( ALPHA == ZERO ) ) return HSCAL_SUCCESS;

/*
*  Get type structure
*/
   type = PB_Cstypeset();

/*
*  Compute descriptor Ad0 for sub( A )
*/
   PB_Cdescribe( N, N, Ai, Aj, Ad, nprow, npcol, myrow, mycol, &Aii, &Ajj,
                 &Ald, &Aimb1, &Ainb1, &Amb, &Anb, &Arow, &Acol, Ad0 );
/*
*  Replicate sub( X ) in process rows (XR) and process columns (XC) spanned by
*  sub( A )
*/
   if( INCX == Xd[M_] )
   {
      char *CONJUG_l   = NOCONJG;
      char *ROWCOL_l_1 = ROW, *ROWCOL_l_2 = ROW; 
      int K_l = 1;
      hscal_PB_CInV_D( 
             type, CONJUG_l, ROWCOL_l_1,    N, N, Ad0, K_l, Xi, Xj,
             Xd,   ROWCOL_l_2,   XRd0, 
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );

      ROWCOL_l_1 = COLUMN;
      ROWCOL_l_2 = ROW;
      hscal_PB_CInV_D( 
             type, CONJUG_l, ROWCOL_l_1, N, N, Ad0, K_l, 0,  0,
             XRd0, ROWCOL_l_2, XCd0, 
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
   else
   {
      char *CONJUG_l   = NOCONJG;
      char *ROWCOL_l_1 = COLUMN, *ROWCOL_l_2 = COLUMN; 
      int K_l = 1;
      hscal_PB_CInV_D( 
        type, CONJUG_l, ROWCOL_l_1, N, N, Ad0, K_l, Xi, Xj,
        Xd,   ROWCOL_l_2, XCd0, 
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );

      ROWCOL_l_1 = ROW;
      ROWCOL_l_2 = COLUMN;
      hscal_PB_CInV_D( 
             type, CONJUG_l, ROWCOL_l_1,    N, N, Ad0, K_l, 0,  0,
             XCd0, ROWCOL_l_2, XRd0,
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
/*
*  Replicate sub( Y ) in process rows (YR) and process columns (YC) spanned by
*  sub( A )
*/
   if( INCY == Yd[M_] )
   {
      char *CONJUG_l   = NOCONJG;
      char *ROWCOL_l_1 = ROW, *ROWCOL_l_2 = ROW; 
      int K_l = 1;
      hscal_PB_CInV_D( 
             type, CONJUG_l, ROWCOL_l_1, N, N, Ad0, K_l, Yi, Yj,
             Yd,   ROWCOL_l_2, YRd0,
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );

      ROWCOL_l_1 = COLUMN;
      ROWCOL_l_2 = ROW;
      hscal_PB_CInV_D( 
             type, CONJUG_l, ROWCOL_l_1, N, N, Ad0, K_l, 0,  0,
             YRd0, ROWCOL_l_2, YCd0, 
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
   else
   {
      char *CONJUG_l   = NOCONJG;
      char *ROWCOL_l_1 = COLUMN, *ROWCOL_l_2 = COLUMN; 
      int K_l = 1;
      hscal_PB_CInV_D(
             type, CONJUG_l, ROWCOL_l_1, N, N, Ad0, K_l, Yi, Yj,
             Yd,   ROWCOL_l_2, YCd0, 
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );

      ROWCOL_l_1 = ROW;
      ROWCOL_l_2 = COLUMN;
      hscal_PB_CInV_D(
             type, CONJUG_l, ROWCOL_l_1, N, N, Ad0, K_l, 0,  0,
             YCd0, ROWCOL_l_2, YRd0, 
             nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
/*
*  Local rank-2 update if I own some data
*/
   Amp = PB_Cnumroc( N, 0, Aimb1, Amb, myrow, Arow, nprow );
   Anq = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );

   if( ( Amp > 0 ) && ( Anq > 0 ) )
   {
      size = type->size;

      XCld = XCd0[LLD_]; YCld = YCd0[LLD_];
      XRld = XRd0[LLD_]; YRld = YRd0[LLD_];
/*
*  Computational partitioning size is computed as the product of the logical
*  value returned by pilaenv_ and 2 * lcm( nprow, npcol ).
*/
      nb = 2 * hscal_pilaenv( C2F_CHAR( &type->type ) ) *
           PB_Clcm( ( Arow >= 0 ? nprow : 1 ), ( Acol >= 0 ? npcol : 1 ) );
      if( upper )
      {
         for( k = 0; k < N; k += nb )
         {
            kb   = N - k; kb = MIN( kb, nb );
            Akp  = PB_Cnumroc( k,  0, Aimb1, Amb, myrow, Arow, nprow );
            Akq  = PB_Cnumroc( k,  0, Ainb1, Anb, mycol, Acol, npcol );
            Anq0 = PB_Cnumroc( kb, k, Ainb1, Anb, mycol, Acol, npcol );
            if( Akp > 0 && Anq0 > 0 )
            {
               sger_tcomp_macro( 
                   &Akp, &Anq0, ((char *) &ALPHA), &ione,
                   &YRld,
                   &Ald,
                   nprow, npcol, myrow, mycol, tcomp
               );
               sger_tcomp_macro( 
                   &Akp, &Anq0, ((char *) &ALPHA), &ione,
                   &XRld,
                   &Ald,
                   nprow, npcol, myrow, mycol, tcomp
               );
            }
            hscal_PB_Cpsyr2(
                     type, UPPER, kb, 1, ((char *) &ALPHA),
                     XCld,
                     XRld,
                     YCld,
                     YRld,
                     k, k, Ad0, hscal_PB_Ctzsyr2_S,
                     nprow, npcol, myrow, mycol, tcomp
            );
         }
      }
      else
      {
         for( k = 0; k < N; k += nb )
         {
            kb  = N - k; ktmp = k + ( kb = MIN( kb, nb ) );
            Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
            Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
            hscal_PB_Cpsyr2(
                     type, LOWER, kb, 1, ((char *) &ALPHA),
                     XCld,
                     XRld,
                     YCld,
                     YRld,
                     k, k, Ad0, hscal_PB_Ctzsyr2_S,
                     nprow, npcol, myrow, mycol, tcomp
            );
            Akp  = PB_Cnumroc( ktmp, 0, Aimb1, Amb, myrow, Arow, nprow );
            Amp0 = Amp - Akp;
            Anq0 = PB_Cnumroc( kb,   k, Ainb1, Anb, mycol, Acol, npcol );
            if( Amp0 > 0 && Anq0 > 0 )
            {
               sger_tcomp_macro(
                   &Amp0, &Anq0, ((char *) &ALPHA),
                   &ione,
                   &YRld,
                   &Ald,
                   nprow, npcol, myrow, mycol, tcomp
               );
               sger_tcomp_macro(
                   &Amp0, &Anq0, ((char *) &ALPHA),
                   &ione,
                   &XRld,
                   &Ald,
                   nprow, npcol, myrow, mycol, tcomp
               );
            }
         }
      }
   }

/*
*  End of PSSYR2
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pssyr2_tcomp_tcomm(
   int UPLOI,
   int N,
   int ALPHA,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pssyr2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, N, 
                ALPHA, 
                IX, JX, DESCX, INCX,
                IY, JY, DESCY, INCY,
                IA, JA, DESCA, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pssyr2_info(
   F_CHAR_T UPLO,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI = *UPLO;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pssyr2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, *N, 
                alphai,
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                *IA, *JA, DESCA, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pssyr2_info_i(
   F_CHAR_T UPLO,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI = *UPLO;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pssyr2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, *N, 
                alphai,
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                *IA, *JA, DESCA, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
