
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzdotu_recon_tcomp(
   int n,
   int incx,
   int incy
)
{
    double tcomp = 0;

    int num_of_computers = HMPI_Get_number_of_computers();
    int reconn = (((n/num_of_computers) >= 1) ? (n/num_of_computers) : 1);
    
    hscal_zdotu(
         &reconn,
         &incx,
         &incy,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pzdotu_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzdotu_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzdotu_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzdotu_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdotu_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pzdotu_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, 
          IX, JX, DESCX, INCX,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pzdotu_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           scope, * top;
   int            OneBlock, OneDgrid, RRorCC, Square, Xcol, Xi, Xii, XinbD,
                  Xinb1D, XisD, XisR, XisRow, Xj, Xjj, Xld, Xlinc, XmyprocD,
                  XmyprocR, XnbD, XnpD, XnprocsD, XnprocsR, XprocD, XprocR,
                  Xrow, Ycol, Yi, Yii, YinbD, Yinb1D, YisD, YisR, YisRow, Yj,
                  Yjj, Yld, Ylinc, YmyprocD, YmyprocR, YnbD, YnpD, YnprocsD,
                  YnprocsR, YprocD, YprocR, Yrow, cdst, csrc, ctxt, dst, info,
                  ione=1, rdst, rsrc, size, src;
   PBTYP_T        * type;
   HSCALBLACSCONTEXT   * blacsctxt;

/*
*  .. Local Arrays ..
*/
   int            Xd[DLEN_], Yd[DLEN_], dbuf[ DLEN_ ];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

/* ..
*  .. Executable Statements ..
*
*/
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
   PB_CargFtoC( IY, JY, DESCY, &Yi, &Yj, Yd );

   if (init_llds_f == 1)
   {
   HSCAL_PXDOTU_INIT_LLDS(
        N,
        Xd, IX, INCX, 
        Yd, IY, INCY, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Yd[M_]=%d Yd[MB_]=%d myrow=%d Yd[RSRC_]=%d nprow=%d Yd[LLD_]=%d", 
        Yd[M_], Yd[MB_], myrow, Yd[RSRC_], nprow, Yd[LLD_]
   );

/*
*  Test the input parameters
*/
   if( !( info = ( ( nprow == -1 ) ? -( 601 + CTXT_ ) : 0 ) ) )
   {
      hscal_PB_Cchkvec( 
           "PZDOTU", "X", N, 1, Xi, Xj, Xd, INCX,  6, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec( 
           "PZDOTU", "Y", N, 1, Yi, Yj, Yd, INCY, 11, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info ) 
   { 
      hscal_PB_Cabort( 
           "PZDOTU", info,
           nprow, npcol, myrow, mycol
      ); 

      return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( N == 0 ) return HSCAL_SUCCESS;

/*
*  Handle degenerate case
*/
   if( ( N == 1 ) && ( ( Xd[ M_ ] == 1 ) || ( Yd[ M_ ] == 1 ) ) )
   {
      type = PB_Cztypeset();
      hscal_PB_Cpdot11_Z( 
           type, N, Xi, Xj, Xd, INCX,
           Yi, Yj, Yd, INCY, hscal_zvvdotu,
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );

      return HSCAL_SUCCESS;
   }

/*
*  Retrieve process grid information
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*
*  Start the operations
*/

/*
*  Determine if sub( X ) is distributed or not
*/
   if( ( XisRow = ( INCX == Xd[M_] ) ) != 0 )
      XisD = ( ( Xd[CSRC_] >= 0 ) && ( ( XnprocsD = npcol ) > 1 ) );
   else
      XisD = ( ( Xd[RSRC_] >= 0 ) && ( ( XnprocsD = nprow ) > 1 ) );
/*
*  Determine if sub( Y ) is distributed or not
*/
   if( ( YisRow = ( INCY == Yd[M_] ) ) != 0 )
      YisD = ( ( Yd[CSRC_] >= 0 ) && ( ( YnprocsD = npcol ) > 1 ) );
   else
      YisD = ( ( Yd[RSRC_] >= 0 ) && ( ( YnprocsD = nprow ) > 1 ) );
/*
*  Are sub( X ) and sub( Y ) both row or column vectors ?
*/
   RRorCC = ( ( XisRow && YisRow ) || ( !( XisRow ) && !( YisRow ) ) );
/*
*  XisD && YisD <=> both vector operands are indeed distributed
*/
   if( XisD && YisD )
   {
/*
*  Retrieve sub( X )'s local information: Xii, Xjj, Xrow, Xcol
*/
      PB_Cinfog2l( Xi, Xj, Xd, nprow, npcol, myrow, mycol, &Xii, &Xjj,
                   &Xrow, &Xcol );
      if( XisRow )
      {
         XinbD  = Xd[INB_]; XnbD     = Xd[NB_];
         Xld    = Xd[LLD_]; Xlinc    = Xld;
         XprocD = Xcol;     XmyprocD = mycol;
         XprocR = Xrow;     XmyprocR = myrow;   XnprocsR = nprow;
         XisR   = ( ( Xrow == -1 ) || ( XnprocsR == 1 ) );
         Mfirstnb( Xinb1D, N, Xj, XinbD, XnbD );
      }
      else
      {
         XinbD  = Xd[IMB_]; XnbD     = Xd[MB_];
         Xld    = Xd[LLD_]; Xlinc    = 1;
         XprocD = Xrow;     XmyprocD = myrow;
         XprocR = Xcol;     XmyprocR = mycol;   XnprocsR = npcol;
         XisR   = ( ( Xcol == -1 ) || ( XnprocsR == 1 ) );
         Mfirstnb( Xinb1D, N, Xi, XinbD, XnbD );
      }
/*
*  Retrieve sub( Y )'s local information: Yii, Yjj, Yrow, Ycol
*/
      PB_Cinfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, &Yii, &Yjj,
                   &Yrow, &Ycol );
      if( YisRow )
      {
         YinbD  = Yd[INB_]; YnbD     = Yd[NB_];
         Yld    = Yd[LLD_]; Ylinc    = Yld;
         YprocD = Ycol;     YmyprocD = mycol;
         YprocR = Yrow;     YmyprocR = myrow;   YnprocsR = nprow;
         YisR   = ( ( Yrow == -1 ) || ( YnprocsR == 1 ) );
         Mfirstnb( Yinb1D, N, Yj, YinbD, YnbD );
      }
      else
      {
         YinbD  = Yd[IMB_]; YnbD     = Yd[MB_];
         Yld    = Yd[LLD_]; Ylinc    = 1;
         YprocD = Yrow;     YmyprocD = myrow;
         YprocR = Ycol;     YmyprocR = mycol;   YnprocsR = npcol;
         YisR   = ( ( Ycol == -1 ) || ( YnprocsR == 1 ) );
         Mfirstnb( Yinb1D, N, Yi, YinbD, YnbD );
      }
/*
*  Do sub( X ) and sub( Y ) span more than one process ?
*/
      OneDgrid = ( ( XnprocsD ==  1 ) && ( YnprocsD ==  1 ) );
      OneBlock = ( ( Xinb1D   >= N ) && ( Yinb1D   >= N ) );
/*
*  Are sub( X ) and sub( Y ) distributed in the same manner ?
*/
      Square   = ( ( Xinb1D   ==   Yinb1D ) && ( XnbD == YnbD ) &&
                   ( XnprocsD == YnprocsD ) );

      if( !( XisR ) )
      {
/*
*  sub( X ) is not replicated
*/
         if( YisR )
         {
/*
*  If sub( X ) is not replicated, but sub( Y ) is, a process row or column
*  YprocR need to be selected. It will contain the non-replicated vector used
*  to perform the dot product computation.
*/
            if( RRorCC )
            {
/*
*  sub( X ) and sub( Y ) are both row or column vectors
*/
               if( ( OneDgrid || OneBlock || Square ) && ( XprocD == YprocD ) )
               {
/*
*  sub( X ) and sub( Y ) start in the same process row or column XprocD=YprocD.
*  Enforce a purely local operation by choosing YprocR to be equal to XprocR.
*/
                  YprocR = XprocR;
               }
               else
               {
/*
*  Otherwise, communication has to occur, so choose the next process row or
*  column for YprocR to maximize the number of links, i.e reduce contention.
*/
                  YprocR = MModAdd1( XprocR, XnprocsR );
               }
            }
            else
            {
/*
*  sub( X ) and sub( Y ) are distributed in orthogonal directions, what is
*  chosen for YprocR does not really matter. Select the process origin.
*/
               YprocR = XprocD;
            }
         }
         else
         {
/*
*  Neither sub( X ) nor sub( Y ) are replicated. If I am not in process row or
*  column XprocR and not in process row or column YprocR, then quick return.
*/
            if( ( XmyprocR != XprocR ) && ( YmyprocR != YprocR ) )
            {
               Cblacs_gridexit_nocomm(blacsctxt);
               return HSCAL_SUCCESS;
            }
         }
      }
      else
      {
/*
*  sub( X ) is distributed and replicated (so no quick return possible)
*/
         if( YisR )
         {
/*
*  sub( Y ) is distributed and replicated as well
*/
            if( RRorCC )
            {
/*
*  sub( X ) and sub( Y ) are both row or column vectors
*/
               if( ( OneDgrid || OneBlock || Square ) && ( XprocD == YprocD ) )
               {
/*
*  sub( X ) and sub( Y ) start in the same process row or column XprocD=YprocD.
*  Enforce a purely local operation by choosing XprocR and YprocR to be equal
*  to zero.
*/
                  XprocR = YprocR = 0;
               }
               else
               {
/*
*  Otherwise, communication has to occur, so select YprocR to be zero and the
*  next process row or column for XprocR in order to maximize the number of
*  used links, i.e reduce contention.
*/
                  YprocR = 0;
                  XprocR = MModAdd1( YprocR, YnprocsR );
               }
            }
            else
            {
/*
*  sub( X ) and sub( Y ) are distributed in orthogonal directions, select the
*  origin processes.
*/
               XprocR = YprocD;
               YprocR = XprocD;
            }
         }
         else
         {
/*
*  sub( Y ) is distributed, but not replicated
*/
            if( RRorCC )
            {
/*
*  sub( X ) and sub( Y ) are both row or column vectors
*/
               if( ( OneDgrid || OneBlock || Square ) && ( XprocD == YprocD ) )
               {
/*
*  sub( X ) and sub( Y ) start in the same process row or column XprocD=YprocD.
*  Enforce a purely local operation by choosing XprocR to be equal to YprocR.
*/
                  XprocR = YprocR;
               }
               else
               {
/*
*  Otherwise, communication has to occur, so choose the next process row or
*  column for XprocR to maximize the number of links, i.e reduce contention.
*/
                  XprocR = MModAdd1( YprocR, YnprocsR );
               }
            }
            else
            {
/*
*  sub( X ) and sub( Y ) are distributed in orthogonal directions, what is
*  chosen for XprocR does not really matter. Select the origin process.
*/
               XprocR = YprocD;
            }
         }
      }
/*
*  Even if sub( X ) and/or sub( Y ) are replicated, only two process row or
*  column are active, namely XprocR and YprocR. If any of those operands is
*  replicated, broadcast will occur (unless there is an easy way out).
*/
      type = PB_Cztypeset(); size = type->size;
/*
*  A purely operation occurs iff the operands start in the same process and if
*  either the grid is mono-dimensional or there is a single local block to be
*  operated with or if both operands are aligned.
*/
      if( ( (    RRorCC   && ( XprocD == YprocD ) && ( XprocR == YprocR ) ) ||
            ( !( RRorCC ) && ( XprocD == YprocR ) && ( XprocR == YprocD ) ) ) &&
          ( OneDgrid || OneBlock || ( RRorCC && Square ) ) )
      {
         if( ( !XisR &&         ( XmyprocR == XprocR ) &&
               !YisR &&         ( YmyprocR == YprocR ) ) ||
             ( !XisR && YisR && ( YmyprocR == YprocR ) ) ||
             ( !YisR && XisR && ( XmyprocR == XprocR ) ) ||
             (  XisR && YisR                           ) )
         {
            XnpD = PB_Cnumroc( N, 0, Xinb1D, XnbD, XmyprocD, XprocD,
                               XnprocsD );
            YnpD = PB_Cnumroc( N, 0, Yinb1D, YnbD, YmyprocD, YprocD,
                               YnprocsD );
            if( ( XnpD > 0 ) && ( YnpD > 0 ) )
            {
               zvvdotu_tcomp_macro( 
                      &XnpD, &Xlinc, &Ylinc,
                      nprow, npcol, myrow, mycol, tcomp_l
               );
            }
         }
/*
*  Combine the local results in sub( X )'s scope
*/
         if( ( XisR && YisR ) || ( XmyprocR == XprocR ) )
         {
            scope = ( XisRow ? CROW : CCOLUMN );
            top = PB_Ctop( &ctxt, COMBINE, &scope, TOP_GET );
            hscal_Czgsum2d( 
                 blacsctxt, &scope, top, 1, 1, 1, -1, 0,
                 nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
            );
         }
         if( RRorCC && XisR && YisR )
         {
            Cblacs_gridexit_nocomm(blacsctxt);
            return HSCAL_SUCCESS;
         }
      }
      else if( ( RRorCC && OneDgrid ) || OneBlock || Square )
      {
/*
*  Otherwise, it may be possible to compute the desired dot-product in a single
*  message exchange iff the grid is mono-dimensional and the operands are
*  distributed in the same direction, or there is just one block to be exchanged
*  or if both operands are similarly distributed in their respective direction.
*/
         if( ( YmyprocR == YprocR ) )
         {
/*
*  The processes owning a piece of sub( Y ) send it to the corresponding
*  process owning s piece of sub ( X ).
*/
            YnpD = PB_Cnumroc( N, 0, Yinb1D, YnbD, YmyprocD, YprocD,
                               YnprocsD );
            if( YnpD > 0 )
            {
               dst = XprocD + MModSub( YmyprocD, YprocD, YnprocsD );
               dst = MPosMod( dst, XnprocsD );
               if( XisRow ) { rdst = XprocR; cdst = dst; }
               else         { rdst = dst; cdst = XprocR; }

               if( ( myrow == rdst ) && ( mycol == cdst ) )
               {
                  zvvdotu_tcomp_macro( 
                         &YnpD, &Xlinc, &Ylinc,
                         nprow, npcol, myrow, mycol, tcomp_l 
                  );
               }
               else
               {
                  if( YisRow )
                     hscal_Czgesd2d( 
                          blacsctxt, 1, YnpD, Yld, rdst, cdst,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  else
                     hscal_Czgesd2d( 
                          blacsctxt, YnpD, 1, Yld, rdst, cdst,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
               }
            }
         }
         if( XmyprocR == XprocR )
         {
/*
*  The processes owning a piece of sub( X ) receive the corresponding local
*  piece of sub( Y ), compute the local dot product and combine the results
*  within sub( X )'s scope.
*/
            XnpD = PB_Cnumroc( N, 0, Xinb1D, XnbD, XmyprocD, XprocD,
                               XnprocsD );
            if( XnpD > 0 )
            {
               src = YprocD + MModSub( XmyprocD, XprocD, XnprocsD );
               src = MPosMod( src, YnprocsD );
               if( YisRow ) { rsrc = YprocR; csrc = src; }
               else         { rsrc = src; csrc = YprocR; }
               if( ( myrow != rsrc ) || ( mycol != csrc ) )
               {
                  if( YisRow )
                     ;
                  else
                     ;
                  zvvdotu_tcomp_macro( 
                         &XnpD, &Xlinc, &ione,
                         nprow, npcol, myrow, mycol, tcomp_l
                  );
               }
            }
            if( XisRow )
            {
               top = PB_Ctop( &ctxt, COMBINE, ROW, TOP_GET );
               hscal_Czgsum2d( 
                    blacsctxt, ROW,    top, 1, 1, 1, -1, 0,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
            }
            else
            {
               top = PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );
               hscal_Czgsum2d( 
                    blacsctxt, COLUMN, top, 1, 1, 1, -1, 0,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
            }
         }
      }
      else
      {
/*
*  General case, copy sub( Y ) within sub( X )'s scope, compute the local
*  results and combine them within sub( X )'s scope.
*/
         XnpD = PB_Cnumroc( N, 0, Xinb1D, XnbD, XmyprocD, XprocD, XnprocsD );

         if( XisRow )
         {
            PB_Cdescset( dbuf, 1, N, 1, Xinb1D, 1, XnbD, XprocR, XprocD, ctxt,
                         1 );
         }
         else
         {
            PB_Cdescset( dbuf, N, 1, Xinb1D, 1, XnbD, 1, XprocD, XprocR, ctxt,
                         MAX( 1, XnpD ) );
         }
         if( ( XmyprocR == XprocR ) && ( XnpD > 0 ) )
            ;

         if( YisRow )
         {
            hscal_PB_Cpaxpby_Z( 
                 type, NOCONJG, 1, N, type->one, Yi, Yj,
                 Yd, ROW, type->zero, 0, 0, dbuf, ( XisRow ? ROW :
                 COLUMN ),
                 nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
            );
         }
         else
         {
            hscal_PB_Cpaxpby_Z( 
                 type, NOCONJG, N, 1, type->one, Yi, Yj,
                 Yd, COLUMN, type->zero, 0, 0, dbuf, ( XisRow ?
                 ROW : COLUMN ),
                 nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
            );
         }

         if( XmyprocR == XprocR )
         {
            if( XnpD > 0 )
            {
               zvvdotu_tcomp_macro( 
               &XnpD, &Xlinc, &ione,
               nprow, npcol, myrow, mycol, tcomp_l
               );
            }
            if( XisRow )
            {
               top = PB_Ctop( &ctxt, COMBINE, ROW,    TOP_GET );
               hscal_Czgsum2d( 
                    blacsctxt, ROW,    top, 1, 1, 1, -1, 0,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
            }
            else
            {
               top = PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );
               hscal_Czgsum2d( 
                    blacsctxt, COLUMN, top, 1, 1, 1, -1, 0,
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
               );
            }
         }
      }
/*
*  Send the DOT product result within sub( Y )'s scope
*/
      if( XisR || YisR )
      {
/*
*  Either sub( X ) or sub( Y ) are replicated, so that every process should have
*  the result -> broadcast it orthogonally from sub( X )'s direction.
*/
         if( XisRow )
         {
           top = PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );
           if( XmyprocR == XprocR )
              hscal_Czgebs2d( 
                   blacsctxt, COLUMN, top, 1, 1, 1,
                   nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
              );
           else
              hscal_Czgebr2d( 
                   blacsctxt, COLUMN, top, 1, 1, 1, XprocR,
                   XmyprocD,
                   nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
              );
         }
         else
         {
           top = PB_Ctop( &ctxt, BCAST, ROW,    TOP_GET );
           if( XmyprocR == XprocR )
              hscal_Czgebs2d( 
                   blacsctxt, ROW,    top, 1, 1, 1,
                   nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
              );
           else
              hscal_Czgebr2d( 
                   blacsctxt, ROW,    top, 1, 1, 1, XmyprocD,
                   XprocR,
                   nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
              );
         }
      }
      else
      {
/*
*  Neither sub( X ) nor sub( Y ) are replicated
*/
         if( RRorCC )
         {
/*
*  Both sub( X ) are distributed in the same direction -> the process row or
*  column XprocR sends the result to the process row or column YprocR.
*/
            if( XprocR != YprocR )
            {
               if( XmyprocR == XprocR )
               {
                  if( XisRow )
                     hscal_Czgesd2d( 
                          blacsctxt, 1, 1, 1, YprocR,
                          YmyprocD,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  else
                     hscal_Czgesd2d( 
                          blacsctxt, 1, 1, 1, YmyprocD,
                          YprocR,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
               }
               else if( YmyprocR == YprocR )
               {
                  if( XisRow )
                     ;
                  else
                     ;
               }
            }
         }
         else
         {
/*
*  Otherwise, the process at the intersection of sub( X )'s and sub( Y )'s
*  scope, broadcast the result within sub( Y )'s scope.
*/
            if( YmyprocR == YprocR )
            {
               if( YisRow )
               {
                  top = PB_Ctop( &ctxt, BCAST, ROW,    TOP_GET );
                  if( YmyprocD == XprocR )
                     hscal_Czgebs2d( 
                          blacsctxt, ROW,    top, 1, 1, 1,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  else
                     hscal_Czgebr2d( 
                          blacsctxt, ROW,    top, 1, 1, 1,
                          YprocR, XprocR,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
               }
               else
               {
                  top = PB_Ctop( &ctxt, BCAST, COLUMN, TOP_GET );
                  if( YmyprocD == XprocR )
                     hscal_Czgebs2d( 
                          blacsctxt, COLUMN, top, 1, 1, 1,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
                  else
                     hscal_Czgebr2d( 
                          blacsctxt, COLUMN, top, 1, 1, 1,
                          XprocR, YprocR,
                          nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                     );
               }
            }
         }
      }
   }
   else if( !( XisD ) && YisD )
   {
/*
*  sub( X ) is not distributed and sub( Y ) is distributed.
*/
      type = PB_Cztypeset();
      hscal_PB_CpdotND_Z( 
           type, N, Xi, Xj, Xd, INCX,
           Yi, Yj, Yd, INCY, hscal_zvvdotu,
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
   }
   else if( XisD && !( YisD ) )
   {
/*
*  sub( X ) is distributed and sub( Y ) is not distributed.
*/
      type = PB_Cztypeset();
/*
*     Compute DOT := sub( Y )**H * sub( X )
*/
      hscal_PB_CpdotND_Z( 
           type, N, Yi, Yj, Yd, INCY,
           Xi, Xj, Xd, INCX, hscal_zvvdotu,
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
   }
   else
   {
/*
*     Neither sub( X ) nor sub( Y ) are distributed
*/
      type = PB_Cztypeset();
      hscal_PB_CpdotNN_Z( 
           type, N, Xi, Xj, Xd, INCX,
           Yi, Yj, Yd, INCY, hscal_zvvdotu,
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );
   }

   Cblacs_gridexit_nocomm(blacsctxt);

/*
*  End of PZDOTU
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdotu_tcomp_tcomm(
   int N,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzdotu_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N, 
                IX, JX, DESCX, INCX,
                IY, JY, DESCY, INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdotu_info(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzdotu_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdotu_info_i(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzdotu_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, 
                *IX, *JX, DESCX, *INCX,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
