
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzdscal_recon_tcomp(
   int n,
   int alphai,
   int incx
)
{
    double tcomp = 0;

    int num_of_computers = HMPI_Get_number_of_computers();
    int reconn = (((n/num_of_computers) >= 1) ? (n/num_of_computers) : 1);
    double alpha = alphai;

    hscal_zdscal(
         &reconn,
         (char*)&alpha,
         &incx,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

int ihscal_pzdscal_tcomp(
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int ALPHAI,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp
)
{
/*
*  .. Local Scalars ..
*/
   int            Xcol, Xi, Xii, Xj, Xjj, Xld, Xnp, Xnq, Xrow, ctxt, info;
   PBTYP_T        * type;
/*
*  .. Local Arrays ..
*/
   int            Xd[DLEN_];
   double         * tcomp_l = tcomp;
   double         ALPHA     = ALPHAI;

/* ..
*  .. Executable Statements ..
*
*/
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
/*
*  Test the input parameters
*/
   if (init_llds_f == 1)
   {
   HSCAL_PXSCAL_INIT_LLDS(
        N,
        Xd, IX, INCX,
        myrow, nprow
   );
   }

   if( !( info = ( ( nprow == -1 ) ? -( 601 + CTXT_ ) : 0 ) ) )
   {
      hscal_PB_Cchkvec(
           "PZDSCAL", "X", N, 1, Xi, Xj, Xd, INCX, 6, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
      hscal_PB_Cabort(
           "PZDSCAL",
           info,
           nprow, npcol, myrow, mycol
      );

      return HSCAL_SUCCESS;
   }

/*
*  Quick return if possible
*/
   if( ( N == 0 ) || ( ALPHA == ONE ) ) 
   {
      return HSCAL_SUCCESS;
   }

/*
*  Retrieve sub( X )'s local information: Xii, Xjj, Xrow, Xcol
*/
   PB_Cinfog2l( Xi, Xj, Xd, nprow, npcol, myrow, mycol, &Xii, &Xjj, &Xrow,
                &Xcol );
/*
*  Start the operations
*/
   if( INCX == Xd[M_] )
   {
/*
*  sub( X ) resides in (a) process row(s)
*/
      if( ( myrow == Xrow ) || ( Xrow < 0 ) )
      {
/*
*  Make sure I own some data and scale sub( X )
*/
         Xnq = PB_Cnumroc( N, Xj, Xd[INB_], Xd[NB_], mycol, Xd[CSRC_], npcol );
         if( Xnq > 0 )
         {
            Xld = Xd[LLD_];
            type = PB_Cztypeset();
            if( ALPHA == ZERO )
            {
               zset_tcomp_macro( 
                   &Xnq, type->zero, &Xld,
                   nprow, npcol, myrow, mycol, tcomp_l
               );
            }
            else
            {
               zdscal_tcomp_macro( 
                     &Xnq, ((char *) &ALPHA), &Xld,
                     nprow, npcol, myrow, mycol, tcomp_l
               );
            }
         }
      }

      return HSCAL_SUCCESS;
   }
   else
   {
/*
*  sub( X ) resides in (a) process column(s)
*/
      if( ( mycol == Xcol ) || ( Xcol < 0 ) )
      {
/*
*  Make sure I own some data and scale sub( X )
*/
         Xnp = PB_Cnumroc( N, Xi, Xd[IMB_], Xd[MB_], myrow, Xd[RSRC_], nprow );
         if( Xnp > 0 )
         {
            type = PB_Cztypeset();
            if( ALPHA == ZERO )
            {
               zset_tcomp_macro( 
                   &Xnp, type->zero, &INCX,
                   nprow, npcol, myrow, mycol, tcomp_l
               );
            }
            else
            {
               zdscal_tcomp_macro( 
                     &Xnp, ((char *) &ALPHA), &INCX,
                     nprow, npcol, myrow, mycol, tcomp_l
               );
            }
         }
      }

      return HSCAL_SUCCESS;
   }

   return HSCAL_SUCCESS;
/*
*  End of PZDSCAL
*/
}

/*------------------------------------------------------------------*/

double hscal_pzdscal_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int ALPHAI,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = ihscal_pzdscal_tcomp
        (
          1,
          myrow, mycol,
          N, ALPHAI,
          IX, JX, DESCX, INCX,
          nprow, npcol,
          tcomp
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzdscal_tcomp(
   int N,
   int ALPHAI,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = ihscal_pzdscal_tcomp
              (
                1,
                myrow, mycol,
                N, ALPHAI,
                IX, JX, DESCX, INCX,
                nprow, npcol,
                tcomp
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdscal_info(
   int * N,
   double* ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp
)
{
   int i, rc, myrow, mycol;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.f)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.f)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = ihscal_pzdscal_tcomp
              (
                1,
                myrow, mycol,
                *N, alphai,
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdscal_info_i(
   int * N,
   double* ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp
)
{
   int i, rc, myrow, mycol;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha
   */
   int alphai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = ihscal_pzdscal_tcomp
              (
                1,
                myrow, mycol,
                *N, alphai,
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
