
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzgemm_recon_tcomp(
   int transai,
   int transbi,
   int m,
   int n,
   int k, 
   int alphai[CNARG],
   int betai[CNARG]
)
{
    double tcomp = 0;

    char   transac = transai;
    char   *transa = &transac;

    char   transbc = transbi;
    char   *transb = &transbc;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;
    int reconm           = m/p;
    int reconn           = n/q;
    int reconk            = hscal_pilaenv_recon("Z");

    int    lda, ldb, ldc;

    double alpha[CNARG], beta[CNARG];

    alpha[REAL_PART] = alphai[REAL_PART];
    alpha[IMAG_PART] = alphai[IMAG_PART];

    beta[REAL_PART]  = betai[REAL_PART];
    beta[IMAG_PART]  = betai[IMAG_PART];

    if (transac == 'N')
    {
       lda = reconm;
    }
    else
    {
       lda = reconk;
    }

    if (transbc == 'N')
    {
       ldb = reconk;
    }
    else
    {
       ldb = reconn;
    }

    ldc = reconm;

    hscal_zgemm(
         transa,
         transb,
         &reconm,
         &reconn,
         &reconk,
         ((char*)alpha),
         &lda,
         &ldb,
         ((char*)beta),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pzgemm_my_tcomp(
   int myrow,
   int mycol,
   int TRANSAI,
   int TRANSBI,
   int M,
   int N,
   int K,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzgemm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSAI, TRANSBI,
          M, N, K,
          ALPHA, 
          IA, JA, DESCA, 
          IB, JB, DESCB, 
          BETA,
          IC, JC, DESCC, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_my_tcomm(
   int myrow,
   int mycol,
   int TRANSAI,
   int TRANSBI,
   int M,
   int N,
   int K,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzgemm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSAI, TRANSBI,
          M, N, K,
          ALPHA, 
          IA, JA, DESCA,
          IB, JB, DESCB,
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int TRANSAI,
   int TRANSBI,
   int M,
   int N,
   int K,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pzgemm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSAI, TRANSBI,
          M, N, K,
          ALPHA, 
          IA, JA, DESCA,
          IB, JB, DESCB,
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int TRANSAI,
   int TRANSBI,
   int M,
   int N,
   int K,
   int ALPHAI[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETAI[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           DirAB, DirBC, OpC, OpR, TrA, TrB, ctop, ctopsave, rtop,
                  rtopsave;
   int            Ai, Aj, Bi, Bj, ChooseAB, ChooseBC, Ci, Cj, ForceTop, ctxt,
                  info, nb, nota, notb;
   double         ABest, ACest, BCest, tmp1, tmp2, tmp3, tmp4;
   PBTYP_T        * type;
/*
*  .. Local Arrays ..
*/
   int            Ad[DLEN_], Bd[DLEN_], Cd[DLEN_];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

   char TRANSAC   = TRANSAI, *TRANSA = &TRANSAC;
   char TRANSBC   = TRANSBI, *TRANSB = &TRANSBC;

   double ALPHA[CNARG], BETA[CNARG];

   ALPHA[REAL_PART]       = ALPHAI[REAL_PART];
   ALPHA[IMAG_PART]       = ALPHAI[IMAG_PART];

   BETA[REAL_PART]        = BETAI[REAL_PART];
   BETA[IMAG_PART]        = BETAI[IMAG_PART];

/* ..
*  .. Executable Statements ..
*
*/
   nota = ( ( TrA = Mupcase( F2C_CHAR( TRANSA )[0] ) ) == CNOTRAN );
   notb = ( ( TrB = Mupcase( F2C_CHAR( TRANSB )[0] ) ) == CNOTRAN );

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IB, JB, DESCB, &Bi, &Bj, Bd );
   PB_CargFtoC( IC, JC, DESCC, &Ci, &Cj, Cd );

   if (init_llds_f == 1)
   {
   HSCAL_PXGEMM_INIT_LLDS(
        Ad,
        Bd,
        Cd,
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Bd[M_]=%d Bd[MB_]=%d myrow=%d Bd[RSRC_]=%d nprow=%d Bd[LLD_]=%d",
        Bd[M_], Bd[MB_], myrow, Bd[RSRC_], nprow, Bd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Cd[M_]=%d Cd[MB_]=%d myrow=%d Cd[RSRC_]=%d nprow=%d Cd[LLD_]=%d", 
        Cd[M_], Cd[MB_], myrow, Cd[RSRC_], nprow, Cd[LLD_]
   );

/*
*  Test the input parameters
*/
   if( !( info = ( ( nprow == -1 ) ? -( 1001 + CTXT_ ) : 0 ) ) )
   {
      if( ( !nota ) && ( TrA != CTRAN ) && ( TrA != CCOTRAN ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZGEMM", "Illegal TRANSA = %c\n", TrA
         );
         info = -1;
      }
      else if( ( !notb ) && ( TrB != CTRAN ) && ( TrB != CCOTRAN ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZGEMM", "Illegal TRANSB = %c\n", TrB 
         );
         info = -2;
      }

      if( nota )
      {
         hscal_PB_Cchkmat( 
              "PZGEMM", "A", M, 3, K, 5, Ai, Aj, Ad, 10,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      else
      {
         hscal_PB_Cchkmat( 
              "PZGEMM", "A", K, 5, M, 3, Ai, Aj, Ad, 10,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }

      if( notb )
      {
         hscal_PB_Cchkmat(
              "PZGEMM", "B", K, 5, N, 4, Bi, Bj, Bd, 14,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      else
      {
         hscal_PB_Cchkmat( 
              "PZGEMM", "B", N, 4, K, 5, Bi, Bj, Bd, 14,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }

      hscal_PB_Cchkmat(
           "PZGEMM", "C", M, 3, N, 4, Ci, Cj, Cd, 19,
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info ) 
   { 
       hscal_PB_Cabort( 
            "PZGEMM", 
            info,
            nprow, npcol, myrow, mycol
       ); 

       return HSCAL_ERR_INPUT;
   }

   if( ( M == 0 ) || ( N == 0 ) ||
       ( ( ( ALPHA[REAL_PART] == ZERO && ALPHA[IMAG_PART] == ZERO ) ||
           ( K == 0                                              ) ) &&
           ( BETA [REAL_PART] == ONE  && BETA [IMAG_PART] == ZERO ) ) )
   {
      return HSCAL_SUCCESS;
   }

/*
*  Get type structure
*/
   type = PB_Cztypeset();
/*
*  If alpha or K is zero, sub( C ) := beta * sub( C ).
*/
   if( ( ( ALPHA[REAL_PART] == ZERO ) &&
         ( ALPHA[IMAG_PART] == ZERO ) ) || ( K == 0 ) )
   {
      if( ( BETA[REAL_PART] == ZERO ) && ( BETA[IMAG_PART] == ZERO ) )
      {
         hscal_PB_Cplapad_Z( 
              type, ALL, NOCONJG, M, N, type->zero, type->zero,
              Ci, Cj, Cd,
              nprow, npcol, myrow, mycol, tcomp_l
         );
      }
      else if( !( ( BETA[REAL_PART] ==  ONE ) &&
                  ( BETA[IMAG_PART] == ZERO ) ) )
      {
         hscal_PB_Cplascal_Z( 
              type, ALL, NOCONJG, M, N, ((char *) BETA),
              Ci, Cj, Cd,
              nprow, npcol, myrow, mycol, tcomp_l
         );
      }

      return HSCAL_SUCCESS;
   }
/*
*  Start the operations
*/

/*
*  Algorithm selection is based on approximation of the communication volume
*  for distributed and aligned operands.
*
*  ABest: both operands sub( A ) and sub( B ) are communicated (M, N >> K)
*  ACest: both operands sub( A ) and sub( C ) are communicated (K, N >> M)
*  BCest: both operands sub( B ) and sub( C ) are communicated (M, K >> N)
*/
   ABest = (double)(K);
   ACest = (double)(M);
   BCest = (double)(N);

   if( notb )
   {
      if( nota )
      {
         tmp1 = DNROC( M, Cd[MB_], nprow ); tmp2 = DNROC( N, Cd[NB_], npcol );
         ABest *= ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 ) +
                  ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 );

         tmp1 = DNROC( K, Bd[MB_], nprow ); tmp2 = DNROC( N, Bd[NB_], npcol );
         tmp3 = DNROC( K, Ad[NB_], npcol );
         ACest *= ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) +
                  CBRATIO * ( nprow == 1 ? ZERO : tmp2 );

         tmp1 = DNROC( M, Ad[MB_], nprow ); tmp2 = DNROC( K, Ad[NB_], npcol );
                                             tmp4 = DNROC( K, Bd[MB_], nprow );
         BCest *= CBRATIO * ( npcol == 1 ? ZERO : tmp1 ) +
                  ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 );
      }
      else
      {
         tmp1 = DNROC( M, Cd[MB_], nprow ); tmp2 = DNROC( N, Cd[NB_], npcol );
         tmp3 = DNROC( M, Ad[NB_], npcol );
         ABest *= ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) +
                  ( nprow == 1 ? ZERO : tmp2 );

         tmp1 = DNROC( K, Bd[MB_], nprow ); tmp2 = DNROC( N, Bd[NB_], npcol );
         ACest *= ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 ) +
                  CBRATIO *
                  ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 );

         tmp1 = DNROC( K, Ad[MB_], nprow ); tmp2 = DNROC( M, Bd[NB_], npcol );
                                             tmp4 = DNROC( M, Cd[MB_], nprow );
         BCest *= ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 ) +
                  CBRATIO * ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 );
      }
   }
   else
   {
      if( nota )
      {
         tmp1 = DNROC( M, Cd[MB_], nprow ); tmp2 = DNROC( N, Cd[NB_], npcol );
                                             tmp4 = DNROC( N, Bd[MB_], nprow );
         ABest *= ( npcol == 1 ? ZERO : tmp1 ) +
                  ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 );

         tmp1 = DNROC( N, Bd[MB_], nprow ); tmp2 = DNROC( K, Bd[NB_], npcol );
         tmp3 = DNROC( N, Cd[NB_], npcol );
         ACest *= CBRATIO * ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) +
                  ( ( ( Ad[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 );

         tmp1 = DNROC( M, Ad[MB_], nprow ); tmp2 = DNROC( K, Ad[NB_], npcol );
         BCest *= CBRATIO *
                  ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 ) +
                  ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 );
      }
      else
      {
         tmp1 = DNROC( M, Cd[MB_], nprow ); tmp2 = DNROC( N, Cd[NB_], npcol );
         tmp3 = DNROC( M, Ad[NB_], npcol ); tmp4 = DNROC( N, Bd[MB_], nprow );
         ABest *= ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) +
                  ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 );

         tmp1 = DNROC( N, Bd[MB_], nprow ); tmp2 = DNROC( K, Bd[NB_], npcol );
         tmp3 = DNROC( N, Cd[NB_], npcol ); tmp4 = DNROC( K, Ad[MB_], nprow );
         ACest *= CBRATIO * ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) +
                  ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 );

         tmp1 = DNROC( K, Ad[MB_], nprow ); tmp2 = DNROC( M, Ad[NB_], npcol );
         tmp3 = DNROC( K, Bd[NB_], npcol ); tmp4 = DNROC( M, Cd[MB_], nprow );
         BCest *= ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) +
                  CBRATIO * ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 );
      }
   }
   ChooseAB = ( ( ABest <= ( 1.3 * BCest ) ) && ( ABest <= ( 1.3 * ACest ) ) );
   ChooseBC = ( ( BCest <= ACest           ) && ( ( 1.3 * BCest ) <= ABest ) );
/*
*  BLACS topologies are enforced iff M, N and K are strictly greater than the
*  logical block size returned by pilaenv_. Otherwise, it is assumed that the
*  routine calling this routine has already selected an adequate topology.
*/
   nb       = hscal_pilaenv( C2F_CHAR( &type->type ) );
   ForceTop = ( ( M > nb ) && ( N > nb ) && ( K > nb ) );

   if( ChooseAB )
   {
      OpR = CBCAST;
      OpC = CBCAST;
   }
   else if( ChooseBC )
   {
      if( nota ) { OpR = CCOMBINE; OpC = CBCAST; }
      else       { OpR = CBCAST; OpC = CCOMBINE; }
   }
   else
   {
      if( notb ) { OpR = CBCAST; OpC = CCOMBINE; }
      else       { OpR = CCOMBINE; OpC = CBCAST; }
   }

   rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_GET );
   ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_GET );

   if( ForceTop )
   {
      rtopsave = rtop;
      ctopsave = ctop;
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
      if( ( rtop != CTOP_DRING ) && ( rtop != CTOP_IRING ) &&
          ( rtop != CTOP_SRING ) )
         rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_IRING );
      if( ( ctop != CTOP_DRING ) && ( ctop != CTOP_IRING ) &&
          ( ctop != CTOP_SRING ) )
         ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_IRING );
/*
*  Remove the next 4 lines when the BLACS combine operations support ring
*  topologies
*/
/*
*      if( OpR == CCOMBINE )
*         rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_DEFAULT );
*      if( OpC == CCOMBINE )
*         ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_DEFAULT );
*/
   }

   DirAB = ( rtop == CTOP_DRING ? CBACKWARD : CFORWARD );
   DirBC = ( ctop == CTOP_DRING ? CBACKWARD : CFORWARD );

   if( ChooseAB )
   {
      char *DIRECA = ( nota ? NOTRAN : ( ( TrA == CCOTRAN ) ? COTRAN : TRAN ) );
      char *DIRECB = ( notb ? NOTRAN : ( ( TrB == CCOTRAN ) ? COTRAN : TRAN ) );

      hscal_PB_CpgemmAB_Z(
           type, &DirAB, &DirBC, DIRECA, DIRECB, M, N, K,
           ((char *)ALPHA), Ai, Aj, Ad, Bi,
           Bj, Bd, ((char *)BETA), Ci, Cj, Cd,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
   else if( ChooseBC )
   {
      char *DIRECA = ( nota ? NOTRAN : ( ( TrA == CCOTRAN ) ? COTRAN : TRAN ) );
      char *DIRECB = ( notb ? NOTRAN : ( ( TrB == CCOTRAN ) ? COTRAN : TRAN ) );

      hscal_PB_CpgemmBC_Z( 
           type, &DirAB, &DirBC, DIRECA, DIRECB, M, N, K,
           ((char *)ALPHA), Ai, Aj, Ad, Bi,
           Bj, Bd, ((char *)BETA), Ci, Cj, Cd,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
   else
   {
      char *DIRECA = ( nota ? NOTRAN : ( ( TrA == CCOTRAN ) ? COTRAN : TRAN ) );
      char *DIRECB = ( notb ? NOTRAN : ( ( TrB == CCOTRAN ) ? COTRAN : TRAN ) );

      hscal_PB_CpgemmAC_Z( 
           type, &DirAB, &DirBC, DIRECA, DIRECB, M, N, K,
           ((char *)ALPHA), Ai, Aj, Ad, Bi,
           Bj, Bd, ((char *)BETA), Ci, Cj, Cd,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
/*
*  Restore the BLACS topologies when necessary.
*/
   if( ForceTop )
   {
      rtopsave = *PB_Ctop( &ctxt, &OpR, ROW,    &rtopsave );
      ctopsave = *PB_Ctop( &ctxt, &OpC, COLUMN, &ctopsave );
   }

/*
*  End of PZGEMM
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_tcomp_tcomm(
   int TRANSAI,
   int TRANSBI,
   int M,
   int N,
   int K,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgemm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSAI, TRANSBI,
                M, N, K,
                ALPHA, 
                IA, JA, DESCA, 
                IB, JB, DESCB, 
                BETA,
                IC, JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_info(
   F_CHAR_T TRANSA,
   F_CHAR_T TRANSB,
   int * M,
   int * N,
   int * K,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double* BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int TRANSAI = *TRANSA;
   int TRANSBI = *TRANSB;

   /*
   * Just get a indicative multiplication scalar of
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgemm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSAI, TRANSBI,
                *M, *N, *K,
                alphai,
                *IA, *JA, DESCA, 
                *IB, *JB, DESCB, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemm_info_i(
   F_CHAR_T TRANSA,
   F_CHAR_T TRANSB,
   int * M,
   int * N,
   int * K,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double* BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int TRANSAI = *TRANSA;
   int TRANSBI = *TRANSB;

   /*
   * Just get a indicative multiplication scalar of
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgemm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSAI, TRANSBI,
                *M, *N, *K,
                alphai,
                *IA, *JA, DESCA, 
                *IB, *JB, DESCB, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
