
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzgemv_recon_tcomp(
   int transi,
   int m,
   int n,
   int alphai[CNARG],
   int incx,
   int betai[CNARG],
   int incy
)
{
    char transc  = transi;
    char *trans  = &transc;
    double tcomp = 0;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;
    int reconm           = m/p;
    int reconn           = n/q;

    int lda      = reconm;
    double alpha[CNARG], beta[CNARG];

    alpha[REAL_PART]       = alphai[REAL_PART];
    alpha[IMAG_PART]       = alphai[IMAG_PART];

    beta[REAL_PART]        = betai[REAL_PART];
    beta[IMAG_PART]        = betai[IMAG_PART];
    
    hscal_zgemv(
         trans,
         &reconm,
         &reconn,
         ((char*)alpha),
         &lda,
         &incx,
         ((char*)beta),
         &incy,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__, __LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pzgemv_my_tcomp(
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA[CNARG],
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzgemv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          M, N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_my_tcomm(
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA[CNARG],
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzgemv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          M, N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int TRANSI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA[CNARG],
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pzgemv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          M, N, 
          ALPHA, IA, JA, DESCA, 
          IX, JX, DESCX, INCX, BETA,
          IY, JY, DESCY, INCY,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %f",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int ALPHAI[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETAI[CNARG],
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   char TRANSC = TRANSI;
   char *TRANS = &TRANSC;

   double ALPHA[CNARG], BETA[CNARG];
   ALPHA[REAL_PART] = ALPHAI[REAL_PART];
   ALPHA[IMAG_PART] = ALPHAI[IMAG_PART];

   BETA[REAL_PART]  = BETAI[REAL_PART];
   BETA[IMAG_PART]  = BETAI[IMAG_PART];

   {
     /*
     *  .. Local Scalars ..
     */
     char           TrA, Yroc, * tbeta, top;
     int            Acol, Ai, Aii, Aimb1, Ainb1, Aj, Ajj, Ald, Amb, Amp, Anb,
                    Anq, Arow, XAfr, Xi, Xj, YAfr, YApbY, YAsum, Ycol, Yi, Yii,
                    Yj, Yjj, Yld, Ynp, Ynq, Yrow, ctxt, info, ione=1, nota;
     PBTYP_T        * type;
     HSCALBLACSCONTEXT   * blacsctxt;

     /*
     *  .. Local Arrays ..
     */
     int            Ad [DLEN_], Ad0[DLEN_], XAd[DLEN_], Xd[DLEN_], YAd[DLEN_],
                    Yd [DLEN_];

     /* ..
     *  .. Executable Statements ..
     *
     */
     nota = ( ( TrA = Mupcase( F2C_CHAR( TRANS )[0] ) ) == CNOTRAN );
     HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "TRANSPOSE %s", ((nota) ? "YES" : "NO"));

     PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
     PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );
     PB_CargFtoC( IY, JY, DESCY, &Yi, &Yj, Yd );

   if (init_llds_f == 1)
   {
     HSCAL_PXGEMV_INIT_LLDS(
          nota, M, N, 
          Ad, 
          Xd, IX, INCX, 
          Yd, IY, INCY, 
          myrow, nprow
     );
   }

     HSCAL_DEBUG(
          __FILE__, __FUNCTION__,__LINE__, 
          "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
          Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
     );

     HSCAL_DEBUG(
          __FILE__, __FUNCTION__,__LINE__, 
          "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d",
          Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
     );

     HSCAL_DEBUG(
          __FILE__, __FUNCTION__,__LINE__, 
          "Yd[M_]=%d Yd[MB_]=%d myrow=%d Yd[RSRC_]=%d nprow=%d Yd[LLD_]=%d", 
          Yd[M_], Yd[MB_], myrow, Yd[RSRC_], nprow, Yd[LLD_]
     );

/*
*  Test the input parameters
*/
   if( !( info = ( ( nprow == -1 ) ? -( 801 + CTXT_ ) : 0 ) ) )
   {
      if( ( !nota ) && ( TrA != CTRAN ) && ( TrA != CCOTRAN ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZGEMV", "Illegal TRANS=%c\n", TrA 
         );
         info = -1;
      }

      hscal_PB_Cchkmat(
           "PZGEMV", "A", M, 2, N, 3, Ai, Aj, Ad,  8,
           nprow, npcol, myrow, mycol,
           &info
      );

      if( nota )
      {
         hscal_PB_Cchkvec( 
              "PZGEMV", "X", N, 3, Xi, Xj, Xd, INCX, 12,
              nprow, npcol, myrow, mycol,
              &info 
         );
         hscal_PB_Cchkvec( 
              "PZGEMV", "Y", M, 2, Yi, Yj, Yd, INCY, 18,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      else
      {
         hscal_PB_Cchkvec( 
              "PZGEMV", "X", M, 2, Xi, Xj, Xd, INCX, 12,
              nprow, npcol, myrow, mycol,
              &info 
         );
         hscal_PB_Cchkvec( 
              "PZGEMV", "Y", N, 3, Yi, Yj, Yd, INCY, 18,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PZGEMV",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }

     /*
     *  Quick return if possible
     */
     if( ( M == 0 ) || ( N == 0 ) ||
         ( ( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) ) &&
           ( ( BETA [REAL_PART] ==  ONE ) && ( BETA [IMAG_PART] == ZERO ) ) ) )
     {
        return HSCAL_SUCCESS;
     }

/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
     Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

     /*
     *  Get type structure
     */
     type = PB_Cztypeset();

     /*
     *  When alpha is zero
     */
     if( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) )
     {
        HSCAL_DEBUG(__FILE__, __FUNCTION__,__LINE__, "ALPHA IS ZERO");

        if( INCY == Yd[M_] )
        {
           /*
           *  sub( Y ) resides in (a) process row(s)
           */
           {
              {
                 /*
                 *  Retrieve sub( Y )'s local information: Yii, Yjj, Yrow, Ycol
                 */
                 PB_Cinfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, &Yii, &Yjj,
                              &Yrow, &Ycol );
 
                 if( ( myrow == Yrow ) || ( Yrow < 0 ) )
                 {
                    /*
                    *  Make sure I own some data and scale sub( Y )
                    */
                    Ynq = PB_Cnumroc( ( nota ? M : N ), Yj, Yd[INB_], Yd[NB_], mycol,
                                      Yd[CSRC_], npcol );
                    if( Ynq > 0 )
                    {
                        Yld = Yd[LLD_];
                        if( ( BETA[REAL_PART] == ZERO ) && ( BETA[IMAG_PART] == ZERO ) )
                        {
                           zset_tcomp_macro(
                               &Ynq, ((char *) BETA), &Yld,
                               nprow, npcol, myrow, mycol,
                               tcomp
                           );
                        }
                        else
                        {
                           zscal_tcomp_macro(
                               &Ynq, ((char *) BETA), &Yld,
                               nprow, npcol, myrow, mycol,
                               tcomp                               
                           );
                        }
                    }
                 }
              }
           }
        }
        else
        {
           /*
           *  sub( Y ) resides in (a) process column(s)
           */
           {
              {
                 /*
                 *  Retrieve sub( Y )'s local information: Yii, Yjj, Yrow, Ycol
                 */
                 PB_Cinfog2l( Yi, Yj, Yd, nprow, npcol, myrow, mycol, &Yii, &Yjj,
                              &Yrow, &Ycol );
 
                 if( ( mycol == Ycol ) || ( Ycol < 0 ) )
                 {
                    /*
                    *  Make sure I own some data and scale sub( Y )
                    */
                    Ynp = PB_Cnumroc( ( nota ? M : N ), Yi, Yd[IMB_], Yd[MB_], myrow,
                                        Yd[RSRC_], nprow );
                    if( Ynp > 0 )
                    {
                        if( ( BETA[REAL_PART] == ZERO ) && ( BETA[IMAG_PART] == ZERO ) )
                        {
                           zset_tcomp_macro(
                               &Ynp, ((char *) BETA), &INCY,
                               nprow, npcol, myrow, mycol,
                               tcomp
                           );
                        }
                        else
                        {
                           zscal_tcomp_macro(
                               &Ynp, ((char *) BETA), &INCY,
                               nprow, npcol, myrow, mycol,
                               tcomp                               
                           );
                        }
                    }
                 }
              }
           }
        }

        Cblacs_gridexit_nocomm(blacsctxt);

        return HSCAL_SUCCESS;
     }

     Yroc = ( INCY == Yd[M_] ? CROW : CCOLUMN );

     if( nota )
     {
        {
           {
              /*
              *  Compute descriptor Ad0 for sub( A )
              */
              PB_Cdescribe( 
                 M, N, Ai, Aj, Ad, 
                 nprow, npcol, myrow, mycol, 
                 &Aii, &Ajj, &Ald, &Aimb1, &Ainb1, 
                 &Amb, &Anb, &Arow, &Acol, Ad0 );

              /*
              *  Reuse sub( Y ) and/or create vector YA in process columns spanned by sub( A )
              */
              {
                 int K_l = 1;
                 char *ROWCOL_l = COLUMN;
                 hscal_PB_CInOutV_Z(
                      type, ROWCOL_l, M, N, Ad0, K_l,
                      ((char *) BETA), Yi, Yj, Yd, 
                      (&Yroc), (&tbeta), YAd,
                      (&YAsum), (&YApbY), nprow, npcol, myrow, mycol, tcomp);
              }

              /*
              *  Replicate sub( X ) in process rows spanned by sub( A ) -> XA
              */
              {
                 char *CONJUG_l = NOCONJG;
                 char *ROWCOL_l = ROW;
                 int K_l = 1;
                 char *XROC_l = ( INCX == Xd[M_] ? ROW : COLUMN );
                 hscal_PB_CInV_Z( 
                      type, CONJUG_l, ROWCOL_l, M, N, Ad0, 
                      K_l, Xi, Xj, Xd, XROC_l,
                      XAd, nprow, npcol, myrow, mycol, tcomp, tcomm );
              }

              /*
              *  Local matrix-vector multiply iff I own some data
              */
              Amp = PB_Cnumroc( M, 0, Ad0[IMB_], Ad0[MB_], myrow, Ad0[RSRC_], nprow );
              Anq = PB_Cnumroc( N, 0, Ad0[INB_], Ad0[NB_], mycol, Ad0[CSRC_], npcol );
              if( ( Amp > 0 ) && ( Anq > 0 ) )
              {
                 zgemv_tcomp_macro( TRANS, &Amp, &Anq, ((char *) ALPHA), &Ald, 
                              &XAd[LLD_], tbeta, &ione, 
                              nprow, npcol, myrow, mycol, tcomp );
              }

              /*
              *  Combine the partial column results into YA
              */
              if( YAsum && ( Amp > 0 ) )
              {
                 top = *PB_Ctop( &ctxt, COMBINE, ROW, TOP_GET );
                 hscal_Czgsum2d(
                    blacsctxt, ROW, &top, Amp, 
                    1, YAd[LLD_], myrow, YAd[CSRC_],
                    nprow, npcol, myrow, mycol, tcomp, tcomm
                 );
              }

              /*
              *  sub( Y ) := beta * sub( Y ) + YA (if necessary)
              */
              if( YApbY )
              {
                 int N_l = 1;
                 int IA_l = 0;
                 int JA_l = 0;
                 char *CONJUG_l = NOCONJG;
                 hscal_PB_Cpaxpby_Z( 
                    type, CONJUG_l, M, N_l, 
                    type->one, IA_l, JA_l, YAd, COLUMN,
                    ((char *) BETA), Yi, Yj, Yd, (&Yroc),
                    nprow, npcol, myrow, mycol, tcomp, tcomm
                 );
              }
           }
        }     
     }
     else
     {
        {
           {
              /*
              *  Compute descriptor Ad0 for sub( A )
              */
              PB_Cdescribe( 
                 M, N, Ai, Aj, Ad, 
                 nprow, npcol, myrow, mycol, 
                 &Aii, &Ajj, &Ald, &Aimb1, &Ainb1, 
                 &Amb, &Anb, &Arow, &Acol, Ad0 );

              /*
              *  Reuse sub( Y ) and/or create vector YA in process rows spanned by sub( A )
              */
              {
                 int K_l = 1;
                 char *ROWCOL_l = ROW;
                 hscal_PB_CInOutV_Z( 
                     type, ROWCOL_l, M, N, Ad0, K_l, 
                     ((char *) BETA), Yi, Yj, Yd, (&Yroc), 
                     (&tbeta), YAd, (&YAsum), (&YApbY),
                     nprow, npcol, myrow, mycol, tcomp);
              }

              /*
              *  Replicate sub( X ) in process columns spanned by sub( A ) -> XA
              */
              {
                 char *CONJUG_l = NOCONJG;
                 char *ROWCOL_l = COLUMN; 
                 int K_l = 1;
                 char *XROC_l = ( INCX == Xd[M_] ? ROW : COLUMN );
                 hscal_PB_CInV_Z( 
                     type, CONJUG_l, ROWCOL_l, M, N, 
                     Ad0, K_l, Xi, Xj, Xd, XROC_l,
                     XAd, nprow, npcol, myrow, mycol, tcomp, tcomm );
              }

              /*
              *  Local matrix-vector multiply iff I own some data
              */
              Amp = PB_Cnumroc( M, 0, Ad0[IMB_], Ad0[MB_], myrow, Ad0[RSRC_], nprow );
              Anq = PB_Cnumroc( N, 0, Ad0[INB_], Ad0[NB_], mycol, Ad0[CSRC_], npcol );

              if( ( Amp > 0 ) && ( Anq > 0 ) )
              {
                 zgemv_tcomp_macro( TRANS, &Amp, &Anq, ((char *) ALPHA), &Ald, 
                              &ione, tbeta, &YAd[LLD_],
                              nprow, npcol, myrow, mycol, tcomp );
              }

              /*
              *  Combine the partial row results into YA
              */
              if( YAsum && ( Anq > 0 ) )
              {
                 top = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );
                 hscal_Czgsum2d( 
                         blacsctxt, COLUMN, &top, 1, 
                         Anq, YAd[LLD_], YAd[RSRC_], mycol,
                         nprow, npcol, myrow, mycol, tcomp, tcomm
                 );
              }

              /*
              *  sub( Y ) := beta * sub( Y ) + YA (if necessary)
              */
              if( YApbY )
              {
                 int M_l = 1;
                 int IA_l = 0;
                 int JA_l = 0;
                 char *CONJUG_l = NOCONJG;
                 hscal_PB_Cpaxpby_Z( 
                    type, CONJUG_l, M_l, N, type->one, 
                    IA_l, JA_l, YAd, ROW,
                    ((char *) BETA), Yi, Yj, Yd, (&Yroc),
                    nprow, npcol, myrow, mycol, tcomp, tcomm
                 );
              }
           }
        }
     }

     Cblacs_gridexit_nocomm(blacsctxt);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_tcomp_tcomm(
   int TRANSI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int BETA[CNARG],
   int IY,
   int JY,
   int DESCY[DLEN_],
   int INCY,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgemv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSI,
                M, N, 
                ALPHA, IA, JA, DESCA, 
                IX, JX, DESCX, INCX, BETA,
                IY, JY, DESCY, INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_info(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double* BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int TRANSI = *TRANS;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgemv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSI,
                *M, *N, 
                alphai, *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX, betai,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgemv_info_i(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double* BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int TRANSI = *TRANS;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgemv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSI,
                *M, *N, 
                alphai, *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX, betai,
                *IY, *JY, DESCY, *INCY,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
