
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzhemm_recon_tcomp(
   int sidei,
   int uploi,
   int m,
   int n, 
   int alphai[CNARG],
   int betai[CNARG]
)
{
    double tcomp = 0;

    char   sidec   = sidei;
    char   *side   = &sidec;

    char   uploc  = uploi;
    char   *uplo  = &uploc;

    int    lda, ldb, ldc;

    double alpha[CNARG], beta[CNARG];

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;
    int reconm           = m/p;
    int reconn           = n/q;
    int reconk            = hscal_pilaenv_recon("Z");

    alpha[REAL_PART] = alphai[REAL_PART];
    alpha[IMAG_PART] = alphai[IMAG_PART];

    beta[REAL_PART]  = betai[REAL_PART];
    beta[IMAG_PART]  = betai[IMAG_PART];

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_zgemm(
         "N",
         "N",
         &reconm,
         &reconn,
         &reconk,
         ((char*)alpha),
         &lda,
         &ldb,
         ((char*)beta),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pzhemm_my_tcomp(
   int myrow,
   int mycol,
   int SIDEI,
   int UPLOI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzhemm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, UPLOI,
          M, N,
          ALPHA, 
          IA, JA, DESCA, 
          IB, JB, DESCB, 
          BETA,
          IC, JC, DESCC, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzhemm_my_tcomm(
   int myrow,
   int mycol,
   int SIDEI,
   int UPLOI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzhemm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, UPLOI,
          M, N,
          ALPHA, 
          IA, JA, DESCA,
          IB, JB, DESCB,
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzhemm_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int SIDEI,
   int UPLOI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pzhemm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, UPLOI,
          M, N,
          ALPHA, 
          IA, JA, DESCA,
          IB, JB, DESCB,
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pzhemm_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int SIDEI,
   int UPLOI,
   int M,
   int N,
   int ALPHAI[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETAI[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           DirAB, SideOp, UploA, cbtop, cbtopsave, cctop, cctopsave,
                  rbtop, rbtopsave, rctop, rctopsave;
   int            Ai, Aj, Bi, Bj, ChooseABC, Ci, Cj, ForceTop, ctxt, info,
                  lside, nb, upper;
   double         ABCest, BCest, tmp1, tmp2, tmp3, tmp4;
   PBTYP_T        * type;
/*
*  .. Local Arrays ..
*/
   int            Ad[DLEN_], Bd[DLEN_], Cd[DLEN_];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

   char SIDEC     = SIDEI, *SIDE = &SIDEC;
   char UPLOC     = UPLOI, *UPLO = &UPLOC;

   double ALPHA[CNARG], BETA[CNARG];

   ALPHA[REAL_PART]       = ALPHAI[REAL_PART];
   ALPHA[IMAG_PART]       = ALPHAI[IMAG_PART];

   BETA[REAL_PART]        = BETAI[REAL_PART];
   BETA[IMAG_PART]        = BETAI[IMAG_PART];

/* ..
*  .. Executable Statements ..
*
*/
   lside = ( ( SideOp = Mupcase( F2C_CHAR( SIDE )[0] ) ) ==  CLEFT );
   upper = ( ( UploA  = Mupcase( F2C_CHAR( UPLO )[0] ) ) == CUPPER );

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IB, JB, DESCB, &Bi, &Bj, Bd );
   PB_CargFtoC( IC, JC, DESCC, &Ci, &Cj, Cd );

   if (init_llds_f == 1)
   {
   HSCAL_PXHEMM_INIT_LLDS(
        Ad,
        Bd,
        Cd,
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Bd[M_]=%d Bd[MB_]=%d myrow=%d Bd[RSRC_]=%d nprow=%d Bd[LLD_]=%d",
        Bd[M_], Bd[MB_], myrow, Bd[RSRC_], nprow, Bd[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Cd[M_]=%d Cd[MB_]=%d myrow=%d Cd[RSRC_]=%d nprow=%d Cd[LLD_]=%d", 
        Cd[M_], Cd[MB_], myrow, Cd[RSRC_], nprow, Cd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 901 + CTXT_ ) : 0 ) ) )
   {
      if( ( !lside ) && ( SideOp != CRIGHT ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZHEMM", "Illegal SIDE = %c\n", SideOp 
         );
         info = -1;
      }
      else if( ( !upper ) && ( UploA != CLOWER ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZHEMM", "Illegal UPLO = %c\n", UploA 
         );
         info = -2;
      }
      if( lside )
      {
         hscal_PB_Cchkmat( 
              "PZHEMM", "A", M, 3, M, 3, Ai, Aj, Ad,  9,
              nprow, npcol, myrow, mycol,
              &info 
         );
         hscal_PB_Cchkmat( 
              "PZHEMM", "B", M, 3, N, 4, Bi, Bj, Bd, 13,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      else
      {
         hscal_PB_Cchkmat( 
              "PZHEMM", "A", N, 4, N, 4, Ai, Aj, Ad,  9,
              nprow, npcol, myrow, mycol,
              &info 
         );
         hscal_PB_Cchkmat( 
              "PZHEMM", "B", M, 3, N, 4, Bi, Bj, Bd, 13,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      hscal_PB_Cchkmat(
           "PZHEMM", "C", M, 3, N, 4, Ci, Cj, Cd, 18,
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PZHEMM",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( ( M == 0 ) || ( N == 0 ) ||
       ( ( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) ) &&
         ( ( BETA [REAL_PART] ==  ONE ) && ( BETA [IMAG_PART] == ZERO ) ) ) )
   {
      return HSCAL_SUCCESS;
   }

/*
*  Get type structure
*/
   type = PB_Cztypeset();
/*
*  If alpha is zero, sub( C ) := beta * sub( C ).
*/
   if( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) )
   {
      if( ( BETA[REAL_PART] == ZERO ) && ( BETA[IMAG_PART] == ZERO ) )
      {
         hscal_PB_Cplapad_Z( 
              type, ALL, NOCONJG, M, N, type->zero, type->zero,
              Ci, Cj, Cd,
              nprow, npcol, myrow, mycol, tcomp_l
         );
      }
      else if( !( ( BETA[REAL_PART] ==  ONE ) && ( BETA[IMAG_PART] == ZERO ) ) )
      {
         hscal_PB_Cplascal_Z( 
              type, ALL, NOCONJG, M, N, ((char *) BETA), 
              Ci, Cj, Cd,
              nprow, npcol, myrow, mycol, tcomp_l
         );
      }

      return HSCAL_SUCCESS;
   }
/*
*  Start the operations
*/

/*
*  Algorithm selection is based on approximation of the communication volume
*  for distributed and aligned operands.
*
*  ABCest: operands sub( A ), sub( B ) and sub( C ) are communicated (N >> M)
*  BCest : Both operands sub( B ) and sub( C ) are communicated      (M >> N)
*/
   if( lside )
   {
      tmp1   = DNROC( M, Ad[MB_], nprow ); tmp2 = DNROC( N, Bd[NB_], npcol );
      ABCest = (double)(M) *
         ( ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 / TWO ) +
           ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO :
             tmp2 + tmp2 * CBRATIO ) );
      tmp1  = DNROC( M, Ad[MB_], nprow ); tmp2 = DNROC( M, Ad[NB_], npcol );
      tmp3  = DNROC( M, Bd[MB_], nprow ); tmp4 = DNROC( M, Cd[MB_], nprow );
      BCest = (double)(N) *
              ( CBRATIO * ( npcol == 1 ? ZERO : tmp1 ) +
                ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp3 ) +
                ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 ) +
                CBRATIO * ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 ) );
   }
   else
   {
      tmp1   = DNROC( N, Ad[NB_], npcol ); tmp2 = DNROC( M, Bd[MB_], nprow );
      ABCest = (double)(N) *
         ( ( ( ( Ad[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp1 / TWO ) +
           ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO :
             tmp2 + tmp2 * CBRATIO ) );
      tmp1  = DNROC( N, Ad[MB_], nprow ); tmp2 = DNROC( N, Ad[NB_], npcol );
      tmp3  = DNROC( N, Bd[NB_], npcol ); tmp4 = DNROC( N, Cd[NB_], npcol );
      BCest = (double)(M) *
              ( ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) +
                CBRATIO * ( nprow == 1 ? ZERO : tmp2 ) +
                ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 ) +
                CBRATIO * ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp4 ) );
   }
/*
*  Shift a little the cross-over point between both algorithms.
*/
   ChooseABC = ( ( 1.5 * ABCest ) <= BCest );
/*
*  BLACS topologies are enforced iff M and N are strictly greater than the
*  logical block size returned by pilaenv_. Otherwise, it is assumed that the
*  routine calling this routine has already selected an adequate topology.
*/
   nb       = hscal_pilaenv( C2F_CHAR( &type->type ) );
   ForceTop = ( ( M > nb ) && ( N > nb ) );

   rbtop = *PB_Ctop( &ctxt, BCAST,   ROW,    TOP_GET );
   rctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_GET );
   cbtop = *PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_GET );
   cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET );

   if( ChooseABC )
   {
      if( ForceTop )
      {
         rbtopsave = rbtop; rctopsave = rctop;
         cbtopsave = cbtop; cctopsave = cctop;

         if( lside )
         {
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
            if( ( rbtop != CTOP_DRING ) && ( rbtop != CTOP_IRING ) &&
                ( rbtop != CTOP_SRING ) )
               rbtop = *PB_Ctop( &ctxt, BCAST,   ROW,    TOP_IRING );
            if( ( ( cbtop != CTOP_DRING ) && ( cbtop != CTOP_IRING ) &&
                  ( cbtop != CTOP_SRING ) ) || ( cbtop != cctop ) )
            {
               cbtop = *PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_IRING   );
               cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_IRING   );
/*
*  Remove the next 2 lines when the BLACS combine operations support ring
*  topologies
*
*               rctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_DEFAULT );
*               cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_DEFAULT );
*/
            }
         }
         else
         {
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
            if( ( cbtop != CTOP_DRING ) && ( cbtop != CTOP_IRING ) &&
                ( cbtop != CTOP_SRING ) )
               cbtop = *PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_IRING );
            if( ( ( rbtop != CTOP_DRING ) && ( rbtop != CTOP_IRING ) &&
                  ( rbtop != CTOP_SRING ) ) || ( rbtop != rctop ) )
            {
               rbtop = *PB_Ctop( &ctxt, BCAST,   ROW,    TOP_IRING   );
               rctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_IRING   );
/*
*  Remove the next 2 lines when the BLACS combine operations support ring
*  topologies
*
*               rctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_DEFAULT );
*               cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_DEFAULT );
*/
            }
         }
      }
      if( lside )
         DirAB = ( rbtop == CTOP_DRING ? CBACKWARD : CFORWARD );
      else
         DirAB = ( cbtop == CTOP_DRING ? CBACKWARD : CFORWARD );

      {
      char *CONJUG_l = CONJG;

      hscal_PB_CpsymmAB_Z( 
           type, (&DirAB), CONJUG_l, (&SideOp), (&UploA), M, N,
           ((char *)ALPHA), Ai, Aj, Ad, Bi,
           Bj, Bd, ((char *)BETA), Ci, Cj, Cd,
           nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      }
   }
   else
   {
      if( ForceTop )
      {
         rbtopsave = rbtop; rctopsave = rctop;
         cbtopsave = cbtop; cctopsave = cctop;

         if( lside )
         {
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
            if( ( ( rbtop != CTOP_DRING ) && ( rbtop != CTOP_IRING ) &&
                  ( rbtop != CTOP_SRING ) ) || ( rbtop != rctop ) )
            {
               rbtop = *PB_Ctop( &ctxt, BCAST,   ROW,    TOP_IRING   );
               rctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_IRING   );
/*
*  Remove the next 2 lines when the BLACS combine operations support ring
*  topologies
*
*               rctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_DEFAULT );
*               cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_DEFAULT );
*/
            }
            cbtop = *PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_DEFAULT );
            cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_DEFAULT );
         }
         else
         {
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
            if( ( ( cbtop != CTOP_DRING ) && ( cbtop != CTOP_IRING ) &&
                  ( cbtop != CTOP_SRING ) ) || ( cbtop != cctop ) )
            {
               cbtop = *PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_IRING   );
               cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_IRING   );
/*
*  Remove the next 2 lines when the BLACS combine operations support ring
*  topologies
*
*               rctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_DEFAULT );
*               cctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_DEFAULT );
*/
            }
            rbtop = *PB_Ctop( &ctxt, BCAST,   ROW,    TOP_DEFAULT );
            rctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_DEFAULT );
         }
      }
      if( lside )
         DirAB = ( ( rbtop == CTOP_DRING || rctop == CTOP_DRING ) ?
                   CBACKWARD : CFORWARD );
      else
         DirAB = ( ( cbtop == CTOP_DRING || cctop == CTOP_DRING ) ?
                   CBACKWARD : CFORWARD );

      {
      char *CONJUG_l = CONJG;

      hscal_PB_CpsymmBC_Z( 
          type, (&DirAB), CONJUG_l, (&SideOp), (&UploA), M, N,
          ((char *)ALPHA), Ai, Aj, Ad, Bi,
          Bj, Bd, ((char *)BETA), Ci, Cj, Cd,
          nprow, npcol, myrow, mycol, tcomp, tcomm
      );
      }
   }
/*
*  Restore the BLACS topologies when necessary.
*/
   if( ForceTop )
   {
      rbtopsave = *PB_Ctop( &ctxt, BCAST,   ROW,    &rbtopsave );
      rctopsave = *PB_Ctop( &ctxt, COMBINE, ROW,    &rctopsave );
      cbtopsave = *PB_Ctop( &ctxt, BCAST,   COLUMN, &cbtopsave );
      cctopsave = *PB_Ctop( &ctxt, COMBINE, COLUMN, &cctopsave );
   }

/*
*  End of PZHEMM
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzhemm_tcomp_tcomm(
   int SIDEI,
   int UPLOI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzhemm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, UPLOI,
                M, N,
                ALPHA, 
                IA, JA, DESCA, 
                IB, JB, DESCB, 
                BETA,
                IC, JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzhemm_info(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double* BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI = *SIDE;
   int UPLOI = *UPLO;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzhemm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, UPLOI,
                *M, *N,
                alphai,
                *IA, *JA, DESCA, 
                *IB, *JB, DESCB, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzhemm_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double* BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI = *SIDE;
   int UPLOI = *UPLO;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzhemm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, UPLOI,
                *M, *N,
                alphai,
                *IA, *JA, DESCA, 
                *IB, *JB, DESCB, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
