
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pztranu_recon_tcomp(
   int m,
   int n, 
   int alphai[CNARG],
   int betai[CNARG]
)
{
    double tcomp = 0;

    char   transac = 'N';
    char   *transa = &transac;

    char   transbc = 'N';
    char   *transb = &transbc;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;
    int reconm           = m/p;
    int reconn           = n/q;
    int reconk           = hscal_pilaenv_recon("Z");

    int    lda, ldb, ldc;

    double alpha[CNARG], beta[CNARG];

    alpha[REAL_PART] = alphai[REAL_PART];
    alpha[IMAG_PART] = alphai[IMAG_PART];

    beta[REAL_PART]  = betai[REAL_PART];
    beta[IMAG_PART]  = betai[IMAG_PART];

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_zgemm(
         transa,
         transb,
         &reconm,
         &reconn,
         &reconk,
         ((char*)alpha),
         &lda,
         &ldb,
         ((char*)beta),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pztranu_my_tcomp(
   int myrow,
   int mycol,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pztranu_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          ALPHA, 
          IA, JA, DESCA, 
          BETA,
          IC, JC, DESCC, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_my_tcomm(
   int myrow,
   int mycol,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pztranu_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          ALPHA, 
          IA, JA, DESCA,
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pztranu_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          ALPHA, 
          IA, JA, DESCA, 
          BETA,
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pztranu_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int M,
   int N,
   int ALPHAI[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETAI[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   int            Ai, Aj, Ci, Cj, ctxt, info;
/*
*  .. Local Arrays ..
*/
   int            Ad[DLEN_], Cd[DLEN_];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

   double ALPHA[CNARG], BETA[CNARG];

   ALPHA[REAL_PART]       = ALPHAI[REAL_PART];
   ALPHA[IMAG_PART]       = ALPHAI[IMAG_PART];

   BETA[REAL_PART]        = BETAI[REAL_PART];
   BETA[IMAG_PART]        = BETAI[IMAG_PART];

/* ..
*  .. Executable Statements ..
*
*/
   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IC, JC, DESCC, &Ci, &Cj, Cd );

   if (init_llds_f == 1)
   {
   HSCAL_PXTRANU_INIT_LLDS(
        Ad, 
        Cd, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Cd[M_]=%d Cd[MB_]=%d myrow=%d Cd[RSRC_]=%d nprow=%d Cd[LLD_]=%d", 
        Cd[M_], Cd[MB_], myrow, Cd[RSRC_], nprow, Cd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 701 + CTXT_ ) : 0 ) ) )
   {
      hscal_PB_Cchkmat( 
           "PZTRANU", "A", N, 2, M, 1, Ai, Aj, Ad,  7,
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkmat( 
           "PZTRANU", "C", M, 1, N, 2, Ci, Cj, Cd, 12, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PZTRANU",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( ( M == 0 ) || ( N == 0 ) ||
       ( ( ALPHA[REAL_PART] == ZERO && ALPHA[IMAG_PART] == ZERO ) &&
         ( BETA [REAL_PART] ==  ONE && BETA [IMAG_PART] == ZERO ) ) )
   {
      return HSCAL_SUCCESS;
   }

/*
*  And when alpha is zero
*/
   if( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) )
   {
      if( ( BETA[REAL_PART] == ZERO ) && ( BETA[IMAG_PART] == ZERO ) )
      {
         hscal_PB_Cplapad_Z( 
              PB_Cztypeset(), ALL, NOCONJG, M, N, ((char *)BETA),
              ((char *)BETA), Ci, Cj, Cd,
              nprow, npcol, myrow, mycol, tcomp_l
         );
      }
      else
      {
         hscal_PB_Cplascal_Z( 
              PB_Cztypeset(), ALL, NOCONJG, M, N, ((char *)BETA),
              Ci, Cj, Cd,
              nprow, npcol, myrow, mycol, tcomp_l
         );
      }

      return HSCAL_SUCCESS;
   }
/*
*  Start the operations
*/
   {
   PBTYP_T        * type;
   char *CONJUG_l = NOCONJG;

/*
*  Get type structure
*/
   type = PB_Cztypeset();

   hscal_PB_Cptran_Z( 
     type, CONJUG_l,   M, N, ((char *) ALPHA),
     Ai, Aj, Ad, ((char *)  BETA),
     Ci, Cj, Cd,
     nprow, npcol, myrow, mycol, tcomp, tcomm
   );
   }

/*
*  End of PZTRANU
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_tcomp_tcomm(
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int BETA[CNARG],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztranu_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                M, N,
                ALPHA, 
                IA, JA, DESCA, 
                BETA,
                IC, JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_info(
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double* BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztranu_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N,
                alphai,
                *IA, *JA, DESCA, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztranu_info_i(
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double* BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG], betai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   if (BETA[REAL_PART] == 0.)
   {
      betai[REAL_PART] = 0;
   }
   else if (BETA[REAL_PART] == 1.)
   {
      betai[REAL_PART] = 1;
   }
   else
   {
      betai[REAL_PART] = 2;
   }

   if (BETA[IMAG_PART] == 0.)
   {
      betai[IMAG_PART] = 0;
   }
   else if (BETA[IMAG_PART] == 1.)
   {
      betai[IMAG_PART] = 1;
   }
   else
   {
      betai[IMAG_PART] = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztranu_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N,
                alphai,
                *IA, *JA, DESCA, 
                betai,
                *IC, *JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
