
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pztrmm_recon_tcomp(
   int sidei,
   int uploi,
   int transi,
   int diagi,
   int m,
   int n, 
   int alphai[CNARG]
)
{
    double tcomp   = 0;

    char   sidec   = sidei;
    char   *side   = &sidec;

    char   uploc   = uploi;
    char   *uplo   = &uploc;

    char   transc  = transi;
    char   *trans   = &transc;

    char   diagc   = diagi;
    char   *diag   = &diagc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    double alpha[CNARG];

    alpha[REAL_PART] = alphai[REAL_PART];
    alpha[IMAG_PART] = alphai[IMAG_PART];

    if (transc == 'N')
    {
       reconm  = m/p;
       reconn  = hscal_pilaenv_recon("Z");
       reconk  = n/q;
    }
    else
    {
       reconm  = hscal_pilaenv_recon("Z");
       reconn  = m/p;
       reconk  = n/q;
    }

    if (transc == 'N')
    {
       lda = reconm;
    }
    else
    {
       lda = reconk;
    }

    if (transc == 'N')
    {
       ldb = reconk;
    }
    else
    {
       ldb = reconn;
    }

    ldc = reconm;

    hscal_zgemm(
         trans,
         trans,
         &reconm,
         &reconn,
         &reconk,
         ((char*)alpha),
         &lda,
         &ldb,
         ((char*)alpha),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pztrmm_my_tcomp(
   int myrow,
   int mycol,
   int SIDEI,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pztrmm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, UPLOI,
          TRANSI, DIAGI,
          M, N,
          ALPHA, 
          IA, JA, DESCA, 
          IB, JB, DESCB, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_my_tcomm(
   int myrow,
   int mycol,
   int SIDEI,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pztrmm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, UPLOI,
          TRANSI, DIAGI,
          M, N,
          ALPHA, 
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int SIDEI,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pztrmm_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, UPLOI,
          TRANSI, DIAGI,
          M, N,
          ALPHA, 
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int SIDEI,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int M,
   int N,
   int ALPHAI[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           DiagA, DirB, OpC, OpR, SideOp, TopC, TopR, TranOp, UploA,
                  Var, ctop, ctopsave, rtop, rtopsave;
   int            Ai, Aj, Bi, Bj, ChooseAB, ForceTop, ctxt, info, lside,
                  nb, notran, nounit, upper;
   double         ABestL, ABestR, Best, tmp1, tmp2, tmp3, tmp4;
   PBTYP_T        * type;
/*
*  .. Local Arrays ..
*/
   int            Ad[DLEN_], Bd[DLEN_];

   double         * tcomp_l = tcomp;
   int            * tcomm_l = tcomm;

   char SIDEC     = SIDEI, *SIDE = &SIDEC;
   char UPLOC     = UPLOI, *UPLO = &UPLOC;
   char TRANSC    = TRANSI, *TRANS = &TRANSC;
   char DIAGC     = DIAGI, *DIAG = &DIAGC;

   double ALPHA[CNARG];

   ALPHA[REAL_PART]       = ALPHAI[REAL_PART];
   ALPHA[IMAG_PART]       = ALPHAI[IMAG_PART];

/* ..
*  .. Executable Statements ..
*
*/
   lside  = ( ( SideOp = Mupcase( F2C_CHAR( SIDE  )[0] ) ) ==   CLEFT );
   upper  = ( ( UploA  = Mupcase( F2C_CHAR( UPLO  )[0] ) ) ==  CUPPER );
   notran = ( ( TranOp = Mupcase( F2C_CHAR( TRANS )[0] ) ) == CNOTRAN );
   nounit = ( ( DiagA  = Mupcase( F2C_CHAR( DIAG  )[0] ) ) == CNOUNIT );

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IB, JB, DESCB, &Bi, &Bj, Bd );

   if (init_llds_f == 1)
   {
   HSCAL_PXTRMM_INIT_LLDS(
        Ad,
        Bd,
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Bd[M_]=%d Bd[MB_]=%d myrow=%d Bd[RSRC_]=%d nprow=%d Bd[LLD_]=%d",
        Bd[M_], Bd[MB_], myrow, Bd[RSRC_], nprow, Bd[LLD_]
   );

   if( !( info = ( ( nprow == -1 ) ? -( 1101 + CTXT_ ) : 0 ) ) )
   {
      if( ( !lside ) && ( SideOp != CRIGHT ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZTRMM", "Illegal SIDE = %c\n", SideOp 
         );
         info = -1;
      }
      else if( ( !upper ) && ( UploA != CLOWER ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZTRMM", "Illegal UPLO = %c\n", UploA 
         );
         info = -2;
      }
      else if( ( !notran ) && ( TranOp != CTRAN ) && ( TranOp != CCOTRAN ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZTRMM", "Illegal TRANS = %c\n", TranOp 
         );
         info = -3;
      }
      if( ( !nounit ) && ( DiagA != CUNIT ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZTRMM", "Illegal DIAG = %c\n", DiagA 
         );
         info = -4;
      }
      if( lside )
      {
         hscal_PB_Cchkmat(
              "PZTRMM", "A", M, 5, M, 5, Ai, Aj, Ad, 11,
              nprow, npcol, myrow, mycol,
              &info 
         );
      }
      else
      {
         hscal_PB_Cchkmat(
              "PZTRMM", "A", N, 6, N, 6, Ai, Aj, Ad, 11,
              nprow, npcol, myrow, mycol,
              &info );
      }
      hscal_PB_Cchkmat(
           "PZTRMM", "B", M, 5, N, 6, Bi, Bj, Bd, 15,
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PZTRMM",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }
/*
*  Quick return if possible
*/
   if( M == 0 || N == 0 ) 
   {
      return HSCAL_SUCCESS;
   }

/*
*  Get type structure
*/
   type = PB_Cztypeset();
/*
*  And when alpha is zero
*/
   if( ( ALPHA[REAL_PART] == ZERO ) && ( ALPHA[IMAG_PART] == ZERO ) )
   {
      hscal_PB_Cplapad_C( 
           type, ALL, NOCONJG, M, N, type->zero, type->zero,
           Bi, Bj, Bd,
           nprow, npcol, myrow, mycol, tcomp_l
      );

      return HSCAL_SUCCESS;
   }
/*
*  Start the operations
*/

/*
*  Algorithm selection is based on approximation of the communication volume
*  for distributed and aligned operands.
*
*  ABestR, ABestL : both operands sub( A ) and sub( B ) are communicated
*                   ( N >> M when SIDE is left and M >> N otherwise )
*  Best           : only sub( B ) is communicated
*                   ( M >> N when SIDE is left and N >> M otherwise )
*/
   if( lside )
   {
      if( notran )
      {
         tmp1 = DNROC( M, Ad[MB_], nprow ); tmp4 = DNROC( N, Bd[NB_], npcol );
         ABestR = (double)(M) *
           ( ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 / TWO ) +
             ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp4 ) );
         tmp1 = DNROC( M, Ad[MB_], nprow ); tmp2 = DNROC( M, Ad[NB_], npcol );
                                             tmp4 = DNROC( M, Bd[MB_], nprow );
         Best = (double)(N) *
                ( CBRATIO * ( npcol == 1 ? ZERO : tmp1 ) +
                  ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 ) );
         ChooseAB = ( ( 1.1 * ABestR ) <= Best );
      }
      else
      {
         tmp1 = DNROC( M, Ad[MB_], nprow ); tmp2 = DNROC( M, Ad[NB_], npcol );
         tmp4 = DNROC( N, Bd[NB_], npcol );
         ABestL = (double)(M) *
           ( ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 / TWO ) +
             CBRATIO *
             ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp4 ) );
         ABestR = (double)(M) *
           ( ( ( ( Ad[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 / TWO ) +
             ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp4       ) +
             MAX( tmp2, tmp1 ) / TWO );
         tmp1 = DNROC( M, Ad[MB_], nprow ); tmp2 = DNROC( M, Ad[NB_], npcol );
                                             tmp4 = DNROC( M, Bd[MB_], nprow );
         Best = (double)(N) *
                ( ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp1 ) +
                  CBRATIO * ( nprow == 1 ? ZERO : tmp2 ) + MAX( tmp2, tmp4 ) );
         ChooseAB = ( ( ( 1.1 * ABestL ) <= Best ) ||
                      ( ( 1.1 * ABestR ) <= Best ) );
      }
   }
   else
   {
      if( notran )
      {
         tmp2 = DNROC( N, Ad[NB_], npcol ); tmp3 = DNROC( M, Bd[MB_], nprow );
         ABestR = (double)(N) *
           ( ( ( ( Ad[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 / TWO ) +
             ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp3 ) );
         tmp1 = DNROC( N, Ad[MB_], nprow ); tmp2 = DNROC( N, Ad[NB_], npcol );
         tmp3 = DNROC( N, Bd[NB_], npcol );
         Best = (double)(M) *
                ( CBRATIO * ( nprow == 1 ? ZERO : tmp2 ) +
                  ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) );
         ChooseAB = ( ( 1.1 * ABestR ) <= Best );
      }
      else
      {
         tmp1 = DNROC( N, Ad[MB_], nprow ); tmp2 = DNROC( N, Ad[NB_], npcol );
         tmp3 = DNROC( M, Bd[MB_], nprow );
         ABestL = (double)(N) *
           ( ( ( ( Ad[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 / TWO ) +
             CBRATIO *
             ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp3 ) );
         ABestR = (double)(N) *
           ( ( ( ( Ad[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 / TWO ) +
             ( ( ( Bd[CSRC_] == -1 ) || ( npcol == 1 ) ) ? ZERO : tmp3       ) +
             MAX( tmp2, tmp1 ) / TWO );
         tmp1 = DNROC( N, Ad[MB_], nprow ); tmp2 = DNROC( N, Ad[NB_], npcol );
         tmp3 = DNROC( N, Bd[NB_], npcol );
         Best = (double)(M) *
                ( ( ( ( Bd[RSRC_] == -1 ) || ( nprow == 1 ) ) ? ZERO : tmp2 ) +
                  CBRATIO * ( npcol == 1 ? ZERO : tmp1 ) + MAX( tmp1, tmp3 ) );
         ChooseAB = ( ( ( 1.1 * ABestL ) <= Best ) ||
                      ( ( 1.1 * ABestR ) <= Best ) );
      }
   }
/*
*  BLACS topologies are enforced iff M and N are strictly greater than the
*  logical block size returned by pilaenv_. Otherwise, it is assumed that the
*  routine calling this routine has already selected an adequate topology.
*/
   nb       = hscal_pilaenv( C2F_CHAR( &type->type ) );
   ForceTop = ( ( M > nb ) && ( N > nb ) );

   if( ChooseAB )
   {
      if( lside )
      {
         if( notran )
         {
            OpR = CBCAST; OpC = CBCAST; Var = CRIGHT;
            if( upper ) { TopR = TopC = CTOP_IRING; }
            else        { TopR = TopC = CTOP_DRING; }
         }
         else
         {
            if( ABestL <= ABestR )
            {
               OpR = CBCAST; OpC = CCOMBINE; Var = CLEFT;
               if( upper ) { TopR = CTOP_DRING; TopC = CTOP_IRING; }
               else        { TopR = CTOP_IRING; TopC = CTOP_DRING; }
            }
            else
            {
               OpR = CBCAST; OpC = CBCAST;  Var = CRIGHT;
               if( upper ) { TopR = TopC = CTOP_DRING; }
               else        { TopR = TopC = CTOP_IRING; }
            }
         }
      }
      else
      {
         if( notran )
         {
            OpR = CBCAST; OpC = CBCAST; Var = CRIGHT;
            if( upper ) { TopR = TopC = CTOP_DRING; }
            else        { TopR = TopC = CTOP_IRING; }
         }
         else
         {
            if( ABestL <= ABestR )
            {
               OpR = CCOMBINE; OpC = CBCAST; Var = CLEFT;
               if( upper ) { TopR = CTOP_DRING; TopC = CTOP_IRING; }
               else        { TopR = CTOP_IRING; TopC = CTOP_DRING; }
            }
            else
            {
               OpR = CBCAST; OpC = CBCAST;  Var = CRIGHT;
               if( upper ) { TopR = TopC = CTOP_IRING; }
               else        { TopR = TopC = CTOP_DRING; }
            }
         }
      }

      rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_GET );
      ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_GET );

      if( ForceTop )
      {
         if( ( rtopsave = rtop ) != TopR )
            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    &TopR );
         if( ( ctopsave = ctop ) != TopC )
            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, &TopC );
/*
*  Remove the next 4 lines when the BLACS combine operations support ring
*  topologies
*
*         if( OpR == CCOMBINE )
*            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_DEFAULT );
*         if( OpC == CCOMBINE )
*            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_DEFAULT );
*/
      }

      hscal_PB_CptrmmAB_Z( 
        type, (&Var), (&SideOp), (&UploA), (&TranOp), (&DiagA), M, N,
        ((char *)ALPHA), Ai, Aj, Ad, Bi,
        Bj, Bd,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
   else
   {
      if( ( lside && notran ) || ( !( lside ) && !( notran ) ) )
      {
         OpR = CCOMBINE; OpC = CBCAST;
         rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_GET );
         ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_GET );

         if( ForceTop )
         {
            rtopsave = rtop;
            ctopsave = ctop;
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
            if( ( rtop != CTOP_DRING ) && ( rtop != CTOP_IRING ) &&
                ( rtop != CTOP_SRING ) )
               rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_SRING );
            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_DEFAULT );
/*
*  Remove the next line when the BLACS combine operations support ring
*  topologies
*
*            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_DEFAULT );
*/
         }
      }
      else
      {
         OpR = CBCAST; OpC = CCOMBINE;
         rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_GET );
         ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_GET );

         if( ForceTop )
         {
            rtopsave = rtop;
            ctopsave = ctop;
/*
*  No clear winner for the ring topologies, so that if a ring topology is
*  already selected, keep it.
*/
            if( ( ctop != CTOP_DRING ) && ( ctop != CTOP_IRING ) &&
                ( ctop != CTOP_SRING ) )
               ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_SRING );
            rtop = *PB_Ctop( &ctxt, &OpR, ROW,    TOP_DEFAULT );
/*
*  Remove the next line when the BLACS combine operations support ring
*  topologies
*
*            ctop = *PB_Ctop( &ctxt, &OpC, COLUMN, TOP_DEFAULT );
*/
         }
      }

      if( lside )
         DirB = ( rtop == CTOP_DRING ? CBACKWARD : CFORWARD );
      else
         DirB = ( ctop == CTOP_DRING ? CBACKWARD : CFORWARD );

      hscal_PB_CptrmmB_Z(
        type, (&DirB), (&SideOp), (&UploA), (&TranOp), (&DiagA), M, N,
        ((char *)ALPHA), Ai, Aj, Ad, Bi,
        Bj, Bd,
        nprow, npcol, myrow, mycol, tcomp, tcomm
      );
   }
/*
*  Restore the BLACS topologies when necessary.
*/
   if( ForceTop )
   {
      rtopsave = *PB_Ctop( &ctxt, &OpR, ROW,    &rtopsave );
      ctopsave = *PB_Ctop( &ctxt, &OpC, COLUMN, &ctopsave );
   }

/*
*  End of PZTRMM
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_tcomp_tcomm(
   int SIDEI,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int M,
   int N,
   int ALPHA[CNARG],
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IB,
   int JB,
   int DESCB[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztrmm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, UPLOI,
                TRANSI, DIAGI,
                M, N,
                ALPHA, 
                IA, JA, DESCA, 
                IB, JB, DESCB, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_info(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;
   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;
   int DIAGI  = *DIAG;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztrmm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, UPLOI,
                TRANSI, DIAGI,
                *M, *N,
                alphai,
                *IA, *JA, DESCA, 
                *IB, *JB, DESCB, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrmm_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   double* ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;
   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;
   int DIAGI  = *DIAG;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai[CNARG];

   if (ALPHA[REAL_PART] == 0.)
   {
      alphai[REAL_PART] = 0;
   }
   else if (ALPHA[REAL_PART] == 1.)
   {
      alphai[REAL_PART] = 1;
   }
   else
   {
      alphai[REAL_PART] = 2;
   }

   if (ALPHA[IMAG_PART] == 0.)
   {
      alphai[IMAG_PART] = 0;
   }
   else if (ALPHA[IMAG_PART] == 1.)
   {
      alphai[IMAG_PART] = 1;
   }
   else
   {
      alphai[IMAG_PART] = 2;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztrmm_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, UPLOI,
                TRANSI, DIAGI,
                *M, *N,
                alphai,
                *IA, *JA, DESCA, 
                *IB, *JB, DESCB, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
