
#include <math.h>
#include <stdlib.h>
#include <stdio.h>

#include "../../include/scalapack.h"

/*------------------------------------------------------------------*/

double hscal_pztrsv_recon_tcomp(
   int uploi,
   int transi,
   int diagi,
   int n,
   int incx
)
{
    double tcomp = 0;

    char uploc   = uploi;
    char *uplo   = &uploc;
    char transc  = transi;
    char *trans  = &transc;
    char diagc   = diagi;
    char *diag   = &diagc;

    int reconn   = hscal_pilaenv_recon("Z");
    int lda      = reconn;
    
    hscal_ztrsv(
         uplo,
         trans,
         diag,
         &reconn,
         &lda,
         &incx,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f", tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pztrsv_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pztrsv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          IA, JA, DESCA, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pztrsv_my_tcomm(
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pztrsv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          IA, JA, DESCA, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          NULL, tcomm
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrsv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pztrsv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, TRANSI, DIAGI,
          N, 
          IA, JA, DESCA, 
          IX, JX, DESCX, INCX,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pztrsv_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
/*
*  .. Local Scalars ..
*/
   char           DiagA, TranOp, UploA, Xroc, btop, ctop, * negone, * one,
                  * zero;
   int            Acol, Ai, Aii, Aimb1, Ainb1, Aj, Ajj, Akp, Akq, Ald, Amb, Anb,
                  Anp, Anp0, Anq, Anq0, Arow, Asrc, XACapbX, XACfr, XACld,
                  XACsum, XARapbX, XARfr, XARld, XARsum, Xi, Xj, ctxt, info,
                  ione=1, k, kb, kbnext, kbprev, ktmp, nb, notran,
                  nounit, size, upper;
   PBTYP_T        * type;
   HSCALBLACSCONTEXT   * blacsctxt;

   char UPLOC    = UPLOI, *UPLO = &UPLOC;
   char TRANSC   = TRANSI, *TRANS = &TRANSC;
   char DIAGC    = DIAGI, *DIAG = &DIAGC;

/*
*  .. Local Arrays ..
*/
   int            Ad[DLEN_], Ad0[DLEN_], XACd[DLEN_], XARd[DLEN_], Xd[DLEN_];
/* ..
*  .. Executable Statements ..
*
*/
   upper  = ( ( UploA  = Mupcase( F2C_CHAR( UPLO  )[0] ) ) ==  CUPPER );
   notran = ( ( TranOp = Mupcase( F2C_CHAR( TRANS )[0] ) ) == CNOTRAN );
   nounit = ( ( DiagA  = Mupcase( F2C_CHAR( DIAG  )[0] ) ) == CNOUNIT );

   PB_CargFtoC( IA, JA, DESCA, &Ai, &Aj, Ad );
   PB_CargFtoC( IX, JX, DESCX, &Xi, &Xj, Xd );

   if (init_llds_f == 1)
   {
   HSCAL_PXTRSV_INIT_LLDS(
        N, 
        Ad, 
        Xd, IX, INCX, 
        myrow, nprow
   );
   }

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Ad[M_]=%d Ad[MB_]=%d myrow=%d Ad[RSRC_]=%d nprow=%d Ad[LLD_]=%d", 
        Ad[M_], Ad[MB_], myrow, Ad[RSRC_], nprow, Ad[LLD_]
   );

   HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "Xd[M_]=%d Xd[MB_]=%d myrow=%d Xd[RSRC_]=%d nprow=%d Xd[LLD_]=%d", 
        Xd[M_], Xd[MB_], myrow, Xd[RSRC_], nprow, Xd[LLD_]
   );

/*
*  Test the input parameters
*/
   if( !( info = ( ( nprow == -1 ) ? -( 801 + CTXT_ ) : 0 ) ) )
   {
      if( ( !upper ) && ( UploA != CLOWER ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZTRSV", "Illegal UPLO = %c\n", UploA 
         );
         info = -1;
      }
      else if( ( !notran ) && ( TranOp != CTRAN ) && ( TranOp != CCOTRAN ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZTRSV", "Illegal TRANS = %c\n", TranOp 
         );
         info = -2;
      }
      else if( ( !nounit ) && ( DiagA != CUNIT ) )
      {
         hscal_PB_Cwarn(
              nprow, npcol, myrow, mycol,
              __LINE__, "PZTRSV", "Illegal DIAG = %c\n", DiagA 
         );
         info = -3;
      }
      hscal_PB_Cchkmat(
           "PZTRSV", "A", N, 4, N, 4, Ai, Aj, Ad,  8, 
           nprow, npcol, myrow, mycol,
           &info 
      );
      hscal_PB_Cchkvec(
           "PZTRSV", "X", N, 4, Xi, Xj, Xd, INCX, 12, 
           nprow, npcol, myrow, mycol,
           &info 
      );
   }

   if( info )
   {
       hscal_PB_Cabort(
            "PZTRSV",
            info,
            nprow, npcol, myrow, mycol
       );

       return HSCAL_ERR_INPUT;
   }

/*
*  Quick return if possible
*/
   if( N == 0 ) return HSCAL_SUCCESS;
/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*
*  Get type structure
*/
   type = PB_Cztypeset();
   size = type->size; one = type->one; zero = type->zero; negone = type->negone;

/*
*  Compute descriptor Ad0 for sub( A )
*/

   PB_Cdescribe( N, N, Ai, Aj, Ad, nprow, npcol, myrow, mycol, &Aii, &Ajj,
                 &Ald, &Aimb1, &Ainb1, &Amb, &Anb, &Arow, &Acol, Ad0 );
/*
*  Computational partitioning size is computed as the product of the logical
*  value returned by pilaenv_ and 2 * lcm( nprow, npcol )
*/
   nb = 2 * hscal_pilaenv( C2F_CHAR( &type->type ) ) *
        PB_Clcm( ( Arow >= 0 ? nprow : 1 ), ( Acol >= 0 ? npcol : 1 ) );

   Xroc = ( INCX == Xd[M_] ? CROW : CCOLUMN );

   if( notran )
   {
      if( upper )
      {
/*
*  Save current and enforce ring BLACS topologies
*/
         btop = *PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_GET   );
         ctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_GET   );
         (void)  PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_DRING );
         (void)  PB_Ctop( &ctxt, COMBINE, ROW,    TOP_DRING );
/*
*  Remove next line when BLACS combine operations support ring topologies.
*
*        (void)  PB_Ctop( &ctxt, COMBINE, ROW,    TOP_DEFAULT );
*/

/*
*  Reuse sub( X ) and/or create vector XAC in process column owning the last
*  column of sub( A )
*/
         {
         char *CONJUG_l = NOCONJG;
         char *ROWCOL_l = COLUMN;
         int K_l = 1;
         hscal_PB_CInOutV2_Z( 
           type, CONJUG_l, ROWCOL_l, N, N, N-1, Ad0, K_l,
           Xi, Xj, Xd, (&Xroc), XACd,
           (&XACsum), (&XACapbX),
           nprow, npcol, myrow, mycol, tcomp, tcomm           
         );
         }
/*
*  Create vector XAR in process rows spanned by sub( A )
*/
         {
         char * ZEROIT_l = INIT;
         int K_l = 1;
         char *ROWCOL_l = ROW;

         hscal_PB_COutV_Z( 
           type, ROWCOL_l,    ZEROIT_l, N, N, Ad0, K_l, XARd,
           (&XARsum),
           nprow, npcol, myrow, mycol, tcomp
         );
         }
/*
*  Retrieve local quantities related to Ad0 -> sub( A )
*/
         Aimb1 = Ad0[IMB_ ]; Ainb1 = Ad0[INB_ ];
         Amb   = Ad0[MB_  ]; Anb   = Ad0[NB_  ];
         Arow  = Ad0[RSRC_]; Acol  = Ad0[CSRC_]; Ald = Ad0[LLD_];
         Anp   = PB_Cnumroc( N, 0, Aimb1, Amb, myrow, Arow, nprow );
         Anq   = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );
         if( ( Anp > 0 ) && ( Anq > 0 ) )
         {
            ;
         }

         XACld = XACd[LLD_]; XARld = XARd[LLD_];

         for( k = ( ( N - 1 ) / nb ) * nb; k >= 0; k -= nb )
         {
            ktmp = N - k; kb = MIN( ktmp, nb );
/*
*  Solve logical diagonal block, XAC contains the solution scattered in multiple
*  process columns and XAR contains the solution replicated in the process rows.
*/
            Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
            Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
            hscal_PB_Cptrsv_Z( 
              type, XARsum, (&UploA), (&TranOp), (&DiagA), kb, k, k,
              Ad0, 1, XARld,
              nprow, npcol, myrow, mycol, tcomp, tcomm           
            );
/*
*  Update: only the part of sub( X ) to be solved at the next step is locally
*  updated and combined, the remaining part of the vector to be solved later is
*  only locally updated.
*/
            if( Akp > 0 )
            {
               Anq0 = PB_Cnumroc( kb, k, Ainb1, Anb, mycol, Acol, npcol );
               if( XACsum )
               {
                  kbprev = MIN( k, nb );
                  ktmp   = PB_Cnumroc( kbprev, k-kbprev, Aimb1, Amb, myrow,
                                       Arow, nprow );
                  Akp   -= ktmp;

                  if( ktmp > 0 )
                  {
                     if( Anq0 > 0 )
                     {
                        zgemv_tcomp_macro( 
                           TRANS, &ktmp, &Anq0, negone,
                           &Ald,
                           &XARld, one,
                           &ione,
                           nprow, npcol, myrow, mycol, tcomp 
                        );
                     }
                     Asrc = PB_Cindxg2p( k-1, Ainb1, Anb, Acol, Acol, npcol );
                     hscal_Czgsum2d( 
                             blacsctxt, ROW, &ctop, ktmp, 1, XACld, myrow, Asrc,
                             nprow, npcol, myrow, mycol, tcomp, tcomm           
                     );
                     if( mycol != Asrc )
                     {
                        zset_tcomp_macro( 
                            &ktmp, zero, &ione,
                            nprow, npcol, myrow, mycol, tcomp
                        );
                     }
                  }
                  if( Akp > 0 && Anq0 > 0 )
                  {
                     zgemv_tcomp_macro( 
                          TRANS, &Akp, &Anq0, negone,
                          &Ald,
                          &XARld, one,
                          &ione,
                          nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
               else
               {
                  if( Anq0 > 0 )
                  {
                     zgemv_tcomp_macro( 
                          TRANS, &Akp, &Anq0, negone,
                          &Ald,
                          &XARld, one,
                          &ione,
                          nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
            }
         }
/*
*  Combine the scattered resulting vector XAC
*/
         if( XACsum && ( Anp > 0 ) )
         {
            hscal_Czgsum2d( 
                    blacsctxt, ROW, &ctop, Anp, 1, XACld, myrow,
                    XACd[CSRC_],
                    nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
/*
*  sub( X ) := XAC (if necessary)
*/
         if( XACapbX )
         {
            int N_l = 1;
            int IA_l = 0;
            int JA_l = 0;
            char *CONJUG_l = NOCONJG;

            hscal_PB_Cpaxpby_Z( 
               type, CONJUG_l, N, N_l, one, IA_l, JA_l, XACd, COLUMN,
               zero, Xi, Xj, Xd, &Xroc,
               nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
/*
*  Restore BLACS topologies
*/
         (void) PB_Ctop( &ctxt, BCAST,   COLUMN, &btop );
         (void) PB_Ctop( &ctxt, COMBINE, ROW,    &ctop );
      }
      else
      {
/*
*  Save current and enforce ring BLACS topologies
*/
         btop = *PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_GET   );
         ctop = *PB_Ctop( &ctxt, COMBINE, ROW,    TOP_GET   );
         (void)  PB_Ctop( &ctxt, BCAST,   COLUMN, TOP_IRING );
         (void)  PB_Ctop( &ctxt, COMBINE, ROW,    TOP_IRING );
/*
*  Remove next line when BLACS combine operations support ring topologies.
*
*        (void)  PB_Ctop( &ctxt, COMBINE, ROW,    TOP_DEFAULT );
*/
/*
*  Reuse sub( X ) and/or create vector XAC in process column owning the first
*  column of sub( A )
*/
         {
         char *CONJUG_l = NOCONJG;
         char *ROWCOL_l = COLUMN;
         int K_l = 1;
         hscal_PB_CInOutV2_Z( 
           type, CONJUG_l, ROWCOL_l, N, N, 0, Ad0, K_l,
           Xi, Xj, Xd, (&Xroc), XACd,
           (&XACsum), (&XACapbX),
           nprow, npcol, myrow, mycol, tcomp, tcomm           
         );
         }
/*
*  Create vector XAR in process rows spanned by sub( A )
*/
         {
         char * ZEROIT_l = INIT;
         int K_l = 1;
         char *ROWCOL_l = ROW;

         hscal_PB_COutV_Z( 
           type, ROWCOL_l,    ZEROIT_l, N, N, Ad0, K_l, XARd,
           (&XARsum),
           nprow, npcol, myrow, mycol, tcomp
         );
         }
/*
*  Retrieve local quantities related to Ad0 -> sub( A )
*/
         Aimb1 = Ad0[IMB_ ]; Ainb1 = Ad0[INB_ ];
         Amb   = Ad0[MB_  ]; Anb   = Ad0[NB_  ];
         Arow  = Ad0[RSRC_]; Acol  = Ad0[CSRC_]; Ald = Ad0[LLD_];
         Anp   = PB_Cnumroc( N, 0, Aimb1, Amb, myrow, Arow, nprow );
         Anq   = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );
         if( ( Anp > 0 ) && ( Anq > 0 ) )
         {
            ;
         }

         XACld = XACd[LLD_]; XARld = XARd[LLD_];

         for( k = 0; k < N; k += nb )
         {
            ktmp = N - k; kb = MIN( ktmp, nb );
/*
*  Solve logical diagonal block, XAC contains the solution scattered in multiple
*  process columns and XAR contains the solution replicated in the process rows.
*/
            Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
            Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
            hscal_PB_Cptrsv_Z( 
                     type, XARsum, (&UploA), (&TranOp), (&DiagA), kb, k, k,
                     Ad0, 1, XARld,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
            );
/*
*  Update: only the part of sub( X ) to be solved at the next step is locally
*  updated and combined, the remaining part of the vector to be solved later is
*  only locally updated.
*/
            Akp = PB_Cnumroc( k+kb, 0, Aimb1, Amb, myrow, Arow, nprow );
            if( ( Anp0 = Anp - Akp ) > 0 )
            {
               Anq0 = PB_Cnumroc( kb, k, Ainb1, Anb, mycol, Acol, npcol );
               if( XACsum )
               {
                  kbnext = ktmp - kb; kbnext = MIN( kbnext, nb );
                  ktmp   = PB_Cnumroc( kbnext, k+kb, Aimb1, Amb, myrow, Arow,
                                       nprow );
                  Anp0  -= ktmp;

                  if( ktmp > 0 )
                  {
                     if( Anq0 > 0 )
                     {
                        zgemv_tcomp_macro( 
                          TRANS, &ktmp, &Anq0, negone,
                          &Ald,
                          &XARld, one,
                          &ione,
                          nprow, npcol, myrow, mycol, tcomp
                        );
                     }
                     Asrc = PB_Cindxg2p( k+kb, Ainb1, Anb, Acol, Acol, npcol );
                     hscal_Czgsum2d( 
                          blacsctxt, ROW, &ctop, ktmp, 1, XACld, myrow, Asrc,
                          nprow, npcol, myrow, mycol, tcomp, tcomm
                     );
                     if( mycol != Asrc )
                     {
                        zset_tcomp_macro( 
                            &ktmp, zero, &ione,
                            nprow, npcol, myrow, mycol, tcomp
                        );
                     }
                  }
                  if( Anp0 > 0 && Anq0 > 0 )
                  {
                     zgemv_tcomp_macro( 
                      TRANS, &Anp0, &Anq0, negone,
                      &Ald,
                      &XARld,
                      one,
                      &ione,
                      nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
               else
               {
                  if( Anq0 > 0 )
                  {
                     zgemv_tcomp_macro( 
                          TRANS, &Anp0, &Anq0, negone,
                          &Ald,
                          &XARld,
                          one,
                          &ione,
                          nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
            }
         }
/*
*  Combine the scattered resulting vector XAC
*/
         if( XACsum && ( Anp > 0 ) )
         {
            hscal_Czgsum2d( 
                    blacsctxt, ROW, &ctop, Anp, 1, XACld, myrow,
                      XACd[CSRC_],
                    nprow, npcol, myrow, mycol, tcomp, tcomm

            );
         }
/*
*  sub( X ) := XAC (if necessary)
*/
         if( XACapbX )
         {
            int N_l = 1;
            int IA_l = 0;
            int JA_l = 0;
            char *CONJUG_l = NOCONJG;

            hscal_PB_Cpaxpby_Z( 
              type, CONJUG_l, N, N_l, one, IA_l, JA_l, XACd,
              COLUMN, zero, Xi, Xj, Xd, (&Xroc),
              nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
/*
*  Restore BLACS topologies
*/
         (void) PB_Ctop( &ctxt, BCAST,   COLUMN, &btop );
         (void) PB_Ctop( &ctxt, COMBINE, ROW,    &ctop );
      }
   }
   else
   {
      if( upper )
      {
/*
*  Save current and enforce ring BLACS topologies
*/
         btop = *PB_Ctop( &ctxt, BCAST,   ROW,    TOP_GET   );
         ctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET   );
         (void)  PB_Ctop( &ctxt, BCAST,   ROW,    TOP_IRING );
         (void)  PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_IRING );
/*
*  Remove next line when BLACS combine operations support ring topologies.
*
*        (void)  PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_DEFAULT );
*/
/*
*  Reuse sub( X ) and/or create vector XAR in process row owning the first row
*  of sub( A )
*/
         {
         char *CONJUG_l = NOCONJG;
         char *ROWCOL_l = ROW;
         int K_l = 1;
         hscal_PB_CInOutV2_Z( 
           type, CONJUG_l, ROWCOL_l, N, N, 0, Ad0, K_l,
           Xi, Xj, Xd, (&Xroc), XARd,
           (&XARsum), (&XARapbX),
           nprow, npcol, myrow, mycol, tcomp, tcomm
         );
         }
/*
*  Create vector XAC in process columns spanned by sub( A )
*/
         {
         char * ZEROIT_l = INIT;
         int K_l = 1;
         char *ROWCOL_l = COLUMN;

         hscal_PB_COutV_Z( 
           type, ROWCOL_l, ZEROIT_l, N, N, Ad0, K_l, XACd,
           (&XACsum),
           nprow, npcol, myrow, mycol, tcomp
         );
         }
/*
*  Retrieve local quantities related to Ad0 -> sub( A )
*/
         Aimb1 = Ad0[IMB_ ]; Ainb1 = Ad0[INB_ ];
         Amb   = Ad0[MB_  ]; Anb   = Ad0[NB_  ];
         Arow  = Ad0[RSRC_]; Acol  = Ad0[CSRC_]; Ald = Ad0[LLD_];
         Anp   = PB_Cnumroc( N, 0, Aimb1, Amb, myrow, Arow, nprow );
         Anq   = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );
         if( ( Anp > 0 ) && ( Anq > 0 ) )
         {
            ;
         }

         XACld = XACd[LLD_]; XARld = XARd[LLD_];

         for( k = 0; k < N; k += nb )
         {
            ktmp = N - k; kb = MIN( ktmp, nb );
/*
*  Solve logical diagonal block, XAR contains the solution scattered in multiple
*  process rows and XAC contains the solution replicated in the process columns.
*/
            Akp = PB_Cnumroc( k, 0, Aimb1, Amb, myrow, Arow, nprow );
            Akq = PB_Cnumroc( k, 0, Ainb1, Anb, mycol, Acol, npcol );
            hscal_PB_Cptrsv_Z( 
                     type, XACsum, (&UploA), (&TranOp), (&DiagA), kb, k, k,
                     Ad0, 1, XARld,
                     nprow, npcol, myrow, mycol, tcomp, tcomm           
            );
/*
*  Update: only the part of sub( X ) to be solved at the next step is locally
*  updated and combined, the remaining part of the vector to be solved later is
*  only locally updated.
*/
            Akq = PB_Cnumroc( k+kb, 0, Ainb1, Anb, mycol, Acol, npcol );
            if( ( Anq0 = Anq - Akq ) > 0 )
            {
               Anp0 = PB_Cnumroc( kb, k, Aimb1, Amb, myrow, Arow, nprow );
               if( XARsum )
               {
                  kbnext = ktmp - kb; kbnext = MIN( kbnext, nb );
                  ktmp   = PB_Cnumroc( kbnext, k+kb, Ainb1, Anb, mycol, Acol,
                                       npcol );
                  Anq0  -= ktmp;

                  if( ktmp > 0 )
                  {
                     if( Anp0 > 0 )
                     {
                        zgemv_tcomp_macro( 
                         TRANS, &Anp0, &ktmp, negone,
                         &Ald,
                         &ione, one,
                         &XARld,
                         nprow, npcol, myrow, mycol, tcomp
                        );
                     }
                     Asrc = PB_Cindxg2p( k+kb, Aimb1, Amb, Arow, Arow, nprow );
                     hscal_Czgsum2d( 
                         blacsctxt, COLUMN, &ctop, 1, ktmp, XARld, Asrc, mycol,
                         nprow, npcol, myrow, mycol, tcomp, tcomm           
                     );
                     if( myrow != Asrc )
                     {
                        zset_tcomp_macro( 
                            &ktmp, zero, &XARld,
                            nprow, npcol, myrow, mycol, tcomp
                        );
                     }
                  }
                  if( Anp0 > 0 && Anq0 > 0 )
                  {
                     zgemv_tcomp_macro( 
                     TRANS, &Anp0, &Anq0, negone,
                     &Ald,
                     &ione, one,
                     &XARld,
                     nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
               else
               {
                  if( Anp0 > 0 )
                  {
                     zgemv_tcomp_macro( 
                        TRANS, &Anp0, &Anq0, negone,
                        &Ald,
                        &ione, one,
                        &XARld,
                        nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
            }
         }
/*
*  Combine the scattered resulting vector XAR
*/
         if( XARsum && ( Anq > 0 ) )
         {
            hscal_Czgsum2d( 
                blacsctxt, COLUMN, &ctop, 1, Anq, XARld, XARd[RSRC_],
                mycol,
                nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
/*
*  sub( X ) := XAR (if necessary)
*/
         if( XARapbX )
         {
            int N_l = 1;
            int IA_l = 0;
            int JA_l = 0;
            char *CONJUG_l = NOCONJG;
            hscal_PB_Cpaxpby_Z( 
               type, CONJUG_l, N_l, N, one, IA_l, JA_l, XARd, ROW, zero,
               Xi, Xj, Xd, (&Xroc),
               nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
/*
*  Restore BLACS topologies
*/
         (void) PB_Ctop( &ctxt, BCAST,   ROW,    &btop );
         (void) PB_Ctop( &ctxt, COMBINE, COLUMN, &ctop );
      }
      else
      {
/*
*  Save current and enforce ring BLACS topologies
*/
         btop = *PB_Ctop( &ctxt, BCAST,   ROW,    TOP_GET   );
         ctop = *PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_GET   );
         (void)  PB_Ctop( &ctxt, BCAST,   ROW,    TOP_DRING );
         (void)  PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_DRING );
/*
*  Remove next line when BLACS combine operations support ring topologies.
*
*        (void)  PB_Ctop( &ctxt, COMBINE, COLUMN, TOP_DEFAULT );
*/
/*
*  Reuse sub( X ) and/or create vector XAC in process row owning the last row
*  of sub( A )
*/
         {
         char *CONJUG_l = NOCONJG;
         char *ROWCOL_l = ROW;
         int K_l = 1;
         hscal_PB_CInOutV2_Z( 
               type, CONJUG_l, ROWCOL_l, N, N, N-1, Ad0, K_l,
               Xi, Xj, Xd, (&Xroc), XARd,
               (&XARsum), (&XARapbX),
               nprow, npcol, myrow, mycol, tcomp, tcomm           
         );
         }
/*
*  Create vector XAC in process columns spanned by sub( A )
*/
         {
         char * ZEROIT_l = INIT;
         int K_l = 1;
         char *ROWCOL_l = COLUMN;
         hscal_PB_COutV_Z( 
                type, ROWCOL_l, ZEROIT_l, N, N, Ad0, K_l, XACd, 
                &XACsum,
                nprow, npcol, myrow, mycol, tcomp
         );
         }
/*
*  Retrieve local quantities related to Ad0 -> sub( A )
*/
         Aimb1 = Ad0[IMB_ ]; Ainb1 = Ad0[INB_ ];
         Amb   = Ad0[MB_  ]; Anb   = Ad0[NB_  ];
         Arow  = Ad0[RSRC_]; Acol  = Ad0[CSRC_]; Ald = Ad0[LLD_];
         Anp   = PB_Cnumroc( N, 0, Aimb1, Amb, myrow, Arow, nprow );
         Anq   = PB_Cnumroc( N, 0, Ainb1, Anb, mycol, Acol, npcol );
         if( ( Anp > 0 ) && ( Anq > 0 ) )
         {
            ;
         }

         XACld = XACd[LLD_]; XARld = XARd[LLD_];

         for( k = ( ( N - 1 ) / nb ) * nb; k >= 0; k -= nb )
         {
            ktmp = N - k; kb = MIN( ktmp, nb );
/*
*  Solve logical diagonal block, XAR contains the solution scattered in multiple
*  process rows and XAC contains the solution replicated in the process columns.
*/
            Akp = PB_Cnumroc( k,  0, Aimb1, Amb, myrow, Arow, nprow );
            Akq = PB_Cnumroc( k,  0, Ainb1, Anb, mycol, Acol, npcol );
            hscal_PB_Cptrsv_Z( 
                     type, XACsum, (&UploA), (&TranOp), (&DiagA), kb, k, k,
                     Ad0, 1, XARld,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
            );
/*
*  Update: only the part of sub( X ) to be solved at the next step is locally
*  updated and combined, the remaining part of the vector to be solved later
*  is only locally updated.
*/
            if( Akq > 0 )
            {
               Anp0 = PB_Cnumroc( kb, k, Aimb1, Amb, myrow, Arow, nprow );
               if( XARsum )
               {
                  kbprev = MIN( k, nb );
                  ktmp   = PB_Cnumroc( kbprev, k-kbprev, Ainb1, Anb, mycol,
                                       Acol, npcol );
                  Akq   -= ktmp;

                  if( ktmp > 0 )
                  {
                     if( Anp0 > 0 )
                     {
                        zgemv_tcomp_macro( 
                             TRANS, &Anp0, &ktmp, negone,
                             &Ald,
                             &ione, one,
                             &XARld,
                             nprow, npcol, myrow, mycol, tcomp
                        );
                     }
                     Asrc = PB_Cindxg2p( k-1, Aimb1, Amb, Arow, Arow, nprow );
                     hscal_Czgsum2d( 
                         blacsctxt, COLUMN, &ctop, 1, ktmp, XARld, Asrc, mycol,
                         nprow, npcol, myrow, mycol, tcomp, tcomm
                     );
                     if( myrow != Asrc )
                     {
                        zset_tcomp_macro( 
                            &ktmp, zero, &XARld,
                            nprow, npcol, myrow, mycol, tcomp
                        );
                     }
                  }
                  if( Anp0 > 0 && Akq > 0 )
                  {
                     zgemv_tcomp_macro( 
                          TRANS, &Anp0, &Akq, negone,
                          &Ald,
                          &ione,
                          one, &XARld,
                          nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
               else
               {
                  if( Anp0 > 0 )
                  {
                     zgemv_tcomp_macro( 
                          TRANS, &Anp0, &Akq, negone,
                          &Ald,
                          &ione,
                          one, &XARld,
                          nprow, npcol, myrow, mycol, tcomp
                     );
                  }
               }
            }
         }
/*
*  Combine the scattered resulting vector XAR
*/
         if( XARsum && ( Anq > 0 ) )
         {
            hscal_Czgsum2d( 
               blacsctxt, COLUMN, &ctop, 1, Anq, XARld, XARd[RSRC_],
                      mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
/*
*  sub( X ) := XAR (if necessary)
*/
         if( XARapbX )
         {
            int N_l = 1;
            int IA_l = 0;
            int JA_l = 0;
            char *CONJUG_l = NOCONJG;
            hscal_PB_Cpaxpby_Z( 
               type, CONJUG_l, N_l, N, one, IA_l, JA_l, XARd, ROW, zero,
               Xi, Xj, Xd, (&Xroc),
               nprow, npcol, myrow, mycol, tcomp, tcomm
            );
         }
/*
*  Restore BLACS topologies
*/
         (void) PB_Ctop( &ctxt, BCAST,   ROW,    &btop );
         (void) PB_Ctop( &ctxt, COMBINE, COLUMN, &ctop );
      }
   }
/*
*  End of PZTRSV
*/

   Cblacs_gridexit_nocomm(blacsctxt);


   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrsv_tcomp_tcomm(
   int UPLOI,
   int TRANSI,
   int DIAGI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztrsv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                N, 
                IA, JA, DESCA, 
                IX, JX, DESCX, INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrsv_info(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;
   int DIAGI  = *DIAG;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztrsv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                *N, 
                *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pztrsv_info_i(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int UPLOI  = *UPLO;
   int TRANSI = *TRANS;
   int DIAGI  = *DIAG;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pztrsv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, TRANSI, DIAGI,
                *N, 
                *IA, *JA, DESCA, 
                *IX, *JX, DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
