
#ifndef __HSCAL_CLAPACK_CLASET_HH
#define __HSCAL_CLAPACK_CLASET_HH

#include "../../include/hscalapack.h"

int hscal_claset(
    char *uplo, 
    int *m, int *n,
    char *alpha, char *beta, 
    hscalinteger *lda,
    double *tcomp
)
{
/*  -- LAPACK auxiliary routine (version 3.0) --   
       Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,   
       Courant Institute, Argonne National Lab, and Rice University   
       October 31, 1992   


    Purpose   
    =======   

    CLASET initializes a 2-D array A to BETA on the diagonal and   
    ALPHA on the offdiagonals.   

    Arguments   
    =========   

    UPLO    (input) CHARACTER*1   
            Specifies the part of the matrix A to be set.   
            = 'U':      Upper triangular part is set. The lower triangle   
                        is unchanged.   
            = 'L':      Lower triangular part is set. The upper triangle   
                        is unchanged.   
            Otherwise:  All of the matrix A is set.   

    M       (input) INTEGER   
            On entry, M specifies the number of rows of A.   

    N       (input) INTEGER   
            On entry, N specifies the number of columns of A.   

    ALPHA   (input) COMPLEX   
            All the offdiagonal array elements are set to ALPHA.   

    BETA    (input) COMPLEX   
            All the diagonal array elements are set to BETA.   

    LDA     (input) INTEGER   
            The leading dimension of the array A.  LDA >= max(1,M).   

    =====================================================================   


       Parameter adjustments */
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2, i__3;
    /* Local variables */
    hscalinteger i__, j;

    *tcomp = 0.;

    /* Function Body */
    if (hscal_lsame(uplo, "U")) {

/*        Set the diagonal to BETA and the strictly upper triangular   
          part of the array to ALPHA. */
	i__1 = *n;
	for (j = 2; j <= i__1; ++j) {
/* Computing MIN */
	    i__3 = j - 1;
	    i__2 = min(i__3,*m);
          *tcomp += 4*i__2;
	}
	i__1 = min(*n,*m);
      *tcomp += 4*i__1;
    } else if (hscal_lsame(uplo, "L")) {

/*        Set the diagonal to BETA and the strictly lower triangular   
          part of the array to ALPHA. */

	i__1 = min(*m,*n);
	for (j = 1; j <= i__1; ++j) {
	    i__2 = *m;
          *tcomp += 4*(i__2 - j);
	}
	i__1 = min(*n,*m);
      *tcomp += 4*i__1;
    } else {

/*        Set the array to BETA on the diagonal and ALPHA on the   
          offdiagonal. */
      *tcomp += 4*(*n)*(*m);
	i__1 = min(*m,*n);
      *tcomp += 4*i__1;
    }

    return HSCAL_SUCCESS;

/*     End of CLASET */

}

#endif
