
#ifndef __HSCAL_CLAPACK_CTBTRS_HH
#define __HSCAL_CLAPACK_CTBTRS_HH

#include "../../include/hscalapack.h"

int hscal_ctbtrs(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *kd, 
    int *nrhs, 
    int *ldab,
    int *ldb, 
    int *info,
    double *tcomp
)
{
/*  -- LAPACK routine (version 3.0) --   
       Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,   
       Courant Institute, Argonne National Lab, and Rice University   
       September 30, 1994   


    Purpose   
    =======   

    CTBTRS solves a triangular system of the form   

       A * X = B,  A**T * X = B,  or  A**H * X = B,   

    where A is a triangular band matrix of order N, and B is an   
    N-by-NRHS matrix.  A check is made to verify that A is nonsingular.   

    Arguments   
    =========   

    UPLO    (input) CHARACTER*1   
            = 'U':  A is upper triangular;   
            = 'L':  A is lower triangular.   

    TRANS   (input) CHARACTER*1   
            Specifies the form of the system of equations:   
            = 'N':  A * X = B     (No transpose)   
            = 'T':  A**T * X = B  (Transpose)   
            = 'C':  A**H * X = B  (Conjugate transpose)   

    DIAG    (input) CHARACTER*1   
            = 'N':  A is non-unit triangular;   
            = 'U':  A is unit triangular.   

    N       (input) INTEGER   
            The order of the matrix A.  N >= 0.   

    KD      (input) INTEGER   
            The number of superdiagonals or subdiagonals of the   
            triangular band matrix A.  KD >= 0.   

    NRHS    (input) INTEGER   
            The number of right hand sides, i.e., the number of columns   
            of the matrix B.  NRHS >= 0.   

    AB      (input) COMPLEX array, dimension (LDAB,N)   
            The upper or lower triangular band matrix A, stored in the   
            first kd+1 rows of AB.  The j-th column of A is stored   
            in the j-th column of the array AB as follows:   
            if UPLO = 'U', AB(kd+1+i-j,j) = A(i,j) for max(1,j-kd)<=i<=j;   
            if UPLO = 'L', AB(1+i-j,j)    = A(i,j) for j<=i<=min(n,j+kd).   
            If DIAG = 'U', the diagonal elements of A are not referenced   
            and are assumed to be 1.   

    LDAB    (input) INTEGER   
            The leading dimension of the array AB.  LDAB >= KD+1.   

    B       (input/output) COMPLEX array, dimension (LDB,NRHS)   
            On entry, the right hand side matrix B.   
            On exit, if INFO = 0, the solution matrix X.   

    LDB     (input) INTEGER   
            The leading dimension of the array B.  LDB >= max(1,N).   

    INFO    (output) INTEGER   
            = 0:  successful exit   
            < 0:  if INFO = -i, the i-th argument had an illegal value   
            > 0:  if INFO = i, the i-th diagonal element of A is zero,   
                  indicating that the matrix is singular and the   
                  solutions X have not been computed.   

    =====================================================================   


       Test the input parameters.   

       Parameter adjustments */
    /* Table of constant values */
    hscalinteger c__1 = 1;
    
    /* System generated locals */
    hscalinteger ab_dim1, ab_offset, b_dim1, b_offset, i__1, i__2;
    /* Local variables */
    hscalinteger j;
    hscallogical upper;
    hscallogical nounit;

    /* Function Body */
    *tcomp = 0.;
    *info = 0;
    nounit = hscal_lsame(diag, "N");
    upper = hscal_lsame(uplo, "U");
    if (! upper && ! hscal_lsame(uplo, "L")) {
	*info = -1;
    } else if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, 
	    "T") && ! hscal_lsame(trans, "C")) {
	*info = -2;
    } else if (! nounit && ! hscal_lsame(diag, "U")) {
	*info = -3;
    } else if (*n < 0) {
	*info = -4;
    } else if (*kd < 0) {
	*info = -5;
    } else if (*nrhs < 0) {
	*info = -6;
    } else if (*ldab < *kd + 1) {
	*info = -8;
    } else if (*ldb < max(1,*n)) {
	*info = -10;
    }
    if (*info != 0) {
	i__1 = -(*info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible */

    if (*n == 0) {
	return HSCAL_SUCCESS;
    }

/*     Check for singularity. */

    if (nounit) {
	if (upper) {
	    i__1 = *n;
          *tcomp += 2*i__1;
	} else {
	    i__1 = *n;
          *tcomp += 2*i__1;
	}
    }
    *info = 0;

/*     Solve A * X = B,  A**T * X = B,  or  A**H * X = B. */

    {
    double t;
    i__1 = *nrhs;
    {
	hscal_ctbsv(
           uplo, trans, diag, 
           n, kd, ldab, &c__1,
           &t
      );
    }
    *tcomp += i__1*t;
    }

    return HSCAL_SUCCESS;

/*     End of CTBTRS */

}

#endif
