
#ifndef __HSCAL_CLAPACK_DDBTF2_HH
#define __HSCAL_CLAPACK_DDBTF2_HH

#include "../../include/hscalapack.h"

int hscal_ddbtf2(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
)
{
    hscalinteger c__1 = 1;
    hscaldoublereal c_b4 = -1.;

    /* System generated locals */
    hscalinteger ab_dim1, ab_offset, i__1, i__2, i__3, i__4;
    hscaldoublereal d__1;

    /* Local variables */
    hscalinteger j, km, jp, ju, kv;

/*     Modified by Andrew J. Cleary in November, 96 from: */
/*  -- LAPACK auxiliary routine (preliminary version) -- */
/*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd., */
/*     Courant Institute, Argonne National Lab, and Rice University */
/*     August 6, 1991 */


/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  Ddbtrf computes an LU factorization of a real m-by-n band matrix A */
/*  without using partial pivoting with row interchanges. */

/*  This is the unblocked version of the algorithm, calling Level 2 BLAS. */

/*  Arguments */
/*  ========= */

/*  M       (input) INTEGER */
/*          The number of rows of the matrix A.  M >= 0. */

/*  N       (input) INTEGER */
/*          The number of columns of the matrix A.  N >= 0. */

/*  KL      (input) INTEGER */
/*          The number of subdiagonals within the band of A.  KL >= 0. */

/*  KU      (input) INTEGER */
/*          The number of superdiagonals within the band of A.  KU >= 0. */

/*  AB      (input/output) DOUBLE PRECISION   array, dimension (LDAB,N) */
/*          On entry, the matrix A in band storage, in rows KL+1 to */
/*          2*KL+KU+1; rows 1 to KL of the array need not be set. */
/*          The j-th column of A is stored in the j-th column of the */
/*          array AB as follows: */
/*          AB(kl+ku+1+i-j,j) = A(i,j) for max(1,j-ku)<=i<=min(m,j+kl) */

/*          On exit, details of the factorization: U is stored as an */
/*          upper triangular band matrix with KL+KU superdiagonals in */
/*          rows 1 to KL+KU+1, and the multipliers used during the */
/*          factorization are stored in rows KL+KU+2 to 2*KL+KU+1. */
/*          See below for further details. */

/*  LDAB    (input) INTEGER */
/*          The leading dimension of the array AB.  LDAB >= 2*KL+KU+1. */

/*  INFO    (output) INTEGER */
/*          = 0: successful exit */
/*          < 0: if INFO = -i, the i-th argument had an illegal value */
/*          > 0: if INFO = +i, U(i,i) is exactly zero. The factorization */
/*               has been completed, but the factor U is exactly */
/*               singular, and division by zero will occur if it is used */
/*               to solve a system of equations. */

/*  Further Details */
/*  =============== */

/*  The band storage scheme is illustrated by the following example, when */
/*  M = N = 6, KL = 2, KU = 1: */

/*  On entry:                       On exit: */

/*      *   a12  a23  a34  a45  a56      *   u12  u23  u34  u45  u56 */
/*     a11  a22  a33  a44  a55  a66     u11  u22  u33  u44  u55  u66 */
/*     a21  a32  a43  a54  a65   *      m21  m32  m43  m54  m65   * */
/*     a31  a42  a53  a64   *    *      m31  m42  m53  m64   *    * */

/*  Array elements marked * are not used by the routine; elements marked */
/*  + need not be set on entry, but are required by the routine to store */
/*  elements of U, because of fill-in resulting from the row */
/*  interchanges. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     KV is the number of superdiagonals in the factor U, allowing for */
/*     fill-in. */

    *tcomp = 0.;

    /* Function Body */
    kv = *ku;

/*     Test the input parameters. */

    *info = 0;
/* ECA  IF( M.LT.0 ) THEN */
/* ECA     INFO = -1 */
/* ECA  ELSE IF( N.LT.0 ) THEN */
/* ECA     INFO = -2 */
/* ECA  ELSE IF( KL.LT.0 ) THEN */
/* ECA     INFO = -3 */
/* ECA  ELSE IF( KU.LT.0 ) THEN */
/* ECA     INFO = -4 */
/* ECA  ELSE IF( LDAB.LT.KL+KV+1 ) THEN */
/* ECA     INFO = -6 */
/* ECA  END IF */
/* ECA  IF( INFO.NE.0 ) THEN */
/* ECA     CALL XERBLA( 'DDBTF2', -INFO ) */
/* ECA     RETURN */
/* ECA  END IF */

/*     Quick return if possible */

    if (*m == 0 || *n == 0) {
	return HSCAL_SUCCESS;
    }

/*     Gaussian elimination without partial pivoting */

/*     JU is the index of the last column affected by the current stage */
/*     of the factorization. */

    ju = 1;

    i__1 = min(*m,*n);
    for (j = 1; j <= i__1; ++j) {

/*        Test for singularity. KM is the number of */
/*        subdiagonal elements in the current column. */

/* Computing MIN */
	i__2 = *kl, i__3 = *m - j;
	km = min(i__2,i__3);
	jp = 1;
	{
/* Computing MAX */
/* Computing MIN */
	    i__4 = j + *ku;
	    i__2 = ju, i__3 = min(i__4,*n);
	    ju = max(i__2,i__3);

	    if (km > 0) {

/*              Compute multipliers. */

		d__1 = .5;
            {
            double t;
		hscal_dscal(&km, (char*)&d__1, &c__1, &t);
            *tcomp += t;
            }

/*              Update trailing submatrix within the band. */

		if (ju > j) {
		    i__2 = ju - j;
		    i__3 = *ldab - 1;
		    i__4 = *ldab - 1;
                {
                double t;
		    hscal_dger(&km, &i__2, (char*)&c_b4, &c__1, &i__3, &i__4, &t);
                *tcomp += t;
                }
		}
	    }
	}
/* L40: */
    }
    return HSCAL_SUCCESS;

/*     End of DDBTF2 */

}

#endif
