
#ifndef __HSCAL_CLAPACK_DLARFG_HH
#define __HSCAL_CLAPACK_DLARFG_HH

#include "../../include/hscalapack.h"

int hscal_dlarfg(
    int *n, hscaldoublereal *alpha, int *incx,
    double *tcomp
)
{
/*  -- LAPACK auxiliary routine (version 3.0) --   
       Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,   
       Courant Institute, Argonne National Lab, and Rice University   
       September 30, 1994   


    Purpose   
    =======   

    DLARFG generates a real elementary reflector H of order n, such   
    that   

          H * ( alpha ) = ( beta ),   H' * H = I.   
              (   x   )   (   0  )   

    where alpha and beta are scalars, and x is an (n-1)-element real   
    vector. H is represented in the form   

          H = I - tau * ( 1 ) * ( 1 v' ) ,   
                        ( v )   

    where tau is a real scalar and v is a real (n-1)-element   
    vector.   

    If the elements of x are all zero, then tau = 0 and H is taken to be   
    the unit matrix.   

    Otherwise  1 <= tau <= 2.   

    Arguments   
    =========   

    N       (input) INTEGER   
            The order of the elementary reflector.   

    ALPHA   (input/output) DOUBLE PRECISION   
            On entry, the value alpha.   
            On exit, it is overwritten with the value beta.   

    X       (input/output) DOUBLE PRECISION array, dimension   
                           (1+(N-2)*abs(INCX))   
            On entry, the vector x.   
            On exit, it is overwritten with the vector v.   

    INCX    (input) INTEGER   
            The increment between elements of X. INCX > 0.   

    TAU     (output) DOUBLE PRECISION   
            The value tau.   

    =====================================================================   


       Parameter adjustments */
    /* System generated locals */
    hscalinteger i__1;

    *tcomp = 0.;

    /* Function Body */
    if (*n <= 1) {
	return HSCAL_SUCCESS;
    }

    i__1 = *n - 1;
    {
    double t;
    hscal_dnrm2(&i__1, incx,
         &t
    );
    *tcomp += t;
    }

    {
	{
	    i__1 = *n - 1;
            {
            double t;
            hscaldoublereal rsafmn = 0.5;
	    hscal_dscal(
                 &i__1, (char*)&rsafmn, incx,
                 &t
            );
            *tcomp += t;
            }

/*           New BETA is at most 1, at least SAFMIN */

	    i__1 = *n - 1;
            {
            double t;
	    hscal_dnrm2(
                 &i__1, incx,
                 &t
            );
            *tcomp += t;
            }

	    i__1 = *n - 1;
            {
            double t;
            hscaldoublereal d__1 = 0.5;
	    hscal_dscal(
                 &i__1, (char*)&d__1, incx,
                 &t
            );
            *tcomp += t;
            }
	}
    }

    return HSCAL_SUCCESS;

/*     End of DLARFG */

} /* dlarfg_ */

#endif

