
#ifndef __HSCAL_CLAPACK_DTBSV_HH
#define __HSCAL_CLAPACK_DTBSV_HH

#include "../../include/hscalapack.h"

int hscal_dtbsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *k, 
    int *lda,
    int *incx, 
    double *tcomp
)
{
    /* System generated locals */
    hscalinteger a_dim1, a_offset, i__1, i__2, i__3, i__4;
    /* Local variables */
    hscalinteger info;
    hscaldoublereal temp;
    hscalinteger i__, j, l;
    hscalinteger kplus1, ix, jx, kx;
    hscallogical nounit;

/*  Purpose   
    =======   
    DTBSV  solves one of the systems of equations   
       A*x = b,   or   A'*x = b,   
    where b and x are n element vectors and A is an n by n unit, or   
    non-unit, upper or lower triangular band matrix, with ( k + 1 )   
    diagonals.   
    No test for singularity or near-singularity is included in this   
    routine. Such tests must be performed before calling this routine.   
    Parameters   
    ==========   
    UPLO   - CHARACTER*1.   
             On entry, UPLO specifies whether the matrix is an upper or   
             lower triangular matrix as follows:   
                UPLO = 'U' or 'u'   A is an upper triangular matrix.   
                UPLO = 'L' or 'l'   A is a lower triangular matrix.   
             Unchanged on exit.   
    TRANS  - CHARACTER*1.   
             On entry, TRANS specifies the equations to be solved as   
             follows:   
                TRANS = 'N' or 'n'   A*x = b.   
                TRANS = 'T' or 't'   A'*x = b.   
                TRANS = 'C' or 'c'   A'*x = b.   
             Unchanged on exit.   
    DIAG   - CHARACTER*1.   
             On entry, DIAG specifies whether or not A is unit   
             triangular as follows:   
                DIAG = 'U' or 'u'   A is assumed to be unit triangular.   
                DIAG = 'N' or 'n'   A is not assumed to be unit   
                                    triangular.   
             Unchanged on exit.   
    N      - INTEGER.   
             On entry, N specifies the order of the matrix A.   
             N must be at least zero.   
             Unchanged on exit.   
    K      - INTEGER.   
             On entry with UPLO = 'U' or 'u', K specifies the number of   
             super-diagonals of the matrix A.   
             On entry with UPLO = 'L' or 'l', K specifies the number of   
             sub-diagonals of the matrix A.   
             K must satisfy  0 .le. K.   
             Unchanged on exit.   
    A      - DOUBLE PRECISION array of DIMENSION ( LDA, n ).   
             Before entry with UPLO = 'U' or 'u', the leading ( k + 1 )   
             by n part of the array A must contain the upper triangular   
             band part of the matrix of coefficients, supplied column by   
             column, with the leading diagonal of the matrix in row   
             ( k + 1 ) of the array, the first super-diagonal starting at   
             position 2 in row k, and so on. The top left k by k triangle   
             of the array A is not referenced.   
             The following program segment will transfer an upper   
             triangular band matrix from conventional full matrix storage   
             to band storage:   
                   DO 20, J = 1, N   
                      M = K + 1 - J   
                      DO 10, I = MAX( 1, J - K ), J   
                         A( M + I, J ) = matrix( I, J )   
                10    CONTINUE   
                20 CONTINUE   
             Before entry with UPLO = 'L' or 'l', the leading ( k + 1 )   
             by n part of the array A must contain the lower triangular   
             band part of the matrix of coefficients, supplied column by   
             column, with the leading diagonal of the matrix in row 1 of   
             the array, the first sub-diagonal starting at position 1 in   
             row 2, and so on. The bottom right k by k triangle of the   
             array A is not referenced.   
             The following program segment will transfer a lower   
             triangular band matrix from conventional full matrix storage   
             to band storage:   
                   DO 20, J = 1, N   
                      M = 1 - J   
                      DO 10, I = J, MIN( N, J + K )   
                         A( M + I, J ) = matrix( I, J )   
                10    CONTINUE   
                20 CONTINUE   
             Note that when DIAG = 'U' or 'u' the elements of the array A   
             corresponding to the diagonal elements of the matrix are not   
             referenced, but are assumed to be unity.   
             Unchanged on exit.   
    LDA    - INTEGER.   
             On entry, LDA specifies the first dimension of A as declared   
             in the calling (sub) program. LDA must be at least   
             ( k + 1 ).   
             Unchanged on exit.   
    X      - DOUBLE PRECISION array of dimension at least   
             ( 1 + ( n - 1 )*abs( INCX ) ).   
             Before entry, the incremented array X must contain the n   
             element right-hand side vector b. On exit, X is overwritten   
             with the solution vector x.   
    INCX   - INTEGER.   
             On entry, INCX specifies the increment for the elements of   
             X. INCX must not be zero.   
             Unchanged on exit.   
    Level 2 Blas routine.   
    -- Written on 22-October-1986.   
       Jack Dongarra, Argonne National Lab.   
       Jeremy Du Croz, Nag Central Office.   
       Sven Hammarling, Nag Central Office.   
       Richard Hanson, Sandia National Labs.   
       Test the input parameters.   
       Parameter adjustments */
    /* Function Body */
    info = 0;
    *tcomp = 0.;
    if (! hscal_lsame(uplo, "U") && ! hscal_lsame(uplo, "L")) {
	info = 1;
    } else if (! hscal_lsame(trans, "N") && ! hscal_lsame(trans, 
	    "T") && ! hscal_lsame(trans, "C")) {
	info = 2;
    } else if (! hscal_lsame(diag, "U") && ! hscal_lsame(diag, 
	    "N")) {
	info = 3;
    } else if (*n < 0) {
	info = 4;
    } else if (*k < 0) {
	info = 5;
    } else if (*lda < *k + 1) {
	info = 7;
    } else if (*incx == 0) {
	info = 9;
    }
    if (info != 0) {
	return HSCAL_SUCCESS;
    }
/*     Quick return if possible. */
    if (*n == 0) {
	return HSCAL_SUCCESS;
    }
    nounit = hscal_lsame(diag, "N");

/*     Start the operations. In this version the elements of A are   
       accessed by sequentially with one pass through A. */
    if (hscal_lsame(trans, "N")) {
/*        Form  x := inv( A )*x. */
	if (hscal_lsame(uplo, "U")) {
	    if (*incx == 1) {
		for (j = *n; j >= 1; --j) {
		    {
			if (nounit) {
                      *tcomp += 3;
			}
/* Computing MAX */
			i__2 = 1, i__3 = j - *k;
			i__1 = max(i__2,i__3);
                  *tcomp += 4*(i__1 - j + 2);
		    }
/* L20: */
		}
	    } else {
		for (j = *n; j >= 1; --j) {
		    {
			if (nounit) {
                      *tcomp += 3;
			}
/* Computing MAX */
			i__2 = 1, i__3 = j - *k;
			i__1 = max(i__2,i__3);
                  *tcomp += 5*(i__1 - j + 2);
		    }
/* L40: */
		}
	    }
	} else {
	    if (*incx == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    {
			if (nounit) {
                      *tcomp += 3;
			}
/* Computing MIN */
			i__3 = *n, i__4 = j + *k;
			i__2 = min(i__3,i__4);
                  *tcomp += 4*(i__2 - j);
		    }
/* L60: */
		}
	    } else {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
		    {
			if (nounit) {
                      *tcomp += 3;
			}
/* Computing MIN */
			i__3 = *n, i__4 = j + *k;
			i__2 = min(i__3,i__4);
                  *tcomp += 5*(i__2 - j);
		    }
/* L80: */
		}
	    }
	}
    } else {
/*        Form  x := inv( A')*x. */
	if (hscal_lsame(uplo, "U")) {
	    if (*incx == 1) {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
/* Computing MAX */
		    i__2 = 1, i__3 = j - *k;
		    i__4 = j - 1;
                *tcomp += 4*(i__4 - max(i__2,i__3) + 1);
		    if (nounit) {
                   *tcomp += 3;
		    }
		}
	    } else {
		i__1 = *n;
		for (j = 1; j <= i__1; ++j) {
/* Computing MAX */
		    i__4 = 1, i__2 = j - *k;
		    i__3 = j - 1;
                *tcomp += 5*(i__3 - max(i__4,i__2) + 1);
		    if (nounit) {
                   *tcomp += 3;
		    }
                *tcomp += 2;
		}
	    }
	} else {
	    if (*incx == 1) {
		for (j = *n; j >= 1; --j) {
/* Computing MIN */
		    i__1 = *n, i__3 = j + *k;
		    i__4 = j + 1;
                *tcomp += 4*(i__4 - min(i__1,i__3) + 1);
		    if (nounit) {
                   *tcomp += 3;
		    }
		}
	    } else {
		for (j = *n; j >= 1; --j) {
/* Computing MIN */
		    i__4 = *n, i__1 = j + *k;
		    i__3 = j + 1;
                *tcomp += 5*(i__3 - min(i__4,i__1) + 1);
		    if (nounit) {
                   *tcomp += 3;
		    }
                *tcomp += 2;
		}
	    }
	}
    }
    return HSCAL_SUCCESS;
/*     End of DTBSV . */
}

#endif
