
#ifndef __HSCAL_CLAPACK_DDBTRF_HH
#define __HSCAL_CLAPACK_DDBTRF_HH

#include "../../include/hscalapack.h"

int hscal_sdbtrf(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
)
{
    hscalinteger c__1 = 1;
    hscalreal c_b13 = -1.f;
    hscalreal c_b21 = 1.f;
    hscalinteger c__65 = 65;

    /* System generated locals */
    hscalinteger ab_dim1, ab_offset, i__1, i__2, i__3, i__4, i__5, i__6;
    hscalreal r__1;

    /* Local variables */
    hscalinteger i__, j, i2, i3, j2, j3, jb, nb, ii, jj, jm, km, jp, ju, kv,
	     nw;

/*     Written by Andrew J. Cleary, University of Tennessee. */
/*     August, 1996. */
/*     Modified from SGBTRF: */
/*  -- LAPACK routine (preliminary version) -- */
/*     Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd., */
/*     Courant Institute, Argonne National Lab, and Rice University */
/*     August 6, 1991 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  Sdbtrf computes an LU factorization of a real m-by-n band matrix A */
/*  without using partial pivoting or row interchanges. */

/*  This is the blocked version of the algorithm, calling Level 3 BLAS. */

/*  Arguments */
/*  ========= */

/*  M       (input) INTEGER */
/*          The number of rows of the matrix A.  M >= 0. */

/*  N       (input) INTEGER */
/*          The number of columns of the matrix A.  N >= 0. */

/*  KL      (input) INTEGER */
/*          The number of subdiagonals within the band of A.  KL >= 0. */

/*  KU      (input) INTEGER */
/*          The number of superdiagonals within the band of A.  KU >= 0. */

/*  AB      (input/output) REAL array, dimension (LDAB,N) */
/*          On entry, the matrix A in band storage, in rows KL+1 to */
/*          2*KL+KU+1; rows 1 to KL of the array need not be set. */
/*          The j-th column of A is stored in the j-th column of the */
/*          array AB as follows: */
/*          AB(kl+ku+1+i-j,j) = A(i,j) for max(1,j-ku)<=i<=min(m,j+kl) */

/*          On exit, details of the factorization: U is stored as an */
/*          upper triangular band matrix with KL+KU superdiagonals in */
/*          rows 1 to KL+KU+1, and the multipliers used during the */
/*          factorization are stored in rows KL+KU+2 to 2*KL+KU+1. */
/*          See below for further details. */

/*  LDAB    (input) INTEGER */
/*          The leading dimension of the array AB.  LDAB >= 2*KL+KU+1. */

/*  INFO    (output) INTEGER */
/*          = 0: successful exit */
/*          < 0: if INFO = -i, the i-th argument had an illegal value */
/*          > 0: if INFO = +i, U(i,i) is exactly zero. The factorization */
/*               has been completed, but the factor U is exactly */
/*               singular, and division by zero will occur if it is used */
/*               to solve a system of equations. */

/*  Further Details */
/*  =============== */

/*  The band storage scheme is illustrated by the following example, when */
/*  M = N = 6, KL = 2, KU = 1: */

/*  On entry:                       On exit: */

/*      *   a12  a23  a34  a45  a56      *   u12  u23  u34  u45  u56 */
/*     a11  a22  a33  a44  a55  a66     u11  u22  u33  u44  u55  u66 */
/*     a21  a32  a43  a54  a65   *      m21  m32  m43  m54  m65   * */
/*     a31  a42  a53  a64   *    *      m31  m42  m53  m64   *    * */

/*  Array elements marked * are not used by the routine. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. Local Arrays .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     KV is the number of superdiagonals in the factor U */

    *tcomp = 0.;

    /* Function Body */
    kv = *ku;

/*     Test the input parameters. */

    *info = 0;
    if (*m < 0) {
	*info = -1;
    } else if (*n < 0) {
	*info = -2;
    } else if (*kl < 0) {
	*info = -3;
    } else if (*ku < 0) {
	*info = -4;
    } else /* if(complicated condition) */ {
/* Computing MIN */
/* Computing MIN */
	i__2 = *kl + kv + 1;
	i__1 = min(i__2,*m);
	if (*ldab < min(i__1,*n)) {
	    *info = -6;
	}
    }
    if (*info != 0) {
	i__1 = -(*info);
	return HSCAL_SUCCESS;
    }

/*     Quick return if possible */

    if (*m == 0 || *n == 0) {
	return HSCAL_SUCCESS;
    }

/*     Determine the block size for this environment */

    nb = ilaenv_(&c__1, "SDBTRF", " ", m, n, kl, ku);

/*     The block size must not exceed the limit set by the size of the */
/*     local arrays WORK13 and WORK31. */

    nb = min(nb,64);

    if (nb <= 1 || nb > *kl) {

/*        Use unblocked code */

      double t;
	hscal_sdbtf2(m, n, kl, ku, ldab, info, &t);
      *tcomp += t;
    } else {

/*        Use blocked code */

/*        Zero the superdiagonal elements of the work array WORK13 */

	i__1 = nb;
      *tcomp += 4*i__1*(i__1 - 1)/2;

/*        Zero the subdiagonal elements of the work array WORK31 */

	i__1 = nb;
      *tcomp += 4*i__1*(i__1 - 1)/2;

/*        JU is the index of the last column affected by the current */
/*        stage of the factorization */

	ju = 1;

	i__1 = min(*m,*n);
	i__2 = nb;
	for (j = 1; i__2 < 0 ? j >= i__1 : j <= i__1; j += i__2) {
/* Computing MIN */
	    i__3 = nb, i__4 = min(*m,*n) - j + 1;
	    jb = min(i__3,i__4);

/*           The active part of the matrix is partitioned */

/*              A11   A12   A13 */
/*              A21   A22   A23 */
/*              A31   A32   A33 */

/*           Here A11, A21 and A31 denote the current block of JB columns */
/*           which is about to be factorized. The number of rows in the */
/*           partitioning are JB, I2, I3 respectively, and the numbers */
/*           of columns are JB, J2, J3. The superdiagonal elements of A13 */
/*           and the subdiagonal elements of A31 lie outside the band. */

/* Computing MIN */
	    i__3 = *kl - jb, i__4 = *m - j - jb + 1;
	    i2 = min(i__3,i__4);
/* Computing MIN */
	    i__3 = jb, i__4 = *m - j - *kl + 1;
	    i3 = min(i__3,i__4);

/*           J2 and J3 are computed after JU has been updated. */

/*           Factorize the current block of JB columns */

	    i__3 = j + jb - 1;
	    for (jj = j; jj <= i__3; ++jj) {

/*              Find pivot and test for singularity. KM is the number of */
/*              subdiagonal elements in the current column. */

/* Computing MIN */
		i__4 = *kl, i__5 = *m - jj;
		km = min(i__4,i__5);
		jp = 1;
		{
/* Computing MAX */
/* Computing MIN */
		    i__6 = jj + *ku + jp - 1;
		    i__4 = ju, i__5 = min(i__6,*n);
		    ju = max(i__4,i__5);

/*                 Compute multipliers */

		    r__1 = 0.5f;
                {
                double t;
		    hscal_sscal(&km, (char*)&r__1, &c__1, &t);
                *tcomp += t;
                }

/*                 Update trailing submatrix within the band and within */
/*                 the current block. JM is the index of the last column */
/*                 which needs to be updated. */

/* Computing MIN */
		    i__4 = ju, i__5 = j + jb - 1;
		    jm = min(i__4,i__5);
		    if (jm > jj) {
			i__4 = jm - jj;
			i__5 = *ldab - 1;
			i__6 = *ldab - 1;
                  {
                  double t;
			hscal_sger(&km, &i__4, (char*)&c_b13,
				&c__1, &i__5, &i__6, &t);
                  *tcomp += t;
                  }
		    }
		}

/*              Copy current column of A31 into the work array WORK31 */

/* Computing MIN */
		i__4 = jj - j + 1;
		nw = min(i__4,i3);
		if (nw > 0) {
                double t;
		    hscal_scopy(&nw, &c__1, &c__1, &t);
                *tcomp += t;
		}
/* L80: */
	    }
	    if (j + jb <= *n) {

/*              Apply the row interchanges to the other blocks. */

/* Computing MIN */
		i__3 = ju - j + 1;
		j2 = min(i__3,kv) - jb;
/* Computing MAX */
		i__3 = 0, i__4 = ju - j - kv + 1;
		j3 = max(i__3,i__4);

/*              Update the relevant part of the trailing submatrix */

		if (j2 > 0) {

/*                 Update A12 */

		    i__3 = *ldab - 1;
		    i__4 = *ldab - 1;
                {
                double t;
		    hscal_strsm("Left", "Lower", "No transpose", "Unit", &jb, &j2, 
			    (char*)&c_b21, &i__3, &i__4, &t);
                *tcomp += t;
                }

		    if (i2 > 0) {

/*                    Update A22 */

			i__3 = *ldab - 1;
			i__4 = *ldab - 1;
			i__5 = *ldab - 1;
                  {
                  double t;
			hscal_sgemm("No transpose", "No transpose", &i2, &j2, &jb, 
				(char*)&c_b13, &i__3,
				&i__4,
				(char*)&c_b21, &i__5, &t);
                  *tcomp += t;
                  }
		    }

		    if (i3 > 0) {

/*                    Update A32 */

			i__3 = *ldab - 1;
			i__4 = *ldab - 1;
                  {
                  double t;
			hscal_sgemm("No transpose", "No transpose", &i3, &j2, &jb, 
				(char*)&c_b13, &c__65, &i__3, (char*)&c_b21, &i__4, &t);
                  *tcomp += t;
                  }
		    }
		}

		if (j3 > 0) {

/*                 Copy the lower triangle of A13 into the work array */
/*                 WORK13 */

		    i__3 = j3;
                *tcomp += 8*(i__3*jb - (i__3*(i__3-1)/2));

/*                 Update A13 in the work array */

		    i__3 = *ldab - 1;
                {
                double t;
		    hscal_strsm("Left", "Lower", "No transpose", "Unit", &jb, &j3, 
			    (char*)&c_b21, &i__3,
			    &c__65, &t);
                *tcomp += t;
                }

		    if (i2 > 0) {

/*                    Update A23 */

			i__3 = *ldab - 1;
			i__4 = *ldab - 1;
                  {
                  double t;
			hscal_sgemm("No transpose", "No transpose", &i2, &j3, &jb, 
				(char*)&c_b13, &i__3,
				&c__65, (char*)&c_b21, &i__4, &t);
                  *tcomp += t;
                  }
		    }

		    if (i3 > 0) {

/*                    Update A33 */

			i__3 = *ldab - 1;
                  {
                  double t;
			hscal_sgemm("No transpose", "No transpose", &i3, &j3, &jb, 
				(char*)&c_b13, &c__65, &c__65, (char*)&c_b21, &i__3, &t);
                  *tcomp += t;
                  }
		    }

/*                 Copy the lower triangle of A13 back into place */

		    i__3 = j3;
                *tcomp += 8*(i__3*jb - (i__3*(i__3-1)/2));
		}
	    } else {
	    }

/*           copy the upper triangle of A31 back into place */

	    i__3 = j;
	    for (jj = j + jb - 1; jj >= i__3; --jj) {

/*              Copy the current column of A31 back into place */

/* Computing MIN */
		i__4 = i3, i__5 = jj - j + 1;
		nw = min(i__4,i__5);
		if (nw > 0) {
                double t;
		    hscal_scopy(&nw, &c__1, &c__1, &t);
                *tcomp += t;
		}
/* L170: */
	    }
/* L180: */
	}
    }

    return HSCAL_SUCCESS;

/*     End of SDBTRF */

}

#endif
