
#ifndef __HSCAL_CLAPACK_ZLARFG_HH
#define __HSCAL_CLAPACK_ZLARFG_HH

#include "../../include/hscalapack.h"

int hscal_zlarfg(
    int *n, hscaldoublecomplex *alpha, int *incx,
    double *tcomp
)
{
/*  -- LAPACK auxiliary routine (version 3.0) --   
       Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,   
       Courant Institute, Argonne National Lab, and Rice University   
       September 30, 1994   


    Purpose   
    =======   

    ZLARFG generates a complex elementary reflector H of order n, such   
    that   

          H' * ( alpha ) = ( beta ),   H' * H = I.   
               (   x   )   (   0  )   

    where alpha and beta are scalars, with beta real, and x is an   
    (n-1)-element complex vector. H is represented in the form   

          H = I - tau * ( 1 ) * ( 1 v' ) ,   
                        ( v )   

    where tau is a complex scalar and v is a complex (n-1)-element   
    vector. Note that H is not hermitian.   

    If the elements of x are all zero and alpha is real, then tau = 0   
    and H is taken to be the unit matrix.   

    Otherwise  1 <= real(tau) <= 2  and  abs(tau-1) <= 1 .   

    Arguments   
    =========   

    N       (input) INTEGER   
            The order of the elementary reflector.   

    ALPHA   (input/output) COMPLEX*16   
            On entry, the value alpha.   
            On exit, it is overwritten with the value beta.   

    X       (input/output) COMPLEX*16 array, dimension   
                           (1+(N-2)*abs(INCX))   
            On entry, the vector x.   
            On exit, it is overwritten with the vector v.   

    INCX    (input) INTEGER   
            The increment between elements of X. INCX > 0.   

    TAU     (output) COMPLEX*16   
            The value tau.   

    =====================================================================   


       Parameter adjustments */
    
    /* System generated locals */
    hscalinteger i__1;

    *tcomp = 0.;

    /* Function Body */
    if (*n <= 0) {
	return HSCAL_SUCCESS;
    }

    i__1 = *n - 1;
    {
    double t;
    hscal_dznrm2(
         &i__1, incx,
         &t
    );
    *tcomp += t;
    }

    {
	 {

	    i__1 = *n - 1;
            {
            double t;
            hscaldoublereal rsafmn = 0.5;
	    hscal_zdscal(
                 &i__1, (char*)&rsafmn, incx,
                 &t
            );
            *tcomp += t;
            }

	    i__1 = *n - 1;
            {
            double t;
	    hscal_dznrm2(
                 &i__1, incx,
                 &t
            );
            *tcomp += t;
            }

	    i__1 = *n - 1;
            {
            double t;
	    hscal_zscal(
                 &i__1, (char*)alpha, incx,
                 &t
            );
            *tcomp += t;
            }
	}
    }

    return HSCAL_SUCCESS;

/*     End of ZLARFG */

} /* zlarfg_ */

#endif

