
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <time.h>

#include "../include/hscalapack.h"

int hscal_debug_flag = 0;

/*---------------------------------------------------------------------------*/

int hscal_set_debug(
    int debug_level
)
{
    hscal_debug_flag = debug_level;

    return HSCAL_SUCCESS;
}

/*---------------------------------------------------------------------------*/

int hscal_get_debug()
{
    return hscal_debug_flag;
}

/*---------------------------------------------------------------------------*/

void HSCAL_DEBUG(
     const char *file, const char *function, int line,
     const char *form, ...
)
{
     if (hscal_debug_flag & HSCAL_LOG_DEBUG)
     {
        va_list argptr;

        char dtstr[HSCAL_DATE_STR_LENGTH];
        char cline[HSCAL_LOG_LENGTH];

        struct tm *tp;

        time_t curtime;
        time(&curtime);
        tp = localtime(&curtime);
        strftime(dtstr, sizeof(dtstr), "%I:%M:%S %p %B %d, %Y", tp);

        va_start(argptr, form);
        vsprintf(cline, form, argptr);
        va_end(argptr);

        printf("[HSCALAPACK %s %s:%s:%d]===> ", dtstr, file, function, line);

        printf(cline); printf("\n");

        HMPI_Debug(1);
     }
}

/*---------------------------------------------------------------------------*/

void HSCAL_TODO(
     const char *file, const char *function, int line,
     const char *form, ...
)
{
     if (hscal_debug_flag & HSCAL_LOG_TODO)
     {
        va_list argptr;

        char dtstr[HSCAL_DATE_STR_LENGTH];
        char cline[HSCAL_LOG_LENGTH];

        struct tm *tp;

        time_t curtime;
        time(&curtime);
        tp = localtime(&curtime);
        strftime(dtstr, sizeof(dtstr), "%I:%M:%S %p %B %d, %Y", tp);

        va_start(argptr, form);
        vsprintf(cline, form, argptr);
        va_end(argptr);

        printf("[HSCALAPACK TODO %s %s:%s:%d]===> ", dtstr, file, function, line);

        printf(cline); printf("\n");
     }
}

/*---------------------------------------------------------------------------*/

void HSCAL_TIMINGS(
     const char *form, ...
)
{
     if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
     {
        va_list argptr;

        char dtstr[HSCAL_DATE_STR_LENGTH];
        char cline[HSCAL_LOG_LENGTH];

        struct tm *tp;

        time_t curtime;
        time(&curtime);
        tp = localtime(&curtime);
        strftime(dtstr, sizeof(dtstr), "%I:%M:%S %p %B %d, %Y", tp);

        va_start(argptr, form);
        vsprintf(cline, form, argptr);
        va_end(argptr);

        printf("[HSCALAPACK TIMINGS %s]===> ", dtstr);

        printf(cline); printf("\n");
     }
}

/*---------------------------------------------------------------------------*/
