 
#include "../include/hscalapack.h"
#include "hscal_hashtab.h"

hscal_hashtable *__hscal_hasht;

/*---------------------------------------------------------------------------*/

int hscal_init
(
    int *pargc,
    char*** pargv
)
{
    int rc;

    rc = HMPI_Init(
             pargc, 
             pargv
    );

    if (rc != HMPI_OK)
    {
        return HSCAL_ERR_INITIALIZATION;
    }

    __hscal_hasht = hscal_create_hashtable(
                         HSCAL_HASH_ENTRIES, 
                         hscal_hashfromkey, 
                         hscal_equalkeys
    );


    if (__hscal_hasht == NULL)
    {
       return HSCAL_ERR_LOW_MEMORY;
    }

    return HSCAL_SUCCESS;
}

/*---------------------------------------------------------------------------*/

int hscal_finalize(
    int exit_code
)
{
    int rc;

    /*
    * Free the map of blacs context to HeteroMPI gid
    */
    hscal_hashtable_destroy(__hscal_hasht, 0);

    rc = HMPI_Finalize(exit_code);

    if (rc != HMPI_OK)
    {
       return rc;
    }

    return HSCAL_SUCCESS;
}

/*---------------------------------------------------------------------------*/

double hscal_timeof
(
    int *ctxt
)
{
    HMPI_Group *gid;
    HSCALCONTEXT *found;

    if (ctxt == NULL)
    {
       return -1.;
    }

    if (*ctxt == HSCAL_CTXT_UNDEFINED)
    {
       return -1.;
    }

    if (*ctxt == HSCAL_CTXT_FOR_FREE_PROCS)
    {
       return -1.;
    }

    if (NULL == (found = hscal_hashtable_search(__hscal_hasht, ctxt) ))
    {
       return -1.;
    }

    gid = (HMPI_Group*)(found->gid);

    if (HMPI_Is_member(gid))
    {
       return found->time;
    }

    return -1.;
}

/*---------------------------------------------------------------------------*/

int hscal_in_ctxt
(
    int *ctxt
)
{
    HMPI_Group *gid;
    HSCALCONTEXT *found;

    if (ctxt == NULL)
    {
       return 0;
    }

    if (*ctxt == HSCAL_CTXT_UNDEFINED)
    {
       return 0;
    }

    if (*ctxt == HSCAL_CTXT_FOR_FREE_PROCS)
    {
       return 0;
    }

    if (NULL == (found = hscal_hashtable_search(__hscal_hasht, ctxt) ))
    {
       return 0;
    }

    gid = (HMPI_Group*)(found->gid);

    if (HMPI_Is_member(gid))
    {
       return 1;
    }

    return 0;
}

/*---------------------------------------------------------------------------*/

int hscal_store_gid_to_ctxt_mapping
(
    int *ctxt,
    HSCALCONTEXT *hctxt
)
{
    if (!hscal_hashtable_insert(__hscal_hasht, ctxt, hctxt))
    {
       return HSCAL_ERR_CTXT_STORE;
    }

    return HSCAL_SUCCESS;
}

/*---------------------------------------------------------------------------*/

MPI_Comm* hscal_get_comm
(
    int *ctxt
)
{
    HMPI_Group *gid;
    HSCALCONTEXT *found;

    if (ctxt == NULL)
    {
       return NULL;
    }

    if (*ctxt == HSCAL_CTXT_UNDEFINED)
    {
       return NULL;
    }

    if (*ctxt == HSCAL_CTXT_FOR_FREE_PROCS)
    {
       return NULL;
    }

    if (NULL == (found = hscal_hashtable_search(__hscal_hasht, ctxt) ))
    {
       return NULL;
    }

    gid = (HMPI_Group*)(found->gid);

    return (MPI_Comm*)HMPI_Get_comm(gid);
}

/*---------------------------------------------------------------------------*/

int hscal_free_ctxt
(
    int *ctxt
)
{
    HMPI_Group *gid;
    HSCALCONTEXT *found;

    if (*ctxt == HSCAL_CTXT_UNDEFINED)
    {
       return HSCAL_CTXT_UNDEFINED;
    }

    if (NULL == (found = hscal_hashtable_remove(__hscal_hasht, ctxt) ))
    {
       return HSCAL_ERR_CTXT_FIND;
    }

    gid = (HMPI_Group*)(found->gid);

    /*
    * Free the BLACS context
    */
    if (HMPI_Is_member(gid))
    {
       Cblacs_gridexit(*ctxt);
    }

    /*
    * Free the HeteroMPI group
    */
    HMPI_Group_free(
        gid
    );

    free(gid);

    free(found);

    return HSCAL_SUCCESS;
}

/*---------------------------------------------------------------------------*/

int hscal_timeof_slaves_exp()
{
    int hscal_env_timeof_slaves_exp_yn = 0;
    char *envv = (char*)getenv("HSCAL_TSLAVES_EXP");

    if (envv == NULL)
    {
       return hscal_env_timeof_slaves_exp_yn;
    }

    if ((strcmp(envv, "Y") == 0)
        || (strcmp(envv, "y") == 0)
        || (strcmp(envv, "YES") == 0)
        || (strcmp(envv, "yes") == 0
       )
    )
    {
       hscal_env_timeof_slaves_exp_yn = 1;
    }

    return hscal_env_timeof_slaves_exp_yn;
}

/*---------------------------------------------------------------------------*/

