/*
* Ravi Reddy
* Version 1.0
* School of Computer Science and Informatics, 
* UCD, 
* Dublin,
* May 12, 2008
*/

/*
*  Include files
*/
#include "../include/hscalapack.h"

int hscal_pcgels_ctxt(
   char   * TRANS,
   int    * M,
   int    * N,
   int    * NRHS,
   int    * IA,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * JB,
   int    * DESCB,
   int    * ICTXT
)
{
/*  -- ScaLAPACK routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PCGELS solves overdetermined or underdetermined complex linear */
/*  systems involving an M-by-N matrix sub( A ) = A(IA:IA+M-1,JA:JA+N-1), */
/*  or its conjugate-transpose, using a QR or LQ factorization of */
/*  sub( A ).  It is assumed that sub( A ) has full rank. */

/*  The following options are provided: */

/*  1. If TRANS = 'N' and m >= n:  find the least squares solution of */
/*     an overdetermined system, i.e., solve the least squares problem */
/*                  minimize || sub( B ) - sub( A )*X ||. */

/*  2. If TRANS = 'N' and m < n:  find the minimum norm solution of */
/*     an underdetermined system sub( A ) * X = sub( B ). */

/*  3. If TRANS = 'C' and m >= n:  find the minimum norm solution of */
/*     an undetermined system sub( A )**H * X = sub( B ). */

/*  4. If TRANS = 'C' and m < n:  find the least squares solution of */
/*     an overdetermined system, i.e., solve the least squares problem */
/*                  minimize || sub( B ) - sub( A )**H * X ||. */

/*  where sub( B ) denotes B( IB:IB+M-1, JB:JB+NRHS-1 ) when TRANS = 'N' */
/*  and B( IB:IB+N-1, JB:JB+NRHS-1 ) otherwise. Several right hand side */
/*  vectors b and solution vectors x can be handled in a single call; */
/*  When TRANS = 'N', the solution vectors are stored as the columns of */
/*  the N-by-NRHS right hand side matrix sub( B ) and the M-by-NRHS */
/*  right hand side matrix sub( B ) otherwise. */

/*  Notes */
/*  ===== */

/*  Each global data object is described by an associated description */
/*  vector.  This vector stores the information required to establish */
/*  the mapping between an object element and its corresponding process */
/*  and memory location. */

/*  Let A be a generic term for any 2D block cyclicly distributed array. */
/*  Such a global array has an associated description vector DESCA. */
/*  In the following comments, the character _ should be read as */
/*  "of the global array". */

/*  NOTATION        STORED IN      EXPLANATION */
/*  --------------- -------------- -------------------------------------- */
/*  DTYPE_A(global) DESCA( DTYPE_ )The descriptor type.  In this case, */
/*                                 DTYPE_A = 1. */
/*  CTXT_A (global) DESCA( CTXT_ ) The BLACS context handle, indicating */
/*                                 the BLACS process grid A is distribu- */
/*                                 ted over. The context itself is glo- */
/*                                 bal, but the handle (the integer */
/*                                 value) may vary. */
/*  M_A    (global) DESCA( M_ )    The number of rows in the global */
/*                                 array A. */
/*  N_A    (global) DESCA( N_ )    The number of columns in the global */
/*                                 array A. */
/*  MB_A   (global) DESCA( MB_ )   The blocking factor used to distribute */
/*                                 the rows of the array. */
/*  NB_A   (global) DESCA( NB_ )   The blocking factor used to distribute */
/*                                 the columns of the array. */
/*  RSRC_A (global) DESCA( RSRC_ ) The process row over which the first */
/*                                 row of the array A is distributed. */
/*  CSRC_A (global) DESCA( CSRC_ ) The process column over which the */
/*                                 first column of the array A is */
/*                                 distributed. */
/*  LLD_A  (local)  DESCA( LLD_ )  The leading dimension of the local */
/*                                 array.  LLD_A >= MAX(1,LOCr(M_A)). */

/*  Let K be the number of rows or columns of a distributed matrix, */
/*  and assume that its process grid has dimension p x q. */
/*  LOCr( K ) denotes the number of elements of K that a process */
/*  would receive if K were distributed over the p processes of its */
/*  process column. */
/*  Similarly, LOCc( K ) denotes the number of elements of K that a */
/*  process would receive if K were distributed over the q processes of */
/*  its process row. */
/*  The values of LOCr() and LOCc() may be determined via a call to the */
/*  ScaLAPACK tool function, NUMROC: */
/*          LOCr( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ), */
/*          LOCc( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ). */
/*  An upper bound for these quantities may be computed by: */
/*          LOCr( M ) <= ceil( ceil(M/MB_A)/NPROW )*MB_A */
/*          LOCc( N ) <= ceil( ceil(N/NB_A)/NPCOL )*NB_A */

/*  Arguments */
/*  ========= */

/*  TRANS   (global input) CHARACTER */
/*          = 'N': the linear system involves sub( A ); */
/*          = 'C': the linear system involves sub( A )**H. */

/*  M       (global input) INTEGER */
/*          The number of rows to be operated on, i.e. the number of */
/*          rows of the distributed submatrix sub( A ). M >= 0. */

/*  N       (global input) INTEGER */
/*          The number of columns to be operated on, i.e. the number of */
/*          columns of the distributed submatrix sub( A ). N >= 0. */

/*  NRHS    (global input) INTEGER */
/*          The number of right hand sides, i.e. the number of columns */
/*          of the distributed submatrices sub( B ) and X.  NRHS >= 0. */

/*  IA      (global input) INTEGER */
/*          The row index in the global array A indicating the first */
/*          row of sub( A ). */

/*  JA      (global input) INTEGER */
/*          The column index in the global array A indicating the */
/*          first column of sub( A ). */

/*  DESCA   (global and local input) INTEGER array of dimension DLEN_. */
/*          The array descriptor for the distributed matrix A. */

/*  IB      (global input) INTEGER */
/*          The row index in the global array B indicating the first */
/*          row of sub( B ). */

/*  JB      (global input) INTEGER */
/*          The column index in the global array B indicating the */
/*          first column of sub( B ). */

/*  DESCB   (global and local input) INTEGER array of dimension DLEN_. */
/*          The array descriptor for the distributed matrix B. */

/*  LWORK   (local or global input) INTEGER */
/*          The dimension of the array WORK. */
/*          LWORK is local input and must be at least */
/*          LWORK >= LTAU + MAX( LWF, LWS ) where */
/*          If M >= N, then */
/*            LTAU = NUMROC( JA+MIN(M,N)-1, NB_A, MYCOL, CSRC_A, NPCOL ), */
/*            LWF  = NB_A * ( MpA0 + NqA0 + NB_A ) */
/*            LWS  = MAX( (NB_A*(NB_A-1))/2, (NRHSqB0 + MpB0)*NB_A ) + */
/*                   NB_A * NB_A */
/*          Else */
/*            LTAU = NUMROC( IA+MIN(M,N)-1, MB_A, MYROW, RSRC_A, NPROW ), */
/*            LWF  = MB_A * ( MpA0 + NqA0 + MB_A ) */
/*            LWS  = MAX( (MB_A*(MB_A-1))/2, ( NpB0 + MAX( NqA0 + */
/*                   NUMROC( NUMROC( N+IROFFB, MB_A, 0, 0, NPROW ), */
/*                   MB_A, 0, 0, LCMP ), NRHSqB0 ) )*MB_A ) + */
/*                   MB_A * MB_A */
/*          End if */

/*          where LCMP = LCM / NPROW with LCM = ILCM( NPROW, NPCOL ), */

/*          IROFFA = MOD( IA-1, MB_A ), ICOFFA = MOD( JA-1, NB_A ), */
/*          IAROW = INDXG2P( IA, MB_A, MYROW, RSRC_A, NPROW ), */
/*          IACOL = INDXG2P( JA, NB_A, MYCOL, CSRC_A, NPCOL ), */
/*          MpA0 = NUMROC( M+IROFFA, MB_A, MYROW, IAROW, NPROW ), */
/*          NqA0 = NUMROC( N+ICOFFA, NB_A, MYCOL, IACOL, NPCOL ), */

/*          IROFFB = MOD( IB-1, MB_B ), ICOFFB = MOD( JB-1, NB_B ), */
/*          IBROW = INDXG2P( IB, MB_B, MYROW, RSRC_B, NPROW ), */
/*          IBCOL = INDXG2P( JB, NB_B, MYCOL, CSRC_B, NPCOL ), */
/*          MpB0 = NUMROC( M+IROFFB, MB_B, MYROW, IBROW, NPROW ), */
/*          NpB0 = NUMROC( N+IROFFB, MB_B, MYROW, IBROW, NPROW ), */
/*          NRHSqB0 = NUMROC( NRHS+ICOFFB, NB_B, MYCOL, IBCOL, NPCOL ), */

/*          ILCM, INDXG2P and NUMROC are ScaLAPACK tool functions; */
/*          MYROW, MYCOL, NPROW and NPCOL can be determined by calling */
/*          the subroutine BLACS_GRIDINFO. */

/*          If LWORK = -1, then LWORK is global input and a workspace */
/*          query is assumed; the routine only calculates the minimum */
/*          and optimal size for all work arrays. Each of these */
/*          values is returned in the first entry of the corresponding */
/*          work array, and no error message is issued by PXERBLA. */

/*  ===================================================================== */

/*
*  .. Local Scalars ..
*/
    int            rc, p, q;
    HMPI_Group     *gid;
    MPI_Comm       *pcgelscomm;
    HSCALCONTEXT   *hctxt; 
    double         algotime;
   
/*
* Initialize the context ID
*/
    *ICTXT = HSCAL_CTXT_UNDEFINED;

/*
* Dynamically refresh the speeds of the processors
*/
    rc = hscal_pcgels_recon(
              TRANS,
              M,
              N,
              NRHS,
              IA,
              JA,
              DESCA,
              IB,
              JB,
              DESCB
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_RECON;
    }

/*
* Create a HeteroMPI group with the optimal grid arrangement 
* of processes
*/
    gid  = (HMPI_Group*)malloc(sizeof(HMPI_Group));

    if (gid == NULL)
    {
       return HSCAL_ERR_LOW_MEMORY;
    }

    if (*M >= *N)
    {
       rc = hscal_pcgels_group_heuristic_pauto_create(
                 TRANS,
                 M,
                 N,
                 NRHS,
                 IA,
                 JA,
                 DESCA,
                 IB,
                 JB,
                 DESCB,
                 gid,
                 hscal_model_pcgels_m_ge_n_optimized,
                 hscal_heuristic_fn_nonlinear
       );
    }
    else
    {
       rc = hscal_pcgels_group_heuristic_pauto_create(
                 TRANS,
                 M,
                 N,
                 NRHS,
                 IA,
                 JA,
                 DESCA,
                 IB,
                 JB,
                 DESCB,
                 gid,
                 hscal_model_pcgels_m_lt_n_optimized,
                 hscal_heuristic_fn_nonlinear
       );
    }

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_GROUP_CREATION;
    }

    if (HMPI_Is_free())
    {
       *ICTXT = HSCAL_CTXT_FOR_FREE_PROCS;

       hctxt  = (HSCALCONTEXT*)malloc(sizeof(HSCALCONTEXT));

       if (hctxt == NULL)
       {
          return HSCAL_ERR_LOW_MEMORY;
       }

       hctxt->time = -1.;
       hctxt->optp = -1;
       hctxt->optq = -1;
       hctxt->gid  = gid;

       rc = hscal_store_gid_to_ctxt_mapping(ICTXT, hctxt);

       return rc;
    }

    pcgelscomm = (MPI_Comm*)HMPI_Get_comm(gid);

    if (pcgelscomm == NULL)
    {
       return HSCAL_ERR_NULL_COMM; 
    } 

/* 
* Broadcast the estimated execution time 
*/ 
    algotime = HMPI_Group_timeof(gid);

/*
* Broadcast the optimal grid arrangement
*/
    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       p = d[0];

       free(d);

       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *pcgelscomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *pcgelscomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       q = d[1];

       free(d);

       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *pcgelscomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *pcgelscomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       HSCAL_DEBUG(
            __FILE__, __FUNCTION__, __LINE__,
            "Optimal number of process rows =%d, Optimal number of process columns =%d, " 
            "Estimated execution time=%f ",
            p, q, algotime
       );       
    }

/*
* Translate communicator to a BLACS handle
*/
    *ICTXT = Csys2blacs_handle(*pcgelscomm);

/*
* Form BLACS context
*/
    Cblacs_gridinit(ICTXT, "r", p, q);

/*
* Store the context to gid mapping
*/
    hctxt  = (HSCALCONTEXT*)malloc(sizeof(HSCALCONTEXT)); 
 
    if (hctxt == NULL) 
    { 
       return HSCAL_ERR_LOW_MEMORY; 
    } 
 
    hctxt->time = algotime; 
    hctxt->optp = p; 
    hctxt->optq = q; 
    hctxt->gid  = gid; 
 
    rc = hscal_store_gid_to_ctxt_mapping(ICTXT, hctxt); 

    return rc;

/*     End of PCGELS */

}

