
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcgesv_recon_tcomp(
   int N,
   int NRHS
)
{
    double tcomp = 0;

    char   transc  = 'N';
    char   *trans  = &transc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    float alpha[CNARG];

    alpha[REAL_PART] = 1.f;
    alpha[IMAG_PART] = 1.f;

    if (transc == 'N')
    {
       reconm  = N/p;
       reconn  = N/q;
       reconk  = hscal_pilaenv_recon("C");
    }
    else
    {
       reconm  = hscal_pilaenv_recon("C");
       reconn  = N/p;
       reconk  = N/q;
    }

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_cgemm(
         trans,
         trans,
         &reconm,
         &reconn,
         &reconk,
         ((char*)alpha),
         &lda,
         &ldb,
         ((char*)alpha),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pcgesv_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pcgesv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pcgesv_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pcgesv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgesv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pcgesv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pcgesv_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__1 = 1;
    hscalinteger c__6 = 6;
    hscalinteger c__2 = 2;
    hscalinteger c__11 = 11;
    hscalinteger c__0 = 0;

    int     *n      = &N;
    int     *nrhs   = &NRHS;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *ib     = &IB;
    int     *jb     = &JB;
    int     *descb  = (int*)DESCB;

    int     i__1, idum1[1], idum2[1], iarow, ibrow, ictxt,
            icoffa, iroffa, iroffb;

    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descb;
    --desca;

/*     Test the input parameters */
    if (init_llds_f == 1)
    {
       int *PXGESV_DESCA_l = &desca[1];
       int *PXGESV_DESCB_l = &descb[1];
       HSCAL_PXGESV_INIT_LLDS(
            PXGESV_DESCA_l,
            PXGESV_DESCB_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PCGESV: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d", 
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PCGESV: (%d, %d): (descb+1)[M1_]=%d (descb+1)[MB1_]=%d myrow=%d (descb+1)[RSRC1_]=%d nprow=%d (descb+1)[LLD1_]=%d",
        myrow, mycol, (descb+1)[M1_], (descb+1)[MB1_], myrow, (descb+1)[RSRC1_], nprow, (descb+1)[LLD1_]
    );

    *info = 0;
    if (nprow == -1) {
	*info = -602;
    } else {
	hscal_chk1mat(n, &c__1, n, &c__1, ia, ja, &desca[1], &c__6, info,
                      myrow, mycol, nprow, npcol
        );
	hscal_chk1mat(n, &c__1, nrhs, &c__2, ib, jb, &descb[1], &c__11, info,
                      myrow, mycol, nprow, npcol
        );
	if (*info == 0) {
	    iarow = indxg2p_(ia, &desca[5], &myrow, &desca[7], &nprow);
	    ibrow = indxg2p_(ib, &descb[5], &myrow, &descb[7], &nprow);
	    iroffa = (*ia - 1) % desca[5];
	    icoffa = (*ja - 1) % desca[6];
	    iroffb = (*ib - 1) % descb[5];
	    if (iroffa != 0) {
		*info = -4;
	    } else if (icoffa != 0) {
		*info = -5;
	    } else if (desca[5] != desca[6]) {
		*info = -606;
	    } else if (ibrow != iarow || icoffa != iroffb) {
		*info = -9;
	    } else if (descb[5] != desca[6]) {
		*info = -1106;
	    }
	}

	hscal_pchk2mat(n, &c__1, n, &c__1, ia, ja, &desca[1], &c__6, n, &c__1, 
		nrhs, &c__2, ib, jb, &descb[1], &c__11, &c__0, idum1, idum2, 
		info, &myrow, &mycol, &nprow, &npcol);
    }

    if (*info != 0) {
       i__1 = -(*info);
       hscal_PB_Cabort( 
            "PCGESV", 
            i__1,
            nprow, npcol, myrow, mycol
       ); 
       return HSCAL_ERR_INPUT;
    }

/*     Compute the LU factorization of sub( A ). */

    {
       int *PCGETRF_DESCA_l = (&desca[1]);

       hscal_pcgetrf_my_tcomp_tcomm(
           0,
           myrow, mycol,
           N, N, 
           IA, JA, 
           PCGETRF_DESCA_l,
           nprow, npcol,
           tcomp, tcomm
       );
    }

    {

/*        Solve the system sub( A ) * X = sub( B ), overwriting sub( B ) */
/*        with X. */

        int *PCGETRS_DESCA_l = (&desca[1]);
        int *PCGETRS_DESCB_l = (&descb[1]);
        int TRANSI           = 'N';

	hscal_pcgetrs_my_tcomp_tcomm(
               0,
               myrow, mycol,
               TRANSI, N, NRHS, 
               IA, JA, PCGETRS_DESCA_l,
	       IB, JB, PCGETRS_DESCB_l,
               nprow, npcol,
               tcomp, tcomm
        );
    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgesv_tcomp_tcomm(
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcgesv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N, NRHS,
                IA, JA, DESCA,
                IB, JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgesv_info(
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcgesv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgesv_info_i(
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pcgesv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
