
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pclapiv_recon_tcomp(
   int DIRECI,
   int ROWCOLI,
   int PIVROCI,
   int N
)
{
    double tcomp = 1;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pclapiv_my_tcomp(
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int PIVROCI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCP[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pclapiv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI, PIVROCI,
          M, N,
          IA, JA, DESCA, 
          IP, JP, DESCP, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pclapiv_my_tcomm(
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int PIVROCI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCP[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pclapiv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI, PIVROCI,
          M, N,
          IA, JA, DESCA,
          IP, JP, DESCP,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pclapiv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int DIRECI,
   int ROWCOLI,
   int PIVROCI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCP[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pclapiv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI, PIVROCI,
          M, N,
          IA, JA, DESCA,
          IP, JP, DESCP,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pclapiv_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int PIVROCI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCIP[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__1 = 1;
    hscalinteger c__0 = 0;

    /* System generated locals */
    hscalinteger i__1, i__2;

    /* Local variables */
    hscalinteger i__;
    hscalinteger iip, jjp, ipt, jpt, itmp;
    hscalinteger ictxt;
    hscallogical rowpvt;
    hscalinteger icurcol, icurrow;

/*
* .. Intermediate outputs and inputs ..
*/
    int descpt[DLEN1_];

/*
*  .. Local Arrays ..
*/
    HSCALBLACSCONTEXT   * blacsctxt;

    char    DIRECC  = DIRECI, *direc = &DIRECC;
    char    ROWCOLC = ROWCOLI, *rowcol = &ROWCOLC;
    char    PIVROCC = PIVROCI, *pivroc = &PIVROCC;
    int     *m      = &M;
    int     *n      = &N;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *ip     = &IP;
    int     *jp     = &JP;
    int     *descip = (int*)DESCIP;
    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    --descip;
    --desca;

    if (init_llds_f == 1)
    {
       HSCAL_PXLAPIV_INIT_LLDS(
            (desca+1),
            myrow, nprow
       );
    }

    /* Function Body */
    rowpvt = hscal_lsame(rowcol, "R");

/*     If we're pivoting the rows of sub( A ) */

    if (rowpvt) {
	if (*m <= 1 || *n < 1) {
	   return HSCAL_SUCCESS;
	}

/*        If the pivot vector is already distributed correctly */

	if (hscal_lsame(pivroc, "C")) {
            {
              int *PCLAPV2_DESCA_l = (&desca[1]);
              int *PCLAPV2_DESCIP_l = (descip+1);
	      hscal_pclapv2_my_tcomp_tcomm(
                   0,
                   myrow, mycol,
                   DIRECI, ROWCOLI, M, N, IA, JA, PCLAPV2_DESCA_l, 
	           IP, JP, PCLAPV2_DESCIP_l, nprow, npcol, tcomp, tcomm
              );
            }

/*        Otherwise, we must redistribute IPIV to match PCLAPV2 */

	} else {

/*           Take IPIV distributed over row 0, and store it in */
/*           iwork, distributed over column 0 */

	    ipt = (*jp - 1) % desca[5];
            descpt[2] = *m + ipt + nprow * desca[5];
            descpt[3] = 1;
            descpt[4] = desca[5];
            descpt[5] = 1;
            descpt[6] = indxg2p_(ia, &desca[5], ia, &desca[7], &nprow);
            descpt[7] = mycol;
            descpt[1] = ictxt;
            descpt[8] = numroc_(
                  &descpt[2],
                  &descpt[4],
                  &myrow, &descpt[6], &nprow);
            itmp = numroc_(&descip[4],
                           &descip[6], &mycol,
                           &descip[8], &npcol);
	    i__1 = *jp - ipt;
	    infog2l_(ip, &i__1, &descip[1], &nprow, &npcol, &myrow, &mycol, &
		    iip, &jjp, &icurrow, &icurcol);
	    i__1 = *m + ipt;
	    hscal_pirow2col(&ictxt, &i__1, &c__1, &descip[6], &itmp,
		    &descpt[8], &c__0, &icurcol,
                    &descpt[6], &mycol,
		    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);

/*           Send column-distributed pivots to all columns */

/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
            Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

	    itmp = descpt[8] - descpt[4];
	    if (mycol == 0) {
		hscal_Cigebs2d(blacsctxt, "Row", " ", itmp, c__1, itmp,
                     nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    } else {
		hscal_Cigebr2d(blacsctxt, "Row", " ", itmp, c__1, itmp,
			myrow, c__0, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    }

/*           Adjust pivots so they are relative to the start of IWORK, */
/*           not IPIV */

            ++ipt;
            TCOMPL((myrow), (mycol), (nprow), (npcol)) += 2*itmp;

            {
              int *PCLAPV2_DESCA_l = (&desca[1]);
	      hscal_pclapv2_my_tcomp_tcomm(
                     0,
                     myrow, mycol,
                     DIRECI, ROWCOLI, M, N, IA, JA,
                     PCLAPV2_DESCA_l,
		     ipt, c__1, descpt,
                     nprow, npcol, tcomp, tcomm);
            }

            Cblacs_gridexit_nocomm(blacsctxt);
	}

/*     Otherwise, we're pivoting the columns of sub( A ) */

    } else {
	if (*m < 1 || *n <= 1) {
	    return HSCAL_SUCCESS;
	}

/*        If the pivot vector is already distributed correctly */

	if (hscal_lsame(pivroc, "R")) {
            {
              int *PCLAPV2_DESCA_l = (&desca[1]);
              int *PCLAPV2_DESCIP_l = (descip+1);
	      hscal_pclapv2_my_tcomp_tcomm(
                    0,
                    myrow, mycol,
                    DIRECI, ROWCOLI, M, N, IA, JA,
                    PCLAPV2_DESCA_l, 
		    IP, JP, PCLAPV2_DESCIP_l, nprow, npcol, tcomp, tcomm);
            }

/*        Otherwise, we must redistribute IPIV to match PCLAPV2 */

	} else {

/*           Take IPIV distributed over column 0, and store it in */
/*           iwork, distributed over row 0 */

	    jpt = (*ip - 1) % desca[6];
	    descpt[2] = 1;
	    descpt[3] = *n + jpt + npcol * desca[6];
	    descpt[4] = 1;
	    descpt[5] = desca[6];
	    descpt[6] = myrow;
	    descpt[7] = indxg2p_(ja, &desca[6], ja, &desca[8], &npcol);
	    descpt[1] = ictxt;
	    descpt[8] = 1;
	    i__1 = *ip - jpt;
	    infog2l_(&i__1, jp, &descip[1], &nprow, &npcol, &myrow, &mycol, &
		    iip, &jjp, &icurrow, &icurcol);
	    i__1 = *n + jpt;
	    itmp = numroc_(&i__1, &descpt[5],
                           &mycol, &descpt[7], &npcol);
	    i__1 = *n + jpt;
	    i__2 = max(1,itmp);
	    hscal_picol2row(&ictxt, &i__1, &c__1, &descip[5],
                    &descip[9], &i__2, &icurrow, &c__0, &c__0,
                    &descpt[7],
                    nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);

/*           Send row-distributed pivots to all rows */

/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
            Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

	    if (myrow == 0) {
		hscal_Cigebs2d(blacsctxt, "Column", " ", itmp, c__1,
			itmp, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    } else {
		hscal_Cigebr2d(blacsctxt, "Column", " ", itmp, c__1, 
			itmp, c__0, mycol, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    }

/*           Adjust pivots so they are relative to the start of IWORK, */
/*           not IPIV */

            ++jpt;
            TCOMPL((myrow), (mycol), (nprow), (npcol)) += 2*itmp;

            {
              int *PCLAPV2_DESCA_l = (&desca[1]);
	      hscal_pclapv2_my_tcomp_tcomm(
                   0,
                   myrow, mycol,
                   DIRECI, ROWCOLI, M, N, IA, JA,
                   PCLAPV2_DESCA_l,
		   c__1, jpt, descpt, nprow, npcol, tcomp, tcomm);
            }

/*
* Free the intermediate data
*/
            Cblacs_gridexit_nocomm(blacsctxt);
	}
    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pclapiv_tcomp_tcomm(
   int DIRECI,
   int ROWCOLI,
   int PIVROCI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCP[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pclapiv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                DIRECI, ROWCOLI, PIVROCI,
                M, N,
                IA, JA, DESCA,
                IP, JP, DESCP,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pclapiv_info(
   char *DIRECC,
   char *ROWCOLC,
   char *PIVROCC,
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IP,
   int * JP,
   int * DESCP,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pclapiv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *DIRECC, *ROWCOLC, *PIVROCC,
                *M, *N,
                *IA, *JA, DESCA,
                *IP, *JP, DESCP,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pclapiv_info_i(
   char *DIRECC,
   char *ROWCOLC,
   char *PIVROCC,
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IP,
   int * JP,
   int * DESCP,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pclapiv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *DIRECC, *ROWCOLC, *PIVROCC,
                *M, *N,
                *IA, *JA, DESCA,
                *IP, *JP, DESCP,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

