
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pclarft_recon_tcomp(
   int directi,
   int storei,
   int n,
   int k
)
{
    double tcomp = 0;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pclarft_my_tcomp(
   int myrow,
   int mycol,
   int DIRECTI,
   int STOREI,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pclarft_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECTI, STOREI,
          N, K,
          IV, JV, DESCV,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pclarft_my_tcomm(
   int myrow,
   int mycol,
   int DIRECTI,
   int STOREI,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pclarft_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECTI, STOREI,
          N, K,
          IV, JV, DESCV,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pclarft_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int DIRECTI,
   int STOREI,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pclarft_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECTI, STOREI,
          N, K,
          IV, JV, DESCV,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pclarft_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int DIRECTI,
   int STOREI,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalcomplex c_b2 = {0.f,0.f};
    hscalinteger c__1 = 1;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3;
    hscalcomplex q__1;

    /* Local variables */
    hscalinteger ii, jj, np, iw, nq, ldv, iiv;
    hscalinteger jjv, itmp0, itmp1, icoff;
    hscalinteger micol, iroff, ivcol;
    hscalinteger ictxt;
    hscalinteger mirow, ivrow;
    hscallogical forward;

/*
*  .. Local Arrays ..
*/
    HSCALBLACSCONTEXT   * blacsctxt;

    char directc    = DIRECTI, *direct = &directc;
    char storec     = STOREI, *storev = &storec;

    int     *n      = &N;
    int     *k      = &K;
    int     *iv     = &IV;
    int     *jv     = &JV;
    int     *descv  = (int*)DESCV;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descv;

    if (init_llds_f == 1)
    {
       int *PXLARFT_DESCV_l = &descv[1];
       HSCAL_PXLARFT_INIT_LLDS(
            PXLARFT_DESCV_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pclarft: (%d, %d): (descv+1)[M1_]=%d (descv+1)[MB1_]=%d myrow=%d (descv+1)[RSRC1_]=%d nprow=%d (descv+1)[LLD1_]=%d", 
        myrow, mycol, (descv+1)[M1_], (descv+1)[MB1_], myrow, (descv+1)[RSRC1_], nprow, (descv+1)[LLD1_]
    );


    /* Function Body */
    if (*n <= 0 || *k <= 0) {
	return HSCAL_SUCCESS;
    }

   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

    forward = hscal_lsame(direct, "F");
    infog2l_(iv, jv, &descv[1], &nprow, &npcol, &myrow, &mycol, &iiv, &jjv, &
	    ivrow, &ivcol);

    if (hscal_lsame(storev, "C") && mycol == ivcol) {

	iw = 1;
	ldv = descv[9];
	iroff = (*iv - 1) % descv[5];

	if (forward) {

/*           DIRECT = 'Forward', STOREV = 'Columnwise' */

	    i__1 = *n + iroff;
	    np = numroc_(&i__1, &descv[5], &myrow, &ivrow, &nprow);
	    if (myrow == ivrow) {
		np -= iroff;
		ii = iiv + 1;
	    } else {
		ii = iiv;
	    }
	    if (iroff + 1 == descv[5]) {
		mirow = (ivrow + 1) % nprow;
	    } else {
		mirow = ivrow;
	    }
	    itmp0 = 0;

	    i__1 = jjv + *k - 1;
	    for (jj = jjv + 1; jj <= i__1; ++jj) {

		if (myrow == mirow) {
		    ;
		}

/*              T(1:i-1,i) = -tau( jv+i-1 ) * */
/*              V(iv+i-1:iv+n-1,jv:jv+i-2)' * V(iv+i-1:iv+n-1,jv+i-1) */

		++itmp0;
		if (np - ii + iiv > 0) {
		    i__2 = np - ii + iiv;
		    i__3 = jj;
/*
* Assume some values of q__1 (as these are not real computations)
*/
                q__1.r = -.5f, q__1.i = -.5f;
		    cgemv_tcomp_macro(
                     "Conjugate transpose", &i__2, &itmp0, (char*)&q__1,
			   &ldv,
                     &c__1, (char*)&c_b2, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		} else {
		    claset_tcomp_macro(
                      "All", &itmp0, &c__1, (char*)&c_b2, (char*)&c_b2, 
                      &itmp0,
                      nprow, npcol, myrow, mycol, tcomp
                );
		}

		iw += itmp0;
		if (myrow == mirow) {
		    ++ii;
		}

		if ((*iv + itmp0) % descv[5] == 0) {
		    mirow = (mirow + 1) % nprow;
		}

/* L10: */
	    }

	    i__1 = iw - 1;
	    i__2 = iw - 1;
	    hscal_Ccgsum2d(
               blacsctxt, "Columnwise", " ", i__1, c__1, i__2,
		   ivrow, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

	    if (myrow == ivrow) {

		iw = 1;
		itmp0 = 0;
		itmp1 = 1;

		i__1 = itmp1;
		i__2 = jjv;

		i__1 = jjv + *k - 1;
		for (jj = jjv + 1; jj <= i__1; ++jj) {

/*                 T(1:j-1,j) = T(1:j-1,1:j-1) * T(1:j-1,j) */

		    ++itmp0;
		    itmp1 += descv[6];
		    ccopy_tcomp_macro(
                     &itmp0, &c__1, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    iw += itmp0;

		    ctrmv_tcomp_macro(
                     "Upper", "No transpose", "Non-unit", &itmp0,
			   &descv[6], &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    i__2 = itmp1 + itmp0;
		    i__3 = jj;

/* L20: */
		}

	    }

	} else {

/*           DIRECT = 'Backward', STOREV = 'Columnwise' */

	    i__1 = *n + iroff - 1;
	    np = numroc_(&i__1, &descv[5], &myrow, &ivrow, &nprow);
	    if (myrow == ivrow) {
		np -= iroff;
	    }
	    i__1 = *iv + *n - 2;
	    mirow = indxg2p_(&i__1, &descv[5], &myrow, &descv[7], &nprow);
	    ii = iiv + np - 1;
	    itmp0 = 0;

	    i__1 = jjv;
	    for (jj = jjv + *k - 2; jj >= i__1; --jj) {

		if (myrow == mirow) {
               ;
		}

/*              T(1:i-1,i) = -tau( jv+i-1 ) * */
/*              V(iv:iv+n-k+i-1,jv+i:jv+k-1)' * V(iv:iv+n-k+i-1,jv+i-1) */

		++itmp0;
		if (ii - iiv + 1 > 0) {
		    i__2 = ii - iiv + 1;
		    i__3 = jj;
/*
* Assume some values of q__1 (as these are not real computations)
*/
                q__1.r = -.5f, q__1.i = -.5f;
		    cgemv_tcomp_macro(
                     "Conjugate transpose", &i__2, &itmp0, (char*)&q__1,
			   &ldv, &c__1, (char*)&c_b2, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		} else {
		    claset_tcomp_macro(
                      "All", &itmp0, &c__1, (char*)&c_b2, (char*)&c_b2, 
                      &itmp0,
                      nprow, npcol, myrow, mycol, tcomp
                );
		}

		iw += itmp0;
		if (myrow == mirow) {
		    --ii;
		}

		if ((*iv + *n - itmp0 - 2) % descv[5] == 0) {
		    mirow = (mirow + nprow - 1) % nprow;
		}

/* L30: */
	    }

	    i__1 = iw - 1;
	    i__2 = iw - 1;
	    hscal_Ccgsum2d(
               blacsctxt, "Columnwise", " ", i__1, c__1, i__2,
		   ivrow, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

	    if (myrow == ivrow) {

		iw = 1;
		itmp0 = 0;
		itmp1 = *k + 1 + (*k - 1) * descv[6];

		i__1 = itmp1 - 1;
		i__2 = jjv + *k - 1;

		i__1 = jjv;
		for (jj = jjv + *k - 2; jj >= i__1; --jj) {

/*                 T(j+1:k,j) = T(j+1:k,j+1:k) * T(j+1:k,j) */

		    ++itmp0;
		    itmp1 = itmp1 - descv[6] - 1;
		    ccopy_tcomp_macro(
                     &itmp0, &c__1, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    iw += itmp0;

		    ctrmv_tcomp_macro(
                     "Lower", "No transpose", "Non-unit", &itmp0,
			   &descv[6], &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    i__2 = itmp1 - 1;
		    i__3 = jj;

/* L40: */
		}

	    }

	}

    } else if (hscal_lsame(storev, "R") && myrow == ivrow) {

	iw = 1;
	ldv = descv[9];
	icoff = (*jv - 1) % descv[6];

	if (forward) {

/*           DIRECT = 'Forward', STOREV = 'Rowwise' */

	    i__1 = *n + icoff;
	    nq = numroc_(&i__1, &descv[6], &mycol, &ivcol, &npcol);
	    if (mycol == ivcol) {
		nq -= icoff;
		jj = jjv + 1;
	    } else {
		jj = jjv;
	    }
	    if (icoff + 1 == descv[6]) {
		micol = (ivcol + 1) % npcol;
	    } else {
		micol = ivcol;
	    }
	    itmp0 = 0;

	    i__1 = iiv + *k - 1;
	    for (ii = iiv + 1; ii <= i__1; ++ii) {

		if (mycol == micol) {
               ;
		}

/*              T(1:i-1,i) = -tau( iv+i-1 ) * */
/*              V(iv+i-1,jv+i-1:jv+n-1) * V(iv:iv+i-2,jv+i-1:jv+n-1)' */

		++itmp0;
		if (nq - jj + jjv > 0) {
		    i__2 = nq - jj + jjv;
		    clacgv_tcomp_macro(
                      &i__2, &ldv,
                      nprow, npcol, myrow, mycol, tcomp
                );
		    i__2 = nq - jj + jjv;
		    i__3 = ii;
/*
* Assume some values of q__1 (as these are not real computations)
*/
                q__1.r = -.5f, q__1.i = -.5f;
		    cgemv_tcomp_macro(
                     "No transpose", &itmp0, &i__2, (char*)&q__1, &ldv, &ldv, 
			   (char*)&c_b2, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    i__2 = nq - jj + jjv;
		    clacgv_tcomp_macro(
                      &i__2, &ldv,
                      nprow, npcol, myrow, mycol, tcomp
                );
		} else {
		    claset_tcomp_macro(
                      "All", &itmp0, &c__1, (char*)&c_b2, (char*)&c_b2, 
                      &itmp0,
                      nprow, npcol, myrow, mycol, tcomp
                );
		}

		iw += itmp0;
		if (mycol == micol) {
		    ++jj;
		}

		if ((*jv + itmp0) % descv[6] == 0) {
		    micol = (micol + 1) % npcol;
		}

/* L50: */
	    }

	    i__1 = iw - 1;
	    i__2 = iw - 1;
	    hscal_Ccgsum2d(
               blacsctxt, "Rowwise", " ", i__1, c__1, i__2,
		   myrow, ivcol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

	    if (mycol == ivcol) {

		iw = 1;
		itmp0 = 0;
		itmp1 = 1;

		i__1 = itmp1;
		i__2 = iiv;

		i__1 = iiv + *k - 1;
		for (ii = iiv + 1; ii <= i__1; ++ii) {

/*                 T(1:i-1,i) = T(1:i-1,1:i-1) * T(1:i-1,i) */

		    ++itmp0;
		    itmp1 += descv[5];
		    ccopy_tcomp_macro(
                     &itmp0, &c__1, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    iw += itmp0;

		    ctrmv_tcomp_macro(
                     "Upper", "No transpose", "Non-unit", &itmp0,
			   &descv[5], &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    i__2 = itmp1 + itmp0;
		    i__3 = ii;

/* L60: */
		}

	    }

	} else {

/*           DIRECT = 'Backward', STOREV = 'Rowwise' */

	    i__1 = *n + icoff - 1;
	    nq = numroc_(&i__1, &descv[6], &mycol, &ivcol, &npcol);
	    if (mycol == ivcol) {
		nq -= icoff;
	    }
	    i__1 = *jv + *n - 2;
	    micol = indxg2p_(&i__1, &descv[6], &mycol, &descv[8], &npcol);
	    jj = jjv + nq - 1;
	    itmp0 = 0;

	    i__1 = iiv;
	    for (ii = iiv + *k - 2; ii >= i__1; --ii) {

		if (mycol == micol) {
               ;
		}

/*              T(i+1:k,i) = -tau( iv+i-1 ) * */
/*              V(iv+i:iv+k-1,jv:jv+n-k+i-1)' * V(iv+i-1,jv:jv+n-k+i-1)' */

		++itmp0;
		if (jj - jjv + 1 > 0) {
		    i__2 = jj - jjv + 1;
		    clacgv_tcomp_macro(
                      &i__2, &ldv,
                      nprow, npcol, myrow, mycol, tcomp
                );
		    i__2 = jj - jjv + 1;
		    i__3 = ii;
/*
* Assume some values of q__1 (as these are not real computations)
*/
                q__1.r = -.5f, q__1.i = -.5f;
		    cgemv_tcomp_macro(
                     "No transpose", &itmp0, &i__2, (char*)&q__1,
			   &ldv, &ldv, (char*)&c_b2, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    i__2 = jj - jjv + 1;
		    clacgv_tcomp_macro(
                      &i__2, &ldv,
                      nprow, npcol, myrow, mycol, tcomp
                );
		} else {
		    claset_tcomp_macro(
                      "All", &itmp0, &c__1, (char*)&c_b2, (char*)&c_b2, 
                      &itmp0,
                      nprow, npcol, myrow, mycol, tcomp
                );
		}

		iw += itmp0;
		if (mycol == micol) {
		    --jj;
		}

		if ((*jv + *n - itmp0 - 2) % descv[6] == 0) {
		    micol = (micol + npcol - 1) % npcol;
		}

/* L70: */
	    }

	    i__1 = iw - 1;
	    i__2 = iw - 1;
	    hscal_Ccgsum2d(
               blacsctxt, "Rowwise", " ", i__1, c__1, i__2,
		   myrow, ivcol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

	    if (mycol == ivcol) {

		iw = 1;
		itmp0 = 0;
		itmp1 = *k + 1 + (*k - 1) * descv[5];

		i__1 = itmp1 - 1;
		i__2 = iiv + *k - 1;

		i__1 = iiv;
		for (ii = iiv + *k - 2; ii >= i__1; --ii) {

/*                 T(i+1:k,i) = T(i+1:k,i+1:k) * T(i+1:k,i) */

		    ++itmp0;
		    itmp1 = itmp1 - descv[5] - 1;
		    ccopy_tcomp_macro(
                     &itmp0, &c__1, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    iw += itmp0;

		    ctrmv_tcomp_macro(
                     "Lower", "No transpose", "Non-unit", &itmp0,
			   &descv[5], &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    i__2 = itmp1 - 1;
		    i__3 = ii;

/* L80: */
		}

	    }

	}

    }

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pclarft_tcomp_tcomm(
   int DIRECTI,
   int STOREI,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pclarft_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                DIRECTI, STOREI,
                N, K,
                IV, JV, DESCV,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pclarft_info(
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pclarft_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *DIRECT, *STOREV,
                *N, *K,
                *IV, *JV, DESCV,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pclarft_info_i(
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pclarft_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *DIRECT, *STOREV,
                *N, *K,
                *IV, *JV, DESCV,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
