
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcpotf2_recon_tcomp(
   int UPLOI,
   int N
)
{
    double tcomp = 1;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pcpotf2_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = ihscal_pcpotf2_tcomp
        (
          1,
          myrow, mycol,
          UPLOI, N,
          IA, JA, DESCA,
          nprow, npcol,
          tcomp
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int ihscal_pcpotf2_tcomp(
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp
)
{
    hscalinteger c__2 = 2;
    hscalinteger c__6 = 6;
    hscalinteger c__1 = 1;
    hscalcomplex c_b19 = {-1.f,-0.f};
    hscalcomplex c_b21 = {1.f,0.f};

    /* System generated locals */
    hscalinteger i__1, i__2, i__3;
    hscalreal r__1;

    /* Local variables */
    hscalinteger j;
    hscalinteger lda, iia, jja;
    hscalinteger idiag, ioffa, icoff, iacol;
    hscalinteger iroff, iarow, icurr;
    hscallogical upper;
    hscalinteger ictxt;
    char colbtop[1], rowbtop[1];

/*  -- ScaLAPACK routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PCPOTF2 computes the Cholesky factorization of a complex hermitian */
/*  positive definite distributed matrix sub( A )=A(IA:IA+N-1,JA:JA+N-1). */

/*  The factorization has the form */

/*            sub( A ) = U' * U ,  if UPLO = 'U', or */

/*            sub( A ) = L  * L',  if UPLO = 'L', */

/*  where U is an upper triangular matrix and L is lower triangular. */

/*  Notes */
/*  ===== */

/*  Each global data object is described by an associated description */
/*  vector.  This vector stores the information required to establish */
/*  the mapping between an object element and its corresponding process */
/*  and memory location. */

/*  Let A be a generic term for any 2D block cyclicly distributed array. */
/*  Such a global array has an associated description vector DESCA. */
/*  In the following comments, the character _ should be read as */
/*  "of the global array". */

/*  NOTATION        STORED IN      EXPLANATION */
/*  --------------- -------------- -------------------------------------- */
/*  DTYPE_A(global) DESCA( DTYPE_ )The descriptor type.  In this case, */
/*                                 DTYPE_A = 1. */
/*  CTXT_A (global) DESCA( CTXT_ ) The BLACS context handle, indicating */
/*                                 the BLACS process grid A is distribu- */
/*                                 ted over. The context itself is glo- */
/*                                 bal, but the handle (the integer */
/*                                 value) may vary. */
/*  M_A    (global) DESCA( M_ )    The number of rows in the global */
/*                                 array A. */
/*  N_A    (global) DESCA( N_ )    The number of columns in the global */
/*                                 array A. */
/*  MB_A   (global) DESCA( MB_ )   The blocking factor used to distribute */
/*                                 the rows of the array. */
/*  NB_A   (global) DESCA( NB_ )   The blocking factor used to distribute */
/*                                 the columns of the array. */
/*  RSRC_A (global) DESCA( RSRC_ ) The process row over which the first */
/*                                 row of the array A is distributed. */
/*  CSRC_A (global) DESCA( CSRC_ ) The process column over which the */
/*                                 first column of the array A is */
/*                                 distributed. */
/*  LLD_A  (local)  DESCA( LLD_ )  The leading dimension of the local */
/*                                 array.  LLD_A >= MAX(1,LOCr(M_A)). */

/*  Let K be the number of rows or columns of a distributed matrix, */
/*  and assume that its process grid has dimension p x q. */
/*  LOCr( K ) denotes the number of elements of K that a process */
/*  would receive if K were distributed over the p processes of its */
/*  process column. */
/*  Similarly, LOCc( K ) denotes the number of elements of K that a */
/*  process would receive if K were distributed over the q processes of */
/*  its process row. */
/*  The values of LOCr() and LOCc() may be determined via a call to the */
/*  ScaLAPACK tool function, NUMROC: */
/*          LOCr( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ), */
/*          LOCc( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ). */
/*  An upper bound for these quantities may be computed by: */
/*          LOCr( M ) <= ceil( ceil(M/MB_A)/NPROW )*MB_A */
/*          LOCc( N ) <= ceil( ceil(N/NB_A)/NPCOL )*NB_A */

/*  This routine requires N <= NB_A-MOD(JA-1, NB_A) and square block */
/*  decomposition ( MB_A = NB_A ). */

/*  Arguments */
/*  ========= */

/*  UPLO    (global input) CHARACTER */
/*          = 'U':  Upper triangle of sub( A ) is stored; */
/*          = 'L':  Lower triangle of sub( A ) is stored. */

/*  N       (global input) INTEGER */
/*          The number of rows and columns to be operated on, i.e. the */
/*          order of the distributed submatrix sub( A ). N >= 0. */

/*  IA      (global input) INTEGER */
/*          The row index in the global array A indicating the first */
/*          row of sub( A ). */

/*  JA      (global input) INTEGER */
/*          The column index in the global array A indicating the */
/*          first column of sub( A ). */

/*  DESCA   (global and local input) INTEGER array of dimension DLEN_. */
/*          The array descriptor for the distributed matrix A. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. Executable Statements .. */

    char UPLOC    = UPLOI, *uplo = &UPLOC;
    int     *n      = &N;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;

    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;

    /* Parameter adjustments */
    --desca;

/*     Test the input parameters. */

    if (init_llds_f == 1)
    {
    HSCAL_PXPOTF2_INIT_LLDS(
         (desca+1),
         myrow, nprow
    );
    }

    *info = 0;
    if (nprow == -1) {
	*info = -602;
    } else {
	hscal_chk1mat(n, &c__2, n, &c__2, ia, ja, &desca[1], &c__6, info,
                    myrow, mycol, nprow, npcol
      );
	if (*info == 0) {
	    upper = hscal_lsame(uplo, "U");
	    iroff = (*ia - 1) % desca[5];
	    icoff = (*ja - 1) % desca[6];
	    if (! upper && ! hscal_lsame(uplo, "L")) {
		*info = -1;
	    } else if (*n + icoff > desca[6]) {
		*info = -2;
	    } else if (iroff != 0) {
		*info = -4;
	    } else if (icoff != 0) {
		*info = -5;
	    } else if (desca[5] != desca[6]) {
		*info = -606;
	    }
	}
    }

    if (*info != 0) {
        i__1 = -(*info);
        hscal_PB_Cabort(
             "PCPOTF2",
             i__1,
             nprow, npcol, myrow, mycol
        );
        return HSCAL_ERR_INPUT;
    }

/*     Quick return if possible */

    if (*n == 0) {
	return HSCAL_SUCCESS;
    }

/*     Compute local information */

    infog2l_(ia, ja, &desca[1], &nprow, &npcol, &myrow, &mycol, &iia, &jja, &
             iarow, &iacol);
    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);

    if (upper) {

/*        Process (IAROW, IACOL) owns block to be factorized */

          {
            {

/*              Compute the Cholesky factorization A = U'*U. */

		i__1 = *ja + *n - 1;
		for (j = *ja; j <= i__1; ++j) {

/*                 Compute U(J,J) and test for non-positive-definiteness. */

                if (myrow == iarow) {
                if (mycol == iacol) {

           	    lda = desca[9];
		    i__3 = j - *ja;

		    cdotc_tcomp_macro(
                    &i__3, &c__1, &c__1,
                    nprow, npcol, myrow, mycol, tcomp_l
                );

                }
                }

/*                 Compute elements J+1:JA+N-1 of row J. */

		    if (j < *ja + *n - 1) {
                  if (myrow == iarow) {
                  if (mycol == iacol) {

                  lda = desca[9];
			i__2 = j - *ja;
			clacgv_tcomp_macro(
                        &i__2, &c__1, 
                        nprow, npcol, myrow, mycol, tcomp_l
                  );

                  }
                  }

                  if (myrow == iarow) {
                  if (mycol == iacol) {

                  lda = desca[9];
			i__2 = j - *ja;
			i__3 = *ja + *n - j - 1;
			cgemv_tcomp_macro(
                       "Transpose", &i__2, &i__3, (char*)&c_b19, &lda, &c__1, (char*)&c_b21, 
			     &lda, 
                       nprow, npcol, myrow, mycol, tcomp_l
                  );

                  }
                  }

                  if (myrow == iarow) {
                  if (mycol == iacol) {

			i__2 = j - *ja;
			clacgv_tcomp_macro(
                        &i__2, &c__1,
                        nprow, npcol, myrow, mycol, tcomp_l
                  );

                  }
                  }

                  if (myrow == iarow) {
                  if (mycol == iacol) {

                  lda = desca[9];
			i__2 = *ja + *n - j - 1;
                        r__1 = 0.5f;
			csscal_tcomp_macro(
                        &i__2, (char*)&r__1, &lda,
                       nprow, npcol, myrow, mycol, tcomp_l
                  );

                  }
                  }
		    }
		}

/*
* Ignoring the broadcast of 'info'
* for performance model
*/
	    }
	}
    } else {

/*        Process (IAROW, IACOL) owns block to be factorized */

	{
	    {

/*              Compute the Cholesky factorization A = L*L'. */

		i__1 = *ja + *n - 1;
		for (j = *ja; j <= i__1; ++j) {

/*                 Compute L(J,J) and test for non-positive-definiteness. */

                if (mycol == iacol) {
                if (myrow == iarow) {

           	    lda = desca[9];
		    i__2 = idiag;
		    i__3 = j - *ja;
		    cdotc_tcomp_macro(
                     &i__3, &lda, &lda,
                     nprow, npcol, myrow, mycol, tcomp_l
                );

                }
                }

/*                 Compute elements J+1:JA+N-1 of column J. */

		    if (j < *ja + *n - 1) {

                  if (mycol == iacol) {
                  if (myrow == iarow) {

                  lda = desca[9];
			i__2 = j - *ja;
			clacgv_tcomp_macro(
                        &i__2, &lda,
                      nprow, npcol, myrow, mycol, tcomp_l
                  );

                  }
                  }

                  if (mycol == iacol) {
                  if (myrow == iarow) {

                  lda = desca[9];
			i__2 = *ja + *n - j - 1;
			i__3 = j - *ja;
			cgemv_tcomp_macro(
                       "No transpose", &i__2, &i__3, (char*)&c_b19, 
			     &lda, &lda, (char*)&c_b21, &c__1,
                      nprow, npcol, myrow, mycol, tcomp_l
                  );

                  }
                  }

                  if (mycol == iacol) {
                  if (myrow == iarow) {

                  lda = desca[9];
			i__2 = j - *ja;
			clacgv_tcomp_macro(
                        &i__2, &lda,
                        nprow, npcol, myrow, mycol, tcomp_l
                  );

                  }
                  }

                  if (mycol == iacol) {
                  if (myrow == iarow) {

			i__2 = *ja + *n - j - 1;
                        r__1 = 0.5f;
			csscal_tcomp_macro(
                        &i__2, (char*)&r__1, &c__1,
                        nprow, npcol, myrow, mycol, tcomp_l
                  );

                  }
                  }

		    }
		}
/*
* Ignoring the broadcast of 'info'
* for performance model
*/
	    }
	}
    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcpotf2_tcomp(
   int UPLOI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = ihscal_pcpotf2_tcomp
              (
                1,
                myrow, mycol,
                UPLOI, N,
                IA, JA, DESCA,
                nprow, npcol,
                tcomp
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcpotf2_info(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = ihscal_pcpotf2_tcomp
              (
                1,
                myrow, mycol,
                *UPLO, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcpotf2_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = ihscal_pcpotf2_tcomp
              (
                1,
                myrow, mycol,
                *UPLO, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
