
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pddbsv_recon_tcomp(
   int N,
   int BWL, 
   int BWU,
   int NRHS
)
{
    int max_bw__ = max(BWL, BWU);

    double tcomp = 0;

    char   transac  = 'T';
    char   *transa  = &transac;

    char   transbc  = 'N';
    char   *transb  = &transbc;

    int reconm, reconn, reconk;

    int    lda, ldb, ldc;

    double alpha, beta;

    alpha = -1.f;
    beta  = 1.f;

    reconm  = max_bw__;
    reconn  = max_bw__;
    reconk  = max_bw__;

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_dgemm(
         transa,
         transb,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha),
         &lda,
         &ldb,
         ((char*)&beta),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pddbsv_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pddbsv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, BWL, BWU, NRHS,
          JA, DESCA,
          IB, DESCB,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pddbsv_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pddbsv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, BWL, BWU, NRHS,
          JA, DESCA,
          IB, DESCB,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbsv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pddbsv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, BWL, BWU, NRHS,
          JA, DESCA,
          IB, DESCB,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pddbsv_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    /* System generated locals */
    int i__1, i__2;

    /* Local variables */
    int ws_factor__, nb;

    int     *n      = (int*)&N;
    int     *bwl    = (int*)&BWL;
    int     *bwu    = (int*)&BWU;
    int     *nrhs   = (int*)&NRHS;
    int     *ja     = (int*)&JA;
    int     *desca  = (int*)DESCA;
    int     *ib     = (int*)&IB;
    int     *descb  = (int*)DESCB;

    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descb;
    --desca;

    /* Function Body */
    *info = 0;

/*     Get block size to calculate workspace requirements */

    if (desca[1] == 1) {
	nb = desca[6];
    } else if (desca[1] == 501) {
	nb = desca[4];
    } else {
	*info = -601;
	i__1 = -(*info);
      hscal_PB_Cabort( 
           "PDDBSV", 
           i__1,
           nprow, npcol, 0, 0
      ); 
      return HSCAL_ERR_INPUT;
    }

/*     Test the input parameters */
    if (init_llds_f == 1)
    {
       int *PXDBSV_DESCA_l = &desca[1];
       int *PXDBSV_DESCB_l = &descb[1];
       HSCAL_PXDBSV_INIT_LLDS(
            PXDBSV_DESCA_l,
            PXDBSV_DESCB_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PDDBSV: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d", 
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PDDBSV: (%d, %d): (descb+1)[M1_]=%d (descb+1)[MB1_]=%d myrow=%d (descb+1)[RSRC1_]=%d nprow=%d (descb+1)[LLD1_]=%d",
        myrow, mycol, (descb+1)[M1_], (descb+1)[MB1_], myrow, (descb+1)[RSRC1_], nprow, (descb+1)[LLD1_]
    );

/*     Size needed for AF in factorization */

    ws_factor__ = nb * (*bwl + *bwu) + max(*bwl,*bwu) * 6 * max(*bwl,*bwu);

/*     Factor the matrix */

    {
    int *PDDBTRF_DESCA_l = (&desca[1]);
    hscal_pddbtrf_my_tcomp_tcomm(
           0,
           myrow, mycol, 
           N, BWL, BWU, JA, PDDBTRF_DESCA_l,
           nprow, npcol,
           tcomp, tcomm
    );
    }

/*     Solve the system using the factorization */

    {
    int PDDBTRS_TRANS_l = 'N';
    int *PDDBTRS_DESCA_l = (&desca[1]);
    int *PDDBTRS_DESCB_l = (&descb[1]);

    hscal_pddbtrs_my_tcomp_tcomm(
           0,
           myrow, mycol, 
           PDDBTRS_TRANS_l, N, BWL, BWU, NRHS, JA, PDDBTRS_DESCA_l, IB, PDDBTRS_DESCB_l, 
           nprow, npcol,
           tcomp, tcomm
    );
    }

/*
*  End of PDDBSV
*/

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbsv_tcomp_tcomm(
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pddbsv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N, BWL, BWU, NRHS,
                JA, DESCA,
                IB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbsv_info(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pddbsv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, *BWL, *BWU, *NRHS,
                *JA, DESCA,
                *IB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbsv_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pddbsv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, *BWL, *BWU, *NRHS,
                *JA, DESCA,
                *IB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
