
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pddbtrf_recon_tcomp(
   int N,
   int BWL, 
   int BWU
)
{
    int max_bw__ = max(BWL, BWU);

    double tcomp = 0;

    char   transac  = 'T';
    char   *transa  = &transac;

    char   transbc  = 'N';
    char   *transb  = &transbc;

    int reconm, reconn, reconk;

    int    lda, ldb, ldc;

    double alpha, beta;

    alpha = -1.;
    beta = 1.;

    reconm  = max_bw__;
    reconn  = max_bw__;
    reconk  = max_bw__;

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_dgemm(
         transa,
         transb,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha),
         &lda,
         &ldb,
         ((char*)&beta),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pddbtrf_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int BWL, 
   int BWU,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pddbtrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, BWL, BWU,
          JA, DESCA,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrf_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int BWL, 
   int BWU,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pddbtrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, BWL, BWU,
          JA, DESCA,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrf_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int BWL, 
   int BWU,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pddbtrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N, BWL, BWU,
          JA, DESCA,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pddbtrf_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int BWL, 
   int BWU,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__9 = 9;
    hscalinteger c__1 = 1;
    hscalinteger c__0 = 0;
    hscaldoublereal c_b38 = -1.;
    hscaldoublereal c_b39 = 1.;
    hscaldoublereal c_b62 = 0.;
    hscalinteger c_n1 = -1;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4, i__5, i__6;

    /* Local variables */
    hscalinteger prev_tri_size_m__, prev_tri_size_n__, next_tri_size_m__, 
	    next_tri_size_n__, odd_size__, desca_1xp__[7], i__, store_n_a__, 
	    i1, i2, comm_proc__, part_size__, ictxt_new__, nb, np, 
	    level_dist__, ictxt_save__, first_proc__, param_check__[27]	/* 
	    was [9][3] */, up_prev_tri_size_m__, up_prev_tri_size_n__, mbw2, 
	    return_code__, part_offset__, my_num_cols__, llda, csrc, ofst;
    hscalinteger idum3;
    hscalinteger ictxt, work_size_min__, ja_new__;
    hscalinteger max_bw__;
    hscalinteger work_u__;
    hscalinteger laf_min__;
    hscalinteger np_save__;

    HSCALBLACSCONTEXT   * blacsctxt;

    int     *n      = (int*)&N;
    int     *bwl    = (int*)&BWL;
    int     *bwu    = (int*)&BWU;
    int     *ja     = (int*)&JA;
    int     *desca  = (int*)DESCA;
    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

/*     Test the input parameters */

    /* Parameter adjustments */
    --desca;

    if (init_llds_f == 1)
    {
       HSCAL_PXDBTRF_INIT_LLDS(
            (desca+1),
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PSDBTRF: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d",
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    /* Function Body */
    *info = 0;

/*     Convert descriptor into standard form for easy access to */
/*        parameters, check that grid is of right shape. */

    desca_1xp__[0] = 501;

    desc_convert__(&desca[1], desca_1xp__, &return_code__);

    if (return_code__ != 0) {
	*info = -602;
    }

/*     Get values out of descriptor for use in code. */

    ictxt = desca_1xp__[1];
    csrc = desca_1xp__[4];
    nb = desca_1xp__[3];
    llda = desca_1xp__[5];
    store_n_a__ = desca_1xp__[2];

/*     Get grid parameters */


/*     Size of separator blocks is maximum of bandwidths */

    max_bw__ = max(*bwl,*bwu);
    mbw2 = max_bw__ * max_bw__;

    np = nprow * npcol;



    idum3 = 1;

    if (*n < 0) {
	*info = -1;
    }

    if (*n + *ja - 1 > store_n_a__) {
	*info = -606;
    }

    if (*bwl > *n - 1 || *bwl < 0) {
	*info = -2;
    }

    if (*bwu > *n - 1 || *bwu < 0) {
	*info = -3;
    }

    if (llda < *bwl + *bwu + 1) {
	*info = -606;
    }

    if (nb <= 0) {
	*info = -604;
    }

/*     Argument checking that is specific to Divide & Conquer routine */

    if (nprow != 1) {
	*info = -602;
    }

    if (*n > np * nb - (*ja - 1) % nb) {
	*info = -1;
	i__1 = -(*info);
      hscal_PB_Cabort(
           "PDDBTRF, D&C alg.: only 1 block per proc",
           i__1,
           nprow, npcol, myrow, mycol
      );
      return HSCAL_ERR_INPUT;
    }

    if (*ja + *n - 1 > nb && nb < max(*bwl,*bwu) << 1) {
	*info = -604;
	i__1 = -(*info);
      hscal_PB_Cabort(
           "PDDBTRF, D&C alg.: NB too small",
           i__1,
           nprow, npcol, myrow, mycol
      );
      return HSCAL_ERR_INPUT;
    }


/*     Check auxiliary storage size */

    laf_min__ = nb * (*bwl + *bwu) + max(*bwl,*bwu) * 6 * max(*bwl,*bwu);

/*     Check worksize */

    work_size_min__ = max(*bwl,*bwu) * max(*bwl,*bwu);

/*     Pack params and positions into arrays for global consistency check */

    param_check__[8] = desca[5];
    param_check__[7] = desca[4];
    param_check__[6] = desca[3];
    param_check__[5] = desca[1];
    param_check__[4] = *ja;
    param_check__[3] = *bwu;
    param_check__[2] = *bwl;
    param_check__[1] = *n;
    param_check__[0] = idum3;

    param_check__[17] = 605;
    param_check__[16] = 604;
    param_check__[15] = 603;
    param_check__[14] = 601;
    param_check__[13] = 5;
    param_check__[12] = 3;
    param_check__[11] = 2;
    param_check__[10] = 1;
    param_check__[9] = 10;

/*     Want to find errors with MIN( ), so if no error, set it to a big */
/*     number. If there already is an error, multiply by the the */
/*     descriptor multiplier. */

    if (*info >= 0) {
	*info = 10000;
    } else if (*info < -100) {
	*info = -(*info);
    } else {
	*info = -(*info) * 100;
    }

/*     Check consistency across processors */

    hscal_globchk(&ictxt, &c__9, param_check__, &c__9, &param_check__[18], info,
                  &myrow, &mycol, &nprow, &npcol
    );

/*     Prepare output: set info = 0 if no error, and divide by DESCMULT */
/*     if error is not in a descriptor entry. */

    if (*info == 10000) {
	*info = 0;
    } else if (*info % 100 == 0) {
	*info = -(*info) / 100;
    } else {
	*info = -(*info);
    }

    if (*info < 0) {
	i__1 = -(*info);
      hscal_PB_Cabort(
            "PDDBTRF",
            i__1,
            nprow, npcol, myrow, mycol
      );
	return HSCAL_ERR_INPUT;
    }

/*     Quick return if possible */

    if (*n == 0) {
	return HSCAL_SUCCESS;
    }


/*     Adjust addressing into matrix space to properly get into */
/*        the beginning part of the relevant data */

    part_offset__ = nb * ((*ja - 1) / (npcol * nb));

    if (mycol - csrc < (*ja - part_offset__ - 1) / nb) {
	part_offset__ += nb;
    }

    if (mycol < csrc) {
	part_offset__ -= nb;
    }

/*     Form a new BLACS grid (the "standard form" grid) with only procs */
/*        holding part of the matrix, of size 1xNP where NP is adjusted, */
/*        starting at csrc=0, with JA modified to reflect dropped procs. */

/*     First processor to hold part of the matrix: */

    first_proc__ = ((*ja - 1) / nb + csrc) % npcol;

/*     Calculate new JA one while dropping off unused processors. */

    ja_new__ = (*ja - 1) % nb + 1;

/*     Save and compute new value of NP */

    np_save__ = np;
    np = (ja_new__ + *n - 2) / nb + 1;

/*     Call utility routine that forms "standard-form" grid */
/*
*    reshape_(&ictxt, &c__1, &ictxt_new__, &c__1, &first_proc__, &c__1, &np);
*/
/*     Use new context from standard grid as context. */
/*
*    ictxt_save__ = ictxt;
*    ictxt = ictxt_new__;
*    desca_1xp__[1] = ictxt_new__;
*/
/*     Get information about new grid. */
/*
*    blacs_gridinfo__(&ictxt, &nprow, &npcol, &myrow, &mycol);
*/
/*     Drop out processors that do not have part of the matrix. */
/*
*
*    if (myrow < 0) {
*	goto L140;
*    }
*/

/*
* The new grid is of the shape 1*np, therefore use it.
*/
nprow = 1;
npcol = np;

if ((mycol > (np - 1)))
{
   return HSCAL_SUCCESS;
}

/*     ******************************** */
/*     Values reused throughout routine */

/*     User-input value of partition size */

    part_size__ = nb;

/*     Number of columns in each processor */

    my_num_cols__ = numroc_(n, &part_size__, &mycol, &c__0, &npcol);

/*     Offset in columns to beginning of main partition in each proc */

    if (mycol == 0) {
	part_offset__ += (ja_new__ - 1) % part_size__;
	my_num_cols__ -= (ja_new__ - 1) % part_size__;
    }

/*     Offset in elements */

    ofst = part_offset__ * llda;

/*     Size of main (or odd) partition in each processor */

    odd_size__ = my_num_cols__;
    if (mycol < np - 1) {
	odd_size__ -= max_bw__;
    }

/*     Offset to workspace for Upper triangular factor */

    work_u__ = *bwu * odd_size__ + mbw2 * 3;


/*       Zero out space for fillin */

    i__1 = laf_min__;
    if (tcomp_l != NULL)
    {
       TCOMPL((myrow), (mycol), (nprow), (npcol)) += i__1;
    }

/*       Zero out space for work */

    i__1 = work_size_min__;
    if (tcomp_l != NULL)
    {
       TCOMPL((myrow), (mycol), (nprow), (npcol)) += i__1;
    }

/*     Begin main code */


/* ******************************************************************* */
/*       PHASE 1: Local computation phase. */
/* ******************************************************************* */


/*       Sizes of the extra triangles communicated bewtween processors */

    if (mycol > 0) {
/* Computing MIN */
	i__1 = *bwl, i__2 = numroc_(n, &part_size__, &mycol, &c__0, &npcol);
	prev_tri_size_m__ = min(i__1,i__2);
/* Computing MIN */
	i__3 = mycol - 1;
	i__1 = *bwl, i__2 = numroc_(n, &part_size__, &i__3, &c__0, &npcol);
	prev_tri_size_n__ = min(i__1,i__2);
    }

    if (mycol > 0) {
/* Computing MIN */
	i__1 = *bwu, i__2 = numroc_(n, &part_size__, &mycol, &c__0, &npcol);
	up_prev_tri_size_m__ = min(i__1,i__2);
/* Computing MIN */
	i__3 = mycol - 1;
	i__1 = *bwu, i__2 = numroc_(n, &part_size__, &i__3, &c__0, &npcol);
	up_prev_tri_size_n__ = min(i__1,i__2);
    }

    if (mycol < npcol - 1) {
/* Computing MIN */
	i__3 = mycol + 1;
	i__1 = *bwl, i__2 = numroc_(n, &part_size__, &i__3, &c__0, &npcol);
	next_tri_size_m__ = min(i__1,i__2);
/* Computing MIN */
	i__1 = *bwl, i__2 = numroc_(n, &part_size__, &mycol, &c__0, &npcol);
	next_tri_size_n__ = min(i__1,i__2);
    }

/*
*  Retrieve process grid information
*/
    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

    if (mycol < np - 1) {
/*         Transfer last triangle D_i of local matrix to next processor */
/*         which needs it to calculate fillin due to factorization of */
/*         its main (odd) block A_i. */
/*         Overlap the send with the factorization of A_i. */

	i__1 = llda - 1;
	i__2 = mycol + 1;
	hscal_Cdtrsd2d(
           blacsctxt, "U", "N", next_tri_size_m__, next_tri_size_n__, 
           i__1, c__0, i__2, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);

    }


/*       Factor main partition A_i = L_i {U_i} in each processor */

    ddbtrf_tcomp_macro(
          &odd_size__, &odd_size__, bwl, bwu, &llda, info,
          nprow, npcol, myrow, mycol, tcomp_l
    );

    if (mycol < np - 1) {

/*         Apply factorization to lower connection block BL_i */
/*         transpose the connection block in preparation. */
/*         Apply factorization to upper connection block BU_i */
/*         Move the connection block in preparation. */

	i__1 = llda - 1;
	dlatcpy_tcomp_macro(
             "U", bwl, bwl, &i__1, &max_bw__, 
             nprow, npcol, myrow, mycol, tcomp_l
      );
	i__1 = llda - 1;
	dlacpy_tcomp_macro(
            "L", bwu, bwu, &i__1, &max_bw__, nprow, npcol, myrow, mycol, tcomp_l);

/*         Perform the triangular system solve {L_i}{{BU'}_i} = {B_i} */

	dtbtrs_tcomp_macro(
            "L", "N", "U", bwu, bwl, bwu, &llda, (&max_bw__), info, 
            nprow, npcol, myrow, mycol, tcomp_l
      );

/*         Perform the triangular solve {U_i}^T{BL'}_i^T = {BL_i}^T */

	dtbtrs_tcomp_macro(
            "U", "T", "N", bwl, bwu, bwl, &llda, &max_bw__, info, 
            nprow, npcol, myrow, mycol, tcomp_l
      );

/*         transpose resulting block to its location */
/*           in main storage. */

	i__1 = llda - 1;
	dlatcpy_tcomp_macro(
            "L", bwl, bwl, &max_bw__, &i__1, nprow, npcol, myrow, mycol, tcomp_l);

/*         Move the resulting block back to its location in main storage. */

	i__1 = llda - 1;
	dlacpy_tcomp_macro(
            "L", bwu, bwu, &max_bw__, &i__1, nprow, npcol, myrow, mycol, tcomp_l);


/*         Compute contribution to diagonal block(s) of reduced system. */
/*          {C'}_i = {C_i}-{{BL'}_i}{{BU'}_i} */

/*         The following method uses more flops than necessary but */
/*           does not necessitate the writing of a new BLAS routine. */


	i__1 = llda - 1;
	dgemm_tcomp_macro(
            "T", "N", &max_bw__, &max_bw__, &max_bw__, (char*)&c_b38, &max_bw__, 
            &max_bw__, (char*)&c_b39, &i__1, nprow, npcol, myrow, mycol, tcomp_l);

    }
/*       End of "if ( MYCOL .lt. NP-1 )..." loop */


L30:
/*       If the processor could not locally factor, it jumps here. */

    if (mycol != 0) {
/*         Discard temporary matrix stored beginning in */
/*           AF( (odd_size+2*bwl, bwu)*bwl, bwu+1 ) and use for */
/*           off_diagonal block of reduced system. */

/*         Receive previously transmitted matrix section, which forms */
/*         the right-hand-side for the triangular solve that calculates */
/*         the "spike" fillin. */


	i__1 = mycol - 1;

	{

/*         Calculate the "spike" fillin, ${L_i} {{GU}_i} = {DL_i}$ . */

/*         Transpose transmitted triangular matrix  $DL_i$ */

	    i__1 = *bwl;
          if (tcomp_l != NULL)
          {
             TCOMPL((myrow), (mycol), (nprow), (npcol)) += ((i__1*(i__1 - 1)/2)*2);
          }       

	    i__1 = odd_size__;
	    for (i1 = 2; i1 <= i__1; ++i1) {
/* Computing MIN */
		i__2 = i1 - 1;
		i2 = min(i__2,*bwl);
		i__2 = llda - 1;
		dgemv_tcomp_macro(
                 "N", bwl, &i2, (char*)&c_b38, bwl, &i__2, (char*)&c_b39, &c__1, 
                 nprow, npcol, myrow, mycol, tcomp_l
            );
/* L60: */
	    }


/*         Calculate the "spike" fillin, ${U_i}^T {{GL}_i}^T = {DU_i}^T$ */


/*         Copy D block into AF storage for solve. */

	    i__1 = llda - 1;
	    dlacpy_tcomp_macro(
                "L", &up_prev_tri_size_n__, &up_prev_tri_size_m__, &i__1, bwu, 
                nprow, npcol, myrow, mycol, tcomp_l
          );

	    i__1 = odd_size__;
	    for (i1 = 1; i1 <= i__1; ++i1) {
/* Computing MIN */
		i__2 = *bwu, i__3 = i1 - 1;
		i2 = min(i__2,i__3);
		dgemv_tcomp_macro(
                 "N", bwu, &i2, (char*)&c_b38,  
			bwu, &c__1, (char*)&c_b39, &c__1, 
                 nprow, npcol, myrow, mycol, tcomp_l
            );

		i__2 = *bwu;
            if (tcomp_l != NULL)
            {
               TCOMPL((myrow), (mycol), (nprow), (npcol)) += i__2;
            }
/* L80: */
	    }

/*         Calculate the update block for previous proc, E_i = GL_i{GU_i} */


/*         Zero out space in case result is smaller than storage block */

	    i__1 = mbw2;
          if (tcomp_l != NULL)
          {
             TCOMPL((myrow), (mycol), (nprow), (npcol)) += i__1;
          }

/* Computing MAX */
	    i__1 = 0, i__2 = *bwl - *bwu;
/* Computing MAX */
	    i__3 = 0, i__4 = *bwu - *bwl;
	    dgemm_tcomp_macro(
                "N", "T", bwu, bwl, &odd_size__, (char*)&c_b38, bwu, bwl, 
                (char*)&c_b62, &max_bw__, 
                nprow, npcol, myrow, mycol, tcomp_l
          );


/*         Initiate send of E_i to previous processor to overlap */
/*           with next computation. */

	    i__1 = mycol - 1;
	    hscal_Cdgesd2d(
               blacsctxt, max_bw__, max_bw__, max_bw__, c__0, i__1,
               nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
          );


	    if (mycol < np - 1) {

/*           Calculate off-diagonal block(s) of reduced system. */
/*           Note: for ease of use in solution of reduced system, store */
/*           L's off-diagonal block in transpose form. */

/*           Copy matrix HU_i (the last bwl rows of GU_i) to AFL storage */
/*             as per requirements of BLAS routine DTRMM. */
/*             Since we have GU_i stored, */
/*             transpose HU_i to HU_i^T. */

		dlacpy_tcomp_macro(
                  "N", bwl, bwl, bwl, &max_bw__, 
                  nprow, npcol, myrow, mycol, tcomp_l
            );

		i__1 = llda - 1;
		dtrmm_tcomp_macro(
                 "R", "U", "T", "N", bwl, bwl, (char*)&c_b38, &i__1, &max_bw__,
		     nprow, npcol, myrow, mycol, tcomp_l);


/*           Copy matrix HL_i (the last bwu rows of GL_i^T) to AFU store */
/*             as per requirements of BLAS routine DTRMM. */
/*             Since we have GL_i^T stored, */
/*             transpose HL_i^T to HL_i. */

		dlacpy_tcomp_macro(
                  "N", bwu, bwu, 
			bwu, &max_bw__, 
                  nprow, npcol, myrow, mycol, tcomp_l
            );

		i__1 = llda - 1;
		dtrmm_tcomp_macro(
                  "R", "L", "N", "N", bwu, bwu, (char*)&c_b38, &i__1, &max_bw__, 
			nprow, npcol, myrow, mycol, tcomp_l);

	    }

	}
/*       End of "if ( MYCOL .ne. 0 )..." */

    }
/*       End of "if (info.eq.0) then" */


/*       Check to make sure no processors have found errors */
/*
* Broadcast of 'info' ignored
*/

/*       No errors found, continue */


/* ******************************************************************* */
/*       PHASE 2: Formation and factorization of Reduced System. */
/* ******************************************************************* */

/*       Gather up local sections of reduced system */


/*     The last processor does not participate in the factorization of */
/*       the reduced system, having sent its E_i already. */
    if (mycol == npcol - 1) {
	goto L120;
    }

/*       Initiate send of off-diag block(s) to overlap with next part. */
/*       Off-diagonal block needed on neighboring processor to start */
/*       algorithm. */

    if ((mycol + 1) % 2 == 0 && mycol > 0) {

	i__1 = mycol - 1;
	hscal_Cdgesd2d(
             blacsctxt, max_bw__, max_bw__, max_bw__, c__0, i__1,
             nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );

	i__1 = mycol - 1;
	hscal_Cdgesd2d(
           blacsctxt, max_bw__, max_bw__, max_bw__, c__0, i__1,
            nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );

    }

/*       Copy last diagonal block into AF storage for subsequent */
/*         operations. */

    i__1 = llda - 1;
    dlacpy_tcomp_macro(
          "N", &max_bw__, &max_bw__, 
	    &i__1, &max_bw__, 
          nprow, npcol, myrow, mycol, tcomp_l
    );

/*       Receive cont. to diagonal block that is stored on this proc. */

    if (mycol < npcol - 1) {

	i__1 = mycol + 1;

/*          Add contribution to diagonal block */

	daxpy_tcomp_macro(
           &mbw2, (char*)&c_b39, &c__1,
	     &c__1,
           nprow, npcol, myrow, mycol, tcomp_l
      );

    }


/*       ************************************* */
/*       Modification Loop */

/*       The distance for sending and receiving for each level starts */
/*         at 1 for the first level. */
    level_dist__ = 1;

/*       Do until this proc is needed to modify other procs' equations */

L100:
    if ((mycol + 1) / level_dist__ % 2 != 0) {
	goto L110;
    }

/*         Receive and add contribution to diagonal block from the left */

    if (mycol - level_dist__ >= 0) {
	i__1 = mycol - level_dist__;

	daxpy_tcomp_macro(
           &mbw2, (char*)&c_b39, &c__1, &c__1,
           nprow, npcol, myrow, mycol, tcomp_l
      );

    }

/*         Receive and add contribution to diagonal block from the right */

    if (mycol + level_dist__ < npcol - 1) {
	i__1 = mycol + level_dist__;

	daxpy_tcomp_macro(
           &mbw2, (char*)&c_b39, &c__1, &c__1,
           nprow, npcol, myrow, mycol, tcomp_l
      );

    }

    level_dist__ <<= 1;

    goto L100;
L110:
/*       [End of GOTO Loop] */


/*       ********************************* */
/*       Calculate and use this proc's blocks to modify other procs'... */

/*       Factor diagonal block */

/* Computing MIN */
    i__2 = max_bw__ - 1;
    i__1 = min(i__2,*bwl);
/* Computing MIN */
    i__4 = max_bw__ - 1;
    i__3 = min(i__4,*bwu);
/* Computing MIN */
    i__5 = max_bw__ - 1;
    i__6 = max_bw__ + 1;
    ddbtrf_tcomp_macro(
          &max_bw__, &max_bw__, &i__1, &i__3, &i__6, info,
          nprow, npcol, myrow, mycol, tcomp_l
    );

/*       **************************************************************** */
/*       Receive offdiagonal block from processor to right. */
/*         If this is the first group of processors, the receive comes */
/*         from a different processor than otherwise. */

    if (level_dist__ == 1) {
	comm_proc__ = mycol + 1;

/*           Move block into place that it will be expected to be for */
/*             calcs. */

	dlacpy_tcomp_macro("N", &max_bw__, &max_bw__, 
            &max_bw__, 
            &max_bw__,
            nprow, npcol, myrow, mycol, tcomp_l
      );

	dlacpy_tcomp_macro(
            "N", &max_bw__, &max_bw__, &max_bw__, &max_bw__,
            nprow, npcol, myrow, mycol, tcomp_l
      );

    } else {
	comm_proc__ = mycol + level_dist__ / 2;
    }

    if (mycol / level_dist__ <= (npcol - 1) / level_dist__ - 2) {

	{


/*         Modify upper off_diagonal block with diagonal block */


/* Computing MIN */
	    i__2 = *bwl, i__3 = *bwu - 1;
	    i__1 = min(i__2,i__3);
	    i__4 = max_bw__ + 1;
	    dtbtrs_tcomp_macro(
                "L", "N", "U", bwu, &i__1, bwu, &i__4, 
                &max_bw__, info, nprow, npcol, myrow, mycol, tcomp_l);

/*         Modify lower off_diagonal block with diagonal block */


/* Computing MIN */
	    i__2 = *bwu, i__3 = *bwl - 1;
	    i__1 = min(i__2,i__3);
/* Computing MIN */
	    i__4 = *bwu, i__5 = *bwl - 1;
	    i__6 = max_bw__ + 1;
	    dtbtrs_tcomp_macro(
                "U", "T", "N", bwl, &i__1, bwl, &i__6, 
		    &max_bw__, info, nprow, npcol, myrow, mycol, tcomp_l);

	}
/*         End of "if ( info.eq.0 ) then" */

/*         Calculate contribution from this block to next diagonal block */

	dgemm_tcomp_macro(
           "T", "N", &max_bw__, &max_bw__, &max_bw__, (char*)&c_b38, &max_bw__, 
           &max_bw__, (char*)&c_b62, &max_bw__, nprow, npcol, myrow, mycol, tcomp_l);

/*         Send contribution to diagonal block's owning processor. */

	i__1 = mycol + level_dist__;
	hscal_Cdgesd2d(
           blacsctxt, max_bw__, max_bw__, max_bw__, c__0, i__1,
           nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
      );

    }
/*       End of "if( mycol/level_dist .le. (npcol-1)/level_dist-2 )..." */


/*       **************************************************************** */
/*       Receive off_diagonal block from left and use to finish with this */
/*         processor. */

    if (mycol / level_dist__ > 0 && mycol / level_dist__ <= (npcol - 1) / 
	    level_dist__ - 1) {

	if (level_dist__ > 1) {

/*           Receive offdiagonal block(s) from proc level_dist/2 to the */
/*           left */

	    i__1 = mycol - level_dist__ / 2;

/*           Receive offdiagonal block(s) from proc level_dist/2 to the */
/*           left */

	    i__1 = mycol - level_dist__ / 2;

	}


	{

/*         Use diagonal block(s) to modify this offdiagonal block */


/*         Since DTBTRS has no "left-right" option, we must transpose */

	    dlatcpy_tcomp_macro(
                "N", &max_bw__, &max_bw__, &max_bw__, &max_bw__, 
                nprow, npcol, myrow, mycol, tcomp_l
          );

/* Computing MIN */
	    i__2 = *bwl, i__3 = max_bw__ - 1;
	    i__1 = min(i__2,i__3);
	    i__4 = max_bw__ + 1;
	    dtbtrs_tcomp_macro(
                "L", "N", "U", &max_bw__, &i__1, bwl, &i__4, &max_bw__, info, 
                nprow, npcol, myrow, mycol, tcomp_l
          );

/*         Transpose back */

	    dlatcpy_tcomp_macro(
                "N", &max_bw__, &max_bw__, &max_bw__, 
                &max_bw__, nprow, npcol, myrow, mycol, tcomp_l);



/*         Since DTBTRS has no "left-right" option, we must transpose */

	    dlatcpy_tcomp_macro(
                 "N", &max_bw__, &max_bw__, &max_bw__, &max_bw__,
                 nprow, npcol, myrow, mycol, tcomp_l
          );

/* Computing MIN */
	    i__2 = *bwu, i__3 = max_bw__ - 1;
	    i__1 = min(i__2,i__3);
/* Computing MIN */
	    i__4 = *bwu, i__5 = max_bw__ - 1;
	    i__6 = max_bw__ + 1;
	    dtbtrs_tcomp_macro(
                "U", "T", "N", &max_bw__, &i__1, bwu, &i__6, &max_bw__, info, 
                nprow, npcol, myrow, mycol, tcomp_l
          );

/*         Transpose back */

	    dlatcpy_tcomp_macro(
                 "N", &max_bw__, &max_bw__, &max_bw__, &max_bw__,
                 nprow, npcol, myrow, mycol, tcomp_l
          );


	}
/*         End of "if( info.eq.0 ) then" */

/*         Use offdiag block(s) to calculate modification to diag block */
/*           of processor to the left */

	dgemm_tcomp_macro(
            "N", "T", &max_bw__, &max_bw__, &max_bw__, (char*)&c_b38, 
            &max_bw__, &max_bw__, (char*)&c_b62, &max_bw__, 
            nprow, npcol, myrow, mycol, tcomp_l
      );

/*         Send contribution to diagonal block's owning processor. */

	i__1 = mycol - level_dist__;
	hscal_Cdgesd2d(
           blacsctxt, max_bw__, max_bw__, max_bw__, c__0, 
           i__1, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);

/*         ******************************************************* */

	if (mycol / level_dist__ <= (npcol - 1) / level_dist__ - 2) {

/*           Decide which processor offdiagonal block(s) goes to */

	    if (mycol / (level_dist__ << 1) % 2 == 0) {
		comm_proc__ = mycol + level_dist__;
	    } else {
		comm_proc__ = mycol - level_dist__;
	    }

/*           Use offdiagonal blocks to calculate offdiag */
/*             block to send to neighboring processor. Depending */
/*             on circumstances, may need to transpose the matrix. */

	    dgemm_tcomp_macro(
               "N", "N", &max_bw__, &max_bw__, &max_bw__, (char*)&c_b38, 
               &max_bw__, &max_bw__, (char*)&c_b62, &max_bw__, 
               nprow, npcol, myrow, mycol, tcomp_l
          );

/*           Send contribution to offdiagonal block's owning processor. */

	    hscal_Cdgesd2d(
               blacsctxt, max_bw__, max_bw__, max_bw__, c__0,
		     comm_proc__,
               nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
          );

	    dgemm_tcomp_macro(
               "N", "N", &max_bw__, &max_bw__, &max_bw__, (char*)&c_b38, 
               &max_bw__, &max_bw__, (char*)&c_b62,
		   &max_bw__, nprow, npcol, myrow, mycol, tcomp_l);

/*           Send contribution to offdiagonal block's owning processor. */

	    hscal_Cdgesd2d(
               blacsctxt, max_bw__, max_bw__, max_bw__, c__0,
               comm_proc__,
               nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
          );

	}

    }
/*       End of "if( mycol/level_dist.le. (npcol-1)/level_dist -1 )..." */

L120:

/*     Output minimum worksize */

/*         Make INFO consistent across processors */

/*
* Ignoring the broadcast of 'info'
*/

   Cblacs_gridexit_nocomm(blacsctxt);

/*     End of PDDBTRF */

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrf_tcomp_tcomm(
   int N,
   int BWL, 
   int BWU,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pddbtrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N, BWL, BWU,
                JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrf_info(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * JA,
   int    * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pddbtrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, *BWL, *BWU,
                *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrf_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * JA,
   int    * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pddbtrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N, *BWL, *BWU,
                *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
