
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pddbtrs_recon_tcomp(
   int TRANSI,
   int N,
   int BWL, 
   int BWU,
   int NRHS
)
{
    int max_bw__ = max(BWL, BWU);

    double tcomp = 0;

    char   transac  = 'T';
    char   *transa  = &transac;

    char   transbc  = 'N';
    char   *transb  = &transbc;

    int reconm, reconn, reconk;

    int    lda, ldb, ldc;

    double alpha, beta;

    alpha = -1.;
    beta = 1.;

    reconm  = max_bw__;
    reconn  = max_bw__;
    reconk  = max_bw__;

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_dgemm(
         transa,
         transb,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha),
         &lda,
         &ldb,
         ((char*)&beta),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pddbtrs_my_tcomp(
   int myrow,
   int mycol,
   int TRANSI,
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pddbtrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          N, BWL, BWU, NRHS,
          JA, DESCA,
          IB, DESCB,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_my_tcomm(
   int myrow,
   int mycol,
   int TRANSI,
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pddbtrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          N, BWL, BWU, NRHS,
          JA, DESCA,
          IB, DESCB,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int TRANSI,
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pddbtrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          N, BWL, BWU, NRHS,
          JA, DESCA,
          IB, DESCB,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int TRANSI,
   int N,
   int BWL, 
   int BWU,
   int NRHS,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__17 = 17;
    hscalinteger c__1 = 1;
    /* System generated locals */
    hscalinteger i__1;
    hscalinteger desca_1xp__[7], descb_px1__[7], store_m_b__, store_n_a__, 
	    ictxt_new__, nb, np, ictxt_save__, first_proc__, param_check__[51]
	    	/* was [17][3] */, return_code__, part_offset__, llda, lldb, 
	    csrc;
    hscalinteger idum2, idum3;
    hscalinteger ictxt, work_size_min__, ja_new__;
    hscalinteger np_save__;

    char TRANSC     = TRANSI, *trans = &TRANSC;

    int     *n      = (int*)&N;
    int     *bwl    = (int*)&BWL;
    int     *bwu    = (int*)&BWU;
    int     *nrhs   = (int*)&NRHS;
    int     *ja     = (int*)&JA;
    int     *desca  = (int*)DESCA;
    int     *ib     = (int*)&IB;
    int     *descb  = (int*)DESCB;

    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;


/*     Test the input parameters */

    /* Parameter adjustments */
    --descb;
    --desca;

    /* Function Body */
    *info = 0;

    if (init_llds_f == 1)
    {
       int *PXDBTRS_DESCA_l = &desca[1];
       int *PXDBTRS_DESCB_l = &descb[1];
       HSCAL_PXDBTRS_INIT_LLDS(
            PXDBTRS_DESCA_l,
            PXDBTRS_DESCB_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PDDBTRS: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d", 
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PDDBTRS: (%d, %d): (descb+1)[M1_]=%d (descb+1)[MB1_]=%d myrow=%d (descb+1)[RSRC1_]=%d nprow=%d (descb+1)[LLD1_]=%d",
        myrow, mycol, (descb+1)[M1_], (descb+1)[MB1_], myrow, (descb+1)[RSRC1_], nprow, (descb+1)[LLD1_]
    );

/*     Test the input parameters */

    /* Parameter adjustments */
    --descb;
    --desca;

    /* Function Body */
    *info = 0;

/*     Convert descriptor into standard form for easy access to */
/*        parameters, check that grid is of right shape. */

    desca_1xp__[0] = 501;
    descb_px1__[0] = 502;

    desc_convert__(&desca[1], desca_1xp__, &return_code__);

    if (return_code__ != 0) {
	*info = -802;
    }

    desc_convert__(&descb[1], descb_px1__, &return_code__);

    if (return_code__ != 0) {
	*info = -1102;
    }

/*     Consistency checks for DESCA and DESCB. */

/*     Context must be the same */
    if (desca_1xp__[1] != descb_px1__[1]) {
	*info = -1102;
    }

/*        These are alignment restrictions that may or may not be removed */
/*        in future releases. -Andy Cleary, April 14, 1996. */

/*     Block sizes must be the same */
    if (desca_1xp__[3] != descb_px1__[3]) {
	*info = -1104;
    }

/*     Source processor must be the same */

    if (desca_1xp__[4] != descb_px1__[4]) {
	*info = -1105;
    }

/*     Get values out of descriptor for use in code. */

    ictxt = desca_1xp__[1];
    csrc = desca_1xp__[4];
    nb = desca_1xp__[3];
    llda = desca_1xp__[5];
    store_n_a__ = desca_1xp__[2];
    lldb = descb_px1__[5];
    store_m_b__ = descb_px1__[2];

/*     Get grid parameters */
    np = nprow * npcol;

    if (hscal_lsame(trans, "N")) {
	idum2 = 'N';
    } else if (hscal_lsame(trans, "T")) {
	idum2 = 'T';
    } else if (hscal_lsame(trans, "C")) {
	idum2 = 'T';
    } else {
	*info = -1;
    }

    idum3 = 1;

    if (*n < 0) {
	*info = -2;
    }

    if (*n + *ja - 1 > store_n_a__) {
	*info = -806;
    }

    if (*bwl > *n - 1 || *bwl < 0) {
	*info = -3;
    }

    if (*bwu > *n - 1 || *bwu < 0) {
	*info = -4;
    }

    if (llda < *bwl + *bwu + 1) {
	*info = -806;
    }

    if (nb <= 0) {
	*info = -804;
    }

    if (*n + *ib - 1 > store_m_b__) {
	*info = -1103;
    }

    if (lldb < nb) {
	*info = -1106;
    }

    if (*nrhs < 0) {
	*info = -5;
    }

/*     Current alignment restriction */

    if (*ja != *ib) {
	*info = -7;
    }

/*     Argument checking that is specific to Divide & Conquer routine */

    if (nprow != 1) {
	*info = -802;
    }

    if (*n > np * nb - (*ja - 1) % nb) {
	*info = -2;
	i__1 = -(*info);
        hscal_PB_Cabort(
           "PDDBTRS, D&C alg.: only 1 block per proc",
           i__1,
           nprow, npcol, myrow, mycol
        );
        return HSCAL_ERR_INPUT;
    }

    if (*ja + *n - 1 > nb && nb < max(*bwl,*bwu) << 1) {
	*info = -804;
	i__1 = -(*info);
        hscal_PB_Cabort(
           "PDDBTRS, D&C alg.: NB too small",
           i__1,
           nprow, npcol, myrow, mycol
        );
        return HSCAL_ERR_INPUT;
    }


    work_size_min__ = max(*bwl,*bwu) * *nrhs;

/*     Pack params and positions into arrays for global consistency check */

    param_check__[16] = descb[5];
    param_check__[15] = descb[4];
    param_check__[14] = descb[3];
    param_check__[13] = descb[2];
    param_check__[12] = descb[1];
    param_check__[11] = *ib;
    param_check__[10] = desca[5];
    param_check__[9] = desca[4];
    param_check__[8] = desca[3];
    param_check__[7] = desca[1];
    param_check__[6] = *ja;
    param_check__[5] = *nrhs;
    param_check__[4] = *bwu;
    param_check__[3] = *bwl;
    param_check__[2] = *n;
    param_check__[1] = idum3;
    param_check__[0] = idum2;

    param_check__[33] = 1105;
    param_check__[32] = 1104;
    param_check__[31] = 1103;
    param_check__[30] = 1102;
    param_check__[29] = 1101;
    param_check__[28] = 10;
    param_check__[27] = 805;
    param_check__[26] = 804;
    param_check__[25] = 803;
    param_check__[24] = 801;
    param_check__[23] = 7;
    param_check__[22] = 5;
    param_check__[21] = 4;
    param_check__[20] = 3;
    param_check__[19] = 2;
    param_check__[18] = 15;
    param_check__[17] = 1;

/*     Want to find errors with MIN( ), so if no error, set it to a big */
/*     number. If there already is an error, multiply by the the */
/*     descriptor multiplier. */

    if (*info >= 0) {
	*info = 10000;
    } else if (*info < -100) {
	*info = -(*info);
    } else {
	*info = -(*info) * 100;
    }

/*     Check consistency across processors */

    hscal_globchk(&ictxt, &c__17, param_check__, &c__17, &param_check__[34], info,
                  &myrow, &mycol, &nprow, &npcol
    );

/*     Prepare output: set info = 0 if no error, and divide by DESCMULT */
/*     if error is not in a descriptor entry. */

    if (*info == 10000) {
	*info = 0;
    } else if (*info % 100 == 0) {
	*info = -(*info) / 100;
    } else {
	*info = -(*info);
    }

    if (*info < 0) {
	i__1 = -(*info);
        hscal_PB_Cabort(
            "PDDBTRS",
            i__1,
            nprow, npcol, myrow, mycol
        );
	return HSCAL_ERR_INPUT;
    }

/*     Quick return if possible */

    if (*n == 0) {
	return HSCAL_SUCCESS;
    }

    if (*nrhs == 0) {
	return HSCAL_SUCCESS;
    }


/*     Adjust addressing into matrix space to properly get into */
/*        the beginning part of the relevant data */

    part_offset__ = nb * ((*ja - 1) / (npcol * nb));

    if (mycol - csrc < (*ja - part_offset__ - 1) / nb) {
	part_offset__ += nb;
    }

    if (mycol < csrc) {
	part_offset__ -= nb;
    }

/*     Form a new BLACS grid (the "standard form" grid) with only procs */
/*        holding part of the matrix, of size 1xNP where NP is adjusted, */
/*        starting at csrc=0, with JA modified to reflect dropped procs. */

/*     First processor to hold part of the matrix: */

    first_proc__ = ((*ja - 1) / nb + csrc) % npcol;

/*     Calculate new JA one while dropping off unused processors. */

    ja_new__ = (*ja - 1) % nb + 1;

/*     Save and compute new value of NP */

    np_save__ = np;
    np = (ja_new__ + *n - 2) / nb + 1;

/*     Call utility routine that forms "standard-form" grid */
/*
*    reshape_(&ictxt, &c__1, &ictxt_new__, &c__1, &first_proc__, &c__1, &np);
*/
/*     Use new context from standard grid as context. */
/*
*    ictxt_save__ = ictxt;
*    ictxt = ictxt_new__;
*    desca_1xp__[1] = ictxt_new__;
*    descb_px1__[1] = ictxt_new__;
*/
/*     Get information about new grid. */
/*
*    blacs_gridinfo__(&ictxt, &nprow, &npcol, &myrow, &mycol);
*/
/*     Drop out processors that do not have part of the matrix. */
/*
*    if (myrow < 0) {
*	goto L20;
*    }
*/

/*
* The new grid is of the shape 1*np, therefore use it.
*/
nprow = 1;
npcol = np;

if ((mycol > (np - 1)))
{
   return HSCAL_SUCCESS;
}


/*     Begin main code */

    *info = 0;

/*     Call frontsolve routine */

    if (hscal_lsame(trans, "N")) {

	hscal_pddbtrsv(
           "L", "N", n, bwl, bwu, nrhs, 
           &ja_new__, desca_1xp__, ib, descb_px1__,
	     info, 
           nprow, npcol,
           myrow, mycol,
           tcomp, tcomm
      );

    } else {

	hscal_pddbtrsv(
            "U", "T", n, bwl, bwu, nrhs,
            &ja_new__, desca_1xp__, ib, descb_px1__,
		info,
             nprow, npcol,
             myrow, mycol,
             tcomp, tcomm
      );

    }

/*     Call backsolve routine */

    if (hscal_lsame(trans, "C") || hscal_lsame(trans, "T")) {

	hscal_pddbtrsv(
           "L", "T", n, bwl, bwu, nrhs,
           &ja_new__, desca_1xp__, ib, descb_px1__,
           info,
             nprow, npcol,
           myrow, mycol,
             tcomp, tcomm
      );

    } else {

	hscal_pddbtrsv(
           "U", "N", n, bwl, bwu, nrhs, 
           &ja_new__, desca_1xp__, ib, descb_px1__, info,
             nprow, npcol,
           myrow, mycol,
             tcomp, tcomm
      );

    }
/* L10: */

/*     End of PDDBTRS */
   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_info(
   char   * TRANS,
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pddbtrs_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *TRANS,
                *N, *BWL, *BWU, *NRHS,
                *JA, DESCA,
                *IB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_info_i(
   char   * TRANS,
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pddbtrs_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *TRANS,
                *N, *BWL, *BWU, *NRHS,
                *JA, DESCA,
                *IB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
