
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdgetrf_recon_tcomp(
   int M,
   int N
)
{
    double tcomp = 0.;

    char   transc  = 'N';
    char   *trans   = &transc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    double alpha1 = -1.;
    double alpha2 =  1.;

    reconm  = M/p;
    reconn  = N/q;
    reconk  = hscal_pilaenv_recon("D");

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_dgemm(
         trans,
         trans,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha1),
         &lda,
         &ldb,
         ((char*)&alpha2),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdgetrf_my_tcomp(
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdgetrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdgetrf_my_tcomm(
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdgetrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgetrf_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdgetrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdgetrf_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__1 = 1;
    hscalinteger c__2 = 2;
    hscalinteger c__6 = 6;
    hscalinteger c__0 = 0;
    hscaldoublereal c_b31 = 1.;
    hscaldoublereal c_b34 = -1.;
    hscalinteger c_n1 = -1;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4, i__5, i__6, i__7, i__8;

    /* Local variables */
    hscalinteger i__, j;
    hscalinteger jb, in, jn, mn, idum1[1], idum2[1];
    hscalinteger icoff, iroff, ictxt;
    char colbtop[1], colctop[1];
    char rowbtop[1];

    int     *m      = &M;
    int     *n      = &N;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --desca;

    if (init_llds_f == 1)
    {
       HSCAL_PXGETRF_INIT_LLDS(
            (desca+1),
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PDGETRF: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d",
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    /* Function Body */
    *info = 0;
    if (nprow == -1) {
	*info = -602;
    } else {
	hscal_chk1mat(m, &c__1, n, &c__2, ia, ja, &desca[1], &c__6, info,
                      myrow, mycol, nprow, npcol
        );
	if (*info == 0) {
	    iroff = (*ia - 1) % desca[5];
	    icoff = (*ja - 1) % desca[6];
	    if (iroff != 0) {
		*info = -4;
	    } else if (icoff != 0) {
		*info = -5;
	    } else if (desca[5] != desca[6]) {
		*info = -606;
	    }
	}
	hscal_pchk1mat(m, &c__1, n, &c__2, ia, ja, &desca[1], &c__6, &c__0, idum1, 
		idum2, info, &myrow, &mycol, &nprow, &npcol);
    }

    if (*info != 0) {
       i__1 = -(*info);
       hscal_PB_Cabort(
            "PDGETRF",
            i__1,
            nprow, npcol, myrow, mycol
       );
       return HSCAL_ERR_INPUT;
    }

/*     Quick return if possible */

    if (desca[3] == 1) {
	return HSCAL_SUCCESS;
    } else if (*m == 0 || *n == 0) {
	return HSCAL_SUCCESS;
    }

/*     Split-ring topology for the communication along process rows */

    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);
    pb_topget__(&ictxt, "Combine", "Columnwise", colctop);
    pb_topset__(&ictxt, "Broadcast", "Rowwise", "S-ring");
    pb_topset__(&ictxt, "Broadcast", "Columnwise", " ");
    pb_topset__(&ictxt, "Combine", "Columnwise", " ");

/*     Handle the first block of columns separately */

    mn = min(*m,*n);
/* Computing MIN */
    i__1 = hscal_iceil(ia, &desca[5]) * desca[5], i__2 = *ia + *m - 1;
    in = min(i__1,i__2);
/* Computing MIN */
    i__1 = hscal_iceil(ja, &desca[6]) * desca[6], i__2 = *ja + mn - 1;
    jn = min(i__1,i__2);
    jb = jn - *ja + 1;

/*     Factor diagonal and subdiagonal blocks and test for exact */
/*     singularity. */

    hscal_pdgetf2_my_tcomp_tcomm(
         0,
         myrow, mycol,
         M, 
         jb, 
         IA, 
         JA, 
         &desca[1], 
         nprow, npcol,
         tcomp, tcomm
    );

    if (jb + 1 <= *n) {

/*        Apply interchanges to columns JN+1:JA+N-1. */

	i__1 = *n - jb;
	i__2 = jn + 1;
        {
             int DIRECI  = 'F';
             int ROWCOLI = 'R';

	     hscal_pdlaswp_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
                  DIRECI, ROWCOLI,
                  i__1, 
                  IA, 
                  i__2, 
                  &desca[1], 
                  IA, 
                  in, 
                  nprow, npcol,
                  tcomp, tcomm
             );
        }

/*        Compute block row of U. */

	i__1 = *n - jb;
	i__2 = jn + 1;
        {
             int SIDEI  = 'L';
             int UPLOI  = 'L';
             int TRANSI = 'N';
             int DIAGI  = 'U';             
             int ALPHAI = c_b31;

	     hscal_pdtrsm_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
                  SIDEI, UPLOI,
                  TRANSI, DIAGI,                  
                  jb, 
                  i__1, 
                  ALPHAI,
	          IA, JA, &desca[1], 
                  IA, i__2, &desca[1],
                  nprow, npcol,
                  tcomp, tcomm
             );
        }

	if (jb + 1 <= *m) {

/*           Update trailing submatrix. */

	    i__1 = *m - jb;
	    i__2 = *n - jb;
	    i__3 = in + 1;
	    i__4 = jn + 1;
	    i__5 = in + 1;
	    i__6 = jn + 1;

            {
                 int ALPHA31 = c_b31, ALPHA34 = c_b34;

	         hscal_pdgemm_my_tcomp_tcomm(
                      0,
                      myrow, mycol,
                      'N', 'N', 
                      i__1, i__2, jb, ALPHA34,
	              i__3, JA, &desca[1], IA, i__4, &desca[1],
                      ALPHA31, i__5, i__6, &desca[1],
                      nprow, npcol,
                      tcomp, tcomm
                 );
            }
	}
    }

/*     Loop over the remaining blocks of columns. */

    i__1 = *ja + mn - 1;
    i__2 = desca[6];
    for (j = jn + 1; i__2 < 0 ? j >= i__1 : j <= i__1; j += i__2) {
/* Computing MIN */
	i__3 = mn - j + *ja;
	jb = min(i__3,desca[6]);
	i__ = *ia + j - *ja;

/*        Factor diagonal and subdiagonal blocks and test for exact */
/*        singularity. */

	i__3 = *m - j + *ja;
	hscal_pdgetf2_my_tcomp_tcomm(
             0,
             myrow, mycol,
             i__3, jb, i__, j, &desca[1],
             nprow, npcol,
             tcomp, tcomm
        );

/*        Apply interchanges to columns JA:J-JA. */

	i__3 = j - *ja;
	i__4 = i__ + jb - 1;
	hscal_pdlaswp_my_tcomp_tcomm(
             0,
             myrow, mycol,
             'F', 'R', 
             i__3, IA, IA, &desca[1], i__, 
	     i__4, 
             nprow, npcol,
             tcomp, tcomm
        );

	if (j - *ja + jb + 1 <= *n) {

/*           Apply interchanges to columns J+JB:JA+N-1. */

	    i__3 = *n - j - jb + *ja;
	    i__4 = j + jb;
	    i__5 = i__ + jb - 1;
	    hscal_pdlaswp_my_tcomp_tcomm(
                 0,
                 myrow, mycol,
                 'F', 'R', 
                 i__3, IA, i__4, &desca[1],
		 i__, i__5, 
                 nprow, npcol,
                 tcomp, tcomm
            );

/*           Compute block row of U. */

	    i__3 = *n - j - jb + *ja;
	    i__4 = j + jb;

            {
                 int ALPHA31 = c_b31;

	         hscal_pdtrsm_my_tcomp_tcomm(
                      0,
                      myrow, mycol,
                      'L', 'L', 'N', 'U', 
                      jb, i__3, ALPHA31, i__, j, &desca[1], 
                      i__, i__4, &desca[1], 
                      nprow, npcol,
                      tcomp, tcomm
                 );
            }

	    if (j - *ja + jb + 1 <= *m) {

/*              Update trailing submatrix. */

		i__3 = *m - j - jb + *ja;
		i__4 = *n - j - jb + *ja;
		i__5 = i__ + jb;
		i__6 = j + jb;
		i__7 = i__ + jb;
		i__8 = j + jb;

                {
                     int ALPHA31 = c_b31, ALPHA34 = c_b34;

		     hscal_pdgemm_my_tcomp_tcomm(
                          0,
                          myrow, mycol,
                          'N', 'N', 
                          i__3, i__4, jb, ALPHA34, i__5, j, &desca[1], 
                          i__, i__6, &desca[1], ALPHA31, i__7, i__8, &desca[1],
                          nprow, npcol,
                          tcomp, tcomm
                     );
                }
	    }
	}

/* L10: */
    }

/*
*  Broadcast of 'info' is ignored
*/

    pb_topset__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topset__(&ictxt, "Broadcast", "Columnwise", colbtop);
    pb_topset__(&ictxt, "Combine", "Columnwise", colctop);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgetrf_tcomp_tcomm(
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdgetrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                M, N,
                IA, JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgetrf_info(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdgetrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgetrf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdgetrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

