
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdlapv2_recon_tcomp(
   int DIRECI,
   int ROWCOLI,
   int N
)
{
    double tcomp = 1;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdlapv2_my_tcomp(
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCP[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdlapv2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI,
          M, N,
          IA, JA, DESCA, 
          IP, JP, DESCP, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdlapv2_my_tcomm(
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCP[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdlapv2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI,
          M, N,
          IA, JA, DESCA,
          IP, JP, DESCP,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdlapv2_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int DIRECI,
   int ROWCOLI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCP[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdlapv2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI,
          M, N,
          IA, JA, DESCA,
          IP, JP, DESCP,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdlapv2_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCIP[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__1 = 1;

    /* System generated locals */
    hscalinteger i__1, i__2;

    /* Local variables */
    hscalinteger i__, j, k, ib, jb, ma, ip1, jp1, mba, nba, iip, jjp, itmp;
    hscalinteger ictxt;
    hscallogical forwrd;
    hscalinteger ipvwrk;
    hscallogical rowpvt;
    hscalinteger icurcol, icurrow;

/*
*  .. Local Arrays ..
*/
    HSCALBLACSCONTEXT   * blacsctxt;

    char    DIRECC  = DIRECI, *direc = &DIRECC;
    char    ROWCOLC = ROWCOLI, *rowcol = &ROWCOLC;
    int     *m      = &M;
    int     *n      = &N;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *ip     = &IP;
    int     *jp     = &JP;
    int     *descip = (int*)DESCIP;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descip;
    --desca;

    if (init_llds_f == 1)
    {
       HSCAL_PXLAPV2_INIT_LLDS(
            (desca+1),
            myrow, nprow
       );
    }

    /* Function Body */
    rowpvt = hscal_lsame(rowcol, "R");
    if (rowpvt) {
	if (*m <= 1 || *n < 1) {
	    return HSCAL_SUCCESS;
	}
    } else {
	if (*m < 1 || *n <= 1) {
	    return HSCAL_SUCCESS;
	}
    }
    forwrd = hscal_lsame(direc, "F");


/*     Get grid and matrix parameters */

    ma = desca[3];
    mba = desca[5];
    nba = desca[6];
    ictxt = desca[2];

/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*     If I'm applying pivots from beginning to end (e.g., repeating */
/*     pivoting done earlier).  Thus this section computes P * sub( A ). */

    if (forwrd) {

/*        If I'm pivoting the rows of sub( A ) */

	if (rowpvt) {

/*          Loop over rows of sub( A ) */

	    i__ = *ia;
/* Computing MIN */
	    i__1 = *m, i__2 = hscal_iceil(ia, &mba) * mba - *ia + 1;
	    ib = min(i__1,i__2);

            do {
/*              Find local pointer into IPIV, and broadcast this block's */
/*              pivot information to everyone in process column */

            infog2l_(ip, jp, &descip[1],
                &nprow, &npcol, &myrow, &mycol, &iip, &jjp, &icurrow, &icurcol);

	    if (myrow == icurrow) {
		hscal_Cigebs2d(blacsctxt, "Columnwise", " ", ib, c__1,
			ib, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    } else {
		hscal_Cigebr2d(blacsctxt, "Columnwise", " ", ib, c__1, 
			ib, icurrow, mycol, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    }

/*              Pivot the block of rows */

/*
*  Assuming the worst case scenario for performance models
*  A swap for each iteration
*  Also it does not matter what the indices IX, JX, IY, JY are
*  As long as we know they are aligned the same way
*/
	    i__1 = i__ + ib - 1;
	    for (k = i__; k <= i__1; ++k) {
		{
                    int *PDSWAP_DESCA_l = &desca[1];
		    hscal_pdswap_my_tcomp_tcomm(
                             0,
                             myrow, mycol,
                             N, k, JA, PDSWAP_DESCA_l, ma, k, JA,
			     PDSWAP_DESCA_l, ma, nprow, npcol, tcomp, tcomm);
		}
/* L20: */
	    }

/*              Go on to next row of processes, increment row counter, */
/*              and figure number of rows to pivot next */

	    icurrow = (icurrow + 1) % nprow;
	    i__ += ib;
/* Computing MIN */
	    i__1 = mba, i__2 = *m - i__ + *ia;
	    ib = min(i__1,i__2);
            } while (ib > 0);

/*        If I am pivoting the columns of sub( A ) */

	} else {

/*          Loop over columns of sub( A ) */

	    j = *ja;
/* Computing MIN */
	    i__1 = *n, i__2 = hscal_iceil(ja, &nba) * nba - *ja + 1;
	    jb = min(i__1,i__2);
            do {

/*              Find local pointer into IPIV, and broadcast this block's */
/*              pivot information to everyone in process row */

            infog2l_(ip, jp, &descip[1],
                &nprow, &npcol, &myrow, &mycol, &iip, &jjp, &icurrow, &icurcol);

	    if (mycol == icurcol) {
		hscal_Cigebs2d(blacsctxt, "Rowwise", " ", jb, c__1, jb, 
			nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    } else {
		hscal_Cigebr2d(blacsctxt, "Rowwise", " ", jb, c__1, jb,
			myrow, icurcol, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    }

/*              Pivot the block of columns */

/*
*  Assuming the worst case scenario for performance models
*  A swap for each iteration
*  Also it does not matter what the indices IX, JX, IY, JY are
*  As long as we know they are aligned the same way
*/
	    i__1 = j + jb - 1;
	    for (k = j; k <= i__1; ++k) {
		{
                    int *PDSWAP_DESCA_l = &desca[1];
		    hscal_pdswap_my_tcomp_tcomm(
                            0,
                            myrow, mycol,
                            M, IA, k, PDSWAP_DESCA_l, c__1, IA, k,
			    PDSWAP_DESCA_l, c__1, nprow, npcol, tcomp, tcomm);
		}
/* L40: */
	    }

/*              Go on to next column of processes, increment column */
/*              counter, and figure number of columns to pivot next */

	    icurcol = (icurcol + 1) % npcol;
	    j += jb;
/* Computing MIN */
	    i__1 = nba, i__2 = *n - j + *ja;
	    jb = min(i__1,i__2);
            } while (jb > 0);
	}

/*     If I want to apply pivots in reverse order, i.e. reversing */
/*     pivoting done earlier.  Thus this section computes */
/*     inv( P ) * sub( A ). */

    } else {

/*        If I'm pivoting the rows of sub( A ) */

	if (rowpvt) {

/*           If I'm not in the current process row, my IIP points out */
/*           past end of pivot vector (since I don't own a piece of the */
/*           last row). Adjust IIP so it points at last pivot entry. */

	    {
		;
	    }

/*           Loop over rows in reverse order, starting at last row */

	    i__ = *ia + *m - 1;
	    ib = i__ % mba;
	    if (ib == 0) {
		ib = mba;
	    }
	    ib = min(ib,*m);

            do {

/*              Find local pointer into IPIV, and broadcast this block's */
/*              pivot information to everyone in process column */

	    i__1 = *ip + *m - 1;
	    infog2l_(&i__1, jp, &descip[1], &nprow, &npcol, &myrow, &mycol, &
		    iip, &jjp, &icurrow, &icurcol);

	    if (myrow == icurrow) {
		hscal_Cigebs2d(blacsctxt, "Columnwise", " ", ib, c__1,
			ib, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    } else {
		hscal_Cigebr2d(blacsctxt, "Columnwise", " ", ib, c__1,
			 ib, icurrow, mycol, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    }

/*              Pivot the block of rows */

/*
*  Assuming the worst case scenario for performance models
*  A swap for each iteration
*  Also it does not matter what the indices IX, JX, IY, JY are
*  As long as we know they are aligned the same way
*/
	    i__1 = i__ - ib + 1;
	    for (k = i__; k >= i__1; --k) {
		{
                    int *PDSWAP_DESCA_l = &desca[1];
		    hscal_pdswap_my_tcomp_tcomm(
                             0,
                             myrow, mycol,
                             N, k, JA, PDSWAP_DESCA_l, ma, k, JA,
			     PDSWAP_DESCA_l, ma, nprow, npcol, tcomp, tcomm);
		}
/* L60: */
	    }

/*              Go to previous row of processes, decrement row counter, */
/*              and figure number of rows to be pivoted next */

	    icurrow = (nprow + icurrow - 1) % nprow;
	    i__ -= ib;
/* Computing MIN */
	    i__1 = mba, i__2 = i__ - *ia + 1;
	    ib = min(i__1,i__2);
            } while (ib > 0);

/*        Otherwise, I'm pivoting the columns of sub( A ) */

	} else {

/*           If I'm not in the current process column, my JJP points out */
/*           past end of pivot vector (since I don't own a piece of the */
/*           last column). Adjust JJP so it points at last pivot entry. */

	    {
		;
	    }

/*          Loop over columns in reverse order starting at last column */

	    j = *ja + *n - 1;
	    jb = j % nba;
	    if (jb == 0) {
		jb = nba;
	    }
	    jb = min(jb,*n);

            do {

/*              Find local pointer into IPIV, and broadcast this block's */
/*              pivot information to everyone in process row */

	    i__1 = *jp + *n - 1;
	    infog2l_(ip, &i__1, &descip[1], &nprow, &npcol, &myrow, &mycol, &
		    iip, &jjp, &icurrow, &icurcol);
	    if (mycol == icurcol) {
		hscal_Cigebs2d(blacsctxt, "Rowwise", " ", jb, c__1,
			jb, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    } else {
		hscal_Cigebr2d(blacsctxt, "Rowwise", " ", jb, c__1,
			jb, myrow, icurcol, nprow, npcol, myrow, mycol, tcomp_l, tcomm_l);
	    }

/*              Pivot a block of columns */

/*
*  Assuming the worst case scenario for performance models
*  A swap for each iteration
*  Also it does not matter what the indices IX, JX, IY, JY are
*  As long as we know they are aligned the same way
*/
	    i__1 = j - jb + 1;
	    for (k = j; k >= i__1; --k) {
		{
                    int *PDSWAP_DESCA_l = &desca[1];
		    hscal_pdswap_my_tcomp_tcomm(
                            0,
                            myrow, mycol, 
                            M, IA, k, PDSWAP_DESCA_l, c__1, IA, k,
			    PDSWAP_DESCA_l, c__1, nprow, npcol, tcomp, tcomm);
		}
/* L80: */
	    }

/*              Go to previous row of processes, decrement row counter, */
/*              and figure number of rows to be pivoted next */

	    icurcol = (npcol + icurcol - 1) % npcol;
	    j -= jb;
/* Computing MIN */
	    i__1 = nba, i__2 = j - *ja + 1;
	    jb = min(i__1,i__2);
	    } while (jb > 0);
	}

    }

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdlapv2_tcomp_tcomm(
   int DIRECI,
   int ROWCOLI,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IP,
   int JP,
   int DESCP[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdlapv2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                DIRECI, ROWCOLI,
                M, N,
                IA, JA, DESCA,
                IP, JP, DESCP,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdlapv2_info(
   char *DIRECC,
   char *ROWCOLC,
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IP,
   int * JP,
   int * DESCP,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdlapv2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *DIRECC, *ROWCOLC,
                *M, *N,
                *IA, *JA, DESCA,
                *IP, *JP, DESCP,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdlapv2_info_i(
   char *DIRECC,
   char *ROWCOLC,
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IP,
   int * JP,
   int * DESCP,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdlapv2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *DIRECC, *ROWCOLC,
                *M, *N,
                *IA, *JA, DESCA,
                *IP, *JP, DESCP,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

