
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdlarfg_recon_tcomp(
   int n
)
{
    double tcomp = 0;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdlarfg_my_tcomp(
   int myrow,
   int mycol,
   int N,
   int IAX,
   int JAX,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdlarfg_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N,
          IAX, JAX, 
          IX, JX, 
          DESCX, INCX,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdlarfg_my_tcomm(
   int myrow,
   int mycol,
   int N,
   int IAX,
   int JAX,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdlarfg_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N,
          IAX, JAX, 
          IX, JX, 
          DESCX, INCX,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdlarfg_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int N,
   int IAX,
   int JAX,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdlarfg_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          N,
          IAX, JAX, 
          IX, JX, 
          DESCX, INCX,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdlarfg_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int N,
   int IAX,
   int JAX,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__1 = 1;

    /* System generated locals */
    hscalinteger i__1;

    /* Local variables */
    hscalinteger j, knt;
    hscalinteger iiax, jjax, ixcol, ictxt;
    hscalinteger ixrow;

/*
*  .. Local Arrays ..
*/
    HSCALBLACSCONTEXT   * blacsctxt;

    int     *n      = &N;
    int     *iax    = &IAX;
    int     *jax    = &JAX;
    int     *ix     = &IX;
    int     *jx     = &JX;
    int     *descx  = (int*)DESCX;
    int     *incx   = &INCX;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descx;

    if (init_llds_f == 1)
    {
       int *PXLARFG_DESCX_l = &descx[1];
       HSCAL_PXLARFG_INIT_LLDS(
            PXLARFG_DESCX_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pdlarfg: (%d, %d): (descx+1)[M1_]=%d (descx+1)[MB1_]=%d myrow=%d (descx+1)[RSRC1_]=%d nprow=%d (descx+1)[LLD1_]=%d", 
        myrow, mycol, (descx+1)[M1_], (descx+1)[MB1_], myrow, (descx+1)[RSRC1_], nprow, (descx+1)[LLD1_]
    );

   Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

    if (*incx == descx[3]) {

/*        sub( X ) is distributed across a process row. */

	infog2l_(ix, jax, &descx[1], &nprow, &npcol, &myrow, &mycol, &iiax, &
		jjax, &ixrow, &ixcol);

	if (myrow != ixrow) {
	    return HSCAL_SUCCESS;
	}

/*        Broadcast X(IAX,JAX) across the process row. */

	if (mycol == ixcol) {
	    j = iiax + (jjax - 1) * descx[9];
	    hscal_Cdgebs2d(
               blacsctxt, "Rowwise", " ", c__1, c__1, c__1,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );
	} else {
	    hscal_Cdgebr2d(
               blacsctxt, "Rowwise", " ", c__1, c__1, c__1,
		   myrow, ixcol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );
	}

    } else {

/*        sub( X ) is distributed across a process column. */

	infog2l_(iax, jx, &descx[1], &nprow, &npcol, &myrow, &mycol, &iiax, &
		jjax, &ixrow, &ixcol);

	if (mycol != ixcol) {
	    return HSCAL_SUCCESS;
	}

/*        Broadcast X(IAX,JAX) across the process column. */

	if (myrow == ixrow) {
	    j = iiax + (jjax - 1) * descx[9];
	    hscal_Cdgebs2d(
               blacsctxt, "Columnwise", " ", c__1, c__1, c__1,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );
	} else {
	    hscal_Cdgebr2d(
               blacsctxt, "Columnwise", " ", c__1, c__1, c__1,
		   ixrow, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );
	}

    }

    if (*n <= 0) {
	return HSCAL_SUCCESS;
    }

    i__1 = *n - 1;
    hscal_pdnrm2_my_tcomp_tcomm(
         0,
         myrow, mycol,
         i__1, *ix, *jx, &descx[1], *incx,
         nprow, npcol, tcomp, tcomm
    );

/*
* Use the general case for the performance model
*/
    {

/*        General case */

	{
	    i__1 = *n - 1;
          {
          int ALPHA = 2;
	    ihscal_pdscal_tcomp(
               0,
               myrow, mycol,
               i__1, ALPHA, *ix, *jx, &descx[1], *incx,
               nprow, npcol, tcomp
          );
          }

	    i__1 = *n - 1;
	    hscal_pdnrm2_my_tcomp_tcomm(
               0,
               myrow, mycol,
               i__1, *ix, *jx, &descx[1], *incx,
               nprow, npcol, tcomp, tcomm
          );

          {
/*
* Some ALPHA for performance model purposes (these are 
* not real computations but just a measure)
*/
          int ALPHA = 2;
	    i__1 = *n - 1;
	    ihscal_pdscal_tcomp(
               0,
               myrow, mycol,
               i__1, ALPHA, *ix, *jx, &descx[1], *incx,
               nprow, npcol, tcomp
          );
          }
	}
    }

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdlarfg_tcomp_tcomm(
   int N,
   int IAX,
   int JAX,
   int IX,
   int JX,
   int DESCX[DLEN_],
   int INCX,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdlarfg_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                N,
                IAX, JAX, 
                IX, JX, 
                DESCX, INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdlarfg_info(
   int * N,
   int * IAX,
   int * JAX,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdlarfg_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N,
                *IAX, *JAX,
                *IX, *JX,
                DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdlarfg_info_i(
   int * N,
   int * IAX,
   int * JAX,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdlarfg_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *N,
                *IAX, *JAX,
                *IX, *JX,
                DESCX, *INCX,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
