
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdposv_recon_tcomp(
   int UPLOI,
   int N,
   int NRHS
)
{
    double tcomp = 0;

    char   transc  = 'N';
    char   *trans  = &transc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    double alpha = 1.;

    if (transc == 'N')
    {
       reconm  = N/p;
       reconn  = hscal_pilaenv_recon("D");
       reconk  = N/q;
    }
    else
    {
       reconm  = hscal_pilaenv_recon("D");
       reconn  = N/p;
       reconk  = N/q;
    }

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_dgemm(
         trans,
         trans,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha),
         &lda,
         &ldb,
         ((char*)&alpha),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdposv_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdposv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI,
          N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_my_tcomm(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdposv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI,
          N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLOI,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdposv_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI,
          N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdposv_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__2 = 2;
    hscalinteger c__7 = 7;
    hscalinteger c__3 = 3;
    hscalinteger c__11 = 11;
    hscalinteger c__1 = 1;

    char UPLOC    = UPLOI, *uplo = &UPLOC;

    int     *n      = (int*)&N;
    int     *nrhs   = (int*)&NRHS;
    int     *ia     = (int*)&IA;
    int     *ja     = (int*)&JA;
    int     *desca  = (int*)DESCA;
    int     *ib     = (int*)&IB;
    int     *jb     = (int*)&JB;
    int     *descb  = (int*)DESCB;

    hscalinteger i__1;
    hscalinteger idum1[1], idum2[1];
    hscalinteger iarow, ibrow;
    hscallogical upper;
    hscalinteger ictxt, icoffa, iroffa, iroffb;

    int     iinfo, *info = &iinfo;

    /* Parameter adjustments */
    --descb;
    --desca;

/*     Test the input parameters */
    if (init_llds_f == 1)
    {
       HSCAL_PXPOSV_INIT_LLDS(
         (desca+1),
         (descb+1),
         myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pdposv: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d", 
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pdposv: (%d, %d): (descb+1)[M1_]=%d (descb+1)[MB1_]=%d myrow=%d (descb+1)[RSRC1_]=%d nprow=%d (descb+1)[LLD1_]=%d",
        myrow, mycol, (descb+1)[M1_], (descb+1)[MB1_], myrow, (descb+1)[RSRC1_], nprow, (descb+1)[LLD1_]
    );

    *info = 0;
    if (nprow == -1) {
	*info = -702;
    } else {
	upper = hscal_lsame(uplo, "U");
	hscal_chk1mat(
           n, &c__2, n, &c__2, ia, ja, &desca[1], &c__7, info,
           myrow, mycol, nprow, npcol
      );
	if (*info == 0) {
	    iarow = indxg2p_(ia, &desca[5], &myrow, &desca[7], &nprow);
	    ibrow = indxg2p_(ib, &descb[5], &myrow, &descb[7], &nprow);
	    iroffa = (*ia - 1) % desca[5];
	    iroffb = (*ib - 1) % descb[5];
	    icoffa = (*ja - 1) % desca[6];
	    if (! upper && ! hscal_lsame(uplo, "L")) {
		*info = -1;
	    } else if (iroffa != 0) {
		*info = -5;
	    } else if (icoffa != 0) {
		*info = -6;
	    } else if (desca[5] != desca[6]) {
		*info = -706;
	    } else if (iroffb != 0 || ibrow != iarow) {
		*info = -9;
	    } else if (descb[5] != desca[6]) {
		*info = -1006;
	    }
	}
	if (upper) {
	    idum1[0] = 'U';
	} else {
	    idum1[0] = 'L';
	}
	idum2[0] = 1;
	hscal_pchk2mat(n, &c__2, n, &c__2, ia, ja, &desca[1], &c__7, n, &c__2, 
		nrhs, &c__3, ib, jb, &descb[1], &c__11, &c__1, idum1, idum2, 
		info, &myrow, &mycol, &nprow, &npcol);
    }

    if (*info != 0) {
	i__1 = -(*info);
       hscal_PB_Cabort( 
            "PDPOSV", 
            i__1,
            nprow, npcol, myrow, mycol
       ); 
       return HSCAL_ERR_INPUT;
    }

/*     Compute the Cholesky factorization sub( A ) = U'*U or L*L'. */

   {
       int *PDPOTRF_DESCA_l = (&desca[1]);

       hscal_pdpotrf_my_tcomp_tcomm(
           0, myrow, mycol,
           UPLOI, N, IA, JA, PDPOTRF_DESCA_l, 
           nprow, npcol,
           tcomp, tcomm
       );
   }

   {

/*        Solve the system sub( A ) * X = sub( B ) overwriting sub( B ) */
/*        with X. */

        int *PDPOTRS_DESCA_l = (&desca[1]);
        int *PDPOTRS_DESCB_l = (&descb[1]);

        hscal_pdpotrs_my_tcomp_tcomm(
            0, myrow, mycol,
            UPLOI, N, NRHS, IA, JA, PDPOTRS_DESCA_l, IB, JB, 
            PDPOTRS_DESCB_l, 
            nprow, npcol,
            tcomp, tcomm
        );

    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_tcomp_tcomm(
   int UPLOI,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdposv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI,
                N, NRHS,
                IA, JA, DESCA,
                IB, JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_info(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdposv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *UPLO,
                *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdposv_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *UPLO,
                *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
