
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pdpotrf_recon_tcomp(
   int UPLOI,
   int N
)
{
    double tcomp = 0.;

    char   transc  = 'N';
    char   *trans   = &transc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    double alpha1 = -1.;
    double alpha2 =  1.;

    reconm  = N/p;
    reconn  = hscal_pilaenv_recon("D");
    reconk  = N/q;

    lda = reconm;
    ldb = reconk;
    ldc = reconm;

    hscal_dgemm(
         trans,
         trans,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha1),
         &lda,
         &ldb,
         ((char*)&alpha2),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pdpotrf_my_tcomp(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdpotrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, N,
          IA, JA, DESCA,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pdpotrf_my_tcomm(
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pdpotrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, N,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdpotrf_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLOI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pdpotrf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLOI, N,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pdpotrf_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__2 = 2;
    hscalinteger c__6 = 6;
    hscalinteger c__1 = 1;
    hscaldoublereal c_b28 = 1.;
    hscaldoublereal c_b30 = -1.;
 
    /* System generated locals */ 
    hscalinteger i__1, i__2, i__3, i__4, i__5, i__6;
 
    /* Local variables */ 
    hscalinteger i__, j;
    hscalinteger jb, jn, idum1[1], idum2[1];
    hscalinteger icoff;
    hscalinteger iroff;
    hscallogical upper;
    hscalinteger ictxt;
    char colbtop[1], rowbtop[1];

    int  iinfo, *info = &iinfo; 

    char UPLOC    = UPLOI, *uplo = &UPLOC;

    int     *n      = (int*)&N;
    int     *ia     = (int*)&IA;
    int     *ja     = (int*)&JA;
    int     *desca  = (int*)DESCA;

    /* Parameter adjustments */ 
    --desca; 
 
/*     Test the input parameters */ 

    if (init_llds_f == 1)
    {
    HSCAL_PXPOTRF_INIT_LLDS( 
         (desca+1), 
         myrow, nprow 
    ); 
    }

    *info = 0;
    if (nprow == -1) {
	*info = -602;
    } else {
	hscal_chk1mat(n, &c__2, n, &c__2, ia, ja, &desca[1], &c__6, info,
                    myrow, mycol, nprow, npcol
      );
	upper = hscal_lsame(uplo, "U");
	if (*info == 0) {
	    iroff = (*ia - 1) % desca[5];
	    icoff = (*ja - 1) % desca[6];
	    if (! upper && ! hscal_lsame(uplo, "L")) {
		*info = -1;
	    } else if (iroff != 0) {
		*info = -4;
	    } else if (icoff != 0) {
		*info = -5;
	    } else if (desca[5] != desca[6]) {
		*info = -606;
	    }
	}
	if (upper) {
	    idum1[0] = 'U';
	} else {
	    idum1[0] = 'L';
	}
	idum2[0] = 1;
	hscal_pchk1mat(n, &c__2, n, &c__2, ia, ja, &desca[1], &c__6, &c__1, idum1, 
		idum2, info, &myrow, &mycol, &nprow, &npcol);
    }

    if (*info != 0) { 
	i__1 = -(*info); 
        hscal_PB_Cabort( 
             "PDPOTRF", 
             i__1, 
             nprow, npcol, myrow, mycol 
        ); 
       return HSCAL_ERR_INPUT;
    } 
 
/*     Quick return if possible */ 
 
    if (*n == 0) {
	return HSCAL_SUCCESS;
    }

    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);

    if (upper) {

/*        Split-ring topology for the communication along process */
/*        columns, 1-tree topology along process rows. */

	pb_topset__(&ictxt, "Broadcast", "Rowwise", " ");
	pb_topset__(&ictxt, "Broadcast", "Columnwise", "S-ring");

/*        A is upper triangular, compute Cholesky factorization A = U'*U. */

/*        Handle the first block of columns separately */

/* Computing MIN */
	i__1 = hscal_iceil(ja, &desca[6]) * desca[6], i__2 = *ja + *n - 1;
	jn = min(i__1,i__2);
	jb = jn - *ja + 1;

/*        Perform unblocked Cholesky factorization on JB block */

	ihscal_pdpotf2_tcomp(
           0,
           myrow, mycol, 
           UPLOI, jb, IA, JA, &desca[1], nprow, npcol, tcomp);

	if (jb + 1 <= *n) {

/*           Form the row panel of U using the triangular solver */

	    i__1 = *n - jb;
	    i__2 = *ja + jb;
          {
             int SIDEI  = 'L';
             int TRANSI = 'T';
             int DIAGI  = 'N';             
             int ALPHA28 = c_b28;
             hscal_pdtrsm_my_tcomp_tcomm(
                0,
                myrow, mycol,
                SIDEI, UPLOI,
                TRANSI, DIAGI,
                jb, i__1, ALPHA28,
		    IA, JA, &desca[1], IA, i__2, &desca[1], nprow, npcol, tcomp, tcomm);
          }

/*           Update the trailing matrix, A = A - U'*U */

	    i__1 = *n - jb;
	    i__2 = *ja + jb;
	    i__3 = *ia + jb;
	    i__4 = *ja + jb;
          {
             int TRANSI = 'T';
             int ALPHA30 = c_b30;
             int ALPHA28 = c_b28;

             hscal_pdsyrk_my_tcomp_tcomm(
                0,
                myrow, mycol,
                UPLOI, TRANSI,
                i__1, jb, ALPHA30, IA, i__2, &desca[1], 
                ALPHA28, i__3, i__4, &desca[1], nprow, npcol, tcomp, tcomm);
          }
	}

/*        Loop over remaining block of columns */

	i__1 = *ja + *n - 1;
	i__2 = desca[6];
	for (j = jn + 1; i__2 < 0 ? j >= i__1 : j <= i__1; j += i__2) {
/* Computing MIN */
	    i__3 = *n - j + *ja;
	    jb = min(i__3,desca[6]);
	    i__ = *ia + j - *ja;

/*           Perform unblocked Cholesky factorization on JB block */

	    ihscal_pdpotf2_tcomp(
               0,
               myrow, mycol, 
               UPLOI, jb, i__, j, &desca[1], nprow, npcol, tcomp);

	    if (j - *ja + jb + 1 <= *n) {

/*              Form the row panel of U using the triangular solver */

		i__3 = *n - j - jb + *ja;
		i__4 = j + jb;
            {
             int SIDEI  = 'L';
             int TRANSI = 'T';
             int DIAGI  = 'N';   
             int ALPHA28 = c_b28;

		 hscal_pdtrsm_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
                  SIDEI, UPLOI,
                  TRANSI, DIAGI,
                  jb, i__3, ALPHA28, i__, j, &desca[1], i__, i__4,
			&desca[1], nprow, npcol, tcomp, tcomm);
            }

/*              Update the trailing matrix, A = A - U'*U */

		i__3 = *n - j - jb + *ja;
		i__4 = j + jb;
		i__5 = i__ + jb;
		i__6 = j + jb;

            {
             int TRANSI = 'T';
             int ALPHA30 = c_b30;
             int ALPHA28 = c_b28;
  		 hscal_pdsyrk_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
                  UPLOI, TRANSI,
                  i__3, jb, ALPHA30, i__, i__4, &desca[1], 
                  ALPHA28, i__5, i__6, &desca[1], 
                  nprow, npcol, tcomp, tcomm);
            }
	    }
/* L10: */
	}

    } else {

/*        1-tree topology for the communication along process columns, */
/*        Split-ring topology along process rows. */

	pb_topset__(&ictxt, "Broadcast", "Rowwise", "S-ring");
	pb_topset__(&ictxt, "Broadcast", "Columnwise", " ");

/*        A is lower triangular, compute Cholesky factorization A = L*L' */
/*        (right-looking) */

/*        Handle the first block of columns separately */

/* Computing MIN */
	i__2 = hscal_iceil(ja, &desca[6]) * desca[6], i__1 = *ja + *n - 1;
	jn = min(i__2,i__1);
	jb = jn - *ja + 1;

/*        Perform unblocked Cholesky factorization on JB block */

	ihscal_pdpotf2_tcomp(
           0,
           myrow, mycol,
           UPLOI, jb, IA, JA, &desca[1], nprow, npcol, tcomp
      );

	if (jb + 1 <= *n) {

/*           Form the column panel of L using the triangular solver */

	    i__2 = *n - jb;
	    i__1 = *ia + jb;
          {
             int SIDEI  = 'R';
             int TRANSI = 'T';
             int DIAGI  = 'N';
             int ALPHA28 = c_b28;

             hscal_pdtrsm_my_tcomp_tcomm(
                0,
                myrow, mycol,
                SIDEI, UPLOI,
                TRANSI, DIAGI,
                i__2, jb, ALPHA28, IA, JA, &desca[1], i__1, JA, &desca[1],
                nprow, npcol, tcomp, tcomm
             );
          }

/*           Update the trailing matrix, A = A - L*L' */

	    i__2 = *n - jb;
	    i__1 = *ia + jb;
	    i__3 = *ia + jb;
	    i__4 = *ja + jb;
          {
             int TRANSI = 'N';
             int ALPHA30 = c_b30;
             int ALPHA28 = c_b28;
             hscal_pdsyrk_my_tcomp_tcomm(
                0,
                myrow, mycol,
                UPLOI, TRANSI,
                i__2, jb, ALPHA30, i__1, 
		    JA, &desca[1], ALPHA28, i__3, i__4, &desca[1], nprow, npcol, tcomp, tcomm);
          }

	}

	i__2 = *ja + *n - 1;
	i__1 = desca[6];
	for (j = jn + 1; i__1 < 0 ? j >= i__2 : j <= i__2; j += i__1) {
/* Computing MIN */
	    i__3 = *n - j + *ja;
	    jb = min(i__3,desca[6]);
	    i__ = *ia + j - *ja;

/*           Perform unblocked Cholesky factorization on JB block */

	    ihscal_pdpotf2_tcomp(
               0,
               myrow, mycol,
	         UPLOI, jb, i__, j, &desca[1], nprow, npcol, tcomp);

	    if (j - *ja + jb + 1 <= *n) {

/*              Form the column panel of L using the triangular solver */

		i__3 = *n - j - jb + *ja;
		i__4 = i__ + jb;
            {
             int SIDEI  = 'R';
             int TRANSI = 'T';
             int DIAGI  = 'N';
             int ALPHA28 = c_b28;
            
		 hscal_pdtrsm_my_tcomp_tcomm(
                0,
                myrow, mycol,
                SIDEI, UPLOI,
                TRANSI, DIAGI,
                i__3, jb, ALPHA28, i__, j, &desca[1], i__4, j, &desca[1], 
                nprow, npcol, tcomp, tcomm);
            }

/*              Update the trailing matrix, A = A - L*L' */

		i__3 = *n - j - jb + *ja;
		i__4 = i__ + jb;
		i__5 = i__ + jb;
		i__6 = j + jb;
            {
             int TRANSI = 'N';
             int ALPHA30 = c_b30;
             int ALPHA28 = c_b28;
             hscal_pdsyrk_my_tcomp_tcomm(
                0,
                myrow, mycol,
                UPLOI, TRANSI,
		    i__3, jb, ALPHA30, i__4, j, 
                  &desca[1], ALPHA28, i__5, i__6, &desca[1], 
                  nprow, npcol, tcomp, tcomm);
            }
	    }
/* L20: */
	}

    }

    pb_topset__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topset__(&ictxt, "Broadcast", "Columnwise", colbtop);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdpotrf_tcomp_tcomm(
   int UPLOI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdpotrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLOI, N,
                IA, JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdpotrf_info(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdpotrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *UPLO, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdpotrf_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pdpotrf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *UPLO, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
