
/*
* Ravi Reddy
* Version 1.0
* School of Computer Science and Informatics, 
* UCD, 
* Dublin,
* March 19, 2008
*/

/*
*  Include files
*/
#include "../include/hscalapack.h"

int hscal_psdbtrf_ctxt(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * JA,
   int    * DESCA,
   int    * ICTXT
)
{
/*  -- ScaLAPACK routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     March 12, 2002 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PSDBTRF computes a LU factorization */
/*  of an N-by-N real banded */
/*  diagonally dominant-like distributed matrix */
/*  with bandwidth BWL, BWU: A(1:N, JA:JA+N-1). */
/*  Reordering is used to increase parallelism in the factorization. */
/*  This reordering results in factors that are DIFFERENT from those */
/*  produced by equivalent sequential codes. These factors cannot */
/*  be used directly by users; however, they can be used in */
/*  subsequent calls to PSDBTRS to solve linear systems. */

/*  The factorization has the form */

/*          P A(1:N, JA:JA+N-1) P^T = L U */

/*  where U is a banded upper triangular matrix and L is banded */
/*  lower triangular, and P is a permutation matrix. */

/*  ===================================================================== */

/*  Arguments */
/*  ========= */


/*  N       (global input) INTEGER */
/*          The number of rows and columns to be operated on, i.e. the */
/*          order of the distributed submatrix A(1:N, JA:JA+N-1). N >= 0. */

/*  BWL     (global input) INTEGER */
/*          Number of subdiagonals. 0 <= BWL <= N-1 */

/*  BWU     (global input) INTEGER */
/*          Number of superdiagonals. 0 <= BWU <= N-1 */

/*  JA      (global input) INTEGER */
/*          The index in the global array A that points to the start of */
/*          the matrix to be operated on (which may be either all of A */
/*          or a submatrix of A). */

/*  DESCA   (global and local input) INTEGER array of dimension DLEN. */
/*          if 1D type (DTYPE_A=501), DLEN >= 7; */
/*          if 2D type (DTYPE_A=1), DLEN >= 9 . */
/*          The array descriptor for the distributed matrix A. */
/*          Contains information of mapping of A to memory. Please */
/*          see NOTES below for full description and options. */

/*  LAF     (local input) INTEGER */
/*          Size of user-input Auxiliary Fillin space AF. Must be >= */
/*          NB*(bwl+bwu)+6*max(bwl,bwu)*max(bwl,bwu) */
/*          If LAF is not large enough, an error code will be returned */
/*          and the minimum acceptable size will be returned in AF( 1 ) */

/*  LWORK   (local input or global input) INTEGER */
/*          Size of user-input workspace WORK. */
/*          If LWORK is too small, the minimal acceptable size will be */
/*          returned in WORK(1) and an error code is returned. LWORK>= */
/*          max(bwl,bwu)*max(bwl,bwu) */

/*  ===================================================================== */

/*  Restrictions */
/*  ============ */

/*  The following are restrictions on the input parameters. Some of these */
/*    are temporary and will be removed in future releases, while others */
/*    may reflect fundamental technical limitations. */

/*    Non-cyclic restriction: VERY IMPORTANT! */
/*      P*NB>= mod(JA-1,NB)+N. */
/*      The mapping for matrices must be blocked, reflecting the nature */
/*      of the divide and conquer algorithm as a task-parallel algorithm. */
/*      This formula in words is: no processor may have more than one */
/*      chunk of the matrix. */

/*    Blocksize cannot be too small: */
/*      If the matrix spans more than one processor, the following */
/*      restriction on NB, the size of each block on each processor, */
/*      must hold: */
/*      NB >= 2*MAX(BWL,BWU) */
/*      The bulk of parallel computation is done on the matrix of size */
/*      O(NB) on each processor. If this is too small, divide and conquer */
/*      is a poor choice of algorithm. */

/*    Submatrix reference: */
/*      JA = IB */
/*      Alignment restriction that prevents unnecessary communication. */


/*  ===================================================================== */

/*  Notes */
/*  ===== */

/*  If the factorization routine and the solve routine are to be called */
/*    separately (to solve various sets of righthand sides using the same */
/*    coefficient matrix), the auxiliary space AF *must not be altered* */
/*    between calls to the factorization routine and the solve routine. */

/*  The best algorithm for solving banded and tridiagonal linear systems */
/*    depends on a variety of parameters, especially the bandwidth. */
/*    Currently, only algorithms designed for the case N/P >> bw are */
/*    implemented. These go by many names, including Divide and Conquer, */
/*    Partitioning, domain decomposition-type, etc. */

/*  Algorithm description: Divide and Conquer */

/*    The Divide and Conqer algorithm assumes the matrix is narrowly */
/*      banded compared with the number of equations. In this situation, */
/*      it is best to distribute the input matrix A one-dimensionally, */
/*      with columns atomic and rows divided amongst the processes. */
/*      The basic algorithm divides the banded matrix up into */
/*      P pieces with one stored on each processor, */
/*      and then proceeds in 2 phases for the factorization or 3 for the */
/*      solution of a linear system. */
/*      1) Local Phase: */
/*         The individual pieces are factored independently and in */
/*         parallel. These factors are applied to the matrix creating */
/*         fillin, which is stored in a non-inspectable way in auxiliary */
/*         space AF. Mathematically, this is equivalent to reordering */
/*         the matrix A as P A P^T and then factoring the principal */
/*         leading submatrix of size equal to the sum of the sizes of */
/*         the matrices factored on each processor. The factors of */
/*         these submatrices overwrite the corresponding parts of A */
/*         in memory. */
/*      2) Reduced System Phase: */
/*         A small (max(bwl,bwu)* (P-1)) system is formed representing */
/*         interaction of the larger blocks, and is stored (as are its */
/*         factors) in the space AF. A parallel Block Cyclic Reduction */
/*         algorithm is used. For a linear system, a parallel front solve */
/*         followed by an analagous backsolve, both using the structure */
/*         of the factored matrix, are performed. */
/*      3) Backsubsitution Phase: */
/*         For a linear system, a local backsubstitution is performed on */
/*         each processor in parallel. */

/*  Descriptors */
/*  =========== */

/*  Descriptors now have *types* and differ from ScaLAPACK 1.0. */

/*  Note: banded codes can use either the old two dimensional */
/*    or new one-dimensional descriptors, though the processor grid in */
/*    both cases *must be one-dimensional*. We describe both types below. */

/*  Each global data object is described by an associated description */
/*  vector.  This vector stores the information required to establish */
/*  the mapping between an object element and its corresponding process */
/*  and memory location. */

/*  Let A be a generic term for any 2D block cyclicly distributed array. */
/*  Such a global array has an associated description vector DESCA. */
/*  In the following comments, the character _ should be read as */
/*  "of the global array". */

/*  NOTATION        STORED IN      EXPLANATION */
/*  --------------- -------------- -------------------------------------- */
/*  DTYPE_A(global) DESCA( DTYPE_ )The descriptor type.  In this case, */
/*                                 DTYPE_A = 1. */
/*  CTXT_A (global) DESCA( CTXT_ ) The BLACS context handle, indicating */
/*                                 the BLACS process grid A is distribu- */
/*                                 ted over. The context itself is glo- */
/*                                 bal, but the handle (the integer */
/*                                 value) may vary. */
/*  M_A    (global) DESCA( M_ )    The number of rows in the global */
/*                                 array A. */
/*  N_A    (global) DESCA( N_ )    The number of columns in the global */
/*                                 array A. */
/*  MB_A   (global) DESCA( MB_ )   The blocking factor used to distribute */
/*                                 the rows of the array. */
/*  NB_A   (global) DESCA( NB_ )   The blocking factor used to distribute */
/*                                 the columns of the array. */
/*  RSRC_A (global) DESCA( RSRC_ ) The process row over which the first */
/*                                 row of the array A is distributed. */
/*  CSRC_A (global) DESCA( CSRC_ ) The process column over which the */
/*                                 first column of the array A is */
/*                                 distributed. */
/*  LLD_A  (local)  DESCA( LLD_ )  The leading dimension of the local */
/*                                 array.  LLD_A >= MAX(1,LOCr(M_A)). */

/*  Let K be the number of rows or columns of a distributed matrix, */
/*  and assume that its process grid has dimension p x q. */
/*  LOCr( K ) denotes the number of elements of K that a process */
/*  would receive if K were distributed over the p processes of its */
/*  process column. */
/*  Similarly, LOCc( K ) denotes the number of elements of K that a */
/*  process would receive if K were distributed over the q processes of */
/*  its process row. */
/*  The values of LOCr() and LOCc() may be determined via a call to the */
/*  ScaLAPACK tool function, NUMROC: */
/*          LOCr( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ), */
/*          LOCc( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ). */
/*  An upper bound for these quantities may be computed by: */
/*          LOCr( M ) <= ceil( ceil(M/MB_A)/NPROW )*MB_A */
/*          LOCc( N ) <= ceil( ceil(N/NB_A)/NPCOL )*NB_A */


/*  One-dimensional descriptors: */

/*  One-dimensional descriptors are a new addition to ScaLAPACK since */
/*    version 1.0. They simplify and shorten the descriptor for 1D */
/*    arrays. */

/*  Since ScaLAPACK supports two-dimensional arrays as the fundamental */
/*    object, we allow 1D arrays to be distributed either over the */
/*    first dimension of the array (as if the grid were P-by-1) or the */
/*    2nd dimension (as if the grid were 1-by-P). This choice is */
/*    indicated by the descriptor type (501 or 502) */
/*    as described below. */

/*    IMPORTANT NOTE: the actual BLACS grid represented by the */
/*    CTXT entry in the descriptor may be *either*  P-by-1 or 1-by-P */
/*    irrespective of which one-dimensional descriptor type */
/*    (501 or 502) is input. */
/*    This routine will interpret the grid properly either way. */
/*    ScaLAPACK routines *do not support intercontext operations* so that */
/*    the grid passed to a single ScaLAPACK routine *must be the same* */
/*    for all array descriptors passed to that routine. */

/*    NOTE: In all cases where 1D descriptors are used, 2D descriptors */
/*    may also be used, since a one-dimensional array is a special case */
/*    of a two-dimensional array with one dimension of size unity. */
/*    The two-dimensional array used in this case *must* be of the */
/*    proper orientation: */
/*      If the appropriate one-dimensional descriptor is DTYPEA=501 */
/*      (1 by P type), then the two dimensional descriptor must */
/*      have a CTXT value that refers to a 1 by P BLACS grid; */
/*      If the appropriate one-dimensional descriptor is DTYPEA=502 */
/*      (P by 1 type), then the two dimensional descriptor must */
/*      have a CTXT value that refers to a P by 1 BLACS grid. */


/*  Summary of allowed descriptors, types, and BLACS grids: */
/*  DTYPE           501         502         1         1 */
/*  BLACS grid      1xP or Px1  1xP or Px1  1xP       Px1 */
/*  ----------------------------------------------------- */
/*  A               OK          NO          OK        NO */
/*  B               NO          OK          NO        OK */

/*  Let A be a generic term for any 1D block cyclicly distributed array. */
/*  Such a global array has an associated description vector DESCA. */
/*  In the following comments, the character _ should be read as */
/*  "of the global array". */

/*  NOTATION        STORED IN  EXPLANATION */
/*  --------------- ---------- ------------------------------------------ */
/*  DTYPE_A(global) DESCA( 1 ) The descriptor type. For 1D grids, */
/*                                TYPE_A = 501: 1-by-P grid. */
/*                                TYPE_A = 502: P-by-1 grid. */
/*  CTXT_A (global) DESCA( 2 ) The BLACS context handle, indicating */
/*                                the BLACS process grid A is distribu- */
/*                                ted over. The context itself is glo- */
/*                                bal, but the handle (the integer */
/*                                value) may vary. */
/*  N_A    (global) DESCA( 3 ) The size of the array dimension being */
/*                                distributed. */
/*  NB_A   (global) DESCA( 4 ) The blocking factor used to distribute */
/*                                the distributed dimension of the array. */
/*  SRC_A  (global) DESCA( 5 ) The process row or column over which the */
/*                                first row or column of the array */
/*                                is distributed. */
/*  LLD_A  (local)  DESCA( 6 ) The leading dimension of the local array */
/*                                storing the local blocks of the distri- */
/*                                buted array A. Minimum value of LLD_A */
/*                                depends on TYPE_A. */
/*                                TYPE_A = 501: LLD_A >= */
/*                                   size of undistributed dimension, 1. */
/*                                TYPE_A = 502: LLD_A >=NB_A, 1. */
/*  Reserved        DESCA( 7 ) Reserved for future use. */

/*  ===================================================================== */
/*
*  .. Local Scalars ..
*/
    int            rc, p, q;
    HMPI_Group     *gid;
    MPI_Comm       *psdbtrfcomm;
    HSCALCONTEXT   *hctxt; 
    double         algotime;
   
/*
* Initialize the context ID
*/
    *ICTXT = HSCAL_CTXT_UNDEFINED;

/*
* Dynamically refresh the speeds of the processors
*/
    rc = hscal_psdbtrf_recon(
              N,
              BWL,
              BWU,
              JA,
              DESCA
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_RECON;
    }

/*
* Create a HeteroMPI group with the optimal grid arrangement 
* of processes
*/
    gid  = (HMPI_Group*)malloc(sizeof(HMPI_Group));

    if (gid == NULL)
    {
       return HSCAL_ERR_LOW_MEMORY;
    }

    rc = hscal_psdbtrf_group_heuristic_pauto_create(
              N,
              BWL,
              BWU,
              JA,
              DESCA,
              gid,
              hscal_model_psdbtrf,
              hscal_heuristic_fn_1xp
    );

    if (rc != HSCAL_SUCCESS)
    {
       return HSCAL_ERR_GROUP_CREATION;
    }

    if (HMPI_Is_free())
    {
       *ICTXT = HSCAL_CTXT_FOR_FREE_PROCS;

       hctxt  = (HSCALCONTEXT*)malloc(sizeof(HSCALCONTEXT));

       if (hctxt == NULL)
       {
          return HSCAL_ERR_LOW_MEMORY;
       }

       hctxt->time = -1.;
       hctxt->optp = -1;
       hctxt->optq = -1;
       hctxt->gid  = gid;

       rc = hscal_store_gid_to_ctxt_mapping(ICTXT, hctxt);

       return rc;
    }

    psdbtrfcomm = (MPI_Comm*)HMPI_Get_comm(gid);

    if (psdbtrfcomm == NULL)
    {
       return HSCAL_ERR_NULL_COMM; 
    } 

/* 
* Broadcast the estimated execution time 
*/ 
    algotime = HMPI_Group_timeof(gid);

/*
* Broadcast the optimal grid arrangement
*/
    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       p = d[0];

       free(d);

       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *psdbtrfcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               *psdbtrfcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           gid,
           &nd,
           &d
       );

       q = d[1];

       free(d);

       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *psdbtrfcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               *psdbtrfcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       return HSCAL_ERR_BCAST_OPTIMAL_GRID;
    }

    if (HMPI_Is_parent(gid))
    {
       HSCAL_DEBUG(
            __FILE__, __FUNCTION__, __LINE__,
            "Optimal number of process rows =%d, Optimal number of process columns =%d, " 
            "Estimated execution time=%f ",
            p, q, algotime
       );       
    }

/*
* Translate communicator to a BLACS handle
*/
    *ICTXT = Csys2blacs_handle(*psdbtrfcomm);

/*
* Form BLACS context
*/
    Cblacs_gridinit(ICTXT, "r", p, q);

/*
* Store the context to gid mapping
*/
    hctxt  = (HSCALCONTEXT*)malloc(sizeof(HSCALCONTEXT)); 
 
    if (hctxt == NULL) 
    { 
       return HSCAL_ERR_LOW_MEMORY; 
    } 
 
    hctxt->time = algotime; 
    hctxt->optp = p; 
    hctxt->optq = q; 
    hctxt->gid  = gid; 
 
    rc = hscal_store_gid_to_ctxt_mapping(ICTXT, hctxt); 

    return rc;
}
