
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_psgels_recon_tcomp(
   int TRANSI, 
   int M,
   int N,
   int NRHS
)
{
    double tcomp = 0;

    char   transac  = 'N';
    char   *transa  = &transac;

    char   transbc  = 'T';
    char   *transb  = &transbc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    float alpha, beta;

    alpha = -1.f;
    beta  = 1.f;

    reconm  = M/p;
    reconn  = N/q;
    reconk  = hscal_pilaenv_recon("S");

    lda = reconm;
    ldb = reconn;
    ldc = reconm;

    hscal_sgemm(
         transa,
         transb,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha),
         &lda,
         &ldb,
         ((char*)&beta),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_psgels_my_tcomp(
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psgels_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          M, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_psgels_my_tcomm(
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psgels_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          M, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgels_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int TRANSI,
   int M,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_psgels_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANSI,
          M, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_psgels_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int TRANSI,
   int M,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__2 = 2;
    hscalinteger c__3 = 3;
    hscalinteger c__8 = 8;
    hscalinteger c__4 = 4;
    hscalinteger c__12 = 12;
    hscalinteger c__0 = 0;
    hscalreal c_b26 = 0.f;
    hscalreal c_b46 = 1.f;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4, i__5;

    /* Local variables */
    hscalinteger lcm, lwf, ipw, lws, mpa0, mpb0, npb0, nqa0;
    hscalinteger lcmp;
    hscalinteger ltau, brow;
    hscallogical tpsd;
    hscalinteger idum1[2], idum2[2], iacol, ibcol, iascl, ibscl;
    hscalinteger iarow, ibrow, lwmin, ictxt;
    hscalinteger icoffa, icoffb, iroffa, iroffb, scllen;
    hscalinteger nrhsqb0;
    int     info_l, *info = &info_l;

    char   transc  = TRANSI;
    char   *trans  = &transc;
    int     *m      = &M;
    int     *n      = &N;
    int     *nrhs   = &NRHS;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *ib     = &IB;
    int     *jb     = &JB;
    int     *descb  = (int*)DESCB;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descb;
    --desca;

/*     Test the input parameters */
    if (init_llds_f == 1)
    {
       int *PXGELS_DESCA_l = &desca[1];
       int *PXGELS_DESCB_l = &descb[1];
       HSCAL_PXGELS_INIT_LLDS(
            PXGELS_DESCA_l,
            PXGELS_DESCB_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "psgels: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d", 
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "psgels: (%d, %d): (descb+1)[M1_]=%d (descb+1)[MB1_]=%d myrow=%d (descb+1)[RSRC1_]=%d nprow=%d (descb+1)[LLD1_]=%d",
        myrow, mycol, (descb+1)[M1_], (descb+1)[MB1_], myrow, (descb+1)[RSRC1_], nprow, (descb+1)[LLD1_]
    );

/*     Test the input parameters */

    *info = 0;
    if (nprow == -1) {
	*info = -802;
    } else {
	hscal_chk1mat(m, &c__2, n, &c__3, ia, ja, &desca[1], &c__8, info,
                      myrow, mycol, nprow, npcol
      );
	hscal_chk1mat(n, &c__3, nrhs, &c__4, ib, jb, &descb[1], &c__12, info,
                      myrow, mycol, nprow, npcol
      );
	if (*info == 0) {
	    iroffa = (*ia - 1) % desca[5];
	    icoffa = (*ja - 1) % desca[6];
	    iroffb = (*ib - 1) % descb[5];
	    icoffb = (*jb - 1) % descb[6];
	    iarow = indxg2p_(ia, &desca[5], &myrow, &desca[7], &nprow);
	    iacol = indxg2p_(ia, &desca[6], &mycol, &desca[8], &npcol);
	    i__1 = *m + iroffa;
	    mpa0 = numroc_(&i__1, &desca[5], &myrow, &iarow, &nprow);
	    i__1 = *n + icoffa;
	    nqa0 = numroc_(&i__1, &desca[6], &mycol, &iacol, &npcol);

	    ibrow = indxg2p_(ib, &descb[5], &myrow, &descb[7], &nprow);
	    ibcol = indxg2p_(ib, &descb[6], &mycol, &descb[8], &npcol);
	    i__1 = *nrhs + icoffb;
	    nrhsqb0 = numroc_(&i__1, &descb[6], &mycol, &ibcol, &npcol);
	    if (*m >= *n) {
		i__1 = *m + iroffb;
		mpb0 = numroc_(&i__1, &descb[5], &myrow, &ibrow, &nprow);
		i__1 = *ja + min(*m,*n) - 1;
		ltau = numroc_(&i__1, &desca[6], &mycol, &desca[8], &npcol);
		lwf = desca[6] * (mpa0 + nqa0 + desca[6]);
/* Computing MAX */
		i__1 = desca[6] * (desca[6] - 1) / 2, i__2 = (mpb0 + nrhsqb0) 
			* desca[6];
		lws = max(i__1,i__2) + desca[6] * desca[6];
	    } else {
		lcm = ilcm_(&nprow, &npcol);
		lcmp = lcm / nprow;
		i__1 = *n + iroffb;
		npb0 = numroc_(&i__1, &descb[5], &myrow, &ibrow, &nprow);
		i__1 = *ia + min(*m,*n) - 1;
		ltau = numroc_(&i__1, &desca[5], &myrow, &desca[7], &nprow);
		lwf = desca[5] * (mpa0 + nqa0 + desca[5]);
/* Computing MAX */
/* Computing MAX */
		i__5 = *n + iroffb;
		i__4 = numroc_(&i__5, &desca[5], &c__0, &c__0, &nprow);
		i__3 = nqa0 + numroc_(&i__4, &desca[5], &c__0, &c__0, &lcmp);
		i__1 = desca[5] * (desca[5] - 1) / 2, i__2 = (npb0 + max(i__3,
			nrhsqb0)) * desca[5];
		lws = max(i__1,i__2) + desca[5] * desca[5];
	    }
	    lwmin = ltau + max(lwf,lws);

	    tpsd = 1;
	    if (hscal_lsame(trans, "N")) {
		tpsd = 0;
	    }

	    if (! (hscal_lsame(trans, "N") || hscal_lsame(trans, 
		    "T"))) {
		*info = -1;
	    } else if (*m < 0) {
		*info = -2;
	    } else if (*n < 0) {
		*info = -3;
	    } else if (*nrhs < 0) {
		*info = -4;
	    } else if (*m >= *n && iroffa != iroffb) {
		*info = -10;
	    } else if (*m >= *n && iarow != ibrow) {
		*info = -10;
	    } else if (*m < *n && icoffa != iroffb) {
		*info = -10;
	    } else if (*m >= *n && desca[5] != descb[5]) {
		*info = -1205;
	    } else if (*m < *n && desca[6] != descb[5]) {
		*info = -1205;
	    }
	}

	if (! tpsd) {
	    idum1[0] = 'N';
	} else {
	    idum1[0] = 'T';
	}
	idum2[0] = 1;
        idum1[1] = 1;
	idum2[1] = 14;
	hscal_pchk2mat(
            m, &c__2, n, &c__3, ia, ja, &desca[1], &c__8, n, &c__3, 
		nrhs, &c__4, ib, jb, &descb[1], &c__12, &c__2, idum1, idum2, 
		info, &myrow, &mycol, &nprow, &npcol
      );
    }

    if (*info != 0) {
	i__1 = -(*info);
       hscal_PB_Cabort( 
            "PSGELS", 
            i__1,
            nprow, npcol, myrow, mycol
       ); 
       return HSCAL_SUCCESS;
    }

/*     Quick return if possible */

/* Computing MIN */
    i__1 = min(*m,*n);
    if (min(i__1,*nrhs) == 0) {
	i__1 = max(*m,*n);
      {
         hscal_pslaset(
              "Full", &i__1, nrhs, &c_b26, &c_b26, ib, jb, &descb[1],
              nprow, npcol, myrow, mycol, tcomp
         );
      }
	return HSCAL_SUCCESS;
    }

/*     Scale A, B if max entry outside range [SMLNUM,BIGNUM] */
/*
* For the performance model, just call one scaling routine.
* We ignore the matrix all zeroes situation
*/
    hscal_pslange(
           "M", m, n, ia, ja, &desca[1], 
           nprow, npcol, myrow, mycol, tcomp, tcomm
    );

    {
        hscal_pslascl( 
             "G", m, n, ia, ja, &desca[1],
             nprow, npcol, myrow, mycol, tcomp
        );  
    }

    brow = *m;
    if (tpsd) {
	brow = *n;
    }

/*
* For the performance model, just call one scaling routine.
* We ignore the matrix all zeroes situation
*/
    hscal_pslange(
           "M", &brow, nrhs, ib, jb, &descb[1],
           nprow, npcol, myrow, mycol, tcomp, tcomm
    );

    {
        hscal_pslascl(
             "G", &brow, nrhs, ib, jb, &descb[1],
             nprow, npcol, myrow, mycol, tcomp
        );
    }

    ipw = ltau + 1;

    if (*m >= *n) {

/*        compute QR factorization of A */

	hscal_psgeqrf_my_tcomp_tcomm(
           0,
           myrow, mycol,
           M, N, IA, JA, &desca[1],
           nprow, npcol,
           tcomp, tcomm
	);

/*        workspace at least N, optimally N*NB */

	if (! tpsd) {

/*           Least-Squares Problem min || A * X - B || */

/*           B(IB:IB+M-1,JB:JB+NRHS-1) := Q' * B(IB:IB+M-1,JB:JB+NRHS-1) */

          {
             hscal_psormqr_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
                  'L', 
                  'T',
                  M, NRHS, N, IA, JA,
		      &desca[1], IB, JB, &descb[1],
                  nprow, npcol,
                  tcomp, tcomm
             );
          }

/*           workspace at least NRHS, optimally NRHS*NB */

/*           B(IB:IB+N-1,JB:JB+NRHS-1) := inv(R) * */
/*                                        B(IB:IB+N-1,JB:JB+NRHS-1) */

          {
             int ALPHA = c_b46;

             hscal_pstrsm_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
                  'L', 'U', 'N', 'N', 
                  N, NRHS, ALPHA, IA, JA, &desca[1], IB, JB, &descb[1], 
                  nprow, npcol,
                  tcomp, tcomm
             );
          }

	    scllen = *n;

	} else {

/*           Overdetermined system of equations sub( A )' * X = sub( B ) */

/*           sub( B ) := inv(R') * sub( B ) */

          {
             int ALPHA = c_b46;

	       hscal_pstrsm_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
                  'L', 'U', 'T', 'N', N, 
                  NRHS, ALPHA, IA, JA, &desca[1], IB, JB, &descb[1], 
                  nprow, npcol,
                  tcomp, tcomm
             );
          }

/*           B(IB+N:IB+M-1,JB:JB+NRHS-1) = ZERO */

	    i__1 = *m - *n;
	    i__2 = *ib + *n;
          {
             hscal_pslaset(
                  "All", &i__1, nrhs, &c_b26, &c_b26, &i__2, jb, &descb[1],
                  nprow, npcol, myrow, mycol, tcomp
             );
          }

/*           B(IB:IB+M-1,JB:JB+NRHS-1) := Q(1:N,:) * */
/*                                        B(IB:IB+N-1,JB:JB+NRHS-1) */

          {
             hscal_psormqr_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
  	            'L', 'N', M, NRHS, N, 
                  IA, JA, &desca[1], IB, JB, &descb[1],
                  nprow, npcol,
                  tcomp, tcomm
             );
          }

/*           workspace at least NRHS, optimally NRHS*NB */

	    scllen = *m;

	}

    } else {

/*        Compute LQ factorization of sub( A ) */

	hscal_psgelqf_my_tcomp_tcomm(
           0,
           myrow, mycol,
           M, N, IA, JA, &desca[1],
           nprow, npcol,
           tcomp, tcomm
      );

/*        workspace at least M, optimally M*NB. */

	if (! tpsd) {

/*           underdetermined system of equations sub( A ) * X = sub( B ) */

/*           B(IB:IB+M-1,JB:JB+NRHS-1) := inv(L) * */
/*                                        B(IB:IB+M-1,JB:JB+NRHS-1) */

          {
             int ALPHA = c_b46;
             hscal_pstrsm_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
	            'L', 'L', 'N', 'N', M, NRHS, ALPHA,
		      IA, JA, &desca[1], IB, JB, &descb[1], 
                  nprow, npcol,
                  tcomp, tcomm
             );
          }

/*           B(IB+M:IB+N-1,JB:JB+NRHS-1) = 0 */

	    i__1 = *n - *m;
	    i__2 = *ib + *m;
          {
             hscal_pslaset(
                  "All", &i__1, nrhs, &c_b26, &c_b26, &i__2, jb, &descb[1],
                  nprow, npcol, myrow, mycol, tcomp
             );
          }

/*           B(IB:IB+N-1,JB:JB+NRHS-1) := Q(1:N,:)' * */
/*                                        B(IB:IB+M-1,JB:JB+NRHS-1) */

	    hscal_psormlq_my_tcomp_tcomm(
               0,
               myrow, mycol,
               'L', 'T', N, NRHS, M, IA, JA,
               &desca[1], IB, JB, &descb[1],
               nprow, npcol,
               tcomp, tcomm
          );

/*           workspace at least NRHS, optimally NRHS*NB */

	    scllen = *n;

	} else {

/*           overdetermined system min || A' * X - B || */

/*           B(IB:IB+N-1,JB:JB+NRHS-1) := Q * B(IB:IB+N-1,JB:JB+NRHS-1) */

	    hscal_psormlq_my_tcomp_tcomm(
               0,
               myrow, mycol,
               'L', 'N', N, NRHS, M, IA, JA, &desca[1], 
               IB, JB, &descb[1], 
               nprow, npcol,
               tcomp, tcomm
          );

/*           workspace at least NRHS, optimally NRHS*NB */

/*           B(IB:IB+M-1,JB:JB+NRHS-1) := inv(L') * */
/*                                        B(IB:IB+M-1,JB:JB+NRHS-1) */
          {
             int ALPHA = c_b46;
	       hscal_pstrsm_my_tcomp_tcomm(
                  0,
                  myrow, mycol,
                  'L', 'L', 'T', 'N', M, 
                  NRHS, ALPHA, IA, JA, &desca[1], IB, JB, 
                  &descb[1],
                  nprow, npcol,
                  tcomp, tcomm
             );
          }

	    scllen = *m;

	}

    }

/*     Undo scaling */
/*
* For the performance model, just call one scaling routine.
* We ignore the matrix all zeroes situation
*/
    {
      hscal_pslascl(
           "G", &scllen, nrhs, ib, jb, &descb[1],
           nprow, npcol, myrow, mycol, tcomp
      );
    }

    {
      hscal_pslascl(
           "G", &scllen, nrhs, ib, jb, &descb[1],
           nprow, npcol, myrow, mycol, tcomp
      );
    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgels_tcomp_tcomm(
   int TRANSI,
   int M,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgels_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANSI,
                M, N, NRHS,
                IA, JA, DESCA,
                IB, JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgels_info(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgels_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *TRANS,
                *M, *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgels_info_i(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgels_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *TRANS,
                *M, *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
