
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_psgetf2_recon_tcomp(
   int M,
   int N
)
{
    double tcomp = 1;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_psgetf2_my_tcomp(
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psgetf2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_psgetf2_my_tcomm(
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psgetf2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgetf2_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_psgetf2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_psgetf2_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__1 = 1;
    hscalinteger c__2 = 2;
    hscalinteger c__6 = 6;
    hscalreal c_b12 = -1.f;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4, i__5, i__6, i__7;

    /* Local variables */
    hscalinteger i__, j;
    hscalinteger mn, iia, jja;
    hscalinteger icoff, iacol, iroff, iarow, ictxt;
    char rowbtop[1];
    HSCALBLACSCONTEXT   * blacsctxt;

    int     *m      = &M;
    int     *n      = &N;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --desca;

    if (init_llds_f == 1)
    {
       HSCAL_PXGETF2_INIT_LLDS(
           (desca+1),
           myrow, nprow
       );
    }

    *info = 0;
    if (nprow == -1) {
	*info = -602;
    } else {
	hscal_chk1mat(m, &c__1, n, &c__2, ia, ja, &desca[1], &c__6, info,
                      myrow, mycol, nprow, npcol
        );
	if (*info == 0) {
	    iroff = (*ia - 1) % desca[5];
	    icoff = (*ja - 1) % desca[6];
	    if (*n + icoff > desca[6]) {
		*info = -2;
	    } else if (iroff != 0) {
		*info = -4;
	    } else if (icoff != 0) {
		*info = -5;
	    } else if (desca[5] != desca[6]) {
		*info = -606;
	    }
	}
    }

    if (*info != 0) {
	i__1 = -(*info);
        hscal_PB_Cabort(
             "PSGETF2",
             i__1,
             nprow, npcol, myrow, mycol
        );
        return HSCAL_ERR_INPUT;
    }

/*     Quick return if possible */

    if (*m == 0 || *n == 0) {
	return HSCAL_SUCCESS;
    }

    mn = min(*m,*n);

    {
	i__1 = *ja + mn - 1;
	for (j = *ja; j <= i__1; ++j) {
	    i__ = *ia + j - *ja;

/*           Find pivot and test for singularity. */

	    i__2 = *m - j + *ja;
            {
               int *PSAMAX_DESCA_l = &desca[1];
	       hscal_psamax_my_tcomp_tcomm(
                    0,
                    myrow, mycol, i__2, i__, j, PSAMAX_DESCA_l,
		    c__1, nprow, npcol, tcomp, tcomm);
            }

/*
*  Assuming the worst case scenario for performance models
*  No elements have zeroes
*/
	    {

/*              Apply the row interchanges to columns JA:JA+N-1 */
/*
*  Assuming the worst case scenario for performance models
*  A swap for each iteration
*  Also it does not matter what the indices IX, JX, IY, JY are
*  As long as we know they are aligned the same way
*/
                {
                   int *PSSWAP_DESCA1_l = &desca[1];
                   int PSSWAP_DESCA3_l  = desca[3];
		   hscal_psswap_my_tcomp_tcomm(
                        0,
                        myrow, mycol,
                        N, i__, JA, PSSWAP_DESCA1_l, PSSWAP_DESCA3_l,
		        i__, JA, PSSWAP_DESCA1_l, PSSWAP_DESCA3_l,
                        nprow, npcol, tcomp, tcomm
                   );
                }

/*              Compute elements I+1:IA+M-1 of J-th column. */

		if (j - *ja + 1 < *m) {
		    i__2 = *m - j + *ja - 1;
		    i__3 = i__ + 1;
/*
* Just an indicative multiplicative scalar
*/
                    {
                       int ALPHAI = 1;
                       int *PSSCAL_DESCA_l = &desca[1];

	   	       ihscal_psscal_tcomp(
                            0,
                            myrow, mycol,
                            i__2, ALPHAI, i__3, j, PSSCAL_DESCA_l, c__1,
                            nprow, npcol, tcomp
                       );
                    }
		}
	    }

/*           Update trailing submatrix */

	    if (j - *ja + 1 < mn) {
		i__2 = *m - j + *ja - 1;
		i__3 = *n - j + *ja - 1;
		i__4 = i__ + 1;
		i__5 = j + 1;
		i__6 = i__ + 1;
		i__7 = j + 1;
                {
                   int *PSGERU_DESCA1_l = &desca[1];
                   int PSGERU_DESCA3_l  = desca[3];
                   int ALPHAI = c_b12;

		   hscal_psger_my_tcomp_tcomm(
                        0,
                        myrow, mycol,
                        i__2, i__3, ALPHAI, i__4, j, PSGERU_DESCA1_l, 
			c__1, i__, i__5, PSGERU_DESCA1_l, PSGERU_DESCA3_l,
			i__6, i__7, PSGERU_DESCA1_l, nprow, npcol, tcomp, tcomm);
                }
	    }
/* L10: */
	}
    }

/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);
    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);

    infog2l_(ia, ja, &desca[1], &nprow, &npcol, &myrow, &mycol, &iia, &jja, &
	    iarow, &iacol);
    if (mycol == iacol) {
	hscal_Cigebs2d(blacsctxt, "Rowwise", rowbtop, mn, c__1, mn,
             nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
        );
    } else {
	hscal_Cigebr2d(blacsctxt, "Rowwise", rowbtop, mn, c__1, mn,
		myrow, iacol,
                nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
        );
    }

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgetf2_tcomp_tcomm(
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgetf2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                M, N,
                IA, JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgetf2_info(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgetf2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgetf2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgetf2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

