
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_psgetrs_recon_tcomp(
   int TRANSI,
   int N,
   int NRHS
)
{
    double tcomp   = 0;

    char   transc  = 'N';
    char   *trans   = &transc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    float alpha = 1.;

    if (transc == 'N')
    {
       reconm  = N/p;
       reconn  = NRHS/q;
       reconk  = hscal_pilaenv_recon("S");
    }
    else
    {
       reconm  = hscal_pilaenv_recon("S");
       reconn  = N/p;
       reconk  = NRHS/q;
    }

    if (transc == 'N')
    {
       lda = reconm;
    }
    else
    {
       lda = reconk;
    }

    if (transc == 'N')
    {
       ldb = reconk;
    }
    else
    {
       ldb = reconn;
    }

    ldc = reconm;

    hscal_sgemm(
         trans,
         trans,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha),
         &lda,
         &ldb,
         ((char*)&alpha),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_psgetrs_my_tcomp(
   int myrow,
   int mycol,
   int TRANS,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psgetrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANS, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_psgetrs_my_tcomm(
   int myrow,
   int mycol,
   int TRANS,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psgetrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANS, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgetrs_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int TRANS,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_psgetrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          TRANS, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_psgetrs_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int TRANSI,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__2 = 2;
    hscalinteger c__7 = 7;
    hscalinteger c__3 = 3;
    hscalinteger c__12 = 12;
    hscalinteger c__1 = 1;
    hscalreal c_b33 = 1.f;

    /* System generated locals */
    hscalinteger i__1, i__2;

    /* Local variables */
    hscalinteger idum1[1], idum2[1];
    hscalinteger iarow, ibrow, ictxt, icoffa, 
	    iroffa, iroffb;
    hscallogical notran;

/*
* .. Intermediate outputs and inputs ..
*/
    int descip[DLEN1_];

    int     *n      = &N;
    int     *nrhs   = &NRHS;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *ib     = &IB;
    int     *jb     = &JB;
    int     *descb  = (int*)DESCB;

    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    char transc = TRANSI, *trans = &transc;

    /* Parameter adjustments */
    --descb;
    --desca;

    if (init_llds_f == 1)
    {
       HSCAL_PXGETRS_INIT_LLDS(
            (desca+1),
            (descb+1),
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PSGETRS: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d",
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PSGETRS: (%d, %d): (descb+1)[M1_]=%d (descb+1)[MB1_]=%d myrow=%d (descb+1)[RSRC1_]=%d nprow=%d (descb+1)[LLD1_]=%d",
        myrow, mycol, (descb+1)[M1_], (descb+1)[MB1_], myrow, (descb+1)[RSRC1_], nprow, (descb+1)[LLD1_]
    );

    *info = 0;
    if (nprow == -1) {
	*info = -702;
    } else {
	notran = hscal_lsame(trans, "N");
	hscal_chk1mat(n, &c__2, n, &c__2, ia, ja, &desca[1], &c__7, info,
                      myrow, mycol, nprow, npcol);
	hscal_chk1mat(n, &c__2, nrhs, &c__3, ib, jb, &descb[1], &c__12, info,
                      myrow, mycol, nprow, npcol);
	if (*info == 0) {
	    iarow = indxg2p_(ia, &desca[5], &myrow, &desca[7], &nprow);
	    ibrow = indxg2p_(ib, &descb[5], &myrow, &descb[7], &nprow);
	    iroffa = (*ia - 1) % desca[5];
	    icoffa = (*ja - 1) % desca[6];
	    iroffb = (*ib - 1) % descb[5];
	    if (! notran && ! hscal_lsame(trans, "T") && ! 
		    hscal_lsame(trans, "C")) {
		*info = -1;
	    } else if (iroffa != 0) {
		*info = -5;
	    } else if (icoffa != 0) {
		*info = -6;
	    } else if (desca[5] != desca[6]) {
		*info = -706;
	    } else if (iroffb != 0 || ibrow != iarow) {
		*info = -10;
	    } else if (descb[5] != desca[6]) {
		*info = -1206;
	    }
	}
	if (notran) {
	    idum1[0] = 'N';
	} else if (hscal_lsame(trans, "T")) {
	    idum1[0] = 'T';
	} else {
	    idum1[0] = 'C';
	}
	idum2[0] = 1;
	hscal_pchk2mat(n, &c__2, n, &c__2, ia, ja, &desca[1], &c__7, n, &c__2, 
		nrhs, &c__3, ib, jb, &descb[1], &c__12, &c__1, idum1, idum2, 
		info, &myrow, &mycol, &nprow, &npcol);
    }

    if (*info != 0) {
	i__1 = -(*info);
        hscal_PB_Cabort(
             "PSGETRS",
             i__1,
             nprow, npcol, myrow, mycol
        );
        return HSCAL_ERR_INPUT;
    }

/*     Quick return if possible */

    if (*n == 0 || *nrhs == 0) {
	HSCAL_SUCCESS;
    }

    i__1 = desca[3] + desca[5] * nprow;
    i__2 = desca[5] + numroc_(&desca[3], &desca[5], &myrow, &desca[7], &nprow)
	    ;
    descset_(descip, &i__1, &c__1, &desca[5], &c__1, &desca[7], &mycol, &
	    ictxt, &i__2);

    if (notran) {

/*        Solve sub( A ) * X = sub( B ). */

/*        Apply row interchanges to the right hand sides. */

        {
          int *PSLAPIV_DESCB_l = (&descb[1]);

	  hscal_pslapiv_my_tcomp_tcomm(
                0,
                myrow, mycol,
                'F', 'R', 'C', N, NRHS, IB, JB, PSLAPIV_DESCB_l,
		IA, c__1, descip,
		nprow, npcol, tcomp, tcomm);
        }

/*        Solve L*X = sub( B ), overwriting sub( B ) with X. */

        {
          int *PSTRSM_DESCA_l = (&desca[1]);
          int *PSTRSM_DESCB_l = (&descb[1]);
          int ALPHA33         = c_b33;

          hscal_pstrsm_my_tcomp_tcomm(
                0,
                myrow, mycol,
                'L', 'L', 'N', 'U', N, NRHS, ALPHA33,
		IA, JA, PSTRSM_DESCA_l, IB, JB, PSTRSM_DESCB_l,
		nprow, npcol, tcomp, tcomm);
        }

/*        Solve U*X = sub( B ), overwriting sub( B ) with X. */

        {
          int *PSTRSM_DESCA_l = (&desca[1]);
          int *PSTRSM_DESCB_l = (&descb[1]);
          int ALPHA33         = c_b33;

	  hscal_pstrsm_my_tcomp_tcomm(
                0,
                myrow, mycol,
                'L', 'U', 'N', 'N', N, NRHS, ALPHA33, 
		IA, JA, PSTRSM_DESCA_l, IB, JB, PSTRSM_DESCB_l,
		nprow, npcol, tcomp, tcomm);
        }
    } else {

/*        Solve sub( A )' * X = sub( B ). */

/*        Solve U'*X = sub( B ), overwriting sub( B ) with X. */

        {
          int *PSTRSM_DESCA_l = (&desca[1]);
          int *PSTRSM_DESCB_l = (&descb[1]);
          int ALPHA33         = c_b33;

	  hscal_pstrsm_my_tcomp_tcomm(
                0,
                myrow, mycol,
                'L', 'U', TRANSI, 'N', N, NRHS, ALPHA33,
		IA, JA, PSTRSM_DESCA_l, IB, JB, PSTRSM_DESCB_l,
		nprow, npcol, tcomp, tcomm);
        }

/*        Solve L'*X = sub( B ), overwriting sub( B ) with X. */

        {
          int *PSTRSM_DESCA_l = (&desca[1]);
          int *PSTRSM_DESCB_l = (&descb[1]);
          int ALPHA33         = c_b33;

	  hscal_pstrsm_my_tcomp_tcomm(
                0,
                myrow, mycol,
                'L', 'L', TRANSI, 'U', N, NRHS, ALPHA33, IA, 
		JA, PSTRSM_DESCA_l, IB, JB, PSTRSM_DESCB_l,
		nprow, npcol, tcomp, tcomm);
        }

/*        Apply row interchanges to the solution vectors. */

        {
          int *PSLAPIV_DESCB_l = (&descb[1]);
	  hscal_pslapiv_my_tcomp_tcomm(
                0,
                myrow, mycol,
                'B', 'R', 'C', N, NRHS, IB, JB, PSLAPIV_DESCB_l, 
		IA, c__1, descip,
		nprow, npcol, tcomp, tcomm);
        }

    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgetrs_tcomp_tcomm(
   int TRANS,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgetrs_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                TRANS, N, NRHS,
                IA, JA, DESCA,
                IB, JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgetrs_info(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgetrs_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *TRANS, *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgetrs_info_i(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psgetrs_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *TRANS, *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
