
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pslaswp_recon_tcomp(
   int DIRECI,
   int ROWCOLI,
   int N
)
{
    double tcomp = 1;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pslaswp_my_tcomp(
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int K1,
   int K2,
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pslaswp_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI,
          N,
          IA, JA, DESCA, 
          K1, K2,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pslaswp_my_tcomm(
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int K1,
   int K2,
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pslaswp_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI,
          N,
          IA, JA, DESCA,
          K1, K2,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pslaswp_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int DIRECI,
   int ROWCOLI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int K1,
   int K2,
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pslaswp_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          DIRECI, ROWCOLI,
          N,
          IA, JA, DESCA,
          K1, K2,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pslaswp_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int DIRECI,
   int ROWCOLI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int K1,
   int K2,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__1 = 1;

    /* System generated locals */
    hscalinteger i__1;

    /* Local variables */
    hscalinteger i__, j, ip, jp, iia, jja;
    hscalinteger icurcol, icurrow;

    char    DIRECC  = DIRECI, *direc = &DIRECC;
    char    ROWCOLC = ROWCOLI, *rowcol = &ROWCOLC;
    int     *n      = &N;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *k1     = &K1;
    int     *k2     = &K2;
    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    --desca;

    /* Function Body */
    if (*n == 0) {
       return HSCAL_SUCCESS;
    }

    if (init_llds_f == 1)
    {
       HSCAL_PXLASWP_INIT_LLDS(
           (desca+1),
           myrow, nprow
       );
    }

    if (hscal_lsame(rowcol, "R")) {
	if (hscal_lsame(direc, "F")) {
	    i__1 = *k2;
	    for (i__ = *k1; i__ <= i__1; ++i__) {
/*
*  Assuming the worst case scenario for performance models
*/
		{
                    int *PSSWAP_DESCA1_l = &desca[1];
                    int PSSWAP_DESCA3_l  = desca[3];
		    hscal_pdswap_my_tcomp_tcomm(
                            0,
                            myrow, mycol,
                            N, i__, JA, PSSWAP_DESCA1_l, PSSWAP_DESCA3_l,
			    i__, JA, PSSWAP_DESCA1_l, PSSWAP_DESCA3_l,
                            nprow, npcol, tcomp, tcomm);
		}
/* L10: */
	    }
	} else {
	    i__1 = *k1;
	    for (i__ = *k2; i__ >= i__1; --i__) {
/*
*  Assuming the worst case scenario for performance models
*/
		{
                    int *PSSWAP_DESCA1_l = &desca[1];
                    int PSSWAP_DESCA3_l  = desca[3];
		    hscal_pdswap_my_tcomp_tcomm(
                            0,
                            myrow, mycol,
                            N, i__, JA, PSSWAP_DESCA1_l, PSSWAP_DESCA3_l,
			    i__, JA, PSSWAP_DESCA1_l, PSSWAP_DESCA3_l,
                            nprow, npcol, tcomp, tcomm);
		}
/* L20: */
	    }
	}
    } else {
	if (hscal_lsame(direc, "F")) {
	    i__1 = *k2;
	    for (j = *k1; j <= i__1; ++j) {
/*
*  Assuming the worst case scenario for performance models
*/
		{
                    int *PSSWAP_DESCA1_l = &desca[1];

		    hscal_pdswap_my_tcomp_tcomm(
                            0,
                            myrow, mycol,
                            N, IA, j, PSSWAP_DESCA1_l, c__1, IA, j,
			    PSSWAP_DESCA1_l, c__1,
                            nprow, npcol, tcomp, tcomm);
		}
/* L30: */
	    }
	} else {
	    i__1 = *k1;
	    for (j = *k2; j >= i__1; --j) {
/*
*  Assuming the worst case scenario for performance models
*/
		{
                    int *PSSWAP_DESCA1_l = &desca[1];

		    hscal_pdswap_my_tcomp_tcomm(
                            0,
                            myrow, mycol,
                            N, IA, j, PSSWAP_DESCA1_l, c__1, IA, j,
			    PSSWAP_DESCA1_l, c__1,
                            nprow, npcol, tcomp, tcomm);
		}
/* L40: */
	    }
	}
    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pslaswp_tcomp_tcomm(
   int DIRECI,
   int ROWCOLI,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int K1,
   int K2,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pslaswp_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                DIRECI, ROWCOLI,
                N,
                IA, JA, DESCA,
                K1, K2,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pslaswp_info(
   char *DIRECC,
   char *ROWCOLC,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * K1,
   int * K2,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pslaswp_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *DIRECC, *ROWCOLC,
                *N,
                *IA, *JA, DESCA,
                *K1, *K2,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pslaswp_info_i(
   char *DIRECC,
   char *ROWCOLC,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * K1,
   int * K2,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pslaswp_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *DIRECC, *ROWCOLC,
                *N,
                *IA, *JA, DESCA,
                *K1, *K2,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

