
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_psorm2r_recon_tcomp(
   int sidei,
   int transi,
   int m,
   int n,
   int k
)
{
    double tcomp = 0;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_psorm2r_my_tcomp(
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psorm2r_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_psorm2r_my_tcomm(
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psorm2r_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psorm2r_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_psorm2r_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_psorm2r_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__3 = 3;
    hscalinteger c__5 = 5;
    hscalinteger c__9 = 9;
    hscalinteger c__4 = 4;
    hscalinteger c__14 = 14;
    hscalinteger c__0 = 0;
    hscalinteger c__1 = 1;
    hscalreal c_b52 = 1.f;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4;

    /* Local variables */
    hscalinteger j, j1, j2, j3;
    hscalinteger ii, jj, mi, ni, mp, nq, icc, jcc;
    hscalreal ajj;
    hscalinteger lcm;
    hscalinteger mpc0, nqc0;
    hscallogical left;
    hscalinteger lcmq;
    hscalinteger iacol, iccol;
    hscalinteger iarow, icrow, lwmin, ictxt,
	    icoffc, iroffa, iroffc;
    hscallogical notran;
    char colbtop[1], rowbtop[1];
    int     info_l, *info = &info_l;

/*
*  .. Local Arrays ..
*/
    HSCALBLACSCONTEXT   * blacsctxt;

    char sidec      = SIDEI, *side = &sidec;
    char transc     = TRANSI, *trans = &transc;

    int     *m      = &M;
    int     *n      = &N;
    int     *k      = &K;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *ic     = &IC;
    int     *jc     = &JC;
    int     *descc  = (int*)DESCC;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descc;
    --desca;

    if (init_llds_f == 1)
    {
       int *PXORM2R_DESCA_l = &desca[1];
       int *PXORM2R_DESCC_l = &descc[1];
       HSCAL_PXORM2R_INIT_LLDS(
            PXORM2R_DESCA_l,
            PXORM2R_DESCC_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "psorm2r: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d", 
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "psorm2r: (%d, %d): (descc+1)[M1_]=%d (descc+1)[MB1_]=%d myrow=%d (descc+1)[RSRC1_]=%d nprow=%d (descc+1)[LLD1_]=%d",
        myrow, mycol, (descc+1)[M1_], (descc+1)[MB1_], myrow, (descc+1)[RSRC1_], nprow, (descc+1)[LLD1_]
    );

/*     Test the input parameters */

    *info = 0;
    if (nprow == -1) {
	*info = -902;
    } else {
	left = hscal_lsame(side, "L");
	notran = hscal_lsame(trans, "N");

/*        NQ is the order of Q */

	if (left) {
	    nq = *m;
	    hscal_chk1mat(
               m, &c__3, k, &c__5, ia, ja, &desca[1], &c__9, info,
               myrow, mycol, nprow, npcol
          );
	} else {
	    nq = *n;
	    hscal_chk1mat(
               n, &c__4, k, &c__5, ia, ja, &desca[1], &c__9, info,
               myrow, mycol, nprow, npcol
          );
	}
	hscal_chk1mat(
           m, &c__3, n, &c__4, ic, jc, &descc[1], &c__14, info,
           myrow, mycol, nprow, npcol
      );
	if (*info == 0) {
	    iroffa = (*ia - 1) % desca[5];
	    iroffc = (*ic - 1) % descc[5];
	    icoffc = (*jc - 1) % descc[6];
	    iarow = indxg2p_(ia, &desca[5], &myrow, &desca[7], &nprow);
	    icrow = indxg2p_(ic, &descc[5], &myrow, &descc[7], &nprow);
	    iccol = indxg2p_(jc, &descc[6], &mycol, &descc[8], &npcol);
	    i__1 = *m + iroffc;
	    mpc0 = numroc_(&i__1, &descc[5], &myrow, &icrow, &nprow);
	    i__1 = *n + icoffc;
	    nqc0 = numroc_(&i__1, &descc[6], &mycol, &iccol, &npcol);

	    if (left) {
		lwmin = mpc0 + max(1,nqc0);
	    } else {
		lcm = ilcm_(&nprow, &npcol);
		lcmq = lcm / npcol;
/* Computing MAX */
		i__4 = *n + icoffc;
		i__3 = numroc_(&i__4, &desca[6], &c__0, &c__0, &npcol);
		i__1 = max(1,mpc0), i__2 = numroc_(&i__3, &desca[6], &c__0, &
			c__0, &lcmq);
		lwmin = nqc0 + max(i__1,i__2);
	    }

	    if (! left && ! hscal_lsame(side, "R")) {
		*info = -1;
	    } else if (! notran && ! hscal_lsame(trans, "T"))
		     {
		*info = -2;
	    } else if (*k < 0 || *k > nq) {
		*info = -5;
	    } else if (! left && desca[5] != descc[6]) {
		*info = -906;
	    } else if (left && iroffa != iroffc) {
		*info = -12;
	    } else if (left && iarow != icrow) {
		*info = -12;
	    } else if (! left && iroffa != icoffc) {
		*info = -13;
	    } else if (left && desca[5] != descc[5]) {
		*info = -1405;
	    }
	}
    }

    if (*info != 0) {
	i__1 = -(*info);
       hscal_PB_Cabort( 
            "psorm2r", 
            i__1,
            nprow, npcol, myrow, mycol
       ); 
       return HSCAL_SUCCESS;
    }

/*     Quick return if possible */

    if (*m == 0 || *n == 0 || *k == 0) {
       return HSCAL_SUCCESS;
    }

    if (desca[3] == 1) {
      Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);
	infog2l_(ia, ja, &desca[1], &nprow, &npcol, &myrow, &mycol, &ii, &jj, 
		&iarow, &iacol);
	infog2l_(ic, jc, &descc[1], &nprow, &npcol, &myrow, &mycol, &icc, &
		jcc, &icrow, &iccol);
	if (left) {
	    if (myrow == iarow) {
		i__1 = *jc + *n - 1;
		nq = numroc_(&i__1, &descc[6], &mycol, &descc[8], &npcol);
		if (mycol == iacol) {
		    ajj = .5f;
		    hscal_Csgebs2d(
                     blacsctxt, "Rowwise", " ", c__1, c__1,
			   c__1,
                     nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                );
		    i__1 = nq - jcc + 1;
		    sscal_tcomp_macro(
                     &i__1, (char*)&ajj, &descc[9],
                     nprow, npcol, myrow, mycol, tcomp_l
                );
		} else {
		    ajj = .5f;
		    hscal_Csgebr2d(
                     blacsctxt, "Rowwise", " ", c__1, c__1,
			   c__1, iarow, iacol,
                     nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                );
		    i__1 = nq - jcc + 1;
		    sscal_tcomp_macro(
                     &i__1, (char*)&ajj, &descc[9],
                     nprow, npcol, myrow, mycol, tcomp_l
                );
		}
	    }
	} else {
	    if (mycol == iacol) {
		;
	    }

	    if (iacol != iccol) {
		if (mycol == iacol) {
		    hscal_Csgesd2d(
                     blacsctxt, c__1, c__1, c__1, myrow,
	               iccol,
                     nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                );
		}
		if (mycol == iccol) {
		    ;
		}
	    }

	    if (mycol == iccol) {
            ajj = .5f;
		i__1 = *ic + *m - 1;
		mp = numroc_(&i__1, &descc[5], &myrow, &descc[7], &nprow);
		i__1 = mp - icc + 1;
		sscal_tcomp_macro(
                 &i__1, (char*)&ajj, &c__1,
                 nprow, npcol, myrow, mycol, tcomp_l
            );
	    }

	}

      Cblacs_gridexit_nocomm(blacsctxt);

    } else {

	pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
	pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);

	if (left && ! notran || ! left && notran) {
	    j1 = *ja;
	    j2 = *ja + *k - 1;
	    j3 = 1;
	} else {
	    j1 = *ja + *k - 1;
	    j2 = *ja;
	    j3 = -1;
	}

	if (left) {
	    ni = *n;
	    jcc = *jc;
	    if (notran) {
		pb_topset__(&ictxt, "Broadcast", "Rowwise", "D-ring");
	    } else {
		pb_topset__(&ictxt, "Broadcast", "Rowwise", "I-ring");
	    }
	    pb_topset__(&ictxt, "Broadcast", "Columnwise", " ");
	} else {
	    mi = *m;
	    icc = *ic;
	}

	i__1 = j2;
	i__2 = j3;
	for (j = j1; i__2 < 0 ? j >= i__1 : j <= i__1; j += i__2) {
	    if (left) {

/*              H(j) or H(j)' is applied to C(ic+j-ja:ic+m-1,jc:jc+n-1) */

		mi = *m - j + *ja;
		icc = *ic + j - *ja;
	    } else {

/*              H(j) or H(j)' is applied to C(ic:ic+m-1,jc+j-ja:jc+n-1) */

		ni = *n - j + *ja;
		jcc = *jc + j - *ja;
	    }

/*           Apply H(j) or H(j)' */

	    i__3 = *ia + j - *ja;
          {
	    hscal_pselset2(
               &ajj, &i__3, &j, &desca[1], &c_b52,
               nprow, npcol,
               myrow, mycol,
               tcomp, tcomm
          );
          }
	    i__3 = *ia + j - *ja;
	    hscal_pslarf_my_tcomp_tcomm(
               0,
               myrow, mycol,
               SIDEI, mi, ni, i__3, j, &desca[1], &c__1,
               icc, jcc, &descc[1], 
               nprow, npcol,
               tcomp, tcomm
          );
	    i__3 = *ia + j - *ja;
          {
	    hscal_pselset(
               &i__3, &j, &desca[1], &ajj,
               nprow, npcol,
               myrow, mycol,
               tcomp, tcomm
          );
          }

/* L10: */
	}

	pb_topset__(&ictxt, "Broadcast", "Rowwise", rowbtop);
	pb_topset__(&ictxt, "Broadcast", "Columnwise", colbtop);

    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psorm2r_tcomp_tcomm(
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psorm2r_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                M, N, K,
                IA, JA, DESCA, 
                IC, JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psorm2r_info(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;
   int TRANSI = *TRANS;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psorm2r_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                *M, *N, *K,
                *IA, *JA, DESCA, 
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psorm2r_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;
   int TRANSI = *TRANS;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psorm2r_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                *M, *N, *K,
                *IA, *JA, DESCA, 
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
