
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_psormlq_recon_tcomp(
   int sidei,
   int transi,
   int m,
   int n,
   int k
)
{
    double tcomp = 0;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_psormlq_my_tcomp(
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psormlq_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_psormlq_my_tcomm(
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_psormlq_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psormlq_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_psormlq_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_psormlq_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__5 = 5;
    hscalinteger c__3 = 3;
    hscalinteger c__9 = 9;
    hscalinteger c__4 = 4;
    hscalinteger c__14 = 14;
    hscalinteger c__0 = 0;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4, i__5;

    /* Local variables */
    hscalinteger i__, i1, i2, i3;
    hscalinteger ib, mi, ni, nq, icc, jcc, lcm, ipw, mqa0, mpc0, nqc0;
    hscallogical left;
    hscalinteger lcmp, idum1[4], idum2[4];
    hscalinteger iacol, iccol;
    hscalinteger iinfo, icrow, lwmin, ictxt,
	    icoffa, icoffc, iroffc;
    hscallogical notran;
    char transt[1];
    char colbtop[1], rowbtop[1];
    int     info_l, *info = &info_l;

    char sidec      = SIDEI, *side = &sidec;
    char transc     = TRANSI, *trans = &transc;

    int     *m      = &M;
    int     *n      = &N;
    int     *k      = &K;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *ic     = &IC;
    int     *jc     = &JC;
    int     *descc  = (int*)DESCC;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descc;
    --desca;

    if (init_llds_f == 1)
    {
       int *PXUNMLQ_DESCA_l = &desca[1];
       int *PXUNMLQ_DESCC_l = &descc[1];
       HSCAL_PXUNMLQ_INIT_LLDS(
            PXUNMLQ_DESCA_l,
            PXUNMLQ_DESCC_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "psormlq: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d", 
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "psormlq: (%d, %d): (descc+1)[M1_]=%d (descc+1)[MB1_]=%d myrow=%d (descc+1)[RSRC1_]=%d nprow=%d (descc+1)[LLD1_]=%d",
        myrow, mycol, (descc+1)[M1_], (descc+1)[MB1_], myrow, (descc+1)[RSRC1_], nprow, (descc+1)[LLD1_]
    );

/*     Test the input parameters */

    *info = 0;
    if (nprow == -1) {
	*info = -902;
    } else {
	left = hscal_lsame(side, "L");
	notran = hscal_lsame(trans, "N");

/*        NQ is the order of Q */

	if (left) {
	    nq = *m;
	    hscal_chk1mat(
               k, &c__5, m, &c__3, ia, ja, &desca[1], &c__9, info,
               myrow, mycol, nprow, npcol
          );
	} else {
	    nq = *n;
	    hscal_chk1mat(
               k, &c__5, n, &c__4, ia, ja, &desca[1], &c__9, info,
               myrow, mycol, nprow, npcol
          );
	}
	hscal_chk1mat(
           m, &c__3, n, &c__4, ic, jc, &descc[1], &c__14, info,
           myrow, mycol, nprow, npcol
      );
	if (*info == 0) {
	    icoffa = (*ja - 1) % desca[6];
	    iroffc = (*ic - 1) % descc[5];
	    icoffc = (*jc - 1) % descc[6];
	    iacol = indxg2p_(ja, &desca[6], &mycol, &desca[8], &npcol);
	    icrow = indxg2p_(ic, &descc[5], &myrow, &descc[7], &nprow);
	    iccol = indxg2p_(jc, &descc[6], &mycol, &descc[8], &npcol);
	    i__1 = *m + iroffc;
	    mpc0 = numroc_(&i__1, &descc[5], &myrow, &icrow, &nprow);
	    i__1 = *n + icoffc;
	    nqc0 = numroc_(&i__1, &descc[6], &mycol, &iccol, &npcol);

	    if (left) {
		i__1 = *m + icoffa;
		mqa0 = numroc_(&i__1, &desca[6], &mycol, &iacol, &npcol);
		lcm = ilcm_(&nprow, &npcol);
		lcmp = lcm / nprow;
/* Computing MAX */
/* Computing MAX */
		i__5 = *m + iroffc;
		i__4 = numroc_(&i__5, &desca[5], &c__0, &c__0, &nprow);
		i__3 = mqa0 + numroc_(&i__4, &desca[5], &c__0, &c__0, &lcmp);
		i__1 = desca[5] * (desca[5] - 1) / 2, i__2 = (mpc0 + max(i__3,
			nqc0)) * desca[5];
		lwmin = max(i__1,i__2) + desca[5] * desca[5];
	    } else {
/* Computing MAX */
		i__1 = desca[5] * (desca[5] - 1) / 2, i__2 = (mpc0 + nqc0) * 
			desca[5];
		lwmin = max(i__1,i__2) + desca[5] * desca[5];
	    }

	    if (! left && ! hscal_lsame(side, "R")) {
		*info = -1;
	    } else if (! notran && ! hscal_lsame(trans, "T"))
		     {
		*info = -2;
	    } else if (*k < 0 || *k > nq) {
		*info = -5;
	    } else if (left && desca[6] != descc[5]) {
		*info = -906;
	    } else if (left && icoffa != iroffc) {
		*info = -12;
	    } else if (! left && icoffa != icoffc) {
		*info = -13;
	    } else if (! left && iacol != iccol) {
		*info = -13;
	    } else if (! left && desca[6] != descc[6]) {
		*info = -1406;
	    }
	}
	if (left) {
	    idum1[0] = 'L';
	} else {
	    idum1[0] = 'R';
	}
	idum2[0] = 1;
	if (notran) {
	    idum1[1] = 'N';
	} else {
	    idum1[1] = 'T';
	}
	idum2[1] = 2;
	idum1[2] = *k;
	idum2[2] = 5;
        idum1[3] = 1;
	idum2[3] = 16;
	if (left) {
	    hscal_pchk2mat(
               k, &c__5, m, &c__3, ia, ja, &desca[1], &c__9, m, &c__3, 
		   n, &c__4, ic, jc, &descc[1], &c__14, &c__4, idum1, idum2, 
               info, &myrow, &mycol, &nprow, &npcol
          );
	} else {
	    hscal_pchk2mat(
               k, &c__5, n, &c__4, ia, ja, &desca[1], &c__9, m, &c__3, 
		   n, &c__4, ic, jc, &descc[1], &c__14, &c__4, idum1, idum2, 
               info, &myrow, &mycol, &nprow, &npcol
          );
	}
    }

    if (*info != 0) {
	i__1 = -(*info);
       hscal_PB_Cabort( 
            "PSORMLQ", 
            i__1,
            nprow, npcol, myrow, mycol
       ); 
       return HSCAL_SUCCESS;
    }

/*     Quick return if possible */

    if (*m == 0 || *n == 0 || *k == 0) {
       return HSCAL_SUCCESS;
    }

    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);

    if (left && notran || ! left && ! notran) {
/* Computing MIN */
	i__1 = hscal_iceil(ia, &desca[5]) * desca[5], i__2 = *ia + *k - 1;
	i1 = min(i__1,i__2) + 1;
	i2 = *ia + *k - 1;
	i3 = desca[5];
    } else {
/* Computing MAX */
	i__1 = (*ia + *k - 2) / desca[5] * desca[5] + 1;
	i1 = max(i__1,*ia);
/* Computing MIN */
	i__1 = hscal_iceil(ia, &desca[5]) * desca[5], i__2 = *ia + *k - 1;
	i2 = min(i__1,i__2) + 1;
	i3 = -desca[5];
    }

    if (left) {
	ni = *n;
	jcc = *jc;
    } else {
	mi = *m;
	icc = *ic;
	pb_topset__(&ictxt, "Broadcast", "Rowwise", " ");
	if (notran) {
	    pb_topset__(&ictxt, "Broadcast", "Columnwise", "D-ring");
	} else {
	    pb_topset__(&ictxt, "Broadcast", "Columnwise", "I-ring");
	}
    }

    if (notran) {
	*(unsigned char *)transt = 'T';
    } else {
	*(unsigned char *)transt = 'N';
    }

    if (left && notran || ! left && ! notran) {
	i__1 = i1 - *ia;
	hscal_psorml2_my_tcomp_tcomm(
           0,
           myrow, mycol,
           SIDEI, TRANSI, M, N, i__1, IA, JA, &desca[1],
	     IC, JC, &descc[1],
           nprow, npcol,
           tcomp, tcomm
      );
    }

    ipw = desca[5] * desca[5] + 1;
    i__1 = i2;
    i__2 = i3;
    for (i__ = i1; i__2 < 0 ? i__ >= i__1 : i__ <= i__1; i__ += i__2) {
/* Computing MIN */
	i__3 = desca[5], i__4 = *k - i__ + *ia;
	ib = min(i__3,i__4);

/*        Form the triangular factor of the block reflector */
/*        H = H(i) H(i+1) . . . H(i+ib-1) */

	i__3 = nq - i__ + *ia;
	i__4 = *ja + i__ - *ia;
	hscal_pslarft_my_tcomp_tcomm(
           0,
           myrow, mycol,
           'F', 'R', i__3, ib, i__, i__4, &desca[1], 
           nprow, npcol,
           tcomp, tcomm
      );
	if (left) {

/*           H or H' is applied to C(ic+i-ia:ic+m-1,jc:jc+n-1) */

	    mi = *m - i__ + *ia;
	    icc = *ic + i__ - *ia;
	} else {

/*           H or H' is applied to C(ic:ic+m-1,jc+i-ia:jc+n-1) */

	    ni = *n - i__ + *ia;
	    jcc = *jc + i__ - *ia;
	}

/*        Apply H or H' */

	i__3 = *ja + i__ - *ia;
	hscal_pslarfb_my_tcomp_tcomm(
           0,
           myrow, mycol,
           SIDEI, *transt, 'F', 'R', mi, ni, ib, i__, i__3, &desca[1], 
           icc, jcc, &descc[1],
           nprow, npcol,
           tcomp, tcomm
      );
/* L10: */
    }

    if (left && ! notran || ! left && notran) {
	i__2 = i2 - *ia;
	hscal_psorml2_my_tcomp_tcomm(
           0,
           myrow, mycol,
           SIDEI, TRANSI, M, N, i__2, IA, JA, &desca[1],
	     IC, JC, &descc[1],
           nprow, npcol,
           tcomp, tcomm
      );
    }

    pb_topset__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topset__(&ictxt, "Broadcast", "Columnwise", colbtop);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psormlq_tcomp_tcomm(
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psormlq_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                M, N, K,
                IA, JA, DESCA, 
                IC, JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psormlq_info(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;
   int TRANSI = *TRANS;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psormlq_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                *M, *N, *K,
                *IA, *JA, DESCA, 
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psormlq_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;
   int TRANSI = *TRANS;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_psormlq_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                *M, *N, *K,
                *IA, *JA, DESCA, 
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
