
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pspotrs_recon_tcomp(
   int UPLOI,
   int N,
   int NRHS
)
{
    double tcomp   = 0;

    char   transc  = 'N';
    char   *trans  = &transc;

    int reconm, reconn, reconk;

    int num_of_computers = HMPI_Get_number_of_computers();
    int p                = sqrt(num_of_computers);
    int q                = num_of_computers/p;

    int    lda, ldb, ldc;

    float alpha = 1.;

    if (transc == 'N')
    {
       reconm  = N/p;
       reconn  = hscal_pilaenv_recon("S");
       reconk  = NRHS/q;
    }
    else
    {
       reconm  = hscal_pilaenv_recon("S");
       reconn  = N/p;
       reconk  = NRHS/q;
    }

    if (transc == 'N')
    {
       lda = reconm;
    }
    else
    {
       lda = reconk;
    }

    if (transc == 'N')
    {
       ldb = reconk;
    }
    else
    {
       ldb = reconn;
    }

    ldc = reconm;

    hscal_sgemm(
         trans,
         trans,
         &reconm,
         &reconn,
         &reconk,
         ((char*)&alpha),
         &lda,
         &ldb,
         ((char*)&alpha),
         &ldc,
         &tcomp
    );

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pspotrs_my_tcomp(
   int myrow,
   int mycol,
   int UPLO,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pspotrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLO, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pspotrs_my_tcomm(
   int myrow,
   int mycol,
   int UPLO,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pspotrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLO, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pspotrs_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int UPLO,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pspotrs_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          UPLO, N, NRHS,
          IA, JA, DESCA,
          IB, JB, DESCB,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pspotrs_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int UPLOI,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscalinteger c__2 = 2;
    hscalinteger c__7 = 7;
    hscalinteger c__3 = 3;
    hscalinteger c__11 = 11;
    hscalinteger c__1 = 1;
    hscalreal c_b24 = 1.;

    /* System generated locals */
    hscalinteger i__1;

    /* Local variables */
    hscalinteger idum1[1], idum2[1];
    hscalinteger iarow, ibrow;
    hscallogical upper;
    hscalinteger ictxt, icoffa, iroffa, iroffb;

    int  iinfo, *info = &iinfo;

    char UPLOC    = UPLOI, *uplo = &UPLOC;

    int     *n      = (int*)&N;
    int     *nrhs   = (int*)&NRHS;
    int     *ia     = (int*)&IA;
    int     *ja     = (int*)&JA;
    int     *desca  = (int*)DESCA;
    int     *ib     = (int*)&IB;
    int     *jb     = (int*)&JB;
    int     *descb  = (int*)DESCB;

    /* Parameter adjustments */ 
    --descb; 
    --desca; 

    if (init_llds_f == 1)
    {  
    HSCAL_PXPOTRS_INIT_LLDS( 
         (desca+1), 
         (descb+1), 
         myrow, nprow 
    );
    } 
 
    *info = 0;
    if (nprow == -1) {
	*info = -702;
    } else {
	hscal_chk1mat(n, &c__2, n, &c__2, ia, ja, &desca[1], &c__7, info, 
           myrow, mycol, nprow, npcol
      );
	hscal_chk1mat(n, &c__2, nrhs, &c__3, ib, jb, &descb[1], &c__11, info,
           myrow, mycol, nprow, npcol
      );
	upper = hscal_lsame(uplo, "U");
	if (*info == 0) {
	    iarow = indxg2p_(ia, &desca[5], &myrow, &desca[7], &nprow);
	    ibrow = indxg2p_(ib, &descb[5], &myrow, &descb[7], &nprow);
	    iroffa = (*ia - 1) % desca[5];
	    iroffb = (*ib - 1) % descb[5];
	    icoffa = (*ja - 1) % desca[6];
	    if (! upper && ! hscal_lsame(uplo, "L")) {
		*info = -1;
	    } else if (iroffa != 0) {
		*info = -5;
	    } else if (icoffa != 0) {
		*info = -6;
	    } else if (desca[5] != desca[6]) {
		*info = -706;
	    } else if (iroffb != 0 || ibrow != iarow) {
		*info = -9;
	    } else if (descb[5] != desca[6]) {
		*info = -1106;
	    }
	}
	if (upper) {
	    idum1[0] = 'U';
	} else {
	    idum1[0] = 'L';
	}
	idum2[0] = 1;
	hscal_pchk2mat(n, &c__2, n, &c__2, ia, ja, &desca[1], &c__7, n, &c__2, 
		nrhs, &c__3, ib, jb, &descb[1], &c__11, &c__1, idum1, idum2, 
		info, &myrow, &mycol, &nprow, &npcol);
    }
 
    if (*info != 0) { 
	i__1 = -(*info); 
        hscal_PB_Cabort( 
             "PSPOTRS", 
             i__1, 
             nprow, npcol, myrow, mycol 
        ); 
        return HSCAL_ERR_INPUT;
    } 
 
/*     Quick return if possible */ 
 
    if (*n == 0 || *nrhs == 0) { 
	HSCAL_SUCCESS;
    } 
 
    if (upper) {

/*        Solve sub( A ) * X = sub( B ) where sub( A ) = U'*U. */

/*        Solve U'*X = sub( B ), overwriting sub( B ) with X. */

      int ALPHA24 = c_b24;

	hscal_pstrsm_my_tcomp_tcomm(
           0,
           myrow, mycol,
           'L', 'U', 'T', 'N', N, NRHS, 
           ALPHA24, IA, JA, &desca[1], 
           IB, JB, &descb[1], nprow, npcol, tcomp, tcomm);

/*        Solve U*X = sub( B ), overwriting sub( B ) with X. */
	hscal_pstrsm_my_tcomp_tcomm(
           0,
           myrow, mycol,
           'L', 'U', 'N', 'N', N, NRHS, 
	     ALPHA24, 
	     IA, JA, &desca[1], IB, JB, &descb[1], nprow, npcol, tcomp, tcomm);
    } else {

/*        Solve sub( A ) *X = sub( B ) where sub( A ) = L*L'. */

/*        Solve L*X = sub( B ), overwriting sub( B ) with X. */
      int ALPHA24 = c_b24;

	hscal_pstrsm_my_tcomp_tcomm(
           0,
           myrow, mycol,
           'L', 'L', 'N', 'N', N, NRHS, 
	     ALPHA24, 
           IA, JA, &desca[1], IB, JB, &descb[1], nprow, npcol, tcomp, tcomm);

/*        Solve L'*X = sub( B ), overwriting sub( B ) with X. */

	hscal_pstrsm_my_tcomp_tcomm(
           0,
           myrow, mycol,
           'L', 'L', 'T', 'N', N, NRHS, 
	     ALPHA24, 
           IA, JA, &desca[1], IB, JB, &descb[1], nprow, npcol, tcomp, tcomm);
    }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pspotrs_tcomp_tcomm(
   int UPLO,
   int N,
   int NRHS,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int IB,
   int JB,
   int DESCB[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pspotrs_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                UPLO, N, NRHS,
                IA, JA, DESCA,
                IB, JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pspotrs_info(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pspotrs_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *UPLO, *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pspotrs_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pspotrs_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *UPLO, *N, *NRHS,
                *IA, *JA, DESCA,
                *IB, *JB, DESCB,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
