
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcdbtrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int N          =  params[1];
   int BWL        =  params[1+1];
   int BWU        =  params[1+1+1];
   int NRHS       =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+1+DLEN1_];
   int *DESCB     = &params[1+1+1+1+1+1+DLEN1_+1];

   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };

   hscalcomplex ALPHA, BETA, *a, *b, *c;
   int reconm, reconn, reconk;
   int lda, ldb, ldc;

   int max_bw__ = max(BWL, BWU);

   char TRANSAC    = 'C';
   char TRANSA[2]  = { TRANSAC, '\0' };

   char TRANSBC    = 'N';
   char TRANSB[2]  = { TRANSBC, '\0' };

   ALPHA.r = -1.f;
   ALPHA.i = -0.f;

   BETA.r = 1.f;
   BETA.i = 0.f;

   reconm  = max_bw__;
   reconn  = max_bw__;
   reconk  = max_bw__;

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d BWL %d BWU %d NRHS %d reconm %d reconn %d reconk %d "
           "JA %d IB %d",
           N, BWL, BWU, NRHS, reconm, reconn, reconk,
           JA, IB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cgemm routine
   hscal_cgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcdbtrs_modelp_noauto(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* N */
                  +        
                  1        /* BWL */
                  +        
                  1        /* BWU */
                  +        
                  1        /* NRHS */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANS;
   (*model_params)[1]                                                 = *N;
   (*model_params)[1+1]                                               = *BWL;
   (*model_params)[1+1+1]                                             = *BWU;
   (*model_params)[1+1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+DLEN1_]                       = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+DLEN1_+1]                     = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                    = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+i]           = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcdbtrs_modelp(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* N */
                  +        
                  1        /* BWL */
                  +        
                  1        /* BWU */
                  +        
                  1        /* NRHS */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANS;
   (*model_params)[1]                                                 = *N;
   (*model_params)[1+1]                                               = *BWL;
   (*model_params)[1+1+1]                                             = *BWU;
   (*model_params)[1+1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                                = *IB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                    = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+i]           = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcdbtrs_recon(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pcdbtrs_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcdbtrs_timeof(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcdbtrs_modelp_noauto(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcdbtrs_group_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcdbtrs_modelp_noauto(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcdbtrs_group_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcdbtrs_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcdbtrs_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzdbtrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int N          =  params[1];
   int BWL        =  params[1+1];
   int BWU        =  params[1+1+1];
   int NRHS       =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+1+DLEN1_];
   int *DESCB     = &params[1+1+1+1+1+1+DLEN1_+1];

   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };

   hscaldoublecomplex ALPHA, BETA, *a, *b, *c;
   int reconm, reconn, reconk;
   int lda, ldb, ldc;

   int max_bw__ = max(BWL, BWU);

   char TRANSAC    = 'C';
   char TRANSA[2]  = { TRANSAC, '\0' };

   char TRANSBC    = 'N';
   char TRANSB[2]  = { TRANSBC, '\0' };

   ALPHA.r = -1.;
   ALPHA.i = -0.;

   BETA.r = 1.;
   BETA.i = 0.;

   reconm  = max_bw__;
   reconn  = max_bw__;
   reconk  = max_bw__;

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d BWL %d BWU %d NRHS %d reconm %d reconn %d reconk %d "
           "JA %d IB %d",
           N, BWL, BWU, NRHS, reconm, reconn, reconk,
           JA, IB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zgemm routine
   hscal_zgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzdbtrs_modelp_noauto(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* N */
                  +        
                  1        /* BWL */
                  +        
                  1        /* BWU */
                  +        
                  1        /* NRHS */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANS;
   (*model_params)[1]                                                 = *N;
   (*model_params)[1+1]                                               = *BWL;
   (*model_params)[1+1+1]                                             = *BWU;
   (*model_params)[1+1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+DLEN1_]                       = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+DLEN1_+1]                     = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                    = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+i]           = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzdbtrs_modelp(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* N */
                  +        
                  1        /* BWL */
                  +        
                  1        /* BWU */
                  +        
                  1        /* NRHS */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANS;
   (*model_params)[1]                                                 = *N;
   (*model_params)[1+1]                                               = *BWL;
   (*model_params)[1+1+1]                                             = *BWU;
   (*model_params)[1+1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                                = *IB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                    = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+i]           = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdbtrs_recon(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pzdbtrs_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzdbtrs_timeof(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzdbtrs_modelp_noauto(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzdbtrs_group_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzdbtrs_modelp_noauto(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdbtrs_group_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdbtrs_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzdbtrs_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pddbtrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int N          =  params[1];
   int BWL        =  params[1+1];
   int BWU        =  params[1+1+1];
   int NRHS       =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+1+DLEN1_];
   int *DESCB     = &params[1+1+1+1+1+1+DLEN1_+1];

   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };

   hscaldoublereal ALPHA, BETA, *a, *b, *c;
   int reconm, reconn, reconk;
   int lda, ldb, ldc;

   int max_bw__ = max(BWL, BWU);

   char TRANSAC    = 'C';
   char TRANSA[2]  = { TRANSAC, '\0' };

   char TRANSBC    = 'N';
   char TRANSB[2]  = { TRANSBC, '\0' };

   ALPHA = -1.;
   BETA = 1.;

   reconm  = max_bw__;
   reconn  = max_bw__;
   reconk  = max_bw__;

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d BWL %d BWU %d NRHS %d reconm %d reconn %d reconk %d "
           "JA %d IB %d",
           N, BWL, BWU, NRHS, reconm, reconn, reconk,
           JA, IB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dgemm routine
   hscal_dgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pddbtrs_modelp_noauto(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* N */
                  +        
                  1        /* BWL */
                  +        
                  1        /* BWU */
                  +        
                  1        /* NRHS */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANS;
   (*model_params)[1]                                                 = *N;
   (*model_params)[1+1]                                               = *BWL;
   (*model_params)[1+1+1]                                             = *BWU;
   (*model_params)[1+1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+DLEN1_]                       = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+DLEN1_+1]                     = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                    = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+i]           = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pddbtrs_modelp(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* N */
                  +        
                  1        /* BWL */
                  +        
                  1        /* BWU */
                  +        
                  1        /* NRHS */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANS;
   (*model_params)[1]                                                 = *N;
   (*model_params)[1+1]                                               = *BWL;
   (*model_params)[1+1+1]                                             = *BWU;
   (*model_params)[1+1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                                = *IB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                    = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+i]           = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_recon(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pddbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pddbtrs_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pddbtrs_timeof(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pddbtrs_modelp_noauto(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_group_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pddbtrs_modelp_noauto(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_group_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pddbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pddbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pddbtrs_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pddbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psdbtrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int N          =  params[1];
   int BWL        =  params[1+1];
   int BWU        =  params[1+1+1];
   int NRHS       =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+1+DLEN1_];
   int *DESCB     = &params[1+1+1+1+1+1+DLEN1_+1];

   char TRANSC   = TRANSI;
   char TRANS[2] = { TRANSC, '\0' };

   hscalreal ALPHA, BETA, *a, *b, *c;
   int reconm, reconn, reconk;
   int lda, ldb, ldc;

   int max_bw__ = max(BWL, BWU);

   char TRANSAC    = 'C';
   char TRANSA[2]  = { TRANSAC, '\0' };

   char TRANSBC    = 'N';
   char TRANSB[2]  = { TRANSBC, '\0' };

   ALPHA = -1.f;
   BETA = 1.f;

   reconm  = max_bw__;
   reconn  = max_bw__;
   reconk  = max_bw__;

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "N %d BWL %d BWU %d NRHS %d reconm %d reconn %d reconk %d "
           "JA %d IB %d",
           N, BWL, BWU, NRHS, reconm, reconn, reconk,
           JA, IB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sgemm routine
   hscal_sgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psdbtrs_modelp_noauto(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* N */
                  +        
                  1        /* BWL */
                  +        
                  1        /* BWU */
                  +        
                  1        /* NRHS */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANS;
   (*model_params)[1]                                                 = *N;
   (*model_params)[1+1]                                               = *BWL;
   (*model_params)[1+1+1]                                             = *BWU;
   (*model_params)[1+1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+DLEN1_]                       = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+DLEN1_+1]                     = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                    = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+i]           = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psdbtrs_modelp(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* N */
                  +        
                  1        /* BWL */
                  +        
                  1        /* BWU */
                  +        
                  1        /* NRHS */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                                 = *TRANS;
   (*model_params)[1]                                                 = *N;
   (*model_params)[1+1]                                               = *BWL;
   (*model_params)[1+1+1]                                             = *BWU;
   (*model_params)[1+1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                                = *IB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                    = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+i]           = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psdbtrs_recon(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_psdbtrs_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psdbtrs_timeof(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psdbtrs_modelp_noauto(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psdbtrs_group_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psdbtrs_modelp_noauto(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psdbtrs_group_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psdbtrs_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psdbtrs_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psdbtrs_modelp(
                TRANS,
                N,
                BWL,
                BWU,
                NRHS,
                JA,
                DESCA,
                IB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
