
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcgels_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   int NRHS       =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+1+DLEN1_];
   int JB         =  params[1+1+1+1+1+1+DLEN1_+1];
   int *DESCB     = &params[1+1+1+1+1+1+DLEN1_+1+1];

   hscalcomplex ALPHA, BETA, *a, *b, *c;

   int reconm, reconn, reconk;
   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int lda, ldb, ldc;

   char TRANSAC    = 'N';
   char TRANSA[2]  = { TRANSAC, '\0' };

   char TRANSBC    = 'C';
   char TRANSB[2]  = { TRANSBC, '\0' };

   ALPHA.r        = -1.f;
   ALPHA.i        = -0.f;

   BETA.r         = 1.f;
   BETA.i         = 0.f;

   reconm  = M/p;
   reconn  = N/q;
   reconk  = hscal_pilaenv_recon("C");

   lda = reconm;
   ldb = reconn;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "M %d N %d NRHS %d reconm %d reconn %d reconk %d ALPHA (%f, %f) BETA (%f, %f) "
           "IA %d JA %d IB %d JB %d",
           M, N, NRHS, reconm, reconn, reconk, ALPHA.r, ALPHA.i, BETA.r, BETA.i,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cgemm routine
   hscal_cgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcgels_modelp_noauto(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANS;
   (*model_params)[1]                                               = *M;
   (*model_params)[1+1]                                             = *N;
   (*model_params)[1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1]                                         = *IA;
   (*model_params)[1+1+1+1+1]                                       = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                              = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1]                            = *JB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+DLEN1_]                   = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+DLEN1_+1]                 = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                               = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+i]                    = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcgels_modelp(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANS;
   (*model_params)[1]                                               = *M;
   (*model_params)[1+1]                                             = *N;
   (*model_params)[1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1]                                         = *IA;
   (*model_params)[1+1+1+1+1]                                       = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                              = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1]                            = *JB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                               = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+i]                    = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgels_recon(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pcgels_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcgels_timeof(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcgels_modelp_noauto(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcgels_group_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgels_modelp_noauto(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgels_group_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgels_group_heuristic_pauto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcgels_group_heuristic_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzgels_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   int NRHS       =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+1+DLEN1_];
   int JB         =  params[1+1+1+1+1+1+DLEN1_+1];
   int *DESCB     = &params[1+1+1+1+1+1+DLEN1_+1+1];

   hscaldoublecomplex ALPHA, BETA, *a, *b, *c;

   int reconm, reconn, reconk;
   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int lda, ldb, ldc;

   char TRANSAC    = 'N';
   char TRANSA[2]  = { TRANSAC, '\0' };

   char TRANSBC    = 'C';
   char TRANSB[2]  = { TRANSBC, '\0' };

   ALPHA.r        = -1.;
   ALPHA.i        = -0.;

   BETA.r         = 1.;
   BETA.i         = 0.;

   reconm  = M/p;
   reconn  = N/q;
   reconk  = hscal_pilaenv_recon("Z");

   lda = reconm;
   ldb = reconn;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "M %d N %d NRHS %d reconm %d reconn %d reconk %d ALPHA (%f, %f) BETA (%f, %f) "
           "IA %d JA %d IB %d JB %d",
           M, N, NRHS, reconm, reconn, reconk, ALPHA.r, ALPHA.i, BETA.r, BETA.i,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zgemm routine
   hscal_zgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzgels_modelp_noauto(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANS;
   (*model_params)[1]                                               = *M;
   (*model_params)[1+1]                                             = *N;
   (*model_params)[1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1]                                         = *IA;
   (*model_params)[1+1+1+1+1]                                       = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                              = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1]                            = *JB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+DLEN1_]                   = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+DLEN1_+1]                 = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                               = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+i]                    = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzgels_modelp(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANS;
   (*model_params)[1]                                               = *M;
   (*model_params)[1+1]                                             = *N;
   (*model_params)[1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1]                                         = *IA;
   (*model_params)[1+1+1+1+1]                                       = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                              = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1]                            = *JB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                               = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+i]                    = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgels_recon(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pzgels_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzgels_timeof(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzgels_modelp_noauto(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzgels_group_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgels_modelp_noauto(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgels_group_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgels_group_heuristic_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgels_group_heuristic_pauto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdgels_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   int NRHS       =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+1+DLEN1_];
   int JB         =  params[1+1+1+1+1+1+DLEN1_+1];
   int *DESCB     = &params[1+1+1+1+1+1+DLEN1_+1+1];

   hscaldoublereal ALPHA =  -1.;
   hscaldoublereal BETA  =  1.;

   char TRANSAC    = 'N';
   char TRANSA[2]  = { TRANSAC, '\0' };

   char TRANSBC    = 'T';
   char TRANSB[2]  = { TRANSBC, '\0' };

   double *a, *b, *c;
   int reconm, reconn, reconk;
   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int lda, ldb, ldc;

   reconm  = M/p;
   reconn  = N/q;
   reconk  = hscal_pilaenv_recon("D");

   lda = reconm;
   ldb = reconn;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "M %d N %d NRHS %d reconm %d reconn %d reconk %d ALPHA (%f) BETA (%f) "
           "IA %d JA %d IB %d JB %d",
           M, N, NRHS, reconm, reconn, reconk, ALPHA, BETA,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (double*)malloc(
                 sizeof(double)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (double*)malloc(
                 sizeof(double)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (double*)malloc(
                 sizeof(double)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dgemm routine
   hscal_dgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgels_modelp_noauto(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANS;
   (*model_params)[1]                                               = *M;
   (*model_params)[1+1]                                             = *N;
   (*model_params)[1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1]                                         = *IA;
   (*model_params)[1+1+1+1+1]                                       = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                              = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1]                            = *JB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+DLEN1_]                   = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+DLEN1_+1]                 = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                               = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+i]                    = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgels_modelp(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANS;
   (*model_params)[1]                                               = *M;
   (*model_params)[1+1]                                             = *N;
   (*model_params)[1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1]                                         = *IA;
   (*model_params)[1+1+1+1+1]                                       = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                              = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1]                            = *JB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                               = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+i]                    = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgels_recon(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pdgels_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdgels_timeof(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdgels_modelp_noauto(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdgels_group_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgels_modelp_noauto(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgels_group_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgels_group_heuristic_pauto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgels_group_heuristic_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psgels_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int TRANSI     =  params[0];
   int M          =  params[1];
   int N          =  params[1+1];
   int NRHS       =  params[1+1+1];
   int IA         =  params[1+1+1+1];
   int JA         =  params[1+1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+1+DLEN1_];
   int JB         =  params[1+1+1+1+1+1+DLEN1_+1];
   int *DESCB     = &params[1+1+1+1+1+1+DLEN1_+1+1];

   hscalreal ALPHA  =  1.f;
   hscalreal BETA   =  -1.f;

   char TRANSAC    = 'N';
   char TRANSA[2]  = { TRANSAC, '\0' };

   char TRANSBC    = 'T';
   char TRANSB[2]  = { TRANSBC, '\0' };

   float *a, *b, *c;
   int reconm, reconn, reconk;
   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int lda, ldb, ldc;

   reconm  = M/p;
   reconn  = N/q;
   reconk  = hscal_pilaenv_recon("S");

   lda = reconm;
   ldb = reconn;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "M %d N %d NRHS %d reconm %d reconn %d reconk %d ALPHA (%f) BETA (%f) "
           "IA %d JA %d IB %d JB %d",
           M, N, NRHS, reconm, reconn, reconk, ALPHA, BETA,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (float*)malloc(
                 sizeof(float)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (float*)malloc(
                 sizeof(float)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (float*)malloc(
                 sizeof(float)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sgemm routine
   hscal_sgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psgels_modelp_noauto(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANS;
   (*model_params)[1]                                               = *M;
   (*model_params)[1+1]                                             = *N;
   (*model_params)[1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1]                                         = *IA;
   (*model_params)[1+1+1+1+1]                                       = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                              = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1]                            = *JB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+DLEN1_]                   = *nprow;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+DLEN1_+1]                 = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                               = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+i]                    = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psgels_modelp(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* TRANS */
                  +        
                  1        /* M */
                  +        
                  1        /* N */
                  +
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANS;
   (*model_params)[1]                                               = *M;
   (*model_params)[1+1]                                             = *N;
   (*model_params)[1+1+1]                                           = *NRHS;
   (*model_params)[1+1+1+1]                                         = *IA;
   (*model_params)[1+1+1+1+1]                                       = *JA;
   (*model_params)[1+1+1+1+1+1+DLEN1_]                              = *IB;
   (*model_params)[1+1+1+1+1+1+DLEN1_+1]                            = *JB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+1+i]                               = DESCA[i];
       (*model_params)[1+1+1+1+1+1+DLEN1_+1+1+i]                    = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgels_recon(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_psgels_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psgels_timeof(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psgels_modelp_noauto(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psgels_group_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgels_modelp_noauto(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgels_group_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgels_group_heuristic_pauto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psgels_group_heuristic_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psgels_modelp(
                TRANS,
                M,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
