
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pcposv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int UPLOI      =  params[0];
   int N          =  params[1];
   int NRHS       =  params[1+1];
   int IA         =  params[1+1+1];
   int JA         =  params[1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+DLEN1_];
   int JB         =  params[1+1+1+1+1+DLEN1_+1];
   int *DESCB     = &params[1+1+1+1+1+DLEN1_+1+1];

   hscalcomplex ALPHA, *a, *b, *c;

   int reconm, reconn, reconk;
   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int lda, ldb, ldc;

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };

   char TRANSC    = 'N';
   char TRANS[2]  = { TRANSC, '\0' };

   ALPHA.r        = 1.f;
   ALPHA.i        = 1.f;

   if (TRANSC == 'N')
   {
      reconm  = N/p;
      reconn  = hscal_pilaenv_recon("C");
      reconk  = N/q;
   }
   else
   {
      reconm  = hscal_pilaenv_recon("C");
      reconn  = N/p;
      reconk  = N/q;
   }

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "UPLOC %c N %d NRHS %d reconm %d reconn %d reconk %d ALPHA (%f, %f) "
           "IA %d JA %d IB %d JB %d",
           UPLOC, N, NRHS, reconm, reconn, reconk, ALPHA.r, ALPHA.i,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS cgemm routine
   hscal_cgemm_benchmark(
        TRANS,
        TRANS,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &ALPHA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcposv_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* N */
                  +        
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *N;
   (*model_params)[1+1]                                             = *NRHS;
   (*model_params)[1+1+1]                                           = *IA;
   (*model_params)[1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+DLEN1_+1]                              = *JB;
   (*model_params)[1+1+1+1+1+DLEN1_+1+1+DLEN1_]                     = *nprow;
   (*model_params)[1+1+1+1+1+DLEN1_+1+1+DLEN1_+1]                   = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                                 = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN1_+1+1+i]                      = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pcposv_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* N */
                  +        
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *N;
   (*model_params)[1+1]                                             = *NRHS;
   (*model_params)[1+1+1]                                           = *IA;
   (*model_params)[1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+DLEN1_+1]                              = *JB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                                 = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN1_+1+1+i]                      = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcposv_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pcposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pcposv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pcposv_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcposv_modelp_noauto(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pcposv_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcposv_modelp_noauto(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcposv_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcposv_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pcposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pcposv_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pcposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzposv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int UPLOI      =  params[0];
   int N          =  params[1];
   int NRHS       =  params[1+1];
   int IA         =  params[1+1+1];
   int JA         =  params[1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+DLEN1_];
   int JB         =  params[1+1+1+1+1+DLEN1_+1];
   int *DESCB     = &params[1+1+1+1+1+DLEN1_+1+1];

   hscaldoublecomplex ALPHA, *a, *b, *c;

   int reconm, reconn, reconk;
   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int lda, ldb, ldc;

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };

   char TRANSC    = 'N';
   char TRANS[2]  = { TRANSC, '\0' };

   ALPHA.r        = 1.;
   ALPHA.i        = 1.;

   if (TRANSC == 'N')
   {
      reconm  = N/p;
      reconn  = hscal_pilaenv_recon("Z");
      reconk  = N/q;
   }
   else
   {
      reconm  = hscal_pilaenv_recon("Z");
      reconn  = N/p;
      reconk  = N/q;
   }

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "UPLOC %c N %d NRHS %d reconm %d reconn %d reconk %d ALPHA (%f, %f) "
           "IA %d JA %d IB %d JB %d",
           UPLOC, N, NRHS, reconm, reconn, reconk, ALPHA.r, ALPHA.i,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i].r = 2.;
      a[i].i = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i].r = 2.;
      b[i].i = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i].r = 1.;
      c[i].i = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS zgemm routine
   hscal_zgemm_benchmark(
        TRANS,
        TRANS,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &ALPHA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzposv_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* N */
                  +        
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *N;
   (*model_params)[1+1]                                             = *NRHS;
   (*model_params)[1+1+1]                                           = *IA;
   (*model_params)[1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+DLEN1_+1]                              = *JB;
   (*model_params)[1+1+1+1+1+DLEN1_+1+1+DLEN1_]                     = *nprow;
   (*model_params)[1+1+1+1+1+DLEN1_+1+1+DLEN1_+1]                   = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                                 = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN1_+1+1+i]                      = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pzposv_modelp(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* N */
                  +        
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *N;
   (*model_params)[1+1]                                             = *NRHS;
   (*model_params)[1+1+1]                                           = *IA;
   (*model_params)[1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+DLEN1_+1]                              = *JB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                                 = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN1_+1+1+i]                      = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzposv_recon(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pzposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pzposv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pzposv_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzposv_modelp_noauto(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count
      );

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pzposv_group_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzposv_modelp_noauto(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzposv_group_auto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzposv_group_heuristic_auto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pzposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzposv_group_heuristic_pauto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pzposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdposv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int UPLOI      =  params[0];
   int N          =  params[1];
   int NRHS       =  params[1+1];
   int IA         =  params[1+1+1];
   int JA         =  params[1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+DLEN1_];
   int JB         =  params[1+1+1+1+1+DLEN1_+1];
   int *DESCB     = &params[1+1+1+1+1+DLEN1_+1+1];

   hscaldoublereal ALPHA =  1.;

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };

   char TRANSC   = 'N';
   char TRANS[2] = { TRANSC, '\0' };

   double *a, *b, *c;
   int reconm, reconn, reconk;
   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int lda, ldb, ldc;

   if (TRANSC == 'N')
   {
      reconm  = N/p;
      reconn  = hscal_pilaenv_recon("D");
      reconk  = N/q;
   }
   else
   {
      reconm  = hscal_pilaenv_recon("D");
      reconn  = N/p;
      reconk  = N/q;
   }

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "UPLOC %c N %d NRHS %d reconm %d reconn %d reconk %d ALPHA %f "
           "IA %d JA %d IB %d JB %d",
           UPLOC, N, NRHS, reconm, reconn, reconk, ALPHA,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (double*)malloc(
                 sizeof(double)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (double*)malloc(
                 sizeof(double)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (double*)malloc(
                 sizeof(double)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS dgemm routine
   hscal_dgemm_benchmark(
        TRANS,
        TRANS,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &ALPHA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdposv_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* N */
                  +        
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *N;
   (*model_params)[1+1]                                             = *NRHS;
   (*model_params)[1+1+1]                                           = *IA;
   (*model_params)[1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+DLEN1_+1]                              = *JB;
   (*model_params)[1+1+1+1+1+DLEN1_+1+1+DLEN1_]                     = *nprow;
   (*model_params)[1+1+1+1+1+DLEN1_+1+1+DLEN1_+1]                   = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                                 = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN1_+1+1+i]                      = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdposv_modelp(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* N */
                  +        
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *N;
   (*model_params)[1+1]                                             = *NRHS;
   (*model_params)[1+1+1]                                           = *IA;
   (*model_params)[1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+DLEN1_+1]                              = *JB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                                 = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN1_+1+1+i]                      = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_recon(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_pdposv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_pdposv_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdposv_modelp_noauto(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count
      );

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_group_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdposv_modelp_noauto(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_group_auto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_group_heuristic_pauto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_pdposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdposv_group_heuristic_auto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_pdposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psposv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;
   double btime;

   int UPLOI      =  params[0];
   int N          =  params[1];
   int NRHS       =  params[1+1];
   int IA         =  params[1+1+1];
   int JA         =  params[1+1+1+1];
   int *DESCA     = &params[1+1+1+1+1];
   int IB         =  params[1+1+1+1+1+DLEN1_];
   int JB         =  params[1+1+1+1+1+DLEN1_+1];
   int *DESCB     = &params[1+1+1+1+1+DLEN1_+1+1];

   hscalreal ALPHA  =  1.f;

   char UPLOC    = UPLOI;
   char UPLO[2]  = { UPLOC, '\0' };

   char TRANSC   = 'N';
   char TRANS[2] = { TRANSC, '\0' };

   float *a, *b, *c;
   int reconm, reconn, reconk;
   int num_of_computers = HMPI_Get_number_of_computers();
   int p                = sqrt(num_of_computers);
   int q                = num_of_computers/p;
   int lda, ldb, ldc;

   if (TRANSC == 'N')
   {
      reconm  = N/p;
      reconn  = hscal_pilaenv_recon("S");
      reconk  = N/q;
   }
   else
   {
      reconm  = hscal_pilaenv_recon("S");
      reconn  = N/p;
      reconk  = N/q;
   }

   lda = reconm;
   ldb = reconk;
   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "UPLOC %c N %d NRHS %d reconm %d reconn %d reconk %d ALPHA %f "
           "IA %d JA %d IB %d JB %d",
           UPLOC, N, NRHS, reconm, reconn, reconk, ALPHA,
           IA, JA, IB, JB
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (float*)malloc(
                 sizeof(float)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   b = (float*)malloc(
                 sizeof(float)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   c = (float*)malloc(
                 sizeof(float)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return HSCAL_ERR_RECON_ESTIMATION;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   btime = MPI_Wtime();

   //
   // Call the BLAS sgemm routine
   hscal_sgemm_benchmark(
        TRANS,
        TRANS,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &ALPHA,
        c,
        &ldc
   );

   btime = MPI_Wtime() - btime;

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      {
         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            btime
         );
      }
   }

   return btime;
}

/*------------------------------------------------------------------*/

int hscal_pack_psposv_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* N */
                  +        
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
                  +
                  1        /* nprow */
                  +
                  1        /* npcol */
                  +
                  *nprow * *npcol /* total computations (INTERNAL) */
                  +
                  *nprow * *npcol * *nprow * *npcol /* total communications (INTERNAL) */
                  +
                  *nprow * *npcol /* total communications (INTERNAL) */
   ;

   *model_params = (int*)calloc(
                         (*model_count),
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *N;
   (*model_params)[1+1]                                             = *NRHS;
   (*model_params)[1+1+1]                                           = *IA;
   (*model_params)[1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+DLEN1_+1]                              = *JB;
   (*model_params)[1+1+1+1+1+DLEN1_+1+1+DLEN1_]                     = *nprow;
   (*model_params)[1+1+1+1+1+DLEN1_+1+1+DLEN1_+1]                   = *npcol;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                                 = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN1_+1+1+i]                      = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pack_psposv_modelp(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
)
{
   int i;

   *model_count = 1        /* UPLO */
                  +        
                  1        /* N */
                  +        
                  1        /* NRHS */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN1_   /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN1_   /* DESCB */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *UPLO;
   (*model_params)[1]                                               = *N;
   (*model_params)[1+1]                                             = *NRHS;
   (*model_params)[1+1+1]                                           = *IA;
   (*model_params)[1+1+1+1]                                         = *JA;
   (*model_params)[1+1+1+1+1+DLEN1_]                                = *IB;
   (*model_params)[1+1+1+1+1+DLEN1_+1]                              = *JB;

   for (i = 0; i < DLEN1_; i++)
   {
       (*model_params)[1+1+1+1+1+i]                                 = DESCA[i];
       (*model_params)[1+1+1+1+1+DLEN1_+1+1+i]                      = DESCB[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psposv_recon(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_psposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon_t(
               &hscal_psposv_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

double hscal_psposv_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
)
{
   int rc;
   double derr = -1.;
   double timeof;

   {
      int *model_params, model_count;

      rc = hscal_pack_psposv_modelp_noauto(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count
      );

      if (rc != HSCAL_SUCCESS)
      {
         return derr;
      }

      timeof = HMPI_Timeof(
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return derr;
      }

      free(model_params);
   }

   return timeof;
}

/*------------------------------------------------------------------*/

int hscal_psposv_group_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psposv_modelp_noauto(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                nprow,
                npcol,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psposv_group_auto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_auto_create(
               gid,
               model,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_auto_create(
               gid,
               model,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psposv_group_heuristic_pauto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   {
      int *model_params, model_count;

      rc = hscal_pack_psposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_pauto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_psposv_group_heuristic_auto_create(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
)
{
   int rc;

   if (HMPI_Is_host())
   {
      int *model_params, model_count;

      rc = hscal_pack_psposv_modelp(
                UPLO,
                N,
                NRHS,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               hfunc,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {

         free(model_params);

         return rc;
      }

      free(model_params);
   }

   if (HMPI_Is_free())
   {
      rc = HMPI_Group_heuristic_auto_create(
               gid,
               model,
               NULL,
               NULL,
               0
      );

      if (rc != HMPI_SUCCESS)
      {
         return rc;
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
