
#include <math.h>
#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzgeqr2_recon_tcomp(
   int M,
   int N
)
{
    double tcomp = 0.;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pzgeqr2_my_tcomp(
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzgeqr2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzgeqr2_my_tcomm(
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzgeqr2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgeqr2_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pzgeqr2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          M, N,
          IA, JA, DESCA,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pzgeqr2_my_tcomp_tcomm(
   int init_llds_f,
   int myrow,
   int mycol,
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscaldoublecomplex c_b1 = {1.,0.};
    hscalinteger c__1 = 1;
    hscalinteger c__2 = 2;
    hscalinteger c__6 = 6;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4, i__5;
    hscaldoublereal d__1;
    hscaldoublecomplex z__1, z__2;

    /* Local variables */
    hscalinteger i__, j, k;
    hscalinteger ii, jj, mp, nq;
    hscaldoublecomplex ajj;
    hscaldoublecomplex alpha;
    hscalinteger iacol, iarow, lwmin, ictxt;
    char colbtop[1], rowbtop[1];

/*
*  .. Local Arrays ..
*/
    HSCALBLACSCONTEXT   * blacsctxt;

    int     *m      = &M;
    int     *n      = &N;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     iinfo, *info = &iinfo;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --desca;

    if (init_llds_f == 1)
    {
       HSCAL_PXGEQR2_INIT_LLDS(
            (desca+1),
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "PZGEQR2: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d",
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

/*     Test the input parameters */

    *info = 0;
    if (nprow == -1) {
	*info = -602;
    } else {
	hscal_chk1mat(
           m, &c__1, n, &c__2, ia, ja, &desca[1], &c__6, info,
           myrow, mycol, nprow, npcol
      );
	if (*info == 0) {
	    iarow = indxg2p_(ia, &desca[5], &myrow, &desca[7], &nprow);
	    iacol = indxg2p_(ja, &desca[6], &mycol, &desca[8], &npcol);
	    i__1 = *m + (*ia - 1) % desca[5];
	    mp = numroc_(&i__1, &desca[5], &myrow, &iarow, &nprow);
	    i__1 = *n + (*ja - 1) % desca[6];
	    nq = numroc_(&i__1, &desca[6], &mycol, &iacol, &npcol);
	    lwmin = mp + max(1,nq);

	    d__1 = (hscaldoublereal) lwmin;
          z__1.r = d__1, z__1.i = 0.;
	}
    }

    if (*info != 0) {
	i__1 = -(*info);
       hscal_PB_Cabort(
            "PZGEQR2",
            i__1,
            nprow, npcol, myrow, mycol
       );
       return HSCAL_SUCCESS;
    }

/*     Quick return if possible */

    if (*m == 0 || *n == 0) {
       return HSCAL_SUCCESS;
    }

    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);
    pb_topset__(&ictxt, "Broadcast", "Rowwise", "I-ring");
    pb_topset__(&ictxt, "Broadcast", "Columnwise", " ");

    if (desca[3] == 1) {
      Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);
	infog2l_(ia, ja, &desca[1], &nprow, &npcol, &myrow, &mycol, &ii, &jj, 
		&iarow, &iacol);
	if (myrow == iarow) {
	    i__1 = *ja + *n - 1;
	    nq = numroc_(&i__1, &desca[6], &mycol, &desca[8], &npcol);
	    i__ = ii + (jj - 1) * desca[9];
	    if (mycol == iacol) {
/*
* Assume some values for ajj
*/
		ajj.r = 0.5, ajj.i = 0.5;
		zlarfg_tcomp_macro(
                  &c__1, &ajj, &c__1, 
                  nprow, npcol, myrow, mycol, tcomp_l
            );
		if (*n > 1) {
		    hscal_Czgebs2d(
                     blacsctxt, "Rowwise", " ", c__1, c__1, c__1, 
                     nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                );
		    i__1 = nq - jj;
/*
* Assume some values for alpha
*/
                alpha.r = 0.5, alpha.i = -0.5;
		    zscal_tcomp_macro(
                     &i__1, (char*)&alpha, &desca[9],
                     nprow, npcol, myrow, mycol, tcomp_l
                );
		}
		hscal_Czgebs2d(
                 blacsctxt, "Columnwise", " ", c__1, c__1, 
                 c__1, 
                 nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
            );
		i__1 = i__;
	    } else {
		if (*n > 1) {
		    hscal_Czgebr2d(
                     blacsctxt, "Rowwise", " ", c__1, c__1,
			   c__1, iarow, iacol,
                     nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
                );
		    i__1 = nq - jj + 1;
/*
* Assume some values for alpha
*/
                alpha.r = 0.5, alpha.i = -0.5;
		    zscal_tcomp_macro(
                     &i__1, (char*)&alpha, &desca[9],
                     nprow, npcol, myrow, mycol, tcomp_l
                );
		}
	    }
	} else if (mycol == iacol) {
	    hscal_Czgebr2d(
               blacsctxt, "Columnwise", " ", c__1, c__1, c__1,
		   iarow, iacol,
               nprow, npcol, myrow, mycol, tcomp_l, tcomm_l
          );
	}

      Cblacs_gridexit_nocomm(blacsctxt);

    } else {

	k = min(*m,*n);
	i__1 = *ja + k - 1;
	for (j = *ja; j <= i__1; ++j) {
	    i__ = *ia + j - *ja;

/*           Generate elementary reflector H(j) to annihilate */
/*           A(i+1:ia+m-1,j) */

	    i__2 = *m - j + *ja;
/* Computing MIN */
	    i__4 = i__ + 1, i__5 = *ia + *m - 1;
	    i__3 = min(i__4,i__5);
          {
	    hscal_pzlarfg_my_tcomp_tcomm(
               0,
               myrow, mycol,
               i__2, i__, j, i__3, j, &desca[1], c__1,
               nprow, npcol,
               tcomp, tcomm
          );
          }
	    if (j < *ja + *n - 1) {

/*              Apply H(j)' to A(i:ia+m-1,j+1:ja+n-1) from the left */

            {
		hscal_pzelset(
                 &i__, &j, &desca[1], &c_b1,
                 nprow, npcol,
                 myrow, mycol,
                 tcomp, tcomm
            );
            }

		i__2 = *m - j + *ja;
		i__3 = *n - j + *ja - 1;
		i__4 = j + 1;
		hscal_pzlarfc_my_tcomp_tcomm(
                 0,
                 myrow, mycol,
                 'L', i__2, i__3, i__, j, &desca[1], c__1, i__, 
                 i__4, &desca[1], 
                 nprow, npcol,
                 tcomp, tcomm
            );
	    }
          {
	    hscal_pzelset(
               &i__, &j, &desca[1], &ajj,
               nprow, npcol,
               myrow, mycol,
               tcomp, tcomm
          );
          }

/* L10: */
	}

    }

    pb_topset__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topset__(&ictxt, "Broadcast", "Columnwise", colbtop);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgeqr2_tcomp_tcomm(
   int M,
   int N,
   int IA,
   int JA,
   int DESCA[DLEN1_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgeqr2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                M, N,
                IA, JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgeqr2_info(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgeqr2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzgeqr2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzgeqr2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *M, *N,
                *IA, *JA, DESCA,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

