
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzlarf_recon_tcomp(
   int sidei,
   int m,
   int n
)
{
    double tcomp = 0;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pzlarf_my_tcomp(
   int myrow,
   int mycol,
   int SIDEI,
   int M,
   int N,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int INCV,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzlarf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI,
          M, N,
          IV, JV, DESCV, INCV,
          IC, JC, DESCC, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzlarf_my_tcomm(
   int myrow,
   int mycol,
   int SIDEI,
   int M,
   int N,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int INCV,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzlarf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI,
          M, N,
          IV, JV, DESCV, INCV,
          IC, JC, DESCC, 
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzlarf_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int SIDEI,
   int M,
   int N,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int INCV,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pzlarf_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI,
          M, N,
          IV, JV, DESCV, INCV,
          IC, JC, DESCC, 
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pzlarf_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int SIDEI,
   int M,
   int N,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int INCV,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscaldoublecomplex c_b1 = {1.f,0.f};
    hscaldoublecomplex c_b2 = {0.f,0.f};
    hscalinteger c__1 = 1;
    hscalinteger c_n1 = -1;

    /* System generated locals */
    hscalinteger i__1, i__2;
    hscaldoublecomplex z__1;
    hscalinteger mp, nq, ldc, ncc, iic, jjc, ldv, ncv, iiv, jjv, ipw;
    hscalinteger icoff, ioffc, iccol;
    hscalinteger iroff, ioffv, ivcol;
    hscalinteger rdest, icrow, ictxt, ivrow;
    hscallogical ccblck, crblck;
    char colbtop[1];
    char rowbtop[1];

/*
*  .. Local Arrays ..
*/
    HSCALBLACSCONTEXT   * blacsctxt;

    char sidec      = SIDEI, *side = &sidec;

    int     *m      = &M;
    int     *n      = &N;
    int     *iv     = &IV;
    int     *jv     = &JV;
    int     *descv  = (int*)DESCV;
    int     *incv   = &INCV;
    int     *ic     = &IC;
    int     *jc     = &JC;
    int     *descc  = (int*)DESCC;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descc;
    --descv;

    if (init_llds_f == 1)
    {
       int *PXLARF_DESCV_l = &descv[1];
       int *PXLARF_DESCC_l = &descc[1];
       HSCAL_PXLARF_INIT_LLDS(
            PXLARF_DESCV_l,
            PXLARF_DESCC_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pzlarf: (%d, %d): (descv+1)[M1_]=%d (descv+1)[MB1_]=%d myrow=%d (descv+1)[RSRC1_]=%d nprow=%d (descv+1)[LLD1_]=%d", 
        myrow, mycol, (descv+1)[M1_], (descv+1)[MB1_], myrow, (descv+1)[RSRC1_], nprow, (descv+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pzlarf: (%d, %d): (descc+1)[M1_]=%d (descc+1)[MB1_]=%d myrow=%d (descc+1)[RSRC1_]=%d nprow=%d (descc+1)[LLD1_]=%d",
        myrow, mycol, (descc+1)[M1_], (descc+1)[MB1_], myrow, (descc+1)[RSRC1_], nprow, (descc+1)[LLD1_]
    );

    /* Function Body */
    if (*m <= 0 || *n <= 0) {
	return HSCAL_SUCCESS;
    }

    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*     Figure local indexes */

    infog2l_(ic, jc, &descc[1], &nprow, &npcol, &myrow, &mycol, &iic, &jjc, &
	    icrow, &iccol);
    infog2l_(iv, jv, &descv[1], &nprow, &npcol, &myrow, &mycol, &iiv, &jjv, &
	    ivrow, &ivcol);
    ncc = numroc_(&descc[4], &descc[6], &mycol, &descc[8], &npcol);
    ncv = numroc_(&descv[4], &descv[6], &mycol, &descv[8], &npcol);
    ldc = descc[9];
    ldv = descv[9];
    iic = min(iic,ldc);
    iiv = min(iiv,ldv);
    jjc = min(jjc,ncc);
    jjv = min(jjv,ncv);
    ioffc = iic + (jjc - 1) * ldc;
    ioffv = iiv + (jjv - 1) * ldv;

    iroff = (*ic - 1) % descc[5];
    icoff = (*jc - 1) % descc[6];
    i__1 = *m + iroff;
    mp = numroc_(&i__1, &descc[5], &myrow, &icrow, &nprow);
    i__1 = *n + icoff;
    nq = numroc_(&i__1, &descc[6], &mycol, &iccol, &npcol);
    if (myrow == icrow) {
	mp -= iroff;
    }
    if (mycol == iccol) {
	nq -= icoff;
    }

/*     Is sub( C ) only distributed over a process row ? */

    crblck = *m <= descc[5] - iroff;

/*     Is sub( C ) only distributed over a process column ? */

    ccblck = *n <= descc[6] - icoff;

    if (hscal_lsame(side, "L")) {

	if (crblck) {
	    rdest = icrow;
	} else {
	    rdest = -1;
	}

	if (ccblck) {

/*           sub( C ) is distributed over a process column */

	    if (descv[3] == *incv) {

/*              Transpose row vector V */

		ipw = mp + 1;
		hscal_pbztrnv(
                 &ictxt, "Rowwise", "Transpose", m, &descv[6], &iroff,
                 &ldv, &c_b2, &c__1, &ivrow, &ivcol, &icrow, &iccol, 
                 nprow, npcol,
                 myrow, mycol,
                 tcomp, tcomm
            );

/*              Perform the local computation within a process column */

		if (mycol == iccol) {

		    if (myrow == ivrow) {

			hscal_Czgebs2d(
                       blacsctxt, "Columnwise", " ", c__1, c__1, 
                       c__1,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

		    } else {

			hscal_Czgebr2d(
                       blacsctxt, "Columnwise", " ", c__1, c__1, 
                       c__1, ivrow, mycol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

		    }

		    {

/*                    w := sub( C )' * v */

			if (mp > 0) {
			    zgemv_tcomp_macro(
                           "Conjugate transpose", &mp, &nq, (char*)&c_b1, 
                           &ldc, &c__1, (char*)&c_b2, 
		               &c__1,
                           nprow, npcol, myrow, mycol, tcomp
                      );
			} else {
			    i__1 = max(1,nq);
			    zlaset_tcomp_macro(
                            "All", &nq, &c__1, (char*)&c_b2, (char*)&c_b2, 
                            &i__1,
                            nprow, npcol, myrow, mycol, tcomp
                      );
			}
			i__1 = max(1,nq);
			hscal_Czgsum2d(
                       blacsctxt, "Columnwise", " ", nq, c__1,
			     i__1, rdest, mycol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

/*                    sub( C ) := sub( C ) - v * w' */
/*
* Assume some values of z__1 (as these are not real computations)
*/
			z__1.r = -.5, z__1.i = -.5;
			zgerc_tcomp_macro(
                       &mp, &nq, (char*)&z__1, &c__1, 
                       &c__1, &ldc,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }

		}

	    } else {

/*              V is a column vector */

		if (ivcol == iccol) {

/*                 Perform the local computation within a process column */

		    if (mycol == iccol) {

			{

/*                       w := sub( C )' * v */

			    if (mp > 0) {
				zgemv_tcomp_macro(
                             "Conjugate transpose", &mp, &nq, (char*)&c_b1,
		                 &ldc, &c__1, 
		                 (char*)&c_b2, &c__1, 
                             nprow, npcol, myrow, mycol, tcomp
                        );
			    } else {
				i__1 = max(1,nq);
				zlaset_tcomp_macro(
                              "All", &nq, &c__1, (char*)&c_b2, (char*)&c_b2,
					&i__1,
                              nprow, npcol, myrow, mycol, tcomp
                        );
			    }
			    i__1 = max(1,nq);
			    hscal_Czgsum2d(
                           blacsctxt, "Columnwise", " ", nq, c__1,
		               i__1, rdest, mycol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );

/*                       sub( C ) := sub( C ) - v * w' */
/*
* Assume some values of z__1 (as these are not real computations)
*/
                      z__1.r = -.5, z__1.i = -.5;
			    zgerc_tcomp_macro(
                           &mp, &nq, (char*)&z__1, &c__1, &c__1, 
                           &ldc,
                           nprow, npcol, myrow, mycol, tcomp
                      );
			}

		    }

		} else {

/*                 Send V and TAU to the process column ICCOL */

		    if (mycol == ivcol) {

			ipw = mp + 1;
			zcopy_tcomp_macro(
                       &mp, &c__1, &c__1,
                       nprow, npcol, myrow, mycol, tcomp
                  );
			hscal_Czgesd2d(
                       blacsctxt, ipw, c__1, ipw, myrow, 
			     iccol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

		    } else if (mycol == iccol) {

			ipw = mp + 1;

			{

/*                       w := sub( C )' * v */

			    if (mp > 0) {
				zgemv_tcomp_macro(
                             "Conjugate transpose", &mp, &nq, (char*)&c_b1,
				     &ldc, &c__1, (char*)&c_b2, &c__1,
                             nprow, npcol, myrow, mycol, tcomp
                        );
			    } else {
				i__1 = max(1,nq);
				zlaset_tcomp_macro(
                              "All", &nq, &c__1, (char*)&c_b2, (char*)&c_b2,
					&i__1,
                              nprow, npcol, myrow, mycol, tcomp
                        );
			    }
			    i__1 = max(1,nq);
			    hscal_Czgsum2d(
                           blacsctxt, "Columnwise", " ", nq, c__1,
				   i__1, rdest, mycol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );

/*                       sub( C ) := sub( C ) - v * w' */
/*
* Assume some values of z__1 (as these are not real computations)
*/
                      z__1.r = -.5, z__1.i = -.5;
			    zgerc_tcomp_macro(
                           &mp, &nq, (char*)&z__1, &c__1,
				   &c__1, &ldc,
                           nprow, npcol, myrow, mycol, tcomp
                      );
			}

		    }

		}

	    }

	} else {

/*           sub( C ) is a proper distributed matrix */

	    if (descv[3] == *incv) {

/*              Transpose and broadcast row vector V */

		ipw = mp + 1;
		hscal_pbztrnv(
                 &ictxt, "Rowwise", "Transpose", m, &descv[6], &iroff,
		     &ldv, &c_b2, &c__1, &ivrow, &ivcol, &icrow, &c_n1,
                 nprow, npcol,
                 myrow, mycol,
                 tcomp, tcomm
            );

/*              Perform the local computation within a process column */

		if (myrow == ivrow) {

		    hscal_Czgebs2d(
                     blacsctxt, "Columnwise", " ", c__1, c__1,
			   c__1,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

		} else {

		    hscal_Czgebr2d(
                     blacsctxt, "Columnwise", " ", c__1, c__1,
			   c__1, ivrow, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

		}

		{

/*                 w := sub( C )' * v */

		    if (mp > 0) {
			if (ioffc > 0) {
			    zgemv_tcomp_macro(
                           "Conjugate transpose", &mp, &nq, (char*)&c_b1,
				   &ldc, &c__1, (char*)&c_b2, &c__1, 
                           nprow, npcol, myrow, mycol, tcomp
                      );
			}
		    } else {
			i__1 = max(1,nq);
			zlaset_tcomp_macro(
                        "All", &nq, &c__1, (char*)&c_b2, (char*)&c_b2,
                        &i__1,
                        nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    i__1 = max(1,nq);
		    hscal_Czgsum2d(
                     blacsctxt, "Columnwise", " ", nq, c__1, i__1, 
                     rdest, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

/*                 sub( C ) := sub( C ) - v * w' */

		    if (ioffc > 0) {
/*
* Assume some values of z__1 (as these are not real computations)
*/
			z__1.r = -.5, z__1.i = -.5;
			zgerc_tcomp_macro(
                       &mp, &nq, (char*)&z__1, &c__1, 
                       &c__1, &ldc,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		}

	    } else {

/*              Broadcast column vector V */

		pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
		if (mycol == ivcol) {

		    ipw = mp + 1;
		    zcopy_tcomp_macro(
                     &mp, &c__1, &c__1,
                     nprow, npcol, myrow, mycol, tcomp
                );
		    hscal_Czgebs2d(
                     blacsctxt, "Rowwise", rowbtop, ipw, c__1,
                     ipw,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

		} else {

		    ipw = mp + 1;
		    hscal_Czgebr2d(
                     blacsctxt, "Rowwise", rowbtop, ipw, c__1,
                     ipw, myrow, ivcol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

		}

		{

/*                 w := sub( C )' * v */

		    if (mp > 0) {
			if (ioffc > 0) {
			    zgemv_tcomp_macro(
                           "Conjugate transpose", &mp, &nq, (char*)&c_b1,
				   &ldc, &c__1, (char*)&c_b2, 
				   &c__1,
                           nprow, npcol, myrow, mycol, tcomp
                      );
			}
		    } else {
			i__1 = max(1,nq);
			zlaset_tcomp_macro(
                        "All", &nq, &c__1, (char*)&c_b2, (char*)&c_b2,
                        &i__1,
                        nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    i__1 = max(1,nq);
		    hscal_Czgsum2d(
                     blacsctxt, "Columnwise", " ", nq, c__1,
                     i__1, rdest, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

/*                 sub( C ) := sub( C ) - v * w' */

		    if (ioffc > 0) {
/*
* Assume some values of z__1 (as these are not real computations)
*/
			z__1.r = -.5, z__1.i = -.5;
			zgerc_tcomp_macro(
                       &mp, &nq, (char*)&z__1, &c__1, 
                       &c__1, &ldc,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		}

	    }

	}

    } else {

	if (ccblck) {
	    rdest = myrow;
	} else {
	    rdest = -1;
	}

	if (crblck) {

/*           sub( C ) is distributed over a process row */

	    if (descv[3] == *incv) {

/*              V is a row vector */

		if (ivrow == icrow) {

/*                 Perform the local computation within a process row */

		    if (myrow == icrow) {

			{

/*                       w := sub( C ) * v */

			    if (nq > 0) {
				zgemv_tcomp_macro(
                             "No transpose", &mp, &nq, (char*)&c_b1, &ldc,
                             &ldv, (char*)&c_b2, &c__1,
                             nprow, npcol, myrow, mycol, tcomp
                        );
			    } else {
				i__1 = max(1,mp);
				zlaset_tcomp_macro(
                              "All", &mp, &c__1, (char*)&c_b2, (char*)&c_b2,
					&i__1,
                              nprow, npcol, myrow, mycol, tcomp
                        );
			    }
			    i__1 = max(1,mp);
			    hscal_Czgsum2d(
                           blacsctxt, "Rowwise", " ", mp, c__1, i__1, 
                           rdest, iccol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );

/*                       sub( C ) := sub( C ) - w * v' */

			    if (ioffv > 0 && ioffc > 0) {
/*
* Assume some values of z__1 (as these are not real computations)
*/
                        z__1.r = -.5, z__1.i = -.5;
				zgerc_tcomp_macro(
                             &mp, &nq, (char*)&z__1, &c__1, &ldv, &ldc,
                             nprow, npcol, myrow, mycol, tcomp
                        );
			    }
			}

		    }

		} else {

/*                 Send V and TAU to the process row ICROW */

		    if (myrow == ivrow) {

			ipw = nq + 1;
			zcopy_tcomp_macro(
                       &nq, &ldv, &c__1,
                       nprow, npcol, myrow, mycol, tcomp
                  );
			hscal_Czgesd2d(
                       blacsctxt, ipw, c__1, ipw, icrow, 
			     mycol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

		    } else if (myrow == icrow) {

			ipw = nq + 1;

			{

/*                       w := sub( C ) * v */

			    if (nq > 0) {
				zgemv_tcomp_macro(
                             "No transpose", &mp, &nq, (char*)&c_b1,
				     &ldc, &c__1, (char*)&c_b2, 
				     &c__1,
                             nprow, npcol, myrow, mycol, tcomp
                        );
			    } else {
				i__1 = max(1,mp);
				zlaset_tcomp_macro(
                              "All", &mp, &c__1, (char*)&c_b2, (char*)&c_b2,
					&i__1, 
                              nprow, npcol, myrow, mycol, tcomp
                        );
			    }
			    i__1 = max(1,mp);
			    hscal_Czgsum2d(
                           blacsctxt, "Rowwise", " ", mp, c__1,
				   i__1, rdest, iccol,
                           nprow, npcol, myrow, mycol, tcomp, tcomm
                      );

/*                       sub( C ) := sub( C ) - w * v' */
/*
* Assume some values of z__1 (as these are not real computations)
*/
                      z__1.r = -.5, z__1.i = -.5;
			    zgerc_tcomp_macro(
                           &mp, &nq, (char*)&z__1, &c__1,
				   &c__1, &ldc,
                           nprow, npcol, myrow, mycol, tcomp
                      );
			}

		    }

		}

	    } else {

/*              Transpose column vector V */

		ipw = nq + 1;
		hscal_pbztrnv(
                 &ictxt, "Columnwise", "Transpose", n, &descv[5], 
                 &icoff, &c__1, &c_b2, &c__1, &ivrow, &ivcol, &icrow, &iccol,
                 nprow, npcol,
                 myrow, mycol,
                 tcomp, tcomm
            );

/*              Perform the local computation within a process column */

		if (myrow == icrow) {

		    if (mycol == ivcol) {

			hscal_Czgebs2d(
                       blacsctxt, "Rowwise", " ", c__1, c__1, c__1,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

		    } else {

			hscal_Czgebr2d(
                       blacsctxt, "Rowwise", " ", c__1, c__1, 
			     c__1, myrow, ivcol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

		    }

		    {

/*                    w := sub( C ) * v */

			if (nq > 0) {
			    zgemv_tcomp_macro(
                           "No transpose", &mp, &nq, (char*)&c_b1,
				   &ldc, &c__1, (char*)&c_b2, &c__1,
                           nprow, npcol, myrow, mycol, tcomp
                      );
			} else {
			    i__1 = max(1,mp);
			    zlaset_tcomp_macro(
                            "All", &mp, &c__1, (char*)&c_b2, (char*)&c_b2,
				    &i__1, 
                            nprow, npcol, myrow, mycol, tcomp
                      );
			}
			i__1 = max(1,mp);
			hscal_Czgsum2d(
                       blacsctxt, "Rowwise", " ", mp, c__1,
			     i__1, rdest, iccol,
                       nprow, npcol, myrow, mycol, tcomp, tcomm
                  );

/*                    sub( C ) := sub( C ) - w * v' */
/*
* Assume some values of z__1 (as these are not real computations)
*/
                  z__1.r = -.5, z__1.i = -.5;
			zgerc_tcomp_macro(
                       &mp, &nq, (char*)&z__1, &c__1, 
                       &c__1, &ldc,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }

		}

	    }

	} else {

/*           sub( C ) is a proper distributed matrix */

	    if (descv[3] == *incv) {

/*              Broadcast row vector V */

		pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);
		if (myrow == ivrow) {

		    ipw = nq + 1;
		    if (ioffv > 0) {
			zcopy_tcomp_macro(
                       &nq, &ldv, &c__1,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    hscal_Czgebs2d(
                     blacsctxt, "Columnwise", colbtop, ipw, c__1,
			   ipw,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

		} else {

		    ipw = nq + 1;
		    hscal_Czgebr2d(
                     blacsctxt, "Columnwise", colbtop, ipw, c__1,
			   ipw, ivrow, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

		}

		{

/*                 w := sub( C ) * v */

		    if (nq > 0) {
			zgemv_tcomp_macro(
                       "No Transpose", &mp, &nq, (char*)&c_b1, 
                       &ldc, &c__1, (char*)&c_b2, &c__1,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    } else {
			i__1 = max(1,mp);
			zlaset_tcomp_macro(
                        "All", &mp, &c__1, (char*)&c_b2, (char*)&c_b2,
                        &i__1,
                        nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    i__1 = max(1,mp);
		    hscal_Czgsum2d(
                     blacsctxt, "Rowwise", " ", mp, c__1, 
		         i__1, rdest, iccol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

/*                 sub( C ) := sub( C ) - w * v' */

		    if (ioffc > 0) {
/*
* Assume some values of z__1 (as these are not real computations)
*/
                  z__1.r = -.5, z__1.i = -.5;
			zgerc_tcomp_macro(
                       &mp, &nq, (char*)&z__1, &c__1, 
                       &c__1, &ldc,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		}

	    } else {

/*              Transpose and broadcast column vector V */

		ipw = nq + 1;
		hscal_pbztrnv(
                 &ictxt, "Columnwise", "Transpose", n, &descv[5], 
                 &icoff, &c__1, &c_b2, &c__1, &ivrow, &ivcol, &c_n1, &iccol,
                 nprow, npcol,
                 myrow, mycol,
                 tcomp, tcomm
            );

/*              Perform the local computation within a process column */

		if (mycol == ivcol) {

		    hscal_Czgebs2d(
                     blacsctxt, "Rowwise", " ", c__1, c__1, 
			   c__1,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

		} else {

		    hscal_Czgebr2d(
                     blacsctxt, "Rowwise", " ", c__1, c__1,
                     c__1, myrow, ivcol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

		}

		{

/*                 w := sub( C ) * v */

		    if (nq > 0) {
			zgemv_tcomp_macro(
                       "No transpose", &mp, &nq, (char*)&c_b1, 
                       &ldc, &c__1, (char*)&c_b2,
                       &c__1,
                       nprow, npcol, myrow, mycol, tcomp
                  );
		    } else {
			i__1 = max(1,mp);
			zlaset_tcomp_macro(
                        "All", &mp, &c__1, (char*)&c_b2, (char*)&c_b2,
                        &i__1, 
                        nprow, npcol, myrow, mycol, tcomp
                  );
		    }
		    i__1 = max(1,mp);
		    hscal_Czgsum2d(
                     blacsctxt, "Rowwise", " ", mp, c__1, 
			   i__1, rdest, iccol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );

/*                 sub( C ) := sub( C ) - w * v' */
/*
* Assume some values of z__1 (as these are not real computations)
*/
                z__1.r = -.5, z__1.i = -.5;
		    zgerc_tcomp_macro(
                     &mp, &nq, (char*)&z__1, &c__1, 
                     &c__1, &ldc,
                     nprow, npcol, myrow, mycol, tcomp
                );
		}

	    }

	}

    }

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzlarf_tcomp_tcomm(
   int SIDEI,
   int M,
   int N,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int INCV,
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzlarf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI,
                M, N,
                IV, JV, DESCV, INCV,
                IC, JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzlarf_info(
   F_CHAR_T SIDE,
   int * M,
   int * N,
   int * IV,
   int * JV,
   int * DESCV,
   int * INCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzlarf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI,
                *M, *N,
                *IV, *JV, DESCV, *INCV,
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzlarf_info_i(
   F_CHAR_T SIDE,
   int * M,
   int * N,
   int * IV,
   int * JV,
   int * DESCV,
   int * INCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzlarf_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI,
                *M, *N,
                *IV, *JV, DESCV, *INCV,
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
