
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzlarfb_recon_tcomp(
   int sidei,
   int transi,
   int directi,
   int storei,
   int m,
   int n,
   int k
)
{
    double tcomp = 0;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pzlarfb_my_tcomp(
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int DIRECTI,
   int STOREI,
   int M,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzlarfb_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          DIRECTI, STOREI,
          M, N, K,
          IV, JV, DESCV,
          IC, JC, DESCC, 
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzlarfb_my_tcomm(
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int DIRECTI,
   int STOREI,
   int M,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzlarfb_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          DIRECTI, STOREI,
          M, N, K,
          IV, JV, DESCV,
          IC, JC, DESCC, 
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzlarfb_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int SIDEI,
   int TRANSI,
   int DIRECTI,
   int STOREI,
   int M,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pzlarfb_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          DIRECTI, STOREI,
          M, N, K,
          IV, JV, DESCV,
          IC, JC, DESCC, 
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pzlarfb_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int DIRECTI,
   int STOREI,
   int M,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscaldoublecomplex c_b1 = {1.,0.};
    hscaldoublecomplex c_b2 = {0.,0.};
    hscalinteger c_n1 = -1;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4;
    hscaldoublecomplex z__1;

    /* Local variables */
    hscalinteger ilastcol, ilastrow;
    hscalinteger ii, jj, kp, kq, lv, lw, ldc, iic, jjc, mpc, nqc, mbv, ldv, 
	    nbv, iiv, jjv, ipt, ipv, ipw, mqv, npv, mpc0, nqc0, ipw1, mqv0, 
	    npv0, ioff, wide, itop;
    char uplo[1];
    hscalinteger iibeg, ibase, jjbeg;
    hscalinteger ioffc, iiend, iccol, jjend;
    hscalinteger ileft, ioffv, ivcol;
    hscalinteger icrow, ictxt, iinxt, jjnxt, ivrow,
	    icoffc, height, iroffc;
    hscalinteger icoffv;
    hscalinteger iright, iroffv;
    hscalinteger mydist;
    char transt[1];
    char colbtop[1];
    hscallogical forward;
    char rowbtop[1];

/*
*  .. Local Arrays ..
*/
    HSCALBLACSCONTEXT   * blacsctxt;

    char sidec      = SIDEI, *side = &sidec;
    char transc     = TRANSI, *trans = &transc;
    char directc    = DIRECTI, *direct = &directc;
    char storec     = STOREI, *storev = &storec;

    int     *m      = &M;
    int     *n      = &N;
    int     *k      = &K;
    int     *iv     = &IV;
    int     *jv     = &JV;
    int     *descv  = (int*)DESCV;
    int     *ic     = &IC;
    int     *jc     = &JC;
    int     *descc  = (int*)DESCC;

    double  * tcomp_l = tcomp;
    int     * tcomm_l = tcomm;

    /* Parameter adjustments */
    --descc;
    --descv;

    if (init_llds_f == 1)
    {
       int *PXLARFB_DESCV_l = &descv[1];
       int *PXLARFB_DESCC_l = &descc[1];
       HSCAL_PXLARFB_INIT_LLDS(
            PXLARFB_DESCV_l,
            PXLARFB_DESCC_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pzlarfb: (%d, %d): (descv+1)[M1_]=%d (descv+1)[MB1_]=%d myrow=%d (descv+1)[RSRC1_]=%d nprow=%d (descv+1)[LLD1_]=%d", 
        myrow, mycol, (descv+1)[M1_], (descv+1)[MB1_], myrow, (descv+1)[RSRC1_], nprow, (descv+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pzlarfb: (%d, %d): (descc+1)[M1_]=%d (descc+1)[MB1_]=%d myrow=%d (descc+1)[RSRC1_]=%d nprow=%d (descc+1)[LLD1_]=%d",
        myrow, mycol, (descc+1)[M1_], (descc+1)[MB1_], myrow, (descc+1)[RSRC1_], nprow, (descc+1)[LLD1_]
    );

    /* Function Body */
    if (*m <= 0 || *n <= 0 || *k <= 0) {
	return HSCAL_SUCCESS;
    }

    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

    if (hscal_lsame(trans, "N")) {
	*(unsigned char *)transt = 'C';
    } else {
	*(unsigned char *)transt = 'N';
    }
    forward = hscal_lsame(direct, "F");
    if (forward) {
	*(unsigned char *)uplo = 'U';
    } else {
	*(unsigned char *)uplo = 'L';
    }

    infog2l_(iv, jv, &descv[1], &nprow, &npcol, &myrow, &mycol, &iiv, &jjv, &
	    ivrow, &ivcol);
    infog2l_(ic, jc, &descc[1], &nprow, &npcol, &myrow, &mycol, &iic, &jjc, &
	    icrow, &iccol);
    ldc = descc[9];
    ldv = descv[9];
    iic = min(iic,ldc);
    iiv = min(iiv,ldv);
    iroffc = (*ic - 1) % descc[5];
    icoffc = (*jc - 1) % descc[6];
    mbv = descv[5];
    nbv = descv[6];
    iroffv = (*iv - 1) % mbv;
    icoffv = (*jv - 1) % nbv;
    i__1 = *m + iroffc;
    mpc = numroc_(&i__1, &descc[5], &myrow, &icrow, &nprow);
    i__1 = *n + icoffc;
    nqc = numroc_(&i__1, &descc[6], &mycol, &iccol, &npcol);
    if (mycol == iccol) {
	nqc -= icoffc;
    }
    if (myrow == icrow) {
	mpc -= iroffc;
    }
/* Computing MIN */
/* Computing MAX */
    i__3 = 1, i__4 = jjc + nqc - 1;
    i__1 = jjc, i__2 = max(i__3,i__4);
    jjc = min(i__1,i__2);
/* Computing MIN */
/* Computing MAX */
    i__3 = 1, i__4 = numroc_(&descv[4], &nbv, &mycol, &descv[8], &npcol);
    i__1 = jjv, i__2 = max(i__3,i__4);
    jjv = min(i__1,i__2);
    ioffc = iic + (jjc - 1) * ldc;
    ioffv = iiv + (jjv - 1) * ldv;

    if (hscal_lsame(storev, "C")) {

/*        V is stored columnwise */

	if (hscal_lsame(side, "L")) {

/*           Form  Q*sub( C )  or  Q'*sub( C  ) */

/*           Locally V( IOFFV ) is MPV x K, C( IOFFC ) is MPC x NQC */
/*           WORK( IPV ) is MPC x K = V( IOFFV ), MPC = MPV */
/*           WORK( IPW ) is NQC x K = C( IOFFC )' * V( IOFFV ) */

	    ipv = 1;
	    ipw = ipv + mpc * *k;
	    lv = max(1,mpc);
	    lw = max(1,nqc);

/*           Broadcast V to the other process columns. */

	    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
	    if (mycol == ivcol) {
		hscal_Czgebs2d(
                 blacsctxt, "Rowwise", rowbtop, mpc, *k, ldv,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
		);
		if (myrow == ivrow) {
		    hscal_Cztrbs2d(
                     blacsctxt, "Rowwise", rowbtop, uplo, "Non unit", *k, 
	               *k, nbv, 
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
		zlacpy_tcomp_macro(
                  "All", &mpc, k, &ldv, &lv, 
                  nprow, npcol, myrow, mycol, tcomp
            );
	    } else {
		hscal_Czgebr2d(
                 blacsctxt, "Rowwise", rowbtop, mpc, *k, lv,
		     myrow, ivcol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
		if (myrow == ivrow) {
		    hscal_Cztrbr2d(
                     blacsctxt, "Rowwise", rowbtop, uplo, "Non unit", *k, 
			   *k, nbv, myrow, ivcol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
	    }

	    if (forward) {

/*              WORK(IPV) = ( V1 ) where V1 is unit lower triangular, */
/*                          ( V2 ) zeroes upper triangular part of V1 */

		mydist = (myrow - ivrow + nprow) % nprow;
/* Computing MAX */
		i__1 = 0, i__2 = mydist * mbv - iroffv;
		itop = max(i__1,i__2);
		iibeg = iiv;
		iiend = iibeg + mpc - 1;
/* Computing MIN */
		i__1 = hscal_iceil(&iibeg, &mbv) * mbv;
		iinxt = min(i__1,iiend);

L10:
		if (*k - itop > 0) {
		    i__1 = iinxt - iibeg + 1;
		    i__2 = *k - itop;
		    zlaset_tcomp_macro(
                      "Upper", &i__1, &i__2, (char*)&c_b2, (char*)&c_b1, &lv,
                       nprow, npcol, myrow, mycol, tcomp
                );
		    mydist += nprow;
		    itop = mydist * mbv - iroffv;
		    iibeg = iinxt + 1;
/* Computing MIN */
		    i__1 = iinxt + mbv;
		    iinxt = min(i__1,iiend);
		    goto L10;
		}

	    } else {

/*              WORK(IPV) = ( V1 ) where V2 is unit upper triangular, */
/*                          ( V2 ) zeroes lower triangular part of V2 */

		jj = jjv;
		ioff = (*iv + *m - *k - 1) % mbv;
		i__1 = *iv + *m - *k;
		infog1l_(&i__1, &mbv, &nprow, &myrow, &descv[7], &ii, &
			ilastrow);
		i__1 = *k + ioff;
		kp = numroc_(&i__1, &mbv, &myrow, &ilastrow, &nprow);
		if (myrow == ilastrow) {
		    kp -= ioff;
		}
		mydist = (myrow - ilastrow + nprow) % nprow;
		itop = mydist * mbv - ioff;
/* Computing MIN */
		i__1 = itop + mbv;
		ibase = min(i__1,*k);
/* Computing MIN */
		i__1 = max(0,itop);
		itop = min(i__1,*k);

L20:
		if (jj <= jjv + *k - 1) {
		    height = ibase - itop;
		    i__1 = itop - jj + jjv;
		    zlaset_tcomp_macro(
                      "All", &kp, &i__1, (char*)&c_b2, (char*)&c_b2, &lv,
                       nprow, npcol, myrow, mycol, tcomp
                );
		    zlaset_tcomp_macro(
                      "Lower", &kp, &height, (char*)&c_b2, (char*)&c_b1, &lv,
                      nprow, npcol, myrow, mycol, tcomp
                );
/* Computing MAX */
		    i__1 = 0, i__2 = kp - height;
		    kp = max(i__1,i__2);
		    ii += height;
		    jj = jjv + ibase;
		    mydist += nprow;
		    itop = mydist * mbv - ioff;
/* Computing MIN */
		    i__1 = itop + mbv;
		    ibase = min(i__1,*k);
		    itop = min(itop,*k);
		    goto L20;
		}

	    }

/*           WORK( IPW ) = C( IOFFC )' * V  (NQC x MPC x K) -> NQC x K */

	    if (mpc > 0) {
		zgemm_tcomp_macro(
                 "Conjugate transpose", "No transpose", &nqc, k, &mpc, 
                 (char*)&c_b1, &ldc, &lv, (char*)&c_b2, &lw,
                 nprow, npcol, myrow, mycol, tcomp
            );
	    } else {
		zlaset_tcomp_macro(
                  "All", &nqc, k, (char*)&c_b2, (char*)&c_b2, &lw, 
                  nprow, npcol, myrow, mycol, tcomp
            );
	    }

	    hscal_Czgsum2d(
               blacsctxt, "Columnwise", " ", nqc, *k, lw,
		   ivrow, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

	    if (myrow == ivrow) {

/*              WORK( IPW ) = WORK( IPW ) * T' or WORK( IPW ) * T */

		ztrmm_tcomp_macro(
                 "Right", uplo, transt, "Non unit", &nqc, k, (char*)&c_b1, 
                 &nbv, &lw,
                 nprow, npcol, myrow, mycol, tcomp
            );
		hscal_Czgebs2d(
                 blacsctxt, "Columnwise", " ", nqc, *k, lw, 
                 nprow, npcol, myrow, mycol, tcomp, tcomm
		);
	    } else {
		hscal_Czgebr2d(
                 blacsctxt, "Columnwise", " ", nqc, *k, lw, 
		     ivrow, mycol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
	    }

/*               C            C      -     V       *     W' */
/*           C( IOFFC ) = C( IOFFC ) - WORK( IPV ) * WORK( IPW )' */
/*                        MPC x NQC    MPC x K         K x NQC */

	    z__1.r = -1., z__1.i = -0.;
	    zgemm_tcomp_macro(
               "No transpose", "Conjugate transpose", &mpc, &nqc, k, 
               (char*)&z__1, &lv, &lw, (char*)&c_b1,
		   &ldc,
               nprow, npcol, myrow, mycol, tcomp
          );

	} else {

/*           Form sub( C )*Q or sub( C )*Q' */

/*           ICOFFC = IROFFV is required by the current transposition */
/*           routine PBCTRAN */

	    i__1 = *n + iroffv;
	    npv0 = numroc_(&i__1, &mbv, &myrow, &ivrow, &nprow);
	    if (myrow == ivrow) {
		npv = npv0 - iroffv;
	    } else {
		npv = npv0;
	    }
	    if (mycol == iccol) {
		nqc0 = nqc + icoffc;
	    } else {
		nqc0 = nqc;
	    }

/*           Locally V( IOFFV ) is NPV x K C( IOFFC ) is MPC x NQC */
/*           WORK( IPV ) is K x NQC0 = [ . V( IOFFV ) ]' */
/*           WORK( IPW ) is NPV0 x K = [ . V( IOFFV )' ]' */
/*           WORK( IPT ) is the workspace for PBCTRAN */

	    ipv = 1;
	    ipw = ipv + *k * nqc0;
	    ipt = ipw + npv0 * *k;
	    lv = max(1,*k);
	    lw = max(1,npv0);

	    if (mycol == ivcol) {
		if (myrow == ivrow) {
		    zlaset_tcomp_macro(
                      "All", &iroffv, k, (char*)&c_b2, (char*)&c_b2, &lw, 
                      nprow, npcol, myrow, mycol, tcomp
		    );
		    ipw1 = ipw + iroffv;
		    zlacpy_tcomp_macro(
                      "All", &npv, k, &ldv, &lw,
                      nprow, npcol, myrow, mycol, tcomp
		    );
		} else {
		    ipw1 = ipw;
		    zlacpy_tcomp_macro(
                      "All", &npv, k, &ldv, &lw,
                      nprow, npcol, myrow, mycol, tcomp
		    );
		}

		if (forward) {

/*                 WORK(IPW) = ( . V1' V2' )' where V1 is unit lower */
/*                 triangular, zeroes upper triangular part of V1 */

		    mydist = (myrow - ivrow + nprow) % nprow;
/* Computing MAX */
		    i__1 = 0, i__2 = mydist * mbv - iroffv;
		    itop = max(i__1,i__2);
		    iibeg = iiv;
		    iiend = iibeg + npv - 1;
/* Computing MIN */
		    i__1 = hscal_iceil(&iibeg, &mbv) * mbv;
		    iinxt = min(i__1,iiend);

L30:
		    if (*k - itop > 0) {
			i__1 = iinxt - iibeg + 1;
			i__2 = *k - itop;
			zlaset_tcomp_macro(
                        "Upper", &i__1, &i__2, (char*)&c_b2, (char*)&c_b1, 
                        &lw,
                        nprow, npcol, myrow, mycol, tcomp
                  );
			mydist += nprow;
			itop = mydist * mbv - iroffv;
			iibeg = iinxt + 1;
/* Computing MIN */
			i__1 = iinxt + mbv;
			iinxt = min(i__1,iiend);
			goto L30;
		    }

		} else {

/*                 WORK( IPW ) = ( . V1' V2' )' where V2 is unit upper */
/*                 triangular, zeroes lower triangular part of V2. */

		    jj = jjv;
		    i__1 = *iv + *n - *k;
		    infog1l_(&i__1, &mbv, &nprow, &myrow, &descv[7], &ii, &
			    ilastrow);
		    ioff = (*iv + *n - *k - 1) % mbv;
		    i__1 = *k + ioff;
		    kp = numroc_(&i__1, &mbv, &myrow, &ilastrow, &nprow);
		    if (myrow == ilastrow) {
			kp -= ioff;
		    }
		    mydist = (myrow - ilastrow + nprow) % nprow;
		    itop = mydist * mbv - ioff;
/* Computing MIN */
		    i__1 = itop + mbv;
		    ibase = min(i__1,*k);
/* Computing MIN */
		    i__1 = max(0,itop);
		    itop = min(i__1,*k);

L40:
		    if (jj <= jjv + *k - 1) {
			height = ibase - itop;
			i__1 = itop - jj + jjv;
			zlaset_tcomp_macro(
                        "All", &kp, &i__1, (char*)&c_b2, (char*)&c_b2, &lw,
                        nprow, npcol, myrow, mycol, tcomp
                  );
			zlaset_tcomp_macro(
                        "Lower", &kp, &height, (char*)&c_b2, (char*)&c_b1, &lw,
                        nprow, npcol, myrow, mycol, tcomp
                  );
/* Computing MAX */
			i__1 = 0, i__2 = kp - height;
			kp = max(i__1,i__2);
			ii += height;
			jj = jjv + ibase;
			mydist += nprow;
			itop = mydist * mbv - ioff;
/* Computing MIN */
			i__1 = itop + mbv;
			ibase = min(i__1,*k);
			itop = min(itop,*k);
			goto L40;
		    }
		}
	    }

	    i__1 = *n + iroffv;
	    hscal_pbztran(
               &ictxt, "Columnwise", "Conjugate transpose", &i__1, k, &mbv,
               &lw, &c_b2, &lv, &ivrow, &ivcol, &c_n1, &iccol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

/*           WORK( IPV ) = ( . V' ) -> WORK( IPV ) = V' is K x NQC */

	    if (mycol == iccol) {
		ipv += icoffc * lv;
	    }

/*           WORK( IPW ) becomes MPC x K = C( IOFFC ) * V */
/*           WORK( IPW ) = C( IOFFC ) * V  (MPC x NQC x K) -> MPC x K */

	    lw = max(1,mpc);

	    if (nqc > 0) {
		zgemm_tcomp_macro(
                 "No transpose", "Conjugate transpose", &mpc, k, &nqc, (char*)&c_b1,
                 &ldc, &lv, (char*)&c_b2,
		     &lw,
                 nprow, npcol, myrow, mycol, tcomp
            );
	    } else {
		zlaset_tcomp_macro(
                  "All", &mpc, k, (char*)&c_b2, (char*)&c_b2, &lw,
                  nprow, npcol, myrow, mycol, tcomp
            );
	    }

	    hscal_Czgsum2d(
               blacsctxt, "Rowwise", " ", mpc, *k, lw, myrow,
		   ivcol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

/*           WORK( IPW ) = WORK( IPW ) * T' or WORK( IPW ) * T */

	    if (mycol == ivcol) {
		if (myrow == ivrow) {

/*                 Broadcast the block reflector to the other rows. */

		    hscal_Cztrbs2d(
                     blacsctxt, "Columnwise", " ", uplo, "Non unit", *k, 
		         *k, nbv,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Cztrbr2d(
                     blacsctxt, "Columnwise", " ", uplo, "Non unit", *k, 
			   *k, nbv, ivrow, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
		ztrmm_tcomp_macro(
                 "Right", uplo, trans, "Non unit", &mpc, k, (char*)&c_b1,
		     &nbv, &lw,
                 nprow, npcol, myrow, mycol, tcomp
            );

		hscal_Czgebs2d(
                 blacsctxt, "Rowwise", " ", mpc, *k, lw,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
	    } else {
		hscal_Czgebr2d(
                 blacsctxt, "Rowwise", " ", mpc, *k, lw,
		     myrow, ivcol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
	    }

/*               C            C      -     W       *     V' */
/*           C( IOFFC ) = C( IOFFC ) - WORK( IPW ) * WORK( IPV ) */
/*                        MPC x NQC    MPC x K         K x NQC */

	    z__1.r = -1., z__1.i = -0.;
	    zgemm_tcomp_macro(
               "No transpose", "No transpose", &mpc, &nqc, k, (char*)&z__1,
		   &lw, &lv, (char*)&c_b1, &ldc,
               nprow, npcol, myrow, mycol, tcomp
	    );
	}

    } else {

/*        V is stored rowwise */

	if (hscal_lsame(side, "L")) {

/*           Form Q*sub( C ) or Q'*sub( C ) */

/*           IROFFC = ICOFFV is required by the current transposition */
/*           routine PBCTRAN */

	    i__1 = *m + icoffv;
	    mqv0 = numroc_(&i__1, &nbv, &mycol, &ivcol, &npcol);
	    if (mycol == ivcol) {
		mqv = mqv0 - icoffv;
	    } else {
		mqv = mqv0;
	    }
	    if (myrow == icrow) {
		mpc0 = mpc + iroffc;
	    } else {
		mpc0 = mpc;
	    }

/*           Locally V( IOFFV ) is K x MQV, C( IOFFC ) is MPC x NQC */
/*           WORK( IPV ) is MPC0 x K = [ . V( IOFFV ) ]' */
/*           WORK( IPW ) is K x MQV0 = [ . V( IOFFV ) ] */
/*           WORK( IPT ) is the workspace for PBCTRAN */

	    ipv = 1;
	    ipw = ipv + mpc0 * *k;
	    ipt = ipw + *k * mqv0;
	    lv = max(1,mpc0);
	    lw = max(1,*k);

	    if (myrow == ivrow) {
		if (mycol == ivcol) {
		    zlaset_tcomp_macro(
                      "All", k, &icoffv, (char*)&c_b2, (char*)&c_b2, &lw, 
                      nprow, npcol, myrow, mycol, tcomp
		    );
		    ipw1 = ipw + icoffv * lw;
		    zlacpy_tcomp_macro(
                      "All", k, &mqv, &ldv, &lw,
                      nprow, npcol, myrow, mycol, tcomp
		    );
		} else {
		    ipw1 = ipw;
		    zlacpy_tcomp_macro(
                      "All", k, &mqv, &ldv, &lw,
                      nprow, npcol, myrow, mycol, tcomp
		    );
		}

		if (forward) {

/*                 WORK( IPW ) = ( . V1 V2 ) where V1 is unit upper */
/*                 triangular, zeroes lower triangular part of V1 */

		    mydist = (mycol - ivcol + npcol) % npcol;
/* Computing MAX */
		    i__1 = 0, i__2 = mydist * nbv - icoffv;
		    ileft = max(i__1,i__2);
		    jjbeg = jjv;
		    jjend = jjv + mqv - 1;
/* Computing MIN */
		    i__1 = hscal_iceil(&jjbeg, &nbv) * nbv;
		    jjnxt = min(i__1,jjend);

L50:
		    if (*k - ileft > 0) {
			i__1 = *k - ileft;
			i__2 = jjnxt - jjbeg + 1;
			zlaset_tcomp_macro(
                        "Lower", &i__1, &i__2, (char*)&c_b2, (char*)&c_b1,
				&lw,
                        nprow, npcol, myrow, mycol, tcomp
                  );
			mydist += npcol;
			ileft = mydist * nbv - icoffv;
			jjbeg = jjnxt + 1;
/* Computing MIN */
			i__1 = jjnxt + nbv;
			jjnxt = min(i__1,jjend);
			goto L50;
		    }

		} else {

/*                 WORK( IPW ) = ( . V1 V2 ) where V2 is unit lower */
/*                 triangular, zeroes upper triangular part of V2. */

		    ii = iiv;
		    i__1 = *jv + *m - *k;
		    infog1l_(&i__1, &nbv, &npcol, &mycol, &descv[8], &jj, &
			    ilastcol);
		    ioff = (*jv + *m - *k - 1) % nbv;
		    i__1 = *k + ioff;
		    kq = numroc_(&i__1, &nbv, &mycol, &ilastcol, &npcol);
		    if (mycol == ilastcol) {
			kq -= ioff;
		    }
		    mydist = (mycol - ilastcol + npcol) % npcol;
		    ileft = mydist * nbv - ioff;
/* Computing MIN */
		    i__1 = ileft + nbv;
		    iright = min(i__1,*k);
/* Computing MIN */
		    i__1 = max(0,ileft);
		    ileft = min(i__1,*k);

L60:
		    if (ii <= iiv + *k - 1) {
			wide = iright - ileft;
			i__1 = ileft - ii + iiv;
			zlaset_tcomp_macro(
                        "All", &i__1, &kq, (char*)&c_b2, (char*)&c_b2,
				&lw,
                        nprow, npcol, myrow, mycol, tcomp
                  );
			zlaset_tcomp_macro(
                        "Upper", &wide, &kq, (char*)&c_b2, (char*)&c_b1, &lw,
                        nprow, npcol, myrow, mycol, tcomp
                  );
/* Computing MAX */
			i__1 = 0, i__2 = kq - wide;
			kq = max(i__1,i__2);
			ii = iiv + iright;
			jj += wide;
			mydist += npcol;
			ileft = mydist * nbv - ioff;
/* Computing MIN */
			i__1 = ileft + nbv;
			iright = min(i__1,*k);
			ileft = min(ileft,*k);
			goto L60;
		    }
		}
	    }

/*           WORK( IPV ) = WORK( IPW )' (replicated) is MPC0 x K */

	    i__1 = *m + icoffv;
	    hscal_pbztran(
               &ictxt, "Rowwise", "Conjugate transpose", k, &i__1, &nbv,
		   &lw, &c_b2, &lv, &ivrow, &ivcol, 
               &icrow, &c_n1,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

/*           WORK( IPV ) = ( . V )' -> WORK( IPV ) = V' is MPC x K */

	    if (myrow == icrow) {
		ipv += iroffc;
	    }

/*           WORK( IPW ) becomes NQC x K = C( IOFFC )' * V' */
/*           WORK( IPW ) = C( IOFFC )' * V'  (NQC x MPC x K) -> NQC x K */

	    lw = max(1,nqc);

	    if (mpc > 0) {
		zgemm_tcomp_macro(
                 "Conjugate transpose", "No transpose", &nqc, k, &mpc, (char*)&c_b1,
                 &ldc, &lv, (char*)&c_b2, &lw,
                 nprow, npcol, myrow, mycol, tcomp
            );
	    } else {
		zlaset_tcomp_macro(
                  "All", &nqc, k, (char*)&c_b2, (char*)&c_b2, &lw, 
                  nprow, npcol, myrow, mycol, tcomp
            );
	    }

	    hscal_Czgsum2d(
               blacsctxt, "Columnwise", " ", nqc, *k, lw,
		   ivrow, mycol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

/*           WORK( IPW ) = WORK( IPW ) * T' or WORK( IPW ) * T */

	    if (myrow == ivrow) {
		if (mycol == ivcol) {

/*                 Broadcast the block reflector to the other columns. */

		    hscal_Cztrbs2d(
                     blacsctxt, "Rowwise", " ", uplo, "Non unit", *k, *k,
			   mbv, 
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		} else {
		    hscal_Cztrbr2d(
                     blacsctxt, "Rowwise", " ", uplo, "Non unit", *k, *k,
			   mbv, myrow, ivcol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
		ztrmm_tcomp_macro(
                 "Right", uplo, transt, "Non unit", &nqc, k, (char*)&c_b1,
		     &mbv, &lw,
                 nprow, npcol, myrow, mycol, tcomp
            );

		hscal_Czgebs2d(
                 blacsctxt, "Columnwise", " ", nqc, *k, lw, 
                 nprow, npcol, myrow, mycol, tcomp, tcomm
		);
	    } else {
		hscal_Czgebr2d(
                 blacsctxt, "Columnwise", " ", nqc, *k, lw, 
		     ivrow, mycol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
	    }

/*               C            C      -     V'      *     W' */
/*           C( IOFFC ) = C( IOFFC ) - WORK( IPV ) * WORK( IPW )' */
/*                        MPC x NQC    MPC x K         K x NQC */

	    z__1.r = -1., z__1.i = -0.;
	    zgemm_tcomp_macro(
               "No transpose", "Conjugate transpose", &mpc, &nqc, k, (char*)&z__1, 
               &lv, &lw, (char*)&c_b1, &ldc,
               nprow, npcol, myrow, mycol, tcomp
          );

	} else {

/*           Form Q*sub( C ) or Q'*sub( C ) */

/*           Locally V( IOFFV ) is K x NQV, C( IOFFC ) is MPC x NQC */
/*           WORK( IPV ) is K x NQV = V( IOFFV ), NQV = NQC */
/*           WORK( IPW ) is MPC x K = C( IOFFC ) * V( IOFFV )' */

	    ipv = 1;
	    ipw = ipv + *k * nqc;
	    lv = max(1,*k);
	    lw = max(1,mpc);

/*           Broadcast V to the other process rows. */

	    pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);
	    if (myrow == ivrow) {
		hscal_Czgebs2d(
                 blacsctxt, "Columnwise", colbtop, *k, nqc,
		     ldv,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
		if (mycol == ivcol) {
		    hscal_Cztrbs2d(
                     blacsctxt, "Columnwise", colbtop, uplo, "Non unit", 
			   *k, *k, mbv,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
		zlacpy_tcomp_macro(
                  "All", k, &nqc, &ldv, &lv, 
                  nprow, npcol, myrow, mycol, tcomp
            );
	    } else {
		hscal_Czgebr2d(
                 blacsctxt, "Columnwise", colbtop, *k, nqc,
		     lv, ivrow, mycol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
		if (mycol == ivcol) {
		    hscal_Cztrbr2d(
                     blacsctxt, "Columnwise", colbtop, uplo, "Non unit", 
			   *k, *k, mbv, ivrow, mycol,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
		}
	    }

	    if (forward) {

/*              WORK(IPW) = ( V1 V2 ) where V1 is unit upper */
/*              triangular, zeroes lower triangular part of V1 */

		mydist = (mycol - ivcol + npcol) % npcol;
/* Computing MAX */
		i__1 = 0, i__2 = mydist * nbv - icoffv;
		ileft = max(i__1,i__2);
		jjbeg = jjv;
		jjend = jjv + nqc - 1;
/* Computing MIN */
		i__1 = hscal_iceil(&jjbeg, &nbv) * nbv;
		jjnxt = min(i__1,jjend);

L70:
		if (*k - ileft > 0) {
		    i__1 = *k - ileft;
		    i__2 = jjnxt - jjbeg + 1;
		    zlaset_tcomp_macro(
                      "Lower", &i__1, &i__2, (char*)&c_b2, (char*)&c_b1, &lv,
                      nprow, npcol, myrow, mycol, tcomp
                );
		    mydist += npcol;
		    ileft = mydist * nbv - icoffv;
		    jjbeg = jjnxt + 1;
/* Computing MIN */
		    i__1 = jjnxt + nbv;
		    jjnxt = min(i__1,jjend);
		    goto L70;
		}

	    } else {

/*              WORK( IPW ) = ( . V1 V2 ) where V2 is unit lower */
/*              triangular, zeroes upper triangular part of V2. */

		ii = iiv;
		i__1 = *jv + *n - *k;
		infog1l_(&i__1, &nbv, &npcol, &mycol, &descv[8], &jj, &
			ilastcol);
		ioff = (*jv + *n - *k - 1) % nbv;
		i__1 = *k + ioff;
		kq = numroc_(&i__1, &nbv, &mycol, &ilastcol, &npcol);
		if (mycol == ilastcol) {
		    kq -= ioff;
		}
		mydist = (mycol - ilastcol + npcol) % npcol;
		ileft = mydist * nbv - ioff;
/* Computing MIN */
		i__1 = ileft + nbv;
		iright = min(i__1,*k);
/* Computing MIN */
		i__1 = max(0,ileft);
		ileft = min(i__1,*k);

L80:
		if (ii <= iiv + *k - 1) {
		    wide = iright - ileft;
		    i__1 = ileft - ii + iiv;
		    zlaset_tcomp_macro(
                      "All", &i__1, &kq, (char*)&c_b2, (char*)&c_b2, &lv,
                      nprow, npcol, myrow, mycol, tcomp
                );
		    zlaset_tcomp_macro(
                      "Upper", &wide, &kq, (char*)&c_b2, (char*)&c_b1, &lv,
                      nprow, npcol, myrow, mycol, tcomp
                );
/* Computing MAX */
		    i__1 = 0, i__2 = kq - wide;
		    kq = max(i__1,i__2);
		    ii = iiv + iright;
		    jj += wide;
		    mydist += npcol;
		    ileft = mydist * nbv - ioff;
/* Computing MIN */
		    i__1 = ileft + nbv;
		    iright = min(i__1,*k);
		    ileft = min(ileft,*k);
		    goto L80;
		}

	    }

/*           WORK( IPV ) is K x NQC = V = V( IOFFV ) */
/*           WORK( IPW ) = C( IOFFC ) * V'  (MPC x NQC x K) -> MPC x K */

	    if (nqc > 0) {
		zgemm_tcomp_macro(
                 "No transpose", "Conjugate transpose", &mpc, k, &nqc, (char*)&c_b1,
                 &ldc, &lv, (char*)&c_b2, &lw,
                 nprow, npcol, myrow, mycol, tcomp
            );
	    } else {
		zlaset_tcomp_macro(
                  "All", &mpc, k, (char*)&c_b2, (char*)&c_b2, &lw,
                  nprow, npcol, myrow, mycol, tcomp
            );
	    }

	    hscal_Czgsum2d(
               blacsctxt, "Rowwise", " ", mpc, *k, lw, myrow,
		   ivcol,
               nprow, npcol, myrow, mycol, tcomp, tcomm
          );

/*           WORK( IPW ) = WORK( IPW ) * T' or WORK( IPW ) * T */

	    if (mycol == ivcol) {
		ztrmm_tcomp_macro(
                 "Right", uplo, trans, "Non unit", &mpc, k, (char*)&c_b1,
		     &mbv, &lw,
                 nprow, npcol, myrow, mycol, tcomp
            );
		hscal_Czgebs2d(
                 blacsctxt, "Rowwise", " ", mpc, *k, lw,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
	    } else {
		hscal_Czgebr2d(
                 blacsctxt, "Rowwise", " ", mpc, *k, lw,
		     myrow, ivcol,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
	    }

/*               C            C      -     W       *     V */
/*           C( IOFFC ) = C( IOFFC ) - WORK( IPW ) * WORK( IPV ) */
/*                        MPC x NQC    MPC x K         K x NQC */

	    z__1.r = -1., z__1.i = -0.;
	    zgemm_tcomp_macro(
               "No transpose", "No transpose", &mpc, &nqc, k, (char*)&z__1,
		   &lw, &lv, (char*)&c_b1, &ldc,
               nprow, npcol, myrow, mycol, tcomp
	    );

	}

    }

   Cblacs_gridexit_nocomm(blacsctxt);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzlarfb_tcomp_tcomm(
   int SIDEI,
   int TRANSI,
   int DIRECTI,
   int STOREI,
   int M,
   int N,
   int K,
   int IV,
   int JV,
   int DESCV[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzlarfb_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                DIRECTI, STOREI,
                M, N, K,
                IV, JV, DESCV,
                IC, JC, DESCC, 
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzlarfb_info(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * M,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzlarfb_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *SIDE, *TRANS, *DIRECT, *STOREV,
                *M, *N, *K,
                *IV, *JV, DESCV,
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzlarfb_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * M,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzlarfb_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                *SIDE, *TRANS, *DIRECT, *STOREV,
                *M, *N, *K,
                *IV, *JV, DESCV,
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
