
#include <math.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>

#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

double hscal_pzunml2_recon_tcomp(
   int sidei,
   int transi,
   int m,
   int n,
   int k
)
{
    double tcomp = 0;

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of recon computations %f",
         tcomp
    );

    return tcomp;
}

/*------------------------------------------------------------------*/

double hscal_pzunml2_my_tcomp(
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol
)
{
   int rc;
   double myt, *tcomp = (double*)calloc(nprow*npcol, sizeof(double));
   if (tcomp == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzunml2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          tcomp, NULL
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__, 
         "Number of my computations %f", 
         TCOMP(myrow, mycol, nprow, npcol)
    );

    myt = TCOMP(myrow, mycol, nprow, npcol);
    free(tcomp);
    return myt;
}

/*------------------------------------------------------------------*/

int hscal_pzunml2_my_tcomm(
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm
)
{
   int i, j, rc;

   int *tcomm_l = (int*)calloc(
                        nprow*npcol*nprow*npcol, 
                        sizeof(int)
   );

   if (tcomm_l == NULL)
   {
      return HSCAL_ERR_LOW_MEMORY;
   }

   rc = hscal_pzunml2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm_l
   );

   if (rc != HSCAL_SUCCESS)
   {
      return rc;
   }

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          tcomm[HSCALRIX(i, j, nprow, npcol)] 
          = 
          TCOMML(myrow, mycol, i, j, nprow, npcol);
      }
   }

   free(tcomm_l);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzunml2_my_tcomm_to_ij(
   int myrow,
   int mycol,
   int ireceiver,
   int jreceiver,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   int *tcomm,
   int *tcomm_i
)
{
   int i, j, rc;

   if (TCOMMI(myrow, mycol, nprow, npcol)
       != HSCAL_COMM_UNINITIALIZED)
   {
      return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
   }

   TCOMMI(myrow, mycol, nprow, npcol) = 1;

   for (i = 0; i < nprow; i++)
   {
      for (j = 0; j < npcol; j++)
      {
          TCOMM(myrow, mycol, i, j, nprow, npcol) = 0;
      }
   }

   rc = hscal_pzunml2_my_tcomp_tcomm
        (
          1,
          myrow, mycol,
          SIDEI, TRANSI,
          M, N, K,
          IA, JA, DESCA, 
          IC, JC, DESCC,
          nprow, npcol,
          NULL, tcomm
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HSCAL_DEBUG(
         __FILE__, __FUNCTION__,__LINE__,
         "(%d, %d) -> (%d, %d), (p, q)=(%d, %d) %d",
         myrow, mycol, ireceiver, jreceiver, nprow, npcol,
         TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol)
    );

    return TCOMM(myrow, mycol, ireceiver, jreceiver, nprow, npcol);
}

/*------------------------------------------------------------------*/

int hscal_pzunml2_my_tcomp_tcomm( 
   int init_llds_f,
   int myrow,
   int mycol,
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
    hscaldoublecomplex c_b1 = {1.,0.};
    hscalinteger c__5 = 5;
    hscalinteger c__3 = 3;
    hscalinteger c__9 = 9;
    hscalinteger c__4 = 4;
    hscalinteger c__14 = 14;
    hscalinteger c__0 = 0;
    hscalinteger c__1 = 1;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4;
    hscalinteger i__, i1, i2, i3;
    hscalinteger mi, ni, nq, icc, jcc;
    hscaldoublecomplex aii;
    hscalinteger lcm;
    hscalinteger mpc0, nqc0;
    hscallogical left;
    hscalinteger lcmp, iacol, iccol;
    hscalinteger icrow, lwmin, ictxt, icoffa, 
	    icoffc;
    hscalinteger iroffc;
    hscallogical notran;
    char colbtop[1], rowbtop[1];
    int     info_l, *info = &info_l;

    char sidec      = SIDEI, *side = &sidec;
    char transc     = TRANSI, *trans = &transc;

    int     *m      = &M;
    int     *n      = &N;
    int     *k      = &K;
    int     *ia     = &IA;
    int     *ja     = &JA;
    int     *desca  = (int*)DESCA;
    int     *ic     = &IC;
    int     *jc     = &JC;
    int     *descc  = (int*)DESCC;

    /* Parameter adjustments */
    --descc;
    --desca;

    if (init_llds_f == 1)
    {
       int *PXUNML2_DESCA_l = &desca[1];
       int *PXUNML2_DESCC_l = &descc[1];
       HSCAL_PXUNML2_INIT_LLDS(
            PXUNML2_DESCA_l,
            PXUNML2_DESCC_l,
            myrow, nprow
       );
    }

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pzunml2: (%d, %d): (desca+1)[M1_]=%d (desca+1)[MB1_]=%d myrow=%d (desca+1)[RSRC1_]=%d nprow=%d (desca+1)[LLD1_]=%d", 
        myrow, mycol, (desca+1)[M1_], (desca+1)[MB1_], myrow, (desca+1)[RSRC1_], nprow, (desca+1)[LLD1_]
    );

    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "pzunml2: (%d, %d): (descc+1)[M1_]=%d (descc+1)[MB1_]=%d myrow=%d (descc+1)[RSRC1_]=%d nprow=%d (descc+1)[LLD1_]=%d",
        myrow, mycol, (descc+1)[M1_], (descc+1)[MB1_], myrow, (descc+1)[RSRC1_], nprow, (descc+1)[LLD1_]
    );

/*     Test the input parameters */

    *info = 0;
    if (nprow == -1) {
	*info = -902;
    } else {
	left = hscal_lsame(side, "L");
	notran = hscal_lsame(trans, "N");

/*        NQ is the order of Q */

	if (left) {
	    nq = *m;
	    hscal_chk1mat(
               k, &c__5, m, &c__3, ia, ja, &desca[1], &c__9, info,
               myrow, mycol, nprow, npcol
          );
	} else {
	    nq = *n;
	    hscal_chk1mat(
               k, &c__5, n, &c__4, ia, ja, &desca[1], &c__9, info,
               myrow, mycol, nprow, npcol
          );
	}
	hscal_chk1mat(
           m, &c__3, n, &c__4, ic, jc, &descc[1], &c__14, info,
           myrow, mycol, nprow, npcol
      );
	if (*info == 0) {
	    icoffa = (*ja - 1) % desca[6];
	    iroffc = (*ic - 1) % descc[5];
	    icoffc = (*jc - 1) % descc[6];
	    iacol = indxg2p_(ja, &desca[6], &mycol, &desca[8], &npcol);
	    icrow = indxg2p_(ic, &descc[5], &myrow, &descc[7], &nprow);
	    iccol = indxg2p_(jc, &descc[6], &mycol, &descc[8], &npcol);
	    i__1 = *m + iroffc;
	    mpc0 = numroc_(&i__1, &descc[5], &myrow, &icrow, &nprow);
	    i__1 = *n + icoffc;
	    nqc0 = numroc_(&i__1, &descc[6], &mycol, &iccol, &npcol);

	    if (left) {
		lcm = ilcm_(&nprow, &npcol);
		lcmp = lcm / nprow;
/* Computing MAX */
		i__4 = *m + iroffc;
		i__3 = numroc_(&i__4, &desca[5], &c__0, &c__0, &nprow);
		i__1 = max(1,nqc0), i__2 = numroc_(&i__3, &desca[5], &c__0, &
			c__0, &lcmp);
		lwmin = mpc0 + max(i__1,i__2);
	    } else {
		i__1 = *n + icoffc;
		nqc0 = numroc_(&i__1, &descc[6], &mycol, &iccol, &npcol);
		i__1 = *m + iroffc;
		mpc0 = numroc_(&i__1, &descc[5], &myrow, &icrow, &nprow);
		lwmin = nqc0 + max(1,mpc0);
	    }

	    if (! left && ! hscal_lsame(side, "R")) {
		*info = -1;
	    } else if (! notran && ! hscal_lsame(trans, "C"))
		     {
		*info = -2;
	    } else if (*k < 0 || *k > nq) {
		*info = -5;
	    } else if (left && desca[6] != descc[5]) {
		*info = -906;
	    } else if (left && icoffa != iroffc) {
		*info = -12;
	    } else if (! left && icoffa != icoffc) {
		*info = -13;
	    } else if (! left && iacol != iccol) {
		*info = -13;
	    } else if (! left && desca[6] != descc[6]) {
		*info = -1406;
	    }
	}
    }

    if (*info != 0) {
	i__1 = -(*info);
       hscal_PB_Cabort( 
            "PZUNML2", 
            i__1,
            nprow, npcol, myrow, mycol
       ); 
       return HSCAL_SUCCESS;
    }

/*     Quick return if possible */

    if (*m == 0 || *n == 0 || *k == 0) {
       return HSCAL_SUCCESS;
    }

    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);

    if (left && notran || ! left && ! notran) {
	i1 = *ia;
	i2 = *ia + *k - 1;
	i3 = 1;
    } else {
	i1 = *ia + *k - 1;
	i2 = *ia;
	i3 = -1;
    }

    if (left) {
	ni = *n;
	jcc = *jc;
    } else {
	mi = *m;
	icc = *ic;
	pb_topset__(&ictxt, "Broadcast", "Rowwise", " ");
	if (notran) {
	    pb_topset__(&ictxt, "Broadcast", "Columnwise", "D-ring");
	} else {
	    pb_topset__(&ictxt, "Broadcast", "Columnwise", "I-ring");
	}
    }

    i__1 = i2;
    i__2 = i3;
    for (i__ = i1; i__2 < 0 ? i__ >= i__1 : i__ <= i__1; i__ += i__2) {
	if (left) {

/*           H(i) or H(i)' is applied to C(i:ic+m-1,jc:jc+n-1) */

	    mi = *m - i__ + *ia;
	    icc = *ic + i__ - *ia;
	} else {

/*           H(i) or H(i)' is applied to C(ic:ic+m-1,jc+i-ia:jc+n-1) */

	    ni = *n - i__ + *ia;
	    jcc = *jc + i__ - *ia;
	}

/*        Apply H(i) or H(i)' */

	if (i__ - *ia + 1 < nq) {
	    i__3 = nq - i__ + *ia - 1;
	    i__4 = *ja + i__ - *ia + 1;
	    hscal_pzlacgv(
               &i__3, &i__, &i__4, &desca[1], &desca[3],
               nprow, npcol,
               myrow, mycol,
               tcomp
          );
	}
	i__3 = *ja + i__ - *ia;
      {
	hscal_pzelset2(
           &aii, &i__, &i__3, &desca[1], &c_b1,
           nprow, npcol,
           myrow, mycol,
           tcomp, tcomm
      );
      }
	if (notran) {
	    i__3 = *ja + i__ - *ia;
	    hscal_pzlarfc_my_tcomp_tcomm(
               0,
               myrow, mycol,
               SIDEI, mi, ni, i__, i__3, &desca[1], desca[3],
		   icc, jcc, &descc[1],
               nprow, npcol,
               tcomp, tcomm
          );
	} else {
	    i__3 = *ja + i__ - *ia;
	    hscal_pzlarf_my_tcomp_tcomm(
               0,
               myrow, mycol,
               SIDEI, mi, ni, i__, i__3, &desca[1], desca[3], 
		   icc, jcc, &descc[1],
               nprow, npcol,
               tcomp, tcomm
          );
	}
	i__3 = *ja + i__ - *ia;
      {
	hscal_pzelset(
           &i__, &i__3, &desca[1], &aii,
           nprow, npcol,
           myrow, mycol,
           tcomp, tcomm
      );
      }
	if (i__ - *ia + 1 < nq) {
	    i__3 = nq - i__ + *ia - 1;
	    i__4 = *ja + i__ - *ia + 1;
	    hscal_pzlacgv(
               &i__3, &i__, &i__4, &desca[1], &desca[3],
               nprow, npcol,
               myrow, mycol,
               tcomp
          );
	}

/* L10: */
    }

    pb_topget__(&ictxt, "Broadcast", "Rowwise", rowbtop);
    pb_topget__(&ictxt, "Broadcast", "Columnwise", colbtop);

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzunml2_tcomp_tcomm(
   int SIDEI,
   int TRANSI,
   int M,
   int N,
   int K,
   int IA,
   int JA,
   int DESCA[DLEN_],
   int IC,
   int JC,
   int DESCC[DLEN_],
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
)
{
   int i, rc, myrow, mycol;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzunml2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                M, N, K,
                IA, JA, DESCA, 
                IC, JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzunml2_info(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;
   int TRANSI = *TRANS;

   for (i = 0; i < nprow*npcol; i++)
   {
      tcomp[i] = 0;
   }

   for (i = 0; i < nprow*npcol*nprow*npcol; i++)
   {
      tcomm[i] = 0;
   }

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzunml2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                *M, *N, *K,
                *IA, *JA, DESCA, 
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pzunml2_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
)
{
   int i, rc, myrow, mycol;

   int SIDEI  = *SIDE;
   int TRANSI = *TRANS;

   for (myrow = 0; myrow < nprow; myrow++)
   {
      for (mycol = 0; mycol < npcol; mycol++)
      {
         rc = hscal_pzunml2_my_tcomp_tcomm
              (
                1,
                myrow, mycol,
                SIDEI, TRANSI,
                *M, *N, *K,
                *IA, *JA, DESCA, 
                *IC, *JC, DESCC,
                nprow, npcol,
                tcomp, tcomm
         );

         if (rc != HSCAL_SUCCESS)
         {
            return rc;
         }
      }
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/
