
#include <math.h>

#include "../include/hscalapack.h"

/*---------------------------------------------------------------------------*/

double hscal_d_imag(hscaldoublecomplex *z)
{
   return(z->i);
}

/*---------------------------------------------------------------------------*/

double hscal_r_imag(hscalcomplex *z)
{
   return(z->i);
}

/*---------------------------------------------------------------------------*/

hscaldoublereal hscal_dcabs1(hscaldoublecomplex *z__)
{
    hscaldoublereal ret_val;
    hscaldoublecomplex equiv_0[1];
    /* Local variables */
#define t ((hscaldoublereal *)equiv_0)
#define zz (equiv_0)
    zz->r = z__->r, zz->i = z__->i;
    ret_val = abs(t[0]) + abs(t[1]);
    return ret_val;
}
#undef zz
#undef t

/*---------------------------------------------------------------------------*/

hscallogical hscal_lsame(char *ca, char *cb)
{
/*  -- LAPACK auxiliary routine (version 3.0) --   
       Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,   
       Courant Institute, Argonne National Lab, and Rice University   
       September 30, 1994   


    Purpose   
    =======   

    LSAME returns .TRUE. if CA is the same letter as CB regardless of   
    case.   

    Arguments   
    =========   

    CA      (input) CHARACTER*1   
    CB      (input) CHARACTER*1   
            CA and CB specify the single characters to be compared.   

   ===================================================================== 
  


       Test if the characters are equal */
    /* System generated locals */
    hscallogical ret_val;
    hscalinteger inta, intb, zcode;


    ret_val = *(unsigned char *)ca == *(unsigned char *)cb;
    if (ret_val) {
	return ret_val;
    }

/*     Now test for equivalence if both characters are alphabetic. */

    zcode = 'Z';

/*     Use 'Z' rather than 'A' so that ASCII can be detected on Prime   
       machines, on which ICHAR returns a value with bit 8 set.   
       ICHAR('A') on Prime machines returns 193 which is the same as   
       ICHAR('A') on an EBCDIC machine. */

    inta = *(unsigned char *)ca;
    intb = *(unsigned char *)cb;

    if (zcode == 90 || zcode == 122) {

/*        ASCII is assumed - ZCODE is the ASCII code of either lower o
r   
          upper case 'Z'. */

	if (inta >= 97 && inta <= 122) {
	    inta += -32;
	}
	if (intb >= 97 && intb <= 122) {
	    intb += -32;
	}

    } else if (zcode == 233 || zcode == 169) {

/*        EBCDIC is assumed - ZCODE is the EBCDIC code of either lower
 or   
          upper case 'Z'. */

	if (inta >= 129 && inta <= 137 || inta >= 145 && inta <= 153 || inta 
		>= 162 && inta <= 169) {
	    inta += 64;
	}
	if (intb >= 129 && intb <= 137 || intb >= 145 && intb <= 153 || intb 
		>= 162 && intb <= 169) {
	    intb += 64;
	}

    } else if (zcode == 218 || zcode == 250) {

/*        ASCII is assumed, on Prime machines - ZCODE is the ASCII cod
e   
          plus 128 of either lower or upper case 'Z'. */

	if (inta >= 225 && inta <= 250) {
	    inta += -32;
	}
	if (intb >= 225 && intb <= 250) {
	    intb += -32;
	}
    }
    ret_val = inta == intb;

/*     RETURN   

       End of LSAME */

    return ret_val;
} /* lsame_ */


/*---------------------------------------------------------------------------*/

int hscal_xerbla(char *srname, hscalinteger *info)
{
    HSCAL_DEBUG(
        __FILE__, __FUNCTION__,__LINE__, 
        "** On entry to %6s, parameter number %2i had an illegal value",
        srname, *info);

    return 0;
}

/*---------------------------------------------------------------------------*/

void hscal_d_cnjg(hscaldoublecomplex *r, hscaldoublecomplex *z)
{
        hscaldoublereal zi = z->i;
        r->r = z->r;
        r->i = -zi;
}

/*---------------------------------------------------------------------------*/

void hscal_r_cnjg(hscalcomplex *r, hscalcomplex *z)
{
        hscalreal zi = z->i;
        r->r = z->r;
        r->i = -zi;
}

/*---------------------------------------------------------------------------*/

double hscal_c_abs(hscalcomplex *z)
{
   return( hscal_f__cabs( z->r, z->i ) );
}

/*---------------------------------------------------------------------------*/

double hscal_d_abs(hscaldoublereal *x)
{
   if (*x >= 0)
      return(*x);
   return(- *x);
}

/*---------------------------------------------------------------------------*/

double hscal_r_abs(hscalreal *x)
{
   if (*x >= 0)
      return(*x);
   return(- *x);
}

/*---------------------------------------------------------------------------*/

double hscal_z_abs(hscaldoublecomplex *z)
{
   return( hscal_f__cabs( z->r, z->i ) );
}

/*---------------------------------------------------------------------------*/

double hscal_d_sign(hscaldoublereal *a, hscaldoublereal *b)
{
   double x;
   x = (*a >= 0 ? *a : - *a);
   return( *b >= 0 ? x : -x);
}

/*---------------------------------------------------------------------------*/

double hscal_r_sign(hscalreal *a, hscalreal *b)
{
   double x;
   x = (*a >= 0 ? *a : - *a);
   return( *b >= 0 ? x : -x);
}

/*---------------------------------------------------------------------------*/

void hscal_c_div(hscalcomplex *c, hscalcomplex *a, hscalcomplex *b)
{
        double ratio, den;
        double abr, abi, cr;

        if( (abr = b->r) < 0.)
                abr = - abr;
        if( (abi = b->i) < 0.)
                abi = - abi;
        if( abr <= abi )
                {
                if(abi == 0) {
                        float af, bf;
                        af = bf = abr;
                        if (a->i != 0 || a->r != 0)
                                af = 1.;
                        c->i = c->r = af / bf;
                        return;
                        }
                ratio = (double)b->r / b->i ;
                den = b->i * (1 + ratio*ratio);
                cr = (a->r*ratio + a->i) / den;
                c->i = (a->i*ratio - a->r) / den;
                }

        else
                {
                ratio = (double)b->i / b->r ;
                den = b->r * (1 + ratio*ratio);
                cr = (a->r + a->i*ratio) / den;
                c->i = (a->i - a->r*ratio) / den;
                }
        c->r = cr;
}

/*---------------------------------------------------------------------------*/

void hscal_z_div(hscaldoublecomplex *c, hscaldoublecomplex *a, hscaldoublecomplex *b)
{
        double ratio, den;
        double abr, abi, cr;

        if( (abr = b->r) < 0.)
                abr = - abr;
        if( (abi = b->i) < 0.)
                abi = - abi;
        if( abr <= abi )
                {
                if(abi == 0) {
                        if (a->i != 0 || a->r != 0)
                                abi = 1.;
                        c->i = c->r = abi / abr;
                        return;
                        }
                ratio = b->r / b->i ;
                den = b->i * (1 + ratio*ratio);
                cr = (a->r*ratio + a->i) / den;
                c->i = (a->i*ratio - a->r) / den;
                }

        else
                {
                ratio = b->i / b->r ;
                den = b->r * (1 + ratio*ratio);
                cr = (a->r + a->i*ratio) / den;
                c->i = (a->i - a->r*ratio) / den;
                }
        c->r = cr;
}

/*---------------------------------------------------------------------------*/

hscalinteger hscal_iceil(hscalinteger *inum, hscalinteger *idenom)
{
    /* System generated locals */
    hscalinteger ret_val;


/*  -- ScaLAPACK tools routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Scalar Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  ICEIL returns the ceiling of the division of two integers. */

/*  Arguments */
/*  ========= */

/*  INUM     (local input) INTEGER */
/*           The numerator, */

/*  IDENOM   (local input) INTEGER */
/*           and the denominator of the fraction to be evaluated. */

/*  ===================================================================== */

/*     .. Executable Statements .. */

    ret_val = (*inum + *idenom - 1) / *idenom;

    return ret_val;

/*     End of ICEIL */

}

/*---------------------------------------------------------------------------*/

double hscal_f__cabs(double real, double imag)
{
double temp;

if(real < 0)
        real = -real;
if(imag < 0)
        imag = -imag;
if(imag > real){
        temp = real;
        real = imag;
        imag = temp;
}
if((real+imag) == real)
        return(real);

temp = imag/real;
temp = real*sqrt(1.0 + temp*temp);  /*overflow!!*/
return(temp);
}

/*---------------------------------------------------------------------------*/

