
    #include <math.h>
    #include <stdio.h>
    #include <string.h>
    #include <stdlib.h>
    #include <sys/time.h>

    #include <hscalapack.h>

    #include "hscal_mxm_i.c"
    #include "hscal_mxm_benchmark_dgemm.c"
    #include "hscal_execute_scalapack_outer_product.c"

    /*
    * Performance model of the PDGEMM routine in 
    * mpC's performance model definition language
    */
    #include "pdgemm.c"

    /*----------------------------------------------------------------------------*/

    int hscal_timeof_using_block_size(
            const int num_of_computers,
            const int gme,
            const int myi,
            const int myj,
            const int n,
            const int p, 
            const int q,
            const int optb,
            double *timeof
    )
    {
        int    i, j, k, rc;
        int    *w, *h, *trow, *tcol, colw;
        double *gspeeds = NULL;
        double  mintime = DBL_MAX;

        int reconm           = n/p;
        int reconn           = n/q;
        int reconk           = optb;

        hscal_set_debug(HSCAL_LOG_VERBOSE);

        {
           /*
           * The initial benchmark
           */
           int benchmark_m = reconm;
           int benchmark_n = reconn;
           int kb          = reconk;
           int MSIZE       = n/kb;

           { 
              int output_p;
              int input_p[5] = 
                  {
                   myi,
                   myj,
                   benchmark_m, 
                   benchmark_n, 
                   kb
              };

              rc = HMPI_Recon_t(
                       &Benchmark_function,
                       input_p,
                       5,
                       &output_p
              ); 
   
              if  (rc != HMPI_OK)
              {
                printf("Panic: HMPI_Recon failed \n");
              }
           }
          
           gspeeds = (double*)malloc(
                              sizeof(double)
                              *
                              num_of_computers
           );

           if (gspeeds == NULL)
           {
              printf("(%d): Cannot allocate gspeeds\n", gme);
           }

           rc = HMPI_Group_performances(
                    HMPI_COMM_WORLD_GROUP,
                    gspeeds
           );
    
           if (rc != HMPI_OK)
           {
              printf(
                 "MAIN: Problems getting the performances"
                 "...Exiting...\n"
              );
           }

           if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
           {
              w = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q*p*q)
              );
   
              if (w == NULL)
              {
                 printf("(%d): Cannot allocate w\n", gme);
              }
   
              h = (int*)malloc(
                        sizeof(int)
                        *
                        (p*q*p*q)
              );

              if (h == NULL)
              {
                 printf("(%d): Cannot allocate h\n", gme);
              }

              trow = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (trow == NULL)
              {
                 printf("(%d): Cannot allocate trow\n", gme);
              }

              tcol = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (tcol == NULL)
              {
                 printf("(%d): Cannot allocate tcol\n", gme);
              }

              rc = HMPI_Partition_matrix_2d(
                    p,
                    q,
                    1,
                    gspeeds,
                    NULL,
                    NULL,
                    MSIZE,
                    MSIZE,
                    HMPI_COLUMN_BASED,
                    w,
                    h,
                    trow,
                    tcol,
                    NULL,
                    NULL
              );
   
              if (rc != HMPI_OK)
              {
                 printf("(%d): Problems partitioning\n", gme);
              }

              int *model_params;
              int model_count;

              model_count = 1+1+1+1+(p*q*p*q)+(p*q*p*q);
              model_params = (int*)malloc(
                                   sizeof(int)
                                   *
                                   model_count
              );
    
              if (model_params == NULL)
              {
                 printf("No memory to allocate performanc model parameters\n");
              }          

              model_params[0]                      = p; 
              model_params[1]                      = q; 
              model_params[1+1]                    = n; 
              model_params[1+1+1]                  = kb; 

              for (i = 0; i < (p*q*p*q); i++) 
              {         
                 model_params[1+1+1+1+i]           = w[i];
                 model_params[1+1+1+1+(p*q*p*q)+i] = h[i];
              } 

              *timeof = HMPI_Timeof(
                            &MPC_NetType_hetpdgemm,
                            model_params,
                            model_count
              );

              free(model_params);
              free(w);
              free(h);
              free(trow);
              free(tcol);

              if (hscal_debug_flag & HSCAL_LOG_VERBOSE)
              {
                 printf(
                    "HETEROGENEOUS SCALAPACK TIMEOF: (n):%d (p,q):%d,%d block size:%d time(sec):%f\n", 
                    n,
                    p, q,
                    kb,
                    *timeof
                 );
              }
           }

           free(gspeeds);
       }

       /*
       * Broadcast the block size to all the processes
       */
       rc = MPI_Bcast(
               timeof,
               1,
               MPI_DOUBLE,
               HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
               HMPI_COMM_WORLD
       );

       if (rc != MPI_SUCCESS)
       {
          printf("TIMEOF using block size: Problems broadcasting timeof...\n");
       }

       return HSCAL_SUCCESS;
    }

    /*----------------------------------------------------------------------------*/

    int hscal_determine_optimal_block_size_using_timeof(
             int num_of_computers,
             int gme,
             int myi,
             int myj,
             int n,
             int p, 
             int q,
             int *optb
    )
    {
        int    i, j, k, rc;
        int    *w, *h, *trow, *tcol, colw;
        double *gspeeds = NULL;
        double  mintime = DBL_MAX;

        int reconm           = n/p;
        int reconn           = n/q;
        int min_m_n          = (reconm < reconn) ? reconm : reconn;
        int reconk;

        hscal_set_debug(HSCAL_LOG_VERBOSE);

        /*
        * We know the block size of 1 gives poor performance
        */
        for (reconk = 2; reconk <= min_m_n; reconk++)
        {
           /*
           * A simplication here; (n/p, n/q) must be a multiple of kb
           */
           if (((reconm%reconk) != 0) || ((reconn%reconk) != 0))
           {
              continue;
           }

           /*
           * The initial benchmark
           */
           int benchmark_m = reconm;
           int benchmark_n = reconn;
           int kb          = reconk;
           int MSIZE       = n/kb;

           { 
              int output_p;
              int input_p[5] = 
                  {
                   myi,
                   myj,
                   benchmark_m, 
                   benchmark_n, 
                   kb
              };

              rc = HMPI_Recon_t(
                       &Benchmark_function,
                       input_p,
                       5,
                       &output_p
              ); 
   
              if  (rc != HMPI_OK)
              {
                printf("Panic: HMPI_Recon failed \n");
              }
           }
          
           gspeeds = (double*)malloc(
                              sizeof(double)
                              *
                              num_of_computers
           );

           if (gspeeds == NULL)
           {
              printf("(%d): Cannot allocate gspeeds\n", gme);
           }

           rc = HMPI_Group_performances(
                    HMPI_COMM_WORLD_GROUP,
                    gspeeds
           );
    
           if (rc != HMPI_OK)
           {
              printf(
                 "MAIN: Problems getting the performances"
                 "...Exiting...\n"
              );
           }

           if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
           {
              w = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q*p*q)
              );
   
              if (w == NULL)
              {
                 printf("(%d): Cannot allocate w\n", gme);
              }
   
              h = (int*)malloc(
                        sizeof(int)
                        *
                        (p*q*p*q)
              );

              if (h == NULL)
              {
                 printf("(%d): Cannot allocate h\n", gme);
              }

              trow = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (trow == NULL)
              {
                 printf("(%d): Cannot allocate trow\n", gme);
              }

              tcol = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (tcol == NULL)
              {
                 printf("(%d): Cannot allocate tcol\n", gme);
              }

              rc = HMPI_Partition_matrix_2d(
                    p,
                    q,
                    1,
                    gspeeds,
                    NULL,
                    NULL,
                    MSIZE,
                    MSIZE,
                    HMPI_COLUMN_BASED,
                    w,
                    h,
                    trow,
                    tcol,
                    NULL,
                    NULL
              );
   
              if (rc != HMPI_OK)
              {
                 printf("(%d): Problems partitioning\n", gme);
              }

              int *model_params;
              int model_count;

              model_count = 1+1+1+1+(p*q*p*q)+(p*q*p*q);
              model_params = (int*)malloc(
                                   sizeof(int)
                                   *
                                   model_count
              );
    
              if (model_params == NULL)
              {
                 printf("No memory to allocate performanc model parameters\n");
              }          

              model_params[0]                      = p; 
              model_params[1]                      = q; 
              model_params[1+1]                    = n; 
              model_params[1+1+1]                  = kb; 

              for (i = 0; i < (p*q*p*q); i++) 
              {         
                 model_params[1+1+1+1+i]           = w[i];
                 model_params[1+1+1+1+(p*q*p*q)+i] = h[i];
              } 

              double timeof = HMPI_Timeof(
                                  &MPC_NetType_hetpdgemm,
                                  model_params,
                                  model_count
              );

              free(model_params);
              free(w);
              free(h);
              free(trow);
              free(tcol);

              if (hscal_debug_flag & HSCAL_LOG_VERBOSE)
              {
                 printf(
                    "HETEROGENEOUS SCALAPACK TIMEOF: (n):%d (p,q):%d,%d block size:%d time(sec):%f\n", 
                    n,
                    p, q,
                    kb,
                    timeof
                 );
              }

              if (timeof <= mintime)
              {
                 mintime = timeof;
                 *optb = kb;
              }
           }

           free(gspeeds);
       }

       /*
       * Broadcast the block size to all the processes
       */
       rc = MPI_Bcast(
               optb,
               1,
               MPI_INT,
               HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
               HMPI_COMM_WORLD
       );

       if (rc != MPI_SUCCESS)
       {
          printf("Block size determination: Problems broadcasting optimal block size...\n");
       }

       return HSCAL_SUCCESS;
    }

    /*----------------------------------------------------------------------------*/

    int hscal_determine_best_block_size_executing_algo(
        const int numb,
        const int *ba,
        const int bpsize,
        const int me,
        const int myi,
        const int myj,
        const int p,
        const int q,
        int *bestb
    )
    {
        int i, rc;
        double *times;

        if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
        {
           times   = (double*)malloc(
                              sizeof(double)
                              *
                              (p*q)
           );

           if (times == NULL)
           {
              printf("(%d): Cannot allocate times\n", me);
           }
        }

        for (i = 0; i < numb; i++)
        {
           int *w, *h, *trow, *tcol;
           int KB =  ba[i];
           int MSIZE = bpsize/KB;

           /*
           * Refresh the speeds
           */
           int benchmark_m = bpsize/p;
           int benchmark_n = bpsize/q;

           { 
              int output_p;
              int input_p[5] = 
                  {
                   myi,
                   myj,
                   benchmark_m, 
                   benchmark_n, 
                   KB
              };

              rc = HMPI_Recon_t(
                       &Benchmark_function,
                       input_p,
                       5,
                       &output_p
              ); 
   
              if  (rc != HMPI_OK)
              {
                printf("Panic: HMPI_Recon failed \n");
              }
           }
          
           double start = MPI_Wtime();

           w = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q*p*q)
           );

           if (w == NULL)
           {
              printf("(%d): Cannot allocate w\n", me);
           }

           h = (int*)malloc(
                     sizeof(int)
                     *
                     (p*q*p*q)
           );

           if (h == NULL)
           {
              printf("(%d): Cannot allocate h\n", me);
           }

           trow = (int*)malloc(
                        sizeof(int)
                        *
                        (p*q)
           );

           if (trow == NULL)
           {
              printf("(%d): Cannot allocate trow\n", me);
           }

           tcol = (int*)malloc(
                        sizeof(int)
                        *
                        (p*q)
           );

           if (tcol == NULL)
           {
              printf("(%d): Cannot allocate tcol\n", me);
           }

           double *gspeeds = (double*)malloc(
                                      sizeof(double)
                                      *
                                      (p*q)
           );

           if (gspeeds == NULL)
           {
              printf("(%d): Cannot allocate gspeeds\n", me);
           }

           rc = HMPI_Group_performances(
                    HMPI_COMM_WORLD_GROUP,
                    gspeeds
           );

           if (rc != HMPI_OK)
           {
              printf(
                 "Best even block size: Problems getting the performances...\n"
              );
           }

           if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
           {
              rc = HMPI_Partition_matrix_2d(
                       p, q, 1,
                       gspeeds, NULL, NULL,
                       MSIZE, MSIZE,
                       HMPI_COLUMN_BASED,
                       w, h, trow, tcol,
                       NULL, NULL
              );

              if (rc != HMPI_OK)
              {
                 printf("(%d): Problems partitioning\n", me);
              }

              Print_rectangle_2d_local_stdin(
                   p, q, MSIZE, MSIZE, 
                   w, h
              );
           }

           /*
           * Broadcast the matrix distribution
           * w, h, trow, tcol
           */     
           rc = MPI_Bcast(
                   w,
                   p*q*p*q,
                   MPI_INT,
                   HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
                   HMPI_COMM_WORLD
           );

           if (rc != MPI_SUCCESS)
           {
              printf("(%d):Problems broadcasting w\n", me);
           }

           rc = MPI_Bcast(
                   h,
                   p*q*p*q,
                   MPI_INT,
                   HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
                   HMPI_COMM_WORLD
           );

           if (rc != MPI_SUCCESS)
           {
              printf("(%d):Problems broadcasting h\n", me);
           }

           if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
           {
              printf("HETEROGENEOUS SCALAPACK ALGO: Executing n:%d b:%d\n",
                     bpsize,
                     KB
              );
           }

           Execute_scalapack_outer_product_algo(
                  me, myi, myj,
                  MSIZE, KB,
                  p, q, 
                  w, h
           );

           free(w);
           free(h);
           free(trow);
           free(tcol);
           free(gspeeds);

           double end = MPI_Wtime();

           if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
           {
              times[i] = (end - start);
              printf("HETEROGENEOUS SCALAPACK ALGO: Executed n:%d b:%d, time(sec):%f\n",
                     bpsize,
                     KB,
                     times[i]
              );
           }
        }

        if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
        {
           double t = times[0];
           *bestb = ba[0];
           for (i = 1; i < numb; i++)
           {
              if (times[i] <= t)
              {
                 t = times[i];
                 *bestb = ba[i];
              }
           }

           free(times);
        }

       /*
       * Broadcast the block size to all the processes
       */
       rc = MPI_Bcast(
               bestb,
               1,
               MPI_INT,
               HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
               HMPI_COMM_WORLD
       );

       if (rc != MPI_SUCCESS)
       {
          printf("Best block size determination: Problems broadcasting optimal block size...\n");
       }

        return HSCAL_SUCCESS;
    }

    /*----------------------------------------------------------------------------*/

    int main(int argc, char **argv)
    {
        int i, j, rc, optimal_even_block_size, optimal_odd_block_size, b, loopi;

        rc = hscal_init(&argc, &argv);

        if (rc != HSCAL_SUCCESS)
        {
           printf("MAIN: Problems initializing Heterogeneous ScaLAPACK runtime..Exiting...\n");
        }

        int myi, myj;
        int gme              = HMPI_Group_rank(HMPI_COMM_WORLD_GROUP);
        int num_of_computers = HMPI_Get_number_of_computers();
        int p                = sqrt(num_of_computers);
        int q                = num_of_computers/p;

        /*
        * My global coordinates
        */
        {
           int COORDN = 2;
           const int ppar[] = {p, q};
           int* mycoords = (int*)malloc(
                                 sizeof(int)
                                 *
                                 COORDN
           );

           if (mycoords == NULL)
           {
              printf("%d:Cannot allocate mycoords\n", gme);
           }

           rank2coord(gme, ppar, mycoords);

           myi = mycoords[0];
           myj = mycoords[1];

           free(mycoords);
        }

        /*
        * A sample of problem sizes tested
        */
        int lcm_p_q = hscal_ilcm(&p, &q);
        int hscal_initial_even_n = 128*lcm_p_q;
        int hscal_final_even_n = hscal_initial_even_n*6;
        int hscal_even_n_increment = hscal_initial_even_n*2;
     
        int num_of_even_blocksizes = 0;
        int *even_blocksizes = (int*)malloc(sizeof(int));

        if (even_blocksizes == NULL)
        {
           printf("MAIN: Problems allocating even_blocksizes...\n");
        }

        /*
        * The best block size for each problem size is determined
        * using HMPI_Timeof
        */
        for (loopi = 0, i = hscal_initial_even_n; 
             i <= hscal_final_even_n; 
             i+=hscal_even_n_increment, loopi++
        )
        {
           rc = hscal_determine_optimal_block_size_using_timeof(
                     num_of_computers,
                     gme,
                     myi,
                     myj,
                     i,
                     p,
                     q,
                     &b
           );

           if (rc != HSCAL_SUCCESS)
           {
              printf("MAIN: Problems determining optimal blocking size...\n");
           }

           if (num_of_even_blocksizes == 0)
           {
              even_blocksizes[0] = b;
              num_of_even_blocksizes++;
           }
           else
           {
              for (j = 0; j < loopi; j++)
              {
                 if (b == even_blocksizes[j])
                 {
                    break;
                 }
              }

              if (j == loopi)
              {
                 num_of_even_blocksizes++;
                 even_blocksizes = (int*)realloc(even_blocksizes, 
                                                 sizeof(int)*num_of_even_blocksizes
                 );

                 if (even_blocksizes == NULL)
                 {
                    printf("MAIN: Problems reallocating even_blocksizes...\n");
                 }

                 even_blocksizes[num_of_even_blocksizes-1] = b;
              }
           }
        }

        if ((hscal_debug_flag & HSCAL_LOG_VERBOSE) && (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP)))
        {
            printf("HETEROGENEOUS SCALAPACK: Even block sizes are ");
            for (i = 0; i < num_of_even_blocksizes; i++)
            {
               printf("%d ", even_blocksizes[i]);
            }
            printf("\n");
        }

        /*
        * Find the LCM of the blocking sizes
        */
        int even_blocksizes_lcm_psize;

        if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
        {
           int even_blocksizes_lcm = even_blocksizes[0];
           for (i = 1; i < num_of_even_blocksizes; i++)
           {
              even_blocksizes_lcm = hscal_ilcm(
                                         &even_blocksizes_lcm,
                                         &even_blocksizes[i]
              );
           }

           even_blocksizes_lcm_psize = even_blocksizes_lcm*lcm_p_q*2;

           if ((hscal_debug_flag & HSCAL_LOG_VERBOSE))
           {
              printf("HETEROGENEOUS SCALAPACK: Even block sizes(lcm, psize):%d,%d\n", 
                     even_blocksizes_lcm, even_blocksizes_lcm_psize);
           }
        }

        /*
        * Broadcast the block size psize to execute to all the processes
        */
        rc = MPI_Bcast(
                &even_blocksizes_lcm_psize,
                1,
                MPI_INT,
                HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
                HMPI_COMM_WORLD
        );
 
        if (rc != MPI_SUCCESS)
        {
           printf("MAIN: Problems broadcasting problem size...\n");
        }
 
        /*
        * Execute the MXM algorithm to find the best block size 
        * of all the block sizes
        */
        int best_even_b;
        rc = hscal_determine_best_block_size_executing_algo(
             num_of_even_blocksizes,
             even_blocksizes,
             even_blocksizes_lcm_psize,
             gme, myi, myj,
             p, q,
             &best_even_b
        );

        if (rc != HSCAL_SUCCESS)
        {
           printf("MAIN: Problems determining best even block size...\n");
        }

        if ((hscal_debug_flag & HSCAL_LOG_VERBOSE) && (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP)))
        {
           printf("HETEROGENEOUS SCALAPACK: Best even block size is %d\n", best_even_b);
        }

        free(even_blocksizes);

        /*
        * A sample of problem sizes tested
        */
        int hscal_initial_odd_n = 99*lcm_p_q;
        int hscal_final_odd_n = hscal_initial_odd_n*6;
        int hscal_odd_n_increment = hscal_initial_odd_n*2;

        int num_of_odd_blocksizes = 0;
        int *odd_blocksizes = (int*)malloc(sizeof(int));

        if (odd_blocksizes == NULL)
        {
           printf("MAIN: Problems allocating odd_blocksizes...\n");
        }

        /*
        * The best block size for each problem size is determined
        * using HMPI_Timeof
        */
        for (loopi = 0, i = hscal_initial_odd_n; 
             i <= hscal_final_odd_n; 
             i+=hscal_odd_n_increment, loopi++)
        {
           rc = hscal_determine_optimal_block_size_using_timeof(
                     num_of_computers,
                     gme,
                     myi,
                     myj,
                     i,
                     p,
                     q,
                     &b
           );

           if (rc != HSCAL_SUCCESS)
           {
              printf("MAIN: Problems determining optimal blocking size...\n");
           }

           if (num_of_odd_blocksizes == 0)
           {
              odd_blocksizes[0] = b;
              num_of_odd_blocksizes++;
           }
           else
           {
              for (j = 0; j < loopi; j++)
              {
                 if (b == odd_blocksizes[j])
                 {
                    break;
                 }
              }

              if (j == loopi)
              {
                 num_of_odd_blocksizes++;
                 odd_blocksizes = (int*)realloc(odd_blocksizes, 
                                                 sizeof(int)*num_of_odd_blocksizes
                 );

                 if (odd_blocksizes == NULL)
                 {
                    printf("MAIN: Problems reallocating odd_blocksizes...\n");
                 }

                 odd_blocksizes[num_of_odd_blocksizes-1] = b;
              }
           }
        }

        if ((hscal_debug_flag & HSCAL_LOG_VERBOSE) && (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP)))
        {
            printf("HETEROGENEOUS SCALAPACK: odd block sizes are ");
            for (i = 0; i < num_of_odd_blocksizes; i++)
            {
               printf("%d ", odd_blocksizes[i]);
            }
            printf("\n");
        }

        /*
        * Find the LCM of the blocking sizes
        */
        int odd_blocksizes_lcm_psize;

        if (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP))
        {
           int odd_blocksizes_lcm = odd_blocksizes[0];
           for (i = 1; i < num_of_odd_blocksizes; i++)
           {
              odd_blocksizes_lcm = hscal_ilcm(
                                         &odd_blocksizes_lcm,
                                         &odd_blocksizes[i]
              );
           }

           odd_blocksizes_lcm_psize = odd_blocksizes_lcm*lcm_p_q*2;

           if ((hscal_debug_flag & HSCAL_LOG_VERBOSE))
           {
              printf("HETEROGENEOUS SCALAPACK: odd block sizes(lcm, psize):%d,%d\n", 
                     odd_blocksizes_lcm, odd_blocksizes_lcm_psize);
           }
        }

        /*
        * Broadcast the block size psize to execute to all the processes
        */
        rc = MPI_Bcast(
                &odd_blocksizes_lcm_psize,
                1,
                MPI_INT,
                HMPI_Group_parent(HMPI_COMM_WORLD_GROUP),
                HMPI_COMM_WORLD 
        );

        if (rc != MPI_SUCCESS)
        {
           printf("MAIN: Problems broadcasting problem size...\n");
        }

        /*
        * Execute the MXM algorithm to find the best block size 
        * of all the block sizes
        */
        int best_odd_b;
        rc = hscal_determine_best_block_size_executing_algo(
             num_of_odd_blocksizes,
             odd_blocksizes,
             odd_blocksizes_lcm_psize,
             gme, myi, myj,
             p, q,
             &best_odd_b
        );

        if (rc != HSCAL_SUCCESS)
        {
           printf("MAIN: Problems determining best odd block size...\n");
        }

        if ((hscal_debug_flag & HSCAL_LOG_VERBOSE) && (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP)))
        {
           printf("HETEROGENEOUS SCALAPACK: Best odd block size is %d\n", best_odd_b);
        }

        free(odd_blocksizes);

        int even_odd_b_lcm = hscal_ilcm(
                                  &best_even_b,
                                  &best_odd_b
        );

        if ((hscal_debug_flag & HSCAL_LOG_VERBOSE) && (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP)))
        {
           printf("HETEROGENEOUS SCALAPACK: LCM of best even odd block sizes is %d\n", 
                  even_odd_b_lcm);
        }

        /*
        * The LCM may be quite large
        */

        /*
        * Use HMPI_Timeof now to decide the best block size
        */
        double best_even_b_timeof, best_odd_b_timeof;

        rc = hscal_timeof_using_block_size(
                  num_of_computers,
                  gme,
                  myi,
                  myj,
                  even_odd_b_lcm,
                  p,
                  q,
                  best_even_b,
                  &best_even_b_timeof
        );

        if (rc != HSCAL_SUCCESS)
        {
           printf("MAIN: Problems determining best even b timeof...\n");
        }

        rc = hscal_timeof_using_block_size(
                  num_of_computers,
                  gme,
                  myi,
                  myj,
                  even_odd_b_lcm,
                  p,
                  q,
                  best_odd_b,
                  &best_odd_b_timeof
        );

        if (rc != HSCAL_SUCCESS)
        {
           printf("MAIN: Problems determining best odd b timeof...\n");
        }

        if ((hscal_debug_flag & HSCAL_LOG_VERBOSE) && (HMPI_Is_parent(HMPI_COMM_WORLD_GROUP)))
        {
           if (best_even_b_timeof > best_odd_b_timeof)
           {
              printf("===============================================================\n");
              printf("=    HETEROGENEOUS SCALAPACK: The best block size is %d   =\n",
                     best_even_b);
              printf("===============================================================\n");
           }
           else
           {
              printf("===============================================================\n");
              printf("=    HETEROGENEOUS SCALAPACK: The best block size is %d   =\n",
                     best_odd_b);
              printf("===============================================================\n");
           }
        }        

        hscal_finalize(0);
    }

    /*----------------------------------------------------------------------------*/

