
/*
* ScaLAPACK TOOLS
*
* Ravi Reddy
* Version 1.0
* School of Computer Science and Informatics,
* UCD,
* Dublin,
* November 18, 2007
*/

/*
*  Include files
*/
#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_chk1mat(
         hscalinteger *ma, 
         hscalinteger *mapos0, 
         hscalinteger *na, 
	 hscalinteger *napos0, 
         hscalinteger *ia, 
         hscalinteger *ja, 
         hscalinteger *desca, 
         hscalinteger *descapos0, 
         hscalinteger *info,
         hscalinteger myrow,
         hscalinteger mycol,
         hscalinteger nprow,
         hscalinteger npcol
)
{
    /* System generated locals */
    hscalinteger i__1, i__2;

    /* Local variables */
    hscalinteger descapos, iapos, japos, mapos, napos;

/*  -- ScaLAPACK tools routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  CHK1MAT checks that the values associated with one distributed matrix */
/*  make sense from a local viewpoint */

/*  Arguments */
/*  ========= */

/*  MA      (global input) INTEGER */
/*          The number or matrix rows of A being operated on. */

/*  MAPOS0  (global input) INTEGER */
/*          Where in the calling routine's parameter list MA appears. */

/*  NA      (global input) INTEGER */
/*          The number of matrix columns of A being operated on. */

/*  NAPOS0  (global input) INTEGER */
/*          Where in the calling routine's parameter list NA appears. */

/*  IA      (global input) INTEGER */
/*          The row index in the global array A indicating the first */
/*          row of sub( A ). */

/*  JA      (global input) INTEGER */
/*          The column index in the global array A indicating the */
/*          first column of sub( A ). */

/*  DESCA   (global and local input) INTEGER array of dimension DLEN_. */
/*          The array descriptor for the distributed matrix A. */

/*  DESCAPOS0 (global input) INTEGER */
/*          Where in the calling routine's parameter list DESCA */
/*          appears.  Note that we assume IA and JA are respectively 2 */
/*          and 1 entries behind DESCA. */

/*  INFO    (local input/local output) INTEGER */
/*          = 0:  successful exit */
/*          < 0:  If the i-th argument is an array and the j-entry had */
/*                an illegal value, then INFO = -(i*100+j), if the i-th */
/*                argument is a scalar and had an illegal value, then */
/*                INFO = -i. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Want to find errors with MIN( ), so if no error, set it to a big */
/*     number.  If there already is an error, multiply by the the des- */
/*     criptor multiplier */

    /* Parameter adjustments */
    --desca;

    /* Function Body */
    if (*info >= 0) {
	*info = 10000;
    } else if (*info < -100) {
	*info = -(*info);
    } else {
	*info = -(*info) * 100;
    }

/*     Figure where in parameter list each parameter was, factoring in */
/*     descriptor multiplier */

    mapos = *mapos0 * 100;
    napos = *napos0 * 100;
    iapos = (*descapos0 - 2) * 100;
    japos = (*descapos0 - 1) * 100;
    descapos = *descapos0 * 100;

/*     Check that matrix values make sense from local viewpoint */

    if (desca[1] != 1) {
/* Computing MIN */
	i__1 = *info, i__2 = descapos + 1;
	*info = min(i__1,i__2);
    } else if (*ma < 0) {
	*info = min(*info,mapos);
    } else if (*na < 0) {
	*info = min(*info,napos);
    } else if (*ia < 1) {
	*info = min(*info,iapos);
    } else if (*ja < 1) {
	*info = min(*info,japos);
    } else if (desca[5] < 1) {
/* Computing MIN */
	i__1 = *info, i__2 = descapos + 5;
	*info = min(i__1,i__2);
    } else if (desca[6] < 1) {
/* Computing MIN */
	i__1 = *info, i__2 = descapos + 6;
	*info = min(i__1,i__2);
    } else if (desca[7] < 0 || desca[7] >= nprow) {
/* Computing MIN */
	i__1 = *info, i__2 = descapos + 7;
	*info = min(i__1,i__2);
    } else if (desca[8] < 0 || desca[8] >= npcol) {
/* Computing MIN */
	i__1 = *info, i__2 = descapos + 8;
	*info = min(i__1,i__2);
    } else if (desca[9] < 1) {
/* Computing MIN */
	i__1 = *info, i__2 = descapos + 9;
	*info = min(i__1,i__2);
    } else if (desca[9] < numroc_(&desca[3], &desca[5], &myrow, &desca[7], &
	    nprow)) {
	if (numroc_(&desca[4], &desca[6], &mycol, &desca[8], &npcol) > 0) {
/* Computing MIN */
	    i__1 = *info, i__2 = descapos + 9;
	    *info = min(i__1,i__2);
	}
    }

    if (*ma == 0 || *na == 0) {

/*        NULL matrix, relax some checks */

	if (desca[3] < 0) {
/* Computing MIN */
	    i__1 = *info, i__2 = descapos + 3;
	    *info = min(i__1,i__2);
	}
	if (desca[4] < 0) {
/* Computing MIN */
	    i__1 = *info, i__2 = descapos + 4;
	    *info = min(i__1,i__2);
	}

    } else {

/*        more rigorous checks for non-degenerate matrices */

	if (desca[3] < 1) {
/* Computing MIN */
	    i__1 = *info, i__2 = descapos + 3;
	    *info = min(i__1,i__2);
	} else if (desca[4] < 1) {
/* Computing MIN */
	    i__1 = *info, i__2 = descapos + 4;
	    *info = min(i__1,i__2);
	} else {
	    if (*ia > desca[3]) {
		*info = min(*info,iapos);
	    } else if (*ja > desca[4]) {
		*info = min(*info,japos);
	    } else {
		if (*ia + *ma - 1 > desca[3]) {
		    *info = min(*info,mapos);
		}
		if (*ja + *na - 1 > desca[4]) {
		    *info = min(*info,napos);
		}
	    }
	}

    }

/*     Prepare output: set info = 0 if no error, and divide by */
/*     DESCMULT if error is not in a descriptor entry */

    if (*info == 10000) {
	*info = 0;
    } else if (*info % 100 == 0) {
	*info = -(*info) / 100;
    } else {
	*info = -(*info);
    }

    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

