
/*
* ScaLAPACK TOOLS
*
* Ravi Reddy
* Version 1.0
* School of Computer Science and Informatics,
* UCD,
* Dublin,
* November 18, 2007
*/

/*
*  Include files
*/
#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_descinit(
    hscalinteger *desc, 
    hscalinteger *m,  
    hscalinteger *n, 
    hscalinteger *mb, 
    hscalinteger *nb, 
    hscalinteger *irsrc, 
    hscalinteger *icsrc, 
    hscalinteger *info
)
{
    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4;

/*  -- ScaLAPACK tools routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  DESCINIT initializes the descriptor vector with the 8 input arguments */
/*  M, N, MB, NB, IRSRC, ICSRC, ICTXT, LLD. */

/*  Notes */
/*  ===== */

/*  Each global data object is described by an associated description */
/*  vector.  This vector stores the information required to establish */
/*  the mapping between an object element and its corresponding process */
/*  and memory location. */

/*  Let A be a generic term for any 2D block cyclicly distributed array. */
/*  Such a global array has an associated description vector DESCA. */
/*  In the following comments, the character _ should be read as */
/*  "of the global array". */

/*  NOTATION        STORED IN      EXPLANATION */
/*  --------------- -------------- -------------------------------------- */
/*  DTYPE_A(global) DESCA( DTYPE_ )The descriptor type.  In this case, */
/*                                 DTYPE_A = 1. */
/*  CTXT_A (global) DESCA( CTXT_ ) The BLACS context handle, indicating */
/*                                 the BLACS process grid A is distribu- */
/*                                 ted over. The context itself is glo- */
/*                                 bal, but the handle (the hscalinteger */
/*                                 value) may vary. */
/*  M_A    (global) DESCA( M_ )    The number of rows in the global */
/*                                 array A. */
/*  N_A    (global) DESCA( N_ )    The number of columns in the global */
/*                                 array A. */
/*  MB_A   (global) DESCA( MB_ )   The blocking factor used to distribute */
/*                                 the rows of the array. */
/*  NB_A   (global) DESCA( NB_ )   The blocking factor used to distribute */
/*                                 the columns of the array. */
/*  RSRC_A (global) DESCA( RSRC_ ) The process row over which the first */
/*                                 row of the array A is distributed. */
/*  CSRC_A (global) DESCA( CSRC_ ) The process column over which the */
/*                                 first column of the array A is */
/*                                 distributed. */
/*  LLD_A  (local)  DESCA( LLD_ )  The leading dimension of the local */
/*                                 array.  LLD_A >= MAX(1,LOCr(M_A)). */

/*  Let K be the number of rows or columns of a distributed matrix, */
/*  and assume that its process grid has dimension p x q. */
/*  LOCr( K ) denotes the number of elements of K that a process */
/*  would receive if K were distributed over the p processes of its */
/*  process column. */
/*  Similarly, LOCc( K ) denotes the number of elements of K that a */
/*  process would receive if K were distributed over the q processes of */
/*  its process row. */
/*  The values of LOCr() and LOCc() may be determined via a call to the */
/*  ScaLAPACK tool function, NUMROC: */
/*          LOCr( M ) = NUMROC( M, MB_A, MYROW, RSRC_A, NPROW ), */
/*          LOCc( N ) = NUMROC( N, NB_A, MYCOL, CSRC_A, NPCOL ). */
/*  An upper bound for these quantities may be computed by: */
/*          LOCr( M ) <= ceil( ceil(M/MB_A)/NPROW )*MB_A */
/*          LOCc( N ) <= ceil( ceil(N/NB_A)/NPCOL )*NB_A */

/*  Arguments */
/*  ========= */

/*  DESC    (output) hscalinteger array of dimension DLEN_. */
/*          The array descriptor of a distributed matrix to be set. */

/*  M       (global input) hscalinteger */
/*          The number of rows in the distributed matrix. M >= 0. */

/*  N       (global input) hscalinteger */
/*          The number of columns in the distributed matrix. N >= 0. */

/*  MB      (global input) hscalinteger */
/*          The blocking factor used to distribute the rows of the */
/*          matrix. MB >= 1. */

/*  NB      (global input) hscalinteger */
/*          The blocking factor used to distribute the columns of the */
/*          matrix. NB >= 1. */

/*  IRSRC   (global input) hscalinteger */
/*          The process row over which the first row of the matrix is */
/*          distributed. 0 <= IRSRC < NPROW. */

/*  ICSRC   (global input) hscalinteger */
/*          The process column over which the first column of the */
/*          matrix is distributed. 0 <= ICSRC < NPCOL. */

/*  INFO    (output) hscalinteger */
/*          = 0: successful exit */
/*          < 0: if INFO = -i, the i-th argument had an illegal value */

/*  Note */
/*  ==== */

/*  If the routine can recover from an erroneous input argument, it will */
/*  return an acceptable descriptor vector.  */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     Get grid parameters */

    /* Parameter adjustments */
    --desc;

    *info = 0;
    if (*m < 0) {
	*info = -2;
    } else if (*n < 0) {
	*info = -3;
    } else if (*mb < 1) {
	*info = -4;
    } else if (*nb < 1) {
	*info = -5;
    } else if (*irsrc < 0) {
	*info = -6;
    } else if (*icsrc < 0) {
	*info = -7;
    }

    if (*info != 0) {
	i__1 = -(*info);
      return HSCAL_ERR_INPUT;
    }

    desc[1] = 1;
    desc[3] = max(0,*m);
    desc[4] = max(0,*n);
    desc[5] = max(1,*mb);
    desc[6] = max(1,*nb);
    desc[7] = *irsrc;
    desc[8] = *icsrc;
    desc[2] = -1;
    desc[9] = -1;

    return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

