
/*
*  Include files
*/
#include "../include/hscalapack.h"

/*------------------------------------------------------------------*/

int hscal_pdtreecomb(
    hscalinteger *ictxt, char *scope, hscalinteger *n, 
    hscalinteger *rdest0, hscalinteger *cdest0, HSCAL_S_fp subptr, 
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
)
{
    hscalinteger c__1 = 1;

    hscalinteger i__, np, iam;
    hscaldoublereal his[2];
    hscalinteger dest, dist;
    hscallogical bcast;
    hscalinteger cmssg, rmssg;
    hscallogical cscope;
    hscalinteger tcdest;
    hscallogical rscope;
    hscalinteger trdest, mydist;
    hscalinteger mydist2, hisdist;

    HSCALBLACSCONTEXT   * blacsctxt;

/*  -- ScaLAPACK tools routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */
/*     .. Subroutine Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  PDTREECOMB does a 1-tree parallel combine operation on scalars, */
/*  using the subroutine indicated by SUBPTR to perform the required */
/*  computation. */

/*  Arguments */
/*  ========= */

/*  ICTXT   (global input) INTEGER */
/*          The BLACS context handle, indicating the global context of */
/*          the operation. The context itself is global. */

/*  SCOPE   (global input) CHARACTER */
/*          The scope of the operation:  'Rowwise', 'Columnwise', or */
/*          'All'. */

/*  N       (global input) INTEGER */
/*          The number of elements in MINE.  N = 1 for the norm-2 */
/*          computation and 2 for the sum of square. */

/*  MINE    (local input/global output) DOUBLE PRECISION array of */
/*          dimension at least equal to N. The local data to use in the */
/*          combine. */

/*  RDEST0  (global input) INTEGER */
/*          The process row to receive the answer. If RDEST0 = -1, */
/*          every process in the scope gets the answer. */

/*  CDEST0  (global input) INTEGER */
/*          The process column to receive the answer. If CDEST0 = -1, */
/*          every process in the scope gets the answer. */

/*  SUBPTR  (local input) Pointer to the subroutine to call to perform */
/*          the required combine. */

/*  ===================================================================== */

/*     .. Local Scalars .. */
/*     .. */
/*     .. Local Arrays .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*     See if everyone wants the answer (need to broadcast the answer) */

    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

    /* Function Body */
    bcast = *rdest0 == -1 || *cdest0 == -1;
    if (bcast) {
	trdest = 0;
	tcdest = 0;
    } else {
	trdest = *rdest0;
	tcdest = *cdest0;
    }

/*     Figure scope-dependant variables, or report illegal scope */

    rscope = hscal_lsame(scope, "R");
    cscope = hscal_lsame(scope, "C");

    if (rscope) {
	if (bcast) {
	    trdest = myrow;
	} else if (myrow != trdest) {
	    return 0;
	}
	np = npcol;
	mydist = (npcol + mycol - tcdest) % npcol;
    } else if (cscope) {
	if (bcast) {
	    tcdest = mycol;
	} else if (mycol != tcdest) {
	    return 0;
	}
	np = nprow;
	mydist = (nprow + myrow - trdest) % nprow;
    } else if (hscal_lsame(scope, "A")) {
	np = nprow * npcol;
	iam = myrow * npcol + mycol;
	dest = trdest * npcol + tcdest;
	mydist = (np + iam - dest) % np;
    } else {
	return 0;
    }

    if (np < 2) {
	return 0;
    }

    mydist2 = mydist;
    rmssg = myrow;
    cmssg = mycol;
    i__ = 1;

L10:

    if (mydist % 2 != 0) {

/*           If I am process that sends information */

	dist = i__ * (mydist - mydist % 2);

/*           Figure coordinates of dest of message */

	if (rscope) {
	    cmssg = (tcdest + dist) % np;
	} else if (cscope) {
	    rmssg = (trdest + dist) % np;
	} else {
	    cmssg = (dest + dist) % np;
	    rmssg = cmssg / npcol;
	    cmssg %= npcol;
	}

	hscal_Cdgesd2d(
             blacsctxt, *n, c__1, *n, rmssg, cmssg,
             nprow, npcol, myrow, mycol, tcomp, tcomm
        );

	goto L20;

    } else {

/*           If I am a process receiving information, figure coordinates */
/*           of source of message */

	dist = mydist2 + i__;
	if (rscope) {
	    cmssg = (tcdest + dist) % np;
	    hisdist = (np + cmssg - tcdest) % np;
	} else if (cscope) {
	    rmssg = (trdest + dist) % np;
	    hisdist = (np + rmssg - trdest) % np;
	} else {
	    cmssg = (dest + dist) % np;
	    rmssg = cmssg / npcol;
	    cmssg %= npcol;
	    hisdist = (np + rmssg * npcol + cmssg - dest) % np;
	}

	if (mydist2 < hisdist) {

/*              If I have anyone sending to me */

	    (*subptr)(
                 nprow, npcol, myrow, mycol, tcomp
            );

	}
	mydist /= 2;

    }
    i__ <<= 1;

    if (i__ < np) {
	goto L10;
    }

L20:

    if (bcast) {
	if (mydist2 == 0) {
	    hscal_Cdgebs2d(
                 blacsctxt, scope, " ", *n, c__1, *n, 
                 nprow, npcol, myrow, mycol, tcomp, tcomm
	    );
	} else {
	    hscal_Cdgebr2d(
                 blacsctxt, scope, " ", *n, c__1, *n, trdest, tcdest,
                 nprow, npcol, myrow, mycol, tcomp, tcomm
            );
	}
    }

    Cblacs_gridexit_nocomm(blacsctxt);

    return HSCAL_SUCCESS;

/*     End of PDTREECOMB */

} /* pdtreecomb_ */

/*------------------------------------------------------------------*/

int hscal_dcombamax(
    int nprow, int npcol, int myrow, int mycol, double *tcomp
)
{

/*  -- ScaLAPACK tools routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  DCOMBAMAX finds the element having max. absolute value as well */
/*  as its corresponding globl index. */

/*  Arguments */
/*  ========= */

/*  V1        (local input/local output) DOUBLE PRECISION array of */
/*            dimension 2.  The first maximum absolute value element and */
/*            its global index. V1(1) = AMAX, V1(2) = INDX. */

/*  V2        (local input) DOUBLE PRECISION array of dimension 2. */
/*            The second maximum absolute value element and its global */
/*            index. V2(1) = AMAX, V2(2) = INDX. */

/*  ===================================================================== */

/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

    if (tcomp != NULL)
    {
       TCOMP((myrow), (mycol), (nprow), (npcol)) += 3;
    }

    return HSCAL_SUCCESS;

/*     End of DCOMBAMAX */

} /* dcombamax_ */

/*------------------------------------------------------------------*/

int hscal_dcombssq(
    int nprow, int npcol, int myrow, int mycol, double *tcomp
)
{

/*  -- ScaLAPACK tools routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  DCOMBSSQ does a scaled sum of squares on two scalars. */

/*  Arguments */
/*  ========= */

/*  V1        (local input/local output) DOUBLE PRECISION array of */
/*            dimension 2.  The first scaled sum. V1(1) = SCALE, */
/*            V1(2) = SUMSQ. */

/*  V2        (local input) DOUBLE PRECISION array of dimension 2. */
/*            The second scaled sum. V2(1) = SCALE, V2(2) = SUMSQ. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Executable Statements .. */

    if (tcomp != NULL)
    {
       TCOMP((myrow), (mycol), (nprow), (npcol)) += 5;
    }

    return HSCAL_SUCCESS;

} /* dcombssq_ */

/*------------------------------------------------------------------*/

int hscal_dcombnrm2(
    int nprow, int npcol, int myrow, int mycol, double *tcomp
)
{

/*  -- ScaLAPACK tools routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Scalar Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  DCOMBNRM2 combines local norm 2 results, taking care not to cause */
/*  unnecessary overflow. */

/*  Arguments */
/*  ========= */

/*  X       (local input) DOUBLE PRECISION */
/*  Y       (local input) DOUBLE PRECISION */
/*          X and Y specify the values x and y. X and Y are supposed to */
/*          be >= 0. */

/*  ===================================================================== */

/*     .. Parameters .. */
/*     .. */
/*     .. Local Scalars .. */
/*     .. */
/*     .. Intrinsic Functions .. */
/*     .. */
/*     .. Executable Statements .. */

    if (tcomp != NULL)
    {
       TCOMP((myrow), (mycol), (nprow), (npcol)) += 5;
    }

    return HSCAL_SUCCESS;

/*     End of DCOMBNRM2 */

} /* dcombnrm2_ */

/*------------------------------------------------------------------*/

