
/*
* ScaLAPACK TOOLS
*
* Ravi Reddy
* Version 1.0
* School of Computer Science and Informatics,
* UCD,
* Dublin,
* November 18, 2007
*/

/*
*  Include files
*/
#include "../include/hscalapack.h"

int hscal_pirow2col(
    hscalinteger *ictxt, 
    hscalinteger *m, 
    hscalinteger *n, 
    hscalinteger *nb, 
    hscalinteger *ldvs, 
    hscalinteger *ldvd, 
    hscalinteger *rsrc, 
    hscalinteger *csrc, 
    hscalinteger *rdest, 
    hscalinteger *cdest,
    int nprow,
    int npcol,
    int myrow,
    int mycol,
    double *tcomp,
    int *tcomm
)
{
    hscalinteger c__1 = 1;

    /* System generated locals */
    hscalinteger i__1, i__2, i__3, i__4, i__5;

    /* Local variables */
    hscalinteger cblkskip, rblkskip, k, jb, ii, jj, mp, mq, lcm;
    hscalinteger icpy, icsrc;
    hscalinteger irdest;
    hscalinteger istart, mydist;
    hscalinteger nblocks;

    HSCALBLACSCONTEXT   * blacsctxt;

/*  -- ScaLAPACK tools routine (version 1.7) -- */
/*     University of Tennessee, Knoxville, Oak Ridge National Laboratory, */
/*     and University of California, Berkeley. */
/*     May 1, 1997 */

/*     .. Scalar Arguments .. */
/*     .. */
/*     .. Array Arguments .. */
/*     .. */

/*  Purpose */
/*  ======= */

/*  Take a block of vectors with M total rows which are distributed over */
/*  a row of processes, and distribute those rows over a column of */
/*  processes. This routine minimizes communication by sending all */
/*  information it has that a given process in the CDEST needs at once. */
/*  To do this it uses the least common multiple (LCM) concept.  This is */
/*  simply the realization that if I have part of a vector split over a */
/*  process row consisting of Q processes, and I want to send all of that */
/*  vector that I own to a new vector distributed over P processes within */
/*  a process column, that after I find the process in RDEST that owns */
/*  the row of the vector I'm currently looking at, he will want every */
/*  ( (LCM(P,Q)/Q ) block of my vector (the block being of size NB x N). */

/*  Arguments */
/*  ========= */

/*  Rem:  MP, resp. NQ, denotes the number of local rows, resp. local */
/*  ====  columns, necessary to store a global vector of dimension M */
/*        across P processes, resp. N over Q processes. */

/*  ICTXT   (global input) INTEGER */
/*          The BLACS context handle, indicating the global context of */
/*          the operation. The context itself is global. */

/*  M       (global input) INTEGER */
/*          The number of global rows each vector has. */

/*  N       (global input) INTEGER */
/*          The number of vectors in the vector block */

/*  NB      (global input) INTEGER */
/*          The blocking factor used to divide the rows of the vector */
/*          amongst the processes of a row. */

/*  VS      (local input) @(typec) */
/*          Array of dimension (LDVS,N), the block of vectors stored on */
/*          process row RSRC to be put into memory VD, and stored on */
/*          process column CDEST. */

/*  LDVS    (local input) INTEGER */
/*          The leading dimension of VS. */

/*  VD      (local output) @(typec) */
/*          Array of dimension (LDVD,N), on output, the contents of VD */
/*          stored on process column CDEST will be here. */

/*  LDVD    (local input) INTEGER */
/*          The leading dimension of VD. */

/*  RSRC    (global input) INTEGER */
/*          The process row VS is distributed over. */

/*  CSRC    (global input) INTEGER */
/*          The process column the distributed block of vectors VS */
/*          begins on. */

/*  RDEST   (global input) INTEGER */
/*          The process row that VD begins on. */

/*  CDEST   (global input) INTEGER */
/*          The process column to distribute VD over. */

/*  WORK    (local workspace) @(typec) */
/*          Array, dimension (LDW). The required size of work varies: */
/*          if( nprow.eq.npcol ) then */
/*             LDW = 0; WORK not accessed. */
/*          else */
/*            lcm = least common multiple of process rows and columns. */
/*             Mq  = number of rows of VS on my process. */
/*             npcol = number of process columns */
/*             CEIL = the ceiling of given operation */
/*             LDW = NB*N*CEIL( CEIL( Mq/NB )/(LCM/npcol) ) */
/*          end if */

/*  ===================================================================== */

/*     .. Local Scalars .. */
/*     .. */
/*     .. External Subroutines .. */
/*     .. */
/*     .. External Functions .. */
/*     .. */
/*     .. Executable Statements .. */

/*
*  Retrieve process grid information
*  Context is not really used in performance models
*  Create dummy BLACS context
*/
    Cblacs_gridinit_nocomm(&blacsctxt, "r", nprow, npcol);

/*     If we are not in special case for NPROW = NPCOL where there is no */
/*     copying required */

    if (nprow != npcol) {
	lcm = hscal_ilcm(&nprow, &npcol);
	rblkskip = lcm / npcol;
	cblkskip = lcm / nprow;

/*        If I have part of VS, the source vector(s) */

	if (myrow == *rsrc) {

	    istart = 1;

/*           Figure my distance from CSRC: the process in CDEST the same */
/*           distance from RDEST will want my first block */

	    mydist = (npcol + mycol - *csrc) % npcol;
	    mq = numroc_(m, nb, &mycol, csrc, &npcol);
	    irdest = (*rdest + mydist) % nprow;

/*           Loop over all possible destination processes */

	    i__1 = rblkskip;
	    for (k = 1; k <= i__1; ++k) {
		jj = 1;

/*              If I am not destination process */

		if (myrow != irdest || mycol != *cdest) {

/*                 Pack all data I own that destination needs */

		    i__2 = mq;
		    i__3 = *nb * rblkskip;
		    for (ii = istart; i__3 < 0 ? ii >= i__2 : ii <= i__2; ii 
			    += i__3) {
/* Computing MIN */
			i__4 = *nb, i__5 = mq - ii + 1;
			jb = min(i__4,i__5);
			ilacpy_tcomp_macro(
                              "G", &jb, n, ldvs,
	                      &jb,
                              nprow, npcol, myrow, mycol, tcomp
                        );
			jj += *nb * *n;
/* L10: */
		    }

/*                 Figure how many rows are to be sent and send them if */
/*                 necessary, NOTE: will send extra if NB > JB */

		    --jj;
		    if (jj > 0) {
                        hscal_Cigesd2d(
			     blacsctxt, jj, c__1, jj, 
                             irdest, *cdest,
                             nprow, npcol, myrow, mycol, tcomp, tcomm
                        );
		    }

/*              I am both source and destination, save where to start */
/*              copying from for later use */

		} else {
		    icpy = istart;
		}

		istart += *nb;
		irdest = (irdest + npcol) % nprow;
/* L20: */
	    }
	}

/*        If I should receive info into VD */

	if (mycol == *cdest) {

	    istart = 1;

/*           Figure my distance from CDEST: the process in CSRC the same */
/*           distance from RSRC will have my first block */

	    mydist = (nprow + myrow - *rdest) % nprow;
	    mp = numroc_(m, nb, &myrow, rdest, &nprow);
	    icsrc = (*csrc + mydist) % npcol;

/*           Loop over all sending processes */

	    i__1 = cblkskip;
	    for (k = 1; k <= i__1; ++k) {

/*              If I don't already possess the required data */

		if (myrow != *rsrc || mycol != icsrc) {

/*                 Figure how many rows to receive, and receive them */
/*                 NOTE: may receive to much -- NB instead of JB */

		    nblocks = (mp - istart + *nb) / *nb;
		    jj = (nblocks + cblkskip - 1) / cblkskip * *nb;
		    if (jj > 0) {
			;
		    }

/*                 Copy data to destination vector */

		    jj = 1;
		    i__3 = mp;
		    i__2 = *nb * cblkskip;
		    for (ii = istart; i__2 < 0 ? ii >= i__3 : ii <= i__3; ii 
			    += i__2) {
/* Computing MIN */
			i__4 = *nb, i__5 = mp - ii + 1;
			jb = min(i__4,i__5);
			ilacpy_tcomp_macro(
                              "G", &jb, n, &jb, ldvd,
                              nprow, npcol, myrow, mycol, tcomp
                        );
			jj += *nb * *n;
/* L30: */
		    }

/*              If I am both source and destination */

		} else {
		    jj = icpy;
		    i__2 = mp;
		    i__3 = *nb * cblkskip;
		    for (ii = istart; i__3 < 0 ? ii >= i__2 : ii <= i__2; ii 
			    += i__3) {
/* Computing MIN */
			i__4 = *nb, i__5 = mp - ii + 1;
			jb = min(i__4,i__5);
			ilacpy_tcomp_macro(
                              "G", &jb, n, ldvs,
			      ldvd, 
                              nprow, npcol, myrow, mycol, tcomp
                        );
			jj += *nb * rblkskip;
/* L40: */
		    }
		}
		istart += *nb;
		icsrc = (icsrc + nprow) % npcol;
/* L50: */
	    }
	}

/*     if NPROW = NPCOL, there is a one-to-one correspondance between */
/*     process rows and columns, so no work space or copying required */

    } else {

	if (myrow == *rsrc) {

/*           Figure my distance from CSRC: the process in CDEST the same */
/*           distance from RDEST will want my piece of the vector */

	    mydist = (npcol + mycol - *csrc) % npcol;
	    mq = numroc_(m, nb, &mycol, csrc, &npcol);
	    irdest = (*rdest + mydist) % nprow;
	    if (myrow != irdest || mycol != *cdest) {
                hscal_Cigesd2d(
                     blacsctxt, mq, *n, *ldvs, irdest, *cdest,
                     nprow, npcol, myrow, mycol, tcomp, tcomm
                );
	    } else {
		ilacpy_tcomp_macro(
                      "G", &mq, n, ldvs,
	              ldvd, 
                      nprow, npcol, myrow, mycol, tcomp
                );
	    }
	}
	if (mycol == *cdest) {

/*           Figure my distance from RDEST: the process in RSRC the same */
/*           distance from CSRC will have my piece of the vector */

	    mydist = (nprow + myrow - *rdest) % nprow;
	    mp = numroc_(m, nb, &myrow, rdest, &nprow);
	    icsrc = (*csrc + mydist) % npcol;
	    if (mycol != icsrc || myrow != *rsrc) {
               ;
	    }
	}
    }

    Cblacs_gridexit_nocomm(blacsctxt);

    return HSCAL_SUCCESS;
}

