/*******************************************************************
*                                                                  *
*                       Heterogeneous ScaLAPACK                    *
*                 Recon, Group management Routines                 *
*                             (INTERNAL)                           *
*                 ====================================             *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_GROUP_MGMT_HH
#define __HMPI_SCALAPACK_GROUP_MGMT_HH

/*
* pcgesv
*/
double hscal_pcgesv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cgesv_benchmark(
    int     *n, 
    int     *nrhs, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *b, 
    int     *ldb
);

int hscal_pack_pcgesv_modelp_noauto(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgesv_modelp(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pcgesv_recon(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pcgesv_timeof(
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgesv_group_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgesv_group_auto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgesv_group_heuristic_auto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcgesv_group_heuristic_pauto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdgesv
*/
double hscal_pdgesv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dgesv_benchmark(
    int     *n, 
    int     *nrhs, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *b, 
    int     *ldb
);

int hscal_pack_pdgesv_modelp_noauto(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgesv_modelp(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pdgesv_recon(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pdgesv_timeof(
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgesv_group_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgesv_group_auto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgesv_group_heuristic_auto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdgesv_group_heuristic_pauto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psgesv
*/
double hscal_psgesv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_sgesv_benchmark(
    int     *n, 
    int     *nrhs, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *b, 
    int     *ldb
);

int hscal_pack_psgesv_modelp_noauto(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgesv_modelp(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_psgesv_recon(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_psgesv_timeof(
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgesv_group_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgesv_group_auto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgesv_group_heuristic_auto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psgesv_group_heuristic_pauto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzgesv
*/
double hscal_pzgesv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zgesv_benchmark(
    int     *n, 
    int     *nrhs, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *b, 
    int     *ldb
);

int hscal_pack_pzgesv_modelp_noauto(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgesv_modelp(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pzgesv_recon(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pzgesv_timeof(
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgesv_group_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgesv_group_auto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgesv_group_heuristic_auto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzgesv_group_heuristic_pauto_create(
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcgetrf
*/
double hscal_pcgetrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cgetrf_benchmark(
    int     *m,
    int     *n,
    hscalcomplex *a,
    int     *lda
);

int hscal_pack_pcgetrf_modelp_noauto(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgetrf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcgetrf_recon(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA
);

double hscal_pcgetrf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgetrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgetrf_group_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgetrf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcgetrf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdgetrf
*/
double hscal_pdgetrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dgetrf_benchmark(
    int     *m,
    int     *n,
    hscaldoublereal *a,
    int     *lda
);

int hscal_pack_pdgetrf_modelp_noauto(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgetrf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pdgetrf_recon(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA
);

double hscal_pdgetrf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgetrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgetrf_group_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgetrf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdgetrf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psgetrf
*/
double hscal_psgetrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_sgetrf_benchmark(
    int     *m,
    int     *n,
    hscalreal *a,
    int     *lda
);

int hscal_pack_psgetrf_modelp_noauto(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgetrf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_psgetrf_recon(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA
);

double hscal_psgetrf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgetrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgetrf_group_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgetrf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psgetrf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzgetrf
*/
double hscal_pzgetrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zgetrf_benchmark(
    int     *m,
    int     *n,
    hscaldoublecomplex *a,
    int     *lda
);

int hscal_pack_pzgetrf_modelp_noauto(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgetrf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzgetrf_recon(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA
);

double hscal_pzgetrf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgetrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgetrf_group_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgetrf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzgetrf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcgetrs
*/
double hscal_pcgetrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cgetrs_benchmark(
    char    *trans,
    int     *n,
    int     *nrhs,
    hscalcomplex *a,
    int     *lda,
    hscalcomplex *b,
    int     *ldb
);

int hscal_pack_pcgetrs_modelp_noauto(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgetrs_modelp(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pcgetrs_recon(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pcgetrs_timeof(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgetrs_group_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgetrs_group_auto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgetrs_group_heuristic_auto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcgetrs_group_heuristic_pauto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdgetrs
*/
double hscal_pdgetrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dgetrs_benchmark(
    char    *trans,
    int     *n,
    int     *nrhs,
    hscaldoublereal *a,
    int     *lda,
    hscaldoublereal *b,
    int     *ldb
);

int hscal_pack_pdgetrs_modelp_noauto(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgetrs_modelp(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pdgetrs_recon(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pdgetrs_timeof(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgetrs_group_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgetrs_group_auto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgetrs_group_heuristic_auto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdgetrs_group_heuristic_pauto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psgetrs
*/
double hscal_psgetrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_sgetrs_benchmark(
    char    *trans,
    int     *n,
    int     *nrhs,
    hscalreal *a,
    int     *lda,
    hscalreal *b,
    int     *ldb
);

int hscal_pack_psgetrs_modelp_noauto(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgetrs_modelp(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_psgetrs_recon(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_psgetrs_timeof(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgetrs_group_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgetrs_group_auto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgetrs_group_heuristic_auto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psgetrs_group_heuristic_pauto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzgetrs
*/
double hscal_pzgetrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zgetrs_benchmark(
    char    *trans,
    int     *n,
    int     *nrhs,
    hscaldoublecomplex *a,
    int     *lda,
    hscaldoublecomplex *b,
    int     *ldb
);

int hscal_pack_pzgetrs_modelp_noauto(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgetrs_modelp(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pzgetrs_recon(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pzgetrs_timeof(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgetrs_group_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgetrs_group_auto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgetrs_group_heuristic_auto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzgetrs_group_heuristic_pauto_create(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);


/*
* pxposv
*/
double hscal_pcposv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcposv_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcposv_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pcposv_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pcposv_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcposv_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcposv_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcposv_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcposv_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pdposv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pdposv_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdposv_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pdposv_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pdposv_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdposv_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdposv_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdposv_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdposv_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_psposv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_psposv_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psposv_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_psposv_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_psposv_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psposv_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psposv_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psposv_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psposv_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzposv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzposv_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzposv_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pzposv_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pzposv_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzposv_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzposv_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzposv_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzposv_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);


/*
* pxpotrs
*/
double hscal_pcpotrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcpotrs_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcpotrs_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pcpotrs_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pcpotrs_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcpotrs_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcpotrs_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcpotrs_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcpotrs_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pdpotrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pdpotrs_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdpotrs_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pdpotrs_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pdpotrs_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdpotrs_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdpotrs_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdpotrs_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdpotrs_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pspotrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pspotrs_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pspotrs_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pspotrs_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pspotrs_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pspotrs_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pspotrs_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pspotrs_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pspotrs_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzpotrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzpotrs_modelp_noauto(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzpotrs_modelp(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pzpotrs_recon(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB
);

double hscal_pzpotrs_timeof(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzpotrs_group_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzpotrs_group_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzpotrs_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzpotrs_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pxpotrf
*/
double hscal_pcpotrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcpotrf_modelp_noauto(
   char * UPLO,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcpotrf_modelp(
   char   * UPLO,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcpotrf_recon(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA
);

double hscal_pcpotrf_timeof(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcpotrf_group_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcpotrf_group_auto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcpotrf_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcpotrf_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pdpotrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pdpotrf_modelp_noauto(
   char * UPLO,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdpotrf_modelp(
   char   * UPLO,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pdpotrf_recon(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA
);

double hscal_pdpotrf_timeof(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdpotrf_group_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdpotrf_group_auto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdpotrf_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdpotrf_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pspotrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pspotrf_modelp_noauto(
   char * UPLO,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pspotrf_modelp(
   char   * UPLO,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pspotrf_recon(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA
);

double hscal_pspotrf_timeof(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pspotrf_group_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pspotrf_group_auto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pspotrf_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pspotrf_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzpotrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzpotrf_modelp_noauto(
   char * UPLO,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzpotrf_modelp(
   char   * UPLO,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzpotrf_recon(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA
);

double hscal_pzpotrf_timeof(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzpotrf_group_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzpotrf_group_auto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzpotrf_group_heuristic_pauto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzpotrf_group_heuristic_auto_create(
   char   * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pxdbsv
*/
double hscal_pcdbsv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcdbsv_modelp_noauto(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcdbsv_modelp(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pcdbsv_recon(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
);

double hscal_pcdbsv_timeof(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA,
   int * DESCA,
   int * IB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcdbsv_group_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdbsv_group_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdbsv_group_heuristic_pauto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcdbsv_group_heuristic_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pddbsv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pddbsv_modelp_noauto(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pddbsv_modelp(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pddbsv_recon(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
);

double hscal_pddbsv_timeof(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA,
   int * DESCA,
   int * IB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pddbsv_group_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pddbsv_group_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pddbsv_group_heuristic_pauto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pddbsv_group_heuristic_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_psdbsv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_psdbsv_modelp_noauto(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psdbsv_modelp(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_psdbsv_recon(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
);

double hscal_psdbsv_timeof(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA,
   int * DESCA,
   int * IB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psdbsv_group_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psdbsv_group_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psdbsv_group_heuristic_pauto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psdbsv_group_heuristic_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzdbsv_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzdbsv_modelp_noauto(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzdbsv_modelp(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pzdbsv_recon(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
);

double hscal_pzdbsv_timeof(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA,
   int * DESCA,
   int * IB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzdbsv_group_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdbsv_group_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdbsv_group_heuristic_pauto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzdbsv_group_heuristic_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pxdbtrf
*/
double hscal_pcdbtrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcdbtrf_modelp_noauto(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcdbtrf_modelp(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcdbtrf_recon(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA
);

double hscal_pcdbtrf_timeof(
   int * N,
   int * BWL,
   int * BWU,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcdbtrf_group_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdbtrf_group_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdbtrf_group_heuristic_pauto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcdbtrf_group_heuristic_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pddbtrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pddbtrf_modelp_noauto(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pddbtrf_modelp(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pddbtrf_recon(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA
);

double hscal_pddbtrf_timeof(
   int * N,
   int * BWL,
   int * BWU,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pddbtrf_group_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pddbtrf_group_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pddbtrf_group_heuristic_pauto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pddbtrf_group_heuristic_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_psdbtrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_psdbtrf_modelp_noauto(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psdbtrf_modelp(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_psdbtrf_recon(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA
);

double hscal_psdbtrf_timeof(
   int * N,
   int * BWL,
   int * BWU,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psdbtrf_group_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psdbtrf_group_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psdbtrf_group_heuristic_pauto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psdbtrf_group_heuristic_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzdbtrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzdbtrf_modelp_noauto(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzdbtrf_modelp(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzdbtrf_recon(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA
);

double hscal_pzdbtrf_timeof(
   int * N,
   int * BWL,
   int * BWU,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzdbtrf_group_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdbtrf_group_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdbtrf_group_heuristic_pauto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzdbtrf_group_heuristic_auto_create(
   int * N,
   int * BWL,
   int * BWU,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pxdbtrs
*/
double hscal_pcdbtrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcdbtrs_modelp_noauto(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcdbtrs_modelp(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pcdbtrs_recon(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
);

double hscal_pcdbtrs_timeof(
   F_CHAR_T TRANS,
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA,
   int * DESCA,
   int * IB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcdbtrs_group_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdbtrs_group_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdbtrs_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcdbtrs_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pddbtrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pddbtrs_modelp_noauto(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pddbtrs_modelp(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pddbtrs_recon(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
);

double hscal_pddbtrs_timeof(
   F_CHAR_T TRANS,
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA,
   int * DESCA,
   int * IB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pddbtrs_group_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pddbtrs_group_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pddbtrs_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pddbtrs_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_psdbtrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_psdbtrs_modelp_noauto(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psdbtrs_modelp(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_psdbtrs_recon(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
);

double hscal_psdbtrs_timeof(
   F_CHAR_T TRANS,
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA,
   int * DESCA,
   int * IB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psdbtrs_group_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psdbtrs_group_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psdbtrs_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psdbtrs_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzdbtrs_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzdbtrs_modelp_noauto(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzdbtrs_modelp(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pzdbtrs_recon(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB
);

double hscal_pzdbtrs_timeof(
   F_CHAR_T TRANS,
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA,
   int * DESCA,
   int * IB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzdbtrs_group_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdbtrs_group_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdbtrs_group_heuristic_pauto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzdbtrs_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * N,
   int * BWL,
   int * BWU,
   int * NRHS,
   int * JA, 
   int * DESCA,
   int * IB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pxgels
*/
double hscal_pcgels_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcgels_modelp_noauto(
   char * TRANS,
   int * M, 
   int * N,
   int * NRHS, 
   int * IA, 
   int * JA,
   int * DESCA,
   int * IB, 
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgels_modelp(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pcgels_recon(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA, 
   int * DESCA,
   int * IB,
   int * JB, 
   int * DESCB
);

double hscal_pcgels_timeof(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgels_group_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgels_group_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgels_group_heuristic_pauto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB, 
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcgels_group_heuristic_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pdgels_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pdgels_modelp_noauto(
   char * TRANS,
   int * M, 
   int * N,
   int * NRHS, 
   int * IA, 
   int * JA,
   int * DESCA,
   int * IB, 
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgels_modelp(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pdgels_recon(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA, 
   int * DESCA,
   int * IB,
   int * JB, 
   int * DESCB
);

double hscal_pdgels_timeof(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgels_group_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgels_group_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgels_group_heuristic_pauto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB, 
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdgels_group_heuristic_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_psgels_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_psgels_modelp_noauto(
   char * TRANS,
   int * M, 
   int * N,
   int * NRHS, 
   int * IA, 
   int * JA,
   int * DESCA,
   int * IB, 
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgels_modelp(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_psgels_recon(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA, 
   int * DESCA,
   int * IB,
   int * JB, 
   int * DESCB
);

double hscal_psgels_timeof(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgels_group_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgels_group_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgels_group_heuristic_pauto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB, 
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psgels_group_heuristic_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzgels_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzgels_modelp_noauto(
   char * TRANS,
   int * M, 
   int * N,
   int * NRHS, 
   int * IA, 
   int * JA,
   int * DESCA,
   int * IB, 
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgels_modelp(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pzgels_recon(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA, 
   int * DESCA,
   int * IB,
   int * JB, 
   int * DESCB
);

double hscal_pzgels_timeof(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgels_group_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgels_group_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgels_group_heuristic_pauto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB, 
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzgels_group_heuristic_auto_create(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pxgelqf
*/
double hscal_pcgelqf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcgelqf_modelp_noauto(
   int * M,
   int * N, 
   int * IA,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgelqf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcgelqf_recon(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA
);



int hscal_pcgelqf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

double hscal_pcgelqf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgelqf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgelqf_group_auto_create(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgelqf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcgelqf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pdgelqf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pdgelqf_modelp_noauto(
   int * M,
   int * N, 
   int * IA,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgelqf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pdgelqf_recon(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA
);



int hscal_pdgelqf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

double hscal_pdgelqf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgelqf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgelqf_group_auto_create(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgelqf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdgelqf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_psgelqf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_psgelqf_modelp_noauto(
   int * M,
   int * N, 
   int * IA,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgelqf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_psgelqf_recon(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA
);



int hscal_psgelqf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

double hscal_psgelqf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgelqf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgelqf_group_auto_create(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgelqf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psgelqf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzgelqf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzgelqf_modelp_noauto(
   int * M,
   int * N, 
   int * IA,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgelqf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzgelqf_recon(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA
);



int hscal_pzgelqf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

double hscal_pzgelqf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgelqf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgelqf_group_auto_create(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgelqf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzgelqf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pxgeqrf
*/
double hscal_pcgeqrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pcgeqrf_modelp_noauto(
   int * M,
   int * N, 
   int * IA,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgeqrf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcgeqrf_recon(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA
);



int hscal_pcgeqrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

double hscal_pcgeqrf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgeqrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgeqrf_group_auto_create(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgeqrf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pcgeqrf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pdgeqrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pdgeqrf_modelp_noauto(
   int * M,
   int * N, 
   int * IA,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgeqrf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pdgeqrf_recon(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA
);



int hscal_pdgeqrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

double hscal_pdgeqrf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgeqrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgeqrf_group_auto_create(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgeqrf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pdgeqrf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_psgeqrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_psgeqrf_modelp_noauto(
   int * M,
   int * N, 
   int * IA,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgeqrf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_psgeqrf_recon(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA
);



int hscal_psgeqrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

double hscal_psgeqrf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgeqrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgeqrf_group_auto_create(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgeqrf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_psgeqrf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

double hscal_pzgeqrf_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_pack_pzgeqrf_modelp_noauto(
   int * M,
   int * N, 
   int * IA,
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgeqrf_modelp(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzgeqrf_recon(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA
);



int hscal_pzgeqrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

double hscal_pzgeqrf_timeof(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgeqrf_group_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgeqrf_group_auto_create(
   int * M,
   int * N,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgeqrf_group_heuristic_pauto_create(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

int hscal_pzgeqrf_group_heuristic_auto_create(
   int * M,
   int * N,
   int * IA,
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

#endif

