
/*******************************************************************
*                                                                  *
*                     Heterogeneous ScaLAPACK                      *
*                     auxiliary information functions              *
*                               (INTERNAL)                         *
*                 =====================================            *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_AUX_INFO_I_INCLUDE_HH
#define __HMPI_SCALAPACK_AUX_INFO_I_INCLUDE_HH

/*
* These functions are internal usage.
* _info functions are for the users.
* The differ from the _info functions
* from the only fact that the values in the 
* output parameters (tcomp, tcomm) are accumulated
* and not refreshed every time
*/

/*
* pcgesv
*/
int hscal_pcgesv_info_i(
   int *N,
   int *NRHS,
   int *IA,
   int *JA,
   int *DESCA,
   int *IB,
   int *JB,
   int *DESCB,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
);

/*
* pdgesv
*/
int hscal_pdgesv_info_i(
   int *N,
   int *NRHS,
   int *IA,
   int *JA,
   int *DESCA,
   int *IB,
   int *JB,
   int *DESCB,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
);

/*
* psgesv
*/
int hscal_psgesv_info_i(
   int *N,
   int *NRHS,
   int *IA,
   int *JA,
   int *DESCA,
   int *IB,
   int *JB,
   int *DESCB,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
);

/*
* pzgesv
*/
int hscal_pzgesv_info_i(
   int *N,
   int *NRHS,
   int *IA,
   int *JA,
   int *DESCA,
   int *IB,
   int *JB,
   int *DESCB,
   int nprow,
   int npcol,
   double *tcomp,
   int *tcomm
);

/*
* pcgetrf
*/
int hscal_pcgetrf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pcgetrs
*/
int hscal_pcgetrs_info_i(
   char * TRANS,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);


/*
* pxposv
*/
int hscal_pcposv_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdposv_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psposv_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzposv_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxpotrf
*/
int hscal_pcpotrf_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdpotrf_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pspotrf_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzpotrf_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxpotrs
*/
int hscal_pcpotrs_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdpotrs_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pspotrs_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzpotrs_info_i(
   char * UPLO,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxpotf2
*/
int hscal_pcpotf2_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp
);

int hscal_pdpotf2_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp
);

int hscal_pspotf2_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp
);

int hscal_pzpotf2_info_i(
   char * UPLO,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp
);

/*
* pxdbsv
*/
int hscal_pcdbsv_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pddbsv_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psdbsv_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzdbsv_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxdbtrf
*/
int hscal_pcdbtrf_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * JA,
   int    * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pddbtrf_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * JA,
   int    * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psdbtrf_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * JA,
   int    * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzdbtrf_info_i(
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * JA,
   int    * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxdbtrs
*/
int hscal_pcdbtrs_info_i(
   char   * TRANS,
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pddbtrs_info_i(
   char   * TRANS,
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psdbtrs_info_i(
   char   * TRANS,
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzdbtrs_info_i(
   char   * TRANS,
   int    * N,
   int    * BWL,
   int    * BWU,
   int    * NRHS,
   int    * JA,
   int    * DESCA,
   int    * IB,
   int    * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxgelq2
*/
int hscal_pcgelq2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdgelq2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psgelq2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzgelq2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxlarfc
*/
int hscal_pclarfc_info_i(
   F_CHAR_T SIDE,
   int * M,
   int * N,
   int * IV,
   int * JV,
   int * DESCV,
   int * INCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzlarfc_info_i(
   F_CHAR_T SIDE,
   int * M,
   int * N,
   int * IV,
   int * JV,
   int * DESCV,
   int * INCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxunmlq
*/
int hscal_pcunmlq_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzunmlq_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxgeqrf
*/
int hscal_pcgeqrf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdgeqrf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psgeqrf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzgeqrf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxorml2
*/
int hscal_pdorml2_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psorml2_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxgeqr2
*/
int hscal_pcgeqr2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdgeqr2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow, 
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psgeqr2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzgeqr2_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxorm2r
*/
int hscal_pdorm2r_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psorm2r_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxgels
*/
int hscal_pcgels_info_i(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdgels_info_i(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psgels_info_i(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzgels_info_i(
   char * TRANS,
   int * M,
   int * N,
   int * NRHS,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxlarf
*/
int hscal_pclarf_info_i(
   F_CHAR_T SIDE,
   int * M,
   int * N,
   int * IV,
   int * JV,
   int * DESCV,
   int * INCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdlarf_info_i(
   F_CHAR_T SIDE,
   int * M,
   int * N,
   int * IV,
   int * JV,
   int * DESCV,
   int * INCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pslarf_info_i(
   F_CHAR_T SIDE,
   int * M,
   int * N,
   int * IV,
   int * JV,
   int * DESCV,
   int * INCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzlarf_info_i(
   F_CHAR_T SIDE,
   int * M,
   int * N,
   int * IV,
   int * JV,
   int * DESCV,
   int * INCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxgelqf
*/
int hscal_pcgelqf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdgelqf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psgelqf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzgelqf_info_i(
   int * M,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxlarfg
*/
int hscal_pclarfg_info_i(
   int * N,
   int * IAX,
   int * JAX,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdlarfg_info_i(
   int * N,
   int * IAX,
   int * JAX,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pslarfg_info_i(
   int * N,
   int * IAX,
   int * JAX,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzlarfg_info_i(
   int * N,
   int * IAX,
   int * JAX,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxunmqr
*/
int hscal_pcunmqr_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzunmqr_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxlarfb
*/
int hscal_pclarfb_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * M,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdlarfb_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * M,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pslarfb_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * M,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzlarfb_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * M,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxormlq
*/
int hscal_pdormlq_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psormlq_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxlarft
*/
int hscal_pclarft_info_i(
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pdlarft_info_i(
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pslarft_info_i(
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzlarft_info_i(
   F_CHAR_T DIRECT,
   F_CHAR_T STOREV,
   int * N,
   int * K,
   int * IV,
   int * JV,
   int * DESCV,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxormqr
*/
int hscal_pdormqr_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_psormqr_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxunm2r
*/
int hscal_pcunm2r_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzunm2r_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

/*
* pxunml2
*/
int hscal_pcunml2_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

int hscal_pzunml2_info_i(
   F_CHAR_T SIDE,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   int * K,
   int * IA,
   int * JA,
   int * DESCA,
   int * IC,
   int * JC,
   int * DESCC,
   int   nprow,
   int   npcol,
   double * tcomp,
   int    * tcomm
);

#endif

