
/*******************************************************************
*                                                                  *
*                       Heterogeneous ScaLAPACK                    *
*                                TOOLS                             *
*                             (INTERNAL)                           *
*                 ====================================             *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_TOOLS_HH
#define __HMPI_SCALAPACK_TOOLS_HH

int hscal_pchk1mat(
         hscalinteger *ma,
         hscalinteger *mapos0,
         hscalinteger *na,
         hscalinteger *napos0,
         hscalinteger *ia,
         hscalinteger *ja,
         hscalinteger *desca,
         hscalinteger *descapos0,
         hscalinteger *nextra,
         hscalinteger *ex,
         hscalinteger *expos,
         hscalinteger *info,
         hscalinteger *myrow,
         hscalinteger *mycol,
         hscalinteger *nprow,
         hscalinteger *npcol
);

int hscal_pchk2mat(
         hscalinteger *ma,
         hscalinteger *mapos0,
         hscalinteger *na,
         hscalinteger *napos0,
         hscalinteger *ia,
         hscalinteger *ja,
         hscalinteger *desca,
         hscalinteger *descapos0,
         hscalinteger *mb,
         hscalinteger *mbpos0,
         hscalinteger *nb,
         hscalinteger *nbpos0,
         hscalinteger *ib,
         hscalinteger *jb,
         hscalinteger *descb,
         hscalinteger *descbpos0,
         hscalinteger *nextra,
         hscalinteger *ex,
         hscalinteger *expos,
         hscalinteger *info,
         hscalinteger *myrow,
         hscalinteger *mycol,
         hscalinteger *nprow,
         hscalinteger *npcol
);

int hscal_globchk(
         hscalinteger *ictxt,
         hscalinteger *n,
         hscalinteger *x,
         hscalinteger *ldx,
         hscalinteger *iwork,
         hscalinteger *info,
         hscalinteger *myrow,
         hscalinteger *mycol,
         hscalinteger *nprow,
         hscalinteger *npcol
);

int hscal_picol2row(
         hscalinteger *ictxt,
         hscalinteger *m,
         hscalinteger *n,
         hscalinteger *nb,
         hscalinteger *ldvs,
         hscalinteger *ldvd,
         hscalinteger *rsrc,
         hscalinteger *csrc,
         hscalinteger *rdest,
         hscalinteger *cdest,
         int           nprow,
         int           npcol,
         int           myrow,
         int           mycol,
         double        *tcomp,
         int           *tcomm
);

int hscal_pirow2col(
         hscalinteger *ictxt,
         hscalinteger *m,
         hscalinteger *n,
         hscalinteger *nb,
         hscalinteger *ldvs,
         hscalinteger *ldvd,
         hscalinteger *rsrc,
         hscalinteger *csrc,
         hscalinteger *rdest,
         hscalinteger *cdest,
         int nprow,
         int npcol,
         int myrow,
         int mycol,
         double *tcomp,
         int *tcomm
);

hscalinteger hscal_ilcm(
         hscalinteger *m, 
         hscalinteger *n
);

int hscal_ilacpy(
         char *uplo,
         hscalinteger *m,
         hscalinteger *n,
         hscalinteger *lda,
         hscalinteger *ldb,
         double *tcomp
);

#define ilacpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ilacpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_indicative_complex(
         hscalcomplex *hc,
         hscalinteger *ihc
);

int hscal_indicative_doublecomplex(
         hscaldoublecomplex *hdc,
         hscalinteger *ihdc
);

int hscal_chk1mat(
         hscalinteger *ma,
         hscalinteger *mapos0,
         hscalinteger *na,
         hscalinteger *napos0,
         hscalinteger *ia,
         hscalinteger *ja,
         hscalinteger *desca,
         hscalinteger *descapos0,
         hscalinteger *info,
         hscalinteger myrow,
         hscalinteger mycol,
         hscalinteger nprow,
         hscalinteger npcol
);

int hscal_cmatadd(
    int *m,
    int *n,
    char *calpha,
    hscalinteger *lda,
    char *cbeta,
    hscalinteger *ldc,
    double *tcomp
);

int hscal_dmatadd(
    int *m,
    int *n,
    char *calpha,
    hscalinteger *lda,
    char *cbeta,
    hscalinteger *ldc,
    double *tcomp
);

int hscal_smatadd(
    int *m,
    int *n,
    char *calpha,
    hscalinteger *lda,
    char *cbeta,
    hscalinteger *ldc,
    double *tcomp
);

int hscal_zmatadd(
    int *m,
    int *n,
    char *calpha,
    hscalinteger *lda,
    char *cbeta,
    hscalinteger *ldc,
    double *tcomp
);

int hscal_descinit(
    hscalinteger *desc, 
    hscalinteger *m,  
    hscalinteger *n, 
    hscalinteger *mb, 
    hscalinteger *nb, 
    hscalinteger *irsrc, 
    hscalinteger *icsrc, 
    hscalinteger *info
);

int hscal_pcdbtrsv(
    char *uplo, char *trans, hscalinteger *n, hscalinteger *bwl, hscalinteger *bwu, 
    hscalinteger *nrhs, hscalinteger *ja, 
    hscalinteger *desca, hscalinteger *ib, hscalinteger *descb, 
    hscalinteger *info,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pddbtrsv(
    char *uplo, char *trans, hscalinteger *n, hscalinteger *bwl, hscalinteger *bwu, 
    hscalinteger *nrhs, hscalinteger *ja, 
    hscalinteger *desca, hscalinteger *ib, hscalinteger *descb, 
    hscalinteger *info,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_psdbtrsv(
    char *uplo, char *trans, hscalinteger *n, hscalinteger *bwl, hscalinteger *bwu, 
    hscalinteger *nrhs, hscalinteger *ja, 
    hscalinteger *desca, hscalinteger *ib, hscalinteger *descb, 
    hscalinteger *info,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pzdbtrsv(
    char *uplo, char *trans, hscalinteger *n, hscalinteger *bwl, hscalinteger *bwu, 
    hscalinteger *nrhs, hscalinteger *ja, 
    hscalinteger *desca, hscalinteger *ib, hscalinteger *descb, 
    hscalinteger *info,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pclacgv(
    hscalinteger *n,  hscalinteger *ix, hscalinteger *jx, hscalinteger *descx, hscalinteger *incx,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pzlacgv(
    hscalinteger *n,  hscalinteger *ix, hscalinteger *jx, hscalinteger *descx, hscalinteger *incx,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pclascl(
    char *type__,
    hscalinteger *m, hscalinteger *n,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pdlascl(
    char *type__,
    hscalinteger *m, hscalinteger *n,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pslascl(
    char *type__,
    hscalinteger *m, hscalinteger *n,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pzlascl(
    char *type__,
    hscalinteger *m, hscalinteger *n,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pclange(
    char *norm, hscalinteger *m, hscalinteger *n,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pdlange(
    char *norm, hscalinteger *m, hscalinteger *n,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pslange(
    char *norm, hscalinteger *m, hscalinteger *n,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pzlange(
    char *norm, hscalinteger *m, hscalinteger *n,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pclase2(
    char *uplo, hscalinteger *m, hscalinteger *n, hscalcomplex * alpha, hscalcomplex *beta,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pdlase2(
    char *uplo, hscalinteger *m, hscalinteger *n, hscaldoublereal* alpha, hscaldoublereal *beta,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pslase2(
    char *uplo, hscalinteger *m, hscalinteger *n, hscalreal* alpha, hscalreal* beta,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pzlase2(
    char *uplo, hscalinteger *m, hscalinteger *n, hscaldoublecomplex * alpha, hscaldoublecomplex *beta,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pclaset(
    char *uplo, hscalinteger *m, hscalinteger *n, hscalcomplex * alpha, hscalcomplex *beta,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pdlaset(
    char *uplo, hscalinteger *m, hscalinteger *n, hscaldoublereal* alpha, hscaldoublereal *beta,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pslaset(
    char *uplo, hscalinteger *m, hscalinteger *n, hscalreal* alpha, hscalreal* beta,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pzlaset(
    char *uplo, hscalinteger *m, hscalinteger *n, hscaldoublecomplex * alpha, hscaldoublecomplex *beta,
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

typedef void           (*HSCAL_S_fp)   ( int, int,
                                         int, int, 
                                         double*);

int hscal_pdtreecomb(
    hscalinteger *ictxt, char *scope, hscalinteger *n, 
    hscalinteger *rdest0, hscalinteger *cdest0, HSCAL_S_fp subptr,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_dcombamax(
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_dcombssq(
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_dcombnrm2(
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pstreecomb(
    hscalinteger *ictxt, char *scope, hscalinteger *n,
    hscalinteger *rdest0, hscalinteger *cdest0, HSCAL_S_fp subptr,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_scombamax(
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_scombssq( 
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_scombnrm2(
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pcelset(
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca, hscalcomplex *alpha,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pdelset(
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca, hscaldoublereal *alpha,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pielset(
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca, hscalinteger *alpha,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pselset(
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca, hscalreal *alpha,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pzelset(
    hscalinteger *ia, hscalinteger *ja, hscalinteger *desca, hscaldoublecomplex *alpha,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pcelset2(
    hscalcomplex *alpha, hscalinteger *ia,
    hscalinteger *ja, hscalinteger *desca, hscalcomplex *beta,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pdelset2(
    hscaldoublereal *alpha, hscalinteger *ia,
    hscalinteger *ja, hscalinteger *desca, hscaldoublereal *beta,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pielset2(
    hscalinteger *alpha, hscalinteger *ia,
    hscalinteger *ja, hscalinteger *desca, hscalinteger *beta,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pselset2(
    hscalreal *alpha, hscalinteger *ia,
    hscalinteger *ja, hscalinteger *desca, hscalreal *beta,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

int hscal_pzelset2(
    hscaldoublecomplex *alpha, hscalinteger *ia,
    hscalinteger *ja, hscalinteger *desca, hscaldoublecomplex *beta,
    int nprow, int npcol, int myrow, int mycol,
    double *tcomp, int *tcomm
);

#endif

