
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK CBLAS MACROS             *
*                         (INTERNAL)                               *
*                 ====================================             *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_CBLAS_MACROS_INCLUDE_HH
#define __HMPI_SCALAPACK_CBLAS_MACROS_INCLUDE_HH

/*
* ------------------------------------------------------------------
*       FUNCTION PROTOTYPES and mpC MACROS FOR
*       CBLAS COMPUTATIONS
* ------------------------------------------------------------------
*/
int hscal_caxpy(
    int *n,
    char *ca,
    int *incx,
    int *incy,
    double *tcomp
);

#define caxpy_tcomp_macro( \
             n, ca, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_caxpy( \
         (n), (ca), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_daxpy(
    int *n,
    char *da,
    int *incx,
    int *incy,
    double *tcomp
);

#define daxpy_tcomp_macro( \
             n, da, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_daxpy( \
         (n), (da), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_saxpy(
    int *n,
    char *sa,
    int *incx,
    int *incy,
    double *tcomp
);

#define saxpy_tcomp_macro( \
             n, sa, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_saxpy( \
         (n), (sa), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zaxpy(
    int *n,
    char *za,
    int *incx,
    int *incy,
    double *tcomp
);

#define zaxpy_tcomp_macro( \
             n, za, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zaxpy( \
         (n), (za), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ccopy(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define ccopy_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ccopy( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dcopy(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define dcopy_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dcopy( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_scopy(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define scopy_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_scopy( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zcopy(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define zcopy_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zcopy( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_chemv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta,
    int *incy,
    double *tcomp
);

#define chemv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_chemv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cahemv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta,
    int *incy,
    double *tcomp
);

#define cahemv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cahemv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zhemv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta,
    int *incy,
    double *tcomp
);

#define zhemv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zhemv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zahemv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta,
    int *incy,
    double *tcomp
);

#define zahemv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zahemv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cgemv(
    char *trans, 
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define cgemv_tcomp_macro( \
             trans, m, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cgemv( \
         (trans), (m), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cagemv(
    char *trans, 
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define cagemv_tcomp_macro( \
             trans, m, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cagemv( \
         (trans), (m), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dgemv(
    char *trans, 
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    int *incx, 
    char *beta,
    int *incy,
    double *tcomp
);

#define dgemv_tcomp_macro( \
             trans, m, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dgemv( \
         (trans), (m), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dagemv(
    char *trans, 
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    int *incx, 
    char *beta,
    int *incy,
    double *tcomp
);

#define dagemv_tcomp_macro( \
             trans, m, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dagemv( \
         (trans), (m), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dsymv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
);

#define dsymv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dsymv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ssymv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
);

#define ssymv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ssymv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sgemv(
    char *trans, 
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define sgemv_tcomp_macro( \
             trans, m, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sgemv( \
         (trans), (m), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sagemv(
    char *trans, 
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define sagemv_tcomp_macro( \
             trans, m, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sagemv( \
         (trans), (m), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zgemv(
    char *trans, 
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define zgemv_tcomp_macro( \
             trans, m, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zgemv( \
         (trans), (m), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zagemv(
    char *trans, 
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define zagemv_tcomp_macro( \
             trans, m, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zagemv( \
         (trans), (m), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cscal(
    int *n,
    char *ca,
    int *incx,
    double *tcomp
);

#define cscal_tcomp_macro( \
             n, ca, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cscal( \
         (n), (ca), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dscal(
    int *n,
    char *da,
    int *incx,
    double *tcomp
);

#define dscal_tcomp_macro( \
             n, da, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dscal( \
         (n), (da), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dascal(
    int *n,
    char *da,
    int *incx,
    double *tcomp
);

#define dascal_tcomp_macro( \
             n, da, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dascal( \
         (n), (da), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sscal(
    int *n,
    char *sa,
    int *incx,
    double *tcomp
);

#define sscal_tcomp_macro( \
             n, sa, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sscal( \
         (n), (sa), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sascal(
    int *n,
    char *sa,
    int *incx,
    double *tcomp
);

#define sascal_tcomp_macro( \
             n, sa, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sascal( \
         (n), (sa), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zscal(
    int *n,
    char *za,
    int *incx,
    double *tcomp
);

#define zscal_tcomp_macro( \
             n, za, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zscal( \
         (n), (za), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctrmv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *lda, 
    int *incx,
    double *tcomp
);

#define ctrmv_tcomp_macro( \
             uplo, trans, diag, n, lda, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctrmv( \
         (uplo), (trans), (diag), (n), (lda), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtrmv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *lda, 
    int *incx,
    double *tcomp
);

#define dtrmv_tcomp_macro( \
             uplo, trans, diag, n, lda, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtrmv( \
         (uplo), (trans), (diag), (n), (lda), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_strmv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *lda, 
    int *incx,
    double *tcomp
);

#define strmv_tcomp_macro( \
             uplo, trans, diag, n, lda, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_strmv( \
         (uplo), (trans), (diag), (n), (lda), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztrmv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *lda, 
    int *incx,
    double *tcomp
);

#define ztrmv_tcomp_macro( \
             uplo, trans, diag, n, lda, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztrmv( \
         (uplo), (trans), (diag), (n), (lda), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctrsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *lda, 
    int *incx,
    double *tcomp
);

#define ctrsv_tcomp_macro( \
             uplo, trans, diag, n, lda, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctrsv( \
         (uplo), (trans), (diag), (n), (lda), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtrsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *lda, 
    int *incx,
    double *tcomp
);

#define dtrsv_tcomp_macro( \
             uplo, trans, diag, n, lda, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtrsv( \
         (uplo), (trans), (diag), (n), (lda), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_strsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *lda, 
    int *incx,
    double *tcomp
);

#define strsv_tcomp_macro( \
             uplo, trans, diag, n, lda, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_strsv( \
         (uplo), (trans), (diag), (n), (lda), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztrsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *lda, 
    int *incx,
    double *tcomp
);

#define ztrsv_tcomp_macro( \
             uplo, trans, diag, n, lda, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztrsv( \
         (uplo), (trans), (diag), (n), (lda), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}


int hscal_cgerc(
    int *m, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy,
    int *lda, 
    double *tcomp
);

#define cgerc_tcomp_macro( \
             m, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cgerc( \
         (m), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}


int hscal_zgerc(
    int *m, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy,
    int *lda, 
    double *tcomp
);

#define zgerc_tcomp_macro( \
             m, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zgerc( \
         (m), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cgeru(
    int *m, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy,
    int *lda, 
    double *tcomp
);

#define cgeru_tcomp_macro( \
             m, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cgeru( \
         (m), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zgeru(
    int *m, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy,
    int *lda, 
    double *tcomp
);

#define zgeru_tcomp_macro( \
             m, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zgeru( \
         (m), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dger(
    int *m, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy,
    int *lda, 
    double *tcomp
);

#define dger_tcomp_macro( \
             m, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dger( \
         (m), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sger(
    int *m, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy,
    int *lda, 
    double *tcomp
);

#define sger_tcomp_macro( \
             m, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sger( \
         (m), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cher(
    char *uplo,
    int  *n, 
    char *calpha,
    int  *incx, 
    int  *lda, 
    double *tcomp
);

#define cher_tcomp_macro( \
             uplo, n, calpha, incx, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cher( \
         (uplo), (n), (calpha), (incx), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zher(
    char *uplo,
    int  *n, 
    char *calpha,
    int  *incx, 
    int  *lda, 
    double *tcomp
);

#define zher_tcomp_macro( \
             uplo, n, calpha, incx, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zher( \
         (uplo), (n), (calpha), (incx), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cher2(
    char *uplo,
    int  *n, 
    char *calpha,
    int  *incx, 
    int  *incy, 
    int  *lda, 
    double *tcomp
);

#define cher2_tcomp_macro( \
             uplo, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cher2( \
         (uplo), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zher2(
    char *uplo,
    int  *n, 
    char *calpha,
    int  *incx, 
    int  *incy, 
    int  *lda, 
    double *tcomp
);

#define zher2_tcomp_macro( \
             uplo, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zher2( \
         (uplo), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cgemm(
    char   *transa, 
    char   *transb, 
    int    *m, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb, 
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define cgemm_tcomp_macro( \
             transa, transb, m, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cgemm( \
         (transa), (transb), (m), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dgemm(
    char   *transa, 
    char   *transb, 
    int    *m, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb, 
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define dgemm_tcomp_macro( \
             transa, transb, m, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dgemm( \
         (transa), (transb), (m), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sgemm(
    char   *transa, 
    char   *transb, 
    int    *m, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb, 
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define sgemm_tcomp_macro( \
             transa, transb, m, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sgemm( \
         (transa), (transb), (m), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}


int hscal_zgemm(
    char   *transa, 
    char   *transb, 
    int    *m, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb, 
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define zgemm_tcomp_macro( \
             transa, transb, m, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zgemm( \
         (transa), (transb), (m), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dsyr(
    char *uplo, 
    int *n, 
    char *calpha,
    int *incx, 
    int *lda, 
    double *tcomp
);

#define dsyr_tcomp_macro( \
             uplo, n, calpha, incx, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dsyr( \
         (uplo), (n), (calpha), (incx), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ssyr(
    char *uplo, 
    int *n, 
    char *calpha,
    int *incx, 
    int *lda, 
    double *tcomp
);

#define ssyr_tcomp_macro( \
             uplo, n, calpha, incx, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ssyr( \
         (uplo), (n), (calpha), (incx), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dsyr2(
    char *uplo, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy, 
    int *lda, 
    double *tcomp
);

#define dsyr2_tcomp_macro( \
             uplo, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dsyr2( \
         (uplo), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ssyr2(
    char *uplo, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy, 
    int *lda, 
    double *tcomp
);

#define ssyr2_tcomp_macro( \
             uplo, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ssyr2( \
         (uplo), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_csyrk(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define csyrk_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_csyrk( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dsyrk(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define dsyrk_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dsyrk( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ssyrk(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define ssyrk_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ssyrk( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zsyrk(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define zsyrk_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zsyrk( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_csyr2k(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define csyr2k_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_csyr2k( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dsyr2k(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define dsyr2k_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dsyr2k( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ssyr2k(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define ssyr2k_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ssyr2k( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zsyr2k(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define zsyr2k_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zsyr2k( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cherk(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define cherk_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cherk( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zherk(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define zherk_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zherk( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cher2k(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta,
    int    *ldc, 
    double *tcomp
);

#define cher2k_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cher2k( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zher2k(
    char   *uplo, 
    char   *trans, 
    int    *n, 
    int    *k, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta,
    int    *ldc, 
    double *tcomp
);

#define zher2k_tcomp_macro( \
             uplo, trans, n, k, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zher2k( \
         (uplo), (trans), (n), (k), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_csscal(
    int *n,
    char *ca,
    int *incx,
    double *tcomp
);

#define csscal_tcomp_macro( \
             n, ca, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_csscal( \
         (n), (ca), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zdscal(
    int *n,
    char *ca,
    int *incx,
    double *tcomp
);

#define zdscal_tcomp_macro( \
             n, ca, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zdscal( \
         (n), (ca), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cswap(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define cswap_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cswap( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dswap(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define dswap_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dswap( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sswap(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define sswap_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sswap( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zswap(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define zswap_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zswap( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dasum(
    int *n,
    int *incx,
    double *tcomp
);

#define dasum_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dasum( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dzasum(
    int *n,
    int *incx,
    double *tcomp
);

#define dzasum_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dzasum( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sasum(
    int *n,
    int *incx,
    double *tcomp
);

#define sasum_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sasum( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_scasum(
    int *n,
    int *incx,
    double *tcomp
);

#define scasum_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_scasum( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dnrm2(
    int *n,
    int *incx,
    double *tcomp
);

#define dnrm2_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dnrm2( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dznrm2(
    int *n,
    int *incx,
    double *tcomp
);

#define dznrm2_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dznrm2( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_snrm2(
    int *n,
    int *incx,
    double *tcomp
);

#define snrm2_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_snrm2( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_scnrm2(
    int *n,
    int *incx,
    double *tcomp
);

#define scnrm2_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_scnrm2( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_icamax(
    int *n,
    int *incx,
    double *tcomp
);

#define icamax_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_icamax( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_idamax(
    int *n,
    int *incx,
    double *tcomp
);

#define idamax_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_idamax( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_isamax(
    int *n,
    int *incx,
    double *tcomp
);

#define isamax_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_isamax( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_izamax(
    int *n,
    int *incx,
    double *tcomp
);

#define izamax_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_izamax( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cdotc(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define cdotc_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cdotc( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cdotu(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define cdotu_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cdotu( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zdotc(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define zdotc_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zdotc( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zdotu(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define zdotu_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zdotu( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ddot(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define ddot_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ddot( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sdot(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define sdot_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sdot( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

#define FDOT_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    FDOT( \
      (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_csymm(
    char   *side, 
    char   *uplo, 
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define csymm_tcomp_macro( \
             side, uplo, m, n, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_csymm( \
         (side), (uplo), (m), (n), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dsymm(
    char   *side, 
    char   *uplo, 
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define dsymm_tcomp_macro( \
             side, uplo, m, n, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dsymm( \
         (side), (uplo), (m), (n), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ssymm(
    char   *side, 
    char   *uplo, 
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define ssymm_tcomp_macro( \
             side, uplo, m, n, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ssymm( \
         (side), (uplo), (m), (n), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zsymm(
    char   *side, 
    char   *uplo, 
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define zsymm_tcomp_macro( \
             side, uplo, m, n, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zsymm( \
         (side), (uplo), (m), (n), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_chemm(
    char   *side, 
    char   *uplo, 
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define chemm_tcomp_macro( \
             side, uplo, m, n, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_chemm( \
         (side), (uplo), (m), (n), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zhemm(
    char   *side, 
    char   *uplo, 
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    char   *cbeta, 
    int    *ldc,
    double *tcomp
);

#define zhemm_tcomp_macro( \
             side, uplo, m, n, calpha, \
             lda, ldb, cbeta, ldc, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zhemm( \
         (side), (uplo), (m), (n), \
         (calpha), (lda), (ldb), (cbeta), (ldc), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctrmm(
    char   *side, 
    char   *uplo, 
    char   *transa, 
    char   *diag,
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    double *tcomp
);

#define ctrmm_tcomp_macro( \
             side, uplo, transa, diag, m, n, calpha, \
             lda, ldb, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctrmm( \
         (side), (uplo), (transa), (diag), (m), (n), \
         (calpha), (lda), (ldb), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtrmm(
    char   *side, 
    char   *uplo, 
    char   *transa, 
    char   *diag,
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    double *tcomp
);

#define dtrmm_tcomp_macro( \
             side, uplo, transa, diag, m, n, calpha, \
             lda, ldb, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtrmm( \
         (side), (uplo), (transa), (diag), (m), (n), \
         (calpha), (lda), (ldb), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_strmm(
    char   *side, 
    char   *uplo, 
    char   *transa, 
    char   *diag,
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    double *tcomp
);

#define strmm_tcomp_macro( \
             side, uplo, transa, diag, m, n, calpha, \
             lda, ldb, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_strmm( \
         (side), (uplo), (transa), (diag), (m), (n), \
         (calpha), (lda), (ldb), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztrmm(
    char   *side, 
    char   *uplo, 
    char   *transa, 
    char   *diag,
    int    *m, 
    int    *n, 
    char   *calpha, 
    int    *lda,
    int    *ldb,
    double *tcomp
);

#define ztrmm_tcomp_macro( \
             side, uplo, transa, diag, m, n, calpha, \
             lda, ldb, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztrmm( \
         (side), (uplo), (transa), (diag), (m), (n), \
         (calpha), (lda), (ldb), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctrsm(
    char *side, 
    char *uplo, 
    char *transa, 
    char *diag, 
    int *m, 
    int *n, 
    char *alphac,
    int *lda, 
    int *ldb,
    double *tcomp
);

#define ctrsm_tcomp_macro( \
             side, uplo, transa, diag, m, n, alphac, lda, ldb, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctrsm( \
         (side), (uplo), (transa), (diag), (m), (n), (alphac), (lda), \
         (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtrsm(
    char *side, 
    char *uplo, 
    char *transa, 
    char *diag, 
    int *m, 
    int *n, 
    char *alphac,
    int *lda, 
    int *ldb,
    double *tcomp
);

#define dtrsm_tcomp_macro( \
             side, uplo, transa, diag, m, n, alphac, lda, ldb, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtrsm( \
         (side), (uplo), (transa), (diag), (m), (n), (alphac), (lda), \
         (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_strsm(
    char *side, 
    char *uplo, 
    char *transa, 
    char *diag, 
    int *m, 
    int *n, 
    char *alphac,
    int *lda, 
    int *ldb,
    double *tcomp
);

#define strsm_tcomp_macro( \
             side, uplo, transa, diag, m, n, alphac, lda, ldb, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_strsm( \
         (side), (uplo), (transa), (diag), (m), (n), (alphac), (lda), \
         (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztrsm(
    char *side, 
    char *uplo, 
    char *transa, 
    char *diag, 
    int *m, 
    int *n, 
    char *alphac,
    int *lda, 
    int *ldb,
    double *tcomp
);

#define ztrsm_tcomp_macro( \
             side, uplo, transa, diag, m, n, alphac, lda, ldb, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztrsm( \
         (side), (uplo), (transa), (diag), (m), (n), (alphac), (lda), \
         (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

#endif  /* __HMPI_SCALAPACK_CBLAS_MACROS_INCLUDE_HH */

