
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK CLAPACK MACROS           *
*                         (INTERNAL)                               *
*                 ====================================             *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_CLAPACK_MACROS_INCLUDE_HH
#define __HMPI_SCALAPACK_CLAPACK_MACROS_INCLUDE_HH

int hscal_clacgv(
    int *n,
    int *incx,
    double *tcomp
);

#define clacgv_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_clacgv( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zlacgv(
    int *n,
    int *incx,
    double *tcomp
);

#define zlacgv_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zlacgv( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_clatcpy(
    char *uplo,
    int *m,
    int *n,
    int *lda,
    int *ldb,
    double *tcomp
);

#define clatcpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_clatcpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dlatcpy(
    char *uplo,
    int *m,
    int *n,
    int *lda,
    int *ldb,
    double *tcomp
);

#define dlatcpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dlatcpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_slatcpy(
    char *uplo,
    int *m,
    int *n,
    int *lda,
    int *ldb,
    double *tcomp
);

#define slatcpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_slatcpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zlatcpy(
    char *uplo,
    int *m,
    int *n,
    int *lda,
    int *ldb,
    double *tcomp
);

#define zlatcpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zlatcpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_clacpy(
    char *uplo,
    int *m,
    int *n,
    int *lda,
    int *ldb,
    double *tcomp
);

#define clacpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_clacpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dlacpy(
    char *uplo,
    int *m,
    int *n,
    int *lda,
    int *ldb,
    double *tcomp
);

#define dlacpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dlacpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_slacpy(
    char *uplo,
    int *m,
    int *n,
    int *lda,
    int *ldb,
    double *tcomp
);

#define slacpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_slacpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zlacpy(
    char *uplo,
    int *m,
    int *n,
    int *lda,
    int *ldb,
    double *tcomp
);

#define zlacpy_tcomp_macro( \
             uplo, m, n, lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zlacpy( \
         (uplo), (m), (n), (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cdbtf2(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
);

#define cdbtf2_tcomp_macro( \
             m, n, kl, ku, ldab, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cdbtf2( \
         (m), (n), (kl), (ku), (ldab), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ddbtf2(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
);

#define ddbtf2_tcomp_macro( \
             m, n, kl, ku, ldab, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ddbtf2( \
         (m), (n), (kl), (ku), (ldab), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sdbtf2(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
);

#define sdbtf2_tcomp_macro( \
             m, n, kl, ku, ldab, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sdbtf2( \
         (m), (n), (kl), (ku), (ldab), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zdbtf2(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
);

#define zdbtf2_tcomp_macro( \
             m, n, kl, ku, ldab, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zdbtf2( \
         (m), (n), (kl), (ku), (ldab), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cdbtrf(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
);

#define cdbtrf_tcomp_macro( \
             m, n, kl, ku, ldab, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cdbtrf( \
         (m), (n), (kl), (ku), (ldab), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ddbtrf(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
);

#define ddbtrf_tcomp_macro( \
             m, n, kl, ku, ldab, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ddbtrf( \
         (m), (n), (kl), (ku), (ldab), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sdbtrf(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
);

#define sdbtrf_tcomp_macro( \
             m, n, kl, ku, ldab, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sdbtrf( \
         (m), (n), (kl), (ku), (ldab), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zdbtrf(
    int *m, 
    int *n, 
    int *kl, 
    int *ku,
    int *ldab, 
    int *info,
    double *tcomp
);

#define zdbtrf_tcomp_macro( \
             m, n, kl, ku, ldab, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zdbtrf( \
         (m), (n), (kl), (ku), (ldab), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctbsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *k, 
    int *lda,
    int *incx, 
    double *tcomp
);

#define ctbsv_tcomp_macro( \
             uplo, trans, diag, n, k, lda, incx, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctbsv( \
         (uplo), (trans), (diag), (n), (k), (lda), (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtbsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *k, 
    int *lda,
    int *incx, 
    double *tcomp
);

#define dtbsv_tcomp_macro( \
             uplo, trans, diag, n, k, lda, incx, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtbsv( \
         (uplo), (trans), (diag), (n), (k), (lda), (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_stbsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *k, 
    int *lda,
    int *incx, 
    double *tcomp
);

#define stbsv_tcomp_macro( \
             uplo, trans, diag, n, k, lda, incx, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_stbsv( \
         (uplo), (trans), (diag), (n), (k), (lda), (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztbsv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *k, 
    int *lda,
    int *incx, 
    double *tcomp
);

#define ztbsv_tcomp_macro( \
             uplo, trans, diag, n, k, lda, incx, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztbsv( \
         (uplo), (trans), (diag), (n), (k), (lda), (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctbtrs(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *kd, 
    int *nrhs, 
    int *ldab,
    int *ldb, 
    int *info,
    double *tcomp
);

#define ctbtrs_tcomp_macro( \
             uplo, trans, diag, n, kd, nrhs, ldab, ldb, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctbtrs( \
         (uplo), (trans), (diag), (n), (kd), (nrhs), (ldab), (ldb), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtbtrs(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *kd, 
    int *nrhs, 
    int *ldab,
    int *ldb, 
    int *info,
    double *tcomp
);

#define dtbtrs_tcomp_macro( \
             uplo, trans, diag, n, kd, nrhs, ldab, ldb, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtbtrs( \
         (uplo), (trans), (diag), (n), (kd), (nrhs), (ldab), (ldb), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_stbtrs(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *kd, 
    int *nrhs, 
    int *ldab,
    int *ldb, 
    int *info,
    double *tcomp
);

#define stbtrs_tcomp_macro( \
             uplo, trans, diag, n, kd, nrhs, ldab, ldb, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_stbtrs( \
         (uplo), (trans), (diag), (n), (kd), (nrhs), (ldab), (ldb), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztbtrs(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    int *kd, 
    int *nrhs, 
    int *ldab,
    int *ldb, 
    int *info,
    double *tcomp
);

#define ztbtrs_tcomp_macro( \
             uplo, trans, diag, n, kd, nrhs, ldab, ldb, info, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztbtrs( \
         (uplo), (trans), (diag), (n), (kd), (nrhs), (ldab), (ldb), (info), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_claset(
    char *uplo, 
    int *m, int *n, 
    char *alpha, char *beta, 
    int *lda,
    double *tcomp
);

#define claset_tcomp_macro( \
             uplo, m, n, alpha, beta, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_claset( \
         (uplo), (m), (n), (alpha), (beta), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dlaset(
    char *uplo, 
    int *m, int *n, 
    char *alpha, char *beta, 
    int *lda,
    double *tcomp
);

#define dlaset_tcomp_macro( \
             uplo, m, n, alpha, beta, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dlaset( \
         (uplo), (m), (n), (alpha), (beta), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_slaset(
    char *uplo, 
    int *m, int *n, 
    char *alpha, char *beta, 
    int *lda,
    double *tcomp
);

#define slaset_tcomp_macro( \
             uplo, m, n, alpha, beta, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_slaset( \
         (uplo), (m), (n), (alpha), (beta), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zlaset(
    char *uplo, 
    int *m, int *n, 
    char *alpha, char *beta, 
    int *lda,
    double *tcomp
);

#define zlaset_tcomp_macro( \
             uplo, m, n, alpha, beta, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zlaset( \
         (uplo), (m), (n), (alpha), (beta), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_classq(
    int *n,
    int *incx,
    double *tcomp
);

int hscal_dlassq(
    int *n,
    int *incx,
    double *tcomp
);

int hscal_slassq(
    int *n,
    int *incx,
    double *tcomp
);

int hscal_zlassq(
    int *n,
    int *incx,
    double *tcomp
);

int hscal_clarfg(
    int *n, hscalcomplex *alpha, int *incx,
    double *tcomp
);

#define clarfg_tcomp_macro( \
             n, alpha, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_clarfg( \
         (n), (alpha), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dlarfg(
    int *n, hscaldoublereal *alpha, int *incx,
    double *tcomp
);

#define dlarfg_tcomp_macro( \
             n, alpha, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dlarfg( \
         (n), (alpha), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_slarfg(
    int *n, hscalreal *alpha, int *incx,
    double *tcomp
);

#define slarfg_tcomp_macro( \
             n, alpha, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_slarfg( \
         (n), (alpha), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zlarfg(
    int *n, hscaldoublecomplex *alpha, int *incx,
    double *tcomp
);

#define zlarfg_tcomp_macro( \
             n, alpha, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zlarfg( \
         (n), (alpha), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

#endif

