

/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK PBBLAS                   *
*                 =============================                    *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_PBBLAS_INCLUDE_HH
#define __HMPI_SCALAPACK_PBBLAS_INCLUDE_HH

int hscal_pbcmatadd(
    hscalinteger *icontxt, char *mode, hscalinteger *m,
    hscalinteger *n, hscalcomplex *alpha, hscalinteger *lda, hscalcomplex *beta,
    hscalinteger *ldb,
    double *tcomp
);

#define pbcmatadd_tcomp_macro( \
             icontxt, mode, m, n, alpha, lda, beta, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_pbcmatadd( \
         (icontxt), (mode), (m), (n), (alpha), (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_pbdmatadd(
    hscalinteger *icontxt, char *mode, hscalinteger *m,
    hscalinteger *n, hscaldoublereal *alpha, hscalinteger *lda,
    hscaldoublereal *beta, hscalinteger *ldb,
    double *tcomp
);

#define pbdmatadd_tcomp_macro( \
             icontxt, mode, m, n, alpha, lda, beta, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_pbdmatadd( \
         (icontxt), (mode), (m), (n), (alpha), (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_pbsmatadd(
    hscalinteger *icontxt, char *mode, hscalinteger *m,
    hscalinteger *n, hscalreal *alpha, hscalinteger *lda,
    hscalreal *beta, hscalinteger *ldb,
    double *tcomp
);

#define pbsmatadd_tcomp_macro( \
             icontxt, mode, m, n, alpha, lda, beta, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_pbsmatadd( \
         (icontxt), (mode), (m), (n), (alpha), (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_pbzmatadd(
    hscalinteger *icontxt, char *mode, hscalinteger *m,
    hscalinteger *n, hscaldoublecomplex *alpha, hscalinteger *lda,
    hscaldoublecomplex *beta, hscalinteger *ldb,
    double *tcomp
);

#define pbzmatadd_tcomp_macro( \
             icontxt, mode, m, n, alpha, lda, beta, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_pbzmatadd( \
         (icontxt), (mode), (m), (n), (alpha), (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_pbcvecadd(
    hscalinteger *icontxt, char *mode, hscalinteger *n,
    hscalcomplex *alpha, hscalinteger *incx, hscalcomplex *beta,
    hscalinteger *incy,
    double *tcomp
);

#define pbcvecadd_tcomp_macro( \
             icontxt, mode, n, alpha, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_pbcvecadd( \
         (icontxt), (mode), (n), (alpha), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_pbdvecadd(
    hscalinteger *icontxt, char *mode, hscalinteger *n,
    hscaldoublereal *alpha, hscalinteger *incx, hscaldoublereal *beta,
    hscalinteger *incy,
    double *tcomp
);

#define pbdvecadd_tcomp_macro( \
             icontxt, mode, n, alpha, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_pbdvecadd( \
         (icontxt), (mode), (n), (alpha), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_pbsvecadd(
    hscalinteger *icontxt, char *mode, hscalinteger *n,
    hscalreal *alpha, hscalinteger *incx, hscalreal *beta, 
    hscalinteger *incy,
    double *tcomp
);

#define pbsvecadd_tcomp_macro( \
             icontxt, mode, n, alpha, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_pbsvecadd( \
         (icontxt), (mode), (n), (alpha), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_pbzvecadd(
    hscalinteger *icontxt, char *mode, hscalinteger *n,
    hscaldoublecomplex *alpha, hscalinteger *incx, hscaldoublecomplex *beta,
    hscalinteger *incy,
    double *tcomp
);

#define pbzvecadd_tcomp_macro( \
             icontxt, mode, n, alpha, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_pbzvecadd( \
         (icontxt), (mode), (n), (alpha), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_pbctrnv(
    hscalinteger *icontxt, char *xdist, char *trans,
    hscalinteger *n, hscalinteger *nb, hscalinteger *nz, hscalinteger *incx,
    hscalcomplex *beta, hscalinteger *incy, hscalinteger *ixrow,
    hscalinteger *ixcol, hscalinteger *iyrow, hscalinteger *iycol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbctr2a1(
    hscalinteger *icontxt, hscalinteger *n, hscalinteger *nb,
    hscalinteger *nz, hscalinteger *incx, hscalcomplex *beta,
    hscalinteger *incy, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbctr2b1(
    hscalinteger *icontxt, char *trans, hscalinteger *n,
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, hscalcomplex *beta,
    hscalinteger *incy, hscalinteger *jinx, hscalinteger *jiny,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbdtrnv(
    hscalinteger *icontxt, char *xdist, char *trans,
    hscalinteger *n, hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, 
    hscaldoublereal *beta, hscalinteger *incy, hscalinteger *ixrow,
    hscalinteger *ixcol, hscalinteger *iyrow, hscalinteger *iycol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbdtr2a1(
    hscalinteger *icontxt, hscalinteger *n, hscalinteger *nb,
    hscalinteger *nz, hscalinteger *incx, hscaldoublereal *beta,
    hscalinteger *incy, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbdtr2b1(
    hscalinteger *icontxt, char *trans, hscalinteger *n,
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx,
    hscaldoublereal *beta, hscalinteger *incy, hscalinteger *jinx, hscalinteger *jiny,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbstrnv(
    hscalinteger *icontxt, char *xdist, char *trans,
    hscalinteger *n, hscalinteger *nb, hscalinteger *nz, hscalinteger *incx,
    hscalreal *beta, hscalinteger *incy, hscalinteger *ixrow,
    hscalinteger *ixcol, hscalinteger *iyrow, hscalinteger *iycol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbstr2a1(
    hscalinteger *icontxt, hscalinteger *n, hscalinteger *nb,
    hscalinteger *nz, hscalinteger *incx, hscalreal *beta,
    hscalinteger *incy, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbstr2b1(
    hscalinteger *icontxt, char *trans, hscalinteger *n,
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx,
    hscalreal *beta, hscalinteger *incy, hscalinteger *jinx, hscalinteger *jiny,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbztrnv(
    hscalinteger *icontxt, char *xdist, char *trans,
    hscalinteger *n, hscalinteger *nb, hscalinteger *nz, hscalinteger *incx,
    hscaldoublecomplex *beta, hscalinteger *incy, hscalinteger *ixrow,
    hscalinteger *ixcol, hscalinteger *iyrow, hscalinteger *iycol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbztr2a1(
    hscalinteger *icontxt, hscalinteger *n, hscalinteger *nb,
    hscalinteger *nz, hscalinteger *incx, hscaldoublecomplex *beta,
    hscalinteger *incy, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbztr2b1(
    hscalinteger *icontxt, char *trans, hscalinteger *n,
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, hscaldoublecomplex *beta,
    hscalinteger *incy, hscalinteger *jinx, hscalinteger *jiny,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm 
);

int hscal_pbctran(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscalcomplex *beta, hscalinteger *ldc, hscalinteger *iarow,
    hscalinteger *iacol, hscalinteger *icrow, hscalinteger *iccol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbctr2at(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscalcomplex *beta, hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbctr2bt(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda, 
    hscalcomplex *beta, hscalinteger *ldb, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbctr2af(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda, hscalcomplex *beta,
    hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbdtran(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscaldoublereal *beta, hscalinteger *ldc, hscalinteger *iarow,
    hscalinteger *iacol, hscalinteger *icrow, hscalinteger *iccol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbdtr2at(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscaldoublereal *beta, hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbdtr2bt(
    hscalinteger *icontxt, char *adist, char *trans, 
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscaldoublereal *beta, hscalinteger *ldb, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbdtr2af(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscaldoublereal *beta, hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq,
    hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbstran(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscalreal *beta, hscalinteger *ldc, hscalinteger *iarow,
    hscalinteger *iacol, hscalinteger *icrow, hscalinteger *iccol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbstr2at(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscalreal *beta, hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbstr2bt(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscalreal *beta, hscalinteger *ldb, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbstr2af( 
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscalreal *beta, hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq,
    hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbztran(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscaldoublecomplex *beta, hscalinteger *ldc, hscalinteger *iarow,
    hscalinteger *iacol, hscalinteger *icrow, hscalinteger *iccol,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbztr2at(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda,
    hscaldoublecomplex *beta, hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbztr2bt(
    hscalinteger *icontxt, char *adist, char *trans,
    hscalinteger *m, hscalinteger *n, hscalinteger *nb, hscalinteger *lda, 
    hscaldoublecomplex *beta, hscalinteger *ldb, hscalinteger *intv,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbztr2af(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda, hscaldoublecomplex *beta,
    hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp,
    int *tcomm
);

int hscal_pbctrget(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *mnb, hscalinteger *lda, hscalinteger *mcrow,
    hscalinteger *mccol, hscalinteger *igd,
    hscalinteger *myrow, hscalinteger *mycol, hscalinteger *nprow, hscalinteger *npcol,
    double *tcomp, int *tcomm
);

int hscal_pbdtrget(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *mnb, hscalinteger *lda, hscalinteger *mcrow,
    hscalinteger *mccol, hscalinteger *igd,
    hscalinteger *myrow, hscalinteger *mycol, hscalinteger *nprow, hscalinteger *npcol,
    double *tcomp, int *tcomm
);

int hscal_pbstrget(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *mnb, hscalinteger *lda, hscalinteger *mcrow,
    hscalinteger *mccol, hscalinteger *igd,
    hscalinteger *myrow, hscalinteger *mycol, hscalinteger *nprow, hscalinteger *npcol,
    double *tcomp, int *tcomm
);

int hscal_pbztrget(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *mnb, hscalinteger *lda, hscalinteger *mcrow,
    hscalinteger *mccol, hscalinteger *igd,
    hscalinteger *myrow, hscalinteger *mycol, hscalinteger *nprow, hscalinteger *npcol,
    double *tcomp, int *tcomm
);

int hscal_pbctrsrt(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda, hscalcomplex *beta,
    hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pbdtrsrt(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda, hscaldoublereal *beta,
    hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pbstrsrt(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda, hscalreal *beta,
    hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pbztrsrt(
    hscalinteger *icontxt, char *adist, hscalinteger *m,
    hscalinteger *n, hscalinteger *nb, hscalinteger *lda, hscaldoublecomplex *beta,
    hscalinteger *ldb, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pbctrst1(
    hscalinteger *icontxt, char *xdist, hscalinteger *n,
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, hscalcomplex *beta,
    hscalinteger *incy, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pbdtrst1(
    hscalinteger *icontxt, char *xdist, hscalinteger *n,
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, hscaldoublereal *beta,
    hscalinteger *incy, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pbstrst1(
    hscalinteger *icontxt, char *xdist, hscalinteger *n,
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, hscalreal *beta,
    hscalinteger *incy, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

int hscal_pbztrst1(
    hscalinteger *icontxt, char *xdist, hscalinteger *n,
    hscalinteger *nb, hscalinteger *nz, hscalinteger *incx, hscaldoublecomplex *beta,
    hscalinteger *incy, hscalinteger *lcmp, hscalinteger *lcmq, hscalinteger *nint,
    int nprow, int npcol, int myrow, int mycol, double *tcomp
);

#endif

