
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK PBLAS Level 1            *
*                 =====================================            *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_PBLAS_LEVEL1_INCLUDE_HH
#define __HMPI_SCALAPACK_PBLAS_LEVEL1_INCLUDE_HH

/*
* ------------------------------------------------------------------
*       PBLAS LEVEL 1
*       HeteroMPI group calls
* ------------------------------------------------------------------
*/

/*
*  pcswap
*/
double hscal_pcswap_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cswap_benchmark(
    int     *n, 
    hscalcomplex *cx, 
    int     *incx, 
    hscalcomplex *cy, 
    int     *incy
);

int hscal_pack_pcswap_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcswap_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pcswap_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);
 
double hscal_pcswap_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcswap_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcswap_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcswap_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pdswap
*/
double hscal_pdswap_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dswap_benchmark(
    int     *n, 
    hscaldoublereal *cx, 
    int     *incx, 
    hscaldoublereal *cy, 
    int     *incy
);

int hscal_pack_pdswap_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdswap_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pdswap_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pdswap_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdswap_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdswap_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdswap_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  psswap
*/
double hscal_psswap_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_sswap_benchmark(
    int     *n, 
    hscalreal *cx, 
    int     *incx, 
    hscalreal *cy, 
    int     *incy
);

int hscal_pack_psswap_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psswap_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_psswap_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_psswap_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psswap_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psswap_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psswap_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pzswap
*/
double hscal_pzswap_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zswap_benchmark(
    int     *n, 
    hscaldoublecomplex *cx, 
    int     *incx, 
    hscaldoublecomplex *cy, 
    int     *incy
);

int hscal_pack_pzswap_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzswap_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pzswap_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pzswap_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);


int hscal_pzswap_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzswap_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzswap_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCSCAL
*/
double hscal_pcscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cscal_benchmark(
    int     *n, 
    hscalcomplex *ca, 
    hscalcomplex *cx, 
    int     *incx
);

int hscal_pack_pcscal_modelp_noauto(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcscal_modelp(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pcscal_recon(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pcscal_timeof(
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcscal_group_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcscal_group_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcscal_group_heuristic_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcsscal
*/
double hscal_pcsscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_csscal_benchmark(
    int     *n, 
    hscalreal    *sa, 
    hscalcomplex *cx, 
    int     *incx
);

int hscal_pack_pcsscal_modelp_noauto(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcsscal_modelp(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pcsscal_recon(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pcsscal_timeof(
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcsscal_group_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcsscal_group_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcsscal_group_heuristic_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdscal
*/
double hscal_pdscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dscal_benchmark(
    int     *n, 
    hscaldoublereal *da, 
    hscaldoublereal *dx, 
    int     *incx
);

int hscal_pack_pdscal_modelp_noauto(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdscal_modelp(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pdscal_recon(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pdscal_timeof(
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdscal_group_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdscal_group_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdscal_group_heuristic_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psscal
*/
double hscal_psscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_sscal_benchmark(
    int     *n, 
    hscalreal *sa, 
    hscalreal *sx, 
    int     *incx
);

int hscal_pack_psscal_modelp_noauto(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psscal_modelp(
   int * N, 
   float  * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_psscal_recon(
   int * N, 
   float  * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_psscal_timeof(
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psscal_group_create(
   int * N, 
   float  * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psscal_group_auto_create(
   int * N, 
   float  * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psscal_group_heuristic_auto_create(
   int * N, 
   float  * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzscal
*/
double hscal_pzscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zscal_benchmark(
    int     *n, 
    hscaldoublecomplex *za, 
    hscaldoublecomplex *zx, 
    int     *incx
);

int hscal_pack_pzscal_modelp_noauto(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzscal_modelp(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pzscal_recon(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pzscal_timeof(
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzscal_group_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzscal_group_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzscal_group_heuristic_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzdscal
*/
double hscal_pzdscal_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zdscal_benchmark(
    int           *n, 
    hscaldoublereal    *da, 
    hscaldoublecomplex *zx, 
    int           *incx
);

int hscal_pack_pzdscal_modelp_noauto(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzdscal_modelp(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pzdscal_recon(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pzdscal_timeof(
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzdscal_group_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdscal_group_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdscal_group_heuristic_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pccopy
*/
double hscal_pccopy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ccopy_benchmark(
    int     *n, 
    hscalcomplex *cx, 
    int     *incx, 
    hscalcomplex *cy, 
    int     *incy
);

int hscal_pack_pccopy_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pccopy_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pccopy_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pccopy_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pccopy_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pccopy_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pccopy_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdcopy
*/
double hscal_pdcopy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dcopy_benchmark(
    int     *n, 
    hscaldoublereal *dx, 
    int     *incx, 
    hscaldoublereal *dy, 
    int     *incy
);

int hscal_pack_pdcopy_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdcopy_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pdcopy_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pdcopy_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdcopy_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdcopy_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdcopy_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* picopy
*/
double hscal_picopy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_icopy_benchmark(
    int     *n,
    int     *cx,
    int     *incx,
    int     *cy,
    int     *incy
);

int hscal_pack_picopy_modelp_noauto(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_picopy_modelp(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_picopy_recon(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_picopy_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_picopy_group_create(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_picopy_group_auto_create(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_picopy_group_heuristic_auto_create(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pscopy
*/
double hscal_pscopy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_scopy_benchmark(
    int     *n, 
    hscalreal *sx, 
    int     *incx,
    hscalreal *sy, 
    int     *incy
);

int hscal_pack_pscopy_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pscopy_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pscopy_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pscopy_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pscopy_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pscopy_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pscopy_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzcopy
*/
double hscal_pzcopy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zcopy_benchmark(
    int     *n, 
    hscaldoublecomplex *zx, 
    int     *incx, 
    hscaldoublecomplex *zy, 
    int     *incy
);

int hscal_pack_pzcopy_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzcopy_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pzcopy_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pzcopy_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzcopy_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzcopy_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzcopy_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcaxpy
*/
double hscal_pcaxpy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_caxpy_benchmark(
    int     *n, 
    hscalcomplex *ca, 
    hscalcomplex *cx, 
    int     *incx, 
    hscalcomplex *cy, 
    int     *incy
);

int hscal_pack_pcaxpy_modelp_noauto(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcaxpy_modelp(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pcaxpy_recon(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pcaxpy_timeof(
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcaxpy_group_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcaxpy_group_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcaxpy_group_heuristic_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdaxpy
*/
double hscal_pdaxpy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_daxpy_benchmark(
    int     *n, 
    hscaldoublereal *da, 
    hscaldoublereal *dx, 
    int     *incx, 
    hscaldoublereal *dy, 
    int     *incy
);

int hscal_pack_pdaxpy_modelp_noauto(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdaxpy_modelp(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pdaxpy_recon(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pdaxpy_timeof(
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdaxpy_group_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdaxpy_group_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdaxpy_group_heuristic_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psaxpy
*/
double hscal_psaxpy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_saxpy_benchmark(
    int     *n, 
    hscalreal *sa, 
    hscalreal *sx, 
    int     *incx, 
    hscalreal *sy, 
    int     *incy
);

int hscal_pack_psaxpy_modelp_noauto(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psaxpy_modelp(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_psaxpy_recon(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_psaxpy_timeof(
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psaxpy_group_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psaxpy_group_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psaxpy_group_heuristic_auto_create(
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzaxpy
*/
double hscal_pzaxpy_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zaxpy_benchmark(  
    int     *n, 
    hscaldoublecomplex *za, 
    hscaldoublecomplex *zx, 
    int     *incx, 
    hscaldoublecomplex *zy, 
    int     *incy
);

int hscal_pack_pzaxpy_modelp_noauto(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzaxpy_modelp(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pzaxpy_recon(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pzaxpy_timeof(
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzaxpy_group_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzaxpy_group_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzaxpy_group_heuristic_auto_create(
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCAMAX
*/
double hscal_pcamax_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int     hscal_icamax_benchmark(
        int     *n, 
        hscalcomplex *cx, 
        int     *incx
);

int hscal_pack_pcamax_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcamax_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pcamax_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pcamax_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcamax_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcamax_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcamax_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdamax
*/
double hscal_pdamax_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int     hscal_idamax_benchmark(
        int     *n, 
        hscaldoublereal *dx, 
        int     *incx
);

int hscal_pack_pdamax_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdamax_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pdamax_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pdamax_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdamax_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdamax_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdamax_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psamax
*/
double hscal_psamax_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int     hscal_isamax_benchmark(
        int     *n, 
        hscalreal    *sx, 
        int     *incx
);

int hscal_pack_psamax_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psamax_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_psamax_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_psamax_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psamax_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psamax_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psamax_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzamax
*/
double hscal_pzamax_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int     hscal_izamax_benchmark(
        int           *n, 
        hscaldoublecomplex *zx, 
        int           *incx
);

int hscal_pack_pzamax_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzamax_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pzamax_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pzamax_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzamax_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzamax_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzamax_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdasum
*/
double hscal_pdasum_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_dasum_benchmark(
           int     *n, 
           hscaldoublereal *dx, 
           int     *incx
);

int hscal_pack_pdasum_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdasum_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pdasum_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pdasum_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdasum_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdasum_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdasum_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pscasum
*/
double hscal_pscasum_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_scasum_benchmark(
           int     *n, 
           hscalcomplex *cx, 
           int     *incx
);

int hscal_pack_pscasum_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pscasum_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pscasum_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pscasum_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pscasum_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pscasum_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pscasum_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdzasum
*/
double hscal_pdzasum_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_dzasum_benchmark(
           int     *n, 
           hscaldoublecomplex *zx, 
           int     *incx
);

int hscal_pack_pdzasum_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdzasum_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pdzasum_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pdzasum_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdzasum_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdzasum_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdzasum_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psasum
*/
double hscal_psasum_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_sasum_benchmark(
           int     *n, 
           hscalreal    *sx, 
           int     *incx
);

int hscal_pack_psasum_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psasum_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_psasum_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_psasum_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psasum_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psasum_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psasum_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdnrm2
*/
double hscal_pdnrm2_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_dnrm2_benchmark(
           int     *n, 
           hscaldoublereal *x, 
           int     *incx
);

int hscal_pack_pdnrm2_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdnrm2_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pdnrm2_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pdnrm2_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdnrm2_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdnrm2_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdnrm2_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pscnrm2
*/
double hscal_pscnrm2_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_scnrm2_benchmark(
           int     *n, 
           hscalcomplex *x, 
           int     *incx
);

int hscal_pack_pscnrm2_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pscnrm2_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pscnrm2_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pscnrm2_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pscnrm2_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pscnrm2_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pscnrm2_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdznrm2
*/
double hscal_pdznrm2_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_dznrm2_benchmark(
           int     *n, 
           hscaldoublecomplex *x, 
           int     *incx
);

int hscal_pack_pdznrm2_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdznrm2_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pdznrm2_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pdznrm2_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdznrm2_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdznrm2_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdznrm2_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psnrm2
*/
double hscal_psnrm2_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_snrm2_benchmark(
           int     *n, 
           hscalreal *x, 
           int     *incx
);

int hscal_pack_psnrm2_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psnrm2_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_psnrm2_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_psnrm2_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psnrm2_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psnrm2_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psnrm2_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PDDOT
*/
double hscal_pddot_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_ddot_benchmark(
           int        *n, 
           hscaldoublereal *dx, 
           int        *incx, 
           hscaldoublereal *dy, 
           int        *incy
);

int hscal_pack_pddot_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pddot_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pddot_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pddot_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pddot_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pddot_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pddot_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psdot
*/
double hscal_psdot_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

hscaldoublereal hscal_sdot_benchmark(
           int     *n, 
           hscalreal    *sx, 
           int     *incx, 
           hscalreal    *sy, 
           int     *incy
);

int hscal_pack_psdot_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psdot_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_psdot_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_psdot_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psdot_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psdot_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psdot_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcdotc
*/
double hscal_pcdotc_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

void hscal_cdotc_benchmark(
     hscalcomplex * ret_val, 
     int     *n, 
     hscalcomplex *cx, 
     int     *incx, 
     hscalcomplex *cy, 
     int     *incy
);

int hscal_pack_pcdotc_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcdotc_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pcdotc_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pcdotc_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcdotc_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdotc_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdotc_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcdotu
*/
double hscal_pcdotu_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

void hscal_cdotu_benchmark(
     hscalcomplex * ret_val, 
     int     * n, 
     hscalcomplex * cx, 
     int     * incx, 
     hscalcomplex * cy, 
     int     * incy
);

int hscal_pack_pcdotu_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcdotu_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pcdotu_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pcdotu_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcdotu_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdotu_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcdotu_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzdotc
*/
double hscal_pzdotc_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

void hscal_zdotc_benchmark(
     hscaldoublecomplex * ret_val, 
     int           * n, 
     hscaldoublecomplex * zx, 
     int           * incx, 
     hscaldoublecomplex * zy, 
     int           * incy
);

int hscal_pack_pzdotc_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzdotc_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pzdotc_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pzdotc_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzdotc_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdotc_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdotc_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzdotu
*/
double hscal_pzdotu_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

void hscal_zdotu_benchmark(
     hscaldoublecomplex * ret_val, 
     int           * n, 
     hscaldoublecomplex * zx, 
     int           * incx, 
     hscaldoublecomplex * zy, 
     int           * incy
);

int hscal_pack_pzdotu_modelp_noauto(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzdotu_modelp(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pzdotu_recon(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY
);

double hscal_pzdotu_timeof(
   int * N,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzdotu_group_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdotu_group_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzdotu_group_heuristic_auto_create(
   int * N, 
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

#endif /* __HMPI_SCALAPACK_PBLAS_LEVEL1_INCLUDE_HH */
