
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK LEVEL2                   *
*                         (INTERNAL)                               *
*                 ==============================                   *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_PBLAS_LEVEL2_INCLUDE_HH
#define __HMPI_SCALAPACK_PBLAS_LEVEL2_INCLUDE_HH

/*
* ------------------------------------------------------------------
*       PBLAS LEVEL 2
*       HeteroMPI group calls
* ------------------------------------------------------------------
*/

/*
*  pcgemv
*/
double hscal_pcgemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_cgemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *x, 
    int     *incx, 
    hscalcomplex *beta, 
    hscalcomplex *y, 
    int     *incy
);

int hscal_pack_pcgemv_modelp(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgemv_modelp_noauto(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pcgemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_pcgemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pdgemv
*/
double hscal_pdgemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_dgemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublereal * alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *x, 
    int     *incx,
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    int     *incy
);

int hscal_pack_pdgemv_modelp_noauto(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgemv_modelp(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pdgemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pdgemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  psgemv
*/
double hscal_psgemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_sgemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *x, 
    int     *incx, 
    hscalreal *beta, 
    hscalreal *y,
    int     *incy
);

int hscal_pack_psgemv_modelp_noauto(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgemv_modelp(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_psgemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_psgemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pzgemv
*/
double hscal_pzgemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_zgemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *x, 
    int     *incx, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *y, 
    int     *incy
);

int hscal_pzgemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pack_pzgemv_modelp(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgemv_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

double hscal_pzgemv_timeof(
   F_CHAR_T TRANS,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pcagemv
*/
double hscal_pcagemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_cagemv_benchmark(
    char    *trans, 
    int     *m, 
    int     *n, 
    hscalreal    *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *x, 
    int     *incx, 
    hscalreal    *beta, 
    hscalreal    *y, 
    int     *incy
);

int hscal_pack_pcagemv_modelp(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pack_pcagemv_modelp_noauto(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pcagemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_pcagemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX, 
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcagemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcagemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcagemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pdagemv
*/
double hscal_pdagemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_dagemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublereal * alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *x, 
    int     *incx,
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    int     *incy
);

int hscal_pack_pdagemv_modelp_noauto(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdagemv_modelp(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pdagemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdagemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdagemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pdagemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdagemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  psagemv
*/
double hscal_psagemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_sagemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *x, 
    int     *incx, 
    hscalreal *beta, 
    hscalreal *y,
    int     *incy
);

int hscal_pack_psagemv_modelp_noauto(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psagemv_modelp(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_psagemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_psagemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psagemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psagemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psagemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pzagemv
*/
double hscal_pzagemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_zagemv_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *x, 
    int     *incx, 
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    int     *incy
);

int hscal_pzagemv_recon(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pack_pzagemv_modelp(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pack_pzagemv_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N, 
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   double * BETA,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

double hscal_pzagemv_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzagemv_group_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzagemv_group_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzagemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pdsymv
*/
double hscal_pdsymv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_dsymv_benchmark(
    char *uplo, 
    int     *n, 
    hscaldoublereal * alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *x, 
    int     *incx,
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    int     *incy
);

int hscal_pack_pdsymv_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdsymv_modelp(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pdsymv_timeof(
   F_CHAR_T UPLO,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdsymv_group_create(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsymv_recon(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pdsymv_group_auto_create(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsymv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pssymv
*/
double hscal_pssymv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_ssymv_benchmark(
    char *uplo, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *x, 
    int     *incx, 
    hscalreal *beta, 
    hscalreal *y,
    int     *incy
);

int hscal_pack_pssymv_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pssymv_modelp(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pssymv_recon(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_pssymv_timeof(
   F_CHAR_T UPLO,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pssymv_group_create(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssymv_group_auto_create(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssymv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pzhemv
*/
double hscal_pzhemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_zhemv_benchmark(
    char *uplo,
    int     *n,
    hscaldoublecomplex * alpha,
    hscaldoublecomplex *a,
    int     *lda,
    hscaldoublecomplex *x,
    int     *incx,
    hscaldoublecomplex *beta,
    hscaldoublecomplex *y,
    int     *incy
);

int hscal_pack_pzhemv_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzhemv_modelp(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pzhemv_timeof(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzhemv_group_create(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzhemv_recon(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pzhemv_group_auto_create(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzhemv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pchemv
*/
double hscal_pchemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_chemv_benchmark(
    char *uplo,
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *x, 
    int     *incx, 
    hscalcomplex *beta, 
    hscalcomplex *y,
    int     *incy
);

int hscal_pack_pchemv_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pchemv_modelp(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pchemv_recon(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_pchemv_timeof(
   F_CHAR_T UPLO,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pchemv_group_create(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pchemv_group_auto_create(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pchemv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pzahemv
*/
double hscal_pzahemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_zahemv_benchmark(
    char *uplo, 
    int     *n, 
    hscaldoublereal * alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *x, 
    int     *incx,
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    int     *incy
);

int hscal_pack_pzahemv_modelp_noauto(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzahemv_modelp(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pzahemv_timeof(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzahemv_group_create(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzahemv_recon(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pzahemv_group_auto_create(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzahemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pcahemv
*/
double hscal_pcahemv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_cahemv_benchmark(
    char *trans, 
    int     *n, 
    hscalreal *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *x, 
    int     *incx, 
    hscalreal *beta, 
    hscalreal *y,
    int     *incy
);

int hscal_pack_pcahemv_modelp_noauto(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcahemv_modelp(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_pcahemv_recon(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_pcahemv_timeof(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcahemv_group_create(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcahemv_group_auto_create(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcahemv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pdasymv
*/
double hscal_pdasymv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_dasymv_benchmark(
    char *uplo, 
    int     *n, 
    hscaldoublereal * alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *x, 
    int     *incx,
    hscaldoublereal *beta, 
    hscaldoublereal *y, 
    int     *incy
);

int hscal_pack_pdasymv_modelp_noauto(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdasymv_modelp(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pdasymv_timeof(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdasymv_group_create(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdasymv_recon(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pdasymv_group_auto_create(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdasymv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  psasymv
*/
double hscal_psasymv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_sasymv_benchmark(
    char *uplo, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *x, 
    int     *incx, 
    hscalreal *beta, 
    hscalreal *y,
    int     *incy
);

int hscal_pack_psasymv_modelp_noauto(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psasymv_modelp(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

int hscal_psasymv_recon(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

double hscal_psasymv_timeof(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psasymv_group_create(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psasymv_group_auto_create(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psasymv_group_heuristic_auto_create(
   F_CHAR_T TRANS,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pctrmv
*/
double hscal_pctrmv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_ctrmv_benchmark(
    char *uplo, 
    char *trans, 
    char *diag, 
    int     *n, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *x, 
    int     *incx
);

int hscal_pack_pctrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pctrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pctrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pctrmv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pctrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pdtrmv
*/
double hscal_pdtrmv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_dtrmv_benchmark(
    char *uplo, 
    char *trans, 
    char *diag, 
    int     *n, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *x, 
    int     *incx
);

int hscal_pack_pdtrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdtrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pdtrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pdtrmv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdtrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pstrmv
*/
double hscal_pstrmv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_strmv_benchmark(
    char *uplo, 
    char *trans, 
    char *diag, 
    int     *n, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *x, 
    int     *incx
);

int hscal_pack_pstrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pstrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pstrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pstrmv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pstrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pztrmv
*/
double hscal_pztrmv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_ztrmv_benchmark(
    char *uplo, 
    char *trans, 
    char *diag, 
    int     *n, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *x, 
    int     *incx
);

int hscal_pack_pztrmv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pztrmv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pztrmv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pztrmv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pztrmv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztrmv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pcatrmv
*/
double hscal_pcatrmv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_catrmv_benchmark(
    char *uplo,
    char *trans, 
    char *diag,
    int     *n, 
    hscalreal * alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *x, 
    int     *incx,
    hscalreal *beta, 
    hscalreal *y, 
    int     *incy
);

int hscal_pack_pcatrmv_modelp_noauto(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcatrmv_modelp(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pcatrmv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcatrmv_group_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcatrmv_recon(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pcatrmv_group_auto_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcatrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  psatrmv
*/
double hscal_psatrmv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_satrmv_benchmark(
    char *uplo,
    char *trans, 
    char *diag,
    int     *n, 
    hscalreal * alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *x, 
    int     *incx,
    hscalreal *beta, 
    hscalreal *y, 
    int     *incy
);

int hscal_pack_psatrmv_modelp_noauto(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psatrmv_modelp(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_psatrmv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psatrmv_group_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psatrmv_recon(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_psatrmv_group_auto_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psatrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   float * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pdatrmv
*/
double hscal_pdatrmv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_datrmv_benchmark(
    char *uplo,
    char *trans, 
    char *diag,
    int     *n, 
    double * alpha, 
    double *a, 
    int     *lda, 
    double *x, 
    int     *incx,
    double *beta, 
    double *y, 
    int     *incy
);

int hscal_pack_pdatrmv_modelp_noauto(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdatrmv_modelp(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pdatrmv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdatrmv_group_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdatrmv_recon(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pdatrmv_group_auto_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdatrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pzatrmv
*/
double hscal_pzatrmv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_zatrmv_benchmark(
    char *uplo,
    char *trans, 
    char *diag,
    int     *n, 
    double * alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *x, 
    int     *incx,
    double *beta, 
    double *y, 
    int     *incy
);

int hscal_pack_pzatrmv_modelp_noauto(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzatrmv_modelp(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int **model_params,
   int *model_count
);

double hscal_pzatrmv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzatrmv_group_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzatrmv_recon(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY
);

int hscal_pzatrmv_group_auto_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzatrmv_group_heuristic_auto_create(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   double * BETA,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pctrsv
*/
double hscal_pctrsv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_ctrsv_benchmark(
    char *uplo, 
    char *trans, 
    char *diag, 
    int     *n, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *x, 
    int     *incx
);

int hscal_pack_pctrsv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pctrsv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pctrsv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pctrsv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pctrsv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctrsv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctrsv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pdtrsv
*/
double hscal_pdtrsv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_dtrsv_benchmark(
    char *uplo, 
    char *trans, 
    char *diag, 
    int     *n, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *x, 
    int     *incx
);

int hscal_pack_pdtrsv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdtrsv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pdtrsv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pdtrsv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdtrsv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtrsv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtrsv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pstrsv
*/
double hscal_pstrsv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_strsv_benchmark(
    char *uplo, 
    char *trans, 
    char *diag, 
    int     *n, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *x, 
    int     *incx
);

int hscal_pack_pstrsv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pstrsv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pstrsv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pstrsv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pstrsv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstrsv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstrsv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
*  pztrsv
*/
double hscal_pztrsv_benchmark_function
(
   const void* input_p,
   int num_of_p,
   void* output_p
);

int hscal_ztrsv_benchmark(
    char *uplo, 
    char *trans, 
    char *diag, 
    int     *n, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *x, 
    int     *incx
);

int hscal_pack_pztrsv_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pztrsv_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int **model_params,
   int *model_count
);

int hscal_pztrsv_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX
);

double hscal_pztrsv_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * N,
   int * IA,
   int * JA,
   int * DESCA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pztrsv_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztrsv_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztrsv_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS,
   F_CHAR_T DIAG, 
   int * N, 
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psger
*/
double hscal_psger_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_sger_benchmark(
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *x, 
    int     *incx,
    hscalreal *y, 
    int     *incy, 
    hscalreal *a, 
    int     *lda
);

int hscal_pack_psger_modelp_noauto(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psger_modelp(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_psger_recon(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_psger_timeof(
   int * M,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psger_group_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psger_group_auto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psger_group_heuristic_auto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdger
*/
double hscal_pdger_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dger_benchmark(
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *x, 
    int     *incx, 
    hscaldoublereal *y, 
    int     *incy, 
    hscaldoublereal *a, 
    int     *lda
);

int hscal_pack_pdger_modelp_noauto(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdger_modelp(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pdger_recon(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pdger_timeof(
   int * M,
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdger_group_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdger_group_auto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdger_group_heuristic_auto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcgerc
*/
double hscal_pcgerc_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cgerc_benchmark(
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *x, 
    int     *incx, 
    hscalcomplex *y, 
    int     *incy, 
    hscalcomplex *a, 
    int     *lda
);

int hscal_pack_pcgerc_modelp_noauto(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgerc_modelp(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcgerc_recon(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pcgerc_timeof(
   int * M,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgerc_group_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgerc_group_auto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgerc_group_heuristic_auto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzgerc
*/
double hscal_pzgerc_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zgerc_benchmark(
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *x, 
    int     *incx, 
    hscaldoublecomplex *y, 
    int     *incy, 
    hscaldoublecomplex *a, 
    int     *lda
);

int hscal_pack_pzgerc_modelp_noauto(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgerc_modelp(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzgerc_recon(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pzgerc_timeof(
   int * M,
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgerc_group_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgerc_group_auto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgerc_group_heuristic_auto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcgeru
*/
double hscal_pcgeru_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cgeru_benchmark(
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *x, 
    int     *incx, 
    hscalcomplex *y, 
    int     *incy, 
    hscalcomplex *a, 
    int     *lda
);

int hscal_pack_pcgeru_modelp_noauto(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgeru_modelp(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcgeru_recon(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pcgeru_timeof(
   int * M,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgeru_group_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgeru_group_auto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgeru_group_heuristic_auto_create(
   int * M, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzgeru
*/
double hscal_pzgeru_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zgeru_benchmark(
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *x, 
    int     *incx, 
    hscaldoublecomplex *y, 
    int     *incy, 
    hscaldoublecomplex *a, 
    int     *lda
);

int hscal_pack_pzgeru_modelp_noauto(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgeru_modelp(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzgeru_recon(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pzgeru_timeof(
   int * M,
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgeru_group_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgeru_group_auto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgeru_group_heuristic_auto_create(
   int * M, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCHER
*/
double hscal_pcher_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cher_benchmark(
    char    *uplo, 
    int     *n, 
    hscalreal    *alpha, 
    hscalcomplex *x, 
    int     *incx, 
    hscalcomplex *a, 
    int     *lda
);

int hscal_pack_pcher_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcher_modelp(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcher_recon(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pcher_timeof(
   F_CHAR_T UPLO,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcher_group_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcher_group_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcher_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PZHER
*/
double hscal_pzher_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zher_benchmark(
    char          *uplo, 
    int           *n, 
    hscaldoublereal    *alpha, 
    hscaldoublecomplex *x, 
    int           *incx, 
    hscaldoublecomplex *a, 
    int           *lda
);

int hscal_pack_pzher_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzher_modelp(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzher_recon(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pzher_timeof(
   F_CHAR_T UPLO,
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzher_group_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzher_group_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzher_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);


/*
* PCHER2
*/
double hscal_pcher2_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cher2_benchmark(
    char    *uplo, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *x, 
    int     *incx, 
    hscalcomplex *y, 
    int     *incy, 
    hscalcomplex *a, 
    int     *lda
);

int hscal_pack_pcher2_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcher2_modelp(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pcher2_recon(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pcher2_timeof(
   F_CHAR_T UPLO,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcher2_group_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcher2_group_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcher2_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PZHER2
*/
double hscal_pzher2_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zher2_benchmark(
    char          *uplo, 
    int           *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *x, 
    int           *incx, 
    hscaldoublecomplex *y, 
    int           *incy, 
    hscaldoublecomplex *a, 
    int           *lda
);

int hscal_pack_pzher2_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzher2_modelp(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pzher2_recon(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pzher2_timeof(
   F_CHAR_T UPLO,
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzher2_group_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzher2_group_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzher2_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pssyr
*/
double hscal_pssyr_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ssyr_benchmark(
    char    *uplo, 
    int     *n, 
    hscalreal    *alpha, 
    hscalreal    *x, 
    int     *incx, 
    hscalreal    *a, 
    int     *lda
);

int hscal_pack_pssyr_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pssyr_modelp(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pssyr_recon(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pssyr_timeof(
   F_CHAR_T UPLO,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pssyr_group_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssyr_group_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssyr_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdsyr
*/
double hscal_pdsyr_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dsyr_benchmark(
    char       *uplo, 
    int        *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *x, 
    int        *incx, 
    hscaldoublereal *a, 
    int        *lda
);

int hscal_pack_pdsyr_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdsyr_modelp(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pdsyr_recon(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pdsyr_timeof(
   F_CHAR_T UPLO,
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdsyr_group_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsyr_group_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsyr_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pssyr2
*/
double hscal_pssyr2_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ssyr2_benchmark(
    char    *uplo, 
    int     *n, 
    hscalreal    *alpha, 
    hscalreal    *x, 
    int     *incx, 
    hscalreal    *y, 
    int     *incy, 
    hscalreal    *a, 
    int     *lda
);

int hscal_pack_pssyr2_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pssyr2_modelp(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pssyr2_recon(
   F_CHAR_T UPLO,
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pssyr2_timeof(
   F_CHAR_T UPLO,
   int * N,
   float * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pssyr2_group_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssyr2_group_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssyr2_group_heuristic_auto_create(
   F_CHAR_T UPLO,
   int * N, 
   float * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdsyr2
*/
double hscal_pdsyr2_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dsyr2_benchmark(
    char       *uplo, 
    int        *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *x, 
    int        *incx, 
    hscaldoublereal *y, 
    int        *incy, 
    hscaldoublereal *a, 
    int        *lda
);

int hscal_pack_pdsyr2_modelp_noauto(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdsyr2_modelp(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int **model_params,
   int *model_count
);

int hscal_pdsyr2_recon(
   F_CHAR_T UPLO,
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA
);

double hscal_pdsyr2_timeof(
   F_CHAR_T UPLO,
   int * N,
   double * ALPHA,
   int * IX,
   int * JX,
   int * DESCX,
   int * INCX,
   int * IY,
   int * JY,
   int * DESCY,
   int * INCY,
   int * IA,
   int * JA,
   int * DESCA,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdsyr2_group_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsyr2_group_auto_create(
   F_CHAR_T UPLO, 
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsyr2_group_heuristic_auto_create(
   F_CHAR_T UPLO,
   int * N, 
   double * ALPHA,
   int * IX, 
   int * JX, 
   int * DESCX, 
   int * INCX,
   int * IY, 
   int * JY, 
   int * DESCY, 
   int * INCY,
   int * IA, 
   int * JA, 
   int * DESCA,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

#endif /* __HMPI_SCALAPACK_PBLAS_LEVEL2_INCLUDE_HH */
