
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK                          *
*                         (INTERNAL)                               *
*                 =======================                          *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_PBLAS_LEVEL3_INCLUDE_HH
#define __HMPI_SCALAPACK_PBLAS_LEVEL3_INCLUDE_HH

/*
* ------------------------------------------------------------------
*       PBLAS LEVEL 3
*       HeteroMPI group calls
* ------------------------------------------------------------------
*/
/*
* pcherk
*/
double hscal_pcherk_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cherk_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscalreal *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalreal *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pcherk_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcherk_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pcherk_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pcherk_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcherk_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcherk_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcherk_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzherk
*/
double hscal_pzherk_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zherk_benchmark(
    char          *uplo, 
    char          *trans, 
    int           *n, 
    int           *k, 
    hscaldoublereal    *alpha, 
    hscaldoublecomplex *a, 
    int           *lda, 
    hscaldoublereal    *beta, 
    hscaldoublecomplex *c__, 
    int           *ldc
);

int hscal_pack_pzherk_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzherk_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pzherk_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pzherk_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzherk_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzherk_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzherk_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcsyrk
*/
double hscal_pcsyrk_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_csyrk_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pcsyrk_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcsyrk_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pcsyrk_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pcsyrk_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcsyrk_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcsyrk_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcsyrk_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzsyrk
*/
double hscal_pzsyrk_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zsyrk_benchmark(
    char          *uplo, 
    char          *trans, 
    int           *n, 
    int           *k, 
    hscaldoublecomplex    *alpha, 
    hscaldoublecomplex *a, 
    int           *lda, 
    hscaldoublecomplex    *beta, 
    hscaldoublecomplex *c__, 
    int           *ldc
);

int hscal_pack_pzsyrk_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzsyrk_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pzsyrk_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pzsyrk_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzsyrk_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzsyrk_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzsyrk_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pssyrk
*/
double hscal_pssyrk_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ssyrk_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *beta, 
    hscalreal *c__, 
    int     *ldc
);

int hscal_pack_pssyrk_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pssyrk_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pssyrk_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pssyrk_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pssyrk_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssyrk_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssyrk_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdsyrk
*/
double hscal_pdsyrk_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dsyrk_benchmark(
    char          *uplo, 
    char          *trans, 
    int           *n, 
    int           *k, 
    hscaldoublereal    *alpha, 
    hscaldoublereal    *a, 
    int           *lda, 
    hscaldoublereal    *beta, 
    hscaldoublereal    *c__, 
    int           *ldc
);

int hscal_pack_pdsyrk_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdsyrk_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pdsyrk_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pdsyrk_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdsyrk_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsyrk_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsyrk_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcher2k
*/
double hscal_pcher2k_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cher2k_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *b, 
    int     *ldb, 
    hscalreal *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pcher2k_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcher2k_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pcher2k_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pcher2k_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcher2k_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcher2k_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcher2k_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzher2k
*/
double hscal_pzher2k_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zher2k_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *b, 
    int     *ldb, 
    hscaldoublereal *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
);

int hscal_pack_pzher2k_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzher2k_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pzher2k_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pzher2k_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzher2k_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzher2k_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzher2k_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcsyr2k
*/
double hscal_pcsyr2k_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_csyr2k_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *b, 
    int     *ldb, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pcsyr2k_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcsyr2k_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pcsyr2k_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pcsyr2k_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcsyr2k_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcsyr2k_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcsyr2k_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzsyr2k
*/
double hscal_pzsyr2k_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zsyr2k_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *b, 
    int     *ldb, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__,
    int     *ldc
);

int hscal_pack_pzsyr2k_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzsyr2k_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pzsyr2k_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pzsyr2k_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzsyr2k_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzsyr2k_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzsyr2k_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pssyr2k
*/
double hscal_pssyr2k_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ssyr2k_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *b, 
    int     *ldb, 
    hscalreal *beta,
    hscalreal *c__, 
    int     *ldc
);

int hscal_pack_pssyr2k_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pssyr2k_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pssyr2k_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pssyr2k_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pssyr2k_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssyr2k_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssyr2k_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdsyr2k
*/
double hscal_pdsyr2k_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dsyr2k_benchmark(
    char *uplo, 
    char *trans, 
    int     *n, 
    int     *k, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *b, 
    int     *ldb, 
    hscaldoublereal *beta, 
    hscaldoublereal *c__, 
    int     *ldc
);

int hscal_pack_pdsyr2k_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdsyr2k_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pdsyr2k_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pdsyr2k_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * N,
   int * K,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdsyr2k_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsyr2k_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsyr2k_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCHEMM
*/
double hscal_pchemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_chemm_benchmark(
    char    *side, 
    char    *uplo, 
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *b, 
    int     *ldb, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pchemm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pchemm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pchemm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pchemm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pchemm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pchemm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pchemm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PZHEMM
*/
double hscal_pzhemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zhemm_benchmark(
    char    *side, 
    char    *uplo, 
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *b, 
    int     *ldb, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
);

int hscal_pack_pzhemm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzhemm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pzhemm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pzhemm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzhemm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzhemm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzhemm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCSYMM
*/
double hscal_pcsymm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_csymm_benchmark(
    char    *side, 
    char    *uplo, 
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *b, 
    int     *ldb, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pcsymm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcsymm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pcsymm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pcsymm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcsymm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcsymm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcsymm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PDSYMM
*/
double hscal_pdsymm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dsymm_benchmark(
    char    *side, 
    char    *uplo, 
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *b, 
    int     *ldb, 
    hscaldoublereal *beta, 
    hscaldoublereal *c__, 
    int     *ldc
);

int hscal_pack_pdsymm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdsymm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pdsymm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pdsymm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdsymm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsymm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdsymm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PSSYMM
*/
double hscal_pssymm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ssymm_benchmark(
    char    *side, 
    char    *uplo, 
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *b, 
    int     *ldb, 
    hscalreal *beta, 
    hscalreal *c__, 
    int     *ldc
);

int hscal_pack_pssymm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pssymm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pssymm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pssymm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pssymm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssymm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pssymm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PZSYMM
*/
double hscal_pzsymm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zsymm_benchmark(
    char    *side, 
    char    *uplo, 
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *b, 
    int     *ldb, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
);

int hscal_pack_pzsymm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzsymm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pzsymm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pzsymm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzsymm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzsymm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzsymm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCTRANC
*/
double hscal_pctranc_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ctranc_benchmark(
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pctranc_modelp_noauto(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pctranc_modelp(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pctranc_recon(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pctranc_timeof(
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pctranc_group_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctranc_group_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctranc_group_heuristic_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pztranc
*/
double hscal_pztranc_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ztranc_benchmark(
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
);

int hscal_pack_pztranc_modelp_noauto(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pztranc_modelp(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pztranc_recon(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pztranc_timeof(
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pztranc_group_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztranc_group_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztranc_group_heuristic_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pctranu
*/
double hscal_pctranu_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ctranu_benchmark(
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pctranu_modelp_noauto(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pctranu_modelp(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pctranu_recon(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pctranu_timeof(
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pctranu_group_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctranu_group_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctranu_group_heuristic_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pztranu
*/
double hscal_pztranu_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ztranu_benchmark(
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
);

int hscal_pack_pztranu_modelp_noauto(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pztranu_modelp(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pztranu_recon(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pztranu_timeof(
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pztranu_group_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztranu_group_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztranu_group_heuristic_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pstran
*/
double hscal_pstran_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_stran_benchmark(
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *beta, 
    hscalreal *c__, 
    int     *ldc
);

int hscal_pack_pstran_modelp_noauto(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pstran_modelp(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pstran_recon(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pstran_timeof(
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pstran_group_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstran_group_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstran_group_heuristic_auto_create(
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdtran
*/
double hscal_pdtran_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dtran_benchmark(
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *beta, 
    hscaldoublereal *c__, 
    int     *ldc
);

int hscal_pack_pdtran_modelp_noauto(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdtran_modelp(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pdtran_recon(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pdtran_timeof(
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdtran_group_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtran_group_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtran_group_heuristic_auto_create(
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCTRMM
*/
double hscal_pctrmm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ctrmm_benchmark(
    char    *side, 
    char    *uplo, 
    char    *transa, 
    char    *diag, 
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *b, 
    int     *ldb
);

int hscal_pack_pctrmm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pctrmm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pctrmm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pctrmm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pctrmm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctrmm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctrmm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pstrmm
*/
double hscal_pstrmm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_strmm_benchmark(
    char    *side, 
    char    *uplo, 
    char    *transa, 
    char    *diag, 
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *b, 
    int     *ldb
);

int hscal_pack_pstrmm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pstrmm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pstrmm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pstrmm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pstrmm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstrmm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstrmm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdtrmm
*/
double hscal_pdtrmm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dtrmm_benchmark(
    char    *side, 
    char    *uplo, 
    char    *transa, 
    char    *diag, 
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *b, 
    int     *ldb
);

int hscal_pack_pdtrmm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdtrmm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pdtrmm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pdtrmm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdtrmm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtrmm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtrmm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pztrmm
*/
double hscal_pztrmm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ztrmm_benchmark(
    char    *side, 
    char    *uplo, 
    char    *transa, 
    char    *diag, 
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *b, 
    int     *ldb
);

int hscal_pack_pztrmm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pztrmm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pztrmm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pztrmm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pztrmm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztrmm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztrmm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCGEMM
*/
double hscal_pcgemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cgemm_benchmark(
    char *transa, 
    char *transb, 
    int     *m, 
    int     *n, 
    int     *k, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *b, 
    int     *ldb, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pcgemm_modelp_noauto(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgemm_modelp(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pcgemm_recon(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pcgemm_timeof(
   F_CHAR_T TRANSA,
   F_CHAR_T TRANSB,
   int * M,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgemm_group_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgemm_group_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgemm_group_heuristic_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdgemm
*/
double hscal_pdgemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dgemm_benchmark(
    char *transa, 
    char *transb, 
    int     *m, 
    int     *n, 
    int     *k, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *b, 
    int     *ldb, 
    hscaldoublereal *beta, 
    hscaldoublereal *c__, 
    int     *ldc
);

int hscal_pack_pdgemm_modelp_noauto(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgemm_modelp(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pdgemm_recon(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pdgemm_timeof(
   F_CHAR_T TRANSA,
   F_CHAR_T TRANSB,
   int * M,
   int * N,
   int * K,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgemm_group_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgemm_group_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgemm_group_heuristic_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psgemm
*/
double hscal_psgemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_sgemm_benchmark(
    char *transa, 
    char *transb, 
    int     *m, 
    int     *n, 
    int     *k, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *b, 
    int     *ldb, 
    hscalreal *beta, 
    hscalreal *c__, 
    int     *ldc
);

int hscal_pack_psgemm_modelp_noauto(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgemm_modelp(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_psgemm_recon(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_psgemm_timeof(
   F_CHAR_T TRANSA,
   F_CHAR_T TRANSB,
   int * M,
   int * N,
   int * K,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgemm_group_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgemm_group_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgemm_group_heuristic_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzgemm
*/
double hscal_pzgemm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zgemm_benchmark(
    char *transa, 
    char *transb, 
    int     *m, 
    int     *n, 
    int     *k, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *b, 
    int     *ldb, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
);

int hscal_pack_pzgemm_modelp_noauto(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgemm_modelp(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pzgemm_recon(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pzgemm_timeof(
   F_CHAR_T TRANSA,
   F_CHAR_T TRANSB,
   int * M,
   int * N,
   int * K,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgemm_group_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgemm_group_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgemm_group_heuristic_auto_create(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pctradd
*/
double hscal_pctradd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ctradd_benchmark(
    char *uplo, 
    char *trans, 
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pctradd_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pctradd_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pctradd_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pctradd_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pctradd_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctradd_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctradd_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdtradd
*/
double hscal_pdtradd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dtradd_benchmark(
    char *uplo, 
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *beta, 
    hscaldoublereal *c__, 
    int     *ldc
);

int hscal_pack_pdtradd_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdtradd_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pdtradd_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pdtradd_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdtradd_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtradd_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtradd_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pstradd
*/
double hscal_pstradd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_stradd_benchmark(
    char *uplo, 
    char *trans, 
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *beta, 
    hscalreal *c__, 
    int     *ldc
);

int hscal_pack_pstradd_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pstradd_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pstradd_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pstradd_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pstradd_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstradd_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstradd_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pztradd
*/
double hscal_pztradd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ztradd_benchmark(
    char *uplo, 
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
);

int hscal_pack_pztradd_modelp_noauto(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pztradd_modelp(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pztradd_recon(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pztradd_timeof(
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pztradd_group_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztradd_group_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztradd_group_heuristic_auto_create(
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pcgeadd
*/
double hscal_pcgeadd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_cgeadd_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *beta, 
    hscalcomplex *c__, 
    int     *ldc
);

int hscal_pack_pcgeadd_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pcgeadd_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pcgeadd_recon(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pcgeadd_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pcgeadd_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgeadd_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pcgeadd_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdgeadd
*/
double hscal_pdgeadd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dgeadd_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *beta, 
    hscaldoublereal *c__, 
    int     *ldc
);

int hscal_pack_pdgeadd_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdgeadd_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pdgeadd_recon(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pdgeadd_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdgeadd_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgeadd_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdgeadd_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* psgeadd
*/
double hscal_psgeadd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_sgeadd_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *beta, 
    hscalreal *c__, 
    int     *ldc
);

int hscal_pack_psgeadd_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_psgeadd_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_psgeadd_recon(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_psgeadd_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   float * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_psgeadd_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgeadd_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_psgeadd_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   float * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pzgeadd
*/
double hscal_pzgeadd_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_zgeadd_benchmark(
    char *trans, 
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *beta, 
    hscaldoublecomplex *c__, 
    int     *ldc
);

int hscal_pack_pzgeadd_modelp_noauto(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pzgeadd_modelp(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
);

int hscal_pzgeadd_recon(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
);

double hscal_pzgeadd_timeof(
   F_CHAR_T TRANS,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   double * BETA,
   int * IC,
   int * JC,
   int * DESCC,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pzgeadd_group_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgeadd_group_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pzgeadd_group_heuristic_auto_create(
   F_CHAR_T TRANS, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* PCTRSM
*/
double hscal_pctrsm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ctrsm_benchmark(
    char *side,  
    char *uplo, 
    char *transa, 
    char *diag, 
    int     *m, 
    int     *n, 
    hscalcomplex *alpha, 
    hscalcomplex *a, 
    int     *lda, 
    hscalcomplex *b, 
    int     *ldb
);

int hscal_pack_pctrsm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pctrsm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pctrsm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pctrsm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pctrsm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctrsm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pctrsm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pdtrsm
*/
double hscal_pdtrsm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_dtrsm_benchmark(
    char *side,  
    char *uplo, 
    char *transa, 
    char *diag, 
    int     *m, 
    int     *n, 
    hscaldoublereal *alpha, 
    hscaldoublereal *a, 
    int     *lda, 
    hscaldoublereal *b, 
    int     *ldb
);

int hscal_pack_pdtrsm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pdtrsm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pdtrsm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pdtrsm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pdtrsm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtrsm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pdtrsm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pstrsm
*/
double hscal_pstrsm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_strsm_benchmark(
    char *side,  
    char *uplo, 
    char *transa, 
    char *diag, 
    int     *m, 
    int     *n, 
    hscalreal *alpha, 
    hscalreal *a, 
    int     *lda, 
    hscalreal *b, 
    int     *ldb
);

int hscal_pack_pstrsm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pstrsm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pstrsm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pstrsm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   float * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pstrsm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstrsm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pstrsm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   float * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

/*
* pztrsm
*/
double hscal_pztrsm_benchmark_function
(
     const void* input_p,
     int num_of_p,
     void* output_p
);

int hscal_ztrsm_benchmark(
    char *side,  
    char *uplo, 
    char *transa, 
    char *diag, 
    int     *m, 
    int     *n, 
    hscaldoublecomplex *alpha, 
    hscaldoublecomplex *a, 
    int     *lda, 
    hscaldoublecomplex *b, 
    int     *ldb
);

int hscal_pack_pztrsm_modelp_noauto(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   int **model_params,
   int *model_count
);

int hscal_pack_pztrsm_modelp(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int **model_params,
   int *model_count
);

int hscal_pztrsm_recon(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB
);

double hscal_pztrsm_timeof(
   F_CHAR_T SIDE,
   F_CHAR_T UPLO,
   F_CHAR_T TRANS,
   F_CHAR_T DIAG,
   int * M,
   int * N,
   double * ALPHA,
   int * IA,
   int * JA,
   int * DESCA,
   int * IB,
   int * JB,
   int * DESCB,
   int * nprow,
   int * npcol,
   const HMPI_Model* model
);

int hscal_pztrsm_group_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG,
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   int * nprow,
   int * npcol,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztrsm_group_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model
);

int hscal_pztrsm_group_heuristic_auto_create(
   F_CHAR_T SIDE, 
   F_CHAR_T UPLO, 
   F_CHAR_T TRANS, 
   F_CHAR_T DIAG, 
   int * M, 
   int * N,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   HMPI_Group* gid,
   const HMPI_Model* model,
   HMPI_Heuristic_function hfunc
);

#endif /* __HMPI_SCALAPACK_PBLAS_LEVEL3_INCLUDE_HH */
