
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK                          *
*                 SPECIAL LLD DETERMINATION                        *
*                 FOR SCALAPACK ROUTINES                           *
*                     PBLAS MPC Macros                             *
*                         (INTERNAL)                               *
*                 =======================                          *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HSCALAPACK_PBLAS_MPC_MACROS_SPECIAL_LLD_HH_
#define __HSCALAPACK_PBLAS_MPC_MACROS_SPECIAL_LLD_HH_

/*
* When the PBLAS routines are called from the 
* ScaLAPACK performance models, we should make sure that
* the LLD component of the descriptor array is determined in
* the same way as in the performance models
*
* One way is to use the LLDs determined in the ScaLAPACK 
* performance models to be used in the PBLAS performance models
* but that would mean passing an array of descriptor arrays for all
* the processes to the PBLAS macros, which in turn would pass this on
* to the macros called from it. Maybe looked in the future.
*
*/

/*
* ------------------------------------------------------------------
*                         PBLAS level1
* ------------------------------------------------------------------
*/
#define    HSCAL_PXSWAP_INIT_LLDS( \
                  N, \
                  Xd, IX, INCX, \
                  Yd, IY, INCY, \
                  myrow, nprow \
           ) \
              { \
                 Xd[LLD_] = max(1, PB_Cnumroc(Xd[M_], \
                                              0, \
                                              Xd[IMB_], \
                                              Xd[MB_], \
                                              myrow, \
                                              Xd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Yd[LLD_] = max(1, PB_Cnumroc(Yd[M_], \
                                              0, \
                                              Yd[IMB_], \
                                              Yd[MB_], \
                                              myrow, \
                                              Yd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXSCAL_INIT_LLDS( \
                  N, \
                  Xd, IX, INCX, \
                  myrow, nprow \
           ) \
              { \
                 Xd[LLD_] = max(1, PB_Cnumroc(Xd[M_], \
                                              0, \
                                              Xd[IMB_], \
                                              Xd[MB_], \
                                              myrow, \
                                              Xd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXCOPY_INIT_LLDS           HSCAL_PXSWAP_INIT_LLDS
#define    HSCAL_PXAXPY_INIT_LLDS           HSCAL_PXSWAP_INIT_LLDS
#define    HSCAL_PXASUM_INIT_LLDS           HSCAL_PXSCAL_INIT_LLDS
#define    HSCAL_PXAMAX_INIT_LLDS           HSCAL_PXSCAL_INIT_LLDS
#define    HSCAL_PXNRM2_INIT_LLDS           HSCAL_PXSCAL_INIT_LLDS
#define    HSCAL_PXDOTC_INIT_LLDS           HSCAL_PXSWAP_INIT_LLDS
#define    HSCAL_PXDOTU_INIT_LLDS           HSCAL_PXSWAP_INIT_LLDS
#define    HSCAL_PXDOT_INIT_LLDS            HSCAL_PXSWAP_INIT_LLDS

/*
* ------------------------------------------------------------------
*       PBLAS level2
* ------------------------------------------------------------------
*/
#define    HSCAL_PXAGEMV_INIT_LLDS( \
                  notran, \
                  M, N, \
                  Ad, \
                  Xd, IX, INCX, \
                  Yd, IY, INCY, \
                  myrow, nprow \
           ) \
              { \
                 Ad[LLD_] = max(1, PB_Cnumroc(Ad[M_], \
                                              0, \
                                              Ad[IMB_], \
                                              Ad[MB_], \
                                              myrow, \
                                              Ad[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Xd[LLD_] = max(1, PB_Cnumroc(Xd[M_], \
                                              0, \
                                              Xd[IMB_], \
                                              Xd[MB_], \
                                              myrow, \
                                              Xd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Yd[LLD_] = max(1, PB_Cnumroc(Yd[M_], \
                                              0, \
                                              Yd[IMB_], \
                                              Yd[MB_], \
                                              myrow, \
                                              Yd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXGEMV_INIT_LLDS           HSCAL_PXAGEMV_INIT_LLDS

#define    HSCAL_PXAHEMV_INIT_LLDS( \
                  N, \
                  Ad, \
                  Xd, IX, INCX, \
                  Yd, IY, INCY, \
                  myrow, nprow \
           ) \
              { \
                 Ad[LLD_] = max(1, PB_Cnumroc(Ad[M_], \
                                              0, \
                                              Ad[IMB_], \
                                              Ad[MB_], \
                                              myrow, \
                                              Ad[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Xd[LLD_] = max(1, PB_Cnumroc(Xd[M_], \
                                              0, \
                                              Xd[IMB_], \
                                              Xd[MB_], \
                                              myrow, \
                                              Xd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Yd[LLD_] = max(1, PB_Cnumroc(Yd[M_], \
                                              0, \
                                              Yd[IMB_], \
                                              Yd[MB_], \
                                              myrow, \
                                              Yd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXHEMV_INIT_LLDS            HSCAL_PXAHEMV_INIT_LLDS
#define    HSCAL_PXASYMV_INIT_LLDS           HSCAL_PXAHEMV_INIT_LLDS
#define    HSCAL_PXSYMV_INIT_LLDS            HSCAL_PXASYMV_INIT_LLDS
#define    HSCAL_PXATRMV_INIT_LLDS           HSCAL_PXAHEMV_INIT_LLDS

#define    HSCAL_PXTRMV_INIT_LLDS( \
                  N, \
                  Ad, \
                  Xd, IX, INCX, \
                  myrow, nprow \
           ) \
              { \
                 Ad[LLD_] = max(1, PB_Cnumroc(Ad[M_], \
                                              0, \
                                              Ad[IMB_], \
                                              Ad[MB_], \
                                              myrow, \
                                              Ad[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Xd[LLD_] = max(1, PB_Cnumroc(Xd[M_], \
                                              0, \
                                              Xd[IMB_], \
                                              Xd[MB_], \
                                              myrow, \
                                              Xd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXTRSV_INIT_LLDS          HSCAL_PXTRMV_INIT_LLDS

#define    HSCAL_PXPTRSV_INIT_LLDS( \
                  Ad, \
                  myrow, nprow \
           ) \
              { \
                 Ad[LLD_] = max(1, PB_Cnumroc(Ad[M_], \
                                              0, \
                                              Ad[IMB_], \
                                              Ad[MB_], \
                                              myrow, \
                                              Ad[RSRC_], \
                                              nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXGERC_INIT_LLDS( \
                  N, \
                  Ad, \
                  Xd, IX, INCX, \
                  Yd, IY, INCY, \
                  myrow, nprow \
           ) \
              { \
                 Ad[LLD_] = max(1, PB_Cnumroc(Ad[M_], \
                                              0, \
                                              Ad[IMB_], \
                                              Ad[MB_], \
                                              myrow, \
                                              Ad[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Xd[LLD_] = max(1, PB_Cnumroc(Xd[M_], \
                                              0, \
                                              Xd[IMB_], \
                                              Xd[MB_], \
                                              myrow, \
                                              Xd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Yd[LLD_] = max(1, PB_Cnumroc(Yd[M_], \
                                              0, \
                                              Yd[IMB_], \
                                              Yd[MB_], \
                                              myrow, \
                                              Yd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXGERU_INIT_LLDS            HSCAL_PXGERC_INIT_LLDS
#define    HSCAL_PXGER_INIT_LLDS             HSCAL_PXGERC_INIT_LLDS

#define    HSCAL_PXHER_INIT_LLDS( \
                  N, \
                  Ad, \
                  Xd, IX, INCX, \
                  myrow, nprow \
           ) \
              { \
                 Ad[LLD_] = max(1, PB_Cnumroc(Ad[M_], \
                                              0, \
                                              Ad[IMB_], \
                                              Ad[MB_], \
                                              myrow, \
                                              Ad[RSRC_], \
                                              nprow \
                               ) \
                 ); \
                 Xd[LLD_] = max(1, PB_Cnumroc(Xd[M_], \
                                              0, \
                                              Xd[IMB_], \
                                              Xd[MB_], \
                                              myrow, \
                                              Xd[RSRC_], \
                                              nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXSYR_INIT_LLDS             HSCAL_PXHER_INIT_LLDS
#define    HSCAL_PXHER2_INIT_LLDS            HSCAL_PXAHEMV_INIT_LLDS
#define    HSCAL_PXSYR2_INIT_LLDS            HSCAL_PXHER2_INIT_LLDS

/*
* ------------------------------------------------------------------
*       PBLAS level3
* ------------------------------------------------------------------
*/
#define    HSCAL_PXSYRKA_INIT_LLDS( \
                  notran, \
                  N, K, \
                  DESCA, \
                  DESCC, \
                  myrow, nprow \
           ) \
              { \
                 DESCA[LLD_] = max(1, PB_Cnumroc(DESCA[M_], \
                                                 0, \
                                                 DESCA[IMB_], \
                                                 DESCA[MB_], \
                                                 myrow, \
                                                 DESCA[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCC[LLD_] = max(1, PB_Cnumroc(DESCC[M_], \
                                                 0, \
                                                 DESCC[IMB_], \
                                                 DESCC[MB_], \
                                                 myrow, \
                                                 DESCC[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXSYR2KA_INIT_LLDS( \
                  notran, \
                  N, K, \
                  DESCA, \
                  DESCB, \
                  DESCC, \
                  myrow, nprow \
           ) \
              { \
                 DESCA[LLD_] = max(1, PB_Cnumroc(DESCA[M_], \
                                                 0, \
                                                 DESCA[IMB_], \
                                                 DESCA[MB_], \
                                                 myrow, \
                                                 DESCA[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCB[LLD_] = max(1, PB_Cnumroc(DESCB[M_], \
                                                 0, \
                                                 DESCB[IMB_], \
                                                 DESCB[MB_], \
                                                 myrow, \
                                                 DESCB[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCC[LLD_] = max(1, PB_Cnumroc(DESCC[M_], \
                                                 0, \
                                                 DESCC[IMB_], \
                                                 DESCC[MB_], \
                                                 myrow, \
                                                 DESCC[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXSYRKAC_INIT_LLDS             HSCAL_PXSYRKA_INIT_LLDS
#define    HSCAL_PXSYR2KAC_INIT_LLDS            HSCAL_PXSYR2KA_INIT_LLDS

#define    HSCAL_PXSYRK_INIT_LLDS( \
                  DESCA, \
                  DESCC, \
                  myrow, nprow \
           ) \
              { \
                 DESCA[LLD_] = max(1, PB_Cnumroc(DESCA[M_], \
                                                 0, \
                                                 DESCA[IMB_], \
                                                 DESCA[MB_], \
                                                 myrow, \
                                                 DESCA[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCC[LLD_] = max(1, PB_Cnumroc(DESCC[M_], \
                                                 0, \
                                                 DESCC[IMB_], \
                                                 DESCC[MB_], \
                                                 myrow, \
                                                 DESCC[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXHERK_INIT_LLDS             HSCAL_PXSYRK_INIT_LLDS

#define    HSCAL_PXSYR2K_INIT_LLDS( \
                  DESCA, \
                  DESCB, \
                  DESCC, \
                  myrow, nprow \
           ) \
              { \
                 DESCA[LLD_] = max(1, PB_Cnumroc(DESCA[M_], \
                                                 0, \
                                                 DESCA[IMB_], \
                                                 DESCA[MB_], \
                                                 myrow, \
                                                 DESCA[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCB[LLD_] = max(1, PB_Cnumroc(DESCB[M_], \
                                                 0, \
                                                 DESCB[IMB_], \
                                                 DESCB[MB_], \
                                                 myrow, \
                                                 DESCB[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCC[LLD_] = max(1, PB_Cnumroc(DESCC[M_], \
                                                 0, \
                                                 DESCC[IMB_], \
                                                 DESCC[MB_], \
                                                 myrow, \
                                                 DESCC[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXHER2K_INIT_LLDS             HSCAL_PXSYR2K_INIT_LLDS

#define    HSCAL_PXSYMM_INIT_LLDS( \
                  DESCA, \
                  DESCB, \
                  DESCC, \
                  myrow, nprow \
           ) \
              { \
                 DESCA[LLD_] = max(1, PB_Cnumroc(DESCA[M_], \
                                                 0, \
                                                 DESCA[IMB_], \
                                                 DESCA[MB_], \
                                                 myrow, \
                                                 DESCA[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCB[LLD_] = max(1, PB_Cnumroc(DESCB[M_], \
                                                 0, \
                                                 DESCB[IMB_], \
                                                 DESCB[MB_], \
                                                 myrow, \
                                                 DESCB[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCC[LLD_] = max(1, PB_Cnumroc(DESCC[M_], \
                                                 0, \
                                                 DESCC[IMB_], \
                                                 DESCC[MB_], \
                                                 myrow, \
                                                 DESCC[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXHEMM_INIT_LLDS             HSCAL_PXSYMM_INIT_LLDS
#define    HSCAL_PXTRANC_INIT_LLDS            HSCAL_PXSYRK_INIT_LLDS
#define    HSCAL_PXTRANU_INIT_LLDS            HSCAL_PXSYRK_INIT_LLDS
#define    HSCAL_PXTRAN_INIT_LLDS             HSCAL_PXSYRK_INIT_LLDS
#define    HSCAL_PXPTRAN_INIT_LLDS            HSCAL_PXSYRK_INIT_LLDS

#define    HSCAL_PXPTRMMB_INIT_LLDS( \
                  DESCA, \
                  DESCB, \
                  myrow, nprow \
           ) \
              { \
                 DESCA[LLD_] = max(1, PB_Cnumroc(DESCA[M_], \
                                                 0, \
                                                 DESCA[IMB_], \
                                                 DESCA[MB_], \
                                                 myrow, \
                                                 DESCA[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCB[LLD_] = max(1, PB_Cnumroc(DESCB[M_], \
                                                 0, \
                                                 DESCB[IMB_], \
                                                 DESCB[MB_], \
                                                 myrow, \
                                                 DESCB[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXPTRMMAB_INIT_LLDS            HSCAL_PXPTRMMB_INIT_LLDS
#define    HSCAL_PXTRMM_INIT_LLDS               HSCAL_PXPTRMMB_INIT_LLDS
#define    HSCAL_PXGEMM_INIT_LLDS               HSCAL_PXSYR2K_INIT_LLDS
#define    HSCAL_PXGEADD_INIT_LLDS              HSCAL_PXSYRK_INIT_LLDS
#define    HSCAL_PXTRADD_INIT_LLDS              HSCAL_PXSYRK_INIT_LLDS
#define    HSCAL_PXPTRADD_INIT_LLDS             HSCAL_PXSYRK_INIT_LLDS
#define    HSCAL_PXPGEADD_INIT_LLDS             HSCAL_PXSYRK_INIT_LLDS
#define    HSCAL_PXCPTRSMAB_INIT_LLDS           HSCAL_PXPTRMMB_INIT_LLDS
#define    HSCAL_PXCPTRSMB_INIT_LLDS( \
                  DESCA, \
                  DESCB, \
                  myrow, nprow \
           ) \
              { \
                 DESCA[LLD_] = max(1, PB_Cnumroc(DESCA[M_], \
                                                 0, \
                                                 DESCA[IMB_], \
                                                 DESCA[MB_], \
                                                 myrow, \
                                                 DESCA[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
                 DESCB[LLD_] = max(1, PB_Cnumroc(DESCB[M_], \
                                                 0, \
                                                 DESCB[IMB_], \
                                                 DESCB[MB_], \
                                                 myrow, \
                                                 DESCB[RSRC_], \
                                                 nprow \
                               ) \
                 ); \
              }

#define    HSCAL_PXCPTRSMAB0_INIT_LLDS          HSCAL_PXSYR2K_INIT_LLDS
#define    HSCAL_PXCPTRSMAB00_INIT_LLDS         HSCAL_PXSYRK_INIT_LLDS
#define    HSCAL_PXCPTRSM_INIT_LLDS             HSCAL_PXPTRSV_INIT_LLDS
#define    HSCAL_PXTRSM_INIT_LLDS               HSCAL_PXPTRMMB_INIT_LLDS
#define    HSCAL_PXCPTRSMAB1_INIT_LLDS          HSCAL_PXSYR2K_INIT_LLDS

#endif  /* __HSCALAPACK_PBLAS_MPC_MACROS_SPECIAL_LLD_HH_ */

