
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK PTZBLAS_MACROS           *
*                         (INTERNAL)                               *
*                 ==========================================       *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_PTZBLAS_MACROS_INCLUDE_HH
#define __HMPI_SCALAPACK_PTZBLAS_MACROS_INCLUDE_HH

/*
* ------------------------------------------------------------------
*       FUNCTION PROTOTYPES and mpC MACROS FOR
*       PBLAS COMPUTATIONS (PTZBLAS)
* ------------------------------------------------------------------
*/
#define add_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    add( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cmmadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define cmmadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cmmadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cmmcadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define cmmcadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cmmcadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cmmtadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define cmmtadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cmmtadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cmmtcadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define cmmtcadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cmmtcadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dmmadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define dmmadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dmmadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dmmcadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define dmmcadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dmmcadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dmmtadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define dmmtadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dmmtadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dmmtcadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define dmmtcadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dmmtcadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_immadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define immadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_immadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_immtadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define immtadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_immtadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_smmadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define smmadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_smmadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_smmcadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define smmcadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_smmcadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_smmtadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define smmtadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_smmtadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_smmtcadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define smmtcadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_smmtcadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zmmadd(
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define zmmadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zmmadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zmmcadd(
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define zmmcadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zmmcadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zmmtadd(
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define zmmtadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zmmtadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zmmtcadd(
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define zmmtcadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zmmtcadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cset(
    int *n, 
    char *alpha, 
    int *incx,
    double *tcomp
);

#define cset_tcomp_macro(n, alpha, incx, \
                   nprow, npcol, myrow, mycol, tcomp ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cset( \
         (n), (alpha), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dset(
    int *n, 
    char *alpha,
    int *incx,
    double *tcomp
);

#define dset_tcomp_macro(n, alpha, incx, \
                   nprow, npcol, myrow, mycol, tcomp ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dset( \
         (n), (alpha), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sset(
    int *n, 
    char *alpha, 
    int *incx,
    double *tcomp
);

#define sset_tcomp_macro(n, alpha, incx, \
                   nprow, npcol, myrow, mycol, tcomp ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sset( \
         (n), (alpha), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zset(
    int *n, 
    char *alpha,
    int *incx,
    double *tcomp
);

#define zset_tcomp_macro(n, alpha, incx, \
                   nprow, npcol, myrow, mycol, tcomp ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zset( \
         (n), (alpha), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctzpad(
    char *uplo, 
    char *herm, 
    int *m, 
    int *n,
    int *ioffd, 
    char *alpha, 
    char *beta, 
    int *lda,
    double *tcomp
);

#define ctzpad_tcomp_macro( \
              uplo, herm, m, n, \
              ioffd, alpha, beta, lda, \
              nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctzpad( \
         (uplo), (herm), (m), (n), \
         (ioffd), (alpha), (beta), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtzpad(
    char *uplo, 
    char *herm, 
    int *m, 
    int *n,
    int *ioffd, 
    char *alpha, 
    char *beta, 
    int *lda,
    double *tcomp
);

#define dtzpad_tcomp_macro( \
              uplo, herm, m, n, \
              ioffd, alpha, beta, lda, \
              nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtzpad( \
         (uplo), (herm), (m), (n), \
         (ioffd), (alpha), (beta), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_itzpad(
    char *uplo, 
    char *herm, 
    int *m, 
    int *n,
    int *ioffd, 
    char *alpha, 
    char *beta, 
    int *lda,
    double *tcomp
);

#define itzpad_tcomp_macro( \
              uplo, herm, m, n, \
              ioffd, alpha, beta, lda, \
              nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_itzpad( \
         (uplo), (herm), (m), (n), \
         (ioffd), (alpha), (beta), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_stzpad(
    char *uplo, 
    char *herm, 
    int *m, 
    int *n,
    int *ioffd, 
    char *alpha, 
    char *beta, 
    int *lda,
    double *tcomp
);

#define stzpad_tcomp_macro( \
              uplo, herm, m, n, \
              ioffd, alpha, beta, lda, \
              nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_stzpad( \
         (uplo), (herm), (m), (n), \
         (ioffd), (alpha), (beta), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztzpad(
    char *uplo, 
    char *herm, 
    int *m, 
    int *n,
    int *ioffd, 
    char *alpha, 
    char *beta, 
    int *lda,
    double *tcomp
);

#define ztzpad_tcomp_macro( \
              uplo, herm, m, n, \
              ioffd, alpha, beta, lda, \
              nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztzpad( \
         (uplo), (herm), (m), (n), \
         (ioffd), (alpha), (beta), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctzscal(
    char *uplo, 
    int *m, 
    int *n, 
    int *ioffd, 
    char *alpha, 
    int *lda,
    double *tcomp
);

#define ctzscal_tcomp_macro( \
               uplo, m, n, ioffd, \
               alpha, lda, \
               nprow, npcol, myrow, mycol, \
               tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctzscal( \
         (uplo), (m), (n), \
         (ioffd), (alpha), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtzscal(
    char *uplo, 
    int *m, 
    int *n, 
    int *ioffd, 
    char *alpha, 
    int *lda,
    double *tcomp
);

#define dtzscal_tcomp_macro( \
               uplo, m, n, ioffd, \
               alpha, lda, \
               nprow, npcol, myrow, mycol, \
               tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtzscal( \
         (uplo), (m), (n), \
         (ioffd), (alpha), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_stzscal(
    char *uplo, 
    int *m, 
    int *n, 
    int *ioffd, 
    char *alpha, 
    int *lda,
    double *tcomp
);

#define stzscal_tcomp_macro( \
               uplo, m, n, ioffd, \
               alpha, lda, \
               nprow, npcol, myrow, mycol, \
               tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_stzscal( \
         (uplo), (m), (n), \
         (ioffd), (alpha), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztzscal(
    char *uplo, 
    int *m, 
    int *n, 
    int *ioffd, 
    char *alpha, 
    int *lda,
    double *tcomp
);

#define ztzscal_tcomp_macro( \
               uplo, m, n, ioffd, \
               alpha, lda, \
               nprow, npcol, myrow, mycol, \
               tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztzscal( \
         (uplo), (m), (n), \
         (ioffd), (alpha), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_casymv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
);

#define casymv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_casymv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zasymv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
);

#define zasymv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zasymv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dasymv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
);

#define dasymv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dasymv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_sasymv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
);

#define sasymv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_sasymv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_catrmv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

int hscal_csymv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
);

#define csymv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_csymv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zsymv(
    char *uplo, 
    int *n, 
    char *calpha, 
    int *lda, 
    int *incx, 
    char *cbeta, 
    int *incy,
    double *tcomp
);

#define zsymv_tcomp_macro( \
             uplo, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zsymv( \
         (uplo), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

#define catrmv_tcomp_macro( \
             uplo, trans, diag, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_catrmv( \
         (uplo), (trans), (diag), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_datrmv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define datrmv_tcomp_macro( \
             uplo, trans, diag, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_datrmv( \
         (uplo), (trans), (diag), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_satrmv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define satrmv_tcomp_macro( \
             uplo, trans, diag, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_satrmv( \
         (uplo), (trans), (diag), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zatrmv(
    char *uplo, 
    char *trans, 
    char *diag, 
    int *n, 
    char *alpha, 
    int *lda, 
    int *incx, 
    char *beta, 
    int *incy,
    double *tcomp
);

#define zatrmv_tcomp_macro( \
             uplo, trans, diag, n, alpha, \
             lda, incx, beta, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zatrmv( \
         (uplo), (trans), (diag), (n), (alpha), \
         (lda), (incx), (beta), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctzpadcpy(
    char *uplo, 
    char *diag, 
    int *m, 
    int *n, 
    int *ioffd, 
    int *lda, 
    int *ldb, 
    double *tcomp
);

#define ctzpadcpy_tcomp_macro( \
             uplo, diag, m, n, ioffd, \
             lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctzpadcpy( \
         (uplo), (diag), (m), (n), (ioffd), \
         (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dtzpadcpy(
    char *uplo, 
    char *diag, 
    int *m, 
    int *n, 
    int *ioffd, 
    int *lda, 
    int *ldb, 
    double *tcomp
);

#define dtzpadcpy_tcomp_macro( \
             uplo, diag, m, n, ioffd, \
             lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dtzpadcpy( \
         (uplo), (diag), (m), (n), (ioffd), \
         (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_stzpadcpy(
    char *uplo, 
    char *diag, 
    int *m, 
    int *n, 
    int *ioffd, 
    int *lda, 
    int *ldb, 
    double *tcomp
);

#define stzpadcpy_tcomp_macro( \
             uplo, diag, m, n, ioffd, \
             lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_stzpadcpy( \
         (uplo), (diag), (m), (n), (ioffd), \
         (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztzpadcpy(
    char *uplo, 
    char *diag, 
    int *m, 
    int *n, 
    int *ioffd, 
    int *lda, 
    int *ldb, 
    double *tcomp
);

#define ztzpadcpy_tcomp_macro( \
             uplo, diag, m, n, ioffd, \
             lda, ldb, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztzpadcpy( \
         (uplo), (diag), (m), (n), (ioffd), \
         (lda), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ccshft(
    int *m,
    int *n,
    int *offset,
    int *lda,
    double *tcomp
);

#define ccshft_tcomp_macro( \
             m, n, offset, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ccshft( \
         (m), (n), (offset), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dcshft(
    int *m,
    int *n,
    int *offset,
    int *lda,
    double *tcomp
);

#define dcshft_tcomp_macro( \
             m, n, offset, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dcshft( \
         (m), (n), (offset), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_scshft(
    int *m,
    int *n,
    int *offset,
    int *lda,
    double *tcomp
);

#define scshft_tcomp_macro( \
             m, n, offset, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_scshft( \
         (m), (n), (offset), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zcshft(
    int *m,
    int *n,
    int *offset,
    int *lda,
    double *tcomp
);

#define zcshft_tcomp_macro( \
             m, n, offset, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zcshft( \
         (m), (n), (offset), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_crshft(
    int *m,
    int *n,
    int *offset,
    int *lda,
    double *tcomp
);

#define crshft_tcomp_macro( \
             m, n, offset, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_crshft( \
         (m), (n), (offset), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_drshft(
    int *m,
    int *n,
    int *offset,
    int *lda,
    double *tcomp
);

#define drshft_tcomp_macro( \
             m, n, offset, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_drshft( \
         (m), (n), (offset), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_srshft(
    int *m,
    int *n,
    int *offset,
    int *lda,
    double *tcomp
);

#define srshft_tcomp_macro( \
             m, n, offset, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_srshft( \
         (m), (n), (offset), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zrshft(
    int *m,
    int *n,
    int *offset,
    int *lda,
    double *tcomp
);

#define zrshft_tcomp_macro( \
             m, n, offset, lda, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zrshft( \
         (m), (n), (offset), (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_chescal(
    char *uplo, 
    int *m, 
    int *n, 
    int *ioffd, 
    char *alpha, 
    int *lda,
    double *tcomp
);

#define chescal_tcomp_macro( \
               uplo, m, n, ioffd, \
               alpha, lda, \
               nprow, npcol, myrow, mycol, \
               tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_chescal( \
         (uplo), (m), (n), \
         (ioffd), (alpha), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zhescal(
    char *uplo, 
    int *m, 
    int *n, 
    int *ioffd, 
    char *alpha, 
    int *lda,
    double *tcomp
);

#define zhescal_tcomp_macro( \
               uplo, m, n, ioffd, \
               alpha, lda, \
               nprow, npcol, myrow, mycol, \
               tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zhescal( \
         (uplo), (m), (n), \
         (ioffd), (alpha), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cmmdda(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define cmmdda_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cmmdda( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cmmddac(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define cmmddac_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cmmddac( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cmmddat(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define cmmddat_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cmmddat( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cmmddact(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define cmmddact_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cmmddact( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dmmdda(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define dmmdda_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dmmdda( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dmmddac(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define dmmddac_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dmmddac( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dmmddat(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define dmmddat_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dmmddat( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dmmddact(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define dmmddact_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dmmddact( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_immadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define immadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_immadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_immtadd(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define immtadd_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_immtadd( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_smmdda(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define smmdda_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_smmdda( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_smmddac(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define smmddac_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_smmddac( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_smmddat(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define smmddat_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_smmddat( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_smmddact(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define smmddact_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_smmddact( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zmmdda(
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define zmmdda_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zmmdda( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zmmddac(
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define zmmddac_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zmmddac( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zmmddat(
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define zmmddat_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zmmddat( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zmmddact(
    int *m, 
    int *n, 
    char *alpha,
    int *lda, 
    char *beta,
    int *ldb,
    double *tcomp
);

#define zmmddact_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zmmddact( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_immdda(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define immdda_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_immdda( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_immddat(
    int *m, 
    int *n, 
    char *alpha, 
    int *lda, 
    char *beta, 
    int *ldb,
    double *tcomp
);

#define immddat_tcomp_macro( \
        m, n, alpha, lda, \
        beta, ldb, \
        nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_immddat( \
         (m), (n), (alpha), \
         (lda), (beta), (ldb), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dvasum(
    int *n,
    int *incx,
    double *tcomp
);

#define dvasum_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dvasum( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dzvasum(
    int *n,
    int *incx,
    double *tcomp
);

#define dzvasum_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dzvasum( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_svasum(
    int *n,
    int *incx,
    double *tcomp
);

#define svasum_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_svasum( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_scvasum(
    int *n,
    int *incx,
    double *tcomp
);

#define scvasum_tcomp_macro( \
             n, incx, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_scvasum( \
         (n), \
         (incx), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cvvdotc(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define cvvdotc_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cvvdotc( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_cvvdotu(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define cvvdotu_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_cvvdotu( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zvvdotc(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define zvvdotc_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zvvdotc( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zvvdotu(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define zvvdotu_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zvvdotu( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_dvvdot(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define dvvdot_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_dvvdot( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_svvdot(
    int *n,
    int *incx,
    int *incy,
    double *tcomp
);

#define svvdot_tcomp_macro( \
             n, incx, incy, \
             nprow, npcol, myrow, mycol, tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_svvdot( \
         (n), (incx), (incy), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ctzcnjg(
    char *uplo, 
    int *m, 
    int *n, 
    int *ioffd, 
    char *alpha, 
    int *lda,
    double *tcomp
);

#define ctzcnjg_tcomp_macro( \
               uplo, m, n, ioffd, \
               alpha, lda, \
               nprow, npcol, myrow, mycol, \
               tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ctzcnjg( \
         (uplo), (m), (n), \
         (ioffd), (alpha), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_ztzcnjg(
    char *uplo, 
    int *m, 
    int *n, 
    int *ioffd, 
    char *alpha, 
    int *lda,
    double *tcomp
);

#define ztzcnjg_tcomp_macro( \
               uplo, m, n, ioffd, \
               alpha, lda, \
               nprow, npcol, myrow, mycol, \
               tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_ztzcnjg( \
         (uplo), (m), (n), \
         (ioffd), (alpha), (lda), \
         (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_csyr(
    char *uplo,
    int *n,
    char *calpha,
    int *incx,
    int *lda,
    double *tcomp
);

#define csyr_tcomp_macro( \
             uplo, n, calpha, incx, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_csyr( \
         (uplo), (n), (calpha), (incx), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zsyr(
    char *uplo,
    int *n,
    char *calpha,
    int *incx,
    int *lda,
    double *tcomp
);

#define zsyr_tcomp_macro( \
             uplo, n, calpha, incx, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zsyr( \
         (uplo), (n), (calpha), (incx), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_csyr2(
    char *uplo, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy, 
    int *lda, 
    double *tcomp
);

#define csyr2_tcomp_macro( \
             uplo, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_csyr2( \
         (uplo), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

int hscal_zsyr2(
    char *uplo, 
    int *n, 
    char *calpha,
    int *incx, 
    int *incy, 
    int *lda, 
    double *tcomp
);

#define zsyr2_tcomp_macro( \
             uplo, n, calpha, incx, incy, lda, \
             nprow, npcol, myrow, mycol, \
             tcomp \
        ) \
if (tcomp != NULL) \
{ \
    double t; \
    hscal_zsyr2( \
         (uplo), (n), (calpha), (incx), (incy), \
         (lda), (&t) \
    ); \
    tcomp[HSCALRIX(myrow, mycol, nprow, npcol)] += t; \
}

#endif  /* __HMPI_SCALAPACK_PTZBLAS_MACROS_INCLUDE_HH */

