
#ifndef __HSCAL_BLACS_BDEF_H
#define __HSCAL_BLACS_BDEF_H

#ifndef NULL
#define NULL (void*)0
#endif

/*
 * Include the system dependant and user defined stuff
 */
#include "scalBconfig.h"

/*
 * Data type defining a scope for the BLACS
 */
typedef struct HSCALbLaCsScOpE HSCALBLACSSCOPE;
struct HSCALbLaCsScOpE
{
   int Np, Iam;
   char scope;
};
/*
 * Data type defining a context for the BLACS
 */
typedef struct HSCALbLaCsCoNtExT HSCALBLACSCONTEXT;
struct HSCALbLaCsCoNtExT
{
   HSCALBLACSSCOPE rscp, cscp, ascp, pscp; /* row, column, all, and pt2pt scopes */
   HSCALBLACSSCOPE *scp;                   /* pointer to present scope */
   int TopsRepeat;                    /* Use only repeatable topologies? */
   int TopsCohrnt;                    /* Use only coherent topologies? */
   int Nb_bs, Nr_bs;           /* for bcast general tree and multiring tops */
   int Nb_co, Nr_co;           /* for combine general tree and multiring tops */
};

/*
 *  Define the fortran 77 data types COMPLEX*8 (SCOMPLEX)
 *  and COMPLEX*16 (DCOMPLEX).
 */
typedef struct {double r, i;} DCOMPLEX;
typedef struct {float r, i;} SCOMPLEX;

/*
 * Error codes
 */
#define NORV 1          /* No receiver (only 1 proc in scoped op) */
#define NPOW2 2         /* Number of procs is not a power of 2 */
#define BADSCP 3        /* Scope not row, column or all */

/*
 * Data types
 */
#define INTEGER   3
#define SINGLE    4
#define DOUBLE    6
#define COMPLEX8  5
#define COMPLEX16 7

#define FULLCON 0      /* top is fully connected */

/*
 * Routine types
 */
#define RT_SD    1
#define RT_RV    2
#define RT_BS    3
#define RT_BR    4
#define RT_COMB  5

/*
 * Legal WHAT values for BLACS_SET
 */
#define SGET_SYSCONTXT    0
#define SGET_MSGIDS       1
#define SGET_DEBUGLVL     2
#define SGET_BLACSCONTXT 10
#define SGET_NR_BS       11
#define SGET_NB_BS       12
#define SGET_NR_CO       13
#define SGET_NB_CO       14
#define SGET_TOPSREPEAT  15
#define SGET_TOPSCOHRNT  16

/*
 * If we've got an ANSI standard C compiler, we can use void pointers,
 * otherwise use char pointers
 */
#ifdef __STDC__
#define BVOID void
#else
#define BVOID char
#endif

/*
 * Data types to use in the combine operations
 */
#define BI_DistType                  int

/*
 * ========================================================================
 *     PREPROCESSOR MACRO FUNCTIONS USED FOR OPTIMIZATION & CONVENIENCE
 * ========================================================================
 */

#define Mlowcase(C) ( ((C) > 64 && (C) < 91) ? (C) | 32 : (C) )

/*
 * Slightly modified gridinfo substitute
 */
#define Mgridinfo(ctxt, Ng0, nprow0, npcol0, myrow0, mycol0)\
{\
   (Ng0) = (ctxt)->ascp.Np;\
   (nprow0) = (ctxt)->cscp.Np;\
   (npcol0) = (ctxt)->rscp.Np;\
   (myrow0) = (ctxt)->cscp.Iam;\
   (mycol0) = (ctxt)->rscp.Iam;\
}

/*
 * These routines return coordinates based on nodes number, or node number
 * based on coordinates.  Those routines with v after the M return virtual
 * nodes numbers (i.e., in respect to the grid, not physical node numbers)
 * based on grid coordinates, or grid coordinates based on virtual node numbers.
 */
#define Mpcoord(ctxt, node, prow, pcol)\
{\
   (prow) = (node) / (ctxt)->rscp.Np;\
   (pcol) = (node) % (ctxt)->rscp.Np;\
}
#define Mvpcoord(ctxt, node, prow, pcol) \
        Mpcoord((ctxt), (node), (prow), (pcol));

#define Mkpnum(ctxt, prow, pcol)  ( (prow)*(ctxt)->rscp.Np+(pcol) )
#define Mvkpnum(ctxt, prow, pcol) ( (prow)*(ctxt)->rscp.Np+(pcol) )

/*
 * These macros allow for the funky function declarations and character handling
 * needed on the CRAY to have a C routine callable from fortran
 */
#define F_VOID_FUNC void
#define F_INT_FUNC  int
#define F_DOUBLE_FUNC double
#define F2C_CharTrans(c) *(c)

/*
 *  These macros allow for accessing values and addresses of parameters, which
 *  will be pointers if we're using fortran, and values if we're using C.
 */
#define Mpval(para) (para)
#define Mpaddress(para) (&(para))
#define Mwalltime Cdwalltime00

/*
 * Real and complex absolute values
 */
#define Rabs(x) ( (x) < 0 ? (x) * -1 : (x) )
#define Cabs(z) ( (((z).i) < 0 ? ((z).i) * -1 : ((z).i)) + (((z).r) < 0 ? ((z).r) * -1 : ((z).r)) )

/*
 * Figures the length of packed trapezoidal matrix
 */
#define trsize(diag, m, n, bytes, length)\
{\
   if ( (diag) == 'u' ) (length) = 1;\
   else (length) = 0;\
   if ( (m) > (n) )\
      (length) = ( (n) * ( (m) - (n) ) + ( (n)*(n) ) - ( (n)*(n) )/2 +\
                   (n)/2 - (n) * (length) ) * (bytes);\
   else\
      (length) = ( (m) * ( (n) - (m) ) + ( (m)*(m) ) - ( (m)*(m) )/2 +\
                   (m)/2 - (m) * (length) ) * (bytes);\
}

/*
 * =======================================
 *     FUNCTION PROTOTYPES
 * =======================================
 */
void BI_BlacsErr_nocomm(
       HSCALBLACSCONTEXT *ctxt,
       int line, char *file, char *form, ...);

void BI_BlacsWarn_nocomm(
       HSCALBLACSCONTEXT *ctxt,
       int line, char *file, char *form, ...);

void Cblacs_gridexit_nocomm(
     HSCALBLACSCONTEXT *ctxt
);

int Cblacs_gridinit_nocomm(
    HSCALBLACSCONTEXT **ctxt, 
    char *order, int nprow, int npcol
);

void Cblacs_gridmap_nocomm(
     HSCALBLACSCONTEXT **ctxt, int *usermap,
     int ldup, int nprow0, int npcol0
);

void Cblacs_get_nocomm(
     HSCALBLACSCONTEXT *ctxt, int what, int *val
);

#endif
