
/*******************************************************************
*                                                                  *
*                 Heterogeneous ScaLAPACK                          *
*                 For heterompi performance                        *
*                 model compilation                                *
*                 =========================                        *
*                                                                  *
*  Copyright (c) 2006   School of Computer Science and Informatics *
*                              University College Dublin.          *
*                                                                  *
*  All rights reserved. We assume no responsibility for the use    *
*  or reliability of our software.                                 *
*                                                                  *
********************************************************************/

#ifndef __HMPI_SCALAPACK_HELPERS_INCLUDE_HH
#define __HMPI_SCALAPACK_HELPERS_INCLUDE_HH

#include "scalpblas.h"
#include "scalPBpblas.h"
#include "scalPBtools.h"
#include "scalPBblacs.h"
#include "scalPBblas.h"

/*
* ------------------------------------------------------------------
*       UTILITY MACROS
* ------------------------------------------------------------------
*/
#define   CNARG                         2
#define   TCOMM(a, b, c, d, p, q)       tcomm[(a*p*q*q+b*p*q+c*q+d)]
#define   TCOMML(a, b, c, d, p, q)      tcomm_l[(a*p*q*q+b*p*q+c*q+d)]
#define   TCOMMI(a, b, p, q)            tcomm_i[(a*q+b)]
#define   TCOMP(a, b, p, q)             tcomp[(a*q+b)]
#define   TCOMPL(a, b, p, q)            tcomp_l[(a*q+b)]
#define   PMTCOMM(a, b, c, d, p, q)     pm_tcomm[(a*p*q*q+b*p*q+c*q+d)]
#define   HSCALRIX(a, b, p, q)          (a*q+b)
#define   iceil(a,b) (((a)>0) ? ((((a)+(b)-1)/(b))) : (-((-(a))/(b))))

/*
* ------------------------------------------------------------------
*       ERROR CODES
* ------------------------------------------------------------------
*/
#define   HSCAL_OK                                0
#define   HSCAL_SUCCESS                           0
#define   HSCAL_COMM_UNINITIALIZED                0
#define   HSCAL_ZERO_ESTIMATION                   0.0
#define   HSCAL_ERR_LOW_MEMORY                    -1
#define   HSCAL_ERR_INPUT                         -2
#define   HSCAL_ERR_INITIALIZATION                -3
#define   HSCAL_ERR_RECON                         -4
#define   HSCAL_ERR_RECON_ESTIMATION              -1.0
#define   HSCAL_ERR_GROUP_CREATION                -5
#define   HSCAL_ERR_NULL_COMM                     -6
#define   HSCAL_ERR_BCAST_OPTIMAL_GRID            -7
#define   HSCAL_ERR_CTXT_STORE                    -8
#define   HSCAL_ERR_CTXT_FIND                     -9
#define   HSCAL_CTXT_UNDEFINED                    -10
#define   HSCAL_CTXT_FOR_FREE_PROCS               -11
#define   HSCAL_ERR_BCAST_EXECUTION_TIME          -12

/*
* ------------------------------------------------------------------
*       FUNCTION PROTOTYPES FOR
*       DEBUG
*       CBLAS
*       PTZBLAS
*       PTOOLS
*       BLACS
*       MPC HELPER MACROS
*       COMPUTATIONS AND COMMUNICATIONS
* ------------------------------------------------------------------
*/
#include  "hscalf2c.h"
#include  "hscaldebug.h"
#include  "hscalcblas.h"
#include  "hscalclapack.h"
#include  "hscalptzblas.h"
#include  "hscalptools.h"
#include  "hscalblacs.h"
#include  "hscalpblaslevel1compcomm.h"
#include  "hscalpblaslevel2compcomm.h"
#include  "hscalpblaslevel3compcomm.h"
#include  "hscalpbblas.h"
#include  "hscalapackoptimals.h"
#include  "hscalserialmpcmacros.h"
#include  "hscalparmpcmacros.h"

#include  "hscalpblasintallocmacros.h"

#ifndef __HSCAL_USE_SPECIAL_INIT_LLDS__
#include  "hscalpblaslldmacros.h"
#else
#include  "hscalpblasspeciallldmacros.h"
#endif

#include  "hscalcblasinfo.h"
#include  "hscalpblasinfo.h"
#include  "hscalpblasinfoi.h"
#include  "hscaltest.h"
#include  "hscalapackcompcomm.h"
#include  "hscalapackinfo.h"
#include  "hscalapackinfoi.h"
#include  "hscalapackmpcmacros.h"
#include  "hscalapacktools.h"
#include  "hscalapackenvp.h"

/*
* ------------------------------------------------------------------
*       EXPLICIT BUILT-IN 'C' FUNCTION DECLARATIONS
*       TO AVOID COMPILER WARNINGS
* ------------------------------------------------------------------
*/
#include "hscalcompilerwarnings.h"
#include "hscalblasbuiltins.h"
#include "hscalpblasbuiltins.h"
#include "hscalapackbuiltins.h"

#endif  /* __HMPI_SCALAPACK_HELPERS_HH */

