
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pcswap_parameters.h"

int main(int argc, char **argv)
{
    int     DESCX[DLEN_], DESCY[DLEN_];
    int     i, rc;
    int     iaseed;
    int     ictxt, myrow, mycol;
    int     VERBOSE = 0;
    int     Kx, Ky;

    int     p, q;

    hscalcomplex *X, *Y;

    struct timeval start, end;
    struct timeval astart, aend;
    struct timeval cstart, cend;

    gettimeofday(&start, NULL);

    //
    // Initialize ScaLAPACK runtime
    hscal_init(&argc, &argv);

    DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCX[M_]     = 1;
    DESCX[N_]     = N;
    DESCX[IMB_]   = 1;
    DESCX[INB_]   = NB;
    DESCX[MB_]    = 1;
    DESCX[NB_]    = NB;
    DESCX[RSRC_]  = 0;
    DESCX[CSRC_]  = 0;

    DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCY[M_]     = 1;
    DESCY[N_]     = N;
    DESCY[IMB_]   = 1;
    DESCY[INB_]   = NB;
    DESCY[MB_]    = 1;
    DESCY[NB_]    = NB;
    DESCY[RSRC_]  = 0;
    DESCY[CSRC_]  = 0;

    if (HMPI_Is_host())
    {
       gettimeofday(&cstart, NULL);
    }

    HMPI_Debug(1);

    rc = hscal_pcswap_ctxt(
                 &N,
                 &IX,
                 &JX,
                 DESCX,
                 &INCX,
                 &IY,
                 &JY,
                 DESCY,
                 &INCY,
                 &ictxt
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HMPI_Debug(0);

    if (HMPI_Is_host())
    {
       gettimeofday(&cend, NULL);
    }

    if (!hscal_in_ctxt(&ictxt))
    {
       hscal_free_ctxt(&ictxt);
       hscal_finalize(0);
       return 0;
    }

    if (HMPI_Is_host())
    {
       printf("Estimated execution time=%f\n", hscal_timeof(&ictxt));
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&astart, NULL);
    }

    /*
     * Retrieve the process grid information
     */
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCX[CTXT_]  = ictxt;
    DESCX[M_]     = 1;
    DESCX[N_]     = N;
    DESCX[IMB_]   = 1;
    DESCX[INB_]   = NB;
    DESCX[MB_]    = 1;
    DESCX[NB_]    = NB;
    DESCX[RSRC_]  = 0;
    DESCX[CSRC_]  = 0;
    if (INCX == DESCX[M_])
    {
       DESCX[LLD_] = max(1, PB_Cnumroc(IX,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );

       Kx          = max(1, PB_Cnumroc(JX+N-1,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );
    }
    else
    {
       Kx          = max(1, PB_Cnumroc(JX,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );

       DESCX[LLD_] = max(1, PB_Cnumroc(IX+N-1,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );
    }

    DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCY[CTXT_]  = ictxt;
    DESCY[M_]     = 1;
    DESCY[N_]     = N;
    DESCY[IMB_]   = 1;
    DESCY[INB_]   = NB;
    DESCY[MB_]    = 1;
    DESCY[NB_]    = NB;
    DESCY[RSRC_]  = 0;
    DESCY[CSRC_]  = 0;
    if (INCY == DESCY[M_])
    {
       DESCY[LLD_] = max(1, PB_Cnumroc(IY,
                                       0,
                                       DESCY[IMB_],
                                       DESCY[MB_],
                                       myrow,
                                       DESCY[RSRC_],
                                       p
                        )
       );

       Ky          = max(1, PB_Cnumroc(JY+N-1,
                                       0,
                                       DESCY[INB_],
                                       DESCY[NB_],
                                       mycol,
                                       DESCY[CSRC_],
                                       q
                        )
       );
    }
    else
    {
       Ky          = max(1, PB_Cnumroc(JY,
                                       0,
                                       DESCY[INB_],
                                       DESCY[NB_],
                                       mycol,
                                       DESCY[CSRC_],
                                       q
                        )
       );

       DESCY[LLD_] = max(1, PB_Cnumroc(IY+N-1,
                                       0,
                                       DESCY[IMB_],
                                       DESCY[MB_],
                                       myrow,
                                       DESCY[RSRC_],
                                       p
                        )
       );
    }

    if (VERBOSE > 0)
    {
       printf(
         "(%d, %d): DESCX[LLD_] %d Kx %d, DESCY[LLD_] %d Ky %d\n",
         myrow,
         mycol,
         DESCX[LLD_],
         Kx,
         DESCY[LLD_],
         Ky
       );
    }

    X = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (DESCX[LLD_]*Kx)
    );

    if (X == NULL)
    {
       printf("Cannot allocate X\n");
       HMPI_Finalize(-1);
    }

    Y = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCY[LLD_]*Ky)
    );

    if (Y == NULL)
    {
       printf("Cannot allocate Y\n");
       HMPI_Finalize(-1);
    }

    {
       for (i = 0; i < (DESCX[LLD_]*Kx); i++)
       {
          X[i].r = 2.;
          X[i].i = 2.;
       }

       for (i = 0; i < (DESCY[LLD_]*Ky); i++)
       {
          Y[i].r = 1.;
          Y[i].i = 1.;
       }
    }

    pcswap_(
      &N,
      (float*)X, &IX, &JX, DESCX, &INCX,
      (float*)Y, &IY, &JY, DESCY, &INCY
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCY[LLD_]*Ky); i++)
          {
              printf("(%f, %f) ", Y[i].r, Y[i].i);
          }
          printf("\n");
       }
    }

    free(X);
    free(Y);

    hscal_free_ctxt(&ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double ctstart = cstart.tv_sec + cstart.tv_usec/1000000.;
       double ctend   = cend.tv_sec + cend.tv_usec/1000000.;
       double ct_sec  = ctend - ctstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       printf(
          "N=%d, NB=%d, "
          "Context creation time(sec)=%f, Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          N,
          NB,
          ct_sec,
          at_sec,
          t_sec
       );
    }

    hscal_finalize(0);
}
