
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pdcopy_parameters.h"

int main(int argc, char **argv)
{
    int     DESCX[DLEN_], DESCY[DLEN_];
    int     i, rc;
    int     ictxt, myrow, mycol;
    int     VERBOSE = 0;
    int     Kx, Ky;

    int     p, q;
    HMPI_Group gid;
    MPI_Comm pdcopycomm;

    hscaldoublereal *X, *Y;

    struct timeval start, end;
    struct timeval rstart, rend;
    struct timeval gstart, gend;
    struct timeval astart, aend;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    HMPI_Barrier(HMPI_COMM_WORLD_GROUP);

    DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCX[M_]     = 1;
    DESCX[N_]     = N;
    DESCX[IMB_]   = 1;
    DESCX[INB_]   = NB;
    DESCX[MB_]    = 1;
    DESCX[NB_]    = NB;
    DESCX[RSRC_]  = 0;
    DESCX[CSRC_]  = 0;

    DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCY[M_]     = 1;
    DESCY[N_]     = N;
    DESCY[IMB_]   = 1;
    DESCY[INB_]   = NB;
    DESCY[MB_]    = 1;
    DESCY[NB_]    = NB;
    DESCY[RSRC_]  = 0;
    DESCY[CSRC_]  = 0;

    if (HMPI_Is_host())
    {
       gettimeofday(&rstart, NULL);
    }

    rc = hscal_pdcopy_recon(
              &N,
              &IX,
              &JX,
              DESCX,
              &INCX,
              &IY,
              &JY,
              DESCY,
              &INCY
    );

    if (rc != HMPI_SUCCESS)
    {
       return rc;
    }

    HMPI_Barrier(HMPI_COMM_WORLD_GROUP);

    if (HMPI_Is_host())
    {
       gettimeofday(&rend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&gstart, NULL);
    }

    HMPI_Debug(1);

    rc = hscal_pdcopy_group_heuristic_pauto_create(
              &N,
              &IX,
              &JX,
              DESCX,
              &INCX,
              &IY,
              &JY,
              DESCY,
              &INCY,
              &gid,
              &MPC_NetType_pdcopy,
              hscal_heuristic_fn_all
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HMPI_Debug(0);

    if (HMPI_Is_host())
    {
       gettimeofday(&gend, NULL);
    }

    if (HMPI_Is_free())
    {
       HMPI_Group_free(
           &gid
       );
       HMPI_Finalize(0);
       return 0;
    }

    HMPI_Barrier(&gid);

    if (HMPI_Is_host())
    {
       gettimeofday(&astart, NULL);
    }

    pdcopycomm = *(MPI_Comm*)HMPI_Get_comm(&gid);

    /*
     * Broadcast the number of processes to
     * all the processors
     */
    if (HMPI_Is_parent(&gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           &gid,
           &nd,
           &d
       );

       p = d[0];

       free(d);

       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               pdcopycomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               pdcopycomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       printf("Problems broadcasting p\n");
       MPI_Abort(MPI_COMM_WORLD, -9);
    }

    if (HMPI_Is_parent(&gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           &gid,
           &nd,
           &d
       );

       q = d[1];

       free(d);

       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               pdcopycomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               pdcopycomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       printf("Problems broadcasting q\n");
       MPI_Abort(MPI_COMM_WORLD, -9);
    }

    if (HMPI_Is_parent(&gid))
    {
       printf("Optimal p=%d, Optimal q=%d\n", p, q);
    }

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pdcopycomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCX[CTXT_]  = ictxt;
    DESCX[M_]     = 1;
    DESCX[N_]     = N;
    DESCX[IMB_]   = 1;
    DESCX[INB_]   = NB;
    DESCX[MB_]    = 1;
    DESCX[NB_]    = NB;
    DESCX[RSRC_]  = 0;
    DESCX[CSRC_]  = 0;
    if (INCX == DESCX[M_])
    {
       DESCX[LLD_] = max(1, PB_Cnumroc(IX,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );

       Kx          = max(1, PB_Cnumroc(JX+N-1,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );
    }
    else
    {
       Kx          = max(1, PB_Cnumroc(JX,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );

       DESCX[LLD_] = max(1, PB_Cnumroc(IX+N-1,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );
    }

    DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCY[CTXT_]  = ictxt;
    DESCY[M_]     = 1;
    DESCY[N_]     = N;
    DESCY[IMB_]   = 1;
    DESCY[INB_]   = NB;
    DESCY[MB_]    = 1;
    DESCY[NB_]    = NB;
    DESCY[RSRC_]  = 0;
    DESCY[CSRC_]  = 0;
    if (INCY == DESCY[M_])
    {
       DESCY[LLD_] = max(1, PB_Cnumroc(IY,
                                       0,
                                       DESCY[IMB_],
                                       DESCY[MB_],
                                       myrow,
                                       DESCY[RSRC_],
                                       p
                        )
       );

       Ky          = max(1, PB_Cnumroc(JY+N-1,
                                       0,
                                       DESCY[INB_],
                                       DESCY[NB_],
                                       mycol,
                                       DESCY[CSRC_],
                                       q
                        )
       );
    }
    else
    {
       Ky          = max(1, PB_Cnumroc(JY,
                                       0,
                                       DESCY[INB_],
                                       DESCY[NB_],
                                       mycol,
                                       DESCY[CSRC_],
                                       q
                        )
       );

       DESCY[LLD_] = max(1, PB_Cnumroc(IY+N-1,
                                       0,
                                       DESCY[IMB_],
                                       DESCY[MB_],
                                       myrow,
                                       DESCY[RSRC_],
                                       p
                        )
       );
    }

    if (VERBOSE > 0)
    {
       printf(
         "(%d, %d): DESCX[LLD_] %d Kx %d, DESCY[LLD_] %d Ky %d\n",
         myrow,
         mycol,
         DESCX[LLD_],
         Kx,
         DESCY[LLD_],
         Ky
       );
    }

    X = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (DESCX[LLD_]*Kx)
    );

    if (X == NULL)
    {
       printf("Cannot allocate X\n");
       HMPI_Finalize(-1);
    }

    Y = (hscaldoublereal*)malloc(
                  sizeof(hscaldoublereal)
                  *
                  (DESCY[LLD_]*Ky)
    );

    if (Y == NULL)
    {
       printf("Cannot allocate Y\n");
       HMPI_Finalize(-1);
    }

    for (i = 0; i < (DESCX[LLD_]*Kx); i++)
    {
       X[i] = 2.;
    }

    for (i = 0; i < (DESCY[LLD_]*Ky); i++)
    {
       Y[i] = 1.;
    }

    pdcopy_(
      &N,
      X, &IX, &JX, DESCX, &INCX,
      Y, &IY, &JY, DESCY, &INCY
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCY[LLD_]*Ky); i++)
          {
              printf("%f ", Y[i]);
          }
          printf("\n");
       }
    }

    free(X);
    free(Y);

    Cblacs_gridexit(ictxt);

    HMPI_Barrier(&gid);

    if (HMPI_Is_member(&gid))
    {
       HMPI_Group_free(
           &gid
       );
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double rtstart = rstart.tv_sec + rstart.tv_usec/1000000.;
       double rtend   = rend.tv_sec + rend.tv_usec/1000000.;
       double rt_sec  = rtend - rtstart;

       double gtstart = gstart.tv_sec + gstart.tv_usec/1000000.;
       double gtend   = gend.tv_sec + gend.tv_usec/1000000.;
       double gt_sec  = gtend - gtstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       printf(
          "N=%d, NB=%d, "
          "Recon time(sec)=%f, "
          "Group creation time(sec)=%f, Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          N,
          NB,
          rt_sec,
          gt_sec,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
