
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pccopy_parameters.h"

int main(int argc, char **argv)
{
    int     DESCX[DLEN_], DESCY[DLEN_];
    int     i, rc;
    int     ictxt, myrow, mycol;
    int     VERBOSE = 0;
    int     Kx, Ky;

    int     p = nprow;
    int     q = npcol;
    MPI_Comm pccopycomm;

    hscalcomplex *X, *Y;
 
    struct timeval astart, aend;
    struct timeval start, end;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pccopycomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pccopycomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCX[CTXT_]  = ictxt;
    DESCX[M_]     = N;
    DESCX[N_]     = 1;
    DESCX[IMB_]   = NB;
    DESCX[INB_]   = 1;
    DESCX[MB_]    = NB;
    DESCX[NB_]    = 1;
    DESCX[RSRC_]  = 0;
    DESCX[CSRC_]  = 0;
    if (INCX == DESCX[M_])
    {
       DESCX[LLD_] = max(1, PB_Cnumroc(IX,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );

       Kx          = max(1, PB_Cnumroc(JX+N-1,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );
    }
    else
    {
       Kx          = max(1, PB_Cnumroc(JX,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );

       DESCX[LLD_] = max(1, PB_Cnumroc(IX+N-1,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );
    }

    DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCY[CTXT_]  = ictxt;
    DESCY[M_]     = N;
    DESCY[N_]     = 1;
    DESCY[IMB_]   = NB;
    DESCY[INB_]   = 1;
    DESCY[MB_]    = NB;
    DESCY[NB_]    = 1;
    DESCY[RSRC_]  = 0;
    DESCY[CSRC_]  = 0;
    if (INCY == DESCY[M_])
    {
       DESCY[LLD_] = max(1, PB_Cnumroc(IY,
                                       0,
                                       DESCY[IMB_],
                                       DESCY[MB_],
                                       myrow,
                                       DESCY[RSRC_],
                                       p
                        )
       );

       Ky          = max(1, PB_Cnumroc(JY+N-1,
                                       0,
                                       DESCY[INB_],
                                       DESCY[NB_],
                                       mycol,
                                       DESCY[CSRC_],
                                       q
                        )
       );
    }
    else
    {
       Ky          = max(1, PB_Cnumroc(JY,
                                       0,
                                       DESCY[INB_],
                                       DESCY[NB_],
                                       mycol,
                                       DESCY[CSRC_],
                                       q
                        )
       );

       DESCY[LLD_] = max(1, PB_Cnumroc(IY+N-1,
                                       0,
                                       DESCY[IMB_],
                                       DESCY[MB_],
                                       myrow,
                                       DESCY[RSRC_],
                                       p
                        )
       );
    }

    if (VERBOSE > 0)
    {
       printf(
         "(%d, %d): DESCX[LLD_] %d Kx %d, DESCY[LLD_] %d Ky %d\n",
         myrow,
         mycol,
         DESCX[LLD_],
         Kx,
         DESCY[LLD_],
         Ky
       );
    }

    X = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (DESCX[LLD_]*Kx)
    );

    if (X == NULL)
    {
       printf("Cannot allocate X\n");
       HMPI_Finalize(-1);
    }

    Y = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCY[LLD_]*Ky)
    );

    if (Y == NULL)
    {
       printf("Cannot allocate Y\n");
       HMPI_Finalize(-1);
    }

    {
       for (i = 0; i < (DESCX[LLD_]*Kx); i++)
       {
          X[i].r = 2.;
          X[i].i = 2.;
       }

       for (i = 0; i < (DESCY[LLD_]*Ky); i++)
       {
          Y[i].r = 1.;
          Y[i].i = 1.;
       }
    }

    pccopy_(
      &N,
      (float*)X, &IX, &JX, DESCX, &INCX,
      (float*)Y, &IY, &JY, DESCY, &INCY
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCY[LLD_]*Ky); i++)
          {
              printf("(%f, %f) ", Y[i].r, Y[i].i);
          }
          printf("\n");
       }
    }

    free(X);
    free(Y);

    Cblacs_gridexit(ictxt);

    gettimeofday(&end, NULL);

    if (HMPI_Is_host())
    {
        gettimeofday(&aend, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "p=%d, q=%d, N=%d, NB=%d, algorithm execution time=%f, total time(sec)=%0.6f\n",
          p,
          q,
          N,
          NB,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
