
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pctrsv_parameters.h"

int main(int argc, char **argv)
{
    int     DESCA[DLEN_], DESCX[DLEN_];
    int     i, rc;
    int     iaseed;
    int     ictxt, myrow, mycol;
    int     VERBOSE = 0;
    int     Ka, Kx;
    int     c__0    = 0;

    int     p, q;

    hscalcomplex *A, *X;

    char UPLOC = UPLOI;
    char *UPLO = &UPLOC;
    char TRANSC   = TRANSI;
    char TRANS[2] = { TRANSC, '\0' };
    char DIAGC    = DIAGI;
    char DIAG[2]  = { DIAGC, '\0' };
 
    struct timeval start, end;
    struct timeval cstart, cend;
    struct timeval astart, aend;

    gettimeofday(&start, NULL);

    //
    // Initialize HScaLAPACK runtime
    hscal_init(&argc, &argv);

    DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCA[M_]     = N;
    DESCA[N_]     = N;
    DESCA[IMB_]   = NB;
    DESCA[INB_]   = NB;
    DESCA[MB_]    = NB;
    DESCA[NB_]    = NB;
    DESCA[RSRC_]  = 0;
    DESCA[CSRC_]  = 0;

    DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCX[M_]     = 1;
    DESCX[N_]     = N;
    DESCX[IMB_]   = 1;
    DESCX[INB_]   = NB;
    DESCX[MB_]    = 1;
    DESCX[NB_]    = NB;
    DESCX[RSRC_]  = 0;
    DESCX[CSRC_]  = 0;

    if (HMPI_Is_host())
    {
       gettimeofday(&cstart, NULL);
    }

    HMPI_Debug(1);

    rc = hscal_pctrsv_ctxt(
              UPLO,
              TRANS,
              DIAG,
              &N,
              &IA,
              &JA,
              DESCA,
              &IX,
              &JX,
              DESCX,
              &INCX,
              &ictxt
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HMPI_Debug(0);

    if (HMPI_Is_host())
    {
       gettimeofday(&cend, NULL);
    }

    if (!hscal_in_ctxt(&ictxt))
    {
       hscal_free_ctxt(&ictxt);
       hscal_finalize(0);
       return 0;
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&astart, NULL);
    }

    /*
     * Retrieve the process grid information
     */
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCA[CTXT_]  = ictxt;
    DESCA[M_]     = N;
    DESCA[N_]     = N;
    DESCA[IMB_]   = NB;
    DESCA[INB_]   = NB;
    DESCA[MB_]    = NB;
    DESCA[NB_]    = NB;
    DESCA[RSRC_]  = 0;
    DESCA[CSRC_]  = 0;
    DESCA[LLD_]   = max(1, PB_Cnumroc(DESCA[M_],
                                      0,
                                      DESCA[IMB_],
                                      DESCA[MB_],
                                      myrow,
                                      DESCA[RSRC_],
                                      p
                       )
    );

    Ka            = max(1, PB_Cnumroc(JA+DESCA[N_]-1,
                                      0,
                                      DESCA[INB_],
                                      DESCA[NB_],
                                      mycol,
                                      DESCA[CSRC_],
                                      q
                       )
    );

    DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCX[CTXT_]  = ictxt;
    DESCX[M_]     = 1;
    DESCX[N_]     = N;
    DESCX[IMB_]   = 1;
    DESCX[INB_]   = NB;
    DESCX[MB_]    = 1;
    DESCX[NB_]    = NB;
    DESCX[RSRC_]  = 0;
    DESCX[CSRC_]  = 0;
    if (INCX == DESCX[M_])
    {
       DESCX[LLD_] = max(1, PB_Cnumroc(IX,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );

       Kx          = max(1, PB_Cnumroc(JX+N-1,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );
    }
    else
    {
       Kx          = max(1, PB_Cnumroc(JX,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );

       DESCX[LLD_] = max(1, PB_Cnumroc(IX+N-1,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );
    }

    if (VERBOSE > 0)
    {
       printf(
         "(%d, %d): DESCA[LLD_] %d Ka %d, DESCX[LLD_] %d Kx %d\n",
         myrow,
         mycol,
         DESCA[LLD_],
         Ka,
         DESCX[LLD_],
         Kx
       );
    }

    A = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (DESCA[LLD_]*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    X = (hscalcomplex*)malloc(
                 sizeof(hscalcomplex)
                 *
                 (DESCX[LLD_]*Kx)
    );

    if (X == NULL)
    {
       printf("Cannot allocate X\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pcmatgen(&ictxt, "N", "N", &DESCA[M_],
                 &DESCA[N_], &DESCA[MB_], &DESCA[NB_], A, &DESCA[LLD_],
                 &DESCA[RSRC_], &DESCA[CSRC_], &iaseed, &c__0, &DESCA[LLD_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);

       for (i = 0; i < (DESCX[LLD_]*Kx); i++)
       {
          X[i].r = 1.;
          X[i].i = 1.;
       }
    }

    pctrsv_(
      UPLO, TRANS, DIAG, &N, 
      (float*)A, &IA, &JA, DESCA,
      (float*)X, &IX, &JX, DESCX, &INCX
    );

    free(A);
    free(X);

    hscal_free_ctxt(&ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double ctstart = cstart.tv_sec + cstart.tv_usec/1000000.;
       double ctend = cend.tv_sec + cend.tv_usec/1000000.;
       double ct_sec = ctend - ctstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec = atend - atstart;

       double tstart = start.tv_sec + start.tv_usec/1000000.;
       double tend   = end.tv_sec + end.tv_usec/1000000.;
       double t_sec = tend - tstart;

       printf(
          "UPLOC=%c, TRANS=%c, DIAG=%c, N=%d, NB=%d, "
          "Context creation time(sec)=%f, Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          UPLOC,
          TRANSC,
          DIAGC,
          N,
          NB,
          ct_sec,
          at_sec,
          t_sec
       );
    }

    hscal_finalize(0);
}
