
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pdgemv_parameters.h"

/*------------------------------------------------------------------*/
int main(int argc, char **argv)
{
    int     DESCA[DLEN_], DESCX[DLEN_], DESCY[DLEN_];
    int     i, rc;
    HMPI_Group gid;

    char TRANSC = TRANSI;
    char *TRANS = &TRANSC;

    struct timeval gstart, gend;

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (TRANSC == 'N')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = M;
       DESCA[N_]     = N;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;

       DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCX[M_]     = 1;
       DESCX[N_]     = N;
       DESCX[IMB_]   = 1;
       DESCX[INB_]   = NB;
       DESCX[MB_]    = 1;
       DESCX[NB_]    = NB;
       DESCX[RSRC_]  = 0;
       DESCX[CSRC_]  = 0;

       DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCY[M_]     = 1;
       DESCY[N_]     = M;
       DESCY[IMB_]   = 1;
       DESCY[INB_]   = MB;
       DESCY[MB_]    = 1;
       DESCY[NB_]    = MB;
       DESCY[RSRC_]  = 0;
       DESCY[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = M;
       DESCA[N_]     = N;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;

       DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCX[M_]     = 1;
       DESCX[N_]     = M;
       DESCX[IMB_]   = 1;
       DESCX[INB_]   = MB;
       DESCX[MB_]    = 1;
       DESCX[NB_]    = MB;
       DESCX[RSRC_]  = 0;
       DESCX[CSRC_]  = 0;

       DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCY[M_]     = 1;
       DESCY[N_]     = N;
       DESCY[IMB_]   = 1;
       DESCY[INB_]   = NB;
       DESCY[MB_]    = 1;
       DESCY[NB_]    = NB;
       DESCY[RSRC_]  = 0;
       DESCY[CSRC_]  = 0;
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&gstart, NULL);
    }

    rc = hscal_pdgemv_recon(
              TRANS,
              &M,
              &N,
              &ALPHA,
              &IA,
              &JA,
              DESCA,
              &IX,
              &JX,
              DESCX,
              &INCX,
              &BETA,
              &IY,
              &JY,
              DESCY,
              &INCY
    );

    if (rc != HMPI_SUCCESS)
    {
       return rc;
    }

    {
      int *model_params, model_count;

      rc = hscal_pack_pdgemv_modelp(
                TRANS,
                &M,
                &N,
                &ALPHA,
                &IA,
                &JA,
                DESCA,
                &IX,
                &JX,
                DESCX,
                &INCX,
                &BETA,
                &IY,
                &JY,
                DESCY,
                &INCY,
                &model_params,
                &model_count
      );

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      HMPI_Debug(1);

      rc = HMPI_Group_heuristic_pauto_create(
               &gid,
               hscal_model_pdgemv,
               hscal_heuristic_fn_lean,
               model_params,
               model_count
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);

         return rc;
      }

      free(model_params);
   }

   HMPI_Debug(0);

   if (HMPI_Is_host())
   {
      gettimeofday(&gend, NULL);
   }

   {
      HMPI_Group_free(
          &gid
      );
   }

   /*
    * Print Execution time
    */
   if (HMPI_Is_host())
   {
      double gtstart = gstart.tv_sec + gstart.tv_usec/1000000.;
      double gtend = gend.tv_sec + gend.tv_usec/1000000.;
      double gt_sec = gtend - gtstart;

      printf(
         "M=%d, N=%d, MB=%d, NB=%d, "
         "Group creation time(sec)=%f\n",
         M,
         N,
         MB,
         NB,
         gt_sec
      );
   }

   HMPI_Finalize(0);
}
