
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "scalapack_pdgemv_parameters.h"

/*------------------------------------------------------------------*/
int main(int argc, char **argv)
{
    int     DESCA[DLEN_], DESCX[DLEN_], DESCY[DLEN_];
    int     i, rc;
    double time;

    char TRANSC = TRANSI;
    char *TRANS = &TRANSC;

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (TRANSC == 'N')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = M;
       DESCA[N_]     = N;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;

       DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCX[M_]     = 1;
       DESCX[N_]     = N;
       DESCX[IMB_]   = 1;
       DESCX[INB_]   = NB;
       DESCX[MB_]    = 1;
       DESCX[NB_]    = NB;
       DESCX[RSRC_]  = 0;
       DESCX[CSRC_]  = 0;

       DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCY[M_]     = 1;
       DESCY[N_]     = M;
       DESCY[IMB_]   = 1;
       DESCY[INB_]   = MB;
       DESCY[MB_]    = 1;
       DESCY[NB_]    = MB;
       DESCY[RSRC_]  = 0;
       DESCY[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = M;
       DESCA[N_]     = N;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;

       DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCX[M_]     = 1;
       DESCX[N_]     = M;
       DESCX[IMB_]   = 1;
       DESCX[INB_]   = MB;
       DESCX[MB_]    = 1;
       DESCX[NB_]    = MB;
       DESCX[RSRC_]  = 0;
       DESCX[CSRC_]  = 0;

       DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCY[M_]     = 1;
       DESCY[N_]     = N;
       DESCY[IMB_]   = 1;
       DESCY[INB_]   = NB;
       DESCY[MB_]    = 1;
       DESCY[NB_]    = NB;
       DESCY[RSRC_]  = 0;
       DESCY[CSRC_]  = 0;
    }

    rc = hscal_pdgemv_recon(
              TRANS,
              &M,
              &N,
              &ALPHA,
              &IA,
              &JA,
              DESCA,
              &IX,
              &JX,
              DESCX,
              &INCX,
              &BETA,
              &IY,
              &JY,
              DESCY,
              &INCY
    );

    if (rc != HMPI_SUCCESS)
    {
       return rc;
    }

    {
      int *model_params, model_count;

      rc = hscal_pack_pdgemv_modelp_noauto(
                TRANS,
                &M,
                &N,
                &ALPHA,
                &IA,
                &JA,
                DESCA,
                &IX,
                &JX,
                DESCX,
                &INCX,
                &BETA,
                &IY,
                &JY,
                DESCY,
                &INCY,
                &nprow,
                &npcol,
                &model_params,
                &model_count
      );

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      time = HMPI_Timeof(
                 &MPC_NetType_pdgemv,
                 model_params,
                 model_count
      );

      printf("Time %f\n", time);

      free(model_params);
    }

    HMPI_Finalize(0);
}
