
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pstrmv_parameters.h"

int main(int argc, char **argv)
{
    int     DESCA[DLEN_], DESCX[DLEN_];
    int     i, rc;
    int     ictxt, myrow, mycol;
    int     VERBOSE = 0;
    int     Ka, Kx;

    int     p = nprow;
    int     q = npcol;
    MPI_Comm pstrmvcomm;

    hscalreal *A, *X;
 
    struct timeval astart, aend;
    struct timeval start, end;

    char UPLOC    = UPLOI;
    char UPLO[2]  = { UPLOC, '\0' };
    char TRANSC   = TRANSI;
    char TRANS[2] = { TRANSC, '\0' };
    char DIAGC    = DIAGI;
    char DIAG[2]  = { DIAGC, '\0' };

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pstrmvcomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pstrmvcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCA[CTXT_]  = ictxt;
    DESCA[M_]     = N;
    DESCA[N_]     = N;
    DESCA[IMB_]   = NB;
    DESCA[INB_]   = NB;
    DESCA[MB_]    = NB;
    DESCA[NB_]    = NB;
    DESCA[RSRC_]  = 0;
    DESCA[CSRC_]  = 0;
    DESCA[LLD_]   = max(1, PB_Cnumroc(DESCA[M_],
                                      0,
                                      DESCA[IMB_],
                                      DESCA[MB_],
                                      myrow,
                                      DESCA[RSRC_],
                                      p
                       )
    );

    Ka            = max(1, PB_Cnumroc(JA+DESCA[N_]-1,
                                      0,
                                      DESCA[INB_],
                                      DESCA[NB_],
                                      mycol,
                                      DESCA[CSRC_],
                                      q
                       )
    );

    DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCX[CTXT_]  = ictxt;
    DESCX[M_]     = 1;
    DESCX[N_]     = N;
    DESCX[IMB_]   = 1;
    DESCX[INB_]   = NB;
    DESCX[MB_]    = 1;
    DESCX[NB_]    = NB;
    DESCX[RSRC_]  = 0;
    DESCX[CSRC_]  = 0;
    if (INCX == DESCX[M_])
    {
       DESCX[LLD_] = max(1, PB_Cnumroc(IX,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );

       Kx          = max(1, PB_Cnumroc(JX+N-1,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );
    }
    else
    {
       Kx          = max(1, PB_Cnumroc(JX,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );

       DESCX[LLD_] = max(1, PB_Cnumroc(IX+N-1,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );
    }

    if (VERBOSE > 0)
    {
       printf(
         "(%d, %d): DESCA[LLD_] %d Ka %d, DESCX[LLD_] %d Kx %d\n",
         myrow,
         mycol,
         DESCA[LLD_],
         Ka,
         DESCX[LLD_],
         Kx
       );
    }

    A = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (DESCA[LLD_]*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    X = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (DESCX[LLD_]*Kx)
    );

    if (X == NULL)
    {
       printf("Cannot allocate X\n");
       HMPI_Finalize(-1);
    }

    for (i = 0; i < (DESCA[LLD_]*Ka); i++)
    {
       A[i] = 2.;
    }

    for (i = 0; i < (DESCX[LLD_]*Kx); i++)
    {
       X[i] = 1.;
    }

    pstrmv_(
      UPLO, TRANS, DIAG, &N,
      A, &IA, &JA, DESCA,
      X, &IX, &JX, DESCX, &INCX
    );

    free(A);
    free(X);

    Cblacs_gridexit(ictxt);

    gettimeofday(&end, NULL);

    if (HMPI_Is_host())
    {
        gettimeofday(&aend, NULL);
    }
    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "UPLO=%s, TRANS=%s, DIAG=%s, p=%d, q=%d, N=%d, NB=%d, "
          "algorithm execution time=%f, total time(sec)=%0.6f\n",
          UPLO,
          TRANS,
          DIAG,
          p,
          q,
          N,
          NB,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
