
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pzagemv_parameters.h"

int main(int argc, char **argv)
{
    int     VERBOSE = 0;
    int     DESCA[DLEN_], DESCX[DLEN_], DESCY[DLEN_];
    int     i, rc;
    int     ictxt, myrow, mycol;
    int     Ka, Kx, Ky;

    int     p = nprow;
    int     q = npcol;
    MPI_Comm pzagemvcomm;

    hscaldoublecomplex *A, *X;
    hscaldoublereal    *Y;
 
    struct timeval astart, aend;
    struct timeval start, end;

    char TRANSC = TRANSI;
    char *TRANS = &TRANSC;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pzagemvcomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pzagemvcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    if (TRANSC == 'N')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = M;
       DESCA[N_]     = N;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;

       DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCX[CTXT_]  = ictxt;
       DESCX[M_]     = 1;
       DESCX[N_]     = N;
       DESCX[IMB_]   = 1;
       DESCX[INB_]   = NB;
       DESCX[MB_]    = 1;
       DESCX[NB_]    = NB;
       DESCX[RSRC_]  = 0;
       DESCX[CSRC_]  = 0;

       DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCY[CTXT_]  = ictxt;
       DESCY[M_]     = 1;
       DESCY[N_]     = M;
       DESCY[IMB_]   = 1;
       DESCY[INB_]   = MB;
       DESCY[MB_]    = 1;
       DESCY[NB_]    = MB;
       DESCY[RSRC_]  = 0;
       DESCY[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = M;
       DESCA[N_]     = N;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;

       DESCX[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCX[CTXT_]  = ictxt;
       DESCX[M_]     = 1;
       DESCX[N_]     = M;
       DESCX[IMB_]   = 1;
       DESCX[INB_]   = MB;
       DESCX[MB_]    = 1;
       DESCX[NB_]    = MB;
       DESCX[RSRC_]  = 0;
       DESCX[CSRC_]  = 0;

       DESCY[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCY[CTXT_]  = ictxt;
       DESCY[M_]     = 1;
       DESCY[N_]     = N;
       DESCY[IMB_]   = 1;
       DESCY[INB_]   = NB;
       DESCY[MB_]    = 1;
       DESCY[NB_]    = NB;
       DESCY[RSRC_]  = 0;
       DESCY[CSRC_]  = 0;
    }

    DESCA[LLD_]   = max(1, PB_Cnumroc(DESCA[M_],
                                      0,
                                      DESCA[IMB_],
                                      DESCA[MB_],
                                      myrow,
                                      DESCA[RSRC_],
                                      p
                       )
    );

    Ka            = max(1, PB_Cnumroc(DESCA[N_],
                                      0,
                                      DESCA[INB_],
                                      DESCA[NB_],
                                      mycol,
                                      DESCA[CSRC_],
                                      q
                       )
    );

    if (INCX == DESCX[M_])
    {
       DESCX[LLD_] = max(1, PB_Cnumroc(IX,
                                       0,
                                       DESCX[IMB_],
                                       DESCX[MB_],
                                       myrow,
                                       DESCX[RSRC_],
                                       p
                        )
       );

       if (TRANSC == 'N')
       {
          Kx          = max(1, PB_Cnumroc(JX+N-1,
                                          0,
                                          DESCX[INB_],
                                          DESCX[NB_],
                                          mycol,
                                          DESCX[CSRC_],
                                          q
                           )
          );
       }
       else
       {
          Kx          = max(1, PB_Cnumroc(JX+M-1,
                                          0,
                                          DESCX[INB_],
                                          DESCX[NB_],
                                          mycol,
                                          DESCX[CSRC_],
                                          q
                           )
          );
       }
    }
    else
    {
       Kx          = max(1, PB_Cnumroc(JX,
                                       0,
                                       DESCX[INB_],
                                       DESCX[NB_],
                                       mycol,
                                       DESCX[CSRC_],
                                       q
                        )
       );

       if (TRANSC == 'N')
       {
          DESCX[LLD_] = max(1, PB_Cnumroc(IX+N-1,
                                          0,
                                          DESCX[IMB_],
                                          DESCX[MB_],
                                          myrow,
                                          DESCX[RSRC_],
                                          p
                           )
          );
       }
       else
       {
          DESCX[LLD_] = max(1, PB_Cnumroc(IX+M-1,
                                          0,
                                          DESCX[IMB_],
                                          DESCX[MB_],
                                          myrow,
                                          DESCX[RSRC_],
                                          p
                           )
          );
       }
    }

    if (INCY == DESCY[M_])
    {
       DESCY[LLD_] = max(1, PB_Cnumroc(IY,
                                       0,
                                       DESCY[IMB_],
                                       DESCY[MB_],
                                       myrow,
                                       DESCY[RSRC_],
                                       p
                        )
       );

       if (TRANSC == 'N')
       {
          Ky          = max(1, PB_Cnumroc(JY+M-1,
                                          0,
                                          DESCY[INB_],
                                          DESCY[NB_],
                                          mycol,
                                          DESCY[CSRC_],
                                          q
                           )
          );
       }
       else
       {
          Ky          = max(1, PB_Cnumroc(JY+N-1,
                                          0,
                                          DESCY[INB_],
                                          DESCY[NB_],
                                          mycol,
                                          DESCY[CSRC_],
                                          q
                           )
          );
       }
    }
    else
    {
       Ky          = max(1, PB_Cnumroc(JY,
                                       0,
                                       DESCY[INB_],
                                       DESCY[NB_],
                                       mycol,
                                       DESCY[CSRC_],
                                       q
                        )
       );

       if (TRANSC == 'N')
       {
          DESCY[LLD_] = max(1, PB_Cnumroc(IY+M-1,
                                          0,
                                          DESCY[IMB_],
                                          DESCY[MB_],
                                          myrow,
                                          DESCY[RSRC_],
                                          p
                           )
          );
       }
       else
       {
          DESCY[LLD_] = max(1, PB_Cnumroc(IY+N-1,
                                          0,
                                          DESCY[IMB_],
                                          DESCY[MB_],
                                          myrow,
                                          DESCY[RSRC_],
                                          p
                           )
          );
       }
    }

    if (VERBOSE > 0)
    {
       printf(
         "(%d, %d): DESCA[LLD_] %d Ka %d, DESCX[LLD_] %d Kx %d, DESCY[LLD_] %d Ky %d\n",
         myrow,
         mycol,
         DESCA[LLD_],
         Ka,
         DESCX[LLD_],
         Kx,
         DESCY[LLD_],
         Ky
       );
    }

    A = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (DESCA[LLD_]*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    X = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (DESCX[LLD_]*Kx)
    );

    if (X == NULL)
    {
       printf("Cannot allocate X\n");
       HMPI_Finalize(-1);
    }

    Y = (double*)malloc(
                 sizeof(double)
                 *
                 (DESCY[LLD_]*Ky)
    );

    if (Y == NULL)
    {
       printf("Cannot allocate Y\n");
       HMPI_Finalize(-1);
    }

    {
       for (i = 0; i < (DESCA[LLD_]*Ka); i++)
       {
          A[i].r = 2.;
          A[i].i = 2.;
       }

       for (i = 0; i < (DESCX[LLD_]*Kx); i++)
       {
          X[i].r = 1.;
          X[i].i = 1.;
       }

       for (i = 0; i < (DESCY[LLD_]*Ky); i++)
       {
          Y[i] = 0.;
       }
    }

    pzagemv_(
      TRANS, &M, &N, &ALPHA,
      (double*)A, &IA, &JA, DESCA,
      (double*)X, &IX, &JX, DESCX, &INCX, &BETA,
      Y, &IY, &JY, DESCY, &INCY
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCY[LLD_]*Ky); i++)
          {
              printf("Y[%d]=%0.6f ", i, Y[i]);
          }
          printf("\n");
       }
    }

    free(A);
    free(X);
    free(Y);

    Cblacs_gridexit(ictxt);

    gettimeofday(&end, NULL);

    if (HMPI_Is_host())
    {
        gettimeofday(&aend, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "TRANS=%c, p=%d, q=%d, M=%d, N=%d, MB=%d, NB=%d, algorithm execution time=%f, total time(sec)=%0.6f\n",
          TRANSC,
          p,
          q,
          M,
          N,
          MB,
          NB,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}

