
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pchemm_parameters.h"

int main(int argc, char **argv)
{
    int        DESCA[DLEN_], DESCB[DLEN_], DESCC[DLEN_];
    int        i, rc;
    int        ictxt, myrow, mycol;
    int        Ka, Kb, Kc;
    int        CHECK_RESULT = 0;
    int     c__0    = 0;

    int        p, q;
    HMPI_Group gid;
    MPI_Comm   pchemmcomm;

    hscalcomplex    *A, *B, *C;
 
    struct timeval start, end;
    struct timeval rstart, rend;
    struct timeval gstart, gend;
    struct timeval astart, aend;

    char SIDEC = SIDEI;
    char *SIDE = &SIDEC;
    char UPLOC = UPLOI;
    char *UPLO  = &UPLOC;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (SIDEC == 'L')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = M;
       DESCA[N_]     = M;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = MB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = MB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = N;
       DESCA[N_]     = N;
       DESCA[IMB_]   = NB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = NB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }

    DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCB[M_]     = M;
    DESCB[N_]     = N;
    DESCB[IMB_]   = MB;
    DESCB[INB_]   = NB;
    DESCB[MB_]    = MB;
    DESCB[NB_]    = NB;
    DESCB[RSRC_]  = 0;
    DESCB[CSRC_]  = 0;

    DESCC[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCC[M_]     = M;
    DESCC[N_]     = N;
    DESCC[IMB_]   = MB;
    DESCC[INB_]   = NB;
    DESCC[MB_]    = MB;
    DESCC[NB_]    = NB;
    DESCC[RSRC_]  = 0;
    DESCC[CSRC_]  = 0;

    if (HMPI_Is_host())
    {
       gettimeofday(&rstart, NULL);
    }

    rc = hscal_pchemm_recon(
              SIDE,
              UPLO,
              &M,
              &N,
              (float*)&ALPHA,
              &IA,
              &JA,
              DESCA,
              &IB,
              &JB,
              DESCB,
              (float*)&BETA,
              &IC,
              &JC,
              DESCC
    );

    if (rc != HMPI_SUCCESS)
    {
       return rc;
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&rend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&gstart, NULL);
    }

    HMPI_Debug(1);

    rc = hscal_pchemm_group_heuristic_pauto_create(
              SIDE,
              UPLO,
              &M,
              &N,
              (float*)&ALPHA,
              &IA,
              &JA,
              DESCA,
              &IB,
              &JB,
              DESCB,
              (float*)&BETA,
              &IC,
              &JC,
              DESCC,
              &gid,
              &MPC_NetType_pchemm,
              hscal_heuristic_fn_all
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HMPI_Debug(0);

    if (HMPI_Is_host())
    {
       gettimeofday(&gend, NULL);
    }

    if (HMPI_Is_free())
    {
       HMPI_Group_free(
           &gid
       );
       HMPI_Finalize(0);
       return 0;
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&astart, NULL);
    }

    pchemmcomm = *(MPI_Comm*)HMPI_Get_comm(&gid);

    /*
     * Broadcast the number of processes to
     * all the processors
     */
    if (HMPI_Is_parent(&gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           &gid,
           &nd,
           &d
       );

       p = d[0];

       free(d);

       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               pchemmcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &p,
               1,
               MPI_INT,
               0,
               pchemmcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       printf("Problems broadcasting p\n");
       MPI_Abort(MPI_COMM_WORLD, -9);
    }

    if (HMPI_Is_parent(&gid))
    {
       int nd;
       int *d;

       HMPI_Group_topology(
           &gid,
           &nd,
           &d
       );

       q = d[1];

       free(d);

       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               pchemmcomm
       );
    }
    else
    {
       rc = MPI_Bcast(
               &q,
               1,
               MPI_INT,
               0,
               pchemmcomm
       );
    }

    if (rc != MPI_SUCCESS)
    {
       printf("Problems broadcasting q\n");
       MPI_Abort(MPI_COMM_WORLD, -9);
    }

    if (HMPI_Is_parent(&gid))
    {
       printf("Optimal p=%d, Optimal q=%d\n", p, q);
    }

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pchemmcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    if (SIDEC == 'L')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = M;
       DESCA[N_]     = M;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = MB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = MB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = N;
       DESCA[N_]     = N;
       DESCA[IMB_]   = NB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = NB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }

    DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCB[CTXT_]  = ictxt;
    DESCB[M_]     = M;
    DESCB[N_]     = N;
    DESCB[IMB_]   = MB;
    DESCB[INB_]   = NB;
    DESCB[MB_]    = MB;
    DESCB[NB_]    = NB;
    DESCB[RSRC_]  = 0;
    DESCB[CSRC_]  = 0;

    DESCC[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCC[CTXT_]  = ictxt;
    DESCC[M_]     = M;
    DESCC[N_]     = N;
    DESCC[IMB_]   = MB;
    DESCC[INB_]   = NB;
    DESCC[MB_]    = MB;
    DESCC[NB_]    = NB;
    DESCC[RSRC_]  = 0;
    DESCC[CSRC_]  = 0;

    DESCA[LLD_]   = max(1, PB_Cnumroc(DESCA[M_],
                                      0,
                                      DESCA[IMB_],
                                      DESCA[MB_],
                                      myrow,
                                      DESCA[RSRC_],
                                      p
                       )
    );

    if (SIDEC == 'L')
    {
        Ka            = max(1, PB_Cnumroc(JA+M-1,
                                          0,
                                          DESCA[INB_],
                                          DESCA[NB_],
                                          mycol,
                                          DESCA[CSRC_],
                                          q
                           )
        );
    }
    else
    {
        Ka            = max(1, PB_Cnumroc(JA+N-1,
                                          0,
                                          DESCA[INB_],
                                          DESCA[NB_],
                                          mycol,
                                          DESCA[CSRC_],
                                          q
                           )
        );
    }

    DESCB[LLD_]   = max(1, PB_Cnumroc(DESCB[M_],
                                      0,
                                      DESCB[IMB_],
                                      DESCB[MB_],
                                      myrow,
                                      DESCB[RSRC_],
                                      p
                       )
    );

    Kb            = max(1, PB_Cnumroc(JB+N-1,
                                      0,
                                      DESCB[INB_],
                                      DESCB[NB_],
                                      mycol,
                                      DESCB[CSRC_],
                                      q
                       )
    );

    DESCC[LLD_]   = max(1, PB_Cnumroc(DESCC[M_],
                                      0,
                                      DESCC[IMB_],
                                      DESCC[MB_],
                                      myrow,
                                      DESCC[RSRC_],
                                      p
                       )
    );

    Kc            = max(1, PB_Cnumroc(JC+N-1,
                                      0,
                                      DESCC[INB_],
                                      DESCC[NB_],
                                      mycol,
                                      DESCC[CSRC_],
                                      q
                       )
    );

    A = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCA[LLD_]*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    B = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCB[LLD_]*Kb)
    );

    if (B == NULL)
    {
       printf("Cannot allocate B\n");
       HMPI_Finalize(-1);
    }

    C = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCC[LLD_]*Kc)
    );

    if (C == NULL)
    {
       printf("Cannot allocate C\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pcmatgen(&ictxt, "H", "N", &DESCA[M_],
                 &DESCA[N_], &DESCA[MB_], &DESCA[NB_], A, &DESCA[LLD_],
                 &DESCA[RSRC_], &DESCA[CSRC_], &iaseed, &c__0, &DESCA[LLD_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    {
       int     iaseed = 200;
       hscalt_pcmatgen(&ictxt, "N", "N", &DESCB[M_],
                 &DESCB[N_], &DESCB[MB_], &DESCB[NB_], B, &DESCB[LLD_],
                 &DESCB[RSRC_], &DESCB[CSRC_], &iaseed, &c__0, &DESCB[LLD_], &c__0, &Kb,
                 &myrow, &mycol, &p, &q);
    }

    {
       int     iaseed = 300;
       hscalt_pcmatgen(&ictxt, "N", "N", &DESCC[M_],
                 &DESCC[N_], &DESCC[MB_], &DESCC[NB_], C, &DESCC[LLD_],
                 &DESCC[RSRC_], &DESCC[CSRC_], &iaseed, &c__0, &DESCC[LLD_], &c__0, &Kc,
                 &myrow, &mycol, &p, &q);
    }

    pchemm_(
      SIDE, UPLO, 
      &M, &N,
      (float*)&ALPHA,
      (float*)A, &IA, &JA, DESCA,
      (float*)B, &IB, &JB, DESCB,
      (float*)&BETA,
      (float*)C, &IC, &JC, DESCC
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCC[LLD_]*Kc); i++)
          {
              printf("C[%d]=(%f, %f) ", i, C[i].r, C[i].i);
          }
          printf("\n");
       }
    }

    free(A);
    free(B);
    free(C);

    Cblacs_gridexit(ictxt);

    HMPI_Barrier(&gid);

    if (HMPI_Is_member(&gid))
    {
       HMPI_Group_free(
           &gid
       );
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double rtstart = rstart.tv_sec + rstart.tv_usec/1000000.;
       double rtend = rend.tv_sec + rend.tv_usec/1000000.;
       double rt_sec = rtend - rtstart;

       double gtstart = gstart.tv_sec + gstart.tv_usec/1000000.;
       double gtend = gend.tv_sec + gend.tv_usec/1000000.;
       double gt_sec = gtend - gtstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec = atend - atstart;

       double tstart = start.tv_sec + start.tv_usec/1000000.;
       double tend   = end.tv_sec + end.tv_usec/1000000.;
       double t_sec = tend - tstart;

       printf(
          "SIDE=%c, UPLO=%c, M=%d, N=%d, MB=%d, NB=%d, "
          "Recon time(sec)=%f, "
          "Group creation time(sec)=%f, Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          SIDEC,
          UPLOC,
          M,
          N,
          MB,
          NB,
          rt_sec,
          gt_sec,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
