
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pdtrsm_parameters.h"

/*---------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
    int        DESCA[DLEN_], DESCB[DLEN_];
    int        i, rc;
    int        ictxt, myrow, mycol;
    int        Ka, Kb;
    int        CHECK_RESULT = 0;
    int     c__0    = 0;

    int        p, q;

    hscaldoublereal    *A, *B;
 
    char SIDEC  = SIDEI;
    char *SIDE  = &SIDEC;
    char UPLOC  = UPLOI;
    char *UPLO  = &UPLOC;
    char TRANSC = TRANSI;
    char *TRANS = &TRANSC;
    char DIAGC  = DIAGI;
    char *DIAG  = &DIAGC;

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (SIDEC == 'L')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = M;
       DESCA[N_]     = M;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = MB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = MB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = N;
       DESCA[N_]     = N;
       DESCA[IMB_]   = NB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = NB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }

    DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCB[M_]     = M;
    DESCB[N_]     = N;
    DESCB[IMB_]   = MB;
    DESCB[INB_]   = NB;
    DESCB[MB_]    = MB;
    DESCB[NB_]    = NB;
    DESCB[RSRC_]  = 0;
    DESCB[CSRC_]  = 0;

    rc = hscal_pdtrsm_recon(
              SIDE,
              UPLO,
              TRANS,
              DIAG,
              &M,
              &N,
              &ALPHA,
              &IA,
              &JA,
              DESCA,
              &IB,
              &JB,
              DESCB
    );

    if (rc != HMPI_SUCCESS)
    {
       return rc;
    }

    {
          int dummy, rc, i;
          int num_of_processors = HMPI_Get_number_of_processors();
          double *rperf = (double*)malloc(
                                   sizeof(double)
                                   *
                                   num_of_processors
          );

          rc = HMPI_Get_processors_info(
                   rperf
          );

          if (rc != HMPI_OK)
          {
             printf(
                "MAIN:Problems getting the performances"
                "...Exiting...\n"
             );
             HMPI_Abort(-2);
          }

          if (HMPI_Is_host())
          {
             printf("Updated processor performances are :\n");

             for (i = 0; i < num_of_processors; i++)
             {
                 printf("%0.2f ", rperf[i]);
             }

             printf("\n");
          }

          free(rperf);
    }

    HMPI_Finalize(0);
}
