
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pctradd_parameters.h"

int main(int argc, char **argv)
{
    int        DESCA[DLEN_], DESCC[DLEN_];
    int        i, rc;
    int        ictxt, myrow, mycol;
    int        Ka, Kc;
    int        CHECK_RESULT = 0;
    int     c__0    = 0;


    int     p = nprow;
    int     q = npcol;
    HMPI_Group gid;
    MPI_Comm   pctraddcomm;

    hscalcomplex    *A, *C;
 
    struct timeval start, end;
    struct timeval astart, aend;

    char UPLOC  = UPLOI;
    char *UPLO  = &UPLOC;
    char TRANSC = TRANSI;
    char *TRANS = &TRANSC;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pctraddcomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pctraddcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    if (TRANSC == 'N')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = M;
       DESCA[N_]     = N;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = N;
       DESCA[N_]     = M;
       DESCA[IMB_]   = NB;
       DESCA[INB_]   = MB;
       DESCA[MB_]    = NB;
       DESCA[NB_]    = MB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }

    DESCC[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCC[CTXT_]  = ictxt;
    DESCC[M_]     = M;
    DESCC[N_]     = N;
    DESCC[IMB_]   = MB;
    DESCC[INB_]   = NB;
    DESCC[MB_]    = MB;
    DESCC[NB_]    = NB;
    DESCC[RSRC_]  = 0;
    DESCC[CSRC_]  = 0;

    DESCA[LLD_]   = max(1, PB_Cnumroc(DESCA[M_],
                                      0,
                                      DESCA[IMB_],
                                      DESCA[MB_],
                                      myrow,
                                      DESCA[RSRC_],
                                      p
                       )
    );

    if (TRANSC == 'N')
    {
        Ka            = max(1, PB_Cnumroc(JA+N-1,
                                          0,
                                          DESCA[INB_],
                                          DESCA[NB_],
                                          mycol,
                                          DESCA[CSRC_],
                                          q
                           )
        );
    }
    else
    {
        Ka            = max(1, PB_Cnumroc(JA+M-1,
                                          0,
                                          DESCA[INB_],
                                          DESCA[NB_],
                                          mycol,
                                          DESCA[CSRC_],
                                          q
                           )
        );
    }

    DESCC[LLD_]   = max(1, PB_Cnumroc(DESCC[M_],
                                      0,
                                      DESCC[IMB_],
                                      DESCC[MB_],
                                      myrow,
                                      DESCC[RSRC_],
                                      p
                       )
    );

    Kc            = max(1, PB_Cnumroc(JC+N-1,
                                      0,
                                      DESCC[INB_],
                                      DESCC[NB_],
                                      mycol,
                                      DESCC[CSRC_],
                                      q
                       )
    );

    A = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCA[LLD_]*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    C = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCC[LLD_]*Kc)
    );

    if (C == NULL)
    {
       printf("Cannot allocate C\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pcmatgen(&ictxt, "N", "N", &DESCA[M_],
                 &DESCA[N_], &DESCA[MB_], &DESCA[NB_], A, &DESCA[LLD_],
                 &DESCA[RSRC_], &DESCA[CSRC_], &iaseed, &c__0, &DESCA[LLD_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    {
       int     iaseed = 200;
       hscalt_pcmatgen(&ictxt, "N", "N", &DESCC[M_],
                 &DESCC[N_], &DESCC[MB_], &DESCC[NB_], C, &DESCC[LLD_],
                 &DESCC[RSRC_], &DESCC[CSRC_], &iaseed, &c__0, &DESCC[LLD_], &c__0, &Kc,
                 &myrow, &mycol, &p, &q);
    }

    pctradd_(
      UPLO, TRANS, 
      &M, &N, 
      (float*)&ALPHA,
      (float*)A, &IA, &JA, DESCA,
      (float*)&BETA,
      (float*)C, &IC, &JC, DESCC
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCC[LLD_]*Kc); i++)
          {
              printf("C[%d]=(%f, %f) ", i, C[i].r, C[i].i);
          }
          printf("\n");
       }
    }

    free(A);
    free(C);

    Cblacs_gridexit(ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "UPLO=%s, TRANS=%s, p=%d, q=%d, M=%d, N=%d, MB=%d, NB=%d, "
          "algorithm execution time=%f, total time(sec)=%f\n",  
          UPLO,
          TRANS,
          p, 
          q,
          M,
          N,
          MB,
          NB,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
