
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pctrmm_parameters.h"

int main(int argc, char **argv)
{
    int        DESCA[DLEN_], DESCB[DLEN_];
    int        i, rc;
    int        ictxt, myrow, mycol;
    int        Ka, Kb;
    int        CHECK_RESULT = 0;
    int     c__0    = 0;


    int     p = nprow;
    int     q = npcol;
    HMPI_Group gid;
    MPI_Comm   pctrmmcomm;

    hscalcomplex    *A, *B;
 
    struct timeval start, end;
    struct timeval astart, aend;

    char SIDEC  = SIDEI;
    char *SIDE  = &SIDEC;
    char UPLOC  = UPLOI;
    char *UPLO  = &UPLOC;
    char TRANSC = TRANSI;
    char *TRANS = &TRANSC;
    char DIAGC  = DIAGI;
    char *DIAG  = &DIAGC;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pctrmmcomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pctrmmcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    if (SIDEC == 'L')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = M;
       DESCA[N_]     = M;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = MB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = MB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = N;
       DESCA[N_]     = N;
       DESCA[IMB_]   = NB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = NB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }

    DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCB[CTXT_]  = ictxt;
    DESCB[M_]     = M;
    DESCB[N_]     = N;
    DESCB[IMB_]   = MB;
    DESCB[INB_]   = NB;
    DESCB[MB_]    = MB;
    DESCB[NB_]    = NB;
    DESCB[RSRC_]  = 0;
    DESCB[CSRC_]  = 0;

    DESCA[LLD_]   = max(1, PB_Cnumroc(DESCA[M_],
                                      0,
                                      DESCA[IMB_],
                                      DESCA[MB_],
                                      myrow,
                                      DESCA[RSRC_],
                                      p
                       )
    );

    if (SIDEC == 'L')
    {
        Ka            = max(1, PB_Cnumroc(JA+M-1,
                                          0,
                                          DESCA[INB_],
                                          DESCA[NB_],
                                          mycol,
                                          DESCA[CSRC_],
                                          q
                           )
        );
    }
    else
    {
        Ka            = max(1, PB_Cnumroc(JA+N-1,
                                          0,
                                          DESCA[INB_],
                                          DESCA[NB_],
                                          mycol,
                                          DESCA[CSRC_],
                                          q
                           )
        );
    }

    DESCB[LLD_]   = max(1, PB_Cnumroc(DESCB[M_],
                                      0,
                                      DESCB[IMB_],
                                      DESCB[MB_],
                                      myrow,
                                      DESCB[RSRC_],
                                      p
                       )
    );

    Kb            = max(1, PB_Cnumroc(JB+N-1,
                                      0,
                                      DESCB[INB_],
                                      DESCB[NB_],
                                      mycol,
                                      DESCB[CSRC_],
                                      q
                       )
    );

    A = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCA[LLD_]*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    B = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (DESCB[LLD_]*Kb)
    );

    if (B == NULL)
    {
       printf("Cannot allocate B\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pcmatgen(&ictxt, "N", "N", &DESCA[M_],
                 &DESCA[N_], &DESCA[MB_], &DESCA[NB_], A, &DESCA[LLD_],
                 &DESCA[RSRC_], &DESCA[CSRC_], &iaseed, &c__0, &DESCA[LLD_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    {
       int     iaseed = 200;
       hscalt_pcmatgen(&ictxt, "N", "N", &DESCB[M_],
                 &DESCB[N_], &DESCB[MB_], &DESCB[NB_], B, &DESCB[LLD_],
                 &DESCB[RSRC_], &DESCB[CSRC_], &iaseed, &c__0, &DESCB[LLD_], &c__0, &Kb,
                 &myrow, &mycol, &p, &q);
    }

    pctrmm_(
      SIDE, UPLO, 
      TRANS, DIAG,
      &M, &N,
      (float*)&ALPHA,
      (float*)A, &IA, &JA, DESCA,
      (float*)B, &IB, &JB, DESCB
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCB[LLD_]*Kb); i++)
          {
              printf("B[%d]=(%f, %f) ", i, B[i].r, B[i].i);
          }
          printf("\n");
       }
    }

    free(A);
    free(B);

    Cblacs_gridexit(ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "SIDE=%s, UPLO=%s, TRANS=%s, DIAG=%s, "
          "p=%d, q=%d, M=%d, N=%d, MB=%d, NB=%d, "
          "Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          SIDE,
          UPLO,
          TRANS,
          DIAG,
          p,
          q,
          M,
          N,
          MB,
          NB,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
