
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pzsyr2k_parameters.h"

int main(int argc, char **argv)
{
    int        DESCA[DLEN_], DESCB[DLEN_], DESCC[DLEN_];
    int        i, rc;
    int        ictxt, myrow, mycol;
    int        Ka, Kb, Kc;
    int        CHECK_RESULT = 0;
    int     c__0    = 0;


    int     p = nprow;
    int     q = npcol;
    HMPI_Group gid;
    MPI_Comm   pzsyr2kcomm;

    hscaldoublecomplex    *A, *B, *C;
 
    struct timeval start, end;
    struct timeval astart, aend;

    char UPLOC = UPLOI;
    char *UPLO = &UPLOC;
    char TRANSC = TRANSI;
    char *TRANS = &TRANSC;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pzsyr2kcomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pzsyr2kcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    if (TRANSC == 'N')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = N;
       DESCA[N_]     = K;
       DESCA[IMB_]   = NB;
       DESCA[INB_]   = KB;
       DESCA[MB_]    = NB;
       DESCA[NB_]    = KB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[CTXT_]  = ictxt;
       DESCA[M_]     = K;
       DESCA[N_]     = N;
       DESCA[IMB_]   = KB;
       DESCA[INB_]   = NB;
       DESCA[MB_]    = KB;
       DESCA[NB_]    = NB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }

    if (TRANSC == 'N')
    {
       DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCB[CTXT_]  = ictxt;
       DESCB[M_]     = N;
       DESCB[N_]     = K;
       DESCB[IMB_]   = NB;
       DESCB[INB_]   = KB;
       DESCB[MB_]    = NB;
       DESCB[NB_]    = KB;
       DESCB[RSRC_]  = 0;
       DESCB[CSRC_]  = 0;
    }
    else
    {
       DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCB[CTXT_]  = ictxt;
       DESCB[M_]     = K;
       DESCB[N_]     = N;
       DESCB[IMB_]   = KB;
       DESCB[INB_]   = NB;
       DESCB[MB_]    = KB;
       DESCB[NB_]    = NB;
       DESCB[RSRC_]  = 0;
       DESCB[CSRC_]  = 0;
    }

    DESCC[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCC[CTXT_]  = ictxt;
    DESCC[M_]     = N;
    DESCC[N_]     = N;
    DESCC[IMB_]   = NB;
    DESCC[INB_]   = NB;
    DESCC[MB_]    = NB;
    DESCC[NB_]    = NB;
    DESCC[RSRC_]  = 0;
    DESCC[CSRC_]  = 0;

    DESCA[LLD_]   = max(1, PB_Cnumroc(DESCA[M_],
                                      0,
                                      DESCA[IMB_],
                                      DESCA[MB_],
                                      myrow,
                                      DESCA[RSRC_],
                                      p
                       )
    );

    if (TRANSC == 'N')
    {
        Ka            = max(1, PB_Cnumroc(JA+K-1,
                                          0,
                                          DESCA[INB_],
                                          DESCA[NB_],
                                          mycol,
                                          DESCA[CSRC_],
                                          q
                           )
        );
    }
    else
    {
        Ka            = max(1, PB_Cnumroc(JA+N-1,
                                          0,
                                          DESCA[INB_],
                                          DESCA[NB_],
                                          mycol,
                                          DESCA[CSRC_],
                                          q
                           )
        );
    }

    DESCB[LLD_]   = max(1, PB_Cnumroc(DESCB[M_],
                                      0,
                                      DESCB[IMB_],
                                      DESCB[MB_],
                                      myrow,
                                      DESCB[RSRC_],
                                      p
                       )
    );

    if (TRANSC == 'N')
    {
        Kb            = max(1, PB_Cnumroc(JB+K-1,
                                          0,
                                          DESCB[INB_],
                                          DESCB[NB_],
                                          mycol,
                                          DESCB[CSRC_],
                                          q
                           )
        );
    }
    else
    {
        Kb            = max(1, PB_Cnumroc(JB+N-1,
                                          0,
                                          DESCB[INB_],
                                          DESCB[NB_],
                                          mycol,
                                          DESCB[CSRC_],
                                          q
                           )
        );
    }

    DESCC[LLD_]   = max(1, PB_Cnumroc(DESCC[M_],
                                      0,
                                      DESCC[IMB_],
                                      DESCC[MB_],
                                      myrow,
                                      DESCC[RSRC_],
                                      p
                       )
    );

    Kc            = max(1, PB_Cnumroc(JC+N-1,
                                      0,
                                      DESCC[INB_],
                                      DESCC[NB_],
                                      mycol,
                                      DESCC[CSRC_],
                                      q
                       )
    );

    A = (hscaldoublecomplex*)malloc(
                  sizeof(hscaldoublecomplex)
                  *
                  (DESCA[LLD_]*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    B = (hscaldoublecomplex*)malloc(
                  sizeof(hscaldoublecomplex)
                  *
                  (DESCB[LLD_]*Kb)
    );

    if (B == NULL)
    {
       printf("Cannot allocate B\n");
       HMPI_Finalize(-1);
    }

    C = (hscaldoublecomplex*)malloc(
                  sizeof(hscaldoublecomplex)
                  *
                  (DESCC[LLD_]*Kc)
    );

    if (C == NULL)
    {
       printf("Cannot allocate C\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pzmatgen(&ictxt, "N", "N", &DESCA[M_],
                 &DESCA[N_], &DESCA[MB_], &DESCA[NB_], A, &DESCA[LLD_],
                 &DESCA[RSRC_], &DESCA[CSRC_], &iaseed, &c__0, &DESCA[LLD_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    {
       int     iaseed = 200;
       hscalt_pzmatgen(&ictxt, "N", "N", &DESCB[M_],
                 &DESCB[N_], &DESCB[MB_], &DESCB[NB_], B, &DESCB[LLD_],
                 &DESCB[RSRC_], &DESCB[CSRC_], &iaseed, &c__0, &DESCB[LLD_], &c__0, &Kb,
                 &myrow, &mycol, &p, &q);
    }

    {
       int     iaseed = 300;
       hscalt_pzmatgen(&ictxt, "S", "N", &DESCC[M_],
                 &DESCC[N_], &DESCC[MB_], &DESCC[NB_], C, &DESCC[LLD_],
                 &DESCC[RSRC_], &DESCC[CSRC_], &iaseed, &c__0, &DESCC[LLD_], &c__0, &Kc,
                 &myrow, &mycol, &p, &q);
    }

    pzsyr2k_(
      UPLO, TRANS, 
      &N, &K, 
      (double*)&ALPHA,
      (double*)A, &IA, &JA, DESCA,
      (double*)B, &IB, &JB, DESCB,
      (double*)&BETA,
      (double*)C, &IC, &JC, DESCC
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCC[LLD_]*Kc); i++)
          {
              printf("C[%d]=(%f, %f) ", i, C[i].r, C[i].i);
          }
          printf("\n");
       }
    }

    free(A);
    free(B);
    free(C);

    Cblacs_gridexit(ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "UPLO=%s, TRANS=%s, p=%d, q=%d, N=%d, K=%d, NB=%d, KB=%d, "
          "Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          UPLO,
          TRANS,
          p,
          q,
          N,
          K,
          NB,
          KB,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
