#!/bin/sh

if [ "$1" = "" ]; then
        echo "Usage: $0 {file to test} {parameters file}"
        echo "For example:"
        echo "$0 test_scal_pcgemm scalapack_pcgemm_parameters.h"
        exit 1
fi

if [ "$#" != "2" ]; then
        echo "Usage: $0 {file to test} {parameters file}"
        echo "For example:"
        echo "$0 test_scal_pcgemm scalapack_pcgemm_parameters.h"
        exit 2
fi

for arg in $*
do
    case $arg in 
    -help | -h | --help | --h)
	echo "Usage: $0 {file to test}"
	echo "For example:"
	echo "$0 test_scal_pcgemm scalapack_pcgemm_parameters.h"
	exit 0
        ;;
    esac
done

TESTFILE="$1"
PARAMETERSFILE="$2"

echo "Killing rogue '$TESTFILE' processes"
killall $TESTFILE > /dev/null 2>&1

echo "Removing '$TESTFILE' from MPCLOAD"
/bin/rm -f `find $MPCLOAD/.. -name $TESTFILE`

echo "Broadcasting '${TESTFILE}.c' '${PARAMETERSFILE}'"
hmpibcast ${TESTFILE}.c ${PARAMETERSFILE}

if [ $? -gt 0 ]; then
	echo Error broadcasting the files
	exit 2
fi

echo "Creating '${TESTFILE}'"
hmpiload -g -I\$HSCALAPACKHOME/include -o ${TESTFILE} ${TESTFILE}.c -L\$HSCALAPACKHOME/lib -lhscaltest -lhscalapack -L\$LIBSCALAPACKDIR\$LIBSSCALAPACK -L\$LIBBLACSDIR\$LIBSBLACS -L\$LIBBLASDIR\$LIBSBLAS -L\$LIBF2CDIR\$LIBSF2C

if [ $? -gt 0 ]; then
	echo Error while loading the files
	exit 3
fi

echo "Running '${TESTFILE}'"
hmpirun ${TESTFILE}
