
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pcposv_parameters.h"

int main(int argc, char **argv)
{
    int DESCA[DLEN1_], DESCB[DLEN1_];
    int i, rc;
    int ictxt, myrow, mycol;
    int LLDa, Ka, LLDb, Kb;
    int CHECK_RESULT = 0;
    int c__0    = 0;

    int IRSRC = 0;
    int ICSRC = 0;
    int DLLD  = -1;
    int DCTXT = -1;
    int INFO;

    char UPLOC = UPLOI; char *UPLO = &UPLOC;

    int     p, q;
    hscalcomplex    *A, *B;
 
    struct timeval start, end;
    struct timeval cstart, cend;
    struct timeval astart, aend;

    gettimeofday(&start, NULL);

    //
    // Initialize HScaLAPACK runtime
    hscal_init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    descset_(
        DESCA,
        &N,
        &N,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &DCTXT,
        &DLLD
    );

    descset_(
        DESCB,
        &N,
        &NRHS,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &DCTXT,
        &DLLD
    );

    if (HMPI_Is_host())
    {
       gettimeofday(&cstart, NULL);
    }

    HMPI_Debug(1);
    hscal_set_debug(HSCAL_LOG_NONE);

    rc = hscal_pcposv_ctxt(
              UPLO,
              &N,
              &NRHS,
              &IA,
              &JA,
              DESCA,
              &IB,
              &JB,
              DESCB,
              &ictxt
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    hscal_set_debug(HSCAL_LOG_NONE);
    HMPI_Debug(0);

    if (HMPI_Is_host())
    {
       gettimeofday(&cend, NULL);
    }

    if (!hscal_in_ctxt(&ictxt))
    {
       hscal_free_ctxt(&ictxt);
       hscal_finalize(0);
       return 0;
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&astart, NULL);
    }

    /*
     * Retrieve the process grid information
     */
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    LLDa           = max(1, PB_Cnumroc(N,
                                       0,
                                       NB,
                                       NB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    );

    Ka            = max(1, PB_Cnumroc(JA+N-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    LLDb           = max(1, PB_Cnumroc(N,
                                       0,
                                       NB,
                                       NB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    );

    Kb            = max(1, PB_Cnumroc(JB+NRHS-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    descset_(
        DESCA,
        &N,
        &N,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDa
    );

    descset_(
        DESCB,
        &N,
        &NRHS,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDb
    );

    A = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (LLDa*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    B = (hscalcomplex*)malloc(
                  sizeof(hscalcomplex)
                  *
                  (LLDb*Kb)
    );

    if (B == NULL)
    {
       printf("Cannot allocate B\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pcmatgen(&ictxt, "S", "D", &DESCA[M1_],
                 &DESCA[N1_], &DESCA[MB1_], &DESCA[NB1_], A, &DESCA[LLD1_],
                 &DESCA[RSRC1_], &DESCA[CSRC1_], &iaseed, &c__0, &DESCA[LLD1_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    {
       int     iaseed = 200;
       hscalt_pcmatgen(&ictxt, "N", "N", &DESCB[M1_],
                 &DESCB[N1_], &DESCB[MB1_], &DESCB[NB1_], B, &DESCB[LLD1_],
                 &DESCB[RSRC1_], &DESCB[CSRC1_], &iaseed, &c__0, &DESCB[LLD1_], &c__0, &Kb,
                 &myrow, &mycol, &p, &q);
    }

    pcposv_(
      UPLO,
      &N, &NRHS,
      (float*)A, &IA, &JA, DESCA,
      (float*)B, &IB, &JB, DESCB,
      &INFO
    );

    free(A);
    free(B);

    hscal_free_ctxt(&ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       double ctstart = cstart.tv_sec + cstart.tv_usec/1000000.;
       double ctend = cend.tv_sec + cend.tv_usec/1000000.;
       double ct_sec = ctend - ctstart;

       printf(
          "N=%d, NRHS=%d, NB=%d, INFO=%d, "
          "Context creation time(sec)=%f, Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          N,
          NRHS,
          NB,
          INFO,
          ct_sec,
          at_sec,
          t_sec
       );
    }

    hscal_finalize(0);
}
