
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pdgesv_parameters.h"

int main(int argc, char **argv)
{
    int DESCA[DLEN1_], DESCB[DLEN1_];
    int i, j, rc;
    int ictxt, myrow, mycol;
    int LLDa, Ka, LLDb, Kb, Kipiv;
    int CHECK_RESULT = 0;
    int c__0    = 0;

    int IRSRC = 0;
    int ICSRC = 0;
    int DLLD  = -1;
    int DCTXT = -1;
    int INFO;

    int     p, q;
 
    //
    // Initialize HScaLAPACK runtime
    hscal_init(&argc, &argv);

    descset_(
        DESCA,
        &N,
        &N,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &DCTXT,
        &DLLD
    );

    descset_(
        DESCB,
        &N,
        &NRHS,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &DCTXT,
        &DLLD
    );

    rc = hscal_pdgetrf_ctxt(
              &N,
              &N,
              &IA,
              &JA,
              DESCA,
              &ictxt
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    if (!hscal_in_ctxt(&ictxt))
    {
       hscal_free_ctxt(&ictxt);
       hscal_finalize(0);
       return 0;
    }

    /*
     * Retrieve the process grid information
     */
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    {
       int *counts, *displs;
       char *proc_str;
       char (*proc_table)[MPI_MAX_PROCESSOR_NAME];
       int *proc_array;

       char pname[MPI_MAX_PROCESSOR_NAME];
       int plen;
       MPI_Comm *pdgesvcomm = hscal_get_comm(&ictxt);

       MPI_Get_processor_name(
          pname,
          &plen
       );

       {
           if (HMPI_Is_host())
           {
              proc_str = (char*)malloc(
                         sizeof(char)
                         *
                         (MPI_MAX_PROCESSOR_NAME)*(p*q)
              );

              if (proc_str == NULL)
              {
                 return -1;
              }

              proc_table = (char (*) [MPI_MAX_PROCESSOR_NAME])malloc(
                           sizeof(char[MPI_MAX_PROCESSOR_NAME])
                           *
                           (p*q)
              );

              if (proc_table == NULL)
              {
                 return -1;
              }

              proc_array = (int*)malloc(
                           sizeof(int)
                           *
                           (p*q)
              );

              if (proc_array == NULL)
              {
                 return -1;
              }

              counts = (int*)malloc(
                               sizeof(int)
                               *
                               (p*q)
              );

              if (counts == NULL)
              {
                 return -1;
              }

              displs = (int*)malloc(
                               sizeof(int)
                               *
                               (p*q)
              );

              if (displs == NULL)
              {
                 return MPC_ERR_NOMEM;
              }

              for (i = 0; i < (p*q); i++)
              {
                  counts[i] = MPI_MAX_PROCESSOR_NAME;
                  displs[i] = (i*MPI_MAX_PROCESSOR_NAME);
              }
           }

           if (HMPI_Is_host())
           {
              rc = MPI_Gatherv(
                      &pname,
                      MPI_MAX_PROCESSOR_NAME,
                      MPI_CHAR,
                      proc_str,
                      counts,
                      displs,
                      MPI_CHAR,
                      0,
                      *pdgesvcomm
              );

              if (rc != MPI_SUCCESS)
              {
                 return rc;
              }
           }
           else
           {
              rc = MPI_Gatherv(
                      &pname,
                      MPI_MAX_PROCESSOR_NAME,
                      MPI_CHAR,
                      NULL,
                      NULL,
                      NULL,
                      MPI_CHAR,
                      0,
                      *pdgesvcomm
              );

              if (rc != MPI_SUCCESS)
              {
                 return rc;
              }
           }

           if (HMPI_Is_host())
           {
              int tmpc = 0;

              for (i = 0; i < p; i++)
              {
                  for (j = 0; j < q; j++)
                  {
                      char* tmp = proc_str
                                  +
                                  (tmpc)*(MPI_MAX_PROCESSOR_NAME)
                      ;
                      printf("%s ", tmp);
                      tmpc++;
                  }
                  printf("\n");
              }
              printf("\n");

              free(counts);
              free(displs);
              free(proc_str);
              free(proc_table);
              free(proc_array);
           }
       }
    }

    hscal_free_ctxt(&ictxt);

    if (HMPI_Is_host())
    {
       printf(
          "N=%d, NRHS=%d, NB=%d\n",
          N,
          NRHS,
          NB
       );
    }

    hscal_finalize(0);
}
