
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_psgels_parameters.h"

int main(int argc, char **argv)
{
    int DESCA[DLEN1_], DESCB[DLEN1_];
    int i, rc;
    int ictxt, myrow, mycol;
    int LLDa, Ka, LLDb_mp, LLDb_np, LLDb, Kb, lwork;
    int CHECK_RESULT = 0;
    int c__0    = 0;
    int max_M_N = max(M, N);

    int IRSRC = 0;
    int ICSRC = 0;
    int DLLD  = -1;
    int DCTXT = -1;
    int INFO;

    char TRANSC = TRANSI; 
    char *TRANS = &TRANSC;

    int     p, q, nprow, npcol;

    hscalreal    *A, *B, *work;
 
    struct timeval start, end;
    struct timeval cstart, cend;
    struct timeval astart, aend;

    hscalinteger lcmp, mp, mq, np, nq, nrhsq, ltau, lwf, lws, lcm, mnp,
                 i__1, i__2, i__3, i__4, i__5, i__6, i__7, i__8, i__9;

    gettimeofday(&start, NULL);

    //
    // Initialize HScaLAPACK runtime
    hscal_init(&argc, &argv);

    descset_(
        DESCA,
        &M,
        &N,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &DCTXT,
        &DLLD
    );

    descset_(
        DESCB,
        &max_M_N,
        &NRHS,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &DCTXT,
        &DLLD
    );

    if (HMPI_Is_host())
    {
       gettimeofday(&cstart, NULL);
    }

    HMPI_Debug(1);

    rc = hscal_psgels_ctxt(
              TRANS,
              &M,
              &N,
              &NRHS,
              &IA,
              &JA,
              DESCA,
              &IB,
              &JB,
              DESCB,
              &ictxt
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HMPI_Debug(0);

    if (HMPI_Is_host())
    {
       gettimeofday(&cend, NULL);
    }

    if (!hscal_in_ctxt(&ictxt))
    {
       hscal_free_ctxt(&ictxt);
       hscal_finalize(0);
       return 0;
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&astart, NULL);
    }

    /*
     * Retrieve the process grid information
     */
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    nprow = p;
    npcol = q;

    LLDa           = max(1, PB_Cnumroc(M,
                                       0,
                                       MB,
                                       MB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    );

    Ka            = max(1, PB_Cnumroc(JA+N-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    LLDb_mp        = max(1, PB_Cnumroc(M,
                                       0,
                                       MB,
                                       MB,
                                       myrow,
                                       IRSRC,
                                       p
                         )
    );

    LLDb_np        = max(1, PB_Cnumroc(N,
                                       0,
                                       NB,
                                       NB,
                                       myrow,
                                       IRSRC,
                                       p
                         )
    );

    LLDb = max(LLDb_mp, LLDb_np);

    Kb            = max(1, PB_Cnumroc(JB+NRHS-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    descset_(
        DESCA,
        &M,
        &N,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDa
    );

    descset_(
        DESCB,
        &max_M_N,
        &NRHS,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDb
    );

    A = (hscalreal*)malloc(
                  sizeof(hscalreal)
                  *
                  (LLDa*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    B = (hscalreal*)malloc(
                  sizeof(hscalreal)
                  *
                  (LLDb*Kb)
    );

    if (B == NULL)
    {
       printf("Cannot allocate B\n");
       HMPI_Finalize(-1);
    }

    /*
    * Calculate the amount of workspace for PCGELS 
    */
    mp    = numroc_(&M, &NB, &myrow, &c__0, &nprow);
    mq    = numroc_(&M, &NB, &mycol, &c__0, &npcol);
    np    = numroc_(&N, &NB, &myrow, &c__0, &nprow);
    mnp   = max(mp,np);
    nq    = numroc_(&N, &NB, &mycol, &c__0, &npcol);
    nrhsq = numroc_(&NRHS, &NB, &mycol, &c__0, &npcol);

    if (M >= N) {
       i__6 = min(M,N);
       ltau = numroc_(&i__6, &NB, &mycol, &c__0, &npcol);
       lwf  = NB * (mp + nq + NB);
       i__6 = NB * (NB - 1) / 2;
       i__7 = (mp + nrhsq) * NB;
       lws  = max(i__6,i__7) + NB * NB;
    } else {
       lcm  = ilcm_(&nprow, &npcol);
       lcmp = lcm / nprow;
       i__6 = min(M,N);
       ltau = numroc_(&i__6, &NB, &myrow, &c__0, &nprow);
       lwf  = NB * (mp + nq + NB);
       i__9 = numroc_(&N, &NB, &c__0, &c__0, &nprow);
       i__8 = nq + numroc_(&i__9, &NB, &c__0, &c__0, &lcmp);
       i__6 = NB * (NB - 1) / 2;
       i__7 = (np + max(i__8,nrhsq)) * NB;
       lws = max(i__6,i__7) + NB * NB;
    }

    lwork = ltau + max(lwf,lws);

    work = (hscalreal*)malloc(
                 sizeof(hscalreal)
                 *
                 (lwork)
    );

    if (work == NULL)
    {
       printf("Cannot allocate work\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_psmatgen(&ictxt, "N", "N", &DESCA[M1_],
                 &DESCA[N1_], &DESCA[MB1_], &DESCA[NB1_], A, &DESCA[LLD1_],
                 &DESCA[RSRC1_], &DESCA[CSRC1_], &iaseed, &c__0, &DESCA[LLD1_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    {
       int     iaseed = 200;
       hscalt_psmatgen(&ictxt, "N", "N", &DESCB[M1_],
                 &DESCB[N1_], &DESCB[MB1_], &DESCB[NB1_], B, &DESCB[LLD1_],
                 &DESCB[RSRC1_], &DESCB[CSRC1_], &iaseed, &c__0, &DESCB[LLD1_], &c__0, &Kb,
                 &myrow, &mycol, &p, &q);
    }

    psgels_(
      TRANS,
      &M, &N, &NRHS,
      (float*)A, &IA, &JA, DESCA,
      (float*)B, &IB, &JB, DESCB,
      (float*)work, &lwork,
      &INFO
    );

    free(A);
    free(B);
    free(work);

    hscal_free_ctxt(&ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       double ctstart = cstart.tv_sec + cstart.tv_usec/1000000.;
       double ctend = cend.tv_sec + cend.tv_usec/1000000.;
       double ct_sec = ctend - ctstart;

       printf(
          "M=%d, N=%d, NRHS=%d, MB=%d, NB=%d, INFO=%d, "
          "Context creation time(sec)=%f, Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          M,
          N,
          NRHS,
          MB,
          NB,
          INFO,
          ct_sec,
          at_sec,
          t_sec
       );
    }

    hscal_finalize(0);
}
