
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pzgelqf_parameters.h"

int main(int argc, char **argv)
{
    int DESCA[DLEN1_], DESCB[DLEN1_];
    int i, rc;
    int ictxt, myrow, mycol;
    int LLDa, Ka, LLDb_mp, LLDb_np, LLDb, Kb, lwork;
    int CHECK_RESULT = 0;
    int c__0    = 0;
    int max_M_N = max(M, N);

    int IRSRC = 0;
    int ICSRC = 0;
    int DLLD  = -1;
    int DCTXT = -1;
    int INFO;

    int     p, q, nprow, npcol;

    hscaldoublecomplex    *A, *work, *tau;
 
    struct timeval start, end;
    struct timeval cstart, cend;
    struct timeval astart, aend;

    hscalinteger lcmp, mp, mq, np, nq, nrhsq, ltau, lwf, lws, lcm, mnp,
                 i__1, i__2, i__3, i__4, i__5, i__6, i__7, i__8, i__9;

    gettimeofday(&start, NULL);

    //
    // Initialize HScaLAPACK runtime
    hscal_init(&argc, &argv);

    descset_(
        DESCA,
        &M,
        &N,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &DCTXT,
        &DLLD
    );

    if (HMPI_Is_host())
    {
       gettimeofday(&cstart, NULL);
    }

    HMPI_Debug(1);

    rc = hscal_pzgelqf_ctxt(
              &M,
              &N,
              &IA,
              &JA,
              DESCA,
              &ictxt
    );

    if (rc != HSCAL_SUCCESS)
    {
       return rc;
    }

    HMPI_Debug(0);

    if (HMPI_Is_host())
    {
       gettimeofday(&cend, NULL);
    }

    if (!hscal_in_ctxt(&ictxt))
    {
       hscal_free_ctxt(&ictxt);
       hscal_finalize(0);
       return 0;
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&astart, NULL);
    }

    /*
     * Retrieve the process grid information
     */
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    nprow = p;
    npcol = q;

    LLDa           = max(1, PB_Cnumroc(M,
                                       0,
                                       MB,
                                       MB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    );

    Ka            = max(1, PB_Cnumroc(JA+N-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    descset_(
        DESCA,
        &M,
        &N,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDa
    );

    A = (hscaldoublecomplex*)malloc(
                  sizeof(hscaldoublecomplex)
                  *
                  (LLDa*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    /*
    * Calculate the amount of workspace for PCGELS 
    */
    mp    = numroc_(&M, &NB, &myrow, &c__0, &nprow);
    nq    = numroc_(&N, &NB, &mycol, &c__0, &npcol);
    lwork = DESCA[4] * (mp + nq + DESCA[4]);

    work = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (lwork)
    );

    if (work == NULL)
    {
       printf("Cannot allocate work\n");
       HMPI_Finalize(-1);
    }

    int min_M_N   = ((M <= N) ? M : N);
    ltau      = max(1, PB_Cnumroc(IA+min_M_N-1,
                                  0,
                                  MB,
                                  MB,
                                  myrow,
                                  IRSRC,
                                  p
                    )
    );

    tau  = (hscaldoublecomplex*)malloc(
                 sizeof(hscaldoublecomplex)
                 *
                 (ltau)
    );

    if (tau  == NULL)
    {
       printf("Cannot allocate tau \n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pzmatgen(&ictxt, "N", "N", &DESCA[M1_],
                 &DESCA[N1_], &DESCA[MB1_], &DESCA[NB1_], A, &DESCA[LLD1_],
                 &DESCA[RSRC1_], &DESCA[CSRC1_], &iaseed, &c__0, &DESCA[LLD1_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    pzgelqf_(
      &M, &N,
      (float*)A, &IA, &JA, DESCA,
      (float*)tau,
      (float*)work, &lwork,
      &INFO
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCA[LLD1_]*Ka); i++)
          {
              printf("A[%d]=(%f, %f) ", i, A[i].r, A[i].i);
          }
          printf("\n");
       }
    }

    free(A);
    free(tau);
    free(work);

    hscal_free_ctxt(&ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       double ctstart = cstart.tv_sec + cstart.tv_usec/1000000.;
       double ctend = cend.tv_sec + cend.tv_usec/1000000.;
       double ct_sec = ctend - ctstart;

       printf(
          "M=%d, N=%d, MB=%d, NB=%d, INFO=%d, "
          "Context creation time(sec)=%f, Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          M,
          N,
          MB,
          NB,
          INFO,
          ct_sec,
          at_sec,
          t_sec
       );
    }

    hscal_finalize(0);
}
