
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pdgesvd_parameters.h"

/*-----------------------------------------------------------*/

void printwtime_(char *routine, int *myrow, int *mycol, double etim[2])
{
     char r[8];
     strncpy(r, routine, 7);
     r[7] = '\0';
     double t_sec = etim[0] + etim[1]/1000000.;
     printf("(%d,%d): %s Elapsed time %f\n", *myrow, *mycol, r, t_sec);
     return;
}

/*-----------------------------------------------------------*/

void wtime_(double tim[2], int *ierror)
{
     struct timeval time1; 
     long tim1, tim2; 
     int ierr; 
     double elap; 
     ierr = gettimeofday(&time1, NULL) ; 
     *ierror = ierr; 
     if (ierr != 0 ) printf("bad return of gettimeofday, ierr = %d \n",ierr);  
     tim[0] = time1.tv_sec;
     tim[1] = time1.tv_usec; 
     return;
}

/*-----------------------------------------------------------*/

int main(int argc, char **argv)
{
    int DESCA[DLEN1_], DESCU[DLEN1_], DESCVT[DLEN1_];
    int i, rc;
    int ictxt, myrow, mycol;
    int LLDa, Ka, lwork;
    int CHECK_RESULT = 0;
    int c__0    = 0;

    int IRSRC = 0;
    int ICSRC = 0;
    int INFO;

    char JOBVTC = JOBVTI; char *JOBVT = &JOBVTC;
    char JOBUC = JOBUI; char *JOBU = &JOBUC;

    int     p = nprow;
    int     q = npcol;
    HMPI_Group gid;
    MPI_Comm   pdgesvdcomm;

    hscaldoublereal    *A, *S, *U, *VT, *work;
 
    struct timeval start, end;
    struct timeval astart, aend;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pdgesvdcomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pdgesvdcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    LLDa           = max(1, PB_Cnumroc(IA+M-1,
                                       0,
                                       MB,
                                       MB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    );

    Ka            = max(1, PB_Cnumroc(JA+N-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    descset_(
        DESCA,
        &M,
        &N,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDa
    );

    A = (hscaldoublereal*)malloc(
                  sizeof(hscaldoublereal)
                  *
                  (LLDa*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    int SIZE = ((M <= N) ? M : N);

    S = (hscaldoublereal*)malloc(
                  sizeof(hscaldoublereal)
                  *
                  (SIZE)
    );

    if (S == NULL)
    {
       printf("Cannot allocate S\n");
       HMPI_Finalize(-1);
    }

    int SIZEP = numroc_(&SIZE, &MB, &myrow, &c__0, &nprow);
    int SIZEQ = numroc_(&SIZE, &NB, &mycol, &c__0, &npcol);

    U = (hscaldoublereal*)malloc(
                  sizeof(hscaldoublereal)
                  *
                  (LLDa*SIZEQ)
    );

    if (U == NULL)
    {
       printf("Cannot allocate U\n");
       HMPI_Finalize(-1);
    }

    descset_(
        DESCU,
        &M,
        &SIZE,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDa
    );

    VT = (hscaldoublereal*)malloc(
                  sizeof(hscaldoublereal)
                  *
                  (SIZEP*Ka)
    );

    if (VT== NULL)
    {
       printf("Cannot allocate VT\n");
       HMPI_Finalize(-1);
    }

    descset_(
        DESCVT,
        &SIZE,
        &N,
        &MB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &SIZEP
    );

    lwork = -1;

    work = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
    );

    if (work == NULL)
    {
       printf("Cannot allocate work\n");
       HMPI_Finalize(-1);
    }

    pdgesvd_(
      JOBU,
      JOBVT,
      &M, &N,
      (double*)A, &IA, &JA, DESCA,
      (double*)S,
      (double*)U, &IU, &JU, DESCU,
      (double*)VT, &IVT, &JVT, DESCVT,
      (double*)work, &lwork,
      &INFO
    );

    lwork = work[0];

    free(work);

    work = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (lwork)
    );

    if (work == NULL)
    {
       printf("Cannot allocate work\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pdmatgen(&ictxt, "N", "N", &DESCA[M1_],
                 &DESCA[N1_], &DESCA[MB1_], &DESCA[NB1_], A, &DESCA[LLD1_],
                 &DESCA[RSRC1_], &DESCA[CSRC1_], &iaseed, &c__0, &DESCA[LLD1_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    pdgesvd_(
      JOBU,
      JOBVT,
      &M, &N,
      (double*)A, &IA, &JA, DESCA,
      (double*)S,
      (double*)U, &IU, &JU, DESCU,
      (double*)VT, &IVT, &JVT, DESCVT,
      (double*)work, &lwork,
      &INFO
    );

    free(A);
    free(S);
    free(U);
    free(VT);
    free(work);

    Cblacs_gridexit(ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "p=%d, q=%d, M=%d, N=%d, MB=%d, NB=%d, INFO=%d, "
          "Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          p,
          q,
          M,
          N,
          MB,
          NB,
          INFO,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
