
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pdsyev_parameters.h"

/*-----------------------------------------------------------*/

void printwtime_(char *routine, int *myrow, int *mycol, double etim[2])
{
     char r[8];
     strncpy(r, routine, 7);
     r[7] = '\0';
     double t_sec = etim[0] + etim[1]/1000000.;
     printf("(%d,%d): %s Elapsed time %f\n", *myrow, *mycol, r, t_sec);
     return;
}

/*-----------------------------------------------------------*/

void wtime_(double tim[2], int *ierror)
{
     struct timeval time1; 
     long tim1, tim2; 
     int ierr; 
     double elap; 
     ierr = gettimeofday(&time1, NULL) ; 
     *ierror = ierr; 
     if (ierr != 0 ) printf("bad return of gettimeofday, ierr = %d \n",ierr);  
     tim[0] = time1.tv_sec;
     tim[1] = time1.tv_usec; 
     return;
}

/*-----------------------------------------------------------*/

int main(int argc, char **argv)
{
    int DESCA[DLEN1_], DESCZ[DLEN1_];
    int i, rc;
    int ictxt, myrow, mycol;
    int LLDa, Ka, LLDz, Kz, lwork;
    int CHECK_RESULT = 0;
    int c__0    = 0;

    int IRSRC = 0;
    int ICSRC = 0;
    int INFO;

    char UPLOC = UPLOI; char *UPLO = &UPLOC;
    char JOBZC = JOBZI; char *JOBZ = &JOBZC;

    int     p = nprow;
    int     q = npcol;
    HMPI_Group gid;
    MPI_Comm   pdsyevcomm;

    hscaldoublereal    *A, *W, *Z, *work;
 
    struct timeval start, end;
    struct timeval astart, aend;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pdsyevcomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pdsyevcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    LLDa           = max(1, PB_Cnumroc(N,
                                       0,
                                       NB,
                                       NB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    );

    Ka            = max(1, PB_Cnumroc(JA+N-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    descset_(
        DESCA,
        &N,
        &N,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDa
    );

    A = (hscaldoublereal*)malloc(
                  sizeof(hscaldoublereal)
                  *
                  (LLDa*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    W = (hscaldoublereal*)malloc(
                  sizeof(hscaldoublereal)
                  *
                  (N)
    );

    if (W == NULL)
    {
       printf("Cannot allocate W\n");
       HMPI_Finalize(-1);
    }

    LLDz           = max(1, PB_Cnumroc(N,
                                       0,
                                       NB,
                                       NB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    );

    Kz            = max(1, PB_Cnumroc(JZ+N-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    Z = (hscaldoublereal*)malloc(
                  sizeof(hscaldoublereal)
                  *
                  (LLDz*Kz)
    );

    if (Z == NULL)
    {
       printf("Cannot allocate Z\n");
       HMPI_Finalize(-1);
    }

    descset_(
        DESCZ,
        &N,
        &N,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDz
    );


    /*
    * Calculate the amount of workspace
    */
    int NEIG = N;
    int i__1 = max(N,NB);
    int NN = max(i__1,2);
    int np0 = numroc_(&NN, &NB, &c__0, &c__0, &nprow);
    int i__2 = max(NEIG,NB);
    i__1 = max(i__2,2);
    int mq0 = numroc_(&i__1, &NB, &c__0, &c__0, &npcol);
    i__1 = NN * 5, i__2 = np0 * mq0 + (NB << 1) * NB;
    int i__3 = nprow * npcol;
    int sizesyevx = N * 5 + max(i__1,i__2) + iceil_(&NEIG, &i__3) * NN;

    lwork = sizesyevx;

    work = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (lwork)
    );

    if (work == NULL)
    {
       printf("Cannot allocate work\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pdmatgen(&ictxt, "S", "N", &DESCA[M1_],
                 &DESCA[N1_], &DESCA[MB1_], &DESCA[NB1_], A, &DESCA[LLD1_],
                 &DESCA[RSRC1_], &DESCA[CSRC1_], &iaseed, &c__0, &DESCA[LLD1_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    pdsyev_(
      JOBZ,
      UPLO,
      &N,
      (double*)A, &IA, &JA, DESCA,
      (double*)W,
      (double*)Z, &IZ, &JZ, DESCZ,
      (double*)work, &lwork,
      &INFO
    );

    free(A);
    free(W);
    free(Z);
    free(work);

    Cblacs_gridexit(ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "p=%d, q=%d, N=%d, NB=%d, INFO=%d, "
          "Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          p,
          q,
          N,
          NB,
          INFO,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
