
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hmpi.h>

#include <scalapack.h>

#include "scalapack_pzgetrf_parameters.h"

int main(int argc, char **argv)
{
    int DESCA[DLEN1_];
    int i, rc;
    int ictxt, myrow, mycol;
    int LLDa, Ka, Kipiv;
    int CHECK_RESULT = 0;
    int c__0    = 0;

    int IRSRC = 0;
    int ICSRC = 0;
    int INFO;

    int     p = nprow;
    int     q = npcol;
    HMPI_Group gid;
    MPI_Comm   pzgetrfcomm;

    int             *IPIV;
    hscaldoublecomplex    *A;
 
    struct timeval start, end;
    struct timeval astart, aend;

    gettimeofday(&start, NULL);

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (HMPI_Is_host())
    {
        gettimeofday(&astart, NULL);
    }

    pzgetrfcomm = *(MPI_Comm*)HMPI_Get_comm(HMPI_COMM_WORLD_GROUP);

    /*
     * Translate algocomm to a BLACS handle
     */
    ictxt = Csys2blacs_handle(pzgetrfcomm);

    /*
     * Form BLACS context based on algocomm
     */
    Cblacs_gridinit(&ictxt, "r", p, q);
    Cblacs_gridinfo(ictxt, &p, &q, &myrow, &mycol);

    LLDa           = max(1, PB_Cnumroc(M,
                                       0,
                                       MB,
                                       MB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    );

    Ka            = max(1, PB_Cnumroc(JA+N-1,
                                      0,
                                      NB,
                                      NB,
                                      mycol,
                                      ICSRC,
                                      q
                       )
    );

    descset_(
        DESCA,
        &N,
        &N,
        &NB,
        &NB,
        &IRSRC,
        &ICSRC,
        &ictxt,
        &LLDa
    );

    A = (hscaldoublecomplex*)malloc(
                  sizeof(hscaldoublecomplex)
                  *
                  (LLDa*Ka)
    );

    if (A == NULL)
    {
       printf("Cannot allocate A\n");
       HMPI_Finalize(-1);
    }

    Kipiv          = max(1, PB_Cnumroc(M,
                                       0,
                                       MB,
                                       MB,
                                       myrow,
                                       IRSRC,
                                       p
                        )
    ) + MB;

    IPIV = (int*)malloc(
                 sizeof(int)
                 *
                 (Kipiv)
    );

    if (IPIV == NULL)
    {
       printf("Cannot allocate IPIV\n");
       HMPI_Finalize(-1);
    }

    {
       int     iaseed = 100;
       hscalt_pzmatgen(&ictxt, "N", "N", &DESCA[M1_],
                 &DESCA[N1_], &DESCA[MB1_], &DESCA[NB1_], A, &DESCA[LLD1_],
                 &DESCA[RSRC1_], &DESCA[CSRC1_], &iaseed, &c__0, &DESCA[LLD1_], &c__0, &Ka,
                 &myrow, &mycol, &p, &q);
    }

    pzgetrf_(
      &M, &N,
      (double*)A, &IA, &JA, DESCA,
      IPIV, &INFO
    );

    //
    // Check the results
    if (CHECK_RESULT)
    {
       if (HMPI_Is_host())
       {
          for (i = 0; i < (DESCA[LLD1_]*Ka); i++)
          {
              printf("A[%d]=(%f, %f) ", i, A[i].r, A[i].i);
          }
          printf("\n");
       }
    }

    free(A);
    free(IPIV);

    Cblacs_gridexit(ictxt);

    if (HMPI_Is_host())
    {
       gettimeofday(&aend, NULL);
    }

    if (HMPI_Is_host())
    {
       gettimeofday(&end, NULL);
    }

    /*
     * Print Execution time
     */
    if (HMPI_Is_host())
    {
       double tstart  = start.tv_sec + start.tv_usec/1000000.;
       double tend    = end.tv_sec + end.tv_usec/1000000.;
       double t_sec   = tend - tstart;

       double atstart = astart.tv_sec + astart.tv_usec/1000000.;
       double atend   = aend.tv_sec + aend.tv_usec/1000000.;
       double at_sec  = atend - atstart;

       printf(
          "p=%d, q=%d, M=%d, N=%d, MB=%d, NB=%d, INFO=%d, "
          "Algorithm execution time (sec)=%f, total time(sec)=%f\n",
          p,
          q,
          M,
          N,
          MB,
          NB,
          INFO,
          at_sec,
          t_sec
       );
    }

    HMPI_Finalize(0);
}
